/*(C) 2008-2011 Advanced Micro Devices, Inc. All Rights Reserved.
 *  
 *  This file is part of AMD LibM 3.0.
 *  
 *  AMD LibM 3.0 is free software; you can redistribute it and/or
 *  modify it under the terms of the GNU Lesser General Public
 *  License as published by the Free Software Foundation; either
 *  version 2.1 of the License, or (at your option) any later version.
 *  
 *  AMD LibM 3.0 is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 *  Lesser General Public License for more details.
 *  
 *  You should have received a copy of the GNU Lesser General Public
 *  License along with AMD LibM 3.0.  If not, see
 *  <http://www.gnu.org/licenses/>.
 */
 
 
#include "libm_amd.h"
#include "libm_util_amd.h"
#include "libm_errno_amd.h"
#include "libm_special.h"


float FN_PROTOTYPE(nearbyintf)(float x)
{
    /* Check for input range */
	unsigned int mxscr_stat;
    UT32 checkbits,sign,val_2p23;
    checkbits.f32=x;

    /* Clear the sign bit and check if the value can be rounded(i.e check if exponent less than 23) */
    if( (checkbits.u32 & 0x7FFFFFFF) > 0x4B000000)
    {
      /* take care of nan or inf */
	if(	(checkbits.u32 & EXPBITS_SP32) == EXPBITS_SP32)
	{
        if((checkbits.u32 & MANTBITS_SP32) == 0x0)
        {
            // x is Inf	
#ifdef WINDOWS
            return  _amd_handle_errorf("nearbyintf", _AMD_FPCodeNearbyint, checkbits.u32, _DOMAIN, AMD_F_INVALID, EDOM, x, 0.0, 1);
#else
            return  _amd_handle_errorf("nearbyintf", _AMD_FPCodeNearbyint, checkbits.u32, _DOMAIN, AMD_F_NONE, EDOM, x, 0.0, 1);
#endif
		}
		else 
		{
			// x is NaN
#ifdef WINDOWS
			return  _amd_handle_errorf("nearbyintf", _AMD_FPCodeNearbyint, checkbits.u32 | QNAN_MASK_32 , _DOMAIN, AMD_F_NONE, EDOM, x, 0.0, 1);
#else
			if (checkbits.u32 & QNAN_MASK_32)
			return  _amd_handle_errorf("nearbyintf", _AMD_FPCodeNearbyint, checkbits.u32 | QNAN_MASK_32 , _DOMAIN, AMD_F_NONE, EDOM, x, 0.0, 1);
			else
			return  _amd_handle_errorf("nearbyintf", _AMD_FPCodeNearbyint, checkbits.u32 | QNAN_MASK_32 , _DOMAIN, AMD_F_INVALID, EDOM, x, 0.0, 1);
#endif
		}
	}
	else
		return x;
    }

    sign.u32 =  checkbits.u32 & 0x80000000;   
    val_2p23.u32 = sign.u32 | 0x4B000000;
	 
	mxscr_stat = _mm_getcsr();
	_mm_setcsr(mxscr_stat | 0x00000800);
    val_2p23.f32 = (x + val_2p23.f32) - val_2p23.f32;
	mxscr_stat = mxscr_stat & 0xffffffc0; // clear the exceptions
	_mm_setcsr(mxscr_stat);
    /*This extra line is to take care of denormals and various rounding modes*/
    val_2p23.u32 = ((val_2p23.u32 << 1) >> 1) | sign.u32;

    return (val_2p23.f32);   
}
weak_alias (FN_PROTOTYPE(nearbyintf), nearbyintf)

