/* Mapping tables for Big5-HKSCS handling.
   Copyright (C) 1997, 1998, 2000, 2001, 2002 Free Software Foundation, Inc.
   This file is part of the GNU C Library.
   Contributed by Ulrich Drepper <drepper@cygnus.com>, 1997.
   Modified for Big5-HKSCS by Roger So <roger.so@sw-linux.com>, 2000.
   Updated for HKSCS-2001 by James Su <suzhe@turbolinux.com.cn>
                         and Anthony Fok <anthony@thizlinux.com>, 2002

   The GNU C Library is free software; you can redistribute it and/or
   modify it under the terms of the GNU Lesser General Public
   License as published by the Free Software Foundation; either
   version 2.1 of the License, or (at your option) any later version.

   The GNU C Library is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
   Lesser General Public License for more details.

   You should have received a copy of the GNU Lesser General Public
   License along with the GNU C Library; if not, write to the Free
   Software Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA
   02111-1307 USA.  */

#include <dlfcn.h>
#include <gconv.h>
#include <stdint.h>
#include <stdlib.h>
#include <string.h>
#include <wchar.h>


/* Table for Big5-HKSCS to UCS conversion.

   Original comments by Roger So when he updated the tables for HKSCS-1999:

     With HKSCS mappings 0x8140-0xA0FE and 0xFA40-0xFEFE added; more info:
     http://www.digital21.gov.hk/eng/hkscs/index.html
       - spacehunt 07/01/2000

   The BIG5-HKSCS mapping tables are generated from 950.txt, big5-iso.txt
   and big5cmp.txt using a Perl script while merging C source code from
   other developers.  A copy of the source Perl script is available at:

      http://www.thizlinux.com/~anthony/hkscs/gen-glibc-big5hkscs.pl
      http://people.debian.org/~foka/hkscs/gen-glibc-big5hkscs.pl

  Revisions:
    2001-10-30  made codec for Qt
    2002-03-21  ported to glibc-2.2.5 and added HKSCS-2001

  Todo:
    Use a hash for characters beyond BMP to save space and make it
    more efficient

   - Anthony Fok <anthony@thizlinux.com>  21 Mar 2002
     On behalf of ThizLinux Laboratory Ltd., Hong Kong SAR, China
*/

static const uint16_t big5_hkscs_to_ucs[18683] =
{
	/* Big5-HKSCS 0x8840..0x887E, 0x88A1..0x88FE */
	0xF303,	0xF304,	0xF305,	0xF306,	0xF307,	0xF308,	0xF309,	0xF30A,
	0xF30B,	0xF30C,	0xF30D,	0xF30E,	0xF30F,	0xF310,	0xF311,	0xF312,
	0xF313,	0xF314,	0xF315,	0xF316,	0xF317,	0xF318,	0x0100,	0x00C1,
	0x01CD,	0x00C0,	0x0112,	0x00C9,	0x011A,	0x00C8,	0x014C,	0x00D3,
	0x01D1,	0x00D2,	0xF325,	0x1EBE,	0xF327,	0x1EC0,	0x00CA,	0x0101,
	0x00E1,	0x01CE,	0x00E0,	0x0251,	0x0113,	0x00E9,	0x011B,	0x00E8,
	0x012B,	0x00ED,	0x01D0,	0x00EC,	0x014D,	0x00F3,	0x01D2,	0x00F2,
	0x016B,	0x00FA,	0x01D4,	0x00F9,	0x01D6,	0x01D8,	0x01DA,
		0x01DC,	0x00FC,	0xF344,	0x1EBF,	0xF346,	0x1EC1,	0x00EA,
	0x0261,	0xF34A,	0xF34B,	0xF34C,	0xF34D,	0xF34E,	0xF34F,	0xF350,
	0xF351,	0xF352,	0xF353,	0xF354,	0xF355,	0xF356,	0xF357,	0xF358,
	0xF359,	0xF35A,	0xF35B,	0xF35C,	0xF35D,	0xF35E,	0xF35F,	0xF360,
	0xF361,	0xF362,	0xF363,	0xF364,	0xF365,	0xF366,	0xF367,	0xF368,
	0xF369,	0xF36A,	0xF36B,	0xF36C,	0xF36D,	0xF36E,	0xF36F,	0xF370,
	0xF371,	0xF372,	0xF373,	0xF374,	0xF375,	0xF376,	0xF377,	0xF378,
	0xF379,	0xF37A,	0xF37B,	0xF37C,	0xF37D,	0xF37E,	0xF37F,	0xF380,
	0xF381,	0xF382,	0xF383,	0xF384,	0xF385,	0xF386,	0xF387,	0xF388,
	0xF389,	0xF38A,	0xF38B,	0xF38C,	0xF38D,	0xF38E,	0xF38F,	0xF390,
	0xF391,	0xF392,	0xF393,	0xF394,	0xF395,	0xF396,	0xF397,	0xF398,
	0xF399,	0xF39A,	0xF39B,	0xF39C,	0xF39D,	0xF39E,	0xF39F,

	/* Big5-HKSCS 0x8940..0x897E, 0x89A1..0x89FE */
	0xF3A0,	0xF3A1,	0xF3A2,	0x650A,	0xF3A4,	0xF3A5,	0x4E3D,	0x6EDD,
	0x9D4E,	0x91DF,	0xF3AA,	0xF3AB,	0xF3AC,	0x6491,	0x4F1A,	0x4F28,
	0x4FA8,	0x5156,	0x5174,	0x519C,	0x51E4,	0x52A1,	0x52A8,	0x533B,
	0x534E,	0x53D1,	0x53D8,	0x56E2,	0x58F0,	0x5904,	0x5907,	0x5932,
	0x5934,	0x5B66,	0x5B9E,	0x5B9F,	0x5C9A,	0x5E86,	0x603B,	0x6589,
	0x67FE,	0x6804,	0x6865,	0x6D4E,	0x70BC,	0x7535,	0x7EA4,	0x7EAC,
	0x7EBA,	0x7EC7,	0x7ECF,	0x7EDF,	0x7F06,	0x7F37,	0x827A,	0x82CF,
	0x836F,	0x89C6,	0x8BBE,	0x8BE2,	0x8F66,	0x8F67,	0x8F6E,
		0x7411,	0x7CFC,	0x7DCD,	0x6946,	0x7AC9,	0x5227,	0xF3E5,
	0xF3E6,	0xF3E7,	0xF3E8,	0x918C,	0x78B8,	0x915E,	0x80BC,	0xF3ED,
	0x8D0B,	0x80F6,	0xF3F0,	0xF3F1,	0xF3F2,	0x809F,	0x9EC7,	0x4CCD,
	0x9DC9,	0x9E0C,	0x4C3E,	0xF3F9,	0xF3FA,	0x9E0A,	0xF3FC,	0x35C1,
	0xF3FE,	0x6E9A,	0x823E,	0x7519,	0xF402,	0x4911,	0x9A6C,	0x9A8F,
	0x9F99,	0x7987,	0xF408,	0xF409,	0xF40A,	0xF40B,	0x4E24,	0x4E81,
	0x4E80,	0x4E87,	0x4EBF,	0x4EEB,	0x4F37,	0x344C,	0x4FBD,	0x3E48,
	0x5003,	0x5088,	0x347D,	0x3493,	0x34A5,	0x5186,	0x5905,	0x51DB,
	0x51FC,	0x5205,	0x4E89,	0x5279,	0x5290,	0x5327,	0x35C7,	0x53A9,
	0x3551,	0x53B0,	0x3553,	0x53C2,	0x5423,	0x356D,	0x3572,	0x3681,
	0x5493,	0x54A3,	0x54B4,	0x54B9,	0x54D0,	0x54EF,	0x5518,	0x5523,
	0x5528,	0x3598,	0x553F,	0x35A5,	0x35BF,	0x55D7,	0x35C5,

	/* Big5-HKSCS 0x8A40..0x8A7E, 0x8AA1..0x8AFE */
	0xF43D,	0x5525,	0xF43F,	0xF440,	0xF441,	0xF442,	0x5590,	0xF444,
	0x39EC,	0xF446,	0x8E46,	0xF448,	0xF449,	0x4053,	0xF44B,	0x777A,
	0xF44D,	0x3A34,	0x47D5,	0xF450,	0xF451,	0xF452,	0x64DD,	0xF454,
	0xF455,	0xF456,	0xF457,	0x648D,	0x8E7E,	0xF45A,	0xF45B,	0xF45C,
	0xF45D,	0xF45E,	0xF45F,	0xF460,	0xF461,	0xF462,	0xF463,	0x47F4,
	0xF465,	0xF466,	0x9AB2,	0x3A67,	0xF469,	0x3FED,	0x3506,	0xF46C,
	0xF46D,	0xF46E,	0xF46F,	0x9D6E,	0x9815,	0xF472,	0x43D9,	0xF474,
	0x64B4,	0x54E3,	0xF477,	0xF478,	0xF479,	0x39FB,	0xF47B,
		0xF47C,	0xF47D,	0xF47E,	0x64EA,	0xF480,	0xF481,	0x8E68,
	0xF483,	0xF484,	0xF485,	0xF486,	0x480B,	0xF488,	0x3FFA,	0x5873,
	0xF48B,	0xF48C,	0xF48D,	0xF48E,	0xF48F,	0xF490,	0xF491,	0x5579,
	0x40BB,	0x43BA,	0xF495,	0x4AB4,	0xF497,	0xF498,	0x81AA,	0x98F5,
	0xF49B,	0x6379,	0x39FE,	0xF49E,	0x8DC0,	0x56A1,	0x647C,	0x3E43,
	0xF4A3,	0xF4A4,	0xF4A5,	0xF4A6,	0xF4A7,	0xF4A8,	0xF4A9,	0xF4AA,
	0x3992,	0x3A06,	0xF4AD,	0x3578,	0xF4AF,	0xF4B0,	0x5652,	0xF4B2,
	0xF4B3,	0xF4B4,	0x34BC,	0x6C3D,	0xF4B7,	0xF4B8,	0xF4B9,	0xF4BA,
	0xF4BB,	0xF4BC,	0xF4BD,	0xF4BE,	0xF4BF,	0xF4C0,	0xF4C1,	0x7F93,
	0xF4C3,	0xF4C4,	0xF4C5,	0x35FB,	0xF4C7,	0xF4C8,	0xF4C9,	0xF4CA,
	0x3F93,	0xF4CC,	0xF4CD,	0xF4CE,	0xF4CF,	0xF4D0,	0xF4D1,	0xF4D2,
	0xF4D3,	0xF4D4,	0xF4D5,	0x3FF9,	0xF4D7,	0x6432,	0xF4D9,

	/* Big5-HKSCS 0x8B40..0x8B7E, 0x8BA1..0x8BFE */
	0xF4DA,	0xF4DB,	0xF4DC,	0xF4DD,	0xF4DE,	0xF4DF,	0xF4E0,	0x3A18,
	0xF4E2,	0xF4E3,	0xF4E4,	0xF4E5,	0xF4E6,	0xF4E7,	0xF4E8,	0xF4E9,
	0x95AA,	0x54CC,	0x82C4,	0x55B9,	0xF4EE,	0xF4EF,	0x9C26,	0x9AB6,
	0xF4F2,	0xF4F3,	0x7140,	0x816D,	0x80EC,	0x5C1C,	0xF4F8,	0x8134,
	0x3797,	0x535F,	0xF4FC,	0x91B6,	0xF4FE,	0xF4FF,	0xF500,	0xF501,
	0x35DD,	0xF503,	0x3609,	0xF505,	0x56AF,	0xF507,	0xF508,	0xF509,
	0xF50A,	0xF50B,	0xF50C,	0xF50D,	0xF50E,	0xF50F,	0xF510,	0xF511,
	0x5A54,	0xF513,	0xF514,	0xF515,	0xF516,	0x579C,	0xF518,
		0xF519,	0xF51A,	0xF51B,	0xF51C,	0xF51D,	0x3703,	0xF51F,
	0xF520,	0xF521,	0xF522,	0xF523,	0xF524,	0xF525,	0xF526,	0x5899,
	0x5268,	0x361A,	0xF52A,	0x7BB2,	0x5B68,	0x4800,	0x4B2C,	0x9F27,
	0x49E7,	0x9C1F,	0x9B8D,	0xF533,	0xF534,	0x55FB,	0x35F2,	0x5689,
	0x4E28,	0x5902,	0xF53A,	0xF53B,	0x9751,	0xF53D,	0x4E5B,	0x4EBB,
	0x353E,	0x5C23,	0x5F51,	0x5FC4,	0x38FA,	0x624C,	0x6535,	0x6B7A,
	0x6C35,	0x6C3A,	0x706C,	0x722B,	0x4E2C,	0x72AD,	0xF54E,	0x7F52,
	0x793B,	0x7CF9,	0x7F53,	0xF553,	0x34C1,	0xF555,	0xF556,	0x8002,
	0x8080,	0xF559,	0xF55A,	0x535D,	0x8864,	0x89C1,	0xF55E,	0x8BA0,
	0x8D1D,	0x9485,	0x9578,	0x957F,	0x95E8,	0xF565,	0x97E6,	0x9875,
	0x98CE,	0x98DE,	0x9963,	0xF56B,	0x9C7C,	0x9E1F,	0x9EC4,	0x6B6F,
	0xF907,	0x4E37,	0xF572,	0x961D,	0x6237,	0x94A2,	0xF576,

	/* Big5-HKSCS 0x8C40..0x8C7E, 0x8CA1..0x8CFE */
	0x503B,	0x6DFE,	0xF579,	0xF57A,	0x3DC9,	0x888F,	0xF57D,	0x7077,
	0x5CF5,	0x4B20,	0xF581,	0x3559,	0xF583,	0x6122,	0xF585,	0x8FA7,
	0x91F6,	0x7191,	0x6719,	0x73BA,	0xF58B,	0xF58C,	0x3C8B,	0xF58E,
	0x4B10,	0x78E4,	0x7402,	0x51AE,	0xF593,	0x4009,	0x6A63,	0xF596,
	0x4223,	0x860F,	0xF599,	0x7A2A,	0xF59B,	0xF59C,	0x9755,	0x704D,
	0x5324,	0xF5A0,	0x93F4,	0x76D9,	0xF5A3,	0xF5A4,	0x77DD,	0x4EA3,
	0x4FF0,	0x50BC,	0x4E2F,	0x4F17,	0xF5AB,	0x5434,	0x7D8B,	0x5892,
	0x58D0,	0xF5B0,	0x5E92,	0x5E99,	0x5FC2,	0xF5B4,	0x658B,
		0xF5B6,	0x6919,	0x6A43,	0xF5B9,	0x6CFF,	0xF5BB,	0x7200,
	0xF5BD,	0x738C,	0x3EDB,	0xF5C0,	0x5B15,	0x74B9,	0x8B83,	0xF5C4,
	0xF5C5,	0x7A93,	0x7BEC,	0x7CC3,	0x7E6C,	0x82F8,	0x8597,	0xF5CC,
	0x8890,	0xF5CE,	0x8EB9,	0xF5D0,	0x8FCF,	0x855F,	0x99E0,	0x9221,
	0xF5D5,	0xF5D6,	0xF5D7,	0x4071,	0x42A2,	0x5A1A,	0xF5DB,	0xF5DC,
	0xF5DD,	0x9868,	0x676B,	0x4276,	0x573D,	0xF5E2,	0x85D6,	0xF5E4,
	0x82BF,	0xF5E6,	0x4C81,	0xF5E8,	0x5D7B,	0xF5EA,	0xF5EB,	0xF5EC,
	0xF5ED,	0x5B96,	0xF5EF,	0xF5F0,	0x7E5B,	0xF5F2,	0xF5F3,	0xF5F4,
	0xF5F5,	0xF5F6,	0xF5F7,	0xF5F8,	0xF5F9,	0xF5FA,	0xF5FB,	0xF5FC,
	0xF5FD,	0xF5FE,	0xF5FF,	0xF600,	0xF601,	0xF602,	0xF603,	0xF604,
	0xF605,	0xF606,	0xF607,	0xF608,	0xF609,	0xF60A,	0xF60B,	0xF60C,
	0xF60D,	0xF60E,	0xF60F,	0xF610,	0xF611,	0xF612,	0xF613,

	/* Big5-HKSCS 0x8D40..0x8D7E, 0x8DA1..0x8DFE */
	0xF614,	0xF615,	0xF616,	0xF617,	0xF618,	0xF619,	0xF61A,	0xF61B,
	0xF61C,	0xF61D,	0xF61E,	0xF61F,	0xF620,	0xF621,	0xF622,	0xF623,
	0xF624,	0xF625,	0xF626,	0xF627,	0xF628,	0xF629,	0xF62A,	0xF62B,
	0xF62C,	0xF62D,	0xF62E,	0xF62F,	0xF630,	0xF631,	0xF632,	0xF633,
	0x5D3E,	0x5D48,	0x5D56,	0x3DFC,	0x380F,	0x5DA4,	0x5DB9,	0x3820,
	0x3838,	0x5E42,	0x5EBD,	0x5F25,	0x5F83,	0x3908,	0x3914,	0x393F,
	0x394D,	0x60D7,	0x613D,	0x5CE5,	0x3989,	0x61B7,	0x61B9,	0x61CF,
	0x39B8,	0x622C,	0x6290,	0x62E5,	0x6318,	0x39F8,	0x56B1,
		0x3A03,	0x63E2,	0x63FB,	0x6407,	0x645A,	0x3A4B,	0x64C0,
	0x5D15,	0x5621,	0x9F9F,	0x3A97,	0x6586,	0x3ABD,	0x65FF,	0x6653,
	0x3AF2,	0x6692,	0x3B22,	0x6716,	0x3B42,	0x67A4,	0x6800,	0x3B58,
	0x684A,	0x6884,	0x3B72,	0x3B71,	0x3B7B,	0x6909,	0x6943,	0x725C,
	0x6964,	0x699F,	0x6985,	0x3BBC,	0x69D6,	0x3BDD,	0x6A65,	0x6A74,
	0x6A71,	0x6A82,	0x3BEC,	0x6A99,	0x3BF2,	0x6AAB,	0x6AB5,	0x6AD4,
	0x6AF6,	0x6B81,	0x6BC1,	0x6BEA,	0x6C75,	0x6CAA,	0x3CCB,	0x6D02,
	0x6D06,	0x6D26,	0x6D81,	0x3CEF,	0x6DA4,	0x6DB1,	0x6E15,	0x6E18,
	0x6E29,	0x6E86,	0xF694,	0x6EBB,	0x6EE2,	0x6EDA,	0x9F7F,	0x6EE8,
	0x6EE9,	0x6F24,	0x6F34,	0x3D46,	0xF69E,	0x6F81,	0x6FBE,	0x3D6A,
	0x3D75,	0x71B7,	0x5C99,	0x3D8A,	0x702C,	0x3D91,	0x7050,	0x7054,
	0x706F,	0x707F,	0x7089,	0xF6AD,	0x43C1,	0x35F1,	0xF6B0,

	/* Big5-HKSCS 0x8E40..0x8E7E, 0x8EA1..0x8EFE */
	0xE311,	0x57BE,	0xE313,	0x713E,	0xE315,	0x364E,	0x69A2,	0xE318,
	0x5B74,	0x7A49,	0xE31B,	0xE31C,	0x7A65,	0x7A7D,	0xE31F,	0x7ABB,
	0x7AB0,	0x7AC2,	0x7AC3,	0x71D1,	0xE325,	0x41CA,	0x7ADA,	0x7ADD,
	0x7AEA,	0x41EF,	0x54B2,	0xE32C,	0x7B0B,	0x7B55,	0x7B29,	0xE330,
	0xE331,	0x7BA2,	0x7B6F,	0x839C,	0xE335,	0xE336,	0x7BD0,	0x8421,
	0x7B92,	0x7BB8,	0xE33B,	0x3DAD,	0xE33D,	0x8492,	0x7BFA,	0x7C06,
	0x7C35,	0xE342,	0x7C44,	0x7C83,	0xE345,	0x7CA6,	0x667D,	0xE348,
	0x7CC9,	0x7CC7,	0x7CE6,	0x7C74,	0x7CF3,	0x7CF5,	0x7CCE,
		0x7E67,	0x451D,	0xE352,	0x7D5D,	0xE354,	0x748D,	0x7D89,
	0x7DAB,	0x7135,	0x7DB3,	0x7DD2,	0xE35B,	0xE35C,	0x7DE4,	0x3D13,
	0x7DF5,	0xE360,	0x7DE5,	0xE362,	0x7E1D,	0xE364,	0xE365,	0x7E6E,
	0x7E92,	0x432B,	0x946C,	0x7E27,	0x7F40,	0x7F41,	0x7F47,	0x7936,
	0xE36F,	0x99E1,	0x7F97,	0xE372,	0x7FA3,	0xE374,	0xE375,	0x455C,
	0xE377,	0x4503,	0xE379,	0x7FFA,	0xE37B,	0x8005,	0x8008,	0x801D,
	0x8028,	0x802F,	0xE381,	0xE382,	0x803B,	0x803C,	0x8061,	0xE386,
	0x4989,	0xE388,	0xE389,	0xE38A,	0x6725,	0x80A7,	0xE38D,	0x8107,
	0x811A,	0x58B0,	0xE391,	0x6C7F,	0xE393,	0xE394,	0x64E7,	0xE396,
	0x8218,	0xE398,	0x6A53,	0xE39A,	0xE39B,	0x447A,	0x8229,	0xE39E,
	0xE39F,	0xE3A0,	0x4FF9,	0xE3A2,	0x84E2,	0x8362,	0xE3A5,	0xE3A6,
	0xE3A7,	0xE3A8,	0xE3A9,	0x82AA,	0x691B,	0xE3AC,	0x41DB,

	/* Big5-HKSCS 0x8F40..0x8F7E, 0x8FA1..0x8FFE */
	0x854B,	0x82D0,	0x831A,	0xE3B1,	0xE3B2,	0x36C1,	0xE3B4,	0xE3B5,
	0x827B,	0x82E2,	0x8318,	0xE3B9,	0xE3BA,	0xE3BB,	0xE3BC,	0xE3BD,
	0x3DBF,	0x831D,	0x55EC,	0x8385,	0x450B,	0xE3C3,	0x83AC,	0x83C1,
	0x83D3,	0x347E,	0xE3C8,	0x6A57,	0x855A,	0x3496,	0xE3CC,	0xE3CD,
	0x8458,	0xE3CF,	0x8471,	0x3DD3,	0x44E4,	0x6AA7,	0x844A,	0xE3D5,
	0x7958,	0x84A8,	0xE3D8,	0xE3D9,	0xE3DA,	0x84DE,	0x840F,	0x8391,
	0x44A0,	0x8493,	0x84E4,	0xE3E1,	0x4240,	0xE3E3,	0x4543,	0x8534,
	0x5AF2,	0xE3E7,	0x4527,	0x8573,	0x4516,	0x67BF,	0x8616,
		0xE3ED,	0xE3EE,	0x85C1,	0xE3F0,	0x8602,	0xE3F2,	0xE3F3,
	0xE3F4,	0x456A,	0x8628,	0x3648,	0xE3F8,	0x53F7,	0xE3FA,	0x867E,
	0x8771,	0xE3FD,	0x87EE,	0xE3FF,	0x87B1,	0x87DA,	0x880F,	0x5661,
	0x866C,	0x6856,	0x460F,	0x8845,	0x8846,	0xE409,	0xE40A,	0xE40B,
	0x885E,	0x889C,	0x465B,	0x88B4,	0x88B5,	0x63C1,	0x88C5,	0x7777,
	0xE414,	0x8987,	0x898A,	0x89A6,	0x89A9,	0x89A7,	0x89BC,	0xE41B,
	0x89E7,	0xE41D,	0xE41E,	0x8A9C,	0x7793,	0x91FE,	0x8A90,	0xE423,
	0x7AE9,	0xE425,	0xE426,	0x4713,	0xE428,	0x717C,	0x8B0C,	0x8B1F,
	0xE42C,	0xE42D,	0x8B3F,	0x8B4C,	0x8B4D,	0x8AA9,	0xE432,	0x8B90,
	0x8B9B,	0x8AAF,	0xE436,	0x4615,	0x884F,	0x8C9B,	0xE43A,	0xE43B,
	0xE43C,	0x3725,	0xE43E,	0x8CD6,	0xE440,	0xE441,	0x8D12,	0x8D03,
	0xE444,	0x8CDB,	0x705C,	0x8D11,	0xE448,	0x3ED0,	0x8D77,

	/* Big5-HKSCS 0x9040..0x907E, 0x90A1..0x90FE */
	0x8DA9,	0xE44C,	0xE44D,	0xE44E,	0x3B7C,	0xE450,	0xE451,	0x7AE7,
	0x8EAD,	0x8EB6,	0x8EC3,	0x92D4,	0x8F19,	0x8F2D,	0xE459,	0xE45A,
	0x8FA5,	0x9303,	0xE45D,	0xE45E,	0x8FB3,	0x492A,	0xE461,	0xE462,
	0xE463,	0x5EF8,	0xE465,	0x8FF9,	0xE467,	0xE468,	0xE469,	0xE46A,
	0x3980,	0xE46C,	0x9037,	0xE46E,	0xE46F,	0x9061,	0xE471,	0xE472,
	0x90A8,	0xE474,	0x90C4,	0xE476,	0x90AE,	0x90FD,	0x9167,	0x3AF0,
	0x91A9,	0x91C4,	0x7CAC,	0xE47E,	0xE47F,	0x920E,	0x6C9F,	0x9241,
	0x9262,	0xE484,	0x92B9,	0xE486,	0xE487,	0xE488,	0xE489,
		0xE48A,	0x932C,	0x936B,	0xE48D,	0xE48E,	0x708F,	0x5AC3,
	0xE491,	0xE492,	0x4965,	0x9244,	0xE495,	0xE496,	0xE497,	0x9373,
	0x945B,	0x8EBC,	0x9585,	0x95A6,	0x9426,	0x95A0,	0x6FF6,	0x42B9,
	0xE4A1,	0xE4A2,	0xE4A3,	0xE4A4,	0x49DF,	0x6C1C,	0x967B,	0x9696,
	0x416C,	0x96A3,	0xE4AB,	0x61DA,	0x96B6,	0x78F5,	0xE4AF,	0x96BD,
	0x53CC,	0x49A1,	0xE4B3,	0xE4B4,	0xE4B5,	0xE4B6,	0xE4B7,	0xE4B8,
	0xE4B9,	0xE4BA,	0x9731,	0x8642,	0x9736,	0x4A0F,	0x453D,	0x4585,
	0xE4C1,	0x7075,	0x5B41,	0x971B,	0x975C,	0xE4C6,	0x9757,	0x5B4A,
	0xE4C9,	0x975F,	0x9425,	0x50D0,	0xE4CD,	0xE4CE,	0x9789,	0x979F,
	0x97B1,	0x97BE,	0x97C0,	0x97D2,	0x97E0,	0xE4D6,	0x97EE,	0x741C,
	0xE4D9,	0x97FF,	0x97F5,	0xE4DC,	0xE4DD,	0x4AD1,	0x9834,	0x9833,
	0x984B,	0x9866,	0x3B0E,	0xE4E4,	0x3D51,	0xE4E6,	0xE4E7,

	/* Big5-HKSCS 0x9140..0x917E, 0x91A1..0x91FE */
	0xE4E8,	0x98CA,	0x98B7,	0x98C8,	0x98C7,	0x4AFF,	0xE4EE,	0xE4EF,
	0x55B0,	0x98E1,	0x98E6,	0x98EC,	0x9378,	0x9939,	0xE4F6,	0x4B72,
	0xE4F8,	0xE4F9,	0x99F5,	0x9A0C,	0x9A3B,	0x9A10,	0x9A58,	0xE4FF,
	0x36C4,	0xE501,	0xE502,	0x9AE0,	0x9AE2,	0xE505,	0x9AF4,	0x4C0E,
	0x9B14,	0x9B2D,	0xE50A,	0x5034,	0x9B34,	0xE50D,	0x38C3,	0xE50F,
	0x9B50,	0x9B40,	0xE512,	0x5A45,	0xE514,	0x9B8E,	0xE516,	0x9C02,
	0x9BFF,	0x9C0C,	0xE51A,	0x9DD4,	0xE51C,	0xE51D,	0xE51E,	0xE51F,
	0xE520,	0xE521,	0x9D7E,	0x9D83,	0xE524,	0x9E0E,	0x6888,
		0x9DC4,	0xE528,	0xE529,	0xE52A,	0xE52B,	0xE52C,	0x9D39,
	0xE52E,	0xE52F,	0x9E90,	0x9E95,	0x9E9E,	0x9EA2,	0x4D34,	0x9EAA,
	0x9EAF,	0xE537,	0x9EC1,	0x3B60,	0x39E5,	0x3D1D,	0x4F32,	0x37BE,
	0xE53E,	0x9F02,	0x9F08,	0x4B96,	0x9424,	0xE543,	0x9F17,	0x9F16,
	0x9F39,	0x569F,	0x568A,	0x9F45,	0x99B8,	0xE54B,	0x97F2,	0x847F,
	0x9F62,	0x9F69,	0x7ADC,	0x9F8E,	0x7216,	0x4BBE,	0xE554,	0xE555,
	0x7177,	0xE557,	0xE558,	0xE559,	0x739E,	0xE55B,	0xE55C,	0x799F,
	0xE55E,	0xE55F,	0x9369,	0x93F3,	0xE562,	0x92EC,	0x9381,	0x93CB,
	0xE566,	0xE567,	0x7217,	0x3EEB,	0x7772,	0x7A43,	0x70D0,	0xE56D,
	0xE56E,	0x717E,	0xE570,	0x70A3,	0xE572,	0xE573,	0x3EC7,	0xE575,
	0xE576,	0xE577,	0x3722,	0xE579,	0xE57A,	0x36E1,	0xE57C,	0xE57D,
	0xE57E,	0x3723,	0xE580,	0x575B,	0xE582,	0xE583,	0xE584,

	/* Big5-HKSCS 0x9240..0x927E, 0x92A1..0x92FE */
	0xE585,	0xE586,	0x8503,	0xE588,	0x8503,	0x8455,	0xE58B,	0xE58C,
	0xE58D,	0xE58E,	0xE58F,	0xE590,	0x44F4,	0xE592,	0xE593,	0xE594,
	0x67F9,	0x3733,	0x3C15,	0x3DE7,	0x586C,	0xE59A,	0x6810,	0x4057,
	0xE59D,	0xE59E,	0xE59F,	0xE5A0,	0xE5A1,	0x54CB,	0x569E,	0xE5A4,
	0x5692,	0xE5A6,	0xE5A7,	0xE5A8,	0x93C6,	0xE5AA,	0x939C,	0x4EF8,
	0x512B,	0x3819,	0xE5AF,	0x4EBC,	0xE5B1,	0xE5B2,	0x4F4B,	0x4F8A,
	0xE5B5,	0x5A68,	0xE5B7,	0xE5B8,	0x3999,	0xE5BA,	0xE5BB,	0x3435,
	0x4F29,	0xE5BE,	0xE5BF,	0xE5C0,	0x8ADA,	0xE5C2,	0x4E98,
		0x50CD,	0x510D,	0x4FA2,	0x4F03,	0xE5C8,	0xE5C9,	0x4F42,
	0x502E,	0x506C,	0x5081,	0x4FCC,	0x4FE5,	0x5058,	0x50FC,	0x5159,
	0x515B,	0x515D,	0x515E,	0x6E76,	0xE5D7,	0xE5D8,	0xE5D9,	0x6D72,
	0xE5DB,	0xE5DC,	0x51A8,	0x51C3,	0xE5DF,	0x44DD,	0xE5E1,	0xE5E2,
	0xE5E3,	0x8D7A,	0xE5E5,	0xE5E6,	0x5259,	0x52A4,	0xE5E9,	0x52E1,
	0x936E,	0x467A,	0x718C,	0xE5EE,	0xE5EF,	0xE5F0,	0xE5F1,	0x69D1,
	0xE5F3,	0x7479,	0x3EDE,	0x7499,	0x7414,	0x7456,	0x7398,	0x4B8E,
	0xE5FB,	0xE5FC,	0x53D0,	0x3584,	0x720F,	0xE600,	0x55B4,	0xE602,
	0x54CD,	0xE604,	0x571D,	0x925D,	0x96F4,	0x9366,	0x57DD,	0x578D,
	0x577F,	0x363E,	0x58CB,	0x5A99,	0xE60F,	0xE610,	0xE611,	0xE612,
	0x5A2C,	0x59B8,	0x928F,	0x5A7E,	0x5ACF,	0x5A12,	0xE619,	0xE61A,
	0xE61B,	0xE61C,	0x36F5,	0x6D05,	0x7443,	0x5A21,	0xE621,

	/* Big5-HKSCS 0x9340..0x937E, 0x93A1..0x93FE */
	0x5A81,	0xE623,	0xE624,	0x93E0,	0x748C,	0xE627,	0x7105,	0x4972,
	0x9408,	0xE62B,	0x93BD,	0x37A0,	0x5C1E,	0x5C9E,	0x5E5E,	0x5E48,
	0xE632,	0xE633,	0xE634,	0x5ECD,	0x5B4F,	0xE637,	0xE638,	0x3701,
	0xE63A,	0x36DD,	0xE63C,	0x36D3,	0x812A,	0xE63F,	0xE640,	0xE641,
	0xE642,	0x5F0C,	0x5F0E,	0xE645,	0xE646,	0x5A6B,	0xE648,	0x5B44,
	0x8614,	0xE64B,	0x8860,	0x607E,	0xE64E,	0xE64F,	0x5FDB,	0x3EB8,
	0xE652,	0xE653,	0xE654,	0xE655,	0x61C0,	0xE657,	0xE658,	0xE659,
	0x6199,	0x6198,	0x6075,	0xE65D,	0xE65E,	0xE65F,	0xE660,
		0x6471,	0xE662,	0xE663,	0x3A29,	0xE665,	0xE666,	0xE667,
	0xE668,	0x6337,	0xE66A,	0x64B6,	0x6331,	0x63D1,	0xE66E,	0xE66F,
	0x62A4,	0xE671,	0x643B,	0x656B,	0x6972,	0x3BF4,	0xE676,	0xE677,
	0xE678,	0xE679,	0x550D,	0xE67B,	0xE67C,	0xE67D,	0x66CE,	0xE67F,
	0xE680,	0x3AE0,	0x4190,	0xE683,	0xE684,	0xE685,	0xE686,	0xE687,
	0xE688,	0x78EE,	0xE68A,	0xE68B,	0xE68C,	0x3464,	0xE68E,	0xE68F,
	0xE690,	0x668E,	0xE692,	0x666B,	0x4B93,	0x6630,	0xE696,	0xE697,
	0x6663,	0xE699,	0xE69A,	0x661E,	0xE69C,	0x38D1,	0xE69E,	0xE69F,
	0x3B99,	0xE6A1,	0xE6A2,	0x74D0,	0x3B96,	0x678F,	0xE6A6,	0x68B6,
	0x681E,	0x3BC4,	0x6ABE,	0x3863,	0xE6AC,	0xE6AD,	0x6A33,	0x6A52,
	0x6AC9,	0x6B05,	0xE6B2,	0x6511,	0x6898,	0x6A4C,	0x3BD7,	0x6A7A,
	0x6B57,	0xE6B9,	0xE6BA,	0x93A0,	0x92F2,	0xE6BD,	0xE6BE,

	/* Big5-HKSCS 0x9440..0x947E, 0x94A1..0x94FE */
	0x9289,	0xE6C0,	0xE6C1,	0x9467,	0x6DA5,	0x6F0B,	0xE6C5,	0x6D67,
	0xE6C7,	0x3D8F,	0x6E04,	0xE6CA,	0x5A3D,	0x6E0A,	0x5847,	0x6D24,
	0x7842,	0x713B,	0xE6D1,	0xE6D2,	0x70F1,	0x7250,	0x7287,	0x7294,
	0xE6D7,	0xE6D8,	0x5179,	0xE6DA,	0xE6DB,	0x747A,	0xE6DD,	0xE6DE,
	0xE6DF,	0xE6E0,	0xE6E1,	0x3F06,	0x3EB1,	0xE6E4,	0xE6E5,	0xE6E6,
	0x60A7,	0x3EF3,	0x74CC,	0x743C,	0x9387,	0x7437,	0x449F,	0xE6EE,
	0x4551,	0x7583,	0x3F63,	0xE6F2,	0xE6F3,	0x3F58,	0x7555,	0x7673,
	0xE6F7,	0x3B19,	0x7468,	0xE6FA,	0xE6FB,	0xE6FC,	0x3AFB,
		0x3DCD,	0xE6FF,	0x3EFF,	0xE701,	0xE702,	0x91FA,	0x5732,
	0x9342,	0xE706,	0xE707,	0x50DF,	0xE709,	0xE70A,	0x7778,	0xE70C,
	0x770E,	0x770F,	0x777B,	0xE710,	0xE711,	0x3A5E,	0xE713,	0x7438,
	0x749B,	0x3EBF,	0xE717,	0xE718,	0x40C8,	0xE71A,	0xE71B,	0x9307,
	0xE71D,	0x781E,	0x788D,	0x7888,	0x78D2,	0x73D0,	0x7959,	0xE724,
	0xE725,	0x410E,	0x799B,	0x8496,	0x79A5,	0x6A2D,	0xE72B,	0x7A3A,
	0x79F4,	0x416E,	0xE72F,	0x4132,	0x9235,	0x79F1,	0xE733,	0xE734,
	0xE735,	0xE736,	0xE737,	0x3597,	0x556B,	0x3570,	0x36AA,	0xE73C,
	0xE73D,	0x7AE2,	0x5A59,	0xE740,	0xE741,	0xE742,	0x5A0D,	0xE744,
	0x78F0,	0x5A2A,	0xE747,	0x7AFE,	0x41F9,	0x7C5D,	0x7C6D,	0x4211,
	0xE74D,	0xE74E,	0xE74F,	0x7CCD,	0xE751,	0xE752,	0x7C8E,	0x7C7C,
	0x7CAE,	0x6AB2,	0x7DDC,	0x7E07,	0x7DD3,	0x7F4E,	0xE75B,

	/* Big5-HKSCS 0x9540..0x957E, 0x95A1..0x95FE */
	0xE75C,	0xE75D,	0x7D97,	0xE75F,	0x426A,	0xE761,	0xE762,	0x67D6,
	0xE764,	0xE765,	0x57C4,	0xE767,	0xE768,	0xE769,	0x7FDD,	0x7B27,
	0xE76C,	0xE76D,	0xE76E,	0x7B0C,	0xE770,	0x99E6,	0x8645,	0x9A63,
	0x6A1C,	0xE775,	0x39E2,	0xE777,	0xE778,	0x9A1F,	0xE77A,	0x8480,
	0xE77C,	0xE77D,	0x44EA,	0x8137,	0x4402,	0x80C6,	0x8109,	0x8142,
	0xE784,	0x98C3,	0xE786,	0x8262,	0x8265,	0xE789,	0x8453,	0xE78B,
	0x8610,	0xE78D,	0x5A86,	0x417F,	0xE790,	0x5B2B,	0xE792,	0x5AE4,
	0xE794,	0x86A0,	0xE796,	0xE797,	0x882D,	0xE799,	0x5A02,
		0x886E,	0x4F45,	0x8887,	0x88BF,	0x88E6,	0x8965,	0x894D,
	0xE7A2,	0x8954,	0xE7A4,	0xE7A5,	0xE7A6,	0xE7A7,	0xE7A8,	0xE7A9,
	0x3EAD,	0x84A3,	0x46F5,	0x46CF,	0x37F2,	0x8A3D,	0x8A1C,	0xE7B1,
	0x5F4D,	0x922B,	0xE7B4,	0x65D4,	0x7129,	0x70C4,	0xE7B8,	0x9D6D,
	0x8C9F,	0x8CE9,	0xE7BC,	0x599A,	0x77C3,	0x59F0,	0x436E,	0x36D4,
	0x8E2A,	0x8EA7,	0xE7C4,	0x8F30,	0x8F4A,	0x42F4,	0x6C58,	0x6FBB,
	0xE7CA,	0x489B,	0x6F79,	0x6E8B,	0xE7CE,	0x9BE9,	0x36B5,	0xE7D1,
	0x90BB,	0x9097,	0x5571,	0x4906,	0x91BB,	0x9404,	0xE7D8,	0x4062,
	0xE7DA,	0x9427,	0xE7DC,	0xE7DD,	0x84E5,	0x8A2B,	0x9599,	0x95A7,
	0x9597,	0x9596,	0xE7E4,	0x7445,	0x3EC2,	0xE7E7,	0xE7E8,	0xE7E9,
	0x3EE7,	0xE7EB,	0x968F,	0xE7ED,	0xE7EE,	0xE7EF,	0x3ECC,	0xE7F1,
	0xE7F2,	0xE7F3,	0x7412,	0x746B,	0x3EFC,	0x9741,	0xE7F8,

	/* Big5-HKSCS 0x9640..0x967E, 0x96A1..0x96FE */
	0x6847,	0x4A1D,	0xE7FB,	0xE7FC,	0x975D,	0x9368,	0xE7FF,	0xE800,
	0xE801,	0xE802,	0x92BA,	0x5B11,	0x8B69,	0x493C,	0x73F9,	0xE808,
	0x979B,	0x9771,	0x9938,	0xE80C,	0x5DC1,	0xE80E,	0xE80F,	0x981F,
	0xE811,	0x92F6,	0xE813,	0x91E5,	0x44C0,	0xE816,	0xE817,	0xE818,
	0x98DC,	0xE81A,	0x3F00,	0x922A,	0x4925,	0x8414,	0x993B,	0x994D,
	0xE821,	0x3DFD,	0x999B,	0x4B6F,	0x99AA,	0x9A5C,	0xE827,	0xE828,
	0x6A8F,	0x9A21,	0x5AFE,	0x9A2F,	0xE82D,	0x4B90,	0xE82F,	0x99BC,
	0x4BBD,	0x4B97,	0x937D,	0x5872,	0xE835,	0x5822,	0xE837,
		0xE838,	0x7844,	0xE83A,	0xE83B,	0x68C5,	0x3D7D,	0x9458,
	0x3927,	0x6150,	0xE841,	0xE842,	0x6107,	0x9C4F,	0x9C53,	0x9C7B,
	0x9C35,	0x9C10,	0x9B7F,	0x9BCF,	0xE84B,	0x9B9F,	0xE84D,	0xE84E,
	0x9D21,	0x4CAE,	0xE851,	0x9E18,	0x4CB0,	0x9D0C,	0xE855,	0xE856,
	0xE857,	0xE858,	0x9DA5,	0x84BD,	0xE85B,	0xE85C,	0xE85D,	0x85FC,
	0x4533,	0xE860,	0xE861,	0xE862,	0x8420,	0x85EE,	0xE865,	0xE866,
	0xE867,	0x79E2,	0xE869,	0xE86A,	0x492D,	0xE86C,	0x3D62,	0x93DB,
	0x92BE,	0x9348,	0xE871,	0x78B9,	0x9277,	0x944D,	0x4FE4,	0x3440,
	0x9064,	0xE878,	0x783D,	0x7854,	0x78B6,	0x784B,	0xE87D,	0xE87E,
	0xE87F,	0x369A,	0x4F72,	0x6FDA,	0x6FD9,	0x701E,	0x701E,	0x5414,
	0xE887,	0x57BB,	0x58F3,	0x578A,	0x9D16,	0x57D7,	0x7134,	0x34AF,
	0xE88F,	0x71EB,	0xE891,	0xE892,	0x5B28,	0xE894,	0xE895,

	/* Big5-HKSCS 0x9740..0x977E, 0x97A1..0x97FE */
	0x610C,	0x5ACE,	0x5A0B,	0x42BC,	0xE89A,	0x372C,	0x4B7B,	0xE89D,
	0x93BB,	0x93B8,	0xE8A0,	0xE8A1,	0x8472,	0xE8A3,	0xE8A4,	0xE8A5,
	0xE8A6,	0xE8A7,	0x5994,	0xE8A9,	0xE8AA,	0x7DA8,	0xE8AC,	0xE8AD,
	0xE8AE,	0xE8AF,	0xE8B0,	0x92E5,	0x73E2,	0x3EE9,	0x74B4,	0xE8B5,
	0xE8B6,	0x3EE1,	0xE8B8,	0x6AD8,	0x73F3,	0x73FB,	0x3ED6,	0xE8BD,
	0xE8BE,	0xE8BF,	0xE8C0,	0xE8C1,	0xE8C2,	0xE8C3,	0x7448,	0xE8C5,
	0x70A5,	0xE8C7,	0x9284,	0x73E6,	0x935F,	0xE8CB,	0x9331,	0xE8CD,
	0xE8CE,	0x9386,	0xE8D0,	0xE8D1,	0x4935,	0xE8D3,	0x716B,
		0xE8D5,	0xE8D6,	0x56A4,	0xE8D8,	0xE8D9,	0xE8DA,	0x5502,
	0x79C4,	0xE8DD,	0x7DFE,	0xE8DF,	0xE8E0,	0xE8E1,	0x452E,	0x9401,
	0x370A,	0xE8E5,	0xE8E6,	0x59B0,	0xE8E8,	0xE8E9,	0xE8EA,	0x5AA1,
	0x36E2,	0xE8ED,	0x36B0,	0x925F,	0x5A79,	0xE8F1,	0xE8F2,	0x9374,
	0x3CCD,	0xE8F5,	0x4A96,	0x398A,	0x50F4,	0x3D69,	0x3D4C,	0xE8FB,
	0x7175,	0x42FB,	0xE8FE,	0x6E0F,	0xE900,	0x44EB,	0x6D57,	0xE903,
	0x7067,	0x6CAF,	0x3CD6,	0xE907,	0xE908,	0x6E02,	0x6F0C,	0x3D6F,
	0xE90C,	0x7551,	0x36BC,	0x34C8,	0x4680,	0x3EDA,	0x4871,	0x59C4,
	0x926E,	0x493E,	0x8F41,	0xE917,	0xE918,	0x5812,	0x57C8,	0x36D6,
	0xE91C,	0x70FE,	0xE91E,	0xE91F,	0xE920,	0xE921,	0xE922,	0x68B9,
	0x6967,	0xE925,	0xE926,	0xE927,	0xE928,	0xE929,	0xE92A,	0xE92B,
	0xE92C,	0x6A1A,	0xE92E,	0xE92F,	0x843E,	0x44DF,	0x44CE,

	/* Big5-HKSCS 0x9840..0x987E, 0x98A1..0x98FE */
	0xE933,	0xE934,	0xE935,	0xE936,	0x6F17,	0xE938,	0x833D,	0xE93A,
	0x83ED,	0xE93C,	0xE93D,	0xE93E,	0x5989,	0x5A82,	0xE941,	0x5A61,
	0x5A71,	0xE944,	0xE945,	0x372D,	0x59EF,	0xE948,	0x36C7,	0x718E,
	0x9390,	0x669A,	0xE94D,	0x5A6E,	0x5A2B,	0xE950,	0x6A2B,	0xE952,
	0xE953,	0xE954,	0xE955,	0x711D,	0xE957,	0xE958,	0x4FB0,	0xE95A,
	0x5CC2,	0xE95C,	0xE95D,	0xE95E,	0x6A0C,	0xE960,	0xE961,	0x70A6,
	0x7133,	0xE964,	0x3DA5,	0x6CDF,	0xE967,	0xE968,	0x7E65,	0x59EB,
	0x5D2F,	0x3DF3,	0x5F5C,	0xE96E,	0xE96F,	0x7DA4,	0x8426,
		0x5485,	0xE973,	0xE974,	0xE975,	0x577E,	0xE977,	0xE978,
	0x3FE5,	0xE97A,	0xE97B,	0x7003,	0xE97D,	0x5D70,	0x738F,	0x7CD3,
	0xE981,	0xE982,	0x4FC8,	0x7FE7,	0x72CD,	0x7310,	0xE987,	0x7338,
	0x7339,	0xE98A,	0x7341,	0x7348,	0x3EA9,	0xE98E,	0x906C,	0x71F5,
	0xE991,	0x73E1,	0x81F6,	0x3ECA,	0x770C,	0x3ED1,	0x6CA2,	0x56FD,
	0x7419,	0x741E,	0x741F,	0x3EE2,	0x3EF0,	0x3EF4,	0x3EFA,	0x74D3,
	0x3F0E,	0x3F53,	0x7542,	0x756D,	0x7572,	0x758D,	0x3F7C,	0x75C8,
	0x75DC,	0x3FC0,	0x764D,	0x3FD7,	0x7674,	0x3FDC,	0x767A,	0xE9B0,
	0x7188,	0x5623,	0x8980,	0x5869,	0x401D,	0x7743,	0x4039,	0x6761,
	0x4045,	0x35DB,	0x7798,	0x406A,	0x406F,	0x5C5E,	0x77BE,	0x77CB,
	0x58F2,	0x7818,	0x70B9,	0x781C,	0x40A8,	0x7839,	0x7847,	0x7851,
	0x7866,	0x8448,	0xE9CB,	0x7933,	0x6803,	0x7932,	0x4103,

	/* Big5-HKSCS 0x9940..0x997E, 0x99A1..0x99FE */
	0x4109,	0x7991,	0x7999,	0x8FBB,	0x7A06,	0x8FBC,	0x4167,	0x7A91,
	0x41B2,	0x7ABC,	0x8279,	0x41C4,	0x7ACF,	0x7ADB,	0x41CF,	0x4E21,
	0x7B62,	0x7B6C,	0x7B7B,	0x7C12,	0x7C1B,	0x4260,	0x427A,	0x7C7B,
	0x7C9C,	0x428C,	0x7CB8,	0x4294,	0x7CED,	0x8F93,	0x70C0,	0xE9EF,
	0x7DCF,	0x7DD4,	0x7DD0,	0x7DFD,	0x7FAE,	0x7FB4,	0x729F,	0x4397,
	0x8020,	0x8025,	0x7B39,	0x802E,	0x8031,	0x8054,	0x3DCC,	0x57B4,
	0x70A0,	0x80B7,	0x80E9,	0x43ED,	0x810C,	0x732A,	0x810E,	0x8112,
	0x7560,	0x8114,	0x4401,	0x3B39,	0x8156,	0x8159,	0x815A,
		0x4413,	0x583A,	0x817C,	0x8184,	0x4425,	0x8193,	0x442D,
	0x81A5,	0x57EF,	0x81C1,	0x81E4,	0x8254,	0x448F,	0x82A6,	0x8276,
	0x82CA,	0x82D8,	0x82FF,	0x44B0,	0x8357,	0x9669,	0x698A,	0x8405,
	0x70F5,	0x8464,	0x60E3,	0x8488,	0x4504,	0x84BE,	0x84E1,	0x84F8,
	0x8510,	0x8538,	0x8552,	0x453B,	0x856F,	0x8570,	0x85E0,	0x4577,
	0x8672,	0x8692,	0x86B2,	0x86EF,	0x9645,	0x878B,	0x4606,	0x4617,
	0x88AE,	0x88FF,	0x8924,	0x8947,	0x8991,	0xEA43,	0x8A29,	0x8A38,
	0x8A94,	0x8AB4,	0x8C51,	0x8CD4,	0x8CF2,	0x8D1C,	0x4798,	0x585F,
	0x8DC3,	0x47ED,	0x4EEE,	0x8E3A,	0x55D8,	0x5754,	0x8E71,	0x55F5,
	0x8EB0,	0x4837,	0x8ECE,	0x8EE2,	0x8EE4,	0x8EED,	0x8EF2,	0x8FB7,
	0x8FC1,	0x8FCA,	0x8FCC,	0x9033,	0x99C4,	0x48AD,	0x98E0,	0x9213,
	0x491E,	0x9228,	0x9258,	0x926B,	0x92B1,	0x92AE,	0x92BF,

	/* Big5-HKSCS 0x9A40..0x9A7E, 0x9AA1..0x9AFE */
	0x92E3,	0x92EB,	0x92F3,	0x92F4,	0x92FD,	0x9343,	0x9384,	0x93AD,
	0x4945,	0x4951,	0x9EBF,	0x9417,	0x5301,	0x941D,	0x942D,	0x943E,
	0x496A,	0x9454,	0x9479,	0x952D,	0x95A2,	0x49A7,	0x95F4,	0x9633,
	0x49E5,	0x67A0,	0x4A24,	0x9740,	0x4A35,	0x97B2,	0x97C2,	0x5654,
	0x4AE4,	0x60E8,	0x98B9,	0x4B19,	0x98F1,	0x5844,	0x990E,	0x9919,
	0x51B4,	0x991C,	0x9937,	0x9942,	0x995D,	0x9962,	0x4B70,	0x99C5,
	0x4B9D,	0x9A3C,	0x9B0F,	0x7A83,	0x9B69,	0x9B81,	0x9BDD,	0x9BF1,
	0x9BF4,	0x4C6D,	0x9C20,	0x376F,	0xEAA9,	0x9D49,	0x9C3A,
		0x9EFE,	0x5650,	0x9D93,	0x9DBD,	0x9DC0,	0x9DFC,	0x94F6,
	0x8FB6,	0x9E7B,	0x9EAC,	0x9EB1,	0x9EBD,	0x9EC6,	0x94DC,	0x9EE2,
	0x9EF1,	0x9EF8,	0x7AC8,	0x9F44,	0xEABF,	0xEAC0,	0xEAC1,	0x691A,
	0x94C3,	0x59AC,	0xEAC5,	0x5840,	0x94C1,	0x37B9,	0xEAC9,	0xEACA,
	0xEACB,	0xEACC,	0x5757,	0x7173,	0xEACF,	0xEAD0,	0xEAD1,	0x546A,
	0xEAD3,	0xEAD4,	0x549E,	0xEAD6,	0xEAD7,	0xEAD8,	0xEAD9,	0xEADA,
	0x60E7,	0xEADC,	0x567A,	0xEADE,	0xEADF,	0xEAE0,	0xEAE1,	0xEAE2,
	0xEAE3,	0x6955,	0x9C2F,	0x87A5,	0xEAE7,	0xEAE8,	0xEAE9,	0xEAEA,
	0xEAEB,	0xEAEC,	0x5C20,	0xEAEE,	0x5E0B,	0xEAF0,	0xEAF1,	0xEAF2,
	0x671E,	0xEAF4,	0xEAF5,	0xEAF6,	0x3647,	0xEAF8,	0xEAF9,	0xEAFA,
	0xEAFB,	0x5364,	0x84AD,	0xEAFE,	0xEAFF,	0xEB00,	0x8B81,	0xEB02,
	0xEB03,	0xEB04,	0xEB05,	0x4E78,	0x70BB,	0xEB08,	0xEB09,

	/* Big5-HKSCS 0x9B40..0x9B7E, 0x9BA1..0x9BFE */
	0xEB0A,	0xEB0B,	0xEB0C,	0xEB0D,	0xEB0E,	0xEB0F,	0x62C3,	0xEB11,
	0xEB12,	0x7198,	0x6855,	0xEB15,	0x69E9,	0x36C8,	0xEB18,	0xEB19,
	0xEB1A,	0xEB1B,	0xEB1C,	0xEB1D,	0x82FD,	0xEB1F,	0xEB20,	0xEB21,
	0x89A5,	0xEB23,	0x8FA0,	0xEB25,	0x97B8,	0xEB27,	0x9847,	0x9ABD,
	0xEB2A,	0xEB2B,	0xEB2C,	0xEB2D,	0xEB2E,	0xEB2F,	0xEB30,	0xEB31,
	0xEB32,	0xEB33,	0xEB34,	0xEB35,	0xEB36,	0xEB37,	0xEB38,	0xEB39,
	0x5FB1,	0x6648,	0x66BF,	0xEB3D,	0xEB3E,	0xEB3F,	0x7201,	0xEB41,
	0x77D7,	0xEB43,	0xEB44,	0x7E87,	0xEB46,	0x58B5,	0x670E,
		0x6918,	0xEB4A,	0xEB4B,	0xEB4C,	0xEB4D,	0xEB4E,	0xEB4F,
	0xEB50,	0x48D0,	0x4AB8,	0xEB53,	0xEB54,	0xEB55,	0xEB56,	0xEB57,
	0xEB58,	0xEB59,	0xEB5A,	0xEB5B,	0x51D2,	0xEB5D,	0x599F,	0xEB5F,
	0x3BBE,	0xEB61,	0xEB62,	0xEB63,	0x5788,	0xEB65,	0x399B,	0xEB67,
	0xEB68,	0xEB69,	0x3762,	0xEB6B,	0x8B5E,	0xEB6D,	0x99D6,	0xEB6F,
	0xEB70,	0xEB71,	0x7209,	0xEB73,	0xEB74,	0x5965,	0xEB76,	0xEB77,
	0xEB78,	0x8EDA,	0xEB7A,	0x528F,	0x573F,	0x7171,	0xEB7E,	0xEB7F,
	0xEB80,	0xEB81,	0x55BC,	0xEB83,	0xEB84,	0xEB85,	0x91D4,	0x3473,
	0xEB88,	0xEB89,	0xEB8A,	0x4718,	0xEB8C,	0xEB8D,	0xEB8E,	0xEB8F,
	0xEB90,	0x5066,	0x34FB,	0xEB93,	0x60DE,	0xEB95,	0x477C,	0xEB97,
	0xEB98,	0xEB99,	0xEB9A,	0xEB9B,	0x57A1,	0x7151,	0x6FB6,	0xEB9F,
	0xEBA0,	0x9056,	0xEBA2,	0xEBA3,	0x8B62,	0xEBA5,	0xEBA6,

	/* Big5-HKSCS 0x9C40..0x9C7E, 0x9CA1..0x9CFE */
	0x5D5B,	0xEBA8,	0x8F36,	0xEBAA,	0xEBAB,	0x8AEA,	0xEBAD,	0xEBAE,
	0xEBAF,	0xEBB0,	0x4BC0,	0xEBB2,	0xEBB3,	0xEBB4,	0x9465,	0xEBB6,
	0x6195,	0x5A27,	0xEBB9,	0x4FBB,	0x56B9,	0xEBBC,	0xEBBD,	0x4E6A,
	0xEBBF,	0x9656,	0x6D8F,	0xEBC2,	0x3618,	0x8977,	0xEBC5,	0xEBC6,
	0xEBC7,	0xEBC8,	0x71DF,	0xEBCA,	0x7B42,	0xEBCC,	0xEBCD,	0xEBCE,
	0x9104,	0xEBD0,	0x7A45,	0x9DF0,	0xEBD3,	0x9A26,	0xEBD5,	0x365F,
	0xEBD7,	0xEBD8,	0x7983,	0xEBDA,	0xEBDB,	0x5D2C,	0xEBDD,	0x83CF,
	0xEBDF,	0x46D0,	0xEBE1,	0x753B,	0x8865,	0xEBE4,	0x58B6,
		0x371C,	0xEBE7,	0xEBE8,	0xEBE9,	0x3C54,	0xEBEB,	0xEBEC,
	0x9281,	0xEBEE,	0xEBEF,	0x9330,	0xEBF1,	0xEBF2,	0x6C39,	0x949F,
	0xEBF5,	0xEBF6,	0x8827,	0x88F5,	0xEBF9,	0xEBFA,	0xEBFB,	0x6EB8,
	0xEBFD,	0xEBFE,	0x39A4,	0x36B9,	0x5C10,	0x79E3,	0x453F,	0x66B6,
	0xEC05,	0xEC06,	0x8943,	0xEC08,	0xEC09,	0x56D6,	0x40DF,	0xEC0C,
	0x39A1,	0xEC0E,	0xEC0F,	0xEC10,	0x71AD,	0x8366,	0xEC13,	0xEC14,
	0x5A67,	0x4CB7,	0xEC17,	0xEC18,	0xEC19,	0xEC1A,	0xEC1B,	0xEC1C,
	0xEC1D,	0x7B43,	0x797E,	0xEC20,	0x6FB5,	0xEC22,	0x6A03,	0xEC24,
	0x53A2,	0xEC26,	0x93BF,	0x6836,	0x975D,	0xEC2A,	0xEC2B,	0xEC2C,
	0xEC2D,	0xEC2E,	0xEC2F,	0x5D85,	0xEC31,	0xEC32,	0x5715,	0x9823,
	0xEC35,	0x5DAB,	0xEC37,	0x65BE,	0x69D5,	0x53D2,	0xEC3B,	0xEC3C,
	0x3C11,	0x6736,	0xEC3F,	0xEC40,	0xEC41,	0xEC42,	0xEC43,

	/* Big5-HKSCS 0x9D40..0x9D7E, 0x9DA1..0x9DFE */
	0xEC44,	0xEC45,	0xEC46,	0xEC47,	0xEC48,	0xEC49,	0x35CA,	0xEC4B,
	0xEC4C,	0x48FA,	0x63E6,	0xEC4F,	0x7808,	0x9255,	0xEC52,	0x43F2,
	0xEC54,	0x43DF,	0xEC56,	0xEC57,	0xEC58,	0x59F8,	0xEC5A,	0x8F0B,
	0xEC5C,	0xEC5D,	0x7B51,	0xEC5F,	0xEC60,	0x3DF7,	0xEC62,	0xEC63,
	0x8FD0,	0x728F,	0x568B,	0xEC67,	0xEC68,	0xEC69,	0xEC6A,	0xEC6B,
	0xEC6C,	0xEC6D,	0xEC6E,	0xEC6F,	0xEC70,	0xEC71,	0xEC72,	0xEC73,
	0x7E9F,	0xEC75,	0xEC76,	0x4CA4,	0x9547,	0xEC79,	0x71A2,	0xEC7B,
	0x4D91,	0x9012,	0xEC7E,	0x4D9C,	0xEC80,	0x8FBE,	0x55C1,
		0x8FBA,	0xEC84,	0x8FB9,	0xEC86,	0x4509,	0x7E7F,	0x6F56,
	0x6AB1,	0x4EEA,	0x34E4,	0xEC8D,	0xEC8E,	0x373A,	0x8E80,	0xEC91,
	0xEC92,	0xEC93,	0xEC94,	0xEC95,	0xEC96,	0x3DEB,	0xEC98,	0xEC99,
	0xEC9A,	0xEC9B,	0x4E9A,	0xEC9D,	0xEC9E,	0x56BF,	0xECA0,	0x8E0E,
	0x5B6D,	0xECA3,	0xECA4,	0x63DE,	0x62D0,	0xECA7,	0xECA8,	0x6530,
	0x562D,	0xECAB,	0x541A,	0xECAD,	0x3DC6,	0xECAF,	0x4C7D,	0x5622,
	0x561E,	0x7F49,	0xECB4,	0x5975,	0xECB6,	0x8770,	0x4E1C,	0xECB9,
	0xECBA,	0xECBB,	0x8117,	0x9D5E,	0x8D18,	0x763B,	0x9C45,	0x764E,
	0x77B9,	0x9345,	0x5432,	0x8148,	0x82F7,	0x5625,	0x8132,	0x8418,
	0x80BD,	0x55EA,	0x7962,	0x5643,	0x5416,	0xECCF,	0x35CE,	0x5605,
	0x55F1,	0x66F1,	0xECD4,	0x362D,	0x7534,	0x55F0,	0x55BA,	0x5497,
	0x5572,	0xECDB,	0xECDC,	0x5ED0,	0xECDE,	0xECDF,	0xECE0,

	/* Big5-HKSCS 0x9E40..0x9E7E, 0x9EA1..0x9EFE */
	0xECE1,	0x9EAB,	0x7D5A,	0x55DE,	0xECE5,	0x629D,	0x976D,	0x5494,
	0x8CCD,	0x71F6,	0x9176,	0x63FC,	0x63B9,	0x63FE,	0x5569,	0xECF0,
	0x9C72,	0xECF2,	0x519A,	0x34DF,	0xECF5,	0x51A7,	0x544D,	0x551E,
	0x5513,	0x7666,	0x8E2D,	0xECFC,	0x75B1,	0x80B6,	0x8804,	0x8786,
	0x88C7,	0x81B6,	0x841C,	0xED04,	0x44EC,	0x7304,	0xED07,	0x5B90,
	0x830B,	0xED0A,	0x567B,	0xED0C,	0xED0D,	0xED0E,	0xED0F,	0xED10,
	0xED11,	0x9170,	0xED13,	0x9208,	0xED15,	0xED16,	0xED17,	0xED18,
	0x7266,	0xED1A,	0x474E,	0xED1C,	0xED1D,	0xED1E,	0x40FA,
		0x9C5D,	0x651F,	0xED22,	0x48F3,	0xED24,	0xED25,	0xED26,
	0xED27,	0x6062,	0xED29,	0xED2A,	0xED2B,	0xED2C,	0xED2D,	0x71A3,
	0x7E8E,	0x9D50,	0x4E1A,	0x4E04,	0x3577,	0x5B0D,	0x6CB2,	0x5367,
	0x36AC,	0x39DC,	0x537D,	0x36A5,	0xED3B,	0x589A,	0xED3D,	0x822D,
	0x544B,	0x57AA,	0xED41,	0xED42,	0xED43,	0x3A52,	0xED45,	0x7374,
	0xED47,	0x4D09,	0x9BED,	0xED4A,	0xED4B,	0x4C5B,	0xED4D,	0xED4E,
	0xED4F,	0x845C,	0xED51,	0xED52,	0xED53,	0xED54,	0x632E,	0x7D25,
	0xED57,	0xED58,	0x3A2A,	0x9008,	0x52CC,	0x3E74,	0x367A,	0x45E9,
	0xED5F,	0x7640,	0x5AF0,	0xED62,	0x787A,	0x47B6,	0x58A7,	0x40BF,
	0x567C,	0x9B8B,	0x5D74,	0x7654,	0xED6B,	0x9E85,	0x4CE1,	0x75F9,
	0x37FB,	0x6119,	0xED71,	0xED72,	0xED73,	0x565D,	0xED75,	0x57A7,
	0xED77,	0xED78,	0x5234,	0xED7A,	0x35AD,	0x6C4A,	0x9D7C,

	/* Big5-HKSCS 0x9F40..0x9F7E, 0x9FA1..0x9FFE */
	0x7C56,	0x9B39,	0x57DE,	0xED81,	0x5C53,	0x64D3,	0xED84,	0xED85,
	0xED86,	0x86AD,	0xED88,	0xED89,	0xED8A,	0xED8B,	0xED8C,	0x51FE,
	0xED8E,	0x5D8E,	0x9703,	0xED91,	0x9E81,	0x904C,	0x7B1F,	0x9B02,
	0x5CD1,	0x7BA3,	0x6268,	0x6335,	0x9AFF,	0x7BCF,	0x9B2A,	0x7C7E,
	0x9B2E,	0x7C42,	0x7C86,	0x9C15,	0x7BFC,	0x9B09,	0x9F17,	0x9C1B,
	0xEDA6,	0x9F5A,	0x5573,	0x5BC3,	0x4FFD,	0x9E98,	0x4FF2,	0x5260,
	0x3E06,	0x52D1,	0x5767,	0x5056,	0x59B7,	0x5E12,	0x97C8,	0x9DAB,
	0x8F5C,	0x5469,	0x97B4,	0x9940,	0x97BA,	0x532C,	0x6130,
		0x692C,	0x53DA,	0x9C0A,	0x9D02,	0x4C3B,	0x9641,	0x6980,
	0x50A6,	0x7546,	0xEDC6,	0x99DA,	0x5273,	0xEDC9,	0x9159,	0x9681,
	0x915C,	0xEDCD,	0x9151,	0xEDCF,	0x637F,	0xEDD1,	0x6ACA,	0x5611,
	0x918E,	0x757A,	0x6285,	0xEDD7,	0x734F,	0x7C70,	0xEDDA,	0xEDDB,
	0xEDDC,	0xEDDD,	0x76D6,	0x9B9D,	0x4E2A,	0xEDE1,	0x83BE,	0x8842,
	0xEDE4,	0x5C4A,	0x69C0,	0x50ED,	0x577A,	0x521F,	0x5DF5,	0x4ECE,
	0x6C31,	0xEDED,	0x4F39,	0x549C,	0x54DA,	0x529A,	0x8D82,	0x35FE,
	0x5F0C,	0x35F3,	0xEDF6,	0x6B52,	0x917C,	0x9FA5,	0x9B97,	0x982E,
	0x98B4,	0x9ABA,	0x9EA8,	0x9E84,	0x717A,	0x7B14,	0xEE02,	0x6BFA,
	0x8818,	0x7F78,	0xEE06,	0x5620,	0xEE08,	0x8E77,	0x9F53,	0xEE0B,
	0x8DD4,	0x8E4F,	0x9E1C,	0x8E01,	0x6282,	0xEE11,	0x8E28,	0x8E75,
	0x7AD3,	0xEE15,	0x7A3E,	0x78D8,	0x6CEA,	0x8A67,	0x7607,

	/* Big5-HKSCS 0xA040..0xA07E, 0xA0A1..0xA0FE */
	0xEE1B,	0x9F26,	0x6CCE,	0x87D6,	0x75C3,	0xEE20,	0x7853,	0xEE22,
	0x8D0C,	0x72E2,	0x7371,	0x8B2D,	0x7302,	0x74F1,	0x8CEB,	0xEE2A,
	0x862F,	0x5FBA,	0x88A0,	0x44B7,	0xEE2F,	0xEE30,	0xEE31,	0xEE32,
	0x8A7E,	0xEE34,	0xEE35,	0x60FD,	0x7667,	0x9AD7,	0x9D44,	0x936E,
	0x9B8F,	0x87F5,	0xEE3D,	0x880F,	0x8CF7,	0x732C,	0x9721,	0x9BB0,
	0x35D6,	0x72B2,	0x4C07,	0x7C51,	0x994A,	0xEE48,	0x6159,	0x4C04,
	0x9E96,	0x617D,	0xEE4D,	0x575F,	0x616F,	0x62A6,	0x6239,	0x62CE,
	0x3A5C,	0x61E2,	0x53AA,	0xEE56,	0x6364,	0x6802,	0x35D2,
		0x5D57,	0xEE5B,	0x8FDA,	0xEE5D,	0xEE5E,	0x50D9,	0xEE60,
	0x7906,	0x5332,	0x9638,	0xEE64,	0x4065,	0xEE66,	0x77FE,	0xEE68,
	0x7CC2,	0xEE6A,	0x7CDA,	0x7A2D,	0x8066,	0x8063,	0x7D4D,	0x7505,
	0x74F2,	0x8994,	0x821A,	0x670C,	0x8062,	0xEE76,	0x805B,	0x74F0,
	0x8103,	0x7724,	0x8989,	0xEE7C,	0x7553,	0xEE7E,	0x87A9,	0x87CE,
	0x81C8,	0x878C,	0x8A49,	0x8CAD,	0x8B43,	0x772B,	0x74F8,	0x84DA,
	0x3635,	0x69B2,	0x8DA6,	0xEE8C,	0x89A9,	0x7468,	0x6DB9,	0x87C1,
	0xEE91,	0x74E7,	0x3DDB,	0x7176,	0x60A4,	0x619C,	0x3CD1,	0x7162,
	0x6077,	0xEE9A,	0x7F71,	0xEE9C,	0x7250,	0x60E9,	0x4B7E,	0x5220,
	0x3C18,	0xEEA2,	0xEEA3,	0xEEA4,	0xEEA5,	0xEEA6,	0xEEA7,	0xEEA8,
	0xEEA9,	0xEEAA,	0x5CC1,	0xEEAC,	0xEEAD,	0xEEAE,	0xEEAF,	0xEEB0,
	0xEEB1,	0x4562,	0x5B1F,	0xEEB4,	0x9F50,	0x9EA6,	0xEEB7,

	/* Big5-HKSCS 0xA140..0xA17E, 0xA1A1..0xA1FE */
	0x3000,	0xFF0C,	0x3001,	0x3002,	0xFF0E,	0x2027,	0xFF1B,	0xFF1A,
	0xFF1F,	0xFF01,	0xFE30,	0x2026,	0x2025,	0xFE50,	0xFE51,	0xFE52,
	0x00B7,	0xFE54,	0xFE55,	0xFE56,	0xFE57,	0xFF5C,	0x2013,	0xFE31,
	0x2014,	0xFE33,	0x2574,	0xFE34,	0xFE4F,	0xFF08,	0xFF09,	0xFE35,
	0xFE36,	0xFF5B,	0xFF5D,	0xFE37,	0xFE38,	0x3014,	0x3015,	0xFE39,
	0xFE3A,	0x3010,	0x3011,	0xFE3B,	0xFE3C,	0x300A,	0x300B,	0xFE3D,
	0xFE3E,	0x3008,	0x3009,	0xFE3F,	0xFE40,	0x300C,	0x300D,	0xFE41,
	0xFE42,	0x300E,	0x300F,	0xFE43,	0xFE44,	0xFE59,	0xFE5A,
		0xFE5B,	0xFE5C,	0xFE5D,	0xFE5E,	0x2018,	0x2019,	0x201C,
	0x201D,	0x301D,	0x301E,	0x2035,	0x2032,	0xFF03,	0xFF06,	0xFF0A,
	0x203B,	0x00A7,	0x3003,	0x25CB,	0x25CF,	0x25B3,	0x25B2,	0x25CE,
	0x2606,	0x2605,	0x25C7,	0x25C6,	0x25A1,	0x25A0,	0x25BD,	0x25BC,
	0x32A3,	0x2105,	0x00AF,	0xFFE3,	0xFF3F,	0x02CD,	0xFE49,	0xFE4A,
	0xFE4D,	0xFE4E,	0xFE4B,	0xFE4C,	0xFE5F,	0xFE60,	0xFE61,	0xFF0B,
	0xFF0D,	0x00D7,	0x00F7,	0x00B1,	0x221A,	0xFF1C,	0xFF1E,	0xFF1D,
	0x2266,	0x2267,	0x2260,	0x221E,	0x2252,	0x2261,	0xFE62,	0xFE63,
	0xFE64,	0xFE65,	0xFE66,	0xFF5E,	0x2229,	0x222A,	0x22A5,	0x2220,
	0x221F,	0x22BF,	0x33D2,	0x33D1,	0x222B,	0x222E,	0x2235,	0x2234,
	0x2640,	0x2642,	0x2295,	0x2299,	0x2191,	0x2193,	0x2190,	0x2192,
	0x2196,	0x2197,	0x2199,	0x2198,	0x2225,	0x2223,	0xFF0F,

	/* Big5-HKSCS 0xA240..0xA27E, 0xA2A1..0xA2FE */
	0xFF3C,	0x2215,	0xFE68,	0xFF04,	0xFFE5,	0x3012,	0xFFE0,	0xFFE1,
	0xFF05,	0xFF20,	0x2103,	0x2109,	0xFE69,	0xFE6A,	0xFE6B,	0x33D5,
	0x339C,	0x339D,	0x339E,	0x33CE,	0x33A1,	0x338E,	0x338F,	0x33C4,
	0x00B0,	0x5159,	0x515B,	0x515E,	0x515D,	0x5161,	0x5163,	0x55E7,
	0x74E9,	0x7CCE,	0x2581,	0x2582,	0x2583,	0x2584,	0x2585,	0x2586,
	0x2587,	0x2588,	0x258F,	0x258E,	0x258D,	0x258C,	0x258B,	0x258A,
	0x2589,	0x253C,	0x2534,	0x252C,	0x2524,	0x251C,	0x2594,	0x2500,
	0x2502,	0x2595,	0x250C,	0x2510,	0x2514,	0x2518,	0x256D,
		0x256E,	0x2570,	0x256F,	0x2550,	0x255E,	0x256A,	0x2561,
	0x25E2,	0x25E3,	0x25E5,	0x25E4,	0x2571,	0x2572,	0x2573,	0xFF10,
	0xFF11,	0xFF12,	0xFF13,	0xFF14,	0xFF15,	0xFF16,	0xFF17,	0xFF18,
	0xFF19,	0x2160,	0x2161,	0x2162,	0x2163,	0x2164,	0x2165,	0x2166,
	0x2167,	0x2168,	0x2169,	0x3021,	0x3022,	0x3023,	0x3024,	0x3025,
	0x3026,	0x3027,	0x3028,	0x3029,	0x5341,	0x5344,	0x5345,	0xFF21,
	0xFF22,	0xFF23,	0xFF24,	0xFF25,	0xFF26,	0xFF27,	0xFF28,	0xFF29,
	0xFF2A,	0xFF2B,	0xFF2C,	0xFF2D,	0xFF2E,	0xFF2F,	0xFF30,	0xFF31,
	0xFF32,	0xFF33,	0xFF34,	0xFF35,	0xFF36,	0xFF37,	0xFF38,	0xFF39,
	0xFF3A,	0xFF41,	0xFF42,	0xFF43,	0xFF44,	0xFF45,	0xFF46,	0xFF47,
	0xFF48,	0xFF49,	0xFF4A,	0xFF4B,	0xFF4C,	0xFF4D,	0xFF4E,	0xFF4F,
	0xFF50,	0xFF51,	0xFF52,	0xFF53,	0xFF54,	0xFF55,	0xFF56,

	/* Big5-HKSCS 0xA340..0xA37E, 0xA3A1..0xA3FE */
	0xFF57,	0xFF58,	0xFF59,	0xFF5A,	0x0391,	0x0392,	0x0393,	0x0394,
	0x0395,	0x0396,	0x0397,	0x0398,	0x0399,	0x039A,	0x039B,	0x039C,
	0x039D,	0x039E,	0x039F,	0x03A0,	0x03A1,	0x03A3,	0x03A4,	0x03A5,
	0x03A6,	0x03A7,	0x03A8,	0x03A9,	0x03B1,	0x03B2,	0x03B3,	0x03B4,
	0x03B5,	0x03B6,	0x03B7,	0x03B8,	0x03B9,	0x03BA,	0x03BB,	0x03BC,
	0x03BD,	0x03BE,	0x03BF,	0x03C0,	0x03C1,	0x03C3,	0x03C4,	0x03C5,
	0x03C6,	0x03C7,	0x03C8,	0x03C9,	0x3105,	0x3106,	0x3107,	0x3108,
	0x3109,	0x310A,	0x310B,	0x310C,	0x310D,	0x310E,	0x310F,
		0x3110,	0x3111,	0x3112,	0x3113,	0x3114,	0x3115,	0x3116,
	0x3117,	0x3118,	0x3119,	0x311A,	0x311B,	0x311C,	0x311D,	0x311E,
	0x311F,	0x3120,	0x3121,	0x3122,	0x3123,	0x3124,	0x3125,	0x3126,
	0x3127,	0x3128,	0x3129,	0x02D9,	0x02C9,	0x02CA,	0x02C7,	0x02CB,
	0x0000,	0x0000,	0x0000,	0x0000,	0x0000,	0x0000,	0x0000,	0x0000,
	0x0000,	0x0000,	0x0000,	0x0000,	0x0000,	0x0000,	0x0000,	0x0000,
	0x0000,	0x0000,	0x0000,	0x0000,	0x0000,	0x0000,	0x0000,	0x0000,
	0x0000,	0x0000,	0x0000,	0x0000,	0x0000,	0x0000,	0x0000,	0x0000,
	0x0000,	0x0000,	0x0000,	0x0000,	0x0000,	0x0000,	0x0000,	0x0000,
	0x0000,	0x0000,	0x0000,	0x0000,	0x0000,	0x0000,	0x0000,	0x0000,
	0x0000,	0x0000,	0x0000,	0x0000,	0x0000,	0x0000,	0x0000,	0x0000,
	0x0000,	0x0000,	0x0000,	0x0000,	0x0000,	0x0000,	0x0000,

	/* Big5-HKSCS 0xA440..0xA47E, 0xA4A1..0xA4FE */
	0x4E00,	0x4E59,	0x4E01,	0x4E03,	0x4E43,	0x4E5D,	0x4E86,	0x4E8C,
	0x4EBA,	0x513F,	0x5165,	0x516B,	0x51E0,	0x5200,	0x5201,	0x529B,
	0x5315,	0x5341,	0x535C,	0x53C8,	0x4E09,	0x4E0B,	0x4E08,	0x4E0A,
	0x4E2B,	0x4E38,	0x51E1,	0x4E45,	0x4E48,	0x4E5F,	0x4E5E,	0x4E8E,
	0x4EA1,	0x5140,	0x5203,	0x52FA,	0x5343,	0x53C9,	0x53E3,	0x571F,
	0x58EB,	0x5915,	0x5927,	0x5973,	0x5B50,	0x5B51,	0x5B53,	0x5BF8,
	0x5C0F,	0x5C22,	0x5C38,	0x5C71,	0x5DDD,	0x5DE5,	0x5DF1,	0x5DF2,
	0x5DF3,	0x5DFE,	0x5E72,	0x5EFE,	0x5F0B,	0x5F13,	0x624D,
		0x4E11,	0x4E10,	0x4E0D,	0x4E2D,	0x4E30,	0x4E39,	0x4E4B,
	0x5C39,	0x4E88,	0x4E91,	0x4E95,	0x4E92,	0x4E94,	0x4EA2,	0x4EC1,
	0x4EC0,	0x4EC3,	0x4EC6,	0x4EC7,	0x4ECD,	0x4ECA,	0x4ECB,	0x4EC4,
	0x5143,	0x5141,	0x5167,	0x516D,	0x516E,	0x516C,	0x5197,	0x51F6,
	0x5206,	0x5207,	0x5208,	0x52FB,	0x52FE,	0x52FF,	0x5316,	0x5339,
	0x5348,	0x5347,	0x5345,	0x535E,	0x5384,	0x53CB,	0x53CA,	0x53CD,
	0x58EC,	0x5929,	0x592B,	0x592A,	0x592D,	0x5B54,	0x5C11,	0x5C24,
	0x5C3A,	0x5C6F,	0x5DF4,	0x5E7B,	0x5EFF,	0x5F14,	0x5F15,	0x5FC3,
	0x6208,	0x6236,	0x624B,	0x624E,	0x652F,	0x6587,	0x6597,	0x65A4,
	0x65B9,	0x65E5,	0x66F0,	0x6708,	0x6728,	0x6B20,	0x6B62,	0x6B79,
	0x6BCB,	0x6BD4,	0x6BDB,	0x6C0F,	0x6C34,	0x706B,	0x722A,	0x7236,
	0x723B,	0x7247,	0x7259,	0x725B,	0x72AC,	0x738B,	0x4E19,

	/* Big5-HKSCS 0xA540..0xA57E, 0xA5A1..0xA5FE */
	0x4E16,	0x4E15,	0x4E14,	0x4E18,	0x4E3B,	0x4E4D,	0x4E4F,	0x4E4E,
	0x4EE5,	0x4ED8,	0x4ED4,	0x4ED5,	0x4ED6,	0x4ED7,	0x4EE3,	0x4EE4,
	0x4ED9,	0x4EDE,	0x5145,	0x5144,	0x5189,	0x518A,	0x51AC,	0x51F9,
	0x51FA,	0x51F8,	0x520A,	0x52A0,	0x529F,	0x5305,	0x5306,	0x5317,
	0x531D,	0x4EDF,	0x534A,	0x5349,	0x5361,	0x5360,	0x536F,	0x536E,
	0x53BB,	0x53EF,	0x53E4,	0x53F3,	0x53EC,	0x53EE,	0x53E9,	0x53E8,
	0x53FC,	0x53F8,	0x53F5,	0x53EB,	0x53E6,	0x53EA,	0x53F2,	0x53F1,
	0x53F0,	0x53E5,	0x53ED,	0x53FB,	0x56DB,	0x56DA,	0x5916,
		0x592E,	0x5931,	0x5974,	0x5976,	0x5B55,	0x5B83,	0x5C3C,
	0x5DE8,	0x5DE7,	0x5DE6,	0x5E02,	0x5E03,	0x5E73,	0x5E7C,	0x5F01,
	0x5F18,	0x5F17,	0x5FC5,	0x620A,	0x6253,	0x6254,	0x6252,	0x6251,
	0x65A5,	0x65E6,	0x672E,	0x672C,	0x672A,	0x672B,	0x672D,	0x6B63,
	0x6BCD,	0x6C11,	0x6C10,	0x6C38,	0x6C41,	0x6C40,	0x6C3E,	0x72AF,
	0x7384,	0x7389,	0x74DC,	0x74E6,	0x7518,	0x751F,	0x7528,	0x7529,
	0x7530,	0x7531,	0x7532,	0x7533,	0x758B,	0x767D,	0x76AE,	0x76BF,
	0x76EE,	0x77DB,	0x77E2,	0x77F3,	0x793A,	0x79BE,	0x7A74,	0x7ACB,
	0x4E1E,	0x4E1F,	0x4E52,	0x4E53,	0x4E69,	0x4E99,	0x4EA4,	0x4EA6,
	0x4EA5,	0x4EFF,	0x4F09,	0x4F19,	0x4F0A,	0x4F15,	0x4F0D,	0x4F10,
	0x4F11,	0x4F0F,	0x4EF2,	0x4EF6,	0x4EFB,	0x4EF0,	0x4EF3,	0x4EFD,
	0x4F01,	0x4F0B,	0x5149,	0x5147,	0x5146,	0x5148,	0x5168,

	/* Big5-HKSCS 0xA640..0xA67E, 0xA6A1..0xA6FE */
	0x5171,	0x518D,	0x51B0,	0x5217,	0x5211,	0x5212,	0x520E,	0x5216,
	0x52A3,	0x5308,	0x5321,	0x5320,	0x5370,	0x5371,	0x5409,	0x540F,
	0x540C,	0x540A,	0x5410,	0x5401,	0x540B,	0x5404,	0x5411,	0x540D,
	0x5408,	0x5403,	0x540E,	0x5406,	0x5412,	0x56E0,	0x56DE,	0x56DD,
	0x5733,	0x5730,	0x5728,	0x572D,	0x572C,	0x572F,	0x5729,	0x5919,
	0x591A,	0x5937,	0x5938,	0x5984,	0x5978,	0x5983,	0x597D,	0x5979,
	0x5982,	0x5981,	0x5B57,	0x5B58,	0x5B87,	0x5B88,	0x5B85,	0x5B89,
	0x5BFA,	0x5C16,	0x5C79,	0x5DDE,	0x5E06,	0x5E76,	0x5E74,
		0x5F0F,	0x5F1B,	0x5FD9,	0x5FD6,	0x620E,	0x620C,	0x620D,
	0x6210,	0x6263,	0x625B,	0x6258,	0x6536,	0x65E9,	0x65E8,	0x65EC,
	0x65ED,	0x66F2,	0x66F3,	0x6709,	0x673D,	0x6734,	0x6731,	0x6735,
	0x6B21,	0x6B64,	0x6B7B,	0x6C16,	0x6C5D,	0x6C57,	0x6C59,	0x6C5F,
	0x6C60,	0x6C50,	0x6C55,	0x6C61,	0x6C5B,	0x6C4D,	0x6C4E,	0x7070,
	0x725F,	0x725D,	0x767E,	0x7AF9,	0x7C73,	0x7CF8,	0x7F36,	0x7F8A,
	0x7FBD,	0x8001,	0x8003,	0x800C,	0x8012,	0x8033,	0x807F,	0x8089,
	0x808B,	0x808C,	0x81E3,	0x81EA,	0x81F3,	0x81FC,	0x820C,	0x821B,
	0x821F,	0x826E,	0x8272,	0x827E,	0x866B,	0x8840,	0x884C,	0x8863,
	0x897F,	0x9621,	0x4E32,	0x4EA8,	0x4F4D,	0x4F4F,	0x4F47,	0x4F57,
	0x4F5E,	0x4F34,	0x4F5B,	0x4F55,	0x4F30,	0x4F50,	0x4F51,	0x4F3D,
	0x4F3A,	0x4F38,	0x4F43,	0x4F54,	0x4F3C,	0x4F46,	0x4F63,

	/* Big5-HKSCS 0xA740..0xA77E, 0xA7A1..0xA7FE */
	0x4F5C,	0x4F60,	0x4F2F,	0x4F4E,	0x4F36,	0x4F59,	0x4F5D,	0x4F48,
	0x4F5A,	0x514C,	0x514B,	0x514D,	0x5175,	0x51B6,	0x51B7,	0x5225,
	0x5224,	0x5229,	0x522A,	0x5228,	0x52AB,	0x52A9,	0x52AA,	0x52AC,
	0x5323,	0x5373,	0x5375,	0x541D,	0x542D,	0x541E,	0x543E,	0x5426,
	0x544E,	0x5427,	0x5446,	0x5443,	0x5433,	0x5448,	0x5442,	0x541B,
	0x5429,	0x544A,	0x5439,	0x543B,	0x5438,	0x542E,	0x5435,	0x5436,
	0x5420,	0x543C,	0x5440,	0x5431,	0x542B,	0x541F,	0x542C,	0x56EA,
	0x56F0,	0x56E4,	0x56EB,	0x574A,	0x5751,	0x5740,	0x574D,
		0x5747,	0x574E,	0x573E,	0x5750,	0x574F,	0x573B,	0x58EF,
	0x593E,	0x599D,	0x5992,	0x59A8,	0x599E,	0x59A3,	0x5999,	0x5996,
	0x598D,	0x59A4,	0x5993,	0x598A,	0x59A5,	0x5B5D,	0x5B5C,	0x5B5A,
	0x5B5B,	0x5B8C,	0x5B8B,	0x5B8F,	0x5C2C,	0x5C40,	0x5C41,	0x5C3F,
	0x5C3E,	0x5C90,	0x5C91,	0x5C94,	0x5C8C,	0x5DEB,	0x5E0C,	0x5E8F,
	0x5E87,	0x5E8A,	0x5EF7,	0x5F04,	0x5F1F,	0x5F64,	0x5F62,	0x5F77,
	0x5F79,	0x5FD8,	0x5FCC,	0x5FD7,	0x5FCD,	0x5FF1,	0x5FEB,	0x5FF8,
	0x5FEA,	0x6212,	0x6211,	0x6284,	0x6297,	0x6296,	0x6280,	0x6276,
	0x6289,	0x626D,	0x628A,	0x627C,	0x627E,	0x6279,	0x6273,	0x6292,
	0x626F,	0x6298,	0x626E,	0x6295,	0x6293,	0x6291,	0x6286,	0x6539,
	0x653B,	0x6538,	0x65F1,	0x66F4,	0x675F,	0x674E,	0x674F,	0x6750,
	0x6751,	0x675C,	0x6756,	0x675E,	0x6749,	0x6746,	0x6760,

	/* Big5-HKSCS 0xA840..0xA87E, 0xA8A1..0xA8FE */
	0x6753,	0x6757,	0x6B65,	0x6BCF,	0x6C42,	0x6C5E,	0x6C99,	0x6C81,
	0x6C88,	0x6C89,	0x6C85,	0x6C9B,	0x6C6A,	0x6C7A,	0x6C90,	0x6C70,
	0x6C8C,	0x6C68,	0x6C96,	0x6C92,	0x6C7D,	0x6C83,	0x6C72,	0x6C7E,
	0x6C74,	0x6C86,	0x6C76,	0x6C8D,	0x6C94,	0x6C98,	0x6C82,	0x7076,
	0x707C,	0x707D,	0x7078,	0x7262,	0x7261,	0x7260,	0x72C4,	0x72C2,
	0x7396,	0x752C,	0x752B,	0x7537,	0x7538,	0x7682,	0x76EF,	0x77E3,
	0x79C1,	0x79C0,	0x79BF,	0x7A76,	0x7CFB,	0x7F55,	0x8096,	0x8093,
	0x809D,	0x8098,	0x809B,	0x809A,	0x80B2,	0x826F,	0x8292,
		0x828B,	0x828D,	0x898B,	0x89D2,	0x8A00,	0x8C37,	0x8C46,
	0x8C55,	0x8C9D,	0x8D64,	0x8D70,	0x8DB3,	0x8EAB,	0x8ECA,	0x8F9B,
	0x8FB0,	0x8FC2,	0x8FC6,	0x8FC5,	0x8FC4,	0x5DE1,	0x9091,	0x90A2,
	0x90AA,	0x90A6,	0x90A3,	0x9149,	0x91C6,	0x91CC,	0x9632,	0x962E,
	0x9631,	0x962A,	0x962C,	0x4E26,	0x4E56,	0x4E73,	0x4E8B,	0x4E9B,
	0x4E9E,	0x4EAB,	0x4EAC,	0x4F6F,	0x4F9D,	0x4F8D,	0x4F73,	0x4F7F,
	0x4F6C,	0x4F9B,	0x4F8B,	0x4F86,	0x4F83,	0x4F70,	0x4F75,	0x4F88,
	0x4F69,	0x4F7B,	0x4F96,	0x4F7E,	0x4F8F,	0x4F91,	0x4F7A,	0x5154,
	0x5152,	0x5155,	0x5169,	0x5177,	0x5176,	0x5178,	0x51BD,	0x51FD,
	0x523B,	0x5238,	0x5237,	0x523A,	0x5230,	0x522E,	0x5236,	0x5241,
	0x52BE,	0x52BB,	0x5352,	0x5354,	0x5353,	0x5351,	0x5366,	0x5377,
	0x5378,	0x5379,	0x53D6,	0x53D4,	0x53D7,	0x5473,	0x5475,

	/* Big5-HKSCS 0xA940..0xA97E, 0xA9A1..0xA9FE */
	0x5496,	0x5478,	0x5495,	0x5480,	0x547B,	0x5477,	0x5484,	0x5492,
	0x5486,	0x547C,	0x5490,	0x5471,	0x5476,	0x548C,	0x549A,	0x5462,
	0x5468,	0x548B,	0x547D,	0x548E,	0x56FA,	0x5783,	0x5777,	0x576A,
	0x5769,	0x5761,	0x5766,	0x5764,	0x577C,	0x591C,	0x5949,	0x5947,
	0x5948,	0x5944,	0x5954,	0x59BE,	0x59BB,	0x59D4,	0x59B9,	0x59AE,
	0x59D1,	0x59C6,	0x59D0,	0x59CD,	0x59CB,	0x59D3,	0x59CA,	0x59AF,
	0x59B3,	0x59D2,	0x59C5,	0x5B5F,	0x5B64,	0x5B63,	0x5B97,	0x5B9A,
	0x5B98,	0x5B9C,	0x5B99,	0x5B9B,	0x5C1A,	0x5C48,	0x5C45,
		0x5C46,	0x5CB7,	0x5CA1,	0x5CB8,	0x5CA9,	0x5CAB,	0x5CB1,
	0x5CB3,	0x5E18,	0x5E1A,	0x5E16,	0x5E15,	0x5E1B,	0x5E11,	0x5E78,
	0x5E9A,	0x5E97,	0x5E9C,	0x5E95,	0x5E96,	0x5EF6,	0x5F26,	0x5F27,
	0x5F29,	0x5F80,	0x5F81,	0x5F7F,	0x5F7C,	0x5FDD,	0x5FE0,	0x5FFD,
	0x5FF5,	0x5FFF,	0x600F,	0x6014,	0x602F,	0x6035,	0x6016,	0x602A,
	0x6015,	0x6021,	0x6027,	0x6029,	0x602B,	0x601B,	0x6216,	0x6215,
	0x623F,	0x623E,	0x6240,	0x627F,	0x62C9,	0x62CC,	0x62C4,	0x62BF,
	0x62C2,	0x62B9,	0x62D2,	0x62DB,	0x62AB,	0x62D3,	0x62D4,	0x62CB,
	0x62C8,	0x62A8,	0x62BD,	0x62BC,	0x62D0,	0x62D9,	0x62C7,	0x62CD,
	0x62B5,	0x62DA,	0x62B1,	0x62D8,	0x62D6,	0x62D7,	0x62C6,	0x62AC,
	0x62CE,	0x653E,	0x65A7,	0x65BC,	0x65FA,	0x6614,	0x6613,	0x660C,
	0x6606,	0x6602,	0x660E,	0x6600,	0x660F,	0x6615,	0x660A,

	/* Big5-HKSCS 0xAA40..0xAA7E, 0xAAA1..0xAAFE */
	0x6607,	0x670D,	0x670B,	0x676D,	0x678B,	0x6795,	0x6771,	0x679C,
	0x6773,	0x6777,	0x6787,	0x679D,	0x6797,	0x676F,	0x6770,	0x677F,
	0x6789,	0x677E,	0x6790,	0x6775,	0x679A,	0x6793,	0x677C,	0x676A,
	0x6772,	0x6B23,	0x6B66,	0x6B67,	0x6B7F,	0x6C13,	0x6C1B,	0x6CE3,
	0x6CE8,	0x6CF3,	0x6CB1,	0x6CCC,	0x6CE5,	0x6CB3,	0x6CBD,	0x6CBE,
	0x6CBC,	0x6CE2,	0x6CAB,	0x6CD5,	0x6CD3,	0x6CB8,	0x6CC4,	0x6CB9,
	0x6CC1,	0x6CAE,	0x6CD7,	0x6CC5,	0x6CF1,	0x6CBF,	0x6CBB,	0x6CE1,
	0x6CDB,	0x6CCA,	0x6CAC,	0x6CEF,	0x6CDC,	0x6CD6,	0x6CE0,
		0x7095,	0x708E,	0x7092,	0x708A,	0x7099,	0x722C,	0x722D,
	0x7238,	0x7248,	0x7267,	0x7269,	0x72C0,	0x72CE,	0x72D9,	0x72D7,
	0x72D0,	0x73A9,	0x73A8,	0x739F,	0x73AB,	0x73A5,	0x753D,	0x759D,
	0x7599,	0x759A,	0x7684,	0x76C2,	0x76F2,	0x76F4,	0x77E5,	0x77FD,
	0x793E,	0x7940,	0x7941,	0x79C9,	0x79C8,	0x7A7A,	0x7A79,	0x7AFA,
	0x7CFE,	0x7F54,	0x7F8C,	0x7F8B,	0x8005,	0x80BA,	0x80A5,	0x80A2,
	0x80B1,	0x80A1,	0x80AB,	0x80A9,	0x80B4,	0x80AA,	0x80AF,	0x81E5,
	0x81FE,	0x820D,	0x82B3,	0x829D,	0x8299,	0x82AD,	0x82BD,	0x829F,
	0x82B9,	0x82B1,	0x82AC,	0x82A5,	0x82AF,	0x82B8,	0x82A3,	0x82B0,
	0x82BE,	0x82B7,	0x864E,	0x8671,	0x521D,	0x8868,	0x8ECB,	0x8FCE,
	0x8FD4,	0x8FD1,	0x90B5,	0x90B8,	0x90B1,	0x90B6,	0x91C7,	0x91D1,
	0x9577,	0x9580,	0x961C,	0x9640,	0x963F,	0x963B,	0x9644,

	/* Big5-HKSCS 0xAB40..0xAB7E, 0xABA1..0xABFE */
	0x9642,	0x96B9,	0x96E8,	0x9752,	0x975E,	0x4E9F,	0x4EAD,	0x4EAE,
	0x4FE1,	0x4FB5,	0x4FAF,	0x4FBF,	0x4FE0,	0x4FD1,	0x4FCF,	0x4FDD,
	0x4FC3,	0x4FB6,	0x4FD8,	0x4FDF,	0x4FCA,	0x4FD7,	0x4FAE,	0x4FD0,
	0x4FC4,	0x4FC2,	0x4FDA,	0x4FCE,	0x4FDE,	0x4FB7,	0x5157,	0x5192,
	0x5191,	0x51A0,	0x524E,	0x5243,	0x524A,	0x524D,	0x524C,	0x524B,
	0x5247,	0x52C7,	0x52C9,	0x52C3,	0x52C1,	0x530D,	0x5357,	0x537B,
	0x539A,	0x53DB,	0x54AC,	0x54C0,	0x54A8,	0x54CE,	0x54C9,	0x54B8,
	0x54A6,	0x54B3,	0x54C7,	0x54C2,	0x54BD,	0x54AA,	0x54C1,
		0x54C4,	0x54C8,	0x54AF,	0x54AB,	0x54B1,	0x54BB,	0x54A9,
	0x54A7,	0x54BF,	0x56FF,	0x5782,	0x578B,	0x57A0,	0x57A3,	0x57A2,
	0x57CE,	0x57AE,	0x5793,	0x5955,	0x5951,	0x594F,	0x594E,	0x5950,
	0x59DC,	0x59D8,	0x59FF,	0x59E3,	0x59E8,	0x5A03,	0x59E5,	0x59EA,
	0x59DA,	0x59E6,	0x5A01,	0x59FB,	0x5B69,	0x5BA3,	0x5BA6,	0x5BA4,
	0x5BA2,	0x5BA5,	0x5C01,	0x5C4E,	0x5C4F,	0x5C4D,	0x5C4B,	0x5CD9,
	0x5CD2,	0x5DF7,	0x5E1D,	0x5E25,	0x5E1F,	0x5E7D,	0x5EA0,	0x5EA6,
	0x5EFA,	0x5F08,	0x5F2D,	0x5F65,	0x5F88,	0x5F85,	0x5F8A,	0x5F8B,
	0x5F87,	0x5F8C,	0x5F89,	0x6012,	0x601D,	0x6020,	0x6025,	0x600E,
	0x6028,	0x604D,	0x6070,	0x6068,	0x6062,	0x6046,	0x6043,	0x606C,
	0x606B,	0x606A,	0x6064,	0x6241,	0x62DC,	0x6316,	0x6309,	0x62FC,
	0x62ED,	0x6301,	0x62EE,	0x62FD,	0x6307,	0x62F1,	0x62F7,

	/* Big5-HKSCS 0xAC40..0xAC7E, 0xACA1..0xACFE */
	0x62EF,	0x62EC,	0x62FE,	0x62F4,	0x6311,	0x6302,	0x653F,	0x6545,
	0x65AB,	0x65BD,	0x65E2,	0x6625,	0x662D,	0x6620,	0x6627,	0x662F,
	0x661F,	0x6628,	0x6631,	0x6624,	0x66F7,	0x67FF,	0x67D3,	0x67F1,
	0x67D4,	0x67D0,	0x67EC,	0x67B6,	0x67AF,	0x67F5,	0x67E9,	0x67EF,
	0x67C4,	0x67D1,	0x67B4,	0x67DA,	0x67E5,	0x67B8,	0x67CF,	0x67DE,
	0x67F3,	0x67B0,	0x67D9,	0x67E2,	0x67DD,	0x67D2,	0x6B6A,	0x6B83,
	0x6B86,	0x6BB5,	0x6BD2,	0x6BD7,	0x6C1F,	0x6CC9,	0x6D0B,	0x6D32,
	0x6D2A,	0x6D41,	0x6D25,	0x6D0C,	0x6D31,	0x6D1E,	0x6D17,
		0x6D3B,	0x6D3D,	0x6D3E,	0x6D36,	0x6D1B,	0x6CF5,	0x6D39,
	0x6D27,	0x6D38,	0x6D29,	0x6D2E,	0x6D35,	0x6D0E,	0x6D2B,	0x70AB,
	0x70BA,	0x70B3,	0x70AC,	0x70AF,	0x70AD,	0x70B8,	0x70AE,	0x70A4,
	0x7230,	0x7272,	0x726F,	0x7274,	0x72E9,	0x72E0,	0x72E1,	0x73B7,
	0x73CA,	0x73BB,	0x73B2,	0x73CD,	0x73C0,	0x73B3,	0x751A,	0x752D,
	0x754F,	0x754C,	0x754E,	0x754B,	0x75AB,	0x75A4,	0x75A5,	0x75A2,
	0x75A3,	0x7678,	0x7686,	0x7687,	0x7688,	0x76C8,	0x76C6,	0x76C3,
	0x76C5,	0x7701,	0x76F9,	0x76F8,	0x7709,	0x770B,	0x76FE,	0x76FC,
	0x7707,	0x77DC,	0x7802,	0x7814,	0x780C,	0x780D,	0x7946,	0x7949,
	0x7948,	0x7947,	0x79B9,	0x79BA,	0x79D1,	0x79D2,	0x79CB,	0x7A7F,
	0x7A81,	0x7AFF,	0x7AFD,	0x7C7D,	0x7D02,	0x7D05,	0x7D00,	0x7D09,
	0x7D07,	0x7D04,	0x7D06,	0x7F38,	0x7F8E,	0x7FBF,	0x8004,

	/* Big5-HKSCS 0xAD40..0xAD7E, 0xADA1..0xADFE */
	0x8010,	0x800D,	0x8011,	0x8036,	0x80D6,	0x80E5,	0x80DA,	0x80C3,
	0x80C4,	0x80CC,	0x80E1,	0x80DB,	0x80CE,	0x80DE,	0x80E4,	0x80DD,
	0x81F4,	0x8222,	0x82E7,	0x8303,	0x8305,	0x82E3,	0x82DB,	0x82E6,
	0x8304,	0x82E5,	0x8302,	0x8309,	0x82D2,	0x82D7,	0x82F1,	0x8301,
	0x82DC,	0x82D4,	0x82D1,	0x82DE,	0x82D3,	0x82DF,	0x82EF,	0x8306,
	0x8650,	0x8679,	0x867B,	0x867A,	0x884D,	0x886B,	0x8981,	0x89D4,
	0x8A08,	0x8A02,	0x8A03,	0x8C9E,	0x8CA0,	0x8D74,	0x8D73,	0x8DB4,
	0x8ECD,	0x8ECC,	0x8FF0,	0x8FE6,	0x8FE2,	0x8FEA,	0x8FE5,
		0x8FED,	0x8FEB,	0x8FE4,	0x8FE8,	0x90CA,	0x90CE,	0x90C1,
	0x90C3,	0x914B,	0x914A,	0x91CD,	0x9582,	0x9650,	0x964B,	0x964C,
	0x964D,	0x9762,	0x9769,	0x97CB,	0x97ED,	0x97F3,	0x9801,	0x98A8,
	0x98DB,	0x98DF,	0x9996,	0x9999,	0x4E58,	0x4EB3,	0x500C,	0x500D,
	0x5023,	0x4FEF,	0x5026,	0x5025,	0x4FF8,	0x5029,	0x5016,	0x5006,
	0x503C,	0x501F,	0x501A,	0x5012,	0x5011,	0x4FFA,	0x5000,	0x5014,
	0x5028,	0x4FF1,	0x5021,	0x500B,	0x5019,	0x5018,	0x4FF3,	0x4FEE,
	0x502D,	0x502A,	0x4FFE,	0x502B,	0x5009,	0x517C,	0x51A4,	0x51A5,
	0x51A2,	0x51CD,	0x51CC,	0x51C6,	0x51CB,	0x5256,	0x525C,	0x5254,
	0x525B,	0x525D,	0x532A,	0x537F,	0x539F,	0x539D,	0x53DF,	0x54E8,
	0x5510,	0x5501,	0x5537,	0x54FC,	0x54E5,	0x54F2,	0x5506,	0x54FA,
	0x5514,	0x54E9,	0x54ED,	0x54E1,	0x5509,	0x54EE,	0x54EA,

	/* Big5-HKSCS 0xAE40..0xAE7E, 0xAEA1..0xAEFE */
	0x54E6,	0x5527,	0x5507,	0x54FD,	0x550F,	0x5703,	0x5704,	0x57C2,
	0x57D4,	0x57CB,	0x57C3,	0x5809,	0x590F,	0x5957,	0x5958,	0x595A,
	0x5A11,	0x5A18,	0x5A1C,	0x5A1F,	0x5A1B,	0x5A13,	0x59EC,	0x5A20,
	0x5A23,	0x5A29,	0x5A25,	0x5A0C,	0x5A09,	0x5B6B,	0x5C58,	0x5BB0,
	0x5BB3,	0x5BB6,	0x5BB4,	0x5BAE,	0x5BB5,	0x5BB9,	0x5BB8,	0x5C04,
	0x5C51,	0x5C55,	0x5C50,	0x5CED,	0x5CFD,	0x5CFB,	0x5CEA,	0x5CE8,
	0x5CF0,	0x5CF6,	0x5D01,	0x5CF4,	0x5DEE,	0x5E2D,	0x5E2B,	0x5EAB,
	0x5EAD,	0x5EA7,	0x5F31,	0x5F92,	0x5F91,	0x5F90,	0x6059,
		0x6063,	0x6065,	0x6050,	0x6055,	0x606D,	0x6069,	0x606F,
	0x6084,	0x609F,	0x609A,	0x608D,	0x6094,	0x608C,	0x6085,	0x6096,
	0x6247,	0x62F3,	0x6308,	0x62FF,	0x634E,	0x633E,	0x632F,	0x6355,
	0x6342,	0x6346,	0x634F,	0x6349,	0x633A,	0x6350,	0x633D,	0x632A,
	0x632B,	0x6328,	0x634D,	0x634C,	0x6548,	0x6549,	0x6599,	0x65C1,
	0x65C5,	0x6642,	0x6649,	0x664F,	0x6643,	0x6652,	0x664C,	0x6645,
	0x6641,	0x66F8,	0x6714,	0x6715,	0x6717,	0x6821,	0x6838,	0x6848,
	0x6846,	0x6853,	0x6839,	0x6842,	0x6854,	0x6829,	0x68B3,	0x6817,
	0x684C,	0x6851,	0x683D,	0x67F4,	0x6850,	0x6840,	0x683C,	0x6843,
	0x682A,	0x6845,	0x6813,	0x6818,	0x6841,	0x6B8A,	0x6B89,	0x6BB7,
	0x6C23,	0x6C27,	0x6C28,	0x6C26,	0x6C24,	0x6CF0,	0x6D6A,	0x6D95,
	0x6D88,	0x6D87,	0x6D66,	0x6D78,	0x6D77,	0x6D59,	0x6D93,

	/* Big5-HKSCS 0xAF40..0xAF7E, 0xAFA1..0xAFFE */
	0x6D6C,	0x6D89,	0x6D6E,	0x6D5A,	0x6D74,	0x6D69,	0x6D8C,	0x6D8A,
	0x6D79,	0x6D85,	0x6D65,	0x6D94,	0x70CA,	0x70D8,	0x70E4,	0x70D9,
	0x70C8,	0x70CF,	0x7239,	0x7279,	0x72FC,	0x72F9,	0x72FD,	0x72F8,
	0x72F7,	0x7386,	0x73ED,	0x7409,	0x73EE,	0x73E0,	0x73EA,	0x73DE,
	0x7554,	0x755D,	0x755C,	0x755A,	0x7559,	0x75BE,	0x75C5,	0x75C7,
	0x75B2,	0x75B3,	0x75BD,	0x75BC,	0x75B9,	0x75C2,	0x75B8,	0x768B,
	0x76B0,	0x76CA,	0x76CD,	0x76CE,	0x7729,	0x771F,	0x7720,	0x7728,
	0x77E9,	0x7830,	0x7827,	0x7838,	0x781D,	0x7834,	0x7837,
		0x7825,	0x782D,	0x7820,	0x781F,	0x7832,	0x7955,	0x7950,
	0x7960,	0x795F,	0x7956,	0x795E,	0x795D,	0x7957,	0x795A,	0x79E4,
	0x79E3,	0x79E7,	0x79DF,	0x79E6,	0x79E9,	0x79D8,	0x7A84,	0x7A88,
	0x7AD9,	0x7B06,	0x7B11,	0x7C89,	0x7D21,	0x7D17,	0x7D0B,	0x7D0A,
	0x7D20,	0x7D22,	0x7D14,	0x7D10,	0x7D15,	0x7D1A,	0x7D1C,	0x7D0D,
	0x7D19,	0x7D1B,	0x7F3A,	0x7F5F,	0x7F94,	0x7FC5,	0x7FC1,	0x8006,
	0x8018,	0x8015,	0x8019,	0x8017,	0x803D,	0x803F,	0x80F1,	0x8102,
	0x80F0,	0x8105,	0x80ED,	0x80F4,	0x8106,	0x80F8,	0x80F3,	0x8108,
	0x80FD,	0x810A,	0x80FC,	0x80EF,	0x81ED,	0x81EC,	0x8200,	0x8210,
	0x822A,	0x822B,	0x8228,	0x822C,	0x82BB,	0x832B,	0x8352,	0x8354,
	0x834A,	0x8338,	0x8350,	0x8349,	0x8335,	0x8334,	0x834F,	0x8332,
	0x8339,	0x8336,	0x8317,	0x8340,	0x8331,	0x8328,	0x8343,

	/* Big5-HKSCS 0xB040..0xB07E, 0xB0A1..0xB0FE */
	0x8654,	0x868A,	0x86AA,	0x8693,	0x86A4,	0x86A9,	0x868C,	0x86A3,
	0x869C,	0x8870,	0x8877,	0x8881,	0x8882,	0x887D,	0x8879,	0x8A18,
	0x8A10,	0x8A0E,	0x8A0C,	0x8A15,	0x8A0A,	0x8A17,	0x8A13,	0x8A16,
	0x8A0F,	0x8A11,	0x8C48,	0x8C7A,	0x8C79,	0x8CA1,	0x8CA2,	0x8D77,
	0x8EAC,	0x8ED2,	0x8ED4,	0x8ECF,	0x8FB1,	0x9001,	0x9006,	0x8FF7,
	0x9000,	0x8FFA,	0x8FF4,	0x9003,	0x8FFD,	0x9005,	0x8FF8,	0x9095,
	0x90E1,	0x90DD,	0x90E2,	0x9152,	0x914D,	0x914C,	0x91D8,	0x91DD,
	0x91D7,	0x91DC,	0x91D9,	0x9583,	0x9662,	0x9663,	0x9661,
		0x965B,	0x965D,	0x9664,	0x9658,	0x965E,	0x96BB,	0x98E2,
	0x99AC,	0x9AA8,	0x9AD8,	0x9B25,	0x9B32,	0x9B3C,	0x4E7E,	0x507A,
	0x507D,	0x505C,	0x5047,	0x5043,	0x504C,	0x505A,	0x5049,	0x5065,
	0x5076,	0x504E,	0x5055,	0x5075,	0x5074,	0x5077,	0x504F,	0x500F,
	0x506F,	0x506D,	0x515C,	0x5195,	0x51F0,	0x526A,	0x526F,	0x52D2,
	0x52D9,	0x52D8,	0x52D5,	0x5310,	0x530F,	0x5319,	0x533F,	0x5340,
	0x533E,	0x53C3,	0x66FC,	0x5546,	0x556A,	0x5566,	0x5544,	0x555E,
	0x5561,	0x5543,	0x554A,	0x5531,	0x5556,	0x554F,	0x5555,	0x552F,
	0x5564,	0x5538,	0x552E,	0x555C,	0x552C,	0x5563,	0x5533,	0x5541,
	0x5557,	0x5708,	0x570B,	0x5709,	0x57DF,	0x5805,	0x580A,	0x5806,
	0x57E0,	0x57E4,	0x57FA,	0x5802,	0x5835,	0x57F7,	0x57F9,	0x5920,
	0x5962,	0x5A36,	0x5A41,	0x5A49,	0x5A66,	0x5A6A,	0x5A40,

	/* Big5-HKSCS 0xB140..0xB17E, 0xB1A1..0xB1FE */
	0x5A3C,	0x5A62,	0x5A5A,	0x5A46,	0x5A4A,	0x5B70,	0x5BC7,	0x5BC5,
	0x5BC4,	0x5BC2,	0x5BBF,	0x5BC6,	0x5C09,	0x5C08,	0x5C07,	0x5C60,
	0x5C5C,	0x5C5D,	0x5D07,	0x5D06,	0x5D0E,	0x5D1B,	0x5D16,	0x5D22,
	0x5D11,	0x5D29,	0x5D14,	0x5D19,	0x5D24,	0x5D27,	0x5D17,	0x5DE2,
	0x5E38,	0x5E36,	0x5E33,	0x5E37,	0x5EB7,	0x5EB8,	0x5EB6,	0x5EB5,
	0x5EBE,	0x5F35,	0x5F37,	0x5F57,	0x5F6C,	0x5F69,	0x5F6B,	0x5F97,
	0x5F99,	0x5F9E,	0x5F98,	0x5FA1,	0x5FA0,	0x5F9C,	0x607F,	0x60A3,
	0x6089,	0x60A0,	0x60A8,	0x60CB,	0x60B4,	0x60E6,	0x60BD,
		0x60C5,	0x60BB,	0x60B5,	0x60DC,	0x60BC,	0x60D8,	0x60D5,
	0x60C6,	0x60DF,	0x60B8,	0x60DA,	0x60C7,	0x621A,	0x621B,	0x6248,
	0x63A0,	0x63A7,	0x6372,	0x6396,	0x63A2,	0x63A5,	0x6377,	0x6367,
	0x6398,	0x63AA,	0x6371,	0x63A9,	0x6389,	0x6383,	0x639B,	0x636B,
	0x63A8,	0x6384,	0x6388,	0x6399,	0x63A1,	0x63AC,	0x6392,	0x638F,
	0x6380,	0x637B,	0x6369,	0x6368,	0x637A,	0x655D,	0x6556,	0x6551,
	0x6559,	0x6557,	0x555F,	0x654F,	0x6558,	0x6555,	0x6554,	0x659C,
	0x659B,	0x65AC,	0x65CF,	0x65CB,	0x65CC,	0x65CE,	0x665D,	0x665A,
	0x6664,	0x6668,	0x6666,	0x665E,	0x66F9,	0x52D7,	0x671B,	0x6881,
	0x68AF,	0x68A2,	0x6893,	0x68B5,	0x687F,	0x6876,	0x68B1,	0x68A7,
	0x6897,	0x68B0,	0x6883,	0x68C4,	0x68AD,	0x6886,	0x6885,	0x6894,
	0x689D,	0x68A8,	0x689F,	0x68A1,	0x6882,	0x6B32,	0x6BBA,

	/* Big5-HKSCS 0xB240..0xB27E, 0xB2A1..0xB2FE */
	0x6BEB,	0x6BEC,	0x6C2B,	0x6D8E,	0x6DBC,	0x6DF3,	0x6DD9,	0x6DB2,
	0x6DE1,	0x6DCC,	0x6DE4,	0x6DFB,	0x6DFA,	0x6E05,	0x6DC7,	0x6DCB,
	0x6DAF,	0x6DD1,	0x6DAE,	0x6DDE,	0x6DF9,	0x6DB8,	0x6DF7,	0x6DF5,
	0x6DC5,	0x6DD2,	0x6E1A,	0x6DB5,	0x6DDA,	0x6DEB,	0x6DD8,	0x6DEA,
	0x6DF1,	0x6DEE,	0x6DE8,	0x6DC6,	0x6DC4,	0x6DAA,	0x6DEC,	0x6DBF,
	0x6DE6,	0x70F9,	0x7109,	0x710A,	0x70FD,	0x70EF,	0x723D,	0x727D,
	0x7281,	0x731C,	0x731B,	0x7316,	0x7313,	0x7319,	0x7387,	0x7405,
	0x740A,	0x7403,	0x7406,	0x73FE,	0x740D,	0x74E0,	0x74F6,
		0x74F7,	0x751C,	0x7522,	0x7565,	0x7566,	0x7562,	0x7570,
	0x758F,	0x75D4,	0x75D5,	0x75B5,	0x75CA,	0x75CD,	0x768E,	0x76D4,
	0x76D2,	0x76DB,	0x7737,	0x773E,	0x773C,	0x7736,	0x7738,	0x773A,
	0x786B,	0x7843,	0x784E,	0x7965,	0x7968,	0x796D,	0x79FB,	0x7A92,
	0x7A95,	0x7B20,	0x7B28,	0x7B1B,	0x7B2C,	0x7B26,	0x7B19,	0x7B1E,
	0x7B2E,	0x7C92,	0x7C97,	0x7C95,	0x7D46,	0x7D43,	0x7D71,	0x7D2E,
	0x7D39,	0x7D3C,	0x7D40,	0x7D30,	0x7D33,	0x7D44,	0x7D2F,	0x7D42,
	0x7D32,	0x7D31,	0x7F3D,	0x7F9E,	0x7F9A,	0x7FCC,	0x7FCE,	0x7FD2,
	0x801C,	0x804A,	0x8046,	0x812F,	0x8116,	0x8123,	0x812B,	0x8129,
	0x8130,	0x8124,	0x8202,	0x8235,	0x8237,	0x8236,	0x8239,	0x838E,
	0x839E,	0x8398,	0x8378,	0x83A2,	0x8396,	0x83BD,	0x83AB,	0x8392,
	0x838A,	0x8393,	0x8389,	0x83A0,	0x8377,	0x837B,	0x837C,

	/* Big5-HKSCS 0xB340..0xB37E, 0xB3A1..0xB3FE */
	0x8386,	0x83A7,	0x8655,	0x5F6A,	0x86C7,	0x86C0,	0x86B6,	0x86C4,
	0x86B5,	0x86C6,	0x86CB,	0x86B1,	0x86AF,	0x86C9,	0x8853,	0x889E,
	0x8888,	0x88AB,	0x8892,	0x8896,	0x888D,	0x888B,	0x8993,	0x898F,
	0x8A2A,	0x8A1D,	0x8A23,	0x8A25,	0x8A31,	0x8A2D,	0x8A1F,	0x8A1B,
	0x8A22,	0x8C49,	0x8C5A,	0x8CA9,	0x8CAC,	0x8CAB,	0x8CA8,	0x8CAA,
	0x8CA7,	0x8D67,	0x8D66,	0x8DBE,	0x8DBA,	0x8EDB,	0x8EDF,	0x9019,
	0x900D,	0x901A,	0x9017,	0x9023,	0x901F,	0x901D,	0x9010,	0x9015,
	0x901E,	0x9020,	0x900F,	0x9022,	0x9016,	0x901B,	0x9014,
		0x90E8,	0x90ED,	0x90FD,	0x9157,	0x91CE,	0x91F5,	0x91E6,
	0x91E3,	0x91E7,	0x91ED,	0x91E9,	0x9589,	0x966A,	0x9675,	0x9673,
	0x9678,	0x9670,	0x9674,	0x9676,	0x9677,	0x966C,	0x96C0,	0x96EA,
	0x96E9,	0x7AE0,	0x7ADF,	0x9802,	0x9803,	0x9B5A,	0x9CE5,	0x9E75,
	0x9E7F,	0x9EA5,	0x9EBB,	0x50A2,	0x508D,	0x5085,	0x5099,	0x5091,
	0x5080,	0x5096,	0x5098,	0x509A,	0x6700,	0x51F1,	0x5272,	0x5274,
	0x5275,	0x5269,	0x52DE,	0x52DD,	0x52DB,	0x535A,	0x53A5,	0x557B,
	0x5580,	0x55A7,	0x557C,	0x558A,	0x559D,	0x5598,	0x5582,	0x559C,
	0x55AA,	0x5594,	0x5587,	0x558B,	0x5583,	0x55B3,	0x55AE,	0x559F,
	0x553E,	0x55B2,	0x559A,	0x55BB,	0x55AC,	0x55B1,	0x557E,	0x5589,
	0x55AB,	0x5599,	0x570D,	0x582F,	0x582A,	0x5834,	0x5824,	0x5830,
	0x5831,	0x5821,	0x581D,	0x5820,	0x58F9,	0x58FA,	0x5960,

	/* Big5-HKSCS 0xB440..0xB47E, 0xB4A1..0xB4FE */
	0x5A77,	0x5A9A,	0x5A7F,	0x5A92,	0x5A9B,	0x5AA7,	0x5B73,	0x5B71,
	0x5BD2,	0x5BCC,	0x5BD3,	0x5BD0,	0x5C0A,	0x5C0B,	0x5C31,	0x5D4C,
	0x5D50,	0x5D34,	0x5D47,	0x5DFD,	0x5E45,	0x5E3D,	0x5E40,	0x5E43,
	0x5E7E,	0x5ECA,	0x5EC1,	0x5EC2,	0x5EC4,	0x5F3C,	0x5F6D,	0x5FA9,
	0x5FAA,	0x5FA8,	0x60D1,	0x60E1,	0x60B2,	0x60B6,	0x60E0,	0x611C,
	0x6123,	0x60FA,	0x6115,	0x60F0,	0x60FB,	0x60F4,	0x6168,	0x60F1,
	0x610E,	0x60F6,	0x6109,	0x6100,	0x6112,	0x621F,	0x6249,	0x63A3,
	0x638C,	0x63CF,	0x63C0,	0x63E9,	0x63C9,	0x63C6,	0x63CD,
		0x63D2,	0x63E3,	0x63D0,	0x63E1,	0x63D6,	0x63ED,	0x63EE,
	0x6376,	0x63F4,	0x63EA,	0x63DB,	0x6452,	0x63DA,	0x63F9,	0x655E,
	0x6566,	0x6562,	0x6563,	0x6591,	0x6590,	0x65AF,	0x666E,	0x6670,
	0x6674,	0x6676,	0x666F,	0x6691,	0x667A,	0x667E,	0x6677,	0x66FE,
	0x66FF,	0x671F,	0x671D,	0x68FA,	0x68D5,	0x68E0,	0x68D8,	0x68D7,
	0x6905,	0x68DF,	0x68F5,	0x68EE,	0x68E7,	0x68F9,	0x68D2,	0x68F2,
	0x68E3,	0x68CB,	0x68CD,	0x690D,	0x6912,	0x690E,	0x68C9,	0x68DA,
	0x696E,	0x68FB,	0x6B3E,	0x6B3A,	0x6B3D,	0x6B98,	0x6B96,	0x6BBC,
	0x6BEF,	0x6C2E,	0x6C2F,	0x6C2C,	0x6E2F,	0x6E38,	0x6E54,	0x6E21,
	0x6E32,	0x6E67,	0x6E4A,	0x6E20,	0x6E25,	0x6E23,	0x6E1B,	0x6E5B,
	0x6E58,	0x6E24,	0x6E56,	0x6E6E,	0x6E2D,	0x6E26,	0x6E6F,	0x6E34,
	0x6E4D,	0x6E3A,	0x6E2C,	0x6E43,	0x6E1D,	0x6E3E,	0x6ECB,

	/* Big5-HKSCS 0xB540..0xB57E, 0xB5A1..0xB5FE */
	0x6E89,	0x6E19,	0x6E4E,	0x6E63,	0x6E44,	0x6E72,	0x6E69,	0x6E5F,
	0x7119,	0x711A,	0x7126,	0x7130,	0x7121,	0x7136,	0x716E,	0x711C,
	0x724C,	0x7284,	0x7280,	0x7336,	0x7325,	0x7334,	0x7329,	0x743A,
	0x742A,	0x7433,	0x7422,	0x7425,	0x7435,	0x7436,	0x7434,	0x742F,
	0x741B,	0x7426,	0x7428,	0x7525,	0x7526,	0x756B,	0x756A,	0x75E2,
	0x75DB,	0x75E3,	0x75D9,	0x75D8,	0x75DE,	0x75E0,	0x767B,	0x767C,
	0x7696,	0x7693,	0x76B4,	0x76DC,	0x774F,	0x77ED,	0x785D,	0x786C,
	0x786F,	0x7A0D,	0x7A08,	0x7A0B,	0x7A05,	0x7A00,	0x7A98,
		0x7A97,	0x7A96,	0x7AE5,	0x7AE3,	0x7B49,	0x7B56,	0x7B46,
	0x7B50,	0x7B52,	0x7B54,	0x7B4D,	0x7B4B,	0x7B4F,	0x7B51,	0x7C9F,
	0x7CA5,	0x7D5E,	0x7D50,	0x7D68,	0x7D55,	0x7D2B,	0x7D6E,	0x7D72,
	0x7D61,	0x7D66,	0x7D62,	0x7D70,	0x7D73,	0x5584,	0x7FD4,	0x7FD5,
	0x800B,	0x8052,	0x8085,	0x8155,	0x8154,	0x814B,	0x8151,	0x814E,
	0x8139,	0x8146,	0x813E,	0x814C,	0x8153,	0x8174,	0x8212,	0x821C,
	0x83E9,	0x8403,	0x83F8,	0x840D,	0x83E0,	0x83C5,	0x840B,	0x83C1,
	0x83EF,	0x83F1,	0x83F4,	0x8457,	0x840A,	0x83F0,	0x840C,	0x83CC,
	0x83FD,	0x83F2,	0x83CA,	0x8438,	0x840E,	0x8404,	0x83DC,	0x8407,
	0x83D4,	0x83DF,	0x865B,	0x86DF,	0x86D9,	0x86ED,	0x86D4,	0x86DB,
	0x86E4,	0x86D0,	0x86DE,	0x8857,	0x88C1,	0x88C2,	0x88B1,	0x8983,
	0x8996,	0x8A3B,	0x8A60,	0x8A55,	0x8A5E,	0x8A3C,	0x8A41,

	/* Big5-HKSCS 0xB640..0xB67E, 0xB6A1..0xB6FE */
	0x8A54,	0x8A5B,	0x8A50,	0x8A46,	0x8A34,	0x8A3A,	0x8A36,	0x8A56,
	0x8C61,	0x8C82,	0x8CAF,	0x8CBC,	0x8CB3,	0x8CBD,	0x8CC1,	0x8CBB,
	0x8CC0,	0x8CB4,	0x8CB7,	0x8CB6,	0x8CBF,	0x8CB8,	0x8D8A,	0x8D85,
	0x8D81,	0x8DCE,	0x8DDD,	0x8DCB,	0x8DDA,	0x8DD1,	0x8DCC,	0x8DDB,
	0x8DC6,	0x8EFB,	0x8EF8,	0x8EFC,	0x8F9C,	0x902E,	0x9035,	0x9031,
	0x9038,	0x9032,	0x9036,	0x9102,	0x90F5,	0x9109,	0x90FE,	0x9163,
	0x9165,	0x91CF,	0x9214,	0x9215,	0x9223,	0x9209,	0x921E,	0x920D,
	0x9210,	0x9207,	0x9211,	0x9594,	0x958F,	0x958B,	0x9591,
		0x9593,	0x9592,	0x958E,	0x968A,	0x968E,	0x968B,	0x967D,
	0x9685,	0x9686,	0x968D,	0x9672,	0x9684,	0x96C1,	0x96C5,	0x96C4,
	0x96C6,	0x96C7,	0x96EF,	0x96F2,	0x97CC,	0x9805,	0x9806,	0x9808,
	0x98E7,	0x98EA,	0x98EF,	0x98E9,	0x98F2,	0x98ED,	0x99AE,	0x99AD,
	0x9EC3,	0x9ECD,	0x9ED1,	0x4E82,	0x50AD,	0x50B5,	0x50B2,	0x50B3,
	0x50C5,	0x50BE,	0x50AC,	0x50B7,	0x50BB,	0x50AF,	0x50C7,	0x527F,
	0x5277,	0x527D,	0x52DF,	0x52E6,	0x52E4,	0x52E2,	0x52E3,	0x532F,
	0x55DF,	0x55E8,	0x55D3,	0x55E6,	0x55CE,	0x55DC,	0x55C7,	0x55D1,
	0x55E3,	0x55E4,	0x55EF,	0x55DA,	0x55E1,	0x55C5,	0x55C6,	0x55E5,
	0x55C9,	0x5712,	0x5713,	0x585E,	0x5851,	0x5858,	0x5857,	0x585A,
	0x5854,	0x586B,	0x584C,	0x586D,	0x584A,	0x5862,	0x5852,	0x584B,
	0x5967,	0x5AC1,	0x5AC9,	0x5ACC,	0x5ABE,	0x5ABD,	0x5ABC,

	/* Big5-HKSCS 0xB740..0xB77E, 0xB7A1..0xB7FE */
	0x5AB3,	0x5AC2,	0x5AB2,	0x5D69,	0x5D6F,	0x5E4C,	0x5E79,	0x5EC9,
	0x5EC8,	0x5F12,	0x5F59,	0x5FAC,	0x5FAE,	0x611A,	0x610F,	0x6148,
	0x611F,	0x60F3,	0x611B,	0x60F9,	0x6101,	0x6108,	0x614E,	0x614C,
	0x6144,	0x614D,	0x613E,	0x6134,	0x6127,	0x610D,	0x6106,	0x6137,
	0x6221,	0x6222,	0x6413,	0x643E,	0x641E,	0x642A,	0x642D,	0x643D,
	0x642C,	0x640F,	0x641C,	0x6414,	0x640D,	0x6436,	0x6416,	0x6417,
	0x6406,	0x656C,	0x659F,	0x65B0,	0x6697,	0x6689,	0x6687,	0x6688,
	0x6696,	0x6684,	0x6698,	0x668D,	0x6703,	0x6994,	0x696D,
		0x695A,	0x6977,	0x6960,	0x6954,	0x6975,	0x6930,	0x6982,
	0x694A,	0x6968,	0x696B,	0x695E,	0x6953,	0x6979,	0x6986,	0x695D,
	0x6963,	0x695B,	0x6B47,	0x6B72,	0x6BC0,	0x6BBF,	0x6BD3,	0x6BFD,
	0x6EA2,	0x6EAF,	0x6ED3,	0x6EB6,	0x6EC2,	0x6E90,	0x6E9D,	0x6EC7,
	0x6EC5,	0x6EA5,	0x6E98,	0x6EBC,	0x6EBA,	0x6EAB,	0x6ED1,	0x6E96,
	0x6E9C,	0x6EC4,	0x6ED4,	0x6EAA,	0x6EA7,	0x6EB4,	0x714E,	0x7159,
	0x7169,	0x7164,	0x7149,	0x7167,	0x715C,	0x716C,	0x7166,	0x714C,
	0x7165,	0x715E,	0x7146,	0x7168,	0x7156,	0x723A,	0x7252,	0x7337,
	0x7345,	0x733F,	0x733E,	0x746F,	0x745A,	0x7455,	0x745F,	0x745E,
	0x7441,	0x743F,	0x7459,	0x745B,	0x745C,	0x7576,	0x7578,	0x7600,
	0x75F0,	0x7601,	0x75F2,	0x75F1,	0x75FA,	0x75FF,	0x75F4,	0x75F3,
	0x76DE,	0x76DF,	0x775B,	0x776B,	0x7766,	0x775E,	0x7763,

	/* Big5-HKSCS 0xB840..0xB87E, 0xB8A1..0xB8FE */
	0x7779,	0x776A,	0x776C,	0x775C,	0x7765,	0x7768,	0x7762,	0x77EE,
	0x788E,	0x78B0,	0x7897,	0x7898,	0x788C,	0x7889,	0x787C,	0x7891,
	0x7893,	0x787F,	0x797A,	0x797F,	0x7981,	0x842C,	0x79BD,	0x7A1C,
	0x7A1A,	0x7A20,	0x7A14,	0x7A1F,	0x7A1E,	0x7A9F,	0x7AA0,	0x7B77,
	0x7BC0,	0x7B60,	0x7B6E,	0x7B67,	0x7CB1,	0x7CB3,	0x7CB5,	0x7D93,
	0x7D79,	0x7D91,	0x7D81,	0x7D8F,	0x7D5B,	0x7F6E,	0x7F69,	0x7F6A,
	0x7F72,	0x7FA9,	0x7FA8,	0x7FA4,	0x8056,	0x8058,	0x8086,	0x8084,
	0x8171,	0x8170,	0x8178,	0x8165,	0x816E,	0x8173,	0x816B,
		0x8179,	0x817A,	0x8166,	0x8205,	0x8247,	0x8482,	0x8477,
	0x843D,	0x8431,	0x8475,	0x8466,	0x846B,	0x8449,	0x846C,	0x845B,
	0x843C,	0x8435,	0x8461,	0x8463,	0x8469,	0x846D,	0x8446,	0x865E,
	0x865C,	0x865F,	0x86F9,	0x8713,	0x8708,	0x8707,	0x8700,	0x86FE,
	0x86FB,	0x8702,	0x8703,	0x8706,	0x870A,	0x8859,	0x88DF,	0x88D4,
	0x88D9,	0x88DC,	0x88D8,	0x88DD,	0x88E1,	0x88CA,	0x88D5,	0x88D2,
	0x899C,	0x89E3,	0x8A6B,	0x8A72,	0x8A73,	0x8A66,	0x8A69,	0x8A70,
	0x8A87,	0x8A7C,	0x8A63,	0x8AA0,	0x8A71,	0x8A85,	0x8A6D,	0x8A62,
	0x8A6E,	0x8A6C,	0x8A79,	0x8A7B,	0x8A3E,	0x8A68,	0x8C62,	0x8C8A,
	0x8C89,	0x8CCA,	0x8CC7,	0x8CC8,	0x8CC4,	0x8CB2,	0x8CC3,	0x8CC2,
	0x8CC5,	0x8DE1,	0x8DDF,	0x8DE8,	0x8DEF,	0x8DF3,	0x8DFA,	0x8DEA,
	0x8DE4,	0x8DE6,	0x8EB2,	0x8F03,	0x8F09,	0x8EFE,	0x8F0A,

	/* Big5-HKSCS 0xB940..0xB97E, 0xB9A1..0xB9FE */
	0x8F9F,	0x8FB2,	0x904B,	0x904A,	0x9053,	0x9042,	0x9054,	0x903C,
	0x9055,	0x9050,	0x9047,	0x904F,	0x904E,	0x904D,	0x9051,	0x903E,
	0x9041,	0x9112,	0x9117,	0x916C,	0x916A,	0x9169,	0x91C9,	0x9237,
	0x9257,	0x9238,	0x923D,	0x9240,	0x923E,	0x925B,	0x924B,	0x9264,
	0x9251,	0x9234,	0x9249,	0x924D,	0x9245,	0x9239,	0x923F,	0x925A,
	0x9598,	0x9698,	0x9694,	0x9695,	0x96CD,	0x96CB,	0x96C9,	0x96CA,
	0x96F7,	0x96FB,	0x96F9,	0x96F6,	0x9756,	0x9774,	0x9776,	0x9810,
	0x9811,	0x9813,	0x980A,	0x9812,	0x980C,	0x98FC,	0x98F4,
		0x98FD,	0x98FE,	0x99B3,	0x99B1,	0x99B4,	0x9AE1,	0x9CE9,
	0x9E82,	0x9F0E,	0x9F13,	0x9F20,	0x50E7,	0x50EE,	0x50E5,	0x50D6,
	0x50ED,	0x50DA,	0x50D5,	0x50CF,	0x50D1,	0x50F1,	0x50CE,	0x50E9,
	0x5162,	0x51F3,	0x5283,	0x5282,	0x5331,	0x53AD,	0x55FE,	0x5600,
	0x561B,	0x5617,	0x55FD,	0x5614,	0x5606,	0x5609,	0x560D,	0x560E,
	0x55F7,	0x5616,	0x561F,	0x5608,	0x5610,	0x55F6,	0x5718,	0x5716,
	0x5875,	0x587E,	0x5883,	0x5893,	0x588A,	0x5879,	0x5885,	0x587D,
	0x58FD,	0x5925,	0x5922,	0x5924,	0x596A,	0x5969,	0x5AE1,	0x5AE6,
	0x5AE9,	0x5AD7,	0x5AD6,	0x5AD8,	0x5AE3,	0x5B75,	0x5BDE,	0x5BE7,
	0x5BE1,	0x5BE5,	0x5BE6,	0x5BE8,	0x5BE2,	0x5BE4,	0x5BDF,	0x5C0D,
	0x5C62,	0x5D84,	0x5D87,	0x5E5B,	0x5E63,	0x5E55,	0x5E57,	0x5E54,
	0x5ED3,	0x5ED6,	0x5F0A,	0x5F46,	0x5F70,	0x5FB9,	0x6147,

	/* Big5-HKSCS 0xBA40..0xBA7E, 0xBAA1..0xBAFE */
	0x613F,	0x614B,	0x6177,	0x6162,	0x6163,	0x615F,	0x615A,	0x6158,
	0x6175,	0x622A,	0x6487,	0x6458,	0x6454,	0x64A4,	0x6478,	0x645F,
	0x647A,	0x6451,	0x6467,	0x6434,	0x646D,	0x647B,	0x6572,	0x65A1,
	0x65D7,	0x65D6,	0x66A2,	0x66A8,	0x669D,	0x699C,	0x69A8,	0x6995,
	0x69C1,	0x69AE,	0x69D3,	0x69CB,	0x699B,	0x69B7,	0x69BB,	0x69AB,
	0x69B4,	0x69D0,	0x69CD,	0x69AD,	0x69CC,	0x69A6,	0x69C3,	0x69A3,
	0x6B49,	0x6B4C,	0x6C33,	0x6F33,	0x6F14,	0x6EFE,	0x6F13,	0x6EF4,
	0x6F29,	0x6F3E,	0x6F20,	0x6F2C,	0x6F0F,	0x6F02,	0x6F22,
		0x6EFF,	0x6EEF,	0x6F06,	0x6F31,	0x6F38,	0x6F32,	0x6F23,
	0x6F15,	0x6F2B,	0x6F2F,	0x6F88,	0x6F2A,	0x6EEC,	0x6F01,	0x6EF2,
	0x6ECC,	0x6EF7,	0x7194,	0x7199,	0x717D,	0x718A,	0x7184,	0x7192,
	0x723E,	0x7292,	0x7296,	0x7344,	0x7350,	0x7464,	0x7463,	0x746A,
	0x7470,	0x746D,	0x7504,	0x7591,	0x7627,	0x760D,	0x760B,	0x7609,
	0x7613,	0x76E1,	0x76E3,	0x7784,	0x777D,	0x777F,	0x7761,	0x78C1,
	0x789F,	0x78A7,	0x78B3,	0x78A9,	0x78A3,	0x798E,	0x798F,	0x798D,
	0x7A2E,	0x7A31,	0x7AAA,	0x7AA9,	0x7AED,	0x7AEF,	0x7BA1,	0x7B95,
	0x7B8B,	0x7B75,	0x7B97,	0x7B9D,	0x7B94,	0x7B8F,	0x7BB8,	0x7B87,
	0x7B84,	0x7CB9,	0x7CBD,	0x7CBE,	0x7DBB,	0x7DB0,	0x7D9C,	0x7DBD,
	0x7DBE,	0x7DA0,	0x7DCA,	0x7DB4,	0x7DB2,	0x7DB1,	0x7DBA,	0x7DA2,
	0x7DBF,	0x7DB5,	0x7DB8,	0x7DAD,	0x7DD2,	0x7DC7,	0x7DAC,

	/* Big5-HKSCS 0xBB40..0xBB7E, 0xBBA1..0xBBFE */
	0x7F70,	0x7FE0,	0x7FE1,	0x7FDF,	0x805E,	0x805A,	0x8087,	0x8150,
	0x8180,	0x818F,	0x8188,	0x818A,	0x817F,	0x8182,	0x81E7,	0x81FA,
	0x8207,	0x8214,	0x821E,	0x824B,	0x84C9,	0x84BF,	0x84C6,	0x84C4,
	0x8499,	0x849E,	0x84B2,	0x849C,	0x84CB,	0x84B8,	0x84C0,	0x84D3,
	0x8490,	0x84BC,	0x84D1,	0x84CA,	0x873F,	0x871C,	0x873B,	0x8722,
	0x8725,	0x8734,	0x8718,	0x8755,	0x8737,	0x8729,	0x88F3,	0x8902,
	0x88F4,	0x88F9,	0x88F8,	0x88FD,	0x88E8,	0x891A,	0x88EF,	0x8AA6,
	0x8A8C,	0x8A9E,	0x8AA3,	0x8A8D,	0x8AA1,	0x8A93,	0x8AA4,
		0x8AAA,	0x8AA5,	0x8AA8,	0x8A98,	0x8A91,	0x8A9A,	0x8AA7,
	0x8C6A,	0x8C8D,	0x8C8C,	0x8CD3,	0x8CD1,	0x8CD2,	0x8D6B,	0x8D99,
	0x8D95,	0x8DFC,	0x8F14,	0x8F12,	0x8F15,	0x8F13,	0x8FA3,	0x9060,
	0x9058,	0x905C,	0x9063,	0x9059,	0x905E,	0x9062,	0x905D,	0x905B,
	0x9119,	0x9118,	0x911E,	0x9175,	0x9178,	0x9177,	0x9174,	0x9278,
	0x9280,	0x9285,	0x9298,	0x9296,	0x927B,	0x9293,	0x929C,	0x92A8,
	0x927C,	0x9291,	0x95A1,	0x95A8,	0x95A9,	0x95A3,	0x95A5,	0x95A4,
	0x9699,	0x969C,	0x969B,	0x96CC,	0x96D2,	0x9700,	0x977C,	0x9785,
	0x97F6,	0x9817,	0x9818,	0x98AF,	0x98B1,	0x9903,	0x9905,	0x990C,
	0x9909,	0x99C1,	0x9AAF,	0x9AB0,	0x9AE6,	0x9B41,	0x9B42,	0x9CF4,
	0x9CF6,	0x9CF3,	0x9EBC,	0x9F3B,	0x9F4A,	0x5104,	0x5100,	0x50FB,
	0x50F5,	0x50F9,	0x5102,	0x5108,	0x5109,	0x5105,	0x51DC,

	/* Big5-HKSCS 0xBC40..0xBC7E, 0xBCA1..0xBCFE */
	0x5287,	0x5288,	0x5289,	0x528D,	0x528A,	0x52F0,	0x53B2,	0x562E,
	0x563B,	0x5639,	0x5632,	0x563F,	0x5634,	0x5629,	0x5653,	0x564E,
	0x5657,	0x5674,	0x5636,	0x562F,	0x5630,	0x5880,	0x589F,	0x589E,
	0x58B3,	0x589C,	0x58AE,	0x58A9,	0x58A6,	0x596D,	0x5B09,	0x5AFB,
	0x5B0B,	0x5AF5,	0x5B0C,	0x5B08,	0x5BEE,	0x5BEC,	0x5BE9,	0x5BEB,
	0x5C64,	0x5C65,	0x5D9D,	0x5D94,	0x5E62,	0x5E5F,	0x5E61,	0x5EE2,
	0x5EDA,	0x5EDF,	0x5EDD,	0x5EE3,	0x5EE0,	0x5F48,	0x5F71,	0x5FB7,
	0x5FB5,	0x6176,	0x6167,	0x616E,	0x615D,	0x6155,	0x6182,
		0x617C,	0x6170,	0x616B,	0x617E,	0x61A7,	0x6190,	0x61AB,
	0x618E,	0x61AC,	0x619A,	0x61A4,	0x6194,	0x61AE,	0x622E,	0x6469,
	0x646F,	0x6479,	0x649E,	0x64B2,	0x6488,	0x6490,	0x64B0,	0x64A5,
	0x6493,	0x6495,	0x64A9,	0x6492,	0x64AE,	0x64AD,	0x64AB,	0x649A,
	0x64AC,	0x6499,	0x64A2,	0x64B3,	0x6575,	0x6577,	0x6578,	0x66AE,
	0x66AB,	0x66B4,	0x66B1,	0x6A23,	0x6A1F,	0x69E8,	0x6A01,	0x6A1E,
	0x6A19,	0x69FD,	0x6A21,	0x6A13,	0x6A0A,	0x69F3,	0x6A02,	0x6A05,
	0x69ED,	0x6A11,	0x6B50,	0x6B4E,	0x6BA4,	0x6BC5,	0x6BC6,	0x6F3F,
	0x6F7C,	0x6F84,	0x6F51,	0x6F66,	0x6F54,	0x6F86,	0x6F6D,	0x6F5B,
	0x6F78,	0x6F6E,	0x6F8E,	0x6F7A,	0x6F70,	0x6F64,	0x6F97,	0x6F58,
	0x6ED5,	0x6F6F,	0x6F60,	0x6F5F,	0x719F,	0x71AC,	0x71B1,	0x71A8,
	0x7256,	0x729B,	0x734E,	0x7357,	0x7469,	0x748B,	0x7483,

	/* Big5-HKSCS 0xBD40..0xBD7E, 0xBDA1..0xBDFE */
	0x747E,	0x7480,	0x757F,	0x7620,	0x7629,	0x761F,	0x7624,	0x7626,
	0x7621,	0x7622,	0x769A,	0x76BA,	0x76E4,	0x778E,	0x7787,	0x778C,
	0x7791,	0x778B,	0x78CB,	0x78C5,	0x78BA,	0x78CA,	0x78BE,	0x78D5,
	0x78BC,	0x78D0,	0x7A3F,	0x7A3C,	0x7A40,	0x7A3D,	0x7A37,	0x7A3B,
	0x7AAF,	0x7AAE,	0x7BAD,	0x7BB1,	0x7BC4,	0x7BB4,	0x7BC6,	0x7BC7,
	0x7BC1,	0x7BA0,	0x7BCC,	0x7CCA,	0x7DE0,	0x7DF4,	0x7DEF,	0x7DFB,
	0x7DD8,	0x7DEC,	0x7DDD,	0x7DE8,	0x7DE3,	0x7DDA,	0x7DDE,	0x7DE9,
	0x7D9E,	0x7DD9,	0x7DF2,	0x7DF9,	0x7F75,	0x7F77,	0x7FAF,
		0x7FE9,	0x8026,	0x819B,	0x819C,	0x819D,	0x81A0,	0x819A,
	0x8198,	0x8517,	0x853D,	0x851A,	0x84EE,	0x852C,	0x852D,	0x8513,
	0x8511,	0x8523,	0x8521,	0x8514,	0x84EC,	0x8525,	0x84FF,	0x8506,
	0x8782,	0x8774,	0x8776,	0x8760,	0x8766,	0x8778,	0x8768,	0x8759,
	0x8757,	0x874C,	0x8753,	0x885B,	0x885D,	0x8910,	0x8907,	0x8912,
	0x8913,	0x8915,	0x890A,	0x8ABC,	0x8AD2,	0x8AC7,	0x8AC4,	0x8A95,
	0x8ACB,	0x8AF8,	0x8AB2,	0x8AC9,	0x8AC2,	0x8ABF,	0x8AB0,	0x8AD6,
	0x8ACD,	0x8AB6,	0x8AB9,	0x8ADB,	0x8C4C,	0x8C4E,	0x8C6C,	0x8CE0,
	0x8CDE,	0x8CE6,	0x8CE4,	0x8CEC,	0x8CED,	0x8CE2,	0x8CE3,	0x8CDC,
	0x8CEA,	0x8CE1,	0x8D6D,	0x8D9F,	0x8DA3,	0x8E2B,	0x8E10,	0x8E1D,
	0x8E22,	0x8E0F,	0x8E29,	0x8E1F,	0x8E21,	0x8E1E,	0x8EBA,	0x8F1D,
	0x8F1B,	0x8F1F,	0x8F29,	0x8F26,	0x8F2A,	0x8F1C,	0x8F1E,

	/* Big5-HKSCS 0xBE40..0xBE7E, 0xBEA1..0xBEFE */
	0x8F25,	0x9069,	0x906E,	0x9068,	0x906D,	0x9077,	0x9130,	0x912D,
	0x9127,	0x9131,	0x9187,	0x9189,	0x918B,	0x9183,	0x92C5,	0x92BB,
	0x92B7,	0x92EA,	0x92AC,	0x92E4,	0x92C1,	0x92B3,	0x92BC,	0x92D2,
	0x92C7,	0x92F0,	0x92B2,	0x95AD,	0x95B1,	0x9704,	0x9706,	0x9707,
	0x9709,	0x9760,	0x978D,	0x978B,	0x978F,	0x9821,	0x982B,	0x981C,
	0x98B3,	0x990A,	0x9913,	0x9912,	0x9918,	0x99DD,	0x99D0,	0x99DF,
	0x99DB,	0x99D1,	0x99D5,	0x99D2,	0x99D9,	0x9AB7,	0x9AEE,	0x9AEF,
	0x9B27,	0x9B45,	0x9B44,	0x9B77,	0x9B6F,	0x9D06,	0x9D09,
		0x9D03,	0x9EA9,	0x9EBE,	0x9ECE,	0x58A8,	0x9F52,	0x5112,
	0x5118,	0x5114,	0x5110,	0x5115,	0x5180,	0x51AA,	0x51DD,	0x5291,
	0x5293,	0x52F3,	0x5659,	0x566B,	0x5679,	0x5669,	0x5664,	0x5678,
	0x566A,	0x5668,	0x5665,	0x5671,	0x566F,	0x566C,	0x5662,	0x5676,
	0x58C1,	0x58BE,	0x58C7,	0x58C5,	0x596E,	0x5B1D,	0x5B34,	0x5B78,
	0x5BF0,	0x5C0E,	0x5F4A,	0x61B2,	0x6191,	0x61A9,	0x618A,	0x61CD,
	0x61B6,	0x61BE,	0x61CA,	0x61C8,	0x6230,	0x64C5,	0x64C1,	0x64CB,
	0x64BB,	0x64BC,	0x64DA,	0x64C4,	0x64C7,	0x64C2,	0x64CD,	0x64BF,
	0x64D2,	0x64D4,	0x64BE,	0x6574,	0x66C6,	0x66C9,	0x66B9,	0x66C4,
	0x66C7,	0x66B8,	0x6A3D,	0x6A38,	0x6A3A,	0x6A59,	0x6A6B,	0x6A58,
	0x6A39,	0x6A44,	0x6A62,	0x6A61,	0x6A4B,	0x6A47,	0x6A35,	0x6A5F,
	0x6A48,	0x6B59,	0x6B77,	0x6C05,	0x6FC2,	0x6FB1,	0x6FA1,

	/* Big5-HKSCS 0xBF40..0xBF7E, 0xBFA1..0xBFFE */
	0x6FC3,	0x6FA4,	0x6FC1,	0x6FA7,	0x6FB3,	0x6FC0,	0x6FB9,	0x6FB6,
	0x6FA6,	0x6FA0,	0x6FB4,	0x71BE,	0x71C9,	0x71D0,	0x71D2,	0x71C8,
	0x71D5,	0x71B9,	0x71CE,	0x71D9,	0x71DC,	0x71C3,	0x71C4,	0x7368,
	0x749C,	0x74A3,	0x7498,	0x749F,	0x749E,	0x74E2,	0x750C,	0x750D,
	0x7634,	0x7638,	0x763A,	0x76E7,	0x76E5,	0x77A0,	0x779E,	0x779F,
	0x77A5,	0x78E8,	0x78DA,	0x78EC,	0x78E7,	0x79A6,	0x7A4D,	0x7A4E,
	0x7A46,	0x7A4C,	0x7A4B,	0x7ABA,	0x7BD9,	0x7C11,	0x7BC9,	0x7BE4,
	0x7BDB,	0x7BE1,	0x7BE9,	0x7BE6,	0x7CD5,	0x7CD6,	0x7E0A,
		0x7E11,	0x7E08,	0x7E1B,	0x7E23,	0x7E1E,	0x7E1D,	0x7E09,
	0x7E10,	0x7F79,	0x7FB2,	0x7FF0,	0x7FF1,	0x7FEE,	0x8028,	0x81B3,
	0x81A9,	0x81A8,	0x81FB,	0x8208,	0x8258,	0x8259,	0x854A,	0x8559,
	0x8548,	0x8568,	0x8569,	0x8543,	0x8549,	0x856D,	0x856A,	0x855E,
	0x8783,	0x879F,	0x879E,	0x87A2,	0x878D,	0x8861,	0x892A,	0x8932,
	0x8925,	0x892B,	0x8921,	0x89AA,	0x89A6,	0x8AE6,	0x8AFA,	0x8AEB,
	0x8AF1,	0x8B00,	0x8ADC,	0x8AE7,	0x8AEE,	0x8AFE,	0x8B01,	0x8B02,
	0x8AF7,	0x8AED,	0x8AF3,	0x8AF6,	0x8AFC,	0x8C6B,	0x8C6D,	0x8C93,
	0x8CF4,	0x8E44,	0x8E31,	0x8E34,	0x8E42,	0x8E39,	0x8E35,	0x8F3B,
	0x8F2F,	0x8F38,	0x8F33,	0x8FA8,	0x8FA6,	0x9075,	0x9074,	0x9078,
	0x9072,	0x907C,	0x907A,	0x9134,	0x9192,	0x9320,	0x9336,	0x92F8,
	0x9333,	0x932F,	0x9322,	0x92FC,	0x932B,	0x9304,	0x931A,

	/* Big5-HKSCS 0xC040..0xC07E, 0xC0A1..0xC0FE */
	0x9310,	0x9326,	0x9321,	0x9315,	0x932E,	0x9319,	0x95BB,	0x96A7,
	0x96A8,	0x96AA,	0x96D5,	0x970E,	0x9711,	0x9716,	0x970D,	0x9713,
	0x970F,	0x975B,	0x975C,	0x9766,	0x9798,	0x9830,	0x9838,	0x983B,
	0x9837,	0x982D,	0x9839,	0x9824,	0x9910,	0x9928,	0x991E,	0x991B,
	0x9921,	0x991A,	0x99ED,	0x99E2,	0x99F1,	0x9AB8,	0x9ABC,	0x9AFB,
	0x9AED,	0x9B28,	0x9B91,	0x9D15,	0x9D23,	0x9D26,	0x9D28,	0x9D12,
	0x9D1B,	0x9ED8,	0x9ED4,	0x9F8D,	0x9F9C,	0x512A,	0x511F,	0x5121,
	0x5132,	0x52F5,	0x568E,	0x5680,	0x5690,	0x5685,	0x5687,
		0x568F,	0x58D5,	0x58D3,	0x58D1,	0x58CE,	0x5B30,	0x5B2A,
	0x5B24,	0x5B7A,	0x5C37,	0x5C68,	0x5DBC,	0x5DBA,	0x5DBD,	0x5DB8,
	0x5E6B,	0x5F4C,	0x5FBD,	0x61C9,	0x61C2,	0x61C7,	0x61E6,	0x61CB,
	0x6232,	0x6234,	0x64CE,	0x64CA,	0x64D8,	0x64E0,	0x64F0,	0x64E6,
	0x64EC,	0x64F1,	0x64E2,	0x64ED,	0x6582,	0x6583,	0x66D9,	0x66D6,
	0x6A80,	0x6A94,	0x6A84,	0x6AA2,	0x6A9C,	0x6ADB,	0x6AA3,	0x6A7E,
	0x6A97,	0x6A90,	0x6AA0,	0x6B5C,	0x6BAE,	0x6BDA,	0x6C08,	0x6FD8,
	0x6FF1,	0x6FDF,	0x6FE0,	0x6FDB,	0x6FE4,	0x6FEB,	0x6FEF,	0x6F80,
	0x6FEC,	0x6FE1,	0x6FE9,	0x6FD5,	0x6FEE,	0x6FF0,	0x71E7,	0x71DF,
	0x71EE,	0x71E6,	0x71E5,	0x71ED,	0x71EC,	0x71F4,	0x71E0,	0x7235,
	0x7246,	0x7370,	0x7372,	0x74A9,	0x74B0,	0x74A6,	0x74A8,	0x7646,
	0x7642,	0x764C,	0x76EA,	0x77B3,	0x77AA,	0x77B0,	0x77AC,

	/* Big5-HKSCS 0xC140..0xC17E, 0xC1A1..0xC1FE */
	0x77A7,	0x77AD,	0x77EF,	0x78F7,	0x78FA,	0x78F4,	0x78EF,	0x7901,
	0x79A7,	0x79AA,	0x7A57,	0x7ABF,	0x7C07,	0x7C0D,	0x7BFE,	0x7BF7,
	0x7C0C,	0x7BE0,	0x7CE0,	0x7CDC,	0x7CDE,	0x7CE2,	0x7CDF,	0x7CD9,
	0x7CDD,	0x7E2E,	0x7E3E,	0x7E46,	0x7E37,	0x7E32,	0x7E43,	0x7E2B,
	0x7E3D,	0x7E31,	0x7E45,	0x7E41,	0x7E34,	0x7E39,	0x7E48,	0x7E35,
	0x7E3F,	0x7E2F,	0x7F44,	0x7FF3,	0x7FFC,	0x8071,	0x8072,	0x8070,
	0x806F,	0x8073,	0x81C6,	0x81C3,	0x81BA,	0x81C2,	0x81C0,	0x81BF,
	0x81BD,	0x81C9,	0x81BE,	0x81E8,	0x8209,	0x8271,	0x85AA,
		0x8584,	0x857E,	0x859C,	0x8591,	0x8594,	0x85AF,	0x859B,
	0x8587,	0x85A8,	0x858A,	0x8667,	0x87C0,	0x87D1,	0x87B3,	0x87D2,
	0x87C6,	0x87AB,	0x87BB,	0x87BA,	0x87C8,	0x87CB,	0x893B,	0x8936,
	0x8944,	0x8938,	0x893D,	0x89AC,	0x8B0E,	0x8B17,	0x8B19,	0x8B1B,
	0x8B0A,	0x8B20,	0x8B1D,	0x8B04,	0x8B10,	0x8C41,	0x8C3F,	0x8C73,
	0x8CFA,	0x8CFD,	0x8CFC,	0x8CF8,	0x8CFB,	0x8DA8,	0x8E49,	0x8E4B,
	0x8E48,	0x8E4A,	0x8F44,	0x8F3E,	0x8F42,	0x8F45,	0x8F3F,	0x907F,
	0x907D,	0x9084,	0x9081,	0x9082,	0x9080,	0x9139,	0x91A3,	0x919E,
	0x919C,	0x934D,	0x9382,	0x9328,	0x9375,	0x934A,	0x9365,	0x934B,
	0x9318,	0x937E,	0x936C,	0x935B,	0x9370,	0x935A,	0x9354,	0x95CA,
	0x95CB,	0x95CC,	0x95C8,	0x95C6,	0x96B1,	0x96B8,	0x96D6,	0x971C,
	0x971E,	0x97A0,	0x97D3,	0x9846,	0x98B6,	0x9935,	0x9A01,

	/* Big5-HKSCS 0xC240..0xC27E, 0xC2A1..0xC2FE */
	0x99FF,	0x9BAE,	0x9BAB,	0x9BAA,	0x9BAD,	0x9D3B,	0x9D3F,	0x9E8B,
	0x9ECF,	0x9EDE,	0x9EDC,	0x9EDD,	0x9EDB,	0x9F3E,	0x9F4B,	0x53E2,
	0x5695,	0x56AE,	0x58D9,	0x58D8,	0x5B38,	0x5F5D,	0x61E3,	0x6233,
	0x64F4,	0x64F2,	0x64FE,	0x6506,	0x64FA,	0x64FB,	0x64F7,	0x65B7,
	0x66DC,	0x6726,	0x6AB3,	0x6AAC,	0x6AC3,	0x6ABB,	0x6AB8,	0x6AC2,
	0x6AAE,	0x6AAF,	0x6B5F,	0x6B78,	0x6BAF,	0x7009,	0x700B,	0x6FFE,
	0x7006,	0x6FFA,	0x7011,	0x700F,	0x71FB,	0x71FC,	0x71FE,	0x71F8,
	0x7377,	0x7375,	0x74A7,	0x74BF,	0x7515,	0x7656,	0x7658,
		0x7652,	0x77BD,	0x77BF,	0x77BB,	0x77BC,	0x790E,	0x79AE,
	0x7A61,	0x7A62,	0x7A60,	0x7AC4,	0x7AC5,	0x7C2B,	0x7C27,	0x7C2A,
	0x7C1E,	0x7C23,	0x7C21,	0x7CE7,	0x7E54,	0x7E55,	0x7E5E,	0x7E5A,
	0x7E61,	0x7E52,	0x7E59,	0x7F48,	0x7FF9,	0x7FFB,	0x8077,	0x8076,
	0x81CD,	0x81CF,	0x820A,	0x85CF,	0x85A9,	0x85CD,	0x85D0,	0x85C9,
	0x85B0,	0x85BA,	0x85B9,	0x85A6,	0x87EF,	0x87EC,	0x87F2,	0x87E0,
	0x8986,	0x89B2,	0x89F4,	0x8B28,	0x8B39,	0x8B2C,	0x8B2B,	0x8C50,
	0x8D05,	0x8E59,	0x8E63,	0x8E66,	0x8E64,	0x8E5F,	0x8E55,	0x8EC0,
	0x8F49,	0x8F4D,	0x9087,	0x9083,	0x9088,	0x91AB,	0x91AC,	0x91D0,
	0x9394,	0x938A,	0x9396,	0x93A2,	0x93B3,	0x93AE,	0x93AC,	0x93B0,
	0x9398,	0x939A,	0x9397,	0x95D4,	0x95D6,	0x95D0,	0x95D5,	0x96E2,
	0x96DC,	0x96D9,	0x96DB,	0x96DE,	0x9724,	0x97A3,	0x97A6,

	/* Big5-HKSCS 0xC340..0xC37E, 0xC3A1..0xC3FE */
	0x97AD,	0x97F9,	0x984D,	0x984F,	0x984C,	0x984E,	0x9853,	0x98BA,
	0x993E,	0x993F,	0x993D,	0x992E,	0x99A5,	0x9A0E,	0x9AC1,	0x9B03,
	0x9B06,	0x9B4F,	0x9B4E,	0x9B4D,	0x9BCA,	0x9BC9,	0x9BFD,	0x9BC8,
	0x9BC0,	0x9D51,	0x9D5D,	0x9D60,	0x9EE0,	0x9F15,	0x9F2C,	0x5133,
	0x56A5,	0x58DE,	0x58DF,	0x58E2,	0x5BF5,	0x9F90,	0x5EEC,	0x61F2,
	0x61F7,	0x61F6,	0x61F5,	0x6500,	0x650F,	0x66E0,	0x66DD,	0x6AE5,
	0x6ADD,	0x6ADA,	0x6AD3,	0x701B,	0x701F,	0x7028,	0x701A,	0x701D,
	0x7015,	0x7018,	0x7206,	0x720D,	0x7258,	0x72A2,	0x7378,
		0x737A,	0x74BD,	0x74CA,	0x74E3,	0x7587,	0x7586,	0x765F,
	0x7661,	0x77C7,	0x7919,	0x79B1,	0x7A6B,	0x7A69,	0x7C3E,	0x7C3F,
	0x7C38,	0x7C3D,	0x7C37,	0x7C40,	0x7E6B,	0x7E6D,	0x7E79,	0x7E69,
	0x7E6A,	0x7F85,	0x7E73,	0x7FB6,	0x7FB9,	0x7FB8,	0x81D8,	0x85E9,
	0x85DD,	0x85EA,	0x85D5,	0x85E4,	0x85E5,	0x85F7,	0x87FB,	0x8805,
	0x880D,	0x87F9,	0x87FE,	0x8960,	0x895F,	0x8956,	0x895E,	0x8B41,
	0x8B5C,	0x8B58,	0x8B49,	0x8B5A,	0x8B4E,	0x8B4F,	0x8B46,	0x8B59,
	0x8D08,	0x8D0A,	0x8E7C,	0x8E72,	0x8E87,	0x8E76,	0x8E6C,	0x8E7A,
	0x8E74,	0x8F54,	0x8F4E,	0x8FAD,	0x908A,	0x908B,	0x91B1,	0x91AE,
	0x93E1,	0x93D1,	0x93DF,	0x93C3,	0x93C8,	0x93DC,	0x93DD,	0x93D6,
	0x93E2,	0x93CD,	0x93D8,	0x93E4,	0x93D7,	0x93E8,	0x95DC,	0x96B4,
	0x96E3,	0x972A,	0x9727,	0x9761,	0x97DC,	0x97FB,	0x985E,

	/* Big5-HKSCS 0xC440..0xC47E, 0xC4A1..0xC4FE */
	0x9858,	0x985B,	0x98BC,	0x9945,	0x9949,	0x9A16,	0x9A19,	0x9B0D,
	0x9BE8,	0x9BE7,	0x9BD6,	0x9BDB,	0x9D89,	0x9D61,	0x9D72,	0x9D6A,
	0x9D6C,	0x9E92,	0x9E97,	0x9E93,	0x9EB4,	0x52F8,	0x56A8,	0x56B7,
	0x56B6,	0x56B4,	0x56BC,	0x58E4,	0x5B40,	0x5B43,	0x5B7D,	0x5BF6,
	0x5DC9,	0x61F8,	0x61FA,	0x6518,	0x6514,	0x6519,	0x66E6,	0x6727,
	0x6AEC,	0x703E,	0x7030,	0x7032,	0x7210,	0x737B,	0x74CF,	0x7662,
	0x7665,	0x7926,	0x792A,	0x792C,	0x792B,	0x7AC7,	0x7AF6,	0x7C4C,
	0x7C43,	0x7C4D,	0x7CEF,	0x7CF0,	0x8FAE,	0x7E7D,	0x7E7C,
		0x7E82,	0x7F4C,	0x8000,	0x81DA,	0x8266,	0x85FB,	0x85F9,
	0x8611,	0x85FA,	0x8606,	0x860B,	0x8607,	0x860A,	0x8814,	0x8815,
	0x8964,	0x89BA,	0x89F8,	0x8B70,	0x8B6C,	0x8B66,	0x8B6F,	0x8B5F,
	0x8B6B,	0x8D0F,	0x8D0D,	0x8E89,	0x8E81,	0x8E85,	0x8E82,	0x91B4,
	0x91CB,	0x9418,	0x9403,	0x93FD,	0x95E1,	0x9730,	0x98C4,	0x9952,
	0x9951,	0x99A8,	0x9A2B,	0x9A30,	0x9A37,	0x9A35,	0x9C13,	0x9C0D,
	0x9E79,	0x9EB5,	0x9EE8,	0x9F2F,	0x9F5F,	0x9F63,	0x9F61,	0x5137,
	0x5138,	0x56C1,	0x56C0,	0x56C2,	0x5914,	0x5C6C,	0x5DCD,	0x61FC,
	0x61FE,	0x651D,	0x651C,	0x6595,	0x66E9,	0x6AFB,	0x6B04,	0x6AFA,
	0x6BB2,	0x704C,	0x721B,	0x72A7,	0x74D6,	0x74D4,	0x7669,	0x77D3,
	0x7C50,	0x7E8F,	0x7E8C,	0x7FBC,	0x8617,	0x862D,	0x861A,	0x8823,
	0x8822,	0x8821,	0x881F,	0x896A,	0x896C,	0x89BD,	0x8B74,

	/* Big5-HKSCS 0xC540..0xC57E, 0xC5A1..0xC5FE */
	0x8B77,	0x8B7D,	0x8D13,	0x8E8A,	0x8E8D,	0x8E8B,	0x8F5F,	0x8FAF,
	0x91BA,	0x942E,	0x9433,	0x9435,	0x943A,	0x9438,	0x9432,	0x942B,
	0x95E2,	0x9738,	0x9739,	0x9732,	0x97FF,	0x9867,	0x9865,	0x9957,
	0x9A45,	0x9A43,	0x9A40,	0x9A3E,	0x9ACF,	0x9B54,	0x9B51,	0x9C2D,
	0x9C25,	0x9DAF,	0x9DB4,	0x9DC2,	0x9DB8,	0x9E9D,	0x9EEF,	0x9F19,
	0x9F5C,	0x9F66,	0x9F67,	0x513C,	0x513B,	0x56C8,	0x56CA,	0x56C9,
	0x5B7F,	0x5DD4,	0x5DD2,	0x5F4E,	0x61FF,	0x6524,	0x6B0A,	0x6B61,
	0x7051,	0x7058,	0x7380,	0x74E4,	0x758A,	0x766E,	0x766C,
		0x79B3,	0x7C60,	0x7C5F,	0x807E,	0x807D,	0x81DF,	0x8972,
	0x896F,	0x89FC,	0x8B80,	0x8D16,	0x8D17,	0x8E91,	0x8E93,	0x8F61,
	0x9148,	0x9444,	0x9451,	0x9452,	0x973D,	0x973E,	0x97C3,	0x97C1,
	0x986B,	0x9955,	0x9A55,	0x9A4D,	0x9AD2,	0x9B1A,	0x9C49,	0x9C31,
	0x9C3E,	0x9C3B,	0x9DD3,	0x9DD7,	0x9F34,	0x9F6C,	0x9F6A,	0x9F94,
	0x56CC,	0x5DD6,	0x6200,	0x6523,	0x652B,	0x652A,	0x66EC,	0x6B10,
	0x74DA,	0x7ACA,	0x7C64,	0x7C63,	0x7C65,	0x7E93,	0x7E96,	0x7E94,
	0x81E2,	0x8638,	0x863F,	0x8831,	0x8B8A,	0x9090,	0x908F,	0x9463,
	0x9460,	0x9464,	0x9768,	0x986F,	0x995C,	0x9A5A,	0x9A5B,	0x9A57,
	0x9AD3,	0x9AD4,	0x9AD1,	0x9C54,	0x9C57,	0x9C56,	0x9DE5,	0x9E9F,
	0x9EF4,	0x56D1,	0x58E9,	0x652C,	0x705E,	0x7671,	0x7672,	0x77D7,
	0x7F50,	0x7F88,	0x8836,	0x8839,	0x8862,	0x8B93,	0x8B92,

	/* Big5-HKSCS 0xC640..0xC67E, 0xC6A1..0xC6FE */
	0x8B96,	0x8277,	0x8D1B,	0x91C0,	0x946A,	0x9742,	0x9748,	0x9744,
	0x97C6,	0x9870,	0x9A5F,	0x9B22,	0x9B58,	0x9C5F,	0x9DF9,	0x9DFA,
	0x9E7C,	0x9E7D,	0x9F07,	0x9F77,	0x9F72,	0x5EF3,	0x6B16,	0x7063,
	0x7C6C,	0x7C6E,	0x883B,	0x89C0,	0x8EA1,	0x91C1,	0x9472,	0x9470,
	0x9871,	0x995E,	0x9AD6,	0x9B23,	0x9ECC,	0x7064,	0x77DA,	0x8B9A,
	0x9477,	0x97C9,	0x9A62,	0x9A65,	0x7E9C,	0x8B9C,	0x8EAA,	0x91C5,
	0x947D,	0x947E,	0x947C,	0x9C77,	0x9C78,	0x9EF7,	0x8C54,	0x947F,
	0x9E1A,	0x7228,	0x9A6A,	0x9B31,	0x9E1B,	0x9E1E,	0x7C72,
		0x2460,	0x2461,	0x2462,	0x2463,	0x2464,	0x2465,	0x2466,
	0x2467,	0x2468,	0x2469,	0x2474,	0x2475,	0x2476,	0x2477,	0x2478,
	0x2479,	0x247A,	0x247B,	0x247C,	0x247D,	0x2170,	0x2171,	0x2172,
	0x2173,	0x2174,	0x2175,	0x2176,	0x2177,	0x2178,	0x2179,	0x4E36,
	0x4E3F,	0x4E85,	0x4EA0,	0x5182,	0x5196,	0x51AB,	0x52F9,	0x5338,
	0x5369,	0x53B6,	0x590A,	0x5B80,	0x5DDB,	0x2F33,	0x5E7F,	0xF6DF,
	0x5F50,	0x5F61,	0x6534,	0xF6E3,	0x7592,	0xF6E5,	0x8FB5,	0xF6E7,
	0x00A8,	0x02C6,	0x30FD,	0x30FE,	0x309D,	0x309E,	0xF6EE,	0xF6EF,
	0x3005,	0x3006,	0x3007,	0x30FC,	0xFF3B,	0xFF3D,	0x273D,	0x3041,
	0x3042,	0x3043,	0x3044,	0x3045,	0x3046,	0x3047,	0x3048,	0x3049,
	0x304A,	0x304B,	0x304C,	0x304D,	0x304E,	0x304F,	0x3050,	0x3051,
	0x3052,	0x3053,	0x3054,	0x3055,	0x3056,	0x3057,	0x3058,

	/* Big5-HKSCS 0xC740..0xC77E, 0xC7A1..0xC7FE */
	0x3059,	0x305A,	0x305B,	0x305C,	0x305D,	0x305E,	0x305F,	0x3060,
	0x3061,	0x3062,	0x3063,	0x3064,	0x3065,	0x3066,	0x3067,	0x3068,
	0x3069,	0x306A,	0x306B,	0x306C,	0x306D,	0x306E,	0x306F,	0x3070,
	0x3071,	0x3072,	0x3073,	0x3074,	0x3075,	0x3076,	0x3077,	0x3078,
	0x3079,	0x307A,	0x307B,	0x307C,	0x307D,	0x307E,	0x307F,	0x3080,
	0x3081,	0x3082,	0x3083,	0x3084,	0x3085,	0x3086,	0x3087,	0x3088,
	0x3089,	0x308A,	0x308B,	0x308C,	0x308D,	0x308E,	0x308F,	0x3090,
	0x3091,	0x3092,	0x3093,	0x30A1,	0x30A2,	0x30A3,	0x30A4,
		0x30A5,	0x30A6,	0x30A7,	0x30A8,	0x30A9,	0x30AA,	0x30AB,
	0x30AC,	0x30AD,	0x30AE,	0x30AF,	0x30B0,	0x30B1,	0x30B2,	0x30B3,
	0x30B4,	0x30B5,	0x30B6,	0x30B7,	0x30B8,	0x30B9,	0x30BA,	0x30BB,
	0x30BC,	0x30BD,	0x30BE,	0x30BF,	0x30C0,	0x30C1,	0x30C2,	0x30C3,
	0x30C4,	0x30C5,	0x30C6,	0x30C7,	0x30C8,	0x30C9,	0x30CA,	0x30CB,
	0x30CC,	0x30CD,	0x30CE,	0x30CF,	0x30D0,	0x30D1,	0x30D2,	0x30D3,
	0x30D4,	0x30D5,	0x30D6,	0x30D7,	0x30D8,	0x30D9,	0x30DA,	0x30DB,
	0x30DC,	0x30DD,	0x30DE,	0x30DF,	0x30E0,	0x30E1,	0x30E2,	0x30E3,
	0x30E4,	0x30E5,	0x30E6,	0x30E7,	0x30E8,	0x30E9,	0x30EA,	0x30EB,
	0x30EC,	0x30ED,	0x30EE,	0x30EF,	0x30F0,	0x30F1,	0x30F2,	0x30F3,
	0x30F4,	0x30F5,	0x30F6,	0x0410,	0x0411,	0x0412,	0x0413,	0x0414,
	0x0415,	0x0401,	0x0416,	0x0417,	0x0418,	0x0419,	0x041A,

	/* Big5-HKSCS 0xC840..0xC87E, 0xC8A1..0xC8FE */
	0x041B,	0x041C,	0x041D,	0x041E,	0x041F,	0x0420,	0x0421,	0x0422,
	0x0423,	0x0424,	0x0425,	0x0426,	0x0427,	0x0428,	0x0429,	0x042A,
	0x042B,	0x042C,	0x042D,	0x042E,	0x042F,	0x0430,	0x0431,	0x0432,
	0x0433,	0x0434,	0x0435,	0x0451,	0x0436,	0x0437,	0x0438,	0x0439,
	0x043A,	0x043B,	0x043C,	0x043D,	0x043E,	0x043F,	0x0440,	0x0441,
	0x0442,	0x0443,	0x0444,	0x0445,	0x0446,	0x0447,	0x0448,	0x0449,
	0x044A,	0x044B,	0x044C,	0x044D,	0x044E,	0x044F,	0x21E7,	0x21B8,
	0x21B9,	0xF7E5,	0xF7E6,	0x4E5A,	0xF7E8,	0x5202,	0xF7EA,
		0xF7EB,	0x5188,	0xF7ED,	0xF7EE,	0xF7EF,	0xF7F0,	0xF7F1,
	0xF7F2,	0xF7F3,	0xF7F4,	0xF7F5,	0xF7F6,	0xF7F7,	0xF7F8,	0xF7F9,
	0xF7FA,	0xF7FB,	0xF7FC,	0xF7FD,	0xF7FE,	0xF7FF,	0xF800,	0xF801,
	0xF802,	0xF803,	0xF804,	0xF805,	0xF806,	0xF807,	0xF808,	0xF809,
	0xF80A,	0xF80B,	0xF80C,	0xF80D,	0xF80E,	0xF80F,	0xF810,	0xF811,
	0xF812,	0xF813,	0xF814,	0xF815,	0xF816,	0xFFE2,	0xFFE4,	0xFF07,
	0xFF02,	0x3231,	0x2116,	0x2121,	0x309B,	0x309C,	0x2E80,	0x2E84,
	0x2E86,	0x2E87,	0x2E88,	0x2E8A,	0x2E8C,	0x2E8D,	0x2E95,	0x2E9C,
	0x2E9D,	0x2EA5,	0x2EA7,	0x2EAA,	0x2EAC,	0x2EAE,	0x2EB6,	0x2EBC,
	0x2EBE,	0x2EC6,	0x2ECA,	0x2ECC,	0x2ECD,	0x2ECF,	0x2ED6,	0x2ED7,
	0x2EDE,	0x2EE3,	0xF83C,	0xF83D,	0xF83E,	0x0283,	0x0250,	0x025B,
	0x0254,	0x0275,	0x0153,	0x00F8,	0x014B,	0x028A,	0x026A,

	/* Big5-HKSCS 0xC940..0xC97E, 0xC9A1..0xC9FE */
	0x4E42,	0x4E5C,	0x51F5,	0x531A,	0x5382,	0x4E07,	0x4E0C,	0x4E47,
	0x4E8D,	0x56D7,	0xFA0C,	0x5C6E,	0x5F73,	0x4E0F,	0x5187,	0x4E0E,
	0x4E2E,	0x4E93,	0x4EC2,	0x4EC9,	0x4EC8,	0x5198,	0x52FC,	0x536C,
	0x53B9,	0x5720,	0x5903,	0x592C,	0x5C10,	0x5DFF,	0x65E1,	0x6BB3,
	0x6BCC,	0x6C14,	0x723F,	0x4E31,	0x4E3C,	0x4EE8,	0x4EDC,	0x4EE9,
	0x4EE1,	0x4EDD,	0x4EDA,	0x520C,	0x531C,	0x534C,	0x5722,	0x5723,
	0x5917,	0x592F,	0x5B81,	0x5B84,	0x5C12,	0x5C3B,	0x5C74,	0x5C73,
	0x5E04,	0x5E80,	0x5E82,	0x5FC9,	0x6209,	0x6250,	0x6C15,
		0x6C36,	0x6C43,	0x6C3F,	0x6C3B,	0x72AE,	0x72B0,	0x738A,
	0x79B8,	0x808A,	0x961E,	0x4F0E,	0x4F18,	0x4F2C,	0x4EF5,	0x4F14,
	0x4EF1,	0x4F00,	0x4EF7,	0x4F08,	0x4F1D,	0x4F02,	0x4F05,	0x4F22,
	0x4F13,	0x4F04,	0x4EF4,	0x4F12,	0x51B1,	0x5213,	0x5209,	0x5210,
	0x52A6,	0x5322,	0x531F,	0x534D,	0x538A,	0x5407,	0x56E1,	0x56DF,
	0x572E,	0x572A,	0x5734,	0x593C,	0x5980,	0x597C,	0x5985,	0x597B,
	0x597E,	0x5977,	0x597F,	0x5B56,	0x5C15,	0x5C25,	0x5C7C,	0x5C7A,
	0x5C7B,	0x5C7E,	0x5DDF,	0x5E75,	0x5E84,	0x5F02,	0x5F1A,	0x5F74,
	0x5FD5,	0x5FD4,	0x5FCF,	0x625C,	0x625E,	0x6264,	0x6261,	0x6266,
	0x6262,	0x6259,	0x6260,	0x625A,	0x6265,	0x65EF,	0x65EE,	0x673E,
	0x6739,	0x6738,	0x673B,	0x673A,	0x673F,	0x673C,	0x6733,	0x6C18,
	0x6C46,	0x6C52,	0x6C5C,	0x6C4F,	0x6C4A,	0x6C54,	0x6C4B,

	/* Big5-HKSCS 0xCA40..0xCA7E, 0xCAA1..0xCAFE */
	0x6C4C,	0x7071,	0x725E,	0x72B4,	0x72B5,	0x738E,	0x752A,	0x767F,
	0x7A75,	0x7F51,	0x8278,	0x827C,	0x8280,	0x827D,	0x827F,	0x864D,
	0x897E,	0x9099,	0x9097,	0x9098,	0x909B,	0x9094,	0x9622,	0x9624,
	0x9620,	0x9623,	0x4F56,	0x4F3B,	0x4F62,	0x4F49,	0x4F53,	0x4F64,
	0x4F3E,	0x4F67,	0x4F52,	0x4F5F,	0x4F41,	0x4F58,	0x4F2D,	0x4F33,
	0x4F3F,	0x4F61,	0x518F,	0x51B9,	0x521C,	0x521E,	0x5221,	0x52AD,
	0x52AE,	0x5309,	0x5363,	0x5372,	0x538E,	0x538F,	0x5430,	0x5437,
	0x542A,	0x5454,	0x5445,	0x5419,	0x541C,	0x5425,	0x5418,
		0x543D,	0x544F,	0x5441,	0x5428,	0x5424,	0x5447,	0x56EE,
	0x56E7,	0x56E5,	0x5741,	0x5745,	0x574C,	0x5749,	0x574B,	0x5752,
	0x5906,	0x5940,	0x59A6,	0x5998,	0x59A0,	0x5997,	0x598E,	0x59A2,
	0x5990,	0x598F,	0x59A7,	0x59A1,	0x5B8E,	0x5B92,	0x5C28,	0x5C2A,
	0x5C8D,	0x5C8F,	0x5C88,	0x5C8B,	0x5C89,	0x5C92,	0x5C8A,	0x5C86,
	0x5C93,	0x5C95,	0x5DE0,	0x5E0A,	0x5E0E,	0x5E8B,	0x5E89,	0x5E8C,
	0x5E88,	0x5E8D,	0x5F05,	0x5F1D,	0x5F78,	0x5F76,	0x5FD2,	0x5FD1,
	0x5FD0,	0x5FED,	0x5FE8,	0x5FEE,	0x5FF3,	0x5FE1,	0x5FE4,	0x5FE3,
	0x5FFA,	0x5FEF,	0x5FF7,	0x5FFB,	0x6000,	0x5FF4,	0x623A,	0x6283,
	0x628C,	0x628E,	0x628F,	0x6294,	0x6287,	0x6271,	0x627B,	0x627A,
	0x6270,	0x6281,	0x6288,	0x6277,	0x627D,	0x6272,	0x6274,	0x6537,
	0x65F0,	0x65F4,	0x65F3,	0x65F2,	0x65F5,	0x6745,	0x6747,

	/* Big5-HKSCS 0xCB40..0xCB7E, 0xCBA1..0xCBFE */
	0x6759,	0x6755,	0x674C,	0x6748,	0x675D,	0x674D,	0x675A,	0x674B,
	0x6BD0,	0x6C19,	0x6C1A,	0x6C78,	0x6C67,	0x6C6B,	0x6C84,	0x6C8B,
	0x6C8F,	0x6C71,	0x6C6F,	0x6C69,	0x6C9A,	0x6C6D,	0x6C87,	0x6C95,
	0x6C9C,	0x6C66,	0x6C73,	0x6C65,	0x6C7B,	0x6C8E,	0x7074,	0x707A,
	0x7263,	0x72BF,	0x72BD,	0x72C3,	0x72C6,	0x72C1,	0x72BA,	0x72C5,
	0x7395,	0x7397,	0x7393,	0x7394,	0x7392,	0x753A,	0x7539,	0x7594,
	0x7595,	0x7681,	0x793D,	0x8034,	0x8095,	0x8099,	0x8090,	0x8092,
	0x809C,	0x8290,	0x828F,	0x8285,	0x828E,	0x8291,	0x8293,
		0x828A,	0x8283,	0x8284,	0x8C78,	0x8FC9,	0x8FBF,	0x909F,
	0x90A1,	0x90A5,	0x909E,	0x90A7,	0x90A0,	0x9630,	0x9628,	0x962F,
	0x962D,	0x4E33,	0x4F98,	0x4F7C,	0x4F85,	0x4F7D,	0x4F80,	0x4F87,
	0x4F76,	0x4F74,	0x4F89,	0x4F84,	0x4F77,	0x4F4C,	0x4F97,	0x4F6A,
	0x4F9A,	0x4F79,	0x4F81,	0x4F78,	0x4F90,	0x4F9C,	0x4F94,	0x4F9E,
	0x4F92,	0x4F82,	0x4F95,	0x4F6B,	0x4F6E,	0x519E,	0x51BC,	0x51BE,
	0x5235,	0x5232,	0x5233,	0x5246,	0x5231,	0x52BC,	0x530A,	0x530B,
	0x533C,	0x5392,	0x5394,	0x5487,	0x547F,	0x5481,	0x5491,	0x5482,
	0x5488,	0x546B,	0x547A,	0x547E,	0x5465,	0x546C,	0x5474,	0x5466,
	0x548D,	0x546F,	0x5461,	0x5460,	0x5498,	0x5463,	0x5467,	0x5464,
	0x56F7,	0x56F9,	0x576F,	0x5772,	0x576D,	0x576B,	0x5771,	0x5770,
	0x5776,	0x5780,	0x5775,	0x577B,	0x5773,	0x5774,	0x5762,

	/* Big5-HKSCS 0xCC40..0xCC7E, 0xCCA1..0xCCFE */
	0x5768,	0x577D,	0x590C,	0x5945,	0x59B5,	0x59BA,	0x59CF,	0x59CE,
	0x59B2,	0x59CC,	0x59C1,	0x59B6,	0x59BC,	0x59C3,	0x59D6,	0x59B1,
	0x59BD,	0x59C0,	0x59C8,	0x59B4,	0x59C7,	0x5B62,	0x5B65,	0x5B93,
	0x5B95,	0x5C44,	0x5C47,	0x5CAE,	0x5CA4,	0x5CA0,	0x5CB5,	0x5CAF,
	0x5CA8,	0x5CAC,	0x5C9F,	0x5CA3,	0x5CAD,	0x5CA2,	0x5CAA,	0x5CA7,
	0x5C9D,	0x5CA5,	0x5CB6,	0x5CB0,	0x5CA6,	0x5E17,	0x5E14,	0x5E19,
	0x5F28,	0x5F22,	0x5F23,	0x5F24,	0x5F54,	0x5F82,	0x5F7E,	0x5F7D,
	0x5FDE,	0x5FE5,	0x602D,	0x6026,	0x6019,	0x6032,	0x600B,
		0x6034,	0x600A,	0x6017,	0x6033,	0x601A,	0x601E,	0x602C,
	0x6022,	0x600D,	0x6010,	0x602E,	0x6013,	0x6011,	0x600C,	0x6009,
	0x601C,	0x6214,	0x623D,	0x62AD,	0x62B4,	0x62D1,	0x62BE,	0x62AA,
	0x62B6,	0x62CA,	0x62AE,	0x62B3,	0x62AF,	0x62BB,	0x62A9,	0x62B0,
	0x62B8,	0x653D,	0x65A8,	0x65BB,	0x6609,	0x65FC,	0x6604,	0x6612,
	0x6608,	0x65FB,	0x6603,	0x660B,	0x660D,	0x6605,	0x65FD,	0x6611,
	0x6610,	0x66F6,	0x670A,	0x6785,	0x676C,	0x678E,	0x6792,	0x6776,
	0x677B,	0x6798,	0x6786,	0x6784,	0x6774,	0x678D,	0x678C,	0x677A,
	0x679F,	0x6791,	0x6799,	0x6783,	0x677D,	0x6781,	0x6778,	0x6779,
	0x6794,	0x6B25,	0x6B80,	0x6B7E,	0x6BDE,	0x6C1D,	0x6C93,	0x6CEC,
	0x6CEB,	0x6CEE,	0x6CD9,	0x6CB6,	0x6CD4,	0x6CAD,	0x6CE7,	0x6CB7,
	0x6CD0,	0x6CC2,	0x6CBA,	0x6CC3,	0x6CC6,	0x6CED,	0x6CF2,

	/* Big5-HKSCS 0xCD40..0xCD7E, 0xCDA1..0xCDFE */
	0x6CD2,	0x6CDD,	0x6CB4,	0x6C8A,	0x6C9D,	0x6C80,	0x6CDE,	0x6CC0,
	0x6D30,	0x6CCD,	0x6CC7,	0x6CB0,	0x6CF9,	0x6CCF,	0x6CE9,	0x6CD1,
	0x7094,	0x7098,	0x7085,	0x7093,	0x7086,	0x7084,	0x7091,	0x7096,
	0x7082,	0x709A,	0x7083,	0x726A,	0x72D6,	0x72CB,	0x72D8,	0x72C9,
	0x72DC,	0x72D2,	0x72D4,	0x72DA,	0x72CC,	0x72D1,	0x73A4,	0x73A1,
	0x73AD,	0x73A6,	0x73A2,	0x73A0,	0x73AC,	0x739D,	0x74DD,	0x74E8,
	0x753F,	0x7540,	0x753E,	0x758C,	0x7598,	0x76AF,	0x76F3,	0x76F1,
	0x76F0,	0x76F5,	0x77F8,	0x77FC,	0x77F9,	0x77FB,	0x77FA,
		0x77F7,	0x7942,	0x793F,	0x79C5,	0x7A78,	0x7A7B,	0x7AFB,
	0x7C75,	0x7CFD,	0x8035,	0x808F,	0x80AE,	0x80A3,	0x80B8,	0x80B5,
	0x80AD,	0x8220,	0x82A0,	0x82C0,	0x82AB,	0x829A,	0x8298,	0x829B,
	0x82B5,	0x82A7,	0x82AE,	0x82BC,	0x829E,	0x82BA,	0x82B4,	0x82A8,
	0x82A1,	0x82A9,	0x82C2,	0x82A4,	0x82C3,	0x82B6,	0x82A2,	0x8670,
	0x866F,	0x866D,	0x866E,	0x8C56,	0x8FD2,	0x8FCB,	0x8FD3,	0x8FCD,
	0x8FD6,	0x8FD5,	0x8FD7,	0x90B2,	0x90B4,	0x90AF,	0x90B3,	0x90B0,
	0x9639,	0x963D,	0x963C,	0x963A,	0x9643,	0x4FCD,	0x4FC5,	0x4FD3,
	0x4FB2,	0x4FC9,	0x4FCB,	0x4FC1,	0x4FD4,	0x4FDC,	0x4FD9,	0x4FBB,
	0x4FB3,	0x4FDB,	0x4FC7,	0x4FD6,	0x4FBA,	0x4FC0,	0x4FB9,	0x4FEC,
	0x5244,	0x5249,	0x52C0,	0x52C2,	0x533D,	0x537C,	0x5397,	0x5396,
	0x5399,	0x5398,	0x54BA,	0x54A1,	0x54AD,	0x54A5,	0x54CF,

	/* Big5-HKSCS 0xCE40..0xCE7E, 0xCEA1..0xCEFE */
	0x54C3,	0x830D,	0x54B7,	0x54AE,	0x54D6,	0x54B6,	0x54C5,	0x54C6,
	0x54A0,	0x5470,	0x54BC,	0x54A2,	0x54BE,	0x5472,	0x54DE,	0x54B0,
	0x57B5,	0x579E,	0x579F,	0x57A4,	0x578C,	0x5797,	0x579D,	0x579B,
	0x5794,	0x5798,	0x578F,	0x5799,	0x57A5,	0x579A,	0x5795,	0x58F4,
	0x590D,	0x5953,	0x59E1,	0x59DE,	0x59EE,	0x5A00,	0x59F1,	0x59DD,
	0x59FA,	0x59FD,	0x59FC,	0x59F6,	0x59E4,	0x59F2,	0x59F7,	0x59DB,
	0x59E9,	0x59F3,	0x59F5,	0x59E0,	0x59FE,	0x59F4,	0x59ED,	0x5BA8,
	0x5C4C,	0x5CD0,	0x5CD8,	0x5CCC,	0x5CD7,	0x5CCB,	0x5CDB,
		0x5CDE,	0x5CDA,	0x5CC9,	0x5CC7,	0x5CCA,	0x5CD6,	0x5CD3,
	0x5CD4,	0x5CCF,	0x5CC8,	0x5CC6,	0x5CCE,	0x5CDF,	0x5CF8,	0x5DF9,
	0x5E21,	0x5E22,	0x5E23,	0x5E20,	0x5E24,	0x5EB0,	0x5EA4,	0x5EA2,
	0x5E9B,	0x5EA3,	0x5EA5,	0x5F07,	0x5F2E,	0x5F56,	0x5F86,	0x6037,
	0x6039,	0x6054,	0x6072,	0x605E,	0x6045,	0x6053,	0x6047,	0x6049,
	0x605B,	0x604C,	0x6040,	0x6042,	0x605F,	0x6024,	0x6044,	0x6058,
	0x6066,	0x606E,	0x6242,	0x6243,	0x62CF,	0x630D,	0x630B,	0x62F5,
	0x630E,	0x6303,	0x62EB,	0x62F9,	0x630F,	0x630C,	0x62F8,	0x62F6,
	0x6300,	0x6313,	0x6314,	0x62FA,	0x6315,	0x62FB,	0x62F0,	0x6541,
	0x6543,	0x65AA,	0x65BF,	0x6636,	0x6621,	0x6632,	0x6635,	0x661C,
	0x6626,	0x6622,	0x6633,	0x662B,	0x663A,	0x661D,	0x6634,	0x6639,
	0x662E,	0x670F,	0x6710,	0x67C1,	0x67F2,	0x67C8,	0x67BA,

	/* Big5-HKSCS 0xCF40..0xCF7E, 0xCFA1..0xCFFE */
	0x67DC,	0x67BB,	0x67F8,	0x67D8,	0x67C0,	0x67B7,	0x67C5,	0x67EB,
	0x67E4,	0x67DF,	0x67B5,	0x67CD,	0x67B3,	0x67F7,	0x67F6,	0x67EE,
	0x67E3,	0x67C2,	0x67B9,	0x67CE,	0x67E7,	0x67F0,	0x67B2,	0x67FC,
	0x67C6,	0x67ED,	0x67CC,	0x67AE,	0x67E6,	0x67DB,	0x67FA,	0x67C9,
	0x67CA,	0x67C3,	0x67EA,	0x67CB,	0x6B28,	0x6B82,	0x6B84,	0x6BB6,
	0x6BD6,	0x6BD8,	0x6BE0,	0x6C20,	0x6C21,	0x6D28,	0x6D34,	0x6D2D,
	0x6D1F,	0x6D3C,	0x6D3F,	0x6D12,	0x6D0A,	0x6CDA,	0x6D33,	0x6D04,
	0x6D19,	0x6D3A,	0x6D1A,	0x6D11,	0x6D00,	0x6D1D,	0x6D42,
		0x6D01,	0x6D18,	0x6D37,	0x6D03,	0x6D0F,	0x6D40,	0x6D07,
	0x6D20,	0x6D2C,	0x6D08,	0x6D22,	0x6D09,	0x6D10,	0x70B7,	0x709F,
	0x70BE,	0x70B1,	0x70B0,	0x70A1,	0x70B4,	0x70B5,	0x70A9,	0x7241,
	0x7249,	0x724A,	0x726C,	0x7270,	0x7273,	0x726E,	0x72CA,	0x72E4,
	0x72E8,	0x72EB,	0x72DF,	0x72EA,	0x72E6,	0x72E3,	0x7385,	0x73CC,
	0x73C2,	0x73C8,	0x73C5,	0x73B9,	0x73B6,	0x73B5,	0x73B4,	0x73EB,
	0x73BF,	0x73C7,	0x73BE,	0x73C3,	0x73C6,	0x73B8,	0x73CB,	0x74EC,
	0x74EE,	0x752E,	0x7547,	0x7548,	0x75A7,	0x75AA,	0x7679,	0x76C4,
	0x7708,	0x7703,	0x7704,	0x7705,	0x770A,	0x76F7,	0x76FB,	0x76FA,
	0x77E7,	0x77E8,	0x7806,	0x7811,	0x7812,	0x7805,	0x7810,	0x780F,
	0x780E,	0x7809,	0x7803,	0x7813,	0x794A,	0x794C,	0x794B,	0x7945,
	0x7944,	0x79D5,	0x79CD,	0x79CF,	0x79D6,	0x79CE,	0x7A80,

	/* Big5-HKSCS 0xD040..0xD07E, 0xD0A1..0xD0FE */
	0x7A7E,	0x7AD1,	0x7B00,	0x7B01,	0x7C7A,	0x7C78,	0x7C79,	0x7C7F,
	0x7C80,	0x7C81,	0x7D03,	0x7D08,	0x7D01,	0x7F58,	0x7F91,	0x7F8D,
	0x7FBE,	0x8007,	0x800E,	0x800F,	0x8014,	0x8037,	0x80D8,	0x80C7,
	0x80E0,	0x80D1,	0x80C8,	0x80C2,	0x80D0,	0x80C5,	0x80E3,	0x80D9,
	0x80DC,	0x80CA,	0x80D5,	0x80C9,	0x80CF,	0x80D7,	0x80E6,	0x80CD,
	0x81FF,	0x8221,	0x8294,	0x82D9,	0x82FE,	0x82F9,	0x8307,	0x82E8,
	0x8300,	0x82D5,	0x833A,	0x82EB,	0x82D6,	0x82F4,	0x82EC,	0x82E1,
	0x82F2,	0x82F5,	0x830C,	0x82FB,	0x82F6,	0x82F0,	0x82EA,
		0x82E4,	0x82E0,	0x82FA,	0x82F3,	0x82ED,	0x8677,	0x8674,
	0x867C,	0x8673,	0x8841,	0x884E,	0x8867,	0x886A,	0x8869,	0x89D3,
	0x8A04,	0x8A07,	0x8D72,	0x8FE3,	0x8FE1,	0x8FEE,	0x8FE0,	0x90F1,
	0x90BD,	0x90BF,	0x90D5,	0x90C5,	0x90BE,	0x90C7,	0x90CB,	0x90C8,
	0x91D4,	0x91D3,	0x9654,	0x964F,	0x9651,	0x9653,	0x964A,	0x964E,
	0x501E,	0x5005,	0x5007,	0x5013,	0x5022,	0x5030,	0x501B,	0x4FF5,
	0x4FF4,	0x5033,	0x5037,	0x502C,	0x4FF6,	0x4FF7,	0x5017,	0x501C,
	0x5020,	0x5027,	0x5035,	0x502F,	0x5031,	0x500E,	0x515A,	0x5194,
	0x5193,	0x51CA,	0x51C4,	0x51C5,	0x51C8,	0x51CE,	0x5261,	0x525A,
	0x5252,	0x525E,	0x525F,	0x5255,	0x5262,	0x52CD,	0x530E,	0x539E,
	0x5526,	0x54E2,	0x5517,	0x5512,	0x54E7,	0x54F3,	0x54E4,	0x551A,
	0x54FF,	0x5504,	0x5508,	0x54EB,	0x5511,	0x5505,	0x54F1,

	/* Big5-HKSCS 0xD140..0xD17E, 0xD1A1..0xD1FE */
	0x550A,	0x54FB,	0x54F7,	0x54F8,	0x54E0,	0x550E,	0x5503,	0x550B,
	0x5701,	0x5702,	0x57CC,	0x5832,	0x57D5,	0x57D2,	0x57BA,	0x57C6,
	0x57BD,	0x57BC,	0x57B8,	0x57B6,	0x57BF,	0x57C7,	0x57D0,	0x57B9,
	0x57C1,	0x590E,	0x594A,	0x5A19,	0x5A16,	0x5A2D,	0x5A2E,	0x5A15,
	0x5A0F,	0x5A17,	0x5A0A,	0x5A1E,	0x5A33,	0x5B6C,	0x5BA7,	0x5BAD,
	0x5BAC,	0x5C03,	0x5C56,	0x5C54,	0x5CEC,	0x5CFF,	0x5CEE,	0x5CF1,
	0x5CF7,	0x5D00,	0x5CF9,	0x5E29,	0x5E28,	0x5EA8,	0x5EAE,	0x5EAA,
	0x5EAC,	0x5F33,	0x5F30,	0x5F67,	0x605D,	0x605A,	0x6067,
		0x6041,	0x60A2,	0x6088,	0x6080,	0x6092,	0x6081,	0x609D,
	0x6083,	0x6095,	0x609B,	0x6097,	0x6087,	0x609C,	0x608E,	0x6219,
	0x6246,	0x62F2,	0x6310,	0x6356,	0x632C,	0x6344,	0x6345,	0x6336,
	0x6343,	0x63E4,	0x6339,	0x634B,	0x634A,	0x633C,	0x6329,	0x6341,
	0x6334,	0x6358,	0x6354,	0x6359,	0x632D,	0x6347,	0x6333,	0x635A,
	0x6351,	0x6338,	0x6357,	0x6340,	0x6348,	0x654A,	0x6546,	0x65C6,
	0x65C3,	0x65C4,	0x65C2,	0x664A,	0x665F,	0x6647,	0x6651,	0x6712,
	0x6713,	0x681F,	0x681A,	0x6849,	0x6832,	0x6833,	0x683B,	0x684B,
	0x684F,	0x6816,	0x6831,	0x681C,	0x6835,	0x682B,	0x682D,	0x682F,
	0x684E,	0x6844,	0x6834,	0x681D,	0x6812,	0x6814,	0x6826,	0x6828,
	0x682E,	0x684D,	0x683A,	0x6825,	0x6820,	0x6B2C,	0x6B2F,	0x6B2D,
	0x6B31,	0x6B34,	0x6B6D,	0x8082,	0x6B88,	0x6BE6,	0x6BE4,

	/* Big5-HKSCS 0xD240..0xD27E, 0xD2A1..0xD2FE */
	0x6BE8,	0x6BE3,	0x6BE2,	0x6BE7,	0x6C25,	0x6D7A,	0x6D63,	0x6D64,
	0x6D76,	0x6D0D,	0x6D61,	0x6D92,	0x6D58,	0x6D62,	0x6D6D,	0x6D6F,
	0x6D91,	0x6D8D,	0x6DEF,	0x6D7F,	0x6D86,	0x6D5E,	0x6D67,	0x6D60,
	0x6D97,	0x6D70,	0x6D7C,	0x6D5F,	0x6D82,	0x6D98,	0x6D2F,	0x6D68,
	0x6D8B,	0x6D7E,	0x6D80,	0x6D84,	0x6D16,	0x6D83,	0x6D7B,	0x6D7D,
	0x6D75,	0x6D90,	0x70DC,	0x70D3,	0x70D1,	0x70DD,	0x70CB,	0x7F39,
	0x70E2,	0x70D7,	0x70D2,	0x70DE,	0x70E0,	0x70D4,	0x70CD,	0x70C5,
	0x70C6,	0x70C7,	0x70DA,	0x70CE,	0x70E1,	0x7242,	0x7278,
		0x7277,	0x7276,	0x7300,	0x72FA,	0x72F4,	0x72FE,	0x72F6,
	0x72F3,	0x72FB,	0x7301,	0x73D3,	0x73D9,	0x73E5,	0x73D6,	0x73BC,
	0x73E7,	0x73E3,	0x73E9,	0x73DC,	0x73D2,	0x73DB,	0x73D4,	0x73DD,
	0x73DA,	0x73D7,	0x73D8,	0x73E8,	0x74DE,	0x74DF,	0x74F4,	0x74F5,
	0x7521,	0x755B,	0x755F,	0x75B0,	0x75C1,	0x75BB,	0x75C4,	0x75C0,
	0x75BF,	0x75B6,	0x75BA,	0x768A,	0x76C9,	0x771D,	0x771B,	0x7710,
	0x7713,	0x7712,	0x7723,	0x7711,	0x7715,	0x7719,	0x771A,	0x7722,
	0x7727,	0x7823,	0x782C,	0x7822,	0x7835,	0x782F,	0x7828,	0x782E,
	0x782B,	0x7821,	0x7829,	0x7833,	0x782A,	0x7831,	0x7954,	0x795B,
	0x794F,	0x795C,	0x7953,	0x7952,	0x7951,	0x79EB,	0x79EC,	0x79E0,
	0x79EE,	0x79ED,	0x79EA,	0x79DC,	0x79DE,	0x79DD,	0x7A86,	0x7A89,
	0x7A85,	0x7A8B,	0x7A8C,	0x7A8A,	0x7A87,	0x7AD8,	0x7B10,

	/* Big5-HKSCS 0xD340..0xD37E, 0xD3A1..0xD3FE */
	0x7B04,	0x7B13,	0x7B05,	0x7B0F,	0x7B08,	0x7B0A,	0x7B0E,	0x7B09,
	0x7B12,	0x7C84,	0x7C91,	0x7C8A,	0x7C8C,	0x7C88,	0x7C8D,	0x7C85,
	0x7D1E,	0x7D1D,	0x7D11,	0x7D0E,	0x7D18,	0x7D16,	0x7D13,	0x7D1F,
	0x7D12,	0x7D0F,	0x7D0C,	0x7F5C,	0x7F61,	0x7F5E,	0x7F60,	0x7F5D,
	0x7F5B,	0x7F96,	0x7F92,	0x7FC3,	0x7FC2,	0x7FC0,	0x8016,	0x803E,
	0x8039,	0x80FA,	0x80F2,	0x80F9,	0x80F5,	0x8101,	0x80FB,	0x8100,
	0x8201,	0x822F,	0x8225,	0x8333,	0x832D,	0x8344,	0x8319,	0x8351,
	0x8325,	0x8356,	0x833F,	0x8341,	0x8326,	0x831C,	0x8322,
		0x8342,	0x834E,	0x831B,	0x832A,	0x8308,	0x833C,	0x834D,
	0x8316,	0x8324,	0x8320,	0x8337,	0x832F,	0x8329,	0x8347,	0x8345,
	0x834C,	0x8353,	0x831E,	0x832C,	0x834B,	0x8327,	0x8348,	0x8653,
	0x8652,	0x86A2,	0x86A8,	0x8696,	0x868D,	0x8691,	0x869E,	0x8687,
	0x8697,	0x8686,	0x868B,	0x869A,	0x8685,	0x86A5,	0x8699,	0x86A1,
	0x86A7,	0x8695,	0x8698,	0x868E,	0x869D,	0x8690,	0x8694,	0x8843,
	0x8844,	0x886D,	0x8875,	0x8876,	0x8872,	0x8880,	0x8871,	0x887F,
	0x886F,	0x8883,	0x887E,	0x8874,	0x887C,	0x8A12,	0x8C47,	0x8C57,
	0x8C7B,	0x8CA4,	0x8CA3,	0x8D76,	0x8D78,	0x8DB5,	0x8DB7,	0x8DB6,
	0x8ED1,	0x8ED3,	0x8FFE,	0x8FF5,	0x9002,	0x8FFF,	0x8FFB,	0x9004,
	0x8FFC,	0x8FF6,	0x90D6,	0x90E0,	0x90D9,	0x90DA,	0x90E3,	0x90DF,
	0x90E5,	0x90D8,	0x90DB,	0x90D7,	0x90DC,	0x90E4,	0x9150,

	/* Big5-HKSCS 0xD440..0xD47E, 0xD4A1..0xD4FE */
	0x914E,	0x914F,	0x91D5,	0x91E2,	0x91DA,	0x965C,	0x965F,	0x96BC,
	0x98E3,	0x9ADF,	0x9B2F,	0x4E7F,	0x5070,	0x506A,	0x5061,	0x505E,
	0x5060,	0x5053,	0x504B,	0x505D,	0x5072,	0x5048,	0x504D,	0x5041,
	0x505B,	0x504A,	0x5062,	0x5015,	0x5045,	0x505F,	0x5069,	0x506B,
	0x5063,	0x5064,	0x5046,	0x5040,	0x506E,	0x5073,	0x5057,	0x5051,
	0x51D0,	0x526B,	0x526D,	0x526C,	0x526E,	0x52D6,	0x52D3,	0x532D,
	0x539C,	0x5575,	0x5576,	0x553C,	0x554D,	0x5550,	0x5534,	0x552A,
	0x5551,	0x5562,	0x5536,	0x5535,	0x5530,	0x5552,	0x5545,
		0x550C,	0x5532,	0x5565,	0x554E,	0x5539,	0x5548,	0x552D,
	0x553B,	0x5540,	0x554B,	0x570A,	0x5707,	0x57FB,	0x5814,	0x57E2,
	0x57F6,	0x57DC,	0x57F4,	0x5800,	0x57ED,	0x57FD,	0x5808,	0x57F8,
	0x580B,	0x57F3,	0x57CF,	0x5807,	0x57EE,	0x57E3,	0x57F2,	0x57E5,
	0x57EC,	0x57E1,	0x580E,	0x57FC,	0x5810,	0x57E7,	0x5801,	0x580C,
	0x57F1,	0x57E9,	0x57F0,	0x580D,	0x5804,	0x595C,	0x5A60,	0x5A58,
	0x5A55,	0x5A67,	0x5A5E,	0x5A38,	0x5A35,	0x5A6D,	0x5A50,	0x5A5F,
	0x5A65,	0x5A6C,	0x5A53,	0x5A64,	0x5A57,	0x5A43,	0x5A5D,	0x5A52,
	0x5A44,	0x5A5B,	0x5A48,	0x5A8E,	0x5A3E,	0x5A4D,	0x5A39,	0x5A4C,
	0x5A70,	0x5A69,	0x5A47,	0x5A51,	0x5A56,	0x5A42,	0x5A5C,	0x5B72,
	0x5B6E,	0x5BC1,	0x5BC0,	0x5C59,	0x5D1E,	0x5D0B,	0x5D1D,	0x5D1A,
	0x5D20,	0x5D0C,	0x5D28,	0x5D0D,	0x5D26,	0x5D25,	0x5D0F,

	/* Big5-HKSCS 0xD540..0xD57E, 0xD5A1..0xD5FE */
	0x5D30,	0x5D12,	0x5D23,	0x5D1F,	0x5D2E,	0x5E3E,	0x5E34,	0x5EB1,
	0x5EB4,	0x5EB9,	0x5EB2,	0x5EB3,	0x5F36,	0x5F38,	0x5F9B,	0x5F96,
	0x5F9F,	0x608A,	0x6090,	0x6086,	0x60BE,	0x60B0,	0x60BA,	0x60D3,
	0x60D4,	0x60CF,	0x60E4,	0x60D9,	0x60DD,	0x60C8,	0x60B1,	0x60DB,
	0x60B7,	0x60CA,	0x60BF,	0x60C3,	0x60CD,	0x60C0,	0x6332,	0x6365,
	0x638A,	0x6382,	0x637D,	0x63BD,	0x639E,	0x63AD,	0x639D,	0x6397,
	0x63AB,	0x638E,	0x636F,	0x6387,	0x6390,	0x636E,	0x63AF,	0x6375,
	0x639C,	0x636D,	0x63AE,	0x637C,	0x63A4,	0x633B,	0x639F,
		0x6378,	0x6385,	0x6381,	0x6391,	0x638D,	0x6370,	0x6553,
	0x65CD,	0x6665,	0x6661,	0x665B,	0x6659,	0x665C,	0x6662,	0x6718,
	0x6879,	0x6887,	0x6890,	0x689C,	0x686D,	0x686E,	0x68AE,	0x68AB,
	0x6956,	0x686F,	0x68A3,	0x68AC,	0x68A9,	0x6875,	0x6874,	0x68B2,
	0x688F,	0x6877,	0x6892,	0x687C,	0x686B,	0x6872,	0x68AA,	0x6880,
	0x6871,	0x687E,	0x689B,	0x6896,	0x688B,	0x68A0,	0x6889,	0x68A4,
	0x6878,	0x687B,	0x6891,	0x688C,	0x688A,	0x687D,	0x6B36,	0x6B33,
	0x6B37,	0x6B38,	0x6B91,	0x6B8F,	0x6B8D,	0x6B8E,	0x6B8C,	0x6C2A,
	0x6DC0,	0x6DAB,	0x6DB4,	0x6DB3,	0x6E74,	0x6DAC,	0x6DE9,	0x6DE2,
	0x6DB7,	0x6DF6,	0x6DD4,	0x6E00,	0x6DC8,	0x6DE0,	0x6DDF,	0x6DD6,
	0x6DBE,	0x6DE5,	0x6DDC,	0x6DDD,	0x6DDB,	0x6DF4,	0x6DCA,	0x6DBD,
	0x6DED,	0x6DF0,	0x6DBA,	0x6DD5,	0x6DC2,	0x6DCF,	0x6DC9,

	/* Big5-HKSCS 0xD640..0xD67E, 0xD6A1..0xD6FE */
	0x6DD0,	0x6DF2,	0x6DD3,	0x6DFD,	0x6DD7,	0x6DCD,	0x6DE3,	0x6DBB,
	0x70FA,	0x710D,	0x70F7,	0x7117,	0x70F4,	0x710C,	0x70F0,	0x7104,
	0x70F3,	0x7110,	0x70FC,	0x70FF,	0x7106,	0x7113,	0x7100,	0x70F8,
	0x70F6,	0x710B,	0x7102,	0x710E,	0x727E,	0x727B,	0x727C,	0x727F,
	0x731D,	0x7317,	0x7307,	0x7311,	0x7318,	0x730A,	0x7308,	0x72FF,
	0x730F,	0x731E,	0x7388,	0x73F6,	0x73F8,	0x73F5,	0x7404,	0x7401,
	0x73FD,	0x7407,	0x7400,	0x73FA,	0x73FC,	0x73FF,	0x740C,	0x740B,
	0x73F4,	0x7408,	0x7564,	0x7563,	0x75CE,	0x75D2,	0x75CF,
		0x75CB,	0x75CC,	0x75D1,	0x75D0,	0x768F,	0x7689,	0x76D3,
	0x7739,	0x772F,	0x772D,	0x7731,	0x7732,	0x7734,	0x7733,	0x773D,
	0x7725,	0x773B,	0x7735,	0x7848,	0x7852,	0x7849,	0x784D,	0x784A,
	0x784C,	0x7826,	0x7845,	0x7850,	0x7964,	0x7967,	0x7969,	0x796A,
	0x7963,	0x796B,	0x7961,	0x79BB,	0x79FA,	0x79F8,	0x79F6,	0x79F7,
	0x7A8F,	0x7A94,	0x7A90,	0x7B35,	0x7B47,	0x7B34,	0x7B25,	0x7B30,
	0x7B22,	0x7B24,	0x7B33,	0x7B18,	0x7B2A,	0x7B1D,	0x7B31,	0x7B2B,
	0x7B2D,	0x7B2F,	0x7B32,	0x7B38,	0x7B1A,	0x7B23,	0x7C94,	0x7C98,
	0x7C96,	0x7CA3,	0x7D35,	0x7D3D,	0x7D38,	0x7D36,	0x7D3A,	0x7D45,
	0x7D2C,	0x7D29,	0x7D41,	0x7D47,	0x7D3E,	0x7D3F,	0x7D4A,	0x7D3B,
	0x7D28,	0x7F63,	0x7F95,	0x7F9C,	0x7F9D,	0x7F9B,	0x7FCA,	0x7FCB,
	0x7FCD,	0x7FD0,	0x7FD1,	0x7FC7,	0x7FCF,	0x7FC9,	0x801F,

	/* Big5-HKSCS 0xD740..0xD77E, 0xD7A1..0xD7FE */
	0x801E,	0x801B,	0x8047,	0x8043,	0x8048,	0x8118,	0x8125,	0x8119,
	0x811B,	0x812D,	0x811F,	0x812C,	0x811E,	0x8121,	0x8115,	0x8127,
	0x811D,	0x8122,	0x8211,	0x8238,	0x8233,	0x823A,	0x8234,	0x8232,
	0x8274,	0x8390,	0x83A3,	0x83A8,	0x838D,	0x837A,	0x8373,	0x83A4,
	0x8374,	0x838F,	0x8381,	0x8395,	0x8399,	0x8375,	0x8394,	0x83A9,
	0x837D,	0x8383,	0x838C,	0x839D,	0x839B,	0x83AA,	0x838B,	0x837E,
	0x83A5,	0x83AF,	0x8388,	0x8397,	0x83B0,	0x837F,	0x83A6,	0x8387,
	0x83AE,	0x8376,	0x839A,	0x8659,	0x8656,	0x86BF,	0x86B7,
		0x86C2,	0x86C1,	0x86C5,	0x86BA,	0x86B0,	0x86C8,	0x86B9,
	0x86B3,	0x86B8,	0x86CC,	0x86B4,	0x86BB,	0x86BC,	0x86C3,	0x86BD,
	0x86BE,	0x8852,	0x8889,	0x8895,	0x88A8,	0x88A2,	0x88AA,	0x889A,
	0x8891,	0x88A1,	0x889F,	0x8898,	0x88A7,	0x8899,	0x889B,	0x8897,
	0x88A4,	0x88AC,	0x888C,	0x8893,	0x888E,	0x8982,	0x89D6,	0x89D9,
	0x89D5,	0x8A30,	0x8A27,	0x8A2C,	0x8A1E,	0x8C39,	0x8C3B,	0x8C5C,
	0x8C5D,	0x8C7D,	0x8CA5,	0x8D7D,	0x8D7B,	0x8D79,	0x8DBC,	0x8DC2,
	0x8DB9,	0x8DBF,	0x8DC1,	0x8ED8,	0x8EDE,	0x8EDD,	0x8EDC,	0x8ED7,
	0x8EE0,	0x8EE1,	0x9024,	0x900B,	0x9011,	0x901C,	0x900C,	0x9021,
	0x90EF,	0x90EA,	0x90F0,	0x90F4,	0x90F2,	0x90F3,	0x90D4,	0x90EB,
	0x90EC,	0x90E9,	0x9156,	0x9158,	0x915A,	0x9153,	0x9155,	0x91EC,
	0x91F4,	0x91F1,	0x91F3,	0x91F8,	0x91E4,	0x91F9,	0x91EA,

	/* Big5-HKSCS 0xD840..0xD87E, 0xD8A1..0xD8FE */
	0x91EB,	0x91F7,	0x91E8,	0x91EE,	0x957A,	0x9586,	0x9588,	0x967C,
	0x966D,	0x966B,	0x9671,	0x966F,	0x96BF,	0x976A,	0x9804,	0x98E5,
	0x9997,	0x509B,	0x5095,	0x5094,	0x509E,	0x508B,	0x50A3,	0x5083,
	0x508C,	0x508E,	0x509D,	0x5068,	0x509C,	0x5092,	0x5082,	0x5087,
	0x515F,	0x51D4,	0x5312,	0x5311,	0x53A4,	0x53A7,	0x5591,	0x55A8,
	0x55A5,	0x55AD,	0x5577,	0x5645,	0x55A2,	0x5593,	0x5588,	0x558F,
	0x55B5,	0x5581,	0x55A3,	0x5592,	0x55A4,	0x557D,	0x558C,	0x55A6,
	0x557F,	0x5595,	0x55A1,	0x558E,	0x570C,	0x5829,	0x5837,
		0x5819,	0x581E,	0x5827,	0x5823,	0x5828,	0x57F5,	0x5848,
	0x5825,	0x581C,	0x581B,	0x5833,	0x583F,	0x5836,	0x582E,	0x5839,
	0x5838,	0x582D,	0x582C,	0x583B,	0x5961,	0x5AAF,	0x5A94,	0x5A9F,
	0x5A7A,	0x5AA2,	0x5A9E,	0x5A78,	0x5AA6,	0x5A7C,	0x5AA5,	0x5AAC,
	0x5A95,	0x5AAE,	0x5A37,	0x5A84,	0x5A8A,	0x5A97,	0x5A83,	0x5A8B,
	0x5AA9,	0x5A7B,	0x5A7D,	0x5A8C,	0x5A9C,	0x5A8F,	0x5A93,	0x5A9D,
	0x5BEA,	0x5BCD,	0x5BCB,	0x5BD4,	0x5BD1,	0x5BCA,	0x5BCE,	0x5C0C,
	0x5C30,	0x5D37,	0x5D43,	0x5D6B,	0x5D41,	0x5D4B,	0x5D3F,	0x5D35,
	0x5D51,	0x5D4E,	0x5D55,	0x5D33,	0x5D3A,	0x5D52,	0x5D3D,	0x5D31,
	0x5D59,	0x5D42,	0x5D39,	0x5D49,	0x5D38,	0x5D3C,	0x5D32,	0x5D36,
	0x5D40,	0x5D45,	0x5E44,	0x5E41,	0x5F58,	0x5FA6,	0x5FA5,	0x5FAB,
	0x60C9,	0x60B9,	0x60CC,	0x60E2,	0x60CE,	0x60C4,	0x6114,

	/* Big5-HKSCS 0xD940..0xD97E, 0xD9A1..0xD9FE */
	0x60F2,	0x610A,	0x6116,	0x6105,	0x60F5,	0x6113,	0x60F8,	0x60FC,
	0x60FE,	0x60C1,	0x6103,	0x6118,	0x611D,	0x6110,	0x60FF,	0x6104,
	0x610B,	0x624A,	0x6394,	0x63B1,	0x63B0,	0x63CE,	0x63E5,	0x63E8,
	0x63EF,	0x63C3,	0x649D,	0x63F3,	0x63CA,	0x63E0,	0x63F6,	0x63D5,
	0x63F2,	0x63F5,	0x6461,	0x63DF,	0x63BE,	0x63DD,	0x63DC,	0x63C4,
	0x63D8,	0x63D3,	0x63C2,	0x63C7,	0x63CC,	0x63CB,	0x63C8,	0x63F0,
	0x63D7,	0x63D9,	0x6532,	0x6567,	0x656A,	0x6564,	0x655C,	0x6568,
	0x6565,	0x658C,	0x659D,	0x659E,	0x65AE,	0x65D0,	0x65D2,
		0x667C,	0x666C,	0x667B,	0x6680,	0x6671,	0x6679,	0x666A,
	0x6672,	0x6701,	0x690C,	0x68D3,	0x6904,	0x68DC,	0x692A,	0x68EC,
	0x68EA,	0x68F1,	0x690F,	0x68D6,	0x68F7,	0x68EB,	0x68E4,	0x68F6,
	0x6913,	0x6910,	0x68F3,	0x68E1,	0x6907,	0x68CC,	0x6908,	0x6970,
	0x68B4,	0x6911,	0x68EF,	0x68C6,	0x6914,	0x68F8,	0x68D0,	0x68FD,
	0x68FC,	0x68E8,	0x690B,	0x690A,	0x6917,	0x68CE,	0x68C8,	0x68DD,
	0x68DE,	0x68E6,	0x68F4,	0x68D1,	0x6906,	0x68D4,	0x68E9,	0x6915,
	0x6925,	0x68C7,	0x6B39,	0x6B3B,	0x6B3F,	0x6B3C,	0x6B94,	0x6B97,
	0x6B99,	0x6B95,	0x6BBD,	0x6BF0,	0x6BF2,	0x6BF3,	0x6C30,	0x6DFC,
	0x6E46,	0x6E47,	0x6E1F,	0x6E49,	0x6E88,	0x6E3C,	0x6E3D,	0x6E45,
	0x6E62,	0x6E2B,	0x6E3F,	0x6E41,	0x6E5D,	0x6E73,	0x6E1C,	0x6E33,
	0x6E4B,	0x6E40,	0x6E51,	0x6E3B,	0x6E03,	0x6E2E,	0x6E5E,

	/* Big5-HKSCS 0xDA40..0xDA7E, 0xDAA1..0xDAFE */
	0x6E68,	0x6E5C,	0x6E61,	0x6E31,	0x6E28,	0x6E60,	0x6E71,	0x6E6B,
	0x6E39,	0x6E22,	0x6E30,	0x6E53,	0x6E65,	0x6E27,	0x6E78,	0x6E64,
	0x6E77,	0x6E55,	0x6E79,	0x6E52,	0x6E66,	0x6E35,	0x6E36,	0x6E5A,
	0x7120,	0x711E,	0x712F,	0x70FB,	0x712E,	0x7131,	0x7123,	0x7125,
	0x7122,	0x7132,	0x711F,	0x7128,	0x713A,	0x711B,	0x724B,	0x725A,
	0x7288,	0x7289,	0x7286,	0x7285,	0x728B,	0x7312,	0x730B,	0x7330,
	0x7322,	0x7331,	0x7333,	0x7327,	0x7332,	0x732D,	0x7326,	0x7323,
	0x7335,	0x730C,	0x742E,	0x742C,	0x7430,	0x742B,	0x7416,
		0x741A,	0x7421,	0x742D,	0x7431,	0x7424,	0x7423,	0x741D,
	0x7429,	0x7420,	0x7432,	0x74FB,	0x752F,	0x756F,	0x756C,	0x75E7,
	0x75DA,	0x75E1,	0x75E6,	0x75DD,	0x75DF,	0x75E4,	0x75D7,	0x7695,
	0x7692,	0x76DA,	0x7746,	0x7747,	0x7744,	0x774D,	0x7745,	0x774A,
	0x774E,	0x774B,	0x774C,	0x77DE,	0x77EC,	0x7860,	0x7864,	0x7865,
	0x785C,	0x786D,	0x7871,	0x786A,	0x786E,	0x7870,	0x7869,	0x7868,
	0x785E,	0x7862,	0x7974,	0x7973,	0x7972,	0x7970,	0x7A02,	0x7A0A,
	0x7A03,	0x7A0C,	0x7A04,	0x7A99,	0x7AE6,	0x7AE4,	0x7B4A,	0x7B3B,
	0x7B44,	0x7B48,	0x7B4C,	0x7B4E,	0x7B40,	0x7B58,	0x7B45,	0x7CA2,
	0x7C9E,	0x7CA8,	0x7CA1,	0x7D58,	0x7D6F,	0x7D63,	0x7D53,	0x7D56,
	0x7D67,	0x7D6A,	0x7D4F,	0x7D6D,	0x7D5C,	0x7D6B,	0x7D52,	0x7D54,
	0x7D69,	0x7D51,	0x7D5F,	0x7D4E,	0x7F3E,	0x7F3F,	0x7F65,

	/* Big5-HKSCS 0xDB40..0xDB7E, 0xDBA1..0xDBFE */
	0x7F66,	0x7FA2,	0x7FA0,	0x7FA1,	0x7FD7,	0x8051,	0x804F,	0x8050,
	0x80FE,	0x80D4,	0x8143,	0x814A,	0x8152,	0x814F,	0x8147,	0x813D,
	0x814D,	0x813A,	0x81E6,	0x81EE,	0x81F7,	0x81F8,	0x81F9,	0x8204,
	0x823C,	0x823D,	0x823F,	0x8275,	0x833B,	0x83CF,	0x83F9,	0x8423,
	0x83C0,	0x83E8,	0x8412,	0x83E7,	0x83E4,	0x83FC,	0x83F6,	0x8410,
	0x83C6,	0x83C8,	0x83EB,	0x83E3,	0x83BF,	0x8401,	0x83DD,	0x83E5,
	0x83D8,	0x83FF,	0x83E1,	0x83CB,	0x83CE,	0x83D6,	0x83F5,	0x83C9,
	0x8409,	0x840F,	0x83DE,	0x8411,	0x8406,	0x83C2,	0x83F3,
		0x83D5,	0x83FA,	0x83C7,	0x83D1,	0x83EA,	0x8413,	0x83C3,
	0x83EC,	0x83EE,	0x83C4,	0x83FB,	0x83D7,	0x83E2,	0x841B,	0x83DB,
	0x83FE,	0x86D8,	0x86E2,	0x86E6,	0x86D3,	0x86E3,	0x86DA,	0x86EA,
	0x86DD,	0x86EB,	0x86DC,	0x86EC,	0x86E9,	0x86D7,	0x86E8,	0x86D1,
	0x8848,	0x8856,	0x8855,	0x88BA,	0x88D7,	0x88B9,	0x88B8,	0x88C0,
	0x88BE,	0x88B6,	0x88BC,	0x88B7,	0x88BD,	0x88B2,	0x8901,	0x88C9,
	0x8995,	0x8998,	0x8997,	0x89DD,	0x89DA,	0x89DB,	0x8A4E,	0x8A4D,
	0x8A39,	0x8A59,	0x8A40,	0x8A57,	0x8A58,	0x8A44,	0x8A45,	0x8A52,
	0x8A48,	0x8A51,	0x8A4A,	0x8A4C,	0x8A4F,	0x8C5F,	0x8C81,	0x8C80,
	0x8CBA,	0x8CBE,	0x8CB0,	0x8CB9,	0x8CB5,	0x8D84,	0x8D80,	0x8D89,
	0x8DD8,	0x8DD3,	0x8DCD,	0x8DC7,	0x8DD6,	0x8DDC,	0x8DCF,	0x8DD5,
	0x8DD9,	0x8DC8,	0x8DD7,	0x8DC5,	0x8EEF,	0x8EF7,	0x8EFA,

	/* Big5-HKSCS 0xDC40..0xDC7E, 0xDCA1..0xDCFE */
	0x8EF9,	0x8EE6,	0x8EEE,	0x8EE5,	0x8EF5,	0x8EE7,	0x8EE8,	0x8EF6,
	0x8EEB,	0x8EF1,	0x8EEC,	0x8EF4,	0x8EE9,	0x902D,	0x9034,	0x902F,
	0x9106,	0x912C,	0x9104,	0x90FF,	0x90FC,	0x9108,	0x90F9,	0x90FB,
	0x9101,	0x9100,	0x9107,	0x9105,	0x9103,	0x9161,	0x9164,	0x915F,
	0x9162,	0x9160,	0x9201,	0x920A,	0x9225,	0x9203,	0x921A,	0x9226,
	0x920F,	0x920C,	0x9200,	0x9212,	0x91FF,	0x91FD,	0x9206,	0x9204,
	0x9227,	0x9202,	0x921C,	0x9224,	0x9219,	0x9217,	0x9205,	0x9216,
	0x957B,	0x958D,	0x958C,	0x9590,	0x9687,	0x967E,	0x9688,
		0x9689,	0x9683,	0x9680,	0x96C2,	0x96C8,	0x96C3,	0x96F1,
	0x96F0,	0x976C,	0x9770,	0x976E,	0x9807,	0x98A9,	0x98EB,	0x9CE6,
	0x9EF9,	0x4E83,	0x4E84,	0x4EB6,	0x50BD,	0x50BF,	0x50C6,	0x50AE,
	0x50C4,	0x50CA,	0x50B4,	0x50C8,	0x50C2,	0x50B0,	0x50C1,	0x50BA,
	0x50B1,	0x50CB,	0x50C9,	0x50B6,	0x50B8,	0x51D7,	0x527A,	0x5278,
	0x527B,	0x527C,	0x55C3,	0x55DB,	0x55CC,	0x55D0,	0x55CB,	0x55CA,
	0x55DD,	0x55C0,	0x55D4,	0x55C4,	0x55E9,	0x55BF,	0x55D2,	0x558D,
	0x55CF,	0x55D5,	0x55E2,	0x55D6,	0x55C8,	0x55F2,	0x55CD,	0x55D9,
	0x55C2,	0x5714,	0x5853,	0x5868,	0x5864,	0x584F,	0x584D,	0x5849,
	0x586F,	0x5855,	0x584E,	0x585D,	0x5859,	0x5865,	0x585B,	0x583D,
	0x5863,	0x5871,	0x58FC,	0x5AC7,	0x5AC4,	0x5ACB,	0x5ABA,	0x5AB8,
	0x5AB1,	0x5AB5,	0x5AB0,	0x5ABF,	0x5AC8,	0x5ABB,	0x5AC6,

	/* Big5-HKSCS 0xDD40..0xDD7E, 0xDDA1..0xDDFE */
	0x5AB7,	0x5AC0,	0x5ACA,	0x5AB4,	0x5AB6,	0x5ACD,	0x5AB9,	0x5A90,
	0x5BD6,	0x5BD8,	0x5BD9,	0x5C1F,	0x5C33,	0x5D71,	0x5D63,	0x5D4A,
	0x5D65,	0x5D72,	0x5D6C,	0x5D5E,	0x5D68,	0x5D67,	0x5D62,	0x5DF0,
	0x5E4F,	0x5E4E,	0x5E4A,	0x5E4D,	0x5E4B,	0x5EC5,	0x5ECC,	0x5EC6,
	0x5ECB,	0x5EC7,	0x5F40,	0x5FAF,	0x5FAD,	0x60F7,	0x6149,	0x614A,
	0x612B,	0x6145,	0x6136,	0x6132,	0x612E,	0x6146,	0x612F,	0x614F,
	0x6129,	0x6140,	0x6220,	0x9168,	0x6223,	0x6225,	0x6224,	0x63C5,
	0x63F1,	0x63EB,	0x6410,	0x6412,	0x6409,	0x6420,	0x6424,
		0x6433,	0x6443,	0x641F,	0x6415,	0x6418,	0x6439,	0x6437,
	0x6422,	0x6423,	0x640C,	0x6426,	0x6430,	0x6428,	0x6441,	0x6435,
	0x642F,	0x640A,	0x641A,	0x6440,	0x6425,	0x6427,	0x640B,	0x63E7,
	0x641B,	0x642E,	0x6421,	0x640E,	0x656F,	0x6592,	0x65D3,	0x6686,
	0x668C,	0x6695,	0x6690,	0x668B,	0x668A,	0x6699,	0x6694,	0x6678,
	0x6720,	0x6966,	0x695F,	0x6938,	0x694E,	0x6962,	0x6971,	0x693F,
	0x6945,	0x696A,	0x6939,	0x6942,	0x6957,	0x6959,	0x697A,	0x6948,
	0x6949,	0x6935,	0x696C,	0x6933,	0x693D,	0x6965,	0x68F0,	0x6978,
	0x6934,	0x6969,	0x6940,	0x696F,	0x6944,	0x6976,	0x6958,	0x6941,
	0x6974,	0x694C,	0x693B,	0x694B,	0x6937,	0x695C,	0x694F,	0x6951,
	0x6932,	0x6952,	0x692F,	0x697B,	0x693C,	0x6B46,	0x6B45,	0x6B43,
	0x6B42,	0x6B48,	0x6B41,	0x6B9B,	0xFA0D,	0x6BFB,	0x6BFC,

	/* Big5-HKSCS 0xDE40..0xDE7E, 0xDEA1..0xDEFE */
	0x6BF9,	0x6BF7,	0x6BF8,	0x6E9B,	0x6ED6,	0x6EC8,	0x6E8F,	0x6EC0,
	0x6E9F,	0x6E93,	0x6E94,	0x6EA0,	0x6EB1,	0x6EB9,	0x6EC6,	0x6ED2,
	0x6EBD,	0x6EC1,	0x6E9E,	0x6EC9,	0x6EB7,	0x6EB0,	0x6ECD,	0x6EA6,
	0x6ECF,	0x6EB2,	0x6EBE,	0x6EC3,	0x6EDC,	0x6ED8,	0x6E99,	0x6E92,
	0x6E8E,	0x6E8D,	0x6EA4,	0x6EA1,	0x6EBF,	0x6EB3,	0x6ED0,	0x6ECA,
	0x6E97,	0x6EAE,	0x6EA3,	0x7147,	0x7154,	0x7152,	0x7163,	0x7160,
	0x7141,	0x715D,	0x7162,	0x7172,	0x7178,	0x716A,	0x7161,	0x7142,
	0x7158,	0x7143,	0x714B,	0x7170,	0x715F,	0x7150,	0x7153,
		0x7144,	0x714D,	0x715A,	0x724F,	0x728D,	0x728C,	0x7291,
	0x7290,	0x728E,	0x733C,	0x7342,	0x733B,	0x733A,	0x7340,	0x734A,
	0x7349,	0x7444,	0x744A,	0x744B,	0x7452,	0x7451,	0x7457,	0x7440,
	0x744F,	0x7450,	0x744E,	0x7442,	0x7446,	0x744D,	0x7454,	0x74E1,
	0x74FF,	0x74FE,	0x74FD,	0x751D,	0x7579,	0x7577,	0x6983,	0x75EF,
	0x760F,	0x7603,	0x75F7,	0x75FE,	0x75FC,	0x75F9,	0x75F8,	0x7610,
	0x75FB,	0x75F6,	0x75ED,	0x75F5,	0x75FD,	0x7699,	0x76B5,	0x76DD,
	0x7755,	0x775F,	0x7760,	0x7752,	0x7756,	0x775A,	0x7769,	0x7767,
	0x7754,	0x7759,	0x776D,	0x77E0,	0x7887,	0x789A,	0x7894,	0x788F,
	0x7884,	0x7895,	0x7885,	0x7886,	0x78A1,	0x7883,	0x7879,	0x7899,
	0x7880,	0x7896,	0x787B,	0x797C,	0x7982,	0x797D,	0x7979,	0x7A11,
	0x7A18,	0x7A19,	0x7A12,	0x7A17,	0x7A15,	0x7A22,	0x7A13,

	/* Big5-HKSCS 0xDF40..0xDF7E, 0xDFA1..0xDFFE */
	0x7A1B,	0x7A10,	0x7AA3,	0x7AA2,	0x7A9E,	0x7AEB,	0x7B66,	0x7B64,
	0x7B6D,	0x7B74,	0x7B69,	0x7B72,	0x7B65,	0x7B73,	0x7B71,	0x7B70,
	0x7B61,	0x7B78,	0x7B76,	0x7B63,	0x7CB2,	0x7CB4,	0x7CAF,	0x7D88,
	0x7D86,	0x7D80,	0x7D8D,	0x7D7F,	0x7D85,	0x7D7A,	0x7D8E,	0x7D7B,
	0x7D83,	0x7D7C,	0x7D8C,	0x7D94,	0x7D84,	0x7D7D,	0x7D92,	0x7F6D,
	0x7F6B,	0x7F67,	0x7F68,	0x7F6C,	0x7FA6,	0x7FA5,	0x7FA7,	0x7FDB,
	0x7FDC,	0x8021,	0x8164,	0x8160,	0x8177,	0x815C,	0x8169,	0x815B,
	0x8162,	0x8172,	0x6721,	0x815E,	0x8176,	0x8167,	0x816F,
		0x8144,	0x8161,	0x821D,	0x8249,	0x8244,	0x8240,	0x8242,
	0x8245,	0x84F1,	0x843F,	0x8456,	0x8476,	0x8479,	0x848F,	0x848D,
	0x8465,	0x8451,	0x8440,	0x8486,	0x8467,	0x8430,	0x844D,	0x847D,
	0x845A,	0x8459,	0x8474,	0x8473,	0x845D,	0x8507,	0x845E,	0x8437,
	0x843A,	0x8434,	0x847A,	0x8443,	0x8478,	0x8432,	0x8445,	0x8429,
	0x83D9,	0x844B,	0x842F,	0x8442,	0x842D,	0x845F,	0x8470,	0x8439,
	0x844E,	0x844C,	0x8452,	0x846F,	0x84C5,	0x848E,	0x843B,	0x8447,
	0x8436,	0x8433,	0x8468,	0x847E,	0x8444,	0x842B,	0x8460,	0x8454,
	0x846E,	0x8450,	0x870B,	0x8704,	0x86F7,	0x870C,	0x86FA,	0x86D6,
	0x86F5,	0x874D,	0x86F8,	0x870E,	0x8709,	0x8701,	0x86F6,	0x870D,
	0x8705,	0x88D6,	0x88CB,	0x88CD,	0x88CE,	0x88DE,	0x88DB,	0x88DA,
	0x88CC,	0x88D0,	0x8985,	0x899B,	0x89DF,	0x89E5,	0x89E4,

	/* Big5-HKSCS 0xE040..0xE07E, 0xE0A1..0xE0FE */
	0x89E1,	0x89E0,	0x89E2,	0x89DC,	0x89E6,	0x8A76,	0x8A86,	0x8A7F,
	0x8A61,	0x8A3F,	0x8A77,	0x8A82,	0x8A84,	0x8A75,	0x8A83,	0x8A81,
	0x8A74,	0x8A7A,	0x8C3C,	0x8C4B,	0x8C4A,	0x8C65,	0x8C64,	0x8C66,
	0x8C86,	0x8C84,	0x8C85,	0x8CCC,	0x8D68,	0x8D69,	0x8D91,	0x8D8C,
	0x8D8E,	0x8D8F,	0x8D8D,	0x8D93,	0x8D94,	0x8D90,	0x8D92,	0x8DF0,
	0x8DE0,	0x8DEC,	0x8DF1,	0x8DEE,	0x8DD0,	0x8DE9,	0x8DE3,	0x8DE2,
	0x8DE7,	0x8DF2,	0x8DEB,	0x8DF4,	0x8F06,	0x8EFF,	0x8F01,	0x8F00,
	0x8F05,	0x8F07,	0x8F08,	0x8F02,	0x8F0B,	0x9052,	0x903F,
		0x9044,	0x9049,	0x903D,	0x9110,	0x910D,	0x910F,	0x9111,
	0x9116,	0x9114,	0x910B,	0x910E,	0x916E,	0x916F,	0x9248,	0x9252,
	0x9230,	0x923A,	0x9266,	0x9233,	0x9265,	0x925E,	0x9283,	0x922E,
	0x924A,	0x9246,	0x926D,	0x926C,	0x924F,	0x9260,	0x9267,	0x926F,
	0x9236,	0x9261,	0x9270,	0x9231,	0x9254,	0x9263,	0x9250,	0x9272,
	0x924E,	0x9253,	0x924C,	0x9256,	0x9232,	0x959F,	0x959C,	0x959E,
	0x959B,	0x9692,	0x9693,	0x9691,	0x9697,	0x96CE,	0x96FA,	0x96FD,
	0x96F8,	0x96F5,	0x9773,	0x9777,	0x9778,	0x9772,	0x980F,	0x980D,
	0x980E,	0x98AC,	0x98F6,	0x98F9,	0x99AF,	0x99B2,	0x99B0,	0x99B5,
	0x9AAD,	0x9AAB,	0x9B5B,	0x9CEA,	0x9CED,	0x9CE7,	0x9E80,	0x9EFD,
	0x50E6,	0x50D4,	0x50D7,	0x50E8,	0x50F3,	0x50DB,	0x50EA,	0x50DD,
	0x50E4,	0x50D3,	0x50EC,	0x50F0,	0x50EF,	0x50E3,	0x50E0,

	/* Big5-HKSCS 0xE140..0xE17E, 0xE1A1..0xE1FE */
	0x51D8,	0x5280,	0x5281,	0x52E9,	0x52EB,	0x5330,	0x53AC,	0x5627,
	0x5615,	0x560C,	0x5612,	0x55FC,	0x560F,	0x561C,	0x5601,	0x5613,
	0x5602,	0x55FA,	0x561D,	0x5604,	0x55FF,	0x55F9,	0x5889,	0x587C,
	0x5890,	0x5898,	0x5886,	0x5881,	0x587F,	0x5874,	0x588B,	0x587A,
	0x5887,	0x5891,	0x588E,	0x5876,	0x5882,	0x5888,	0x587B,	0x5894,
	0x588F,	0x58FE,	0x596B,	0x5ADC,	0x5AEE,	0x5AE5,	0x5AD5,	0x5AEA,
	0x5ADA,	0x5AED,	0x5AEB,	0x5AF3,	0x5AE2,	0x5AE0,	0x5ADB,	0x5AEC,
	0x5ADE,	0x5ADD,	0x5AD9,	0x5AE8,	0x5ADF,	0x5B77,	0x5BE0,
		0x5BE3,	0x5C63,	0x5D82,	0x5D80,	0x5D7D,	0x5D86,	0x5D7A,
	0x5D81,	0x5D77,	0x5D8A,	0x5D89,	0x5D88,	0x5D7E,	0x5D7C,	0x5D8D,
	0x5D79,	0x5D7F,	0x5E58,	0x5E59,	0x5E53,	0x5ED8,	0x5ED1,	0x5ED7,
	0x5ECE,	0x5EDC,	0x5ED5,	0x5ED9,	0x5ED2,	0x5ED4,	0x5F44,	0x5F43,
	0x5F6F,	0x5FB6,	0x612C,	0x6128,	0x6141,	0x615E,	0x6171,	0x6173,
	0x6152,	0x6153,	0x6172,	0x616C,	0x6180,	0x6174,	0x6154,	0x617A,
	0x615B,	0x6165,	0x613B,	0x616A,	0x6161,	0x6156,	0x6229,	0x6227,
	0x622B,	0x642B,	0x644D,	0x645B,	0x645D,	0x6474,	0x6476,	0x6472,
	0x6473,	0x647D,	0x6475,	0x6466,	0x64A6,	0x644E,	0x6482,	0x645E,
	0x645C,	0x644B,	0x6453,	0x6460,	0x6450,	0x647F,	0x643F,	0x646C,
	0x646B,	0x6459,	0x6465,	0x6477,	0x6573,	0x65A0,	0x66A1,	0x66A0,
	0x669F,	0x6705,	0x6704,	0x6722,	0x69B1,	0x69B6,	0x69C9,

	/* Big5-HKSCS 0xE240..0xE27E, 0xE2A1..0xE2FE */
	0x69A0,	0x69CE,	0x6996,	0x69B0,	0x69AC,	0x69BC,	0x6991,	0x6999,
	0x698E,	0x69A7,	0x698D,	0x69A9,	0x69BE,	0x69AF,	0x69BF,	0x69C4,
	0x69BD,	0x69A4,	0x69D4,	0x69B9,	0x69CA,	0x699A,	0x69CF,	0x69B3,
	0x6993,	0x69AA,	0x69A1,	0x699E,	0x69D9,	0x6997,	0x6990,	0x69C2,
	0x69B5,	0x69A5,	0x69C6,	0x6B4A,	0x6B4D,	0x6B4B,	0x6B9E,	0x6B9F,
	0x6BA0,	0x6BC3,	0x6BC4,	0x6BFE,	0x6ECE,	0x6EF5,	0x6EF1,	0x6F03,
	0x6F25,	0x6EF8,	0x6F37,	0x6EFB,	0x6F2E,	0x6F09,	0x6F4E,	0x6F19,
	0x6F1A,	0x6F27,	0x6F18,	0x6F3B,	0x6F12,	0x6EED,	0x6F0A,
		0x6F36,	0x6F73,	0x6EF9,	0x6EEE,	0x6F2D,	0x6F40,	0x6F30,
	0x6F3C,	0x6F35,	0x6EEB,	0x6F07,	0x6F0E,	0x6F43,	0x6F05,	0x6EFD,
	0x6EF6,	0x6F39,	0x6F1C,	0x6EFC,	0x6F3A,	0x6F1F,	0x6F0D,	0x6F1E,
	0x6F08,	0x6F21,	0x7187,	0x7190,	0x7189,	0x7180,	0x7185,	0x7182,
	0x718F,	0x717B,	0x7186,	0x7181,	0x7197,	0x7244,	0x7253,	0x7297,
	0x7295,	0x7293,	0x7343,	0x734D,	0x7351,	0x734C,	0x7462,	0x7473,
	0x7471,	0x7475,	0x7472,	0x7467,	0x746E,	0x7500,	0x7502,	0x7503,
	0x757D,	0x7590,	0x7616,	0x7608,	0x760C,	0x7615,	0x7611,	0x760A,
	0x7614,	0x76B8,	0x7781,	0x777C,	0x7785,	0x7782,	0x776E,	0x7780,
	0x776F,	0x777E,	0x7783,	0x78B2,	0x78AA,	0x78B4,	0x78AD,	0x78A8,
	0x787E,	0x78AB,	0x789E,	0x78A5,	0x78A0,	0x78AC,	0x78A2,	0x78A4,
	0x7998,	0x798A,	0x798B,	0x7996,	0x7995,	0x7994,	0x7993,

	/* Big5-HKSCS 0xE340..0xE37E, 0xE3A1..0xE3FE */
	0x7997,	0x7988,	0x7992,	0x7990,	0x7A2B,	0x7A4A,	0x7A30,	0x7A2F,
	0x7A28,	0x7A26,	0x7AA8,	0x7AAB,	0x7AAC,	0x7AEE,	0x7B88,	0x7B9C,
	0x7B8A,	0x7B91,	0x7B90,	0x7B96,	0x7B8D,	0x7B8C,	0x7B9B,	0x7B8E,
	0x7B85,	0x7B98,	0x5284,	0x7B99,	0x7BA4,	0x7B82,	0x7CBB,	0x7CBF,
	0x7CBC,	0x7CBA,	0x7DA7,	0x7DB7,	0x7DC2,	0x7DA3,	0x7DAA,	0x7DC1,
	0x7DC0,	0x7DC5,	0x7D9D,	0x7DCE,	0x7DC4,	0x7DC6,	0x7DCB,	0x7DCC,
	0x7DAF,	0x7DB9,	0x7D96,	0x7DBC,	0x7D9F,	0x7DA6,	0x7DAE,	0x7DA9,
	0x7DA1,	0x7DC9,	0x7F73,	0x7FE2,	0x7FE3,	0x7FE5,	0x7FDE,
		0x8024,	0x805D,	0x805C,	0x8189,	0x8186,	0x8183,	0x8187,
	0x818D,	0x818C,	0x818B,	0x8215,	0x8497,	0x84A4,	0x84A1,	0x849F,
	0x84BA,	0x84CE,	0x84C2,	0x84AC,	0x84AE,	0x84AB,	0x84B9,	0x84B4,
	0x84C1,	0x84CD,	0x84AA,	0x849A,	0x84B1,	0x84D0,	0x849D,	0x84A7,
	0x84BB,	0x84A2,	0x8494,	0x84C7,	0x84CC,	0x849B,	0x84A9,	0x84AF,
	0x84A8,	0x84D6,	0x8498,	0x84B6,	0x84CF,	0x84A0,	0x84D7,	0x84D4,
	0x84D2,	0x84DB,	0x84B0,	0x8491,	0x8661,	0x8733,	0x8723,	0x8728,
	0x876B,	0x8740,	0x872E,	0x871E,	0x8721,	0x8719,	0x871B,	0x8743,
	0x872C,	0x8741,	0x873E,	0x8746,	0x8720,	0x8732,	0x872A,	0x872D,
	0x873C,	0x8712,	0x873A,	0x8731,	0x8735,	0x8742,	0x8726,	0x8727,
	0x8738,	0x8724,	0x871A,	0x8730,	0x8711,	0x88F7,	0x88E7,	0x88F1,
	0x88F2,	0x88FA,	0x88FE,	0x88EE,	0x88FC,	0x88F6,	0x88FB,

	/* Big5-HKSCS 0xE440..0xE47E, 0xE4A1..0xE4FE */
	0x88F0,	0x88EC,	0x88EB,	0x899D,	0x89A1,	0x899F,	0x899E,	0x89E9,
	0x89EB,	0x89E8,	0x8AAB,	0x8A99,	0x8A8B,	0x8A92,	0x8A8F,	0x8A96,
	0x8C3D,	0x8C68,	0x8C69,	0x8CD5,	0x8CCF,	0x8CD7,	0x8D96,	0x8E09,
	0x8E02,	0x8DFF,	0x8E0D,	0x8DFD,	0x8E0A,	0x8E03,	0x8E07,	0x8E06,
	0x8E05,	0x8DFE,	0x8E00,	0x8E04,	0x8F10,	0x8F11,	0x8F0E,	0x8F0D,
	0x9123,	0x911C,	0x9120,	0x9122,	0x911F,	0x911D,	0x911A,	0x9124,
	0x9121,	0x911B,	0x917A,	0x9172,	0x9179,	0x9173,	0x92A5,	0x92A4,
	0x9276,	0x929B,	0x927A,	0x92A0,	0x9294,	0x92AA,	0x928D,
		0x92A6,	0x929A,	0x92AB,	0x9279,	0x9297,	0x927F,	0x92A3,
	0x92EE,	0x928E,	0x9282,	0x9295,	0x92A2,	0x927D,	0x9288,	0x92A1,
	0x928A,	0x9286,	0x928C,	0x9299,	0x92A7,	0x927E,	0x9287,	0x92A9,
	0x929D,	0x928B,	0x922D,	0x969E,	0x96A1,	0x96FF,	0x9758,	0x977D,
	0x977A,	0x977E,	0x9783,	0x9780,	0x9782,	0x977B,	0x9784,	0x9781,
	0x977F,	0x97CE,	0x97CD,	0x9816,	0x98AD,	0x98AE,	0x9902,	0x9900,
	0x9907,	0x999D,	0x999C,	0x99C3,	0x99B9,	0x99BB,	0x99BA,	0x99C2,
	0x99BD,	0x99C7,	0x9AB1,	0x9AE3,	0x9AE7,	0x9B3E,	0x9B3F,	0x9B60,
	0x9B61,	0x9B5F,	0x9CF1,	0x9CF2,	0x9CF5,	0x9EA7,	0x50FF,	0x5103,
	0x5130,	0x50F8,	0x5106,	0x5107,	0x50F6,	0x50FE,	0x510B,	0x510C,
	0x50FD,	0x510A,	0x528B,	0x528C,	0x52F1,	0x52EF,	0x5648,	0x5642,
	0x564C,	0x5635,	0x5641,	0x564A,	0x5649,	0x5646,	0x5658,

	/* Big5-HKSCS 0xE540..0xE57E, 0xE5A1..0xE5FE */
	0x565A,	0x5640,	0x5633,	0x563D,	0x562C,	0x563E,	0x5638,	0x562A,
	0x563A,	0x571A,	0x58AB,	0x589D,	0x58B1,	0x58A0,	0x58A3,	0x58AF,
	0x58AC,	0x58A5,	0x58A1,	0x58FF,	0x5AFF,	0x5AF4,	0x5AFD,	0x5AF7,
	0x5AF6,	0x5B03,	0x5AF8,	0x5B02,	0x5AF9,	0x5B01,	0x5B07,	0x5B05,
	0x5B0F,	0x5C67,	0x5D99,	0x5D97,	0x5D9F,	0x5D92,	0x5DA2,	0x5D93,
	0x5D95,	0x5DA0,	0x5D9C,	0x5DA1,	0x5D9A,	0x5D9E,	0x5E69,	0x5E5D,
	0x5E60,	0x5E5C,	0x7DF3,	0x5EDB,	0x5EDE,	0x5EE1,	0x5F49,	0x5FB2,
	0x618B,	0x6183,	0x6179,	0x61B1,	0x61B0,	0x61A2,	0x6189,
		0x619B,	0x6193,	0x61AF,	0x61AD,	0x619F,	0x6192,	0x61AA,
	0x61A1,	0x618D,	0x6166,	0x61B3,	0x622D,	0x646E,	0x6470,	0x6496,
	0x64A0,	0x6485,	0x6497,	0x649C,	0x648F,	0x648B,	0x648A,	0x648C,
	0x64A3,	0x649F,	0x6468,	0x64B1,	0x6498,	0x6576,	0x657A,	0x6579,
	0x657B,	0x65B2,	0x65B3,	0x66B5,	0x66B0,	0x66A9,	0x66B2,	0x66B7,
	0x66AA,	0x66AF,	0x6A00,	0x6A06,	0x6A17,	0x69E5,	0x69F8,	0x6A15,
	0x69F1,	0x69E4,	0x6A20,	0x69FF,	0x69EC,	0x69E2,	0x6A1B,	0x6A1D,
	0x69FE,	0x6A27,	0x69F2,	0x69EE,	0x6A14,	0x69F7,	0x69E7,	0x6A40,
	0x6A08,	0x69E6,	0x69FB,	0x6A0D,	0x69FC,	0x69EB,	0x6A09,	0x6A04,
	0x6A18,	0x6A25,	0x6A0F,	0x69F6,	0x6A26,	0x6A07,	0x69F4,	0x6A16,
	0x6B51,	0x6BA5,	0x6BA3,	0x6BA2,	0x6BA6,	0x6C01,	0x6C00,	0x6BFF,
	0x6C02,	0x6F41,	0x6F26,	0x6F7E,	0x6F87,	0x6FC6,	0x6F92,

	/* Big5-HKSCS 0xE640..0xE67E, 0xE6A1..0xE6FE */
	0x6F8D,	0x6F89,	0x6F8C,	0x6F62,	0x6F4F,	0x6F85,	0x6F5A,	0x6F96,
	0x6F76,	0x6F6C,	0x6F82,	0x6F55,	0x6F72,	0x6F52,	0x6F50,	0x6F57,
	0x6F94,	0x6F93,	0x6F5D,	0x6F00,	0x6F61,	0x6F6B,	0x6F7D,	0x6F67,
	0x6F90,	0x6F53,	0x6F8B,	0x6F69,	0x6F7F,	0x6F95,	0x6F63,	0x6F77,
	0x6F6A,	0x6F7B,	0x71B2,	0x71AF,	0x719B,	0x71B0,	0x71A0,	0x719A,
	0x71A9,	0x71B5,	0x719D,	0x71A5,	0x719E,	0x71A4,	0x71A1,	0x71AA,
	0x719C,	0x71A7,	0x71B3,	0x7298,	0x729A,	0x7358,	0x7352,	0x735E,
	0x735F,	0x7360,	0x735D,	0x735B,	0x7361,	0x735A,	0x7359,
		0x7362,	0x7487,	0x7489,	0x748A,	0x7486,	0x7481,	0x747D,
	0x7485,	0x7488,	0x747C,	0x7479,	0x7508,	0x7507,	0x757E,	0x7625,
	0x761E,	0x7619,	0x761D,	0x761C,	0x7623,	0x761A,	0x7628,	0x761B,
	0x769C,	0x769D,	0x769E,	0x769B,	0x778D,	0x778F,	0x7789,	0x7788,
	0x78CD,	0x78BB,	0x78CF,	0x78CC,	0x78D1,	0x78CE,	0x78D4,	0x78C8,
	0x78C3,	0x78C4,	0x78C9,	0x799A,	0x79A1,	0x79A0,	0x799C,	0x79A2,
	0x799B,	0x6B76,	0x7A39,	0x7AB2,	0x7AB4,	0x7AB3,	0x7BB7,	0x7BCB,
	0x7BBE,	0x7BAC,	0x7BCE,	0x7BAF,	0x7BB9,	0x7BCA,	0x7BB5,	0x7CC5,
	0x7CC8,	0x7CCC,	0x7CCB,	0x7DF7,	0x7DDB,	0x7DEA,	0x7DE7,	0x7DD7,
	0x7DE1,	0x7E03,	0x7DFA,	0x7DE6,	0x7DF6,	0x7DF1,	0x7DF0,	0x7DEE,
	0x7DDF,	0x7F76,	0x7FAC,	0x7FB0,	0x7FAD,	0x7FED,	0x7FEB,	0x7FEA,
	0x7FEC,	0x7FE6,	0x7FE8,	0x8064,	0x8067,	0x81A3,	0x819F,

	/* Big5-HKSCS 0xE740..0xE77E, 0xE7A1..0xE7FE */
	0x819E,	0x8195,	0x81A2,	0x8199,	0x8197,	0x8216,	0x824F,	0x8253,
	0x8252,	0x8250,	0x824E,	0x8251,	0x8524,	0x853B,	0x850F,	0x8500,
	0x8529,	0x850E,	0x8509,	0x850D,	0x851F,	0x850A,	0x8527,	0x851C,
	0x84FB,	0x852B,	0x84FA,	0x8508,	0x850C,	0x84F4,	0x852A,	0x84F2,
	0x8515,	0x84F7,	0x84EB,	0x84F3,	0x84FC,	0x8512,	0x84EA,	0x84E9,
	0x8516,	0x84FE,	0x8528,	0x851D,	0x852E,	0x8502,	0x84FD,	0x851E,
	0x84F6,	0x8531,	0x8526,	0x84E7,	0x84E8,	0x84F0,	0x84EF,	0x84F9,
	0x8518,	0x8520,	0x8530,	0x850B,	0x8519,	0x852F,	0x8662,
		0x8756,	0x8763,	0x8764,	0x8777,	0x87E1,	0x8773,	0x8758,
	0x8754,	0x875B,	0x8752,	0x8761,	0x875A,	0x8751,	0x875E,	0x876D,
	0x876A,	0x8750,	0x874E,	0x875F,	0x875D,	0x876F,	0x876C,	0x877A,
	0x876E,	0x875C,	0x8765,	0x874F,	0x877B,	0x8775,	0x8762,	0x8767,
	0x8769,	0x885A,	0x8905,	0x890C,	0x8914,	0x890B,	0x8917,	0x8918,
	0x8919,	0x8906,	0x8916,	0x8911,	0x890E,	0x8909,	0x89A2,	0x89A4,
	0x89A3,	0x89ED,	0x89F0,	0x89EC,	0x8ACF,	0x8AC6,	0x8AB8,	0x8AD3,
	0x8AD1,	0x8AD4,	0x8AD5,	0x8ABB,	0x8AD7,	0x8ABE,	0x8AC0,	0x8AC5,
	0x8AD8,	0x8AC3,	0x8ABA,	0x8ABD,	0x8AD9,	0x8C3E,	0x8C4D,	0x8C8F,
	0x8CE5,	0x8CDF,	0x8CD9,	0x8CE8,	0x8CDA,	0x8CDD,	0x8CE7,	0x8DA0,
	0x8D9C,	0x8DA1,	0x8D9B,	0x8E20,	0x8E23,	0x8E25,	0x8E24,	0x8E2E,
	0x8E15,	0x8E1B,	0x8E16,	0x8E11,	0x8E19,	0x8E26,	0x8E27,

	/* Big5-HKSCS 0xE840..0xE87E, 0xE8A1..0xE8FE */
	0x8E14,	0x8E12,	0x8E18,	0x8E13,	0x8E1C,	0x8E17,	0x8E1A,	0x8F2C,
	0x8F24,	0x8F18,	0x8F1A,	0x8F20,	0x8F23,	0x8F16,	0x8F17,	0x9073,
	0x9070,	0x906F,	0x9067,	0x906B,	0x912F,	0x912B,	0x9129,	0x912A,
	0x9132,	0x9126,	0x912E,	0x9185,	0x9186,	0x918A,	0x9181,	0x9182,
	0x9184,	0x9180,	0x92D0,	0x92C3,	0x92C4,	0x92C0,	0x92D9,	0x92B6,
	0x92CF,	0x92F1,	0x92DF,	0x92D8,	0x92E9,	0x92D7,	0x92DD,	0x92CC,
	0x92EF,	0x92C2,	0x92E8,	0x92CA,	0x92C8,	0x92CE,	0x92E6,	0x92CD,
	0x92D5,	0x92C9,	0x92E0,	0x92DE,	0x92E7,	0x92D1,	0x92D3,
		0x92B5,	0x92E1,	0x92C6,	0x92B4,	0x957C,	0x95AC,	0x95AB,
	0x95AE,	0x95B0,	0x96A4,	0x96A2,	0x96D3,	0x9705,	0x9708,	0x9702,
	0x975A,	0x978A,	0x978E,	0x9788,	0x97D0,	0x97CF,	0x981E,	0x981D,
	0x9826,	0x9829,	0x9828,	0x9820,	0x981B,	0x9827,	0x98B2,	0x9908,
	0x98FA,	0x9911,	0x9914,	0x9916,	0x9917,	0x9915,	0x99DC,	0x99CD,
	0x99CF,	0x99D3,	0x99D4,	0x99CE,	0x99C9,	0x99D6,	0x99D8,	0x99CB,
	0x99D7,	0x99CC,	0x9AB3,	0x9AEC,	0x9AEB,	0x9AF3,	0x9AF2,	0x9AF1,
	0x9B46,	0x9B43,	0x9B67,	0x9B74,	0x9B71,	0x9B66,	0x9B76,	0x9B75,
	0x9B70,	0x9B68,	0x9B64,	0x9B6C,	0x9CFC,	0x9CFA,	0x9CFD,	0x9CFF,
	0x9CF7,	0x9D07,	0x9D00,	0x9CF9,	0x9CFB,	0x9D08,	0x9D05,	0x9D04,
	0x9E83,	0x9ED3,	0x9F0F,	0x9F10,	0x511C,	0x5113,	0x5117,	0x511A,
	0x5111,	0x51DE,	0x5334,	0x53E1,	0x5670,	0x5660,	0x566E,

	/* Big5-HKSCS 0xE940..0xE97E, 0xE9A1..0xE9FE */
	0x5673,	0x5666,	0x5663,	0x566D,	0x5672,	0x565E,	0x5677,	0x571C,
	0x571B,	0x58C8,	0x58BD,	0x58C9,	0x58BF,	0x58BA,	0x58C2,	0x58BC,
	0x58C6,	0x5B17,	0x5B19,	0x5B1B,	0x5B21,	0x5B14,	0x5B13,	0x5B10,
	0x5B16,	0x5B28,	0x5B1A,	0x5B20,	0x5B1E,	0x5BEF,	0x5DAC,	0x5DB1,
	0x5DA9,	0x5DA7,	0x5DB5,	0x5DB0,	0x5DAE,	0x5DAA,	0x5DA8,	0x5DB2,
	0x5DAD,	0x5DAF,	0x5DB4,	0x5E67,	0x5E68,	0x5E66,	0x5E6F,	0x5EE9,
	0x5EE7,	0x5EE6,	0x5EE8,	0x5EE5,	0x5F4B,	0x5FBC,	0x619D,	0x61A8,
	0x6196,	0x61C5,	0x61B4,	0x61C6,	0x61C1,	0x61CC,	0x61BA,
		0x61BF,	0x61B8,	0x618C,	0x64D7,	0x64D6,	0x64D0,	0x64CF,
	0x64C9,	0x64BD,	0x6489,	0x64C3,	0x64DB,	0x64F3,	0x64D9,	0x6533,
	0x657F,	0x657C,	0x65A2,	0x66C8,	0x66BE,	0x66C0,	0x66CA,	0x66CB,
	0x66CF,	0x66BD,	0x66BB,	0x66BA,	0x66CC,	0x6723,	0x6A34,	0x6A66,
	0x6A49,	0x6A67,	0x6A32,	0x6A68,	0x6A3E,	0x6A5D,	0x6A6D,	0x6A76,
	0x6A5B,	0x6A51,	0x6A28,	0x6A5A,	0x6A3B,	0x6A3F,	0x6A41,	0x6A6A,
	0x6A64,	0x6A50,	0x6A4F,	0x6A54,	0x6A6F,	0x6A69,	0x6A60,	0x6A3C,
	0x6A5E,	0x6A56,	0x6A55,	0x6A4D,	0x6A4E,	0x6A46,	0x6B55,	0x6B54,
	0x6B56,	0x6BA7,	0x6BAA,	0x6BAB,	0x6BC8,	0x6BC7,	0x6C04,	0x6C03,
	0x6C06,	0x6FAD,	0x6FCB,	0x6FA3,	0x6FC7,	0x6FBC,	0x6FCE,	0x6FC8,
	0x6F5E,	0x6FC4,	0x6FBD,	0x6F9E,	0x6FCA,	0x6FA8,	0x7004,	0x6FA5,
	0x6FAE,	0x6FBA,	0x6FAC,	0x6FAA,	0x6FCF,	0x6FBF,	0x6FB8,

	/* Big5-HKSCS 0xEA40..0xEA7E, 0xEAA1..0xEAFE */
	0x6FA2,	0x6FC9,	0x6FAB,	0x6FCD,	0x6FAF,	0x6FB2,	0x6FB0,	0x71C5,
	0x71C2,	0x71BF,	0x71B8,	0x71D6,	0x71C0,	0x71C1,	0x71CB,	0x71D4,
	0x71CA,	0x71C7,	0x71CF,	0x71BD,	0x71D8,	0x71BC,	0x71C6,	0x71DA,
	0x71DB,	0x729D,	0x729E,	0x7369,	0x7366,	0x7367,	0x736C,	0x7365,
	0x736B,	0x736A,	0x747F,	0x749A,	0x74A0,	0x7494,	0x7492,	0x7495,
	0x74A1,	0x750B,	0x7580,	0x762F,	0x762D,	0x7631,	0x763D,	0x7633,
	0x763C,	0x7635,	0x7632,	0x7630,	0x76BB,	0x76E6,	0x779A,	0x779D,
	0x77A1,	0x779C,	0x779B,	0x77A2,	0x77A3,	0x7795,	0x7799,
		0x7797,	0x78DD,	0x78E9,	0x78E5,	0x78EA,	0x78DE,	0x78E3,
	0x78DB,	0x78E1,	0x78E2,	0x78ED,	0x78DF,	0x78E0,	0x79A4,	0x7A44,
	0x7A48,	0x7A47,	0x7AB6,	0x7AB8,	0x7AB5,	0x7AB1,	0x7AB7,	0x7BDE,
	0x7BE3,	0x7BE7,	0x7BDD,	0x7BD5,	0x7BE5,	0x7BDA,	0x7BE8,	0x7BF9,
	0x7BD4,	0x7BEA,	0x7BE2,	0x7BDC,	0x7BEB,	0x7BD8,	0x7BDF,	0x7CD2,
	0x7CD4,	0x7CD7,	0x7CD0,	0x7CD1,	0x7E12,	0x7E21,	0x7E17,	0x7E0C,
	0x7E1F,	0x7E20,	0x7E13,	0x7E0E,	0x7E1C,	0x7E15,	0x7E1A,	0x7E22,
	0x7E0B,	0x7E0F,	0x7E16,	0x7E0D,	0x7E14,	0x7E25,	0x7E24,	0x7F43,
	0x7F7B,	0x7F7C,	0x7F7A,	0x7FB1,	0x7FEF,	0x802A,	0x8029,	0x806C,
	0x81B1,	0x81A6,	0x81AE,	0x81B9,	0x81B5,	0x81AB,	0x81B0,	0x81AC,
	0x81B4,	0x81B2,	0x81B7,	0x81A7,	0x81F2,	0x8255,	0x8256,	0x8257,
	0x8556,	0x8545,	0x856B,	0x854D,	0x8553,	0x8561,	0x8558,

	/* Big5-HKSCS 0xEB40..0xEB7E, 0xEBA1..0xEBFE */
	0x8540,	0x8546,	0x8564,	0x8541,	0x8562,	0x8544,	0x8551,	0x8547,
	0x8563,	0x853E,	0x855B,	0x8571,	0x854E,	0x856E,	0x8575,	0x8555,
	0x8567,	0x8560,	0x858C,	0x8566,	0x855D,	0x8554,	0x8565,	0x856C,
	0x8663,	0x8665,	0x8664,	0x879B,	0x878F,	0x8797,	0x8793,	0x8792,
	0x8788,	0x8781,	0x8796,	0x8798,	0x8779,	0x8787,	0x87A3,	0x8785,
	0x8790,	0x8791,	0x879D,	0x8784,	0x8794,	0x879C,	0x879A,	0x8789,
	0x891E,	0x8926,	0x8930,	0x892D,	0x892E,	0x8927,	0x8931,	0x8922,
	0x8929,	0x8923,	0x892F,	0x892C,	0x891F,	0x89F1,	0x8AE0,
		0x8AE2,	0x8AF2,	0x8AF4,	0x8AF5,	0x8ADD,	0x8B14,	0x8AE4,
	0x8ADF,	0x8AF0,	0x8AC8,	0x8ADE,	0x8AE1,	0x8AE8,	0x8AFF,	0x8AEF,
	0x8AFB,	0x8C91,	0x8C92,	0x8C90,	0x8CF5,	0x8CEE,	0x8CF1,	0x8CF0,
	0x8CF3,	0x8D6C,	0x8D6E,	0x8DA5,	0x8DA7,	0x8E33,	0x8E3E,	0x8E38,
	0x8E40,	0x8E45,	0x8E36,	0x8E3C,	0x8E3D,	0x8E41,	0x8E30,	0x8E3F,
	0x8EBD,	0x8F36,	0x8F2E,	0x8F35,	0x8F32,	0x8F39,	0x8F37,	0x8F34,
	0x9076,	0x9079,	0x907B,	0x9086,	0x90FA,	0x9133,	0x9135,	0x9136,
	0x9193,	0x9190,	0x9191,	0x918D,	0x918F,	0x9327,	0x931E,	0x9308,
	0x931F,	0x9306,	0x930F,	0x937A,	0x9338,	0x933C,	0x931B,	0x9323,
	0x9312,	0x9301,	0x9346,	0x932D,	0x930E,	0x930D,	0x92CB,	0x931D,
	0x92FA,	0x9325,	0x9313,	0x92F9,	0x92F7,	0x9334,	0x9302,	0x9324,
	0x92FF,	0x9329,	0x9339,	0x9335,	0x932A,	0x9314,	0x930C,

	/* Big5-HKSCS 0xEC40..0xEC7E, 0xECA1..0xECFE */
	0x930B,	0x92FE,	0x9309,	0x9300,	0x92FB,	0x9316,	0x95BC,	0x95CD,
	0x95BE,	0x95B9,	0x95BA,	0x95B6,	0x95BF,	0x95B5,	0x95BD,	0x96A9,
	0x96D4,	0x970B,	0x9712,	0x9710,	0x9799,	0x9797,	0x9794,	0x97F0,
	0x97F8,	0x9835,	0x982F,	0x9832,	0x9924,	0x991F,	0x9927,	0x9929,
	0x999E,	0x99EE,	0x99EC,	0x99E5,	0x99E4,	0x99F0,	0x99E3,	0x99EA,
	0x99E9,	0x99E7,	0x9AB9,	0x9ABF,	0x9AB4,	0x9ABB,	0x9AF6,	0x9AFA,
	0x9AF9,	0x9AF7,	0x9B33,	0x9B80,	0x9B85,	0x9B87,	0x9B7C,	0x9B7E,
	0x9B7B,	0x9B82,	0x9B93,	0x9B92,	0x9B90,	0x9B7A,	0x9B95,
		0x9B7D,	0x9B88,	0x9D25,	0x9D17,	0x9D20,	0x9D1E,	0x9D14,
	0x9D29,	0x9D1D,	0x9D18,	0x9D22,	0x9D10,	0x9D19,	0x9D1F,	0x9E88,
	0x9E86,	0x9E87,	0x9EAE,	0x9EAD,	0x9ED5,	0x9ED6,	0x9EFA,	0x9F12,
	0x9F3D,	0x5126,	0x5125,	0x5122,	0x5124,	0x5120,	0x5129,	0x52F4,
	0x5693,	0x568C,	0x568D,	0x5686,	0x5684,	0x5683,	0x567E,	0x5682,
	0x567F,	0x5681,	0x58D6,	0x58D4,	0x58CF,	0x58D2,	0x5B2D,	0x5B25,
	0x5B32,	0x5B23,	0x5B2C,	0x5B27,	0x5B26,	0x5B2F,	0x5B2E,	0x5B7B,
	0x5BF1,	0x5BF2,	0x5DB7,	0x5E6C,	0x5E6A,	0x5FBE,	0x5FBB,	0x61C3,
	0x61B5,	0x61BC,	0x61E7,	0x61E0,	0x61E5,	0x61E4,	0x61E8,	0x61DE,
	0x64EF,	0x64E9,	0x64E3,	0x64EB,	0x64E4,	0x64E8,	0x6581,	0x6580,
	0x65B6,	0x65DA,	0x66D2,	0x6A8D,	0x6A96,	0x6A81,	0x6AA5,	0x6A89,
	0x6A9F,	0x6A9B,	0x6AA1,	0x6A9E,	0x6A87,	0x6A93,	0x6A8E,

	/* Big5-HKSCS 0xED40..0xED7E, 0xEDA1..0xEDFE */
	0x6A95,	0x6A83,	0x6AA8,	0x6AA4,	0x6A91,	0x6A7F,	0x6AA6,	0x6A9A,
	0x6A85,	0x6A8C,	0x6A92,	0x6B5B,	0x6BAD,	0x6C09,	0x6FCC,	0x6FA9,
	0x6FF4,	0x6FD4,	0x6FE3,	0x6FDC,	0x6FED,	0x6FE7,	0x6FE6,	0x6FDE,
	0x6FF2,	0x6FDD,	0x6FE2,	0x6FE8,	0x71E1,	0x71F1,	0x71E8,	0x71F2,
	0x71E4,	0x71F0,	0x71E2,	0x7373,	0x736E,	0x736F,	0x7497,	0x74B2,
	0x74AB,	0x7490,	0x74AA,	0x74AD,	0x74B1,	0x74A5,	0x74AF,	0x7510,
	0x7511,	0x7512,	0x750F,	0x7584,	0x7643,	0x7648,	0x7649,	0x7647,
	0x76A4,	0x76E9,	0x77B5,	0x77AB,	0x77B2,	0x77B7,	0x77B6,
		0x77B4,	0x77B1,	0x77A8,	0x77F0,	0x78F3,	0x78FD,	0x7902,
	0x78FB,	0x78FC,	0x78F2,	0x7905,	0x78F9,	0x78FE,	0x7904,	0x79AB,
	0x79A8,	0x7A5C,	0x7A5B,	0x7A56,	0x7A58,	0x7A54,	0x7A5A,	0x7ABE,
	0x7AC0,	0x7AC1,	0x7C05,	0x7C0F,	0x7BF2,	0x7C00,	0x7BFF,	0x7BFB,
	0x7C0E,	0x7BF4,	0x7C0B,	0x7BF3,	0x7C02,	0x7C09,	0x7C03,	0x7C01,
	0x7BF8,	0x7BFD,	0x7C06,	0x7BF0,	0x7BF1,	0x7C10,	0x7C0A,	0x7CE8,
	0x7E2D,	0x7E3C,	0x7E42,	0x7E33,	0x9848,	0x7E38,	0x7E2A,	0x7E49,
	0x7E40,	0x7E47,	0x7E29,	0x7E4C,	0x7E30,	0x7E3B,	0x7E36,	0x7E44,
	0x7E3A,	0x7F45,	0x7F7F,	0x7F7E,	0x7F7D,	0x7FF4,	0x7FF2,	0x802C,
	0x81BB,	0x81C4,	0x81CC,	0x81CA,	0x81C5,	0x81C7,	0x81BC,	0x81E9,
	0x825B,	0x825A,	0x825C,	0x8583,	0x8580,	0x858F,	0x85A7,	0x8595,
	0x85A0,	0x858B,	0x85A3,	0x857B,	0x85A4,	0x859A,	0x859E,

	/* Big5-HKSCS 0xEE40..0xEE7E, 0xEEA1..0xEEFE */
	0x8577,	0x857C,	0x8589,	0x85A1,	0x857A,	0x8578,	0x8557,	0x858E,
	0x8596,	0x8586,	0x858D,	0x8599,	0x859D,	0x8581,	0x85A2,	0x8582,
	0x8588,	0x8585,	0x8579,	0x8576,	0x8598,	0x8590,	0x859F,	0x8668,
	0x87BE,	0x87AA,	0x87AD,	0x87C5,	0x87B0,	0x87AC,	0x87B9,	0x87B5,
	0x87BC,	0x87AE,	0x87C9,	0x87C3,	0x87C2,	0x87CC,	0x87B7,	0x87AF,
	0x87C4,	0x87CA,	0x87B4,	0x87B6,	0x87BF,	0x87B8,	0x87BD,	0x87DE,
	0x87B2,	0x8935,	0x8933,	0x893C,	0x893E,	0x8941,	0x8952,	0x8937,
	0x8942,	0x89AD,	0x89AF,	0x89AE,	0x89F2,	0x89F3,	0x8B1E,
		0x8B18,	0x8B16,	0x8B11,	0x8B05,	0x8B0B,	0x8B22,	0x8B0F,
	0x8B12,	0x8B15,	0x8B07,	0x8B0D,	0x8B08,	0x8B06,	0x8B1C,	0x8B13,
	0x8B1A,	0x8C4F,	0x8C70,	0x8C72,	0x8C71,	0x8C6F,	0x8C95,	0x8C94,
	0x8CF9,	0x8D6F,	0x8E4E,	0x8E4D,	0x8E53,	0x8E50,	0x8E4C,	0x8E47,
	0x8F43,	0x8F40,	0x9085,	0x907E,	0x9138,	0x919A,	0x91A2,	0x919B,
	0x9199,	0x919F,	0x91A1,	0x919D,	0x91A0,	0x93A1,	0x9383,	0x93AF,
	0x9364,	0x9356,	0x9347,	0x937C,	0x9358,	0x935C,	0x9376,	0x9349,
	0x9350,	0x9351,	0x9360,	0x936D,	0x938F,	0x934C,	0x936A,	0x9379,
	0x9357,	0x9355,	0x9352,	0x934F,	0x9371,	0x9377,	0x937B,	0x9361,
	0x935E,	0x9363,	0x9367,	0x9380,	0x934E,	0x9359,	0x95C7,	0x95C0,
	0x95C9,	0x95C3,	0x95C5,	0x95B7,	0x96AE,	0x96B0,	0x96AC,	0x9720,
	0x971F,	0x9718,	0x971D,	0x9719,	0x979A,	0x97A1,	0x979C,

	/* Big5-HKSCS 0xEF40..0xEF7E, 0xEFA1..0xEFFE */
	0x979E,	0x979D,	0x97D5,	0x97D4,	0x97F1,	0x9841,	0x9844,	0x984A,
	0x9849,	0x9845,	0x9843,	0x9925,	0x992B,	0x992C,	0x992A,	0x9933,
	0x9932,	0x992F,	0x992D,	0x9931,	0x9930,	0x9998,	0x99A3,	0x99A1,
	0x9A02,	0x99FA,	0x99F4,	0x99F7,	0x99F9,	0x99F8,	0x99F6,	0x99FB,
	0x99FD,	0x99FE,	0x99FC,	0x9A03,	0x9ABE,	0x9AFE,	0x9AFD,	0x9B01,
	0x9AFC,	0x9B48,	0x9B9A,	0x9BA8,	0x9B9E,	0x9B9B,	0x9BA6,	0x9BA1,
	0x9BA5,	0x9BA4,	0x9B86,	0x9BA2,	0x9BA0,	0x9BAF,	0x9D33,	0x9D41,
	0x9D67,	0x9D36,	0x9D2E,	0x9D2F,	0x9D31,	0x9D38,	0x9D30,
		0x9D45,	0x9D42,	0x9D43,	0x9D3E,	0x9D37,	0x9D40,	0x9D3D,
	0x7FF5,	0x9D2D,	0x9E8A,	0x9E89,	0x9E8D,	0x9EB0,	0x9EC8,	0x9EDA,
	0x9EFB,	0x9EFF,	0x9F24,	0x9F23,	0x9F22,	0x9F54,	0x9FA0,	0x5131,
	0x512D,	0x512E,	0x5698,	0x569C,	0x5697,	0x569A,	0x569D,	0x5699,
	0x5970,	0x5B3C,	0x5C69,	0x5C6A,	0x5DC0,	0x5E6D,	0x5E6E,	0x61D8,
	0x61DF,	0x61ED,	0x61EE,	0x61F1,	0x61EA,	0x61F0,	0x61EB,	0x61D6,
	0x61E9,	0x64FF,	0x6504,	0x64FD,	0x64F8,	0x6501,	0x6503,	0x64FC,
	0x6594,	0x65DB,	0x66DA,	0x66DB,	0x66D8,	0x6AC5,	0x6AB9,	0x6ABD,
	0x6AE1,	0x6AC6,	0x6ABA,	0x6AB6,	0x6AB7,	0x6AC7,	0x6AB4,	0x6AAD,
	0x6B5E,	0x6BC9,	0x6C0B,	0x7007,	0x700C,	0x700D,	0x7001,	0x7005,
	0x7014,	0x700E,	0x6FFF,	0x7000,	0x6FFB,	0x7026,	0x6FFC,	0x6FF7,
	0x700A,	0x7201,	0x71FF,	0x71F9,	0x7203,	0x71FD,	0x7376,

	/* Big5-HKSCS 0xF040..0xF07E, 0xF0A1..0xF0FE */
	0x74B8,	0x74C0,	0x74B5,	0x74C1,	0x74BE,	0x74B6,	0x74BB,	0x74C2,
	0x7514,	0x7513,	0x765C,	0x7664,	0x7659,	0x7650,	0x7653,	0x7657,
	0x765A,	0x76A6,	0x76BD,	0x76EC,	0x77C2,	0x77BA,	0x78FF,	0x790C,
	0x7913,	0x7914,	0x7909,	0x7910,	0x7912,	0x7911,	0x79AD,	0x79AC,
	0x7A5F,	0x7C1C,	0x7C29,	0x7C19,	0x7C20,	0x7C1F,	0x7C2D,	0x7C1D,
	0x7C26,	0x7C28,	0x7C22,	0x7C25,	0x7C30,	0x7E5C,	0x7E50,	0x7E56,
	0x7E63,	0x7E58,	0x7E62,	0x7E5F,	0x7E51,	0x7E60,	0x7E57,	0x7E53,
	0x7FB5,	0x7FB3,	0x7FF7,	0x7FF8,	0x8075,	0x81D1,	0x81D2,
		0x81D0,	0x825F,	0x825E,	0x85B4,	0x85C6,	0x85C0,	0x85C3,
	0x85C2,	0x85B3,	0x85B5,	0x85BD,	0x85C7,	0x85C4,	0x85BF,	0x85CB,
	0x85CE,	0x85C8,	0x85C5,	0x85B1,	0x85B6,	0x85D2,	0x8624,	0x85B8,
	0x85B7,	0x85BE,	0x8669,	0x87E7,	0x87E6,	0x87E2,	0x87DB,	0x87EB,
	0x87EA,	0x87E5,	0x87DF,	0x87F3,	0x87E4,	0x87D4,	0x87DC,	0x87D3,
	0x87ED,	0x87D8,	0x87E3,	0x87A4,	0x87D7,	0x87D9,	0x8801,	0x87F4,
	0x87E8,	0x87DD,	0x8953,	0x894B,	0x894F,	0x894C,	0x8946,	0x8950,
	0x8951,	0x8949,	0x8B2A,	0x8B27,	0x8B23,	0x8B33,	0x8B30,	0x8B35,
	0x8B47,	0x8B2F,	0x8B3C,	0x8B3E,	0x8B31,	0x8B25,	0x8B37,	0x8B26,
	0x8B36,	0x8B2E,	0x8B24,	0x8B3B,	0x8B3D,	0x8B3A,	0x8C42,	0x8C75,
	0x8C99,	0x8C98,	0x8C97,	0x8CFE,	0x8D04,	0x8D02,	0x8D00,	0x8E5C,
	0x8E62,	0x8E60,	0x8E57,	0x8E56,	0x8E5E,	0x8E65,	0x8E67,

	/* Big5-HKSCS 0xF140..0xF17E, 0xF1A1..0xF1FE */
	0x8E5B,	0x8E5A,	0x8E61,	0x8E5D,	0x8E69,	0x8E54,	0x8F46,	0x8F47,
	0x8F48,	0x8F4B,	0x9128,	0x913A,	0x913B,	0x913E,	0x91A8,	0x91A5,
	0x91A7,	0x91AF,	0x91AA,	0x93B5,	0x938C,	0x9392,	0x93B7,	0x939B,
	0x939D,	0x9389,	0x93A7,	0x938E,	0x93AA,	0x939E,	0x93A6,	0x9395,
	0x9388,	0x9399,	0x939F,	0x938D,	0x93B1,	0x9391,	0x93B2,	0x93A4,
	0x93A8,	0x93B4,	0x93A3,	0x93A5,	0x95D2,	0x95D3,	0x95D1,	0x96B3,
	0x96D7,	0x96DA,	0x5DC2,	0x96DF,	0x96D8,	0x96DD,	0x9723,	0x9722,
	0x9725,	0x97AC,	0x97AE,	0x97A8,	0x97AB,	0x97A4,	0x97AA,
		0x97A2,	0x97A5,	0x97D7,	0x97D9,	0x97D6,	0x97D8,	0x97FA,
	0x9850,	0x9851,	0x9852,	0x98B8,	0x9941,	0x993C,	0x993A,	0x9A0F,
	0x9A0B,	0x9A09,	0x9A0D,	0x9A04,	0x9A11,	0x9A0A,	0x9A05,	0x9A07,
	0x9A06,	0x9AC0,	0x9ADC,	0x9B08,	0x9B04,	0x9B05,	0x9B29,	0x9B35,
	0x9B4A,	0x9B4C,	0x9B4B,	0x9BC7,	0x9BC6,	0x9BC3,	0x9BBF,	0x9BC1,
	0x9BB5,	0x9BB8,	0x9BD3,	0x9BB6,	0x9BC4,	0x9BB9,	0x9BBD,	0x9D5C,
	0x9D53,	0x9D4F,	0x9D4A,	0x9D5B,	0x9D4B,	0x9D59,	0x9D56,	0x9D4C,
	0x9D57,	0x9D52,	0x9D54,	0x9D5F,	0x9D58,	0x9D5A,	0x9E8E,	0x9E8C,
	0x9EDF,	0x9F01,	0x9F00,	0x9F16,	0x9F25,	0x9F2B,	0x9F2A,	0x9F29,
	0x9F28,	0x9F4C,	0x9F55,	0x5134,	0x5135,	0x5296,	0x52F7,	0x53B4,
	0x56AB,	0x56AD,	0x56A6,	0x56A7,	0x56AA,	0x56AC,	0x58DA,	0x58DD,
	0x58DB,	0x5912,	0x5B3D,	0x5B3E,	0x5B3F,	0x5DC3,	0x5E70,

	/* Big5-HKSCS 0xF240..0xF27E, 0xF2A1..0xF2FE */
	0x5FBF,	0x61FB,	0x6507,	0x6510,	0x650D,	0x6509,	0x650C,	0x650E,
	0x6584,	0x65DE,	0x65DD,	0x66DE,	0x6AE7,	0x6AE0,	0x6ACC,	0x6AD1,
	0x6AD9,	0x6ACB,	0x6ADF,	0x6ADC,	0x6AD0,	0x6AEB,	0x6ACF,	0x6ACD,
	0x6ADE,	0x6B60,	0x6BB0,	0x6C0C,	0x7019,	0x7027,	0x7020,	0x7016,
	0x702B,	0x7021,	0x7022,	0x7023,	0x7029,	0x7017,	0x7024,	0x701C,
	0x702A,	0x720C,	0x720A,	0x7207,	0x7202,	0x7205,	0x72A5,	0x72A6,
	0x72A4,	0x72A3,	0x72A1,	0x74CB,	0x74C5,	0x74B7,	0x74C3,	0x7516,
	0x7660,	0x77C9,	0x77CA,	0x77C4,	0x77F1,	0x791D,	0x791B,
		0x7921,	0x791C,	0x7917,	0x791E,	0x79B0,	0x7A67,	0x7A68,
	0x7C33,	0x7C3C,	0x7C39,	0x7C2C,	0x7C3B,	0x7CEC,	0x7CEA,	0x7E76,
	0x7E75,	0x7E78,	0x7E70,	0x7E77,	0x7E6F,	0x7E7A,	0x7E72,	0x7E74,
	0x7E68,	0x7F4B,	0x7F4A,	0x7F83,	0x7F86,	0x7FB7,	0x7FFD,	0x7FFE,
	0x8078,	0x81D7,	0x81D5,	0x8264,	0x8261,	0x8263,	0x85EB,	0x85F1,
	0x85ED,	0x85D9,	0x85E1,	0x85E8,	0x85DA,	0x85D7,	0x85EC,	0x85F2,
	0x85F8,	0x85D8,	0x85DF,	0x85E3,	0x85DC,	0x85D1,	0x85F0,	0x85E6,
	0x85EF,	0x85DE,	0x85E2,	0x8800,	0x87FA,	0x8803,	0x87F6,	0x87F7,
	0x8809,	0x880C,	0x880B,	0x8806,	0x87FC,	0x8808,	0x87FF,	0x880A,
	0x8802,	0x8962,	0x895A,	0x895B,	0x8957,	0x8961,	0x895C,	0x8958,
	0x895D,	0x8959,	0x8988,	0x89B7,	0x89B6,	0x89F6,	0x8B50,	0x8B48,
	0x8B4A,	0x8B40,	0x8B53,	0x8B56,	0x8B54,	0x8B4B,	0x8B55,

	/* Big5-HKSCS 0xF340..0xF37E, 0xF3A1..0xF3FE */
	0x8B51,	0x8B42,	0x8B52,	0x8B57,	0x8C43,	0x8C77,	0x8C76,	0x8C9A,
	0x8D06,	0x8D07,	0x8D09,	0x8DAC,	0x8DAA,	0x8DAD,	0x8DAB,	0x8E6D,
	0x8E78,	0x8E73,	0x8E6A,	0x8E6F,	0x8E7B,	0x8EC2,	0x8F52,	0x8F51,
	0x8F4F,	0x8F50,	0x8F53,	0x8FB4,	0x9140,	0x913F,	0x91B0,	0x91AD,
	0x93DE,	0x93C7,	0x93CF,	0x93C2,	0x93DA,	0x93D0,	0x93F9,	0x93EC,
	0x93CC,	0x93D9,	0x93A9,	0x93E6,	0x93CA,	0x93D4,	0x93EE,	0x93E3,
	0x93D5,	0x93C4,	0x93CE,	0x93C0,	0x93D2,	0x93E7,	0x957D,	0x95DA,
	0x95DB,	0x96E1,	0x9729,	0x972B,	0x972C,	0x9728,	0x9726,
		0x97B3,	0x97B7,	0x97B6,	0x97DD,	0x97DE,	0x97DF,	0x985C,
	0x9859,	0x985D,	0x9857,	0x98BF,	0x98BD,	0x98BB,	0x98BE,	0x9948,
	0x9947,	0x9943,	0x99A6,	0x99A7,	0x9A1A,	0x9A15,	0x9A25,	0x9A1D,
	0x9A24,	0x9A1B,	0x9A22,	0x9A20,	0x9A27,	0x9A23,	0x9A1E,	0x9A1C,
	0x9A14,	0x9AC2,	0x9B0B,	0x9B0A,	0x9B0E,	0x9B0C,	0x9B37,	0x9BEA,
	0x9BEB,	0x9BE0,	0x9BDE,	0x9BE4,	0x9BE6,	0x9BE2,	0x9BF0,	0x9BD4,
	0x9BD7,	0x9BEC,	0x9BDC,	0x9BD9,	0x9BE5,	0x9BD5,	0x9BE1,	0x9BDA,
	0x9D77,	0x9D81,	0x9D8A,	0x9D84,	0x9D88,	0x9D71,	0x9D80,	0x9D78,
	0x9D86,	0x9D8B,	0x9D8C,	0x9D7D,	0x9D6B,	0x9D74,	0x9D75,	0x9D70,
	0x9D69,	0x9D85,	0x9D73,	0x9D7B,	0x9D82,	0x9D6F,	0x9D79,	0x9D7F,
	0x9D87,	0x9D68,	0x9E94,	0x9E91,	0x9EC0,	0x9EFC,	0x9F2D,	0x9F40,
	0x9F41,	0x9F4D,	0x9F56,	0x9F57,	0x9F58,	0x5337,	0x56B2,

	/* Big5-HKSCS 0xF440..0xF47E, 0xF4A1..0xF4FE */
	0x56B5,	0x56B3,	0x58E3,	0x5B45,	0x5DC6,	0x5DC7,	0x5EEE,	0x5EEF,
	0x5FC0,	0x5FC1,	0x61F9,	0x6517,	0x6516,	0x6515,	0x6513,	0x65DF,
	0x66E8,	0x66E3,	0x66E4,	0x6AF3,	0x6AF0,	0x6AEA,	0x6AE8,	0x6AF9,
	0x6AF1,	0x6AEE,	0x6AEF,	0x703C,	0x7035,	0x702F,	0x7037,	0x7034,
	0x7031,	0x7042,	0x7038,	0x703F,	0x703A,	0x7039,	0x7040,	0x703B,
	0x7033,	0x7041,	0x7213,	0x7214,	0x72A8,	0x737D,	0x737C,	0x74BA,
	0x76AB,	0x76AA,	0x76BE,	0x76ED,	0x77CC,	0x77CE,	0x77CF,	0x77CD,
	0x77F2,	0x7925,	0x7923,	0x7927,	0x7928,	0x7924,	0x7929,
		0x79B2,	0x7A6E,	0x7A6C,	0x7A6D,	0x7AF7,	0x7C49,	0x7C48,
	0x7C4A,	0x7C47,	0x7C45,	0x7CEE,	0x7E7B,	0x7E7E,	0x7E81,	0x7E80,
	0x7FBA,	0x7FFF,	0x8079,	0x81DB,	0x81D9,	0x820B,	0x8268,	0x8269,
	0x8622,	0x85FF,	0x8601,	0x85FE,	0x861B,	0x8600,	0x85F6,	0x8604,
	0x8609,	0x8605,	0x860C,	0x85FD,	0x8819,	0x8810,	0x8811,	0x8817,
	0x8813,	0x8816,	0x8963,	0x8966,	0x89B9,	0x89F7,	0x8B60,	0x8B6A,
	0x8B5D,	0x8B68,	0x8B63,	0x8B65,	0x8B67,	0x8B6D,	0x8DAE,	0x8E86,
	0x8E88,	0x8E84,	0x8F59,	0x8F56,	0x8F57,	0x8F55,	0x8F58,	0x8F5A,
	0x908D,	0x9143,	0x9141,	0x91B7,	0x91B5,	0x91B2,	0x91B3,	0x940B,
	0x9413,	0x93FB,	0x9420,	0x940F,	0x9414,	0x93FE,	0x9415,	0x9410,
	0x9428,	0x9419,	0x940D,	0x93F5,	0x9400,	0x93F7,	0x9407,	0x940E,
	0x9416,	0x9412,	0x93FA,	0x9409,	0x93F8,	0x940A,	0x93FF,

	/* Big5-HKSCS 0xF540..0xF57E, 0xF5A1..0xF5FE */
	0x93FC,	0x940C,	0x93F6,	0x9411,	0x9406,	0x95DE,	0x95E0,	0x95DF,
	0x972E,	0x972F,	0x97B9,	0x97BB,	0x97FD,	0x97FE,	0x9860,	0x9862,
	0x9863,	0x985F,	0x98C1,	0x98C2,	0x9950,	0x994E,	0x9959,	0x994C,
	0x994B,	0x9953,	0x9A32,	0x9A34,	0x9A31,	0x9A2C,	0x9A2A,	0x9A36,
	0x9A29,	0x9A2E,	0x9A38,	0x9A2D,	0x9AC7,	0x9ACA,	0x9AC6,	0x9B10,
	0x9B12,	0x9B11,	0x9C0B,	0x9C08,	0x9BF7,	0x9C05,	0x9C12,	0x9BF8,
	0x9C40,	0x9C07,	0x9C0E,	0x9C06,	0x9C17,	0x9C14,	0x9C09,	0x9D9F,
	0x9D99,	0x9DA4,	0x9D9D,	0x9D92,	0x9D98,	0x9D90,	0x9D9B,
		0x9DA0,	0x9D94,	0x9D9C,	0x9DAA,	0x9D97,	0x9DA1,	0x9D9A,
	0x9DA2,	0x9DA8,	0x9D9E,	0x9DA3,	0x9DBF,	0x9DA9,	0x9D96,	0x9DA6,
	0x9DA7,	0x9E99,	0x9E9B,	0x9E9A,	0x9EE5,	0x9EE4,	0x9EE7,	0x9EE6,
	0x9F30,	0x9F2E,	0x9F5B,	0x9F60,	0x9F5E,	0x9F5D,	0x9F59,	0x9F91,
	0x513A,	0x5139,	0x5298,	0x5297,	0x56C3,	0x56BD,	0x56BE,	0x5B48,
	0x5B47,	0x5DCB,	0x5DCF,	0x5EF1,	0x61FD,	0x651B,	0x6B02,	0x6AFC,
	0x6B03,	0x6AF8,	0x6B00,	0x7043,	0x7044,	0x704A,	0x7048,	0x7049,
	0x7045,	0x7046,	0x721D,	0x721A,	0x7219,	0x737E,	0x7517,	0x766A,
	0x77D0,	0x792D,	0x7931,	0x792F,	0x7C54,	0x7C53,	0x7CF2,	0x7E8A,
	0x7E87,	0x7E88,	0x7E8B,	0x7E86,	0x7E8D,	0x7F4D,	0x7FBB,	0x8030,
	0x81DD,	0x8618,	0x862A,	0x8626,	0x861F,	0x8623,	0x861C,	0x8619,
	0x8627,	0x862E,	0x8621,	0x8620,	0x8629,	0x861E,	0x8625,

	/* Big5-HKSCS 0xF640..0xF67E, 0xF6A1..0xF6FE */
	0x8829,	0x881D,	0x881B,	0x8820,	0x8824,	0x881C,	0x882B,	0x884A,
	0x896D,	0x8969,	0x896E,	0x896B,	0x89FA,	0x8B79,	0x8B78,	0x8B45,
	0x8B7A,	0x8B7B,	0x8D10,	0x8D14,	0x8DAF,	0x8E8E,	0x8E8C,	0x8F5E,
	0x8F5B,	0x8F5D,	0x9146,	0x9144,	0x9145,	0x91B9,	0x943F,	0x943B,
	0x9436,	0x9429,	0x943D,	0x943C,	0x9430,	0x9439,	0x942A,	0x9437,
	0x942C,	0x9440,	0x9431,	0x95E5,	0x95E4,	0x95E3,	0x9735,	0x973A,
	0x97BF,	0x97E1,	0x9864,	0x98C9,	0x98C6,	0x98C0,	0x9958,	0x9956,
	0x9A39,	0x9A3D,	0x9A46,	0x9A44,	0x9A42,	0x9A41,	0x9A3A,
		0x9A3F,	0x9ACD,	0x9B15,	0x9B17,	0x9B18,	0x9B16,	0x9B3A,
	0x9B52,	0x9C2B,	0x9C1D,	0x9C1C,	0x9C2C,	0x9C23,	0x9C28,	0x9C29,
	0x9C24,	0x9C21,	0x9DB7,	0x9DB6,	0x9DBC,	0x9DC1,	0x9DC7,	0x9DCA,
	0x9DCF,	0x9DBE,	0x9DC5,	0x9DC3,	0x9DBB,	0x9DB5,	0x9DCE,	0x9DB9,
	0x9DBA,	0x9DAC,	0x9DC8,	0x9DB1,	0x9DAD,	0x9DCC,	0x9DB3,	0x9DCD,
	0x9DB2,	0x9E7A,	0x9E9C,	0x9EEB,	0x9EEE,	0x9EED,	0x9F1B,	0x9F18,
	0x9F1A,	0x9F31,	0x9F4E,	0x9F65,	0x9F64,	0x9F92,	0x4EB9,	0x56C6,
	0x56C5,	0x56CB,	0x5971,	0x5B4B,	0x5B4C,	0x5DD5,	0x5DD1,	0x5EF2,
	0x6521,	0x6520,	0x6526,	0x6522,	0x6B0B,	0x6B08,	0x6B09,	0x6C0D,
	0x7055,	0x7056,	0x7057,	0x7052,	0x721E,	0x721F,	0x72A9,	0x737F,
	0x74D8,	0x74D5,	0x74D9,	0x74D7,	0x766D,	0x76AD,	0x7935,	0x79B4,
	0x7A70,	0x7A71,	0x7C57,	0x7C5C,	0x7C59,	0x7C5B,	0x7C5A,

	/* Big5-HKSCS 0xF740..0xF77E, 0xF7A1..0xF7FE */
	0x7CF4,	0x7CF1,	0x7E91,	0x7F4F,	0x7F87,	0x81DE,	0x826B,	0x8634,
	0x8635,	0x8633,	0x862C,	0x8632,	0x8636,	0x882C,	0x8828,	0x8826,
	0x882A,	0x8825,	0x8971,	0x89BF,	0x89BE,	0x89FB,	0x8B7E,	0x8B84,
	0x8B82,	0x8B86,	0x8B85,	0x8B7F,	0x8D15,	0x8E95,	0x8E94,	0x8E9A,
	0x8E92,	0x8E90,	0x8E96,	0x8E97,	0x8F60,	0x8F62,	0x9147,	0x944C,
	0x9450,	0x944A,	0x944B,	0x944F,	0x9447,	0x9445,	0x9448,	0x9449,
	0x9446,	0x973F,	0x97E3,	0x986A,	0x9869,	0x98CB,	0x9954,	0x995B,
	0x9A4E,	0x9A53,	0x9A54,	0x9A4C,	0x9A4F,	0x9A48,	0x9A4A,
		0x9A49,	0x9A52,	0x9A50,	0x9AD0,	0x9B19,	0x9B2B,	0x9B3B,
	0x9B56,	0x9B55,	0x9C46,	0x9C48,	0x9C3F,	0x9C44,	0x9C39,	0x9C33,
	0x9C41,	0x9C3C,	0x9C37,	0x9C34,	0x9C32,	0x9C3D,	0x9C36,	0x9DDB,
	0x9DD2,	0x9DDE,	0x9DDA,	0x9DCB,	0x9DD0,	0x9DDC,	0x9DD1,	0x9DDF,
	0x9DE9,	0x9DD9,	0x9DD8,	0x9DD6,	0x9DF5,	0x9DD5,	0x9DDD,	0x9EB6,
	0x9EF0,	0x9F35,	0x9F33,	0x9F32,	0x9F42,	0x9F6B,	0x9F95,	0x9FA2,
	0x513D,	0x5299,	0x58E8,	0x58E7,	0x5972,	0x5B4D,	0x5DD8,	0x882F,
	0x5F4F,	0x6201,	0x6203,	0x6204,	0x6529,	0x6525,	0x6596,	0x66EB,
	0x6B11,	0x6B12,	0x6B0F,	0x6BCA,	0x705B,	0x705A,	0x7222,	0x7382,
	0x7381,	0x7383,	0x7670,	0x77D4,	0x7C67,	0x7C66,	0x7E95,	0x826C,
	0x863A,	0x8640,	0x8639,	0x863C,	0x8631,	0x863B,	0x863E,	0x8830,
	0x8832,	0x882E,	0x8833,	0x8976,	0x8974,	0x8973,	0x89FE,

	/* Big5-HKSCS 0xF840..0xF87E, 0xF8A1..0xF8FE */
	0x8B8C,	0x8B8E,	0x8B8B,	0x8B88,	0x8C45,	0x8D19,	0x8E98,	0x8F64,
	0x8F63,	0x91BC,	0x9462,	0x9455,	0x945D,	0x9457,	0x945E,	0x97C4,
	0x97C5,	0x9800,	0x9A56,	0x9A59,	0x9B1E,	0x9B1F,	0x9B20,	0x9C52,
	0x9C58,	0x9C50,	0x9C4A,	0x9C4D,	0x9C4B,	0x9C55,	0x9C59,	0x9C4C,
	0x9C4E,	0x9DFB,	0x9DF7,	0x9DEF,	0x9DE3,	0x9DEB,	0x9DF8,	0x9DE4,
	0x9DF6,	0x9DE1,	0x9DEE,	0x9DE6,	0x9DF2,	0x9DF0,	0x9DE2,	0x9DEC,
	0x9DF4,	0x9DF3,	0x9DE8,	0x9DED,	0x9EC2,	0x9ED0,	0x9EF2,	0x9EF3,
	0x9F06,	0x9F1C,	0x9F38,	0x9F37,	0x9F36,	0x9F43,	0x9F4F,
		0x9F71,	0x9F70,	0x9F6E,	0x9F6F,	0x56D3,	0x56CD,	0x5B4E,
	0x5C6D,	0x652D,	0x66ED,	0x66EE,	0x6B13,	0x705F,	0x7061,	0x705D,
	0x7060,	0x7223,	0x74DB,	0x74E5,	0x77D5,	0x7938,	0x79B7,	0x79B6,
	0x7C6A,	0x7E97,	0x7F89,	0x826D,	0x8643,	0x8838,	0x8837,	0x8835,
	0x884B,	0x8B94,	0x8B95,	0x8E9E,	0x8E9F,	0x8EA0,	0x8E9D,	0x91BE,
	0x91BD,	0x91C2,	0x946B,	0x9468,	0x9469,	0x96E5,	0x9746,	0x9743,
	0x9747,	0x97C7,	0x97E5,	0x9A5E,	0x9AD5,	0x9B59,	0x9C63,	0x9C67,
	0x9C66,	0x9C62,	0x9C5E,	0x9C60,	0x9E02,	0x9DFE,	0x9E07,	0x9E03,
	0x9E06,	0x9E05,	0x9E00,	0x9E01,	0x9E09,	0x9DFF,	0x9DFD,	0x9E04,
	0x9EA0,	0x9F1E,	0x9F46,	0x9F74,	0x9F75,	0x9F76,	0x56D4,	0x652E,
	0x65B8,	0x6B18,	0x6B19,	0x6B17,	0x6B1A,	0x7062,	0x7226,	0x72AA,
	0x77D8,	0x77D9,	0x7939,	0x7C69,	0x7C6B,	0x7CF6,	0x7E9A,

	/* Big5-HKSCS 0xF940..0xF97E, 0xF9A1..0xF9FE */
	0x7E98,	0x7E9B,	0x7E99,	0x81E0,	0x81E1,	0x8646,	0x8647,	0x8648,
	0x8979,	0x897A,	0x897C,	0x897B,	0x89FF,	0x8B98,	0x8B99,	0x8EA5,
	0x8EA4,	0x8EA3,	0x946E,	0x946D,	0x946F,	0x9471,	0x9473,	0x9749,
	0x9872,	0x995F,	0x9C68,	0x9C6E,	0x9C6D,	0x9E0B,	0x9E0D,	0x9E10,
	0x9E0F,	0x9E12,	0x9E11,	0x9EA1,	0x9EF5,	0x9F09,	0x9F47,	0x9F78,
	0x9F7B,	0x9F7A,	0x9F79,	0x571E,	0x7066,	0x7C6F,	0x883C,	0x8DB2,
	0x8EA6,	0x91C3,	0x9474,	0x9478,	0x9476,	0x9475,	0x9A60,	0x9C74,
	0x9C73,	0x9C71,	0x9C75,	0x9E14,	0x9E13,	0x9EF6,	0x9F0A,
		0x9FA4,	0x7068,	0x7065,	0x7CF7,	0x866A,	0x883E,	0x883D,
	0x883F,	0x8B9E,	0x8C9C,	0x8EA9,	0x8EC9,	0x974B,	0x9873,	0x9874,
	0x98CC,	0x9961,	0x99AB,	0x9A64,	0x9A66,	0x9A67,	0x9B24,	0x9E15,
	0x9E17,	0x9F48,	0x6207,	0x6B1E,	0x7227,	0x864C,	0x8EA8,	0x9482,
	0x9480,	0x9481,	0x9A69,	0x9A68,	0x9B2E,	0x9E19,	0x7229,	0x864B,
	0x8B9F,	0x9483,	0x9C79,	0x9EB7,	0x7675,	0x9A6B,	0x9C7A,	0x9E1D,
	0x7069,	0x706A,	0x9EA4,	0x9F7E,	0x9F49,	0x9F98,	0x7881,	0x92B9,
	0x88CF,	0x58BB,	0x6052,	0x7CA7,	0x5AFA,	0x2554,	0x2566,	0x2557,
	0x2560,	0x256C,	0x2563,	0x255A,	0x2569,	0x255D,	0x2552,	0x2564,
	0x2555,	0x255E,	0x256A,	0x2561,	0x2558,	0x2567,	0x255B,	0x2553,
	0x2565,	0x2556,	0x255F,	0x256B,	0x2562,	0x2559,	0x2568,	0x255C,
	0x2551,	0x2550,	0x256D,	0x256E,	0x2570,	0x256F,	0xFFED,

	/* Big5-HKSCS 0xFA40..0xFA7E, 0xFAA1..0xFAFE */
	0xE000,	0x92DB,	0xE002,	0xE003,	0x854C,	0x42B5,	0x73EF,	0x51B5,
	0x3649,	0xE009,	0xE00A,	0x9344,	0xE00C,	0x82EE,	0xE00E,	0x783C,
	0x6744,	0x62DF,	0xE012,	0xE013,	0xE014,	0xE015,	0xE016,	0x4FAB,
	0xE018,	0x5008,	0xE01A,	0xE01B,	0xE01C,	0xE01D,	0xE01E,	0x5029,
	0xE020,	0x5FA4,	0xE022,	0xE023,	0x6EDB,	0xE025,	0x507D,	0x5101,
	0x347A,	0x510E,	0x986C,	0x3743,	0x8416,	0xE02D,	0xE02E,	0x5160,
	0xE030,	0x516A,	0xE032,	0xE033,	0xE034,	0xE035,	0xE036,	0xE037,
	0xE038,	0x5B82,	0x877D,	0xE03B,	0xE03C,	0x51B2,	0x51B8,
		0x9D34,	0x51C9,	0x51CF,	0x51D1,	0x3CDC,	0x51D3,	0xE045,
	0x51B3,	0x51E2,	0x5342,	0x51ED,	0x83CD,	0x693E,	0xE04C,	0x5F7B,
	0x520B,	0x5226,	0x523C,	0x52B5,	0x5257,	0x5294,	0x52B9,	0x52C5,
	0x7C15,	0x8542,	0x52E0,	0x860D,	0xE05A,	0x5305,	0xE05C,	0x5549,
	0x6ED9,	0xE05F,	0xE060,	0xE061,	0x5333,	0x5344,	0xE064,	0x6CCB,
	0xE066,	0x681B,	0x73D5,	0x604A,	0x3EAA,	0x38CC,	0xE06C,	0x71DD,
	0x44A2,	0x536D,	0x5374,	0xE071,	0x537E,	0x537F,	0xE074,	0xE075,
	0x77E6,	0x5393,	0xE078,	0x53A0,	0x53AB,	0x53AE,	0x73A7,	0xE07D,
	0x3F59,	0x739C,	0x53C1,	0x53C5,	0x6C49,	0x4E49,	0x57FE,	0x53D9,
	0x3AAB,	0xE087,	0x53E0,	0xE089,	0xE08A,	0x53F6,	0xE08C,	0x5413,
	0x7079,	0x552B,	0x6657,	0x6D5B,	0x546D,	0xE093,	0xE094,	0x555D,
	0x548F,	0x54A4,	0x47A6,	0xE099,	0xE09A,	0x3DB4,	0xE09C,

	/* Big5-HKSCS 0xFB40..0xFB7E, 0xFBA1..0xFBFE */
	0xE09D,	0xE09E,	0x5547,	0x4CED,	0x542F,	0x7417,	0x5586,	0x55A9,
	0x5605,	0xE0A6,	0xE0A7,	0x4552,	0xE0A9,	0x66B3,	0xE0AB,	0x5637,
	0x66CD,	0xE0AE,	0x66A4,	0x66AD,	0x564D,	0x564F,	0x78F1,	0x56F1,
	0x9787,	0x53FE,	0x5700,	0x56EF,	0x56ED,	0xE0BA,	0x3623,	0xE0BC,
	0x5746,	0xE0BE,	0x6C6E,	0x708B,	0x5742,	0x36B1,	0xE0C3,	0x57E6,
	0xE0C5,	0x5803,	0xE0C7,	0xE0C8,	0x5826,	0xE0CA,	0x585C,	0x58AA,
	0x3561,	0x58E0,	0x58DC,	0xE0D0,	0x58FB,	0x5BFF,	0x5743,	0xE0D4,
	0xE0D5,	0x93D3,	0x35A1,	0x591F,	0x68A6,	0x36C3,	0x6E59,
		0xE0DC,	0x5A24,	0x5553,	0xE0DF,	0x8505,	0x59C9,	0xE0E2,
	0xE0E3,	0xE0E4,	0xE0E5,	0x59D9,	0xE0E7,	0xE0E8,	0xE0E9,	0x6D71,
	0xE0EB,	0xE0EC,	0x59F9,	0xE0EE,	0x5AAB,	0x5A63,	0x36E6,	0xE0F2,
	0x5A77,	0x3708,	0x5A96,	0x7465,	0x5AD3,	0xE0F8,	0xE0F9,	0x3D85,
	0xE0FB,	0x3732,	0xE0FD,	0x5E83,	0x52D0,	0x5B76,	0x6588,	0x5B7C,
	0xE103,	0x4004,	0x485D,	0xE106,	0x5BD5,	0x6160,	0xE109,	0xE10A,
	0xE10B,	0x5BF3,	0x5B9D,	0x4D10,	0x5C05,	0xE110,	0x5C13,	0x73CE,
	0x5C14,	0xE114,	0xE115,	0x5C49,	0x48DD,	0x5C85,	0x5CE9,	0x5CEF,
	0x5D8B,	0xE11C,	0xE11D,	0x5D10,	0x5D18,	0x5D46,	0xE121,	0x5CBA,
	0x5DD7,	0x82FC,	0x382D,	0xE126,	0xE127,	0xE128,	0x8287,	0x3836,
	0x3BC2,	0x5E2E,	0x6A8A,	0x5E75,	0x5E7A,	0xE130,	0xE131,	0x53A6,
	0x4EB7,	0x5ED0,	0x53A8,	0xE136,	0x5E09,	0x5EF4,	0xE139,

	/* Big5-HKSCS 0xFC40..0xFC7E, 0xFCA1..0xFCFE */
	0x5EF9,	0x5EFB,	0x38A0,	0x5EFC,	0x683E,	0x941B,	0x5F0D,	0xE141,
	0xE142,	0x3ADE,	0x48AE,	0xE145,	0x5F3A,	0xE147,	0xE148,	0x5F58,
	0xE14A,	0x5F63,	0x97BD,	0xE14D,	0x5F72,	0x9340,	0xE150,	0x5FA7,
	0x5DB6,	0x3D5F,	0xE154,	0xE155,	0xE156,	0xE157,	0x91D6,	0xE159,
	0xE15A,	0x6031,	0x6685,	0xE15D,	0x3963,	0x3DC7,	0x3639,	0x5790,
	0xE162,	0x7971,	0x3E40,	0x609E,	0x60A4,	0x60B3,	0xE168,	0xE169,
	0xE16A,	0x74A4,	0x50E1,	0x5AA0,	0x6164,	0x8424,	0x6142,	0xE171,
	0xE172,	0x6181,	0x51F4,	0xE175,	0x6187,	0x5BAA,	0xE178,
		0xE179,	0x61D3,	0xE17B,	0xE17C,	0x61D0,	0x3932,	0xE17F,
	0xE180,	0x6023,	0x615C,	0x651E,	0x638B,	0xE185,	0x62C5,	0xE187,
	0x62D5,	0xE189,	0x636C,	0xE18B,	0x3A17,	0x6438,	0x63F8,	0xE18F,
	0xE190,	0x6490,	0x6F8A,	0xE193,	0x9814,	0xE195,	0xE196,	0x64E1,
	0x64E5,	0x947B,	0x3A66,	0x643A,	0x3A57,	0x654D,	0x6F16,	0xE19F,
	0xE1A0,	0x6585,	0x656D,	0x655F,	0xE1A4,	0x65B5,	0xE1A6,	0x4B37,
	0x65D1,	0x40D8,	0xE1AA,	0x65E0,	0x65E3,	0x5FDF,	0xE1AE,	0x6618,
	0xE1B0,	0xE1B1,	0x6644,	0xE1B3,	0xE1B4,	0x664B,	0xE1B6,	0x6667,
	0xE1B8,	0x6673,	0x6674,	0xE1BB,	0xE1BC,	0xE1BD,	0xE1BE,	0xE1BF,
	0x77C5,	0xE1C1,	0x99A4,	0x6702,	0xE1C4,	0xE1C5,	0x3B2B,	0x69FA,
	0xE1C8,	0x675E,	0x6767,	0x6762,	0xE1CC,	0xE1CD,	0x67D7,	0x44E9,
	0x6822,	0x6E50,	0x923C,	0x6801,	0xE1D4,	0xE1D5,	0x685D,

	/* Big5-HKSCS 0xFD40..0xFD7E, 0xFDA1..0xFDFE */
	0xE1D7,	0x69E1,	0x6A0B,	0xE1DA,	0x6973,	0x68C3,	0xE1DD,	0x6901,
	0x6900,	0x3D32,	0x3A01,	0xE1E2,	0x3B80,	0x67AC,	0x6961,	0xE1E6,
	0x42FC,	0x6936,	0x6998,	0x3BA1,	0xE1EB,	0x8363,	0x5090,	0x69F9,
	0xE1EF,	0xE1F0,	0x6A45,	0xE1F2,	0x6A9D,	0x3BF3,	0x67B1,	0x6AC8,
	0xE1F7,	0x3C0D,	0x6B1D,	0xE1FA,	0x60DE,	0x6B35,	0x6B74,	0xE1FE,
	0x6EB5,	0xE200,	0xE201,	0xE202,	0x3740,	0x5421,	0xE205,	0x6BE1,
	0xE207,	0x6BDC,	0x6C37,	0xE20A,	0xE20B,	0xE20C,	0x6C5A,	0x8226,
	0x6C79,	0xE210,	0x44C5,	0xE212,	0xE213,	0xE214,	0xE215,
		0xE216,	0x36E5,	0x3CEB,	0xE219,	0x9B83,	0xE21B,	0xE21C,
	0x7F8F,	0x6837,	0xE21F,	0xE220,	0xE221,	0x6D96,	0x6D5C,	0x6E7C,
	0x6F04,	0xE226,	0xE227,	0xE228,	0x8533,	0xE22A,	0x51C7,	0x6C9C,
	0x6E1D,	0x842E,	0xE22F,	0x6E2F,	0xE231,	0x7453,	0xE233,	0x79CC,
	0x6E4F,	0x5A91,	0xE237,	0x6FF8,	0x370D,	0x6F9D,	0xE23B,	0x6EFA,
	0xE23D,	0xE23E,	0x4555,	0x93F0,	0x6F44,	0x6F5C,	0x3D4E,	0x6F74,
	0xE245,	0x3D3B,	0x6F9F,	0xE248,	0x6FD3,	0xE24A,	0xE24B,	0xE24C,
	0xE24D,	0xE24E,	0xE24F,	0x51DF,	0xE251,	0xE252,	0xE253,	0xE254,
	0x704B,	0x707E,	0x70A7,	0x7081,	0x70CC,	0x70D5,	0x70D6,	0x70DF,
	0x4104,	0x3DE8,	0x71B4,	0x7196,	0xE261,	0x712B,	0x7145,	0x5A88,
	0x714A,	0x716E,	0x5C9C,	0xE268,	0x714F,	0x9362,	0xE26B,	0x712C,
	0xE26D,	0xE26E,	0xE26F,	0x71BA,	0xE271,	0x70BD,	0x720E,

	/* Big5-HKSCS 0xFE40..0xFE7E, 0xFEA1..0xFEFE */
	0x9442,	0x7215,	0x5911,	0x9443,	0x7224,	0x9341,	0xE27A,	0x722E,
	0x7240,	0xE27D,	0x68BD,	0x7255,	0x7257,	0x3E55,	0xE282,	0x680D,
	0x6F3D,	0x7282,	0x732A,	0x732B,	0xE288,	0xE289,	0x48ED,	0xE28B,
	0x7328,	0x732E,	0x73CF,	0x73AA,	0xE290,	0xE291,	0x73C9,	0x7449,
	0xE294,	0xE295,	0xE296,	0x6623,	0x36C5,	0xE299,	0xE29A,	0xE29B,
	0x73F7,	0x7415,	0x6903,	0xE29F,	0x7439,	0xE2A1,	0x3ED7,	0x745C,
	0xE2A4,	0x7460,	0xE2A6,	0x7447,	0x73E4,	0x7476,	0x83B9,	0x746C,
	0x3730,	0x7474,	0x93F1,	0x6A2C,	0x7482,	0x4953,	0xE2B2,
		0xE2B3,	0xE2B4,	0xE2B5,	0x5B46,	0xE2B7,	0xE2B8,	0x74C8,
	0xE2BA,	0x750E,	0x74E9,	0x751E,	0xE2BE,	0xE2BF,	0x5BD7,	0xE2C1,
	0x9385,	0x754D,	0x754A,	0x7567,	0x756E,	0xE2C7,	0x3F04,	0xE2C9,
	0x758E,	0x745D,	0x759E,	0x75B4,	0x7602,	0x762C,	0x7651,	0x764F,
	0x766F,	0x7676,	0xE2D4,	0x7690,	0x81EF,	0x37F8,	0xE2D8,	0xE2D9,
	0x76A1,	0x76A5,	0x76B7,	0x76CC,	0xE2DE,	0x8462,	0xE2E0,	0xE2E1,
	0xE2E2,	0x771E,	0x7726,	0x7740,	0x64AF,	0xE2E7,	0x7758,	0xE2E9,
	0x77AF,	0xE2EB,	0xE2EC,	0xE2ED,	0x77F4,	0x7809,	0xE2F0,	0xE2F1,
	0x68CA,	0x78AF,	0x78C7,	0x78D3,	0x96A5,	0x792E,	0xE2F8,	0x78D7,
	0x7934,	0x78B1,	0xE2FC,	0x8FB8,	0x8884,	0xE2FF,	0xE300,	0xE301,
	0x7986,	0x8900,	0x6902,	0x7980,	0xE306,	0x799D,	0xE308,	0x793C,
	0x79A9,	0x6E2A,	0xE30C,	0x3EA8,	0x79C6,	0xE30F,	0x79D4,
};

static const unsigned char ucs4_to_big5_hkscs[69373][2] =
{
  /* U+00A1 */		    "\x00\x00", "\xA2\x46", "\xA2\x47",
  /* U+00A4 */	"\x00\x00", "\xA2\x44", "\x00\x00", "\xA1\xB1",
  /* U+00A8 */	"\xC6\xD8", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+00AC */	"\x00\x00", "\x00\x00", "\x00\x00", "\xA1\xC2",
  /* U+00B0 */	"\xA2\x58", "\xA1\xD3", "\x00\x00", "\x00\x00",
  /* U+00B4 */	"\x00\x00", "\x00\x00", "\x00\x00", "\xA1\x50",
  /* U+00B8 */	"\xA1\x4D", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+00BC */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+00C0 */	"\x88\x59", "\x88\x57", "\x00\x00", "\x00\x00",
  /* U+00C4 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+00C8 */	"\x88\x5D", "\x88\x5B", "\x88\x66", "\x00\x00",
  /* U+00CC */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+00D0 */	"\x00\x00", "\x00\x00", "\x88\x61", "\x88\x5F",
  /* U+00D4 */	"\x00\x00", "\x00\x00", "\x00\x00", "\xA1\xD1",
  /* U+00D8 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+00DC */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+00E0 */	"\x88\x6A", "\x88\x68", "\x00\x00", "\x00\x00",
  /* U+00E4 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+00E8 */	"\x88\x6F", "\x88\x6D", "\x88\xA7", "\x00\x00",
  /* U+00EC */	"\x88\x73", "\x88\x71", "\x00\x00", "\x00\x00",
  /* U+00F0 */	"\x00\x00", "\x00\x00", "\x88\x77", "\x88\x75",
  /* U+00F4 */	"\x00\x00", "\x00\x00", "\x00\x00", "\xA1\xD2",
  /* U+00F8 */	"\xC8\xFB", "\x88\x7B", "\x88\x79", "\x00\x00",
  /* U+00FC */	"\x88\xA2", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+0100 */	"\x88\x56", "\x88\x67", "\x00\x00", "\x00\x00",
  /* U+0104 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+0108 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+010C */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+0110 */	"\x00\x00", "\x00\x00", "\x88\x5A", "\x88\x6C",
  /* U+0114 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+0118 */	"\x00\x00", "\x00\x00", "\x88\x5C", "\x88\x6E",
  /* U+011C */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+0120 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+0124 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+0128 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x88\x70",
  /* U+012C */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+0130 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+0134 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+0138 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+013C */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+0140 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+0144 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+0148 */	"\x00\x00", "\x00\x00", "\x00\x00", "\xC8\xFC",
  /* U+014C */	"\x88\x5E", "\x88\x74", "\x00\x00", "\x00\x00",
  /* U+0150 */	"\x00\x00", "\x00\x00", "\x00\x00", "\xC8\xFA",
  /* U+0154 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+0158 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+015C */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+0160 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+0164 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+0168 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x88\x78",
  /* U+016C */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+0170 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+0174 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+0178 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+017C */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+0180 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+0184 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+0188 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+018C */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+0190 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+0194 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+0198 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+019C */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+01A0 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+01A4 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+01A8 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+01AC */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+01B0 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+01B4 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+01B8 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+01BC */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+01C0 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+01C4 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+01C8 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+01CC */	"\x00\x00", "\x88\x58", "\x88\x69", "\x00\x00",
  /* U+01D0 */	"\x88\x72", "\x88\x60", "\x88\x76", "\x00\x00",
  /* U+01D4 */	"\x88\x7A", "\x00\x00", "\x88\x7C", "\x00\x00",
  /* U+01D8 */	"\x88\x7D", "\x00\x00", "\x88\x7E", "\x00\x00",
  /* U+01DC */	"\x88\xA1", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+01E0 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+01E4 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+01E8 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+01EC */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+01F0 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+01F4 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+01F8 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+01FC */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+0200 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+0204 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+0208 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+020C */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+0210 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+0214 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+0218 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+021C */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+0220 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+0224 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+0228 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+022C */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+0230 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+0234 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+0238 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+023C */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+0240 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+0244 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+0248 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+024C */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+0250 */	"\xC8\xF6", "\x88\x6B", "\x00\x00", "\x00\x00",
  /* U+0254 */	"\xC8\xF8", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+0258 */	"\x00\x00", "\x00\x00", "\x00\x00", "\xC8\xF7",
  /* U+025C */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+0260 */	"\x00\x00", "\x88\xA8", "\x00\x00", "\x00\x00",
  /* U+0264 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+0268 */	"\x00\x00", "\x00\x00", "\xC8\xFE", "\x00\x00",
  /* U+026C */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+0270 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+0274 */	"\x00\x00", "\xC8\xF9", "\x00\x00", "\x00\x00",
  /* U+0278 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+027C */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+0280 */	"\x00\x00", "\x00\x00", "\x00\x00", "\xC8\xF5",
  /* U+0284 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+0288 */	"\x00\x00", "\x00\x00", "\xC8\xFD", "\x00\x00",
  /* U+028C */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+0290 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+0294 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+0298 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+029C */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+02A0 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+02A4 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+02A8 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+02AC */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+02B0 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+02B4 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+02B8 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+02BC */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+02C0 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+02C4 */	"\x00\x00", "\x00\x00", "\xC6\xD9", "\xA3\xBE",
  /* U+02C8 */	"\x00\x00", "\xA3\xBC", "\xA3\xBD", "\xA3\xBF",
  /* U+02CC */	"\x00\x00", "\xA1\xC5", "\x00\x00", "\x00\x00",
  /* U+02D0 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+02D4 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+02D8 */	"\x00\x00", "\xA3\xBB", "\x00\x00", "\x00\x00",
  /* U+02DC */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+02E0 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+02E4 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+02E8 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+02EC */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+02F0 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+02F4 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+02F8 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+02FC */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+0300 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+0304 */	"\x00\x00", "\xA1\xC2",

  /* U+0391 */		    "\xA3\x44", "\xA3\x45", "\xA3\x46",
  /* U+0394 */	"\xA3\x47", "\xA3\x48", "\xA3\x49", "\xA3\x4A",
  /* U+0398 */	"\xA3\x4B", "\xA3\x4C", "\xA3\x4D", "\xA3\x4E",
  /* U+039C */	"\xA3\x4F", "\xA3\x50", "\xA3\x51", "\xA3\x52",
  /* U+03A0 */	"\xA3\x53", "\xA3\x54", "\x00\x00", "\xA3\x55",
  /* U+03A4 */	"\xA3\x56", "\xA3\x57", "\xA3\x58", "\xA3\x59",
  /* U+03A8 */	"\xA3\x5A", "\xA3\x5B", "\x00\x00", "\x00\x00",
  /* U+03AC */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+03B0 */	"\x00\x00", "\xA3\x5C", "\xA3\x5D", "\xA3\x5E",
  /* U+03B4 */	"\xA3\x5F", "\xA3\x60", "\xA3\x61", "\xA3\x62",
  /* U+03B8 */	"\xA3\x63", "\xA3\x64", "\xA3\x65", "\xA3\x66",
  /* U+03BC */	"\xA3\x67", "\xA3\x68", "\xA3\x69", "\xA3\x6A",
  /* U+03C0 */	"\xA3\x6B", "\xA3\x6C", "\x00\x00", "\xA3\x6D",
  /* U+03C4 */	"\xA3\x6E", "\xA3\x6F", "\xA3\x70", "\xA3\x71",
  /* U+03C8 */	"\xA3\x72", "\xA3\x73", "\x00\x00", "\x00\x00",
  /* U+03CC */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+03D0 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+03D4 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+03D8 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+03DC */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+03E0 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+03E4 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+03E8 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+03EC */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+03F0 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+03F4 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+03F8 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+03FC */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+0400 */	"\x00\x00", "\xC7\xF9", "\x00\x00", "\x00\x00",
  /* U+0404 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+0408 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+040C */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+0410 */	"\xC7\xF3", "\xC7\xF4", "\xC7\xF5", "\xC7\xF6",
  /* U+0414 */	"\xC7\xF7", "\xC7\xF8", "\xC7\xFA", "\xC7\xFB",
  /* U+0418 */	"\xC7\xFC", "\xC7\xFD", "\xC7\xFE", "\xC8\x40",
  /* U+041C */	"\xC8\x41", "\xC8\x42", "\xC8\x43", "\xC8\x44",
  /* U+0420 */	"\xC8\x45", "\xC8\x46", "\xC8\x47", "\xC8\x48",
  /* U+0424 */	"\xC8\x49", "\xC8\x4A", "\xC8\x4B", "\xC8\x4C",
  /* U+0428 */	"\xC8\x4D", "\xC8\x4E", "\xC8\x4F", "\xC8\x50",
  /* U+042C */	"\xC8\x51", "\xC8\x52", "\xC8\x53", "\xC8\x54",
  /* U+0430 */	"\xC8\x55", "\xC8\x56", "\xC8\x57", "\xC8\x58",
  /* U+0434 */	"\xC8\x59", "\xC8\x5A", "\xC8\x5C", "\xC8\x5D",
  /* U+0438 */	"\xC8\x5E", "\xC8\x5F", "\xC8\x60", "\xC8\x61",
  /* U+043C */	"\xC8\x62", "\xC8\x63", "\xC8\x64", "\xC8\x65",
  /* U+0440 */	"\xC8\x66", "\xC8\x67", "\xC8\x68", "\xC8\x69",
  /* U+0444 */	"\xC8\x6A", "\xC8\x6B", "\xC8\x6C", "\xC8\x6D",
  /* U+0448 */	"\xC8\x6E", "\xC8\x6F", "\xC8\x70", "\xC8\x71",
  /* U+044C */	"\xC8\x72", "\xC8\x73", "\xC8\x74", "\xC8\x75",
  /* U+0450 */	"\x00\x00", "\xC8\x5B",

  /* U+1EBE */				"\x88\x63", "\x88\xA4",
  /* U+1EC0 */	"\x88\x65", "\x88\xA6",

  /* U+2013 */					    "\xA1\x56",
  /* U+2014 */	"\xA1\x58", "\xA2\x77", "\xA1\xFC", "\x00\x00",
  /* U+2018 */	"\xA1\xA5", "\xA1\xA6", "\x00\x00", "\x00\x00",
  /* U+201C */	"\xA1\xA7", "\xA1\xA8", "\x00\x00", "\x00\x00",
  /* U+2020 */	"\x00\x00", "\x00\x00", "\xA1\x45", "\x00\x00",
  /* U+2024 */	"\xA3\xBB", "\xA1\x4C", "\xA1\x4B", "\xA1\x45",
  /* U+2028 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+202C */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+2030 */	"\x00\x00", "\x00\x00", "\xA1\xAC", "\xA1\xB2",
  /* U+2034 */	"\x00\x00", "\xA1\xAB", "\x00\x00", "\x00\x00",
  /* U+2038 */	"\x00\x00", "\x00\x00", "\x00\x00", "\xA1\xB0",
  /* U+203C */	"\x00\x00", "\x00\x00", "\xA1\xC3",

  /* U+2103 */					    "\xA2\x4A",
  /* U+2104 */	"\x00\x00", "\xA1\xC1", "\x00\x00", "\x00\x00",
  /* U+2108 */	"\x00\x00", "\xA2\x4B", "\x00\x00", "\x00\x00",
  /* U+210C */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+2110 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+2114 */	"\x00\x00", "\x00\x00", "\xC8\xD2", "\x00\x00",
  /* U+2118 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+211C */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+2120 */	"\x00\x00", "\xC8\xD3", "\x00\x00", "\x00\x00",
  /* U+2124 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+2128 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+212C */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+2130 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+2134 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+2138 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+213C */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+2140 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+2144 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+2148 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+214C */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+2150 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+2154 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+2158 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+215C */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+2160 */	"\xA2\xB9", "\xA2\xBA", "\xA2\xBB", "\xA2\xBC",
  /* U+2164 */	"\xA2\xBD", "\xA2\xBE", "\xA2\xBF", "\xA2\xC0",
  /* U+2168 */	"\xA2\xC1", "\xA2\xC2", "\x00\x00", "\x00\x00",
  /* U+216C */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+2170 */	"\xC6\xB5", "\xC6\xB6", "\xC6\xB7", "\xC6\xB8",
  /* U+2174 */	"\xC6\xB9", "\xC6\xBA", "\xC6\xBB", "\xC6\xBC",
  /* U+2178 */	"\xC6\xBD", "\xC6\xBE", "\x00\x00", "\x00\x00",
  /* U+217C */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+2180 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+2184 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+2188 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+218C */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+2190 */	"\xA1\xF6", "\xA1\xF4", "\xA1\xF7", "\xA1\xF5",
  /* U+2194 */	"\x00\x00", "\x00\x00", "\xA1\xF8", "\xA1\xF9",
  /* U+2198 */	"\xA1\xFB", "\xA1\xFA", "\x00\x00", "\x00\x00",
  /* U+219C */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+21A0 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+21A4 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+21A8 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+21AC */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+21B0 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+21B4 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+21B8 */	"\xC8\x77", "\xC8\x78", "\x00\x00", "\x00\x00",
  /* U+21BC */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+21C0 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+21C4 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+21C8 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+21CC */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+21D0 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+21D4 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+21D8 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+21DC */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+21E0 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+21E4 */	"\x00\x00", "\x00\x00", "\x00\x00", "\xC8\x76",
  /* U+21E8 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+21EC */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+21F0 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+21F4 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+21F8 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+21FC */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+2200 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+2204 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+2208 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+220C */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+2210 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+2214 */	"\x00\x00", "\xA2\x41", "\xA2\x42", "\x00\x00",
  /* U+2218 */	"\xA2\x58", "\x00\x00", "\xA1\xD4", "\x00\x00",
  /* U+221C */	"\x00\x00", "\x00\x00", "\xA1\xDB", "\xA1\xE8",
  /* U+2220 */	"\xA1\xE7", "\x00\x00", "\x00\x00", "\xA1\xFD",
  /* U+2224 */	"\x00\x00", "\xA1\xFC", "\x00\x00", "\x00\x00",
  /* U+2228 */	"\x00\x00", "\xA1\xE4", "\xA1\xE5", "\xA1\xEC",
  /* U+222C */	"\x00\x00", "\x00\x00", "\xA1\xED", "\x00\x00",
  /* U+2230 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+2234 */	"\xA1\xEF", "\xA1\xEE", "\x00\x00", "\x00\x00",
  /* U+2238 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+223C */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+2240 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+2244 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+2248 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+224C */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+2250 */	"\x00\x00", "\x00\x00", "\xA1\xDC", "\x00\x00",
  /* U+2254 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+2258 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+225C */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+2260 */	"\xA1\xDA", "\xA1\xDD", "\x00\x00", "\xA1\xDD",
  /* U+2264 */	"\x00\x00", "\x00\x00", "\xA1\xD8", "\xA1\xD9",
  /* U+2268 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+226C */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+2270 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+2274 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+2278 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+227C */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+2280 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+2284 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+2288 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+228C */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+2290 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+2294 */	"\x00\x00", "\xA1\xF2", "\x00\x00", "\x00\x00",
  /* U+2298 */	"\x00\x00", "\xA1\xF3", "\x00\x00", "\x00\x00",
  /* U+229C */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+22A0 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+22A4 */	"\x00\x00", "\xA1\xE6", "\x00\x00", "\x00\x00",
  /* U+22A8 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+22AC */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+22B0 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+22B4 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+22B8 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+22BC */	"\x00\x00", "\x00\x00", "\x00\x00", "\xA1\xE9",
  /* U+22C0 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+22C4 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+22C8 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+22CC */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+22D0 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+22D4 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+22D8 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+22DC */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+22E0 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+22E4 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+22E8 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+22EC */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+22F0 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+22F4 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+22F8 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+22FC */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+2300 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+2304 */	"\x00\x00", "\x00\x00", "\x00\x00", "\xA1\x5B",

  /* U+2460 */	"\xC6\xA1", "\xC6\xA2", "\xC6\xA3", "\xC6\xA4",
  /* U+2464 */	"\xC6\xA5", "\xC6\xA6", "\xC6\xA7", "\xC6\xA8",
  /* U+2468 */	"\xC6\xA9", "\xC6\xAA", "\x00\x00", "\x00\x00",
  /* U+246C */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+2470 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+2474 */	"\xC6\xAB", "\xC6\xAC", "\xC6\xAD", "\xC6\xAE",
  /* U+2478 */	"\xC6\xAF", "\xC6\xB0", "\xC6\xB1", "\xC6\xB2",
  /* U+247C */	"\xC6\xB3", "\xC6\xB4",

  /* U+2500 */	"\xA2\x77", "\x00\x00", "\xA2\x78", "\x00\x00",
  /* U+2504 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+2508 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+250C */	"\xA2\x7A", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+2510 */	"\xA2\x7B", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+2514 */	"\xA2\x7C", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+2518 */	"\xA2\x7D", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+251C */	"\xA2\x75", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+2520 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+2524 */	"\xA2\x74", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+2528 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+252C */	"\xA2\x73", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+2530 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+2534 */	"\xA2\x72", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+2538 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+253C */	"\xA2\x71", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+2540 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+2544 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+2548 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+254C */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+2550 */	"\xF9\xF9", "\xF9\xF8", "\xF9\xE6", "\xF9\xEF",
  /* U+2554 */	"\xF9\xDD", "\xF9\xE8", "\xF9\xF1", "\xF9\xDF",
  /* U+2558 */	"\xF9\xEC", "\xF9\xF5", "\xF9\xE3", "\xF9\xEE",
  /* U+255C */	"\xF9\xF7", "\xF9\xE5", "\xF9\xE9", "\xF9\xF2",
  /* U+2560 */	"\xF9\xE0", "\xF9\xEB", "\xF9\xF4", "\xF9\xE2",
  /* U+2564 */	"\xF9\xE7", "\xF9\xF0", "\xF9\xDE", "\xF9\xED",
  /* U+2568 */	"\xF9\xF6", "\xF9\xE4", "\xF9\xEA", "\xF9\xF3",
  /* U+256C */	"\xF9\xE1", "\xF9\xFA", "\xF9\xFB", "\xF9\xFD",
  /* U+2570 */	"\xF9\xFC", "\xA2\xAC", "\xA2\xAD", "\xA2\xAE",
  /* U+2574 */	"\xA1\x5A", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+2578 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+257C */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+2580 */	"\x00\x00", "\xA2\x62", "\xA2\x63", "\xA2\x64",
  /* U+2584 */	"\xA2\x65", "\xA2\x66", "\xA2\x67", "\xA2\x68",
  /* U+2588 */	"\xA2\x69", "\xA2\x70", "\xA2\x6F", "\xA2\x6E",
  /* U+258C */	"\xA2\x6D", "\xA2\x6C", "\xA2\x6B", "\xA2\x6A",
  /* U+2590 */	"\x00\x00", "\x00\x00", "\x00\x00", "\xF9\xFE",
  /* U+2594 */	"\xA2\x76", "\xA2\x79", "\x00\x00", "\x00\x00",
  /* U+2598 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+259C */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+25A0 */	"\xA1\xBD", "\xA1\xBC", "\x00\x00", "\x00\x00",
  /* U+25A4 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+25A8 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+25AC */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+25B0 */	"\x00\x00", "\x00\x00", "\xA1\xB6", "\xA1\xB5",
  /* U+25B4 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+25B8 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+25BC */	"\xA1\xBF", "\xA1\xBE", "\x00\x00", "\x00\x00",
  /* U+25C0 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+25C4 */	"\x00\x00", "\x00\x00", "\xA1\xBB", "\xA1\xBA",
  /* U+25C8 */	"\x00\x00", "\x00\x00", "\x00\x00", "\xA1\xB3",
  /* U+25CC */	"\x00\x00", "\x00\x00", "\xA1\xB7", "\xA1\xB4",
  /* U+25D0 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+25D4 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+25D8 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+25DC */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+25E0 */	"\x00\x00", "\x00\x00", "\xA2\xA8", "\xA2\xA9",
  /* U+25E4 */	"\xA2\xAB", "\xA2\xAA", "\x00\x00", "\x00\x00",
  /* U+25E8 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+25EC */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+25F0 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+25F4 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+25F8 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+25FC */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+2600 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+2604 */	"\x00\x00", "\xA1\xB9", "\xA1\xB8", "\x00\x00",
  /* U+2608 */	"\x00\x00", "\xA1\xF3", "\x00\x00", "\x00\x00",
  /* U+260C */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+2610 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+2614 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+2618 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+261C */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+2620 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+2624 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+2628 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+262C */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+2630 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+2634 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+2638 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+263C */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+2640 */	"\xA1\xF0", "\xA1\xF2", "\xA1\xF1",

  /* U+273D */		    "\xC6\xE6",

  /* U+2E80 */	"\xC8\xD6", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+2E84 */	"\xC8\xD7", "\x00\x00", "\xC8\xD8", "\xC8\xD9",
  /* U+2E88 */	"\xC8\xDA", "\x00\x00", "\xC8\xDB", "\x00\x00",
  /* U+2E8C */	"\xC8\xDC", "\xC8\xDD", "\x00\x00", "\x00\x00",
  /* U+2E90 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+2E94 */	"\x00\x00", "\xC8\xDE", "\x00\x00", "\x00\x00",
  /* U+2E98 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+2E9C */	"\xC8\xDF", "\xC8\xE0", "\x00\x00", "\x00\x00",
  /* U+2EA0 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+2EA4 */	"\x00\x00", "\xC8\xE1", "\x00\x00", "\xC8\xE2",
  /* U+2EA8 */	"\x00\x00", "\x00\x00", "\xC8\xE3", "\x00\x00",
  /* U+2EAC */	"\xC8\xE4", "\x00\x00", "\xC8\xE5", "\x00\x00",
  /* U+2EB0 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+2EB4 */	"\x00\x00", "\x00\x00", "\xC8\xE6", "\x00\x00",
  /* U+2EB8 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+2EBC */	"\xC8\xE7", "\x00\x00", "\xC8\xE8", "\x00\x00",
  /* U+2EC0 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+2EC4 */	"\x00\x00", "\x00\x00", "\xC8\xE9", "\x00\x00",
  /* U+2EC8 */	"\x00\x00", "\x00\x00", "\xC8\xEA", "\x00\x00",
  /* U+2ECC */	"\xC8\xEB", "\xC8\xEC", "\x00\x00", "\xC8\xED",
  /* U+2ED0 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+2ED4 */	"\x00\x00", "\x00\x00", "\xC8\xEE", "\xC8\xEF",
  /* U+2ED8 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+2EDC */	"\x00\x00", "\x00\x00", "\xC8\xF0", "\x00\x00",
  /* U+2EE0 */	"\x00\x00", "\x00\x00", "\x00\x00", "\xC8\xF1",
  /* U+2EE4 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+2EE8 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+2EEC */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+2EF0 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+2EF4 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+2EF8 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+2EFC */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+2F00 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+2F04 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+2F08 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+2F0C */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+2F10 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+2F14 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+2F18 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+2F1C */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+2F20 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+2F24 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+2F28 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+2F2C */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+2F30 */	"\x00\x00", "\x00\x00", "\x00\x00", "\xC6\xCD",

  /* U+3000 */	"\xA1\x40", "\xA1\x42", "\xA1\x43", "\xA1\xB2",
  /* U+3004 */	"\x00\x00", "\xC6\xE0", "\xC6\xE1", "\xC6\xE2",
  /* U+3008 */	"\xA1\x71", "\xA1\x72", "\xA1\x6D", "\xA1\x6E",
  /* U+300C */	"\xA1\x75", "\xA1\x76", "\xA1\x79", "\xA1\x7A",
  /* U+3010 */	"\xA1\x69", "\xA1\x6A", "\xA2\x45", "\x00\x00",
  /* U+3014 */	"\xA1\x65", "\xA1\x66", "\x00\x00", "\x00\x00",
  /* U+3018 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+301C */	"\xA1\xE3", "\xA1\xA9", "\xA1\xAA", "\x00\x00",
  /* U+3020 */	"\x00\x00", "\xA2\xC3", "\xA2\xC4", "\xA2\xC5",
  /* U+3024 */	"\xA2\xC6", "\xA2\xC7", "\xA2\xC8", "\xA2\xC9",
  /* U+3028 */	"\xA2\xCA", "\xA2\xCB", "\x00\x00", "\x00\x00",
  /* U+302C */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+3030 */	"\xA1\xCA", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+3034 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+3038 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+303C */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+3040 */	"\x00\x00", "\xC6\xE7", "\xC6\xE8", "\xC6\xE9",
  /* U+3044 */	"\xC6\xEA", "\xC6\xEB", "\xC6\xEC", "\xC6\xED",
  /* U+3048 */	"\xC6\xEE", "\xC6\xEF", "\xC6\xF0", "\xC6\xF1",
  /* U+304C */	"\xC6\xF2", "\xC6\xF3", "\xC6\xF4", "\xC6\xF5",
  /* U+3050 */	"\xC6\xF6", "\xC6\xF7", "\xC6\xF8", "\xC6\xF9",
  /* U+3054 */	"\xC6\xFA", "\xC6\xFB", "\xC6\xFC", "\xC6\xFD",
  /* U+3058 */	"\xC6\xFE", "\xC7\x40", "\xC7\x41", "\xC7\x42",
  /* U+305C */	"\xC7\x43", "\xC7\x44", "\xC7\x45", "\xC7\x46",
  /* U+3060 */	"\xC7\x47", "\xC7\x48", "\xC7\x49", "\xC7\x4A",
  /* U+3064 */	"\xC7\x4B", "\xC7\x4C", "\xC7\x4D", "\xC7\x4E",
  /* U+3068 */	"\xC7\x4F", "\xC7\x50", "\xC7\x51", "\xC7\x52",
  /* U+306C */	"\xC7\x53", "\xC7\x54", "\xC7\x55", "\xC7\x56",
  /* U+3070 */	"\xC7\x57", "\xC7\x58", "\xC7\x59", "\xC7\x5A",
  /* U+3074 */	"\xC7\x5B", "\xC7\x5C", "\xC7\x5D", "\xC7\x5E",
  /* U+3078 */	"\xC7\x5F", "\xC7\x60", "\xC7\x61", "\xC7\x62",
  /* U+307C */	"\xC7\x63", "\xC7\x64", "\xC7\x65", "\xC7\x66",
  /* U+3080 */	"\xC7\x67", "\xC7\x68", "\xC7\x69", "\xC7\x6A",
  /* U+3084 */	"\xC7\x6B", "\xC7\x6C", "\xC7\x6D", "\xC7\x6E",
  /* U+3088 */	"\xC7\x6F", "\xC7\x70", "\xC7\x71", "\xC7\x72",
  /* U+308C */	"\xC7\x73", "\xC7\x74", "\xC7\x75", "\xC7\x76",
  /* U+3090 */	"\xC7\x77", "\xC7\x78", "\xC7\x79", "\xC7\x7A",
  /* U+3094 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+3098 */	"\x00\x00", "\x00\x00", "\x00\x00", "\xC8\xD4",
  /* U+309C */	"\xC8\xD5", "\xC6\xDC", "\xC6\xDD", "\x00\x00",
  /* U+30A0 */	"\x00\x00", "\xC7\x7B", "\xC7\x7C", "\xC7\x7D",
  /* U+30A4 */	"\xC7\x7E", "\xC7\xA1", "\xC7\xA2", "\xC7\xA3",
  /* U+30A8 */	"\xC7\xA4", "\xC7\xA5", "\xC7\xA6", "\xC7\xA7",
  /* U+30AC */	"\xC7\xA8", "\xC7\xA9", "\xC7\xAA", "\xC7\xAB",
  /* U+30B0 */	"\xC7\xAC", "\xC7\xAD", "\xC7\xAE", "\xC7\xAF",
  /* U+30B4 */	"\xC7\xB0", "\xC7\xB1", "\xC7\xB2", "\xC7\xB3",
  /* U+30B8 */	"\xC7\xB4", "\xC7\xB5", "\xC7\xB6", "\xC7\xB7",
  /* U+30BC */	"\xC7\xB8", "\xC7\xB9", "\xC7\xBA", "\xC7\xBB",
  /* U+30C0 */	"\xC7\xBC", "\xC7\xBD", "\xC7\xBE", "\xC7\xBF",
  /* U+30C4 */	"\xC7\xC0", "\xC7\xC1", "\xC7\xC2", "\xC7\xC3",
  /* U+30C8 */	"\xC7\xC4", "\xC7\xC5", "\xC7\xC6", "\xC7\xC7",
  /* U+30CC */	"\xC7\xC8", "\xC7\xC9", "\xC7\xCA", "\xC7\xCB",
  /* U+30D0 */	"\xC7\xCC", "\xC7\xCD", "\xC7\xCE", "\xC7\xCF",
  /* U+30D4 */	"\xC7\xD0", "\xC7\xD1", "\xC7\xD2", "\xC7\xD3",
  /* U+30D8 */	"\xC7\xD4", "\xC7\xD5", "\xC7\xD6", "\xC7\xD7",
  /* U+30DC */	"\xC7\xD8", "\xC7\xD9", "\xC7\xDA", "\xC7\xDB",
  /* U+30E0 */	"\xC7\xDC", "\xC7\xDD", "\xC7\xDE", "\xC7\xDF",
  /* U+30E4 */	"\xC7\xE0", "\xC7\xE1", "\xC7\xE2", "\xC7\xE3",
  /* U+30E8 */	"\xC7\xE4", "\xC7\xE5", "\xC7\xE6", "\xC7\xE7",
  /* U+30EC */	"\xC7\xE8", "\xC7\xE9", "\xC7\xEA", "\xC7\xEB",
  /* U+30F0 */	"\xC7\xEC", "\xC7\xED", "\xC7\xEE", "\xC7\xEF",
  /* U+30F4 */	"\xC7\xF0", "\xC7\xF1", "\xC7\xF2", "\x00\x00",
  /* U+30F8 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+30FC */	"\xC6\xE3", "\xC6\xDA", "\xC6\xDB", "\x00\x00",
  /* U+3100 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+3104 */	"\x00\x00", "\xA3\x74", "\xA3\x75", "\xA3\x76",
  /* U+3108 */	"\xA3\x77", "\xA3\x78", "\xA3\x79", "\xA3\x7A",
  /* U+310C */	"\xA3\x7B", "\xA3\x7C", "\xA3\x7D", "\xA3\x7E",
  /* U+3110 */	"\xA3\xA1", "\xA3\xA2", "\xA3\xA3", "\xA3\xA4",
  /* U+3114 */	"\xA3\xA5", "\xA3\xA6", "\xA3\xA7", "\xA3\xA8",
  /* U+3118 */	"\xA3\xA9", "\xA3\xAA", "\xA3\xAB", "\xA3\xAC",
  /* U+311C */	"\xA3\xAD", "\xA3\xAE", "\xA3\xAF", "\xA3\xB0",
  /* U+3120 */	"\xA3\xB1", "\xA3\xB2", "\xA3\xB3", "\xA3\xB4",
  /* U+3124 */	"\xA3\xB5", "\xA3\xB6", "\xA3\xB7", "\xA3\xB8",
  /* U+3128 */	"\xA3\xB9", "\xA3\xBA", "\x00\x00", "\x00\x00",
  /* U+312C */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+3130 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+3134 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+3138 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+313C */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+3140 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+3144 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+3148 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+314C */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+3150 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+3154 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+3158 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+315C */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+3160 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+3164 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+3168 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+316C */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+3170 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+3174 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+3178 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+317C */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+3180 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+3184 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+3188 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+318C */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+3190 */	"\x00\x00", "\x00\x00", "\xA4\x40", "\xA4\x47",
  /* U+3194 */	"\xA4\x54", "\xA5\x7C", "\xA4\x57", "\xA4\xA4",
  /* U+3198 */	"\xA4\x55", "\xA5\xD2", "\xA4\x41", "\xA4\xFE",
  /* U+319C */	"\xA4\x42", "\xA4\xD1", "\xA6\x61", "\xA4\x48",

  /* U+3220 */	"\xA4\x40", "\xA4\x47", "\xA4\x54", "\xA5\x7C",
  /* U+3224 */	"\xA4\xAD", "\xA4\xBB", "\xA4\x43", "\xA4\x4B",
  /* U+3228 */	"\xA4\x45", "\xA4\x51", "\xA4\xEB", "\xA4\xF5",
  /* U+322C */	"\xA4\xF4", "\xA4\xEC", "\xAA\xF7", "\xA4\x67",
  /* U+3230 */	"\xA4\xE9", "\xC8\xD1", "\xA6\xB3", "\xAA\xC0",
  /* U+3234 */	"\xA6\x57", "\xAF\x53", "\xB0\x5D", "\xAF\xAC",
  /* U+3238 */	"\xB3\xD2", "\xA5\x4E", "\xA9\x49", "\xBE\xC7",
  /* U+323C */	"\xBA\xCA", "\xA5\xF8", "\xB8\xEA", "\xA8\xF3",
  /* U+3240 */	"\xB2\xBD", "\xA5\xF0", "\xA6\xDB", "\xA6\xDC",
  /* U+3244 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+3248 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+324C */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+3250 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+3254 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+3258 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+325C */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+3260 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+3264 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+3268 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+326C */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+3270 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+3274 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+3278 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+327C */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+3280 */	"\xA4\x40", "\xA4\x47", "\xA4\x54", "\xA5\x7C",
  /* U+3284 */	"\xA4\xAD", "\xA4\xBB", "\xA4\x43", "\xA4\x4B",
  /* U+3288 */	"\xA4\x45", "\xA4\x51", "\xA4\xEB", "\xA4\xF5",
  /* U+328C */	"\xA4\xF4", "\xA4\xEC", "\xAA\xF7", "\xA4\x67",
  /* U+3290 */	"\xA4\xE9", "\xAE\xE8", "\xA6\xB3", "\xAA\xC0",
  /* U+3294 */	"\xA6\x57", "\xAF\x53", "\xB0\x5D", "\xAF\xAC",
  /* U+3298 */	"\xB3\xD2", "\xAF\xB5", "\xA8\x6B", "\xA4\x6B",
  /* U+329C */	"\xBE\x41", "\xC0\x75", "\xA6\x4C", "\xAA\x60",
  /* U+32A0 */	"\xB6\xB5", "\xA5\xF0", "\xBC\x67", "\xA1\xC0",
  /* U+32A4 */	"\xA4\x57", "\xA4\xA4", "\xA4\x55", "\xA5\xAA",
  /* U+32A8 */	"\xA5\x6B", "\xC2\xE5", "\xA9\x76", "\xBE\xC7",
  /* U+32AC */	"\xBA\xCA", "\xA5\xF8", "\xB8\xEA", "\xA8\xF3",
  /* U+32B0 */	"\xA9\x5D",

  /* U+338E */				"\xA2\x55", "\xA2\x56",
  /* U+3390 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+3394 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+3398 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+339C */	"\xA2\x50", "\xA2\x51", "\xA2\x52", "\x00\x00",
  /* U+33A0 */	"\x00\x00", "\xA2\x54", "\x00\x00", "\x00\x00",
  /* U+33A4 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+33A8 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+33AC */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+33B0 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+33B4 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+33B8 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+33BC */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+33C0 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+33C4 */	"\xA2\x57", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+33C8 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+33CC */	"\x00\x00", "\x00\x00", "\xA2\x53", "\x00\x00",
  /* U+33D0 */	"\x00\x00", "\xA1\xEB", "\xA1\xEA", "\x00\x00",
  /* U+33D4 */	"\x00\x00", "\xA2\x4F", "\x00\x00", "\x00\x00",
  /* U+33D8 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+33DC */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+33E0 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+33E4 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+33E8 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+33EC */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+33F0 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+33F4 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+33F8 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+33FC */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+3400 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+3404 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+3408 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+340C */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+3410 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+3414 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+3418 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+341C */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+3420 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+3424 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+3428 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+342C */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+3430 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+3434 */	"\x00\x00", "\x92\x77", "\x00\x00", "\x00\x00",
  /* U+3438 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+343C */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+3440 */	"\x96\xDF", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+3444 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+3448 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+344C */	"\x89\xD5", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+3450 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+3454 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+3458 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+345C */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+3460 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+3464 */	"\x93\xCD", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+3468 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+346C */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+3470 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x9B\xDF",
  /* U+3474 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+3478 */	"\x00\x00", "\x00\x00", "\xFA\x68", "\x00\x00",
  /* U+347C */	"\x00\x00", "\x89\xDA", "\x8F\x59", "\x00\x00",
  /* U+3480 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+3484 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+3488 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+348C */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+3490 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x89\xDB",
  /* U+3494 */	"\x00\x00", "\x00\x00", "\x8F\x5D", "\x00\x00",
  /* U+3498 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+349C */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+34A0 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+34A4 */	"\x00\x00", "\x89\xDC", "\x00\x00", "\x00\x00",
  /* U+34A8 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+34AC */	"\x00\x00", "\x00\x00", "\x00\x00", "\x96\xF7",
  /* U+34B0 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+34B4 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+34B8 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+34BC */	"\x8A\xDA", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+34C0 */	"\x00\x00", "\x8B\xDC", "\x00\x00", "\x00\x00",
  /* U+34C4 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+34C8 */	"\x97\xDB", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+34CC */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+34D0 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+34D4 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+34D8 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+34DC */	"\x00\x00", "\x00\x00", "\x00\x00", "\x9E\x53",
  /* U+34E0 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+34E4 */	"\x9D\xAA", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+34E8 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+34EC */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+34F0 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+34F4 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+34F8 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x9B\xEA",
  /* U+34FC */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+3500 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+3504 */	"\x00\x00", "\x00\x00", "\x8A\x6E", "\x00\x00",
  /* U+3508 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+350C */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+3510 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+3514 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+3518 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+351C */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+3520 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+3524 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+3528 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+352C */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+3530 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+3534 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+3538 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+353C */	"\x00\x00", "\x00\x00", "\x8B\xC8", "\x00\x00",
  /* U+3540 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+3544 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+3548 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+354C */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+3550 */	"\x00\x00", "\x89\xE8", "\x00\x00", "\x89\xEA",
  /* U+3554 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+3558 */	"\x00\x00", "\x8C\x4B", "\x00\x00", "\x00\x00",
  /* U+355C */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+3560 */	"\x00\x00", "\xFB\x70", "\x00\x00", "\x00\x00",
  /* U+3564 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+3568 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+356C */	"\x00\x00", "\x89\xED", "\x00\x00", "\x00\x00",
  /* U+3570 */	"\x94\xDD", "\x00\x00", "\x89\xEE", "\x00\x00",
  /* U+3574 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x9E\xB4",
  /* U+3578 */	"\x8A\xD3", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+357C */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+3580 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+3584 */	"\x92\xDB", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+3588 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+358C */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+3590 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+3594 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x94\xDB",
  /* U+3598 */	"\x89\xF9", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+359C */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+35A0 */	"\x00\x00", "\xFB\x7A", "\x00\x00", "\x00\x00",
  /* U+35A4 */	"\x00\x00", "\x89\xFB", "\x00\x00", "\x00\x00",
  /* U+35A8 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+35AC */	"\x00\x00", "\x9E\xFC", "\x00\x00", "\x00\x00",
  /* U+35B0 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+35B4 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+35B8 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+35BC */	"\x00\x00", "\x00\x00", "\x00\x00", "\x89\xFC",
  /* U+35C0 */	"\x00\x00", "\x89\xBF", "\x00\x00", "\x00\x00",
  /* U+35C4 */	"\x00\x00", "\x89\xFE", "\x00\x00", "\x89\xE6",
  /* U+35C8 */	"\x00\x00", "\x00\x00", "\x9D\x46", "\x00\x00",
  /* U+35CC */	"\x00\x00", "\x00\x00", "\x9D\xEE", "\x00\x00",
  /* U+35D0 */	"\x00\x00", "\x00\x00", "\xA0\x7E", "\x00\x00",
  /* U+35D4 */	"\x00\x00", "\x00\x00", "\xA0\x68", "\x00\x00",
  /* U+35D8 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x98\xE9",
  /* U+35DC */	"\x00\x00", "\x8B\x68", "\x00\x00", "\x00\x00",
  /* U+35E0 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+35E4 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+35E8 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+35EC */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+35F0 */	"\x00\x00", "\x8D\xFD", "\x8B\xBE", "\x9F\xD9",
  /* U+35F4 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+35F8 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x8A\xEB",
  /* U+35FC */	"\x00\x00", "\x00\x00", "\x9F\xD7", "\x00\x00",
  /* U+3600 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+3604 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+3608 */	"\x00\x00", "\x8B\x6A", "\x00\x00", "\x00\x00",
  /* U+360C */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+3610 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+3614 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+3618 */	"\x9C\x5C", "\x00\x00", "\x8B\xB1", "\x00\x00",
  /* U+361C */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+3620 */	"\x00\x00", "\x00\x00", "\x00\x00", "\xFB\x5E",
  /* U+3624 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+3628 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+362C */	"\x00\x00", "\x9D\xF3", "\x00\x00", "\x00\x00",
  /* U+3630 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+3634 */	"\x00\x00", "\xA0\xD0", "\x00\x00", "\x00\x00",
  /* U+3638 */	"\x00\x00", "\xFC\x66", "\x00\x00", "\x00\x00",
  /* U+363C */	"\x00\x00", "\x00\x00", "\x92\xE9", "\x00\x00",
  /* U+3640 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+3644 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x9A\xEC",
  /* U+3648 */	"\x8F\xAB", "\xFA\x48", "\x00\x00", "\x00\x00",
  /* U+364C */	"\x00\x00", "\x00\x00", "\x8E\x45", "\x00\x00",
  /* U+3650 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+3654 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+3658 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+365C */	"\x00\x00", "\x00\x00", "\x00\x00", "\x9C\x6F",
  /* U+3660 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+3664 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+3668 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+366C */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+3670 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+3674 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+3678 */	"\x00\x00", "\x00\x00", "\x9E\xDE", "\x00\x00",
  /* U+367C */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+3680 */	"\x00\x00", "\x89\xEF", "\x00\x00", "\x00\x00",
  /* U+3684 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+3688 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+368C */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+3690 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+3694 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+3698 */	"\x00\x00", "\x00\x00", "\x96\xE9", "\x00\x00",
  /* U+369C */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+36A0 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+36A4 */	"\x00\x00", "\x9E\xBB", "\x00\x00", "\x00\x00",
  /* U+36A8 */	"\x00\x00", "\x00\x00", "\x94\xDE", "\x00\x00",
  /* U+36AC */	"\x9E\xB8", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+36B0 */	"\x97\xBA", "\xFB\x65", "\x00\x00", "\x00\x00",
  /* U+36B4 */	"\x00\x00", "\x95\xD6", "\x00\x00", "\x00\x00",
  /* U+36B8 */	"\x00\x00", "\x9C\xBB", "\x00\x00", "\x00\x00",
  /* U+36BC */	"\x97\xDA", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+36C0 */	"\x00\x00", "\x8F\x45", "\x00\x00", "\xFB\x7D",
  /* U+36C4 */	"\x91\x58", "\xFE\x64", "\x00\x00", "\x98\x56",
  /* U+36C8 */	"\x9B\x4D", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+36CC */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+36D0 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x93\x5B",
  /* U+36D4 */	"\x95\xC7", "\x00\x00", "\x97\xE7", "\x00\x00",
  /* U+36D8 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+36DC */	"\x00\x00", "\x93\x59", "\x00\x00", "\x00\x00",
  /* U+36E0 */	"\x00\x00", "\x91\xF5", "\x97\xB8", "\x00\x00",
  /* U+36E4 */	"\x00\x00", "\xFD\xA2", "\xFB\xB6", "\x00\x00",
  /* U+36E8 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+36EC */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+36F0 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+36F4 */	"\x00\x00", "\x92\xFA", "\x00\x00", "\x00\x00",
  /* U+36F8 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+36FC */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+3700 */	"\x00\x00", "\x93\x57", "\x00\x00", "\x8B\xA6",
  /* U+3704 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+3708 */	"\xFB\xB9", "\x00\x00", "\x97\xB0", "\x00\x00",
  /* U+370C */	"\x00\x00", "\xFD\xC4", "\x00\x00", "\x00\x00",
  /* U+3710 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+3714 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+3718 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+371C */	"\x9C\xA1", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+3720 */	"\x00\x00", "\x00\x00", "\x91\xF2", "\x91\xF9",
  /* U+3724 */	"\x00\x00", "\x8F\xF1", "\x00\x00", "\x00\x00",
  /* U+3728 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+372C */	"\x97\x45", "\x98\x53", "\x00\x00", "\x00\x00",
  /* U+3730 */	"\xFE\x78", "\x00\x00", "\xFB\xC1", "\x92\x51",
  /* U+3734 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+3738 */	"\x00\x00", "\x00\x00", "\x9D\xAD", "\x00\x00",
  /* U+373C */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+3740 */	"\xFD\x6C", "\x00\x00", "\x00\x00", "\xFA\x6B",
  /* U+3744 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+3748 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+374C */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+3750 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+3754 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+3758 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+375C */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+3760 */	"\x00\x00", "\x00\x00", "\x9B\xC2", "\x00\x00",
  /* U+3764 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+3768 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+376C */	"\x00\x00", "\x00\x00", "\x00\x00", "\x9A\x7B",
  /* U+3770 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+3774 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+3778 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+377C */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+3780 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+3784 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+3788 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+378C */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+3790 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+3794 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x8B\x60",
  /* U+3798 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+379C */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+37A0 */	"\x93\x4B", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+37A4 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+37A8 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+37AC */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+37B0 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+37B4 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+37B8 */	"\x00\x00", "\x9A\xBD", "\x00\x00", "\x00\x00",
  /* U+37BC */	"\x00\x00", "\x00\x00", "\x91\xB7", "\x00\x00",
  /* U+37C0 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+37C4 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+37C8 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+37CC */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+37D0 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+37D4 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+37D8 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+37DC */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+37E0 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+37E4 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+37E8 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+37EC */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+37F0 */	"\x00\x00", "\x00\x00", "\x95\xB4", "\x00\x00",
  /* U+37F4 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+37F8 */	"\xFE\xC5", "\x00\x00", "\x00\x00", "\x9E\xF0",
  /* U+37FC */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+3800 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+3804 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+3808 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+380C */	"\x00\x00", "\x00\x00", "\x00\x00", "\x8D\x64",
  /* U+3810 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+3814 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+3818 */	"\x00\x00", "\x92\x69", "\x00\x00", "\x00\x00",
  /* U+381C */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+3820 */	"\x8D\x67", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+3824 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+3828 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+382C */	"\x00\x00", "\xFB\xEA", "\x00\x00", "\x00\x00",
  /* U+3830 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+3834 */	"\x00\x00", "\x00\x00", "\xFB\xEF", "\x00\x00",
  /* U+3838 */	"\x8D\x68", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+383C */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+3840 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+3844 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+3848 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+384C */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+3850 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+3854 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+3858 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+385C */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+3860 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x93\xEB",
  /* U+3864 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+3868 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+386C */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+3870 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+3874 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+3878 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+387C */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+3880 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+3884 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+3888 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+388C */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+3890 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+3894 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+3898 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+389C */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+38A0 */	"\xFC\x42", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+38A4 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+38A8 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+38AC */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+38B0 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+38B4 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+38B8 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+38BC */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+38C0 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x91\x66",
  /* U+38C4 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+38C8 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+38CC */	"\xFA\xCD", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+38D0 */	"\x00\x00", "\x93\xDD", "\x00\x00", "\x00\x00",
  /* U+38D4 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+38D8 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+38DC */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+38E0 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+38E4 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+38E8 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+38EC */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+38F0 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+38F4 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+38F8 */	"\x00\x00", "\x00\x00", "\x8B\xCC", "\x00\x00",
  /* U+38FC */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+3900 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+3904 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+3908 */	"\x8D\x6D", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+390C */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+3910 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+3914 */	"\x8D\x6E", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+3918 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+391C */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+3920 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+3924 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x96\xA8",
  /* U+3928 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+392C */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+3930 */	"\x00\x00", "\x00\x00", "\xFC\xA6", "\x00\x00",
  /* U+3934 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+3938 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+393C */	"\x00\x00", "\x00\x00", "\x00\x00", "\x8D\x6F",
  /* U+3940 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+3944 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+3948 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+394C */	"\x00\x00", "\x8D\x70", "\x00\x00", "\x00\x00",
  /* U+3950 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+3954 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+3958 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+395C */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+3960 */	"\x00\x00", "\x00\x00", "\x00\x00", "\xFC\x64",
  /* U+3964 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+3968 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+396C */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+3970 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+3974 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+3978 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+397C */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+3980 */	"\x90\x60", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+3984 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+3988 */	"\x00\x00", "\x8D\x74", "\x97\xC3", "\x00\x00",
  /* U+398C */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+3990 */	"\x00\x00", "\x00\x00", "\x8A\xD0", "\x00\x00",
  /* U+3994 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+3998 */	"\x00\x00", "\x92\x74", "\x00\x00", "\x9B\xBE",
  /* U+399C */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+39A0 */	"\x00\x00", "\x9C\xC8", "\x00\x00", "\x00\x00",
  /* U+39A4 */	"\x9C\xBA", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+39A8 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+39AC */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+39B0 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+39B4 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+39B8 */	"\x8D\x78", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+39BC */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+39C0 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+39C4 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+39C8 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+39CC */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+39D0 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+39D4 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+39D8 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+39DC */	"\x9E\xB9", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+39E0 */	"\x00\x00", "\x00\x00", "\x95\x5A", "\x00\x00",
  /* U+39E4 */	"\x00\x00", "\x91\xB4", "\x00\x00", "\x00\x00",
  /* U+39E8 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+39EC */	"\x8A\x48", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+39F0 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+39F4 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+39F8 */	"\x8D\x7D", "\x00\x00", "\x00\x00", "\x8A\x7D",
  /* U+39FC */	"\x00\x00", "\x00\x00", "\x8A\xC2", "\x00\x00",
  /* U+3A00 */	"\x00\x00", "\xFD\x4A", "\x00\x00", "\x8D\xA1",
  /* U+3A04 */	"\x00\x00", "\x00\x00", "\x8A\xD1", "\x00\x00",
  /* U+3A08 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+3A0C */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+3A10 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+3A14 */	"\x00\x00", "\x00\x00", "\x00\x00", "\xFC\xB4",
  /* U+3A18 */	"\x8B\x47", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+3A1C */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+3A20 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+3A24 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+3A28 */	"\x00\x00", "\x93\xA4", "\x9E\xDA", "\x00\x00",
  /* U+3A2C */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+3A30 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+3A34 */	"\x8A\x51", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+3A38 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+3A3C */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+3A40 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+3A44 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+3A48 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x8D\xA6",
  /* U+3A4C */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+3A50 */	"\x00\x00", "\x00\x00", "\x9E\xC5", "\x00\x00",
  /* U+3A54 */	"\x00\x00", "\x00\x00", "\x00\x00", "\xFC\xC4",
  /* U+3A58 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+3A5C */	"\xA0\x78", "\x00\x00", "\x94\xB5", "\x00\x00",
  /* U+3A60 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+3A64 */	"\x00\x00", "\x00\x00", "\xFC\xC2", "\x8A\x6B",
  /* U+3A68 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+3A6C */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+3A70 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+3A74 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+3A78 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+3A7C */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+3A80 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+3A84 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+3A88 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+3A8C */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+3A90 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+3A94 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x8D\xAB",
  /* U+3A98 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+3A9C */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+3AA0 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+3AA4 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+3AA8 */	"\x00\x00", "\x00\x00", "\x00\x00", "\xFA\xE8",
  /* U+3AAC */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+3AB0 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+3AB4 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+3AB8 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+3ABC */	"\x00\x00", "\x8D\xAD", "\x00\x00", "\x00\x00",
  /* U+3AC0 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+3AC4 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+3AC8 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+3ACC */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+3AD0 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+3AD4 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+3AD8 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+3ADC */	"\x00\x00", "\x00\x00", "\xFC\x49", "\x00\x00",
  /* U+3AE0 */	"\x93\xC1", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+3AE4 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+3AE8 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+3AEC */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+3AF0 */	"\x90\x6F", "\x00\x00", "\x8D\xB0", "\x00\x00",
  /* U+3AF4 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+3AF8 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x94\x7E",
  /* U+3AFC */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+3B00 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+3B04 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+3B08 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+3B0C */	"\x00\x00", "\x00\x00", "\x90\xFA", "\x00\x00",
  /* U+3B10 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+3B14 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+3B18 */	"\x00\x00", "\x94\x79", "\x00\x00", "\x00\x00",
  /* U+3B1C */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+3B20 */	"\x00\x00", "\x00\x00", "\x8D\xB2", "\x00\x00",
  /* U+3B24 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+3B28 */	"\x00\x00", "\x00\x00", "\x00\x00", "\xFC\xEE",
  /* U+3B2C */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+3B30 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+3B34 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+3B38 */	"\x00\x00", "\x99\x7B", "\x00\x00", "\x00\x00",
  /* U+3B3C */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+3B40 */	"\x00\x00", "\x00\x00", "\x8D\xB4", "\x00\x00",
  /* U+3B44 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+3B48 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+3B4C */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+3B50 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+3B54 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+3B58 */	"\x8D\xB7", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+3B5C */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+3B60 */	"\x91\xB3", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+3B64 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+3B68 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+3B6C */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+3B70 */	"\x00\x00", "\x8D\xBB", "\x8D\xBA", "\x00\x00",
  /* U+3B74 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+3B78 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x8D\xBC",
  /* U+3B7C */	"\x90\x44", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+3B80 */	"\xFD\x4C", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+3B84 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+3B88 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+3B8C */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+3B90 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+3B94 */	"\x00\x00", "\x00\x00", "\x93\xE4", "\x00\x00",
  /* U+3B98 */	"\x00\x00", "\x93\xE0", "\x00\x00", "\x00\x00",
  /* U+3B9C */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+3BA0 */	"\x00\x00", "\xFD\x53", "\x00\x00", "\x00\x00",
  /* U+3BA4 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+3BA8 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+3BAC */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+3BB0 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+3BB4 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+3BB8 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+3BBC */	"\x8D\xC3", "\x00\x00", "\x9B\xB8", "\x00\x00",
  /* U+3BC0 */	"\x00\x00", "\x00\x00", "\xFB\xF0", "\x00\x00",
  /* U+3BC4 */	"\x93\xE9", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+3BC8 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+3BCC */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+3BD0 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+3BD4 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x93\xF6",
  /* U+3BD8 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+3BDC */	"\x00\x00", "\x8D\xC5", "\x00\x00", "\x00\x00",
  /* U+3BE0 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+3BE4 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+3BE8 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+3BEC */	"\x8D\xCA", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+3BF0 */	"\x00\x00", "\x00\x00", "\x8D\xCC", "\xFD\x5D",
  /* U+3BF4 */	"\x93\xB5", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+3BF8 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+3BFC */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+3C00 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+3C04 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+3C08 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+3C0C */	"\x00\x00", "\xFD\x61", "\x00\x00", "\x00\x00",
  /* U+3C10 */	"\x00\x00", "\x9C\xF8", "\x00\x00", "\x00\x00",
  /* U+3C14 */	"\x00\x00", "\x92\x52", "\x00\x00", "\x00\x00",
  /* U+3C18 */	"\xA0\xE8", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+3C1C */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+3C20 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+3C24 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+3C28 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+3C2C */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+3C30 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+3C34 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+3C38 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+3C3C */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+3C40 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+3C44 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+3C48 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+3C4C */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+3C50 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+3C54 */	"\x9C\xA5", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+3C58 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+3C5C */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+3C60 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+3C64 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+3C68 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+3C6C */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+3C70 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+3C74 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+3C78 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+3C7C */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+3C80 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+3C84 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+3C88 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x8C\x56",
  /* U+3C8C */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+3C90 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+3C94 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+3C98 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+3C9C */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+3CA0 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+3CA4 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+3CA8 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+3CAC */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+3CB0 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+3CB4 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+3CB8 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+3CBC */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+3CC0 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+3CC4 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+3CC8 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x8D\xD6",
  /* U+3CCC */	"\x00\x00", "\x97\xC0", "\x00\x00", "\x00\x00",
  /* U+3CD0 */	"\x00\x00", "\xA0\xDE", "\x00\x00", "\x00\x00",
  /* U+3CD4 */	"\x00\x00", "\x00\x00", "\x97\xD2", "\x00\x00",
  /* U+3CD8 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+3CDC */	"\xFA\xA5", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+3CE0 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+3CE4 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+3CE8 */	"\x00\x00", "\x00\x00", "\x00\x00", "\xFD\xA3",
  /* U+3CEC */	"\x00\x00", "\x00\x00", "\x00\x00", "\x8D\xDB",
  /* U+3CF0 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+3CF4 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+3CF8 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+3CFC */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+3D00 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+3D04 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+3D08 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+3D0C */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+3D10 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x8E\xAF",
  /* U+3D14 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+3D18 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+3D1C */	"\x00\x00", "\x91\xB5", "\x00\x00", "\x00\x00",
  /* U+3D20 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+3D24 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+3D28 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+3D2C */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+3D30 */	"\x00\x00", "\x00\x00", "\xFD\x49", "\x00\x00",
  /* U+3D34 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+3D38 */	"\x00\x00", "\x00\x00", "\x00\x00", "\xFD\xD1",
  /* U+3D3C */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+3D40 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+3D44 */	"\x00\x00", "\x00\x00", "\x8D\xEB", "\x00\x00",
  /* U+3D48 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+3D4C */	"\x97\xC6", "\x00\x00", "\xFD\xCE", "\x00\x00",
  /* U+3D50 */	"\x00\x00", "\x90\xFC", "\x00\x00", "\x00\x00",
  /* U+3D54 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+3D58 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+3D5C */	"\x00\x00", "\x00\x00", "\x00\x00", "\xFC\x59",
  /* U+3D60 */	"\x00\x00", "\x00\x00", "\x96\xD6", "\x00\x00",
  /* U+3D64 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+3D68 */	"\x00\x00", "\x97\xC5", "\x8D\xEF", "\x00\x00",
  /* U+3D6C */	"\x00\x00", "\x00\x00", "\x00\x00", "\x97\xD7",
  /* U+3D70 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+3D74 */	"\x00\x00", "\x8D\xF0", "\x00\x00", "\x00\x00",
  /* U+3D78 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+3D7C */	"\x00\x00", "\x96\xA6", "\x00\x00", "\x00\x00",
  /* U+3D80 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+3D84 */	"\x00\x00", "\xFB\xBF", "\x00\x00", "\x00\x00",
  /* U+3D88 */	"\x00\x00", "\x00\x00", "\x8D\xF3", "\x00\x00",
  /* U+3D8C */	"\x00\x00", "\x00\x00", "\x00\x00", "\x94\x49",
  /* U+3D90 */	"\x00\x00", "\x8D\xF5", "\x00\x00", "\x00\x00",
  /* U+3D94 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+3D98 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+3D9C */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+3DA0 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+3DA4 */	"\x00\x00", "\x98\x72", "\x00\x00", "\x00\x00",
  /* U+3DA8 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+3DAC */	"\x00\x00", "\x8E\x6B", "\x00\x00", "\x00\x00",
  /* U+3DB0 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+3DB4 */	"\xFA\xFD", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+3DB8 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+3DBC */	"\x00\x00", "\x00\x00", "\x00\x00", "\x8F\x50",
  /* U+3DC0 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+3DC4 */	"\x00\x00", "\x00\x00", "\x9D\xCC", "\xFC\x65",
  /* U+3DC8 */	"\x00\x00", "\x8C\x44", "\x00\x00", "\x00\x00",
  /* U+3DCC */	"\x99\x6E", "\x94\xA1", "\x00\x00", "\x00\x00",
  /* U+3DD0 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x8F\x63",
  /* U+3DD4 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+3DD8 */	"\x00\x00", "\x00\x00", "\x00\x00", "\xA0\xDA",
  /* U+3DDC */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+3DE0 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+3DE4 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x92\x53",
  /* U+3DE8 */	"\xFD\xE9", "\x00\x00", "\x00\x00", "\x9D\xB5",
  /* U+3DEC */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+3DF0 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x98\x79",
  /* U+3DF4 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x9D\x5D",
  /* U+3DF8 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+3DFC */	"\x8D\x63", "\x96\x69", "\x00\x00", "\x00\x00",
  /* U+3E00 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+3E04 */	"\x00\x00", "\x00\x00", "\x9F\x70", "\x00\x00",
  /* U+3E08 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+3E0C */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+3E10 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+3E14 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+3E18 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+3E1C */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+3E20 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+3E24 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+3E28 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+3E2C */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+3E30 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+3E34 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+3E38 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+3E3C */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+3E40 */	"\xFC\x6A", "\x00\x00", "\x00\x00", "\x8A\xC7",
  /* U+3E44 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+3E48 */	"\x89\xD7", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+3E4C */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+3E50 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+3E54 */	"\x00\x00", "\xFE\x4D", "\x00\x00", "\x00\x00",
  /* U+3E58 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+3E5C */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+3E60 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+3E64 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+3E68 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+3E6C */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+3E70 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+3E74 */	"\x9E\xDD", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+3E78 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+3E7C */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+3E80 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+3E84 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+3E88 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+3E8C */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+3E90 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+3E94 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+3E98 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+3E9C */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+3EA0 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+3EA4 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+3EA8 */	"\xFE\xFB", "\x98\xBC", "\xFA\xCC", "\x00\x00",
  /* U+3EAC */	"\x00\x00", "\x95\xB0", "\x00\x00", "\x00\x00",
  /* U+3EB0 */	"\x00\x00", "\x94\x64", "\x00\x00", "\x00\x00",
  /* U+3EB4 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+3EB8 */	"\x93\x6F", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+3EBC */	"\x00\x00", "\x00\x00", "\x00\x00", "\x94\xB9",
  /* U+3EC0 */	"\x00\x00", "\x00\x00", "\x95\xEC", "\x00\x00",
  /* U+3EC4 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x91\xEE",
  /* U+3EC8 */	"\x00\x00", "\x00\x00", "\x98\xC3", "\x00\x00",
  /* U+3ECC */	"\x95\xF6", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+3ED0 */	"\x8F\xFD", "\x98\xC5", "\x00\x00", "\x00\x00",
  /* U+3ED4 */	"\x00\x00", "\x00\x00", "\x97\x66", "\xFE\x6E",
  /* U+3ED8 */	"\x00\x00", "\x00\x00", "\x97\xDD", "\x8C\xAA",
  /* U+3EDC */	"\x00\x00", "\x00\x00", "\x92\xD2", "\x00\x00",
  /* U+3EE0 */	"\x00\x00", "\x97\x61", "\x98\xCB", "\x00\x00",
  /* U+3EE4 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x95\xF0",
  /* U+3EE8 */	"\x00\x00", "\x97\x5D", "\x00\x00", "\x91\xE3",
  /* U+3EEC */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+3EF0 */	"\x98\xCC", "\x00\x00", "\x00\x00", "\x94\x69",
  /* U+3EF4 */	"\x98\xCD", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+3EF8 */	"\x00\x00", "\x00\x00", "\x98\xCE", "\x00\x00",
  /* U+3EFC */	"\x95\xFC", "\x00\x00", "\x00\x00", "\x94\xA3",
  /* U+3F00 */	"\x96\x62", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+3F04 */	"\xFE\xB6", "\x00\x00", "\x94\x63", "\x00\x00",
  /* U+3F08 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+3F0C */	"\x00\x00", "\x00\x00", "\x98\xD0", "\x00\x00",
  /* U+3F10 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+3F14 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+3F18 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+3F1C */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+3F20 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+3F24 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+3F28 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+3F2C */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+3F30 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+3F34 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+3F38 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+3F3C */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+3F40 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+3F44 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+3F48 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+3F4C */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+3F50 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x98\xD1",
  /* U+3F54 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+3F58 */	"\x94\x75", "\xFA\xE0", "\x00\x00", "\x00\x00",
  /* U+3F5C */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+3F60 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x94\x72",
  /* U+3F64 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+3F68 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+3F6C */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+3F70 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+3F74 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+3F78 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+3F7C */	"\x98\xD6", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+3F80 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+3F84 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+3F88 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+3F8C */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+3F90 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x8A\xF0",
  /* U+3F94 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+3F98 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+3F9C */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+3FA0 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+3FA4 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+3FA8 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+3FAC */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+3FB0 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+3FB4 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+3FB8 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+3FBC */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+3FC0 */	"\x98\xD9", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+3FC4 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+3FC8 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+3FCC */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+3FD0 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+3FD4 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x98\xDB",
  /* U+3FD8 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+3FDC */	"\x98\xDD", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+3FE0 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+3FE4 */	"\x00\x00", "\x98\xA8", "\x00\x00", "\x00\x00",
  /* U+3FE8 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+3FEC */	"\x00\x00", "\x8A\x6D", "\x00\x00", "\x00\x00",
  /* U+3FF0 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+3FF4 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+3FF8 */	"\x00\x00", "\x8A\xFB", "\x8A\xAE", "\x00\x00",
  /* U+3FFC */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+4000 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+4004 */	"\xFB\xC9", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+4008 */	"\x00\x00", "\x8C\x5D", "\x00\x00", "\x00\x00",
  /* U+400C */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+4010 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+4014 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+4018 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+401C */	"\x00\x00", "\x98\xE4", "\x00\x00", "\x00\x00",
  /* U+4020 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+4024 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+4028 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+402C */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+4030 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+4034 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+4038 */	"\x00\x00", "\x98\xE6", "\x00\x00", "\x00\x00",
  /* U+403C */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+4040 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+4044 */	"\x00\x00", "\x98\xE8", "\x00\x00", "\x00\x00",
  /* U+4048 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+404C */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+4050 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x8A\x4D",
  /* U+4054 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x92\x57",
  /* U+4058 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+405C */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+4060 */	"\x00\x00", "\x00\x00", "\x95\xDF", "\x00\x00",
  /* U+4064 */	"\x00\x00", "\xA0\xAC", "\x00\x00", "\x00\x00",
  /* U+4068 */	"\x00\x00", "\x00\x00", "\x98\xEB", "\x00\x00",
  /* U+406C */	"\x00\x00", "\x00\x00", "\x00\x00", "\x98\xEC",
  /* U+4070 */	"\x00\x00", "\x8C\xC3", "\x00\x00", "\x00\x00",
  /* U+4074 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+4078 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+407C */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+4080 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+4084 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+4088 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+408C */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+4090 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+4094 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+4098 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+409C */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+40A0 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+40A4 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+40A8 */	"\x98\xF4", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+40AC */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+40B0 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+40B4 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+40B8 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x8A\xB8",
  /* U+40BC */	"\x00\x00", "\x00\x00", "\x00\x00", "\x9E\xE7",
  /* U+40C0 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+40C4 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+40C8 */	"\x94\xBC", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+40CC */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+40D0 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+40D4 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+40D8 */	"\xFC\xD1", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+40DC */	"\x00\x00", "\x00\x00", "\x00\x00", "\x9C\xC6",
  /* U+40E0 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+40E4 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+40E8 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+40EC */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+40F0 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+40F4 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+40F8 */	"\x00\x00", "\x00\x00", "\x9E\x7E", "\x00\x00",
  /* U+40FC */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+4100 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x98\xFE",
  /* U+4104 */	"\xFD\xE8", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+4108 */	"\x00\x00", "\x99\x40", "\x00\x00", "\x00\x00",
  /* U+410C */	"\x00\x00", "\x00\x00", "\x94\xC9", "\x00\x00",
  /* U+4110 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+4114 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+4118 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+411C */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+4120 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+4124 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+4128 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+412C */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+4130 */	"\x00\x00", "\x00\x00", "\x94\xD3", "\x00\x00",
  /* U+4134 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+4138 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+413C */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+4140 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+4144 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+4148 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+414C */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+4150 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+4154 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+4158 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+415C */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+4160 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+4164 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x99\x46",
  /* U+4168 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+416C */	"\x90\xC0", "\x00\x00", "\x94\xD1", "\x00\x00",
  /* U+4170 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+4174 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+4178 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+417C */	"\x00\x00", "\x00\x00", "\x00\x00", "\x95\x73",
  /* U+4180 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+4184 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+4188 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+418C */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+4190 */	"\x93\xC2", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+4194 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+4198 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+419C */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+41A0 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+41A4 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+41A8 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+41AC */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+41B0 */	"\x00\x00", "\x00\x00", "\x99\x48", "\x00\x00",
  /* U+41B4 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+41B8 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+41BC */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+41C0 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+41C4 */	"\x99\x4B", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+41C8 */	"\x00\x00", "\x00\x00", "\x8E\x55", "\x00\x00",
  /* U+41CC */	"\x00\x00", "\x00\x00", "\x00\x00", "\x99\x4E",
  /* U+41D0 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+41D4 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+41D8 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x8E\xFE",
  /* U+41DC */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+41E0 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+41E4 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+41E8 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+41EC */	"\x00\x00", "\x00\x00", "\x00\x00", "\x8E\x59",
  /* U+41F0 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+41F4 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+41F8 */	"\x00\x00", "\x94\xEC", "\x00\x00", "\x00\x00",
  /* U+41FC */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+4200 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+4204 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+4208 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+420C */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+4210 */	"\x00\x00", "\x94\xEF", "\x00\x00", "\x00\x00",
  /* U+4214 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+4218 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+421C */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+4220 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x8C\x60",
  /* U+4224 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+4228 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+422C */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+4230 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+4234 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+4238 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+423C */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+4240 */	"\x8F\x74", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+4244 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+4248 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+424C */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+4250 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+4254 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+4258 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+425C */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+4260 */	"\x99\x55", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+4264 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+4268 */	"\x00\x00", "\x00\x00", "\x95\x44", "\x00\x00",
  /* U+426C */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+4270 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+4274 */	"\x00\x00", "\x00\x00", "\x8C\xCB", "\x00\x00",
  /* U+4278 */	"\x00\x00", "\x00\x00", "\x99\x56", "\x00\x00",
  /* U+427C */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+4280 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+4284 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+4288 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+428C */	"\x99\x59", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+4290 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+4294 */	"\x99\x5B", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+4298 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+429C */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+42A0 */	"\x00\x00", "\x00\x00", "\x8C\xC4", "\x00\x00",
  /* U+42A4 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+42A8 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+42AC */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+42B0 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+42B4 */	"\x00\x00", "\xFA\x45", "\x00\x00", "\x00\x00",
  /* U+42B8 */	"\x00\x00", "\x90\xB7", "\x00\x00", "\x00\x00",
  /* U+42BC */	"\x97\x43", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+42C0 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+42C4 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+42C8 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+42CC */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+42D0 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+42D4 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+42D8 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+42DC */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+42E0 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+42E4 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+42E8 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+42EC */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+42F0 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+42F4 */	"\x95\xCD", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+42F8 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x97\xC9",
  /* U+42FC */	"\xFD\x50", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+4300 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+4304 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+4308 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+430C */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+4310 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+4314 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+4318 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+431C */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+4320 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+4324 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+4328 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x8E\xB9",
  /* U+432C */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+4330 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+4334 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+4338 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+433C */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+4340 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+4344 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+4348 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+434C */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+4350 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+4354 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+4358 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+435C */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+4360 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+4364 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+4368 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+436C */	"\x00\x00", "\x00\x00", "\x95\xC6", "\x00\x00",
  /* U+4370 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+4374 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+4378 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+437C */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+4380 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+4384 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+4388 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+438C */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+4390 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+4394 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x99\x67",
  /* U+4398 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+439C */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+43A0 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+43A4 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+43A8 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+43AC */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+43B0 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+43B4 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+43B8 */	"\x00\x00", "\x00\x00", "\x8A\xB9", "\x00\x00",
  /* U+43BC */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+43C0 */	"\x00\x00", "\x8D\xFC", "\x00\x00", "\x00\x00",
  /* U+43C4 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+43C8 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+43CC */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+43D0 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+43D4 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+43D8 */	"\x00\x00", "\x8A\x76", "\x00\x00", "\x00\x00",
  /* U+43DC */	"\x00\x00", "\x00\x00", "\x00\x00", "\x9D\x51",
  /* U+43E0 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+43E4 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+43E8 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+43EC */	"\x00\x00", "\x99\x73", "\x00\x00", "\x00\x00",
  /* U+43F0 */	"\x00\x00", "\x00\x00", "\x9D\x4F", "\x00\x00",
  /* U+43F4 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+43F8 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+43FC */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+4400 */	"\x00\x00", "\x99\x7A", "\x95\x64", "\x00\x00",
  /* U+4404 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+4408 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+440C */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+4410 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x99\xA1",
  /* U+4414 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+4418 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+441C */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+4420 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+4424 */	"\x00\x00", "\x99\xA5", "\x00\x00", "\x00\x00",
  /* U+4428 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+442C */	"\x00\x00", "\x99\xA7", "\x00\x00", "\x00\x00",
  /* U+4430 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+4434 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+4438 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+443C */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+4440 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+4444 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+4448 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+444C */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+4450 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+4454 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+4458 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+445C */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+4460 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+4464 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+4468 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+446C */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+4470 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+4474 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+4478 */	"\x00\x00", "\x00\x00", "\x8E\xED", "\x00\x00",
  /* U+447C */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+4480 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+4484 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+4488 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+448C */	"\x00\x00", "\x00\x00", "\x00\x00", "\x99\xAD",
  /* U+4490 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+4494 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+4498 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+449C */	"\x00\x00", "\x00\x00", "\x00\x00", "\x94\x6E",
  /* U+44A0 */	"\x8F\x70", "\x00\x00", "\xFA\xD0", "\x00\x00",
  /* U+44A4 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+44A8 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+44AC */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+44B0 */	"\x99\xB3", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+44B4 */	"\x00\x00", "\x00\x00", "\x00\x00", "\xA0\x53",
  /* U+44B8 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+44BC */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+44C0 */	"\x96\x5C", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+44C4 */	"\x00\x00", "\xFD\x7A", "\x00\x00", "\x00\x00",
  /* U+44C8 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+44CC */	"\x00\x00", "\x00\x00", "\x97\xFE", "\x00\x00",
  /* U+44D0 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+44D4 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+44D8 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+44DC */	"\x00\x00", "\x92\xBD", "\x00\x00", "\x97\xFD",
  /* U+44E0 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+44E4 */	"\x8F\x64", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+44E8 */	"\x00\x00", "\xFC\xF7", "\x95\x62", "\x97\xCD",
  /* U+44EC */	"\x9E\x64", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+44F0 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+44F4 */	"\x92\x4C", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+44F8 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+44FC */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+4500 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x8E\xC9",
  /* U+4504 */	"\x99\xBC", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+4508 */	"\x00\x00", "\x9D\xA5", "\x00\x00", "\x8F\x54",
  /* U+450C */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+4510 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+4514 */	"\x00\x00", "\x00\x00", "\x8F\x7C", "\x00\x00",
  /* U+4518 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+451C */	"\x00\x00", "\x8E\xA2", "\x00\x00", "\x00\x00",
  /* U+4520 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+4524 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x8F\x7A",
  /* U+4528 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+452C */	"\x00\x00", "\x00\x00", "\x97\xAE", "\x00\x00",
  /* U+4530 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x96\xC8",
  /* U+4534 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+4538 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x99\xC3",
  /* U+453C */	"\x00\x00", "\x90\xD6", "\x00\x00", "\x9C\xBE",
  /* U+4540 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x8F\x76",
  /* U+4544 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+4548 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+454C */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+4550 */	"\x00\x00", "\x94\x70", "\xFB\x4B", "\x00\x00",
  /* U+4554 */	"\x00\x00", "\xFD\xCA", "\x00\x00", "\x00\x00",
  /* U+4558 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+455C */	"\x8E\xC7", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+4560 */	"\x00\x00", "\x00\x00", "\xA0\xF9", "\x00\x00",
  /* U+4564 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+4568 */	"\x00\x00", "\x00\x00", "\x8F\xA9", "\x00\x00",
  /* U+456C */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+4570 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+4574 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x99\xC7",
  /* U+4578 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+457C */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+4580 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+4584 */	"\x00\x00", "\x90\xD7", "\x00\x00", "\x00\x00",
  /* U+4588 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+458C */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+4590 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+4594 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+4598 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+459C */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+45A0 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+45A4 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+45A8 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+45AC */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+45B0 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+45B4 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+45B8 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+45BC */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+45C0 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+45C4 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+45C8 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+45CC */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+45D0 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+45D4 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+45D8 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+45DC */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+45E0 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+45E4 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+45E8 */	"\x00\x00", "\x9E\xDF", "\x00\x00", "\x00\x00",
  /* U+45EC */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+45F0 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+45F4 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+45F8 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+45FC */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+4600 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+4604 */	"\x00\x00", "\x00\x00", "\x99\xCE", "\x00\x00",
  /* U+4608 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+460C */	"\x00\x00", "\x00\x00", "\x00\x00", "\x8F\xBA",
  /* U+4610 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+4614 */	"\x00\x00", "\x8F\xEB", "\x00\x00", "\x99\xCF",
  /* U+4618 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+461C */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+4620 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+4624 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+4628 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+462C */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+4630 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+4634 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+4638 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+463C */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+4640 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+4644 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+4648 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+464C */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+4650 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+4654 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+4658 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x8F\xC2",
  /* U+465C */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+4660 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+4664 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+4668 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+466C */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+4670 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+4674 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+4678 */	"\x00\x00", "\x00\x00", "\x92\xC9", "\x00\x00",
  /* U+467C */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+4680 */	"\x97\xDC", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+4684 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+4688 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+468C */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+4690 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+4694 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+4698 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+469C */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+46A0 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+46A4 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+46A8 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+46AC */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+46B0 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+46B4 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+46B8 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+46BC */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+46C0 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+46C4 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+46C8 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+46CC */	"\x00\x00", "\x00\x00", "\x00\x00", "\x95\xB3",
  /* U+46D0 */	"\x9C\x79", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+46D4 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+46D8 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+46DC */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+46E0 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+46E4 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+46E8 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+46EC */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+46F0 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+46F4 */	"\x00\x00", "\x95\xB2", "\x00\x00", "\x00\x00",
  /* U+46F8 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+46FC */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+4700 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+4704 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+4708 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+470C */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+4710 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x8F\xDB",
  /* U+4714 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+4718 */	"\x9B\xE3", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+471C */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+4720 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+4724 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+4728 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+472C */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+4730 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+4734 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+4738 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+473C */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+4740 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+4744 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+4748 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+474C */	"\x00\x00", "\x00\x00", "\x9E\x7A", "\x00\x00",
  /* U+4750 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+4754 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+4758 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+475C */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+4760 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+4764 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+4768 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+476C */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+4770 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+4774 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+4778 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+477C */	"\x9B\xEE", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+4780 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+4784 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+4788 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+478C */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+4790 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+4794 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+4798 */	"\x99\xDE", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+479C */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+47A0 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+47A4 */	"\x00\x00", "\x00\x00", "\xFA\xFA", "\x00\x00",
  /* U+47A8 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+47AC */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+47B0 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+47B4 */	"\x00\x00", "\x00\x00", "\x9E\xE5", "\x00\x00",
  /* U+47B8 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+47BC */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+47C0 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+47C4 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+47C8 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+47CC */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+47D0 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+47D4 */	"\x00\x00", "\x8A\x52", "\x00\x00", "\x00\x00",
  /* U+47D8 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+47DC */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+47E0 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+47E4 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+47E8 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+47EC */	"\x00\x00", "\x99\xE1", "\x00\x00", "\x00\x00",
  /* U+47F0 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+47F4 */	"\x8A\x67", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+47F8 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+47FC */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+4800 */	"\x8B\xB5", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+4804 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+4808 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x8A\xAC",
  /* U+480C */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+4810 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+4814 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+4818 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+481C */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+4820 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+4824 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+4828 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+482C */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+4830 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+4834 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x99\xE9",
  /* U+4838 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+483C */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+4840 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+4844 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+4848 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+484C */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+4850 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+4854 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+4858 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+485C */	"\x00\x00", "\xFB\xCA", "\x00\x00", "\x00\x00",
  /* U+4860 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+4864 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+4868 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+486C */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+4870 */	"\x00\x00", "\x97\xDE", "\x00\x00", "\x00\x00",
  /* U+4874 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+4878 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+487C */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+4880 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+4884 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+4888 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+488C */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+4890 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+4894 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+4898 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x95\xD1",
  /* U+489C */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+48A0 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+48A4 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+48A8 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+48AC */	"\x00\x00", "\x99\xF5", "\xFC\x4A", "\x00\x00",
  /* U+48B0 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+48B4 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+48B8 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+48BC */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+48C0 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+48C4 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+48C8 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+48CC */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+48D0 */	"\x9B\xA9", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+48D4 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+48D8 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+48DC */	"\x00\x00", "\xFB\xDC", "\x00\x00", "\x00\x00",
  /* U+48E0 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+48E4 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+48E8 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+48EC */	"\x00\x00", "\xFE\x56", "\x00\x00", "\x00\x00",
  /* U+48F0 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x9E\xA4",
  /* U+48F4 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+48F8 */	"\x00\x00", "\x00\x00", "\x9D\x49", "\x00\x00",
  /* U+48FC */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+4900 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+4904 */	"\x00\x00", "\x00\x00", "\x95\xDB", "\x00\x00",
  /* U+4908 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+490C */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+4910 */	"\x00\x00", "\x89\xC5", "\x00\x00", "\x00\x00",
  /* U+4914 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+4918 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+491C */	"\x00\x00", "\x00\x00", "\x99\xF8", "\x00\x00",
  /* U+4920 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+4924 */	"\x00\x00", "\x96\x64", "\x00\x00", "\x00\x00",
  /* U+4928 */	"\x00\x00", "\x00\x00", "\x90\x55", "\x00\x00",
  /* U+492C */	"\x00\x00", "\x96\xD4", "\x00\x00", "\x00\x00",
  /* U+4930 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+4934 */	"\x00\x00", "\x97\x7C", "\x00\x00", "\x00\x00",
  /* U+4938 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+493C */	"\x96\x4D", "\x00\x00", "\x97\xE1", "\x00\x00",
  /* U+4940 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+4944 */	"\x00\x00", "\x9A\x48", "\x00\x00", "\x00\x00",
  /* U+4948 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+494C */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+4950 */	"\x00\x00", "\x9A\x49", "\x00\x00", "\xFE\x7D",
  /* U+4954 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+4958 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+495C */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+4960 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+4964 */	"\x00\x00", "\x90\xAA", "\x00\x00", "\x00\x00",
  /* U+4968 */	"\x00\x00", "\x00\x00", "\x9A\x50", "\x00\x00",
  /* U+496C */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+4970 */	"\x00\x00", "\x00\x00", "\x93\x47", "\x00\x00",
  /* U+4974 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+4978 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+497C */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+4980 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+4984 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+4988 */	"\x00\x00", "\x8E\xD8", "\x00\x00", "\x00\x00",
  /* U+498C */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+4990 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+4994 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+4998 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+499C */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+49A0 */	"\x00\x00", "\x90\xC9", "\x00\x00", "\x00\x00",
  /* U+49A4 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x9A\x55",
  /* U+49A8 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+49AC */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+49B0 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+49B4 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+49B8 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+49BC */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+49C0 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+49C4 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+49C8 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+49CC */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+49D0 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+49D4 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+49D8 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+49DC */	"\x00\x00", "\x00\x00", "\x00\x00", "\x90\xBC",
  /* U+49E0 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+49E4 */	"\x00\x00", "\x9A\x58", "\x00\x00", "\x8B\xB8",
  /* U+49E8 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+49EC */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+49F0 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+49F4 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+49F8 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+49FC */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+4A00 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+4A04 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+4A08 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+4A0C */	"\x00\x00", "\x00\x00", "\x00\x00", "\x90\xD5",
  /* U+4A10 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+4A14 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+4A18 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+4A1C */	"\x00\x00", "\x96\x41", "\x00\x00", "\x00\x00",
  /* U+4A20 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+4A24 */	"\x9A\x5A", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+4A28 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+4A2C */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+4A30 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+4A34 */	"\x00\x00", "\x9A\x5C", "\x00\x00", "\x00\x00",
  /* U+4A38 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+4A3C */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+4A40 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+4A44 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+4A48 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+4A4C */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+4A50 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+4A54 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+4A58 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+4A5C */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+4A60 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+4A64 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+4A68 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+4A6C */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+4A70 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+4A74 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+4A78 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+4A7C */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+4A80 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+4A84 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+4A88 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+4A8C */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+4A90 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+4A94 */	"\x00\x00", "\x00\x00", "\x97\xC2", "\x00\x00",
  /* U+4A98 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+4A9C */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+4AA0 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+4AA4 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+4AA8 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+4AAC */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+4AB0 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+4AB4 */	"\x8A\xBB", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+4AB8 */	"\x9B\xAA", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+4ABC */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+4AC0 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+4AC4 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+4AC8 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+4ACC */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+4AD0 */	"\x00\x00", "\x90\xF5", "\x00\x00", "\x00\x00",
  /* U+4AD4 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+4AD8 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+4ADC */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+4AE0 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+4AE4 */	"\x9A\x60", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+4AE8 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+4AEC */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+4AF0 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+4AF4 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+4AF8 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+4AFC */	"\x00\x00", "\x00\x00", "\x00\x00", "\x91\x45",
  /* U+4B00 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+4B04 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+4B08 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+4B0C */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+4B10 */	"\x8C\x58", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+4B14 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+4B18 */	"\x00\x00", "\x9A\x63", "\x00\x00", "\x00\x00",
  /* U+4B1C */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+4B20 */	"\x8C\x49", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+4B24 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+4B28 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+4B2C */	"\x8B\xB6", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+4B30 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+4B34 */	"\x00\x00", "\x00\x00", "\x00\x00", "\xFC\xCF",
  /* U+4B38 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+4B3C */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+4B40 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+4B44 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+4B48 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+4B4C */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+4B50 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+4B54 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+4B58 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+4B5C */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+4B60 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+4B64 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+4B68 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+4B6C */	"\x00\x00", "\x00\x00", "\x00\x00", "\x96\x6B",
  /* U+4B70 */	"\x9A\x6E", "\x00\x00", "\x91\x4F", "\x00\x00",
  /* U+4B74 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+4B78 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x97\x46",
  /* U+4B7C */	"\x00\x00", "\x00\x00", "\xA0\xE6", "\x00\x00",
  /* U+4B80 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+4B84 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+4B88 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+4B8C */	"\x00\x00", "\x00\x00", "\x92\xD7", "\x00\x00",
  /* U+4B90 */	"\x96\x75", "\x00\x00", "\x00\x00", "\x93\xD4",
  /* U+4B94 */	"\x00\x00", "\x00\x00", "\x91\xBB", "\x96\x79",
  /* U+4B98 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+4B9C */	"\x00\x00", "\x9A\x70", "\x00\x00", "\x00\x00",
  /* U+4BA0 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+4BA4 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+4BA8 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+4BAC */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+4BB0 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+4BB4 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+4BB8 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+4BBC */	"\x00\x00", "\x96\x78", "\x91\xCD", "\x00\x00",
  /* U+4BC0 */	"\x9C\x4A", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+4BC4 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+4BC8 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+4BCC */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+4BD0 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+4BD4 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+4BD8 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+4BDC */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+4BE0 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+4BE4 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+4BE8 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+4BEC */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+4BF0 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+4BF4 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+4BF8 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+4BFC */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+4C00 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+4C04 */	"\xA0\x6F", "\x00\x00", "\x00\x00", "\xA0\x6A",
  /* U+4C08 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+4C0C */	"\x00\x00", "\x00\x00", "\x91\x5F", "\x00\x00",
  /* U+4C10 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+4C14 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+4C18 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+4C1C */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+4C20 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+4C24 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+4C28 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+4C2C */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+4C30 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+4C34 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+4C38 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x9F\xA5",
  /* U+4C3C */	"\x00\x00", "\x00\x00", "\x89\xBA", "\x00\x00",
  /* U+4C40 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+4C44 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+4C48 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+4C4C */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+4C50 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+4C54 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+4C58 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x9E\xCD",
  /* U+4C5C */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+4C60 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+4C64 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+4C68 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+4C6C */	"\x00\x00", "\x9A\x79", "\x00\x00", "\x00\x00",
  /* U+4C70 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+4C74 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+4C78 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+4C7C */	"\x00\x00", "\x9D\xCE", "\x00\x00", "\x00\x00",
  /* U+4C80 */	"\x00\x00", "\x8C\xD2", "\x00\x00", "\x00\x00",
  /* U+4C84 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+4C88 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+4C8C */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+4C90 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+4C94 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+4C98 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+4C9C */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+4CA0 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+4CA4 */	"\x9D\x73", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+4CA8 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+4CAC */	"\x00\x00", "\x00\x00", "\x96\xB9", "\x00\x00",
  /* U+4CB0 */	"\x96\xBC", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+4CB4 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x9C\xD1",
  /* U+4CB8 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+4CBC */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+4CC0 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+4CC4 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+4CC8 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+4CCC */	"\x00\x00", "\x89\xB7", "\x00\x00", "\x00\x00",
  /* U+4CD0 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+4CD4 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+4CD8 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+4CDC */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+4CE0 */	"\x00\x00", "\x9E\xEE", "\x00\x00", "\x00\x00",
  /* U+4CE4 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+4CE8 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+4CEC */	"\x00\x00", "\xFB\x43", "\x00\x00", "\x00\x00",
  /* U+4CF0 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+4CF4 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+4CF8 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+4CFC */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+4D00 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+4D04 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+4D08 */	"\x00\x00", "\x9E\xC9", "\x00\x00", "\x00\x00",
  /* U+4D0C */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+4D10 */	"\xFB\xD3", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+4D14 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+4D18 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+4D1C */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+4D20 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+4D24 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+4D28 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+4D2C */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+4D30 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+4D34 */	"\x91\xAE", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+4D38 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+4D3C */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+4D40 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+4D44 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+4D48 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+4D4C */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+4D50 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+4D54 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+4D58 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+4D5C */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+4D60 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+4D64 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+4D68 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+4D6C */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+4D70 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+4D74 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+4D78 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+4D7C */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+4D80 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+4D84 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+4D88 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+4D8C */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+4D90 */	"\x00\x00", "\x9D\x78", "\x00\x00", "\x00\x00",
  /* U+4D94 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+4D98 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+4D9C */	"\x9D\x7B", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+4DA0 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+4DA4 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+4DA8 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+4DAC */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+4DB0 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+4DB4 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+4DB8 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+4DBC */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+4DC0 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+4DC4 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+4DC8 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+4DCC */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+4DD0 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+4DD4 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+4DD8 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+4DDC */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+4DE0 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+4DE4 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+4DE8 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+4DEC */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+4DF0 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+4DF4 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+4DF8 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+4DFC */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+4E00 */	"\xA4\x40", "\xA4\x42", "\x00\x00", "\xA4\x43",
  /* U+4E04 */	"\x9E\xB3", "\x00\x00", "\x00\x00", "\xC9\x45",
  /* U+4E08 */	"\xA4\x56", "\xA4\x54", "\xA4\x57", "\xA4\x55",
  /* U+4E0C */	"\xC9\x46", "\xA4\xA3", "\xC9\x4F", "\xC9\x4D",
  /* U+4E10 */	"\xA4\xA2", "\xA4\xA1", "\x00\x00", "\x00\x00",
  /* U+4E14 */	"\xA5\x42", "\xA5\x41", "\xA5\x40", "\x00\x00",
  /* U+4E18 */	"\xA5\x43", "\xA4\xFE", "\x9E\xB2", "\x00\x00",
  /* U+4E1C */	"\x9D\xD6", "\x00\x00", "\xA5\xE0", "\xA5\xE1",
  /* U+4E20 */	"\x00\x00", "\x99\x4F", "\x00\x00", "\x00\x00",
  /* U+4E24 */	"\x89\xCE", "\x00\x00", "\xA8\xC3", "\x00\x00",
  /* U+4E28 */	"\x8B\xC0", "\x00\x00", "\x9F\xC4", "\xA4\x58",
  /* U+4E2C */	"\x8B\xD4", "\xA4\xA4", "\xC9\x50", "\x8C\x72",
  /* U+4E30 */	"\xA4\xA5", "\xC9\x63", "\xA6\xEA", "\xCB\xB1",
  /* U+4E34 */	"\x00\x00", "\x00\x00", "\xC6\xBF", "\x8B\xF9",
  /* U+4E38 */	"\xA4\x59", "\xA4\xA6", "\x00\x00", "\xA5\x44",
  /* U+4E3C */	"\xC9\x64", "\x89\x46", "\x00\x00", "\xC6\xC0",
  /* U+4E40 */	"\x00\x00", "\x00\x00", "\xC9\x40", "\xA4\x44",
  /* U+4E44 */	"\x00\x00", "\xA4\x5B", "\x00\x00", "\xC9\x47",
  /* U+4E48 */	"\xA4\x5C", "\xFA\xE5", "\x00\x00", "\xA4\xA7",
  /* U+4E4C */	"\x00\x00", "\xA5\x45", "\xA5\x47", "\xA5\x46",
  /* U+4E50 */	"\x00\x00", "\x00\x00", "\xA5\xE2", "\xA5\xE3",
  /* U+4E54 */	"\x00\x00", "\x00\x00", "\xA8\xC4", "\x00\x00",
  /* U+4E58 */	"\xAD\xBC", "\xA4\x41", "\xC8\x7B", "\x8B\xC6",
  /* U+4E5C */	"\xC9\x41", "\xA4\x45", "\xA4\x5E", "\xA4\x5D",
  /* U+4E60 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+4E64 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+4E68 */	"\x00\x00", "\xA5\xE4", "\x9C\x57", "\x00\x00",
  /* U+4E6C */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+4E70 */	"\x00\x00", "\x00\x00", "\x00\x00", "\xA8\xC5",
  /* U+4E74 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+4E78 */	"\x9A\xFB", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+4E7C */	"\x00\x00", "\x00\x00", "\xB0\xAE", "\xD4\x4B",
  /* U+4E80 */	"\x89\xD0", "\x89\xCF", "\xB6\xC3", "\xDC\xB1",
  /* U+4E84 */	"\xDC\xB2", "\xC6\xC1", "\xA4\x46", "\x89\xD1",
  /* U+4E88 */	"\xA4\xA9", "\x89\xE2", "\x00\x00", "\xA8\xC6",
  /* U+4E8C */	"\xA4\x47", "\xC9\x48", "\xA4\x5F", "\x00\x00",
  /* U+4E90 */	"\x00\x00", "\xA4\xAA", "\xA4\xAC", "\xC9\x51",
  /* U+4E94 */	"\xA4\xAD", "\xA4\xAB", "\x00\x00", "\x00\x00",
  /* U+4E98 */	"\x92\x7E", "\xA5\xE5", "\x9D\xBA", "\xA8\xC7",
  /* U+4E9C */	"\x00\x00", "\x00\x00", "\xA8\xC8", "\xAB\x45",
  /* U+4EA0 */	"\xC6\xC2", "\xA4\x60", "\xA4\xAE", "\x8C\x6F",
  /* U+4EA4 */	"\xA5\xE6", "\xA5\xE8", "\xA5\xE7", "\x00\x00",
  /* U+4EA8 */	"\xA6\xEB", "\x00\x00", "\x00\x00", "\xA8\xC9",
  /* U+4EAC */	"\xA8\xCA", "\xAB\x46", "\xAB\x47", "\x00\x00",
  /* U+4EB0 */	"\x00\x00", "\x00\x00", "\x00\x00", "\xAD\xBD",
  /* U+4EB4 */	"\x00\x00", "\x00\x00", "\xDC\xB3", "\xFB\xF8",
  /* U+4EB8 */	"\x00\x00", "\xF6\xD6", "\xA4\x48", "\x8B\xC7",
  /* U+4EBC */	"\x92\x6B", "\x00\x00", "\x00\x00", "\x89\xD2",
  /* U+4EC0 */	"\xA4\xB0", "\xA4\xAF", "\xC9\x52", "\xA4\xB1",
  /* U+4EC4 */	"\xA4\xB7", "\x00\x00", "\xA4\xB2", "\xA4\xB3",
  /* U+4EC8 */	"\xC9\x54", "\xC9\x53", "\xA4\xB5", "\xA4\xB6",
  /* U+4ECC */	"\x00\x00", "\xA4\xB4", "\x9F\xCF", "\x00\x00",
  /* U+4ED0 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+4ED4 */	"\xA5\x4A", "\xA5\x4B", "\xA5\x4C", "\xA5\x4D",
  /* U+4ED8 */	"\xA5\x49", "\xA5\x50", "\xC9\x6A", "\x00\x00",
  /* U+4EDC */	"\xC9\x66", "\xC9\x69", "\xA5\x51", "\xA5\x61",
  /* U+4EE0 */	"\x00\x00", "\xC9\x68", "\x00\x00", "\xA5\x4E",
  /* U+4EE4 */	"\xA5\x4F", "\xA5\x48", "\x00\x00", "\x00\x00",
  /* U+4EE8 */	"\xC9\x65", "\xC9\x67", "\x9D\xA9", "\x89\xD3",
  /* U+4EEC */	"\x00\x00", "\x00\x00", "\x99\xE2", "\x00\x00",
  /* U+4EF0 */	"\xA5\xF5", "\xC9\xB0", "\xA5\xF2", "\xA5\xF6",
  /* U+4EF4 */	"\xC9\xBA", "\xC9\xAE", "\xA5\xF3", "\xC9\xB2",
  /* U+4EF8 */	"\x92\x67", "\x00\x00", "\x00\x00", "\xA5\xF4",
  /* U+4EFC */	"\x00\x00", "\xA5\xF7", "\x00\x00", "\xA5\xE9",
  /* U+4F00 */	"\xC9\xB1", "\xA5\xF8", "\xC9\xB5", "\x92\xA4",
  /* U+4F04 */	"\xC9\xB9", "\xC9\xB6", "\x00\x00", "\x00\x00",
  /* U+4F08 */	"\xC9\xB3", "\xA5\xEA", "\xA5\xEC", "\xA5\xF9",
  /* U+4F0C */	"\x00\x00", "\xA5\xEE", "\xC9\xAB", "\xA5\xF1",
  /* U+4F10 */	"\xA5\xEF", "\xA5\xF0", "\xC9\xBB", "\xC9\xB8",
  /* U+4F14 */	"\xC9\xAF", "\xA5\xED", "\x00\x00", "\x8C\x73",
  /* U+4F18 */	"\xC9\xAC", "\xA5\xEB", "\x89\x4E", "\x00\x00",
  /* U+4F1C */	"\x00\x00", "\xC9\xB4", "\x00\x00", "\x00\x00",
  /* U+4F20 */	"\x00\x00", "\x00\x00", "\xC9\xB7", "\x00\x00",
  /* U+4F24 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+4F28 */	"\x89\x4F", "\x92\x78", "\x00\x00", "\x00\x00",
  /* U+4F2C */	"\xC9\xAD", "\xCA\x66", "\x00\x00", "\xA7\x42",
  /* U+4F30 */	"\xA6\xF4", "\x00\x00", "\x91\xB6", "\xCA\x67",
  /* U+4F34 */	"\xA6\xF1", "\x00\x00", "\xA7\x44", "\x89\xD4",
  /* U+4F38 */	"\xA6\xF9", "\x9F\xD2", "\xA6\xF8", "\xCA\x5B",
  /* U+4F3C */	"\xA6\xFC", "\xA6\xF7", "\xCA\x60", "\xCA\x68",
  /* U+4F40 */	"\x00\x00", "\xCA\x64", "\x92\xA7", "\xA6\xFA",
  /* U+4F44 */	"\x00\x00", "\x95\xA2", "\xA6\xFD", "\xA6\xEE",
  /* U+4F48 */	"\xA7\x47", "\xCA\x5D", "\x00\x00", "\x92\x6E",
  /* U+4F4C */	"\xCB\xBD", "\xA6\xEC", "\xA7\x43", "\xA6\xED",
  /* U+4F50 */	"\xA6\xF5", "\xA6\xF6", "\xCA\x62", "\xCA\x5E",
  /* U+4F54 */	"\xA6\xFB", "\xA6\xF3", "\xCA\x5A", "\xA6\xEF",
  /* U+4F58 */	"\xCA\x65", "\xA7\x45", "\xA7\x48", "\xA6\xF2",
  /* U+4F5C */	"\xA7\x40", "\xA7\x46", "\xA6\xF0", "\xCA\x63",
  /* U+4F60 */	"\xA7\x41", "\xCA\x69", "\xCA\x5C", "\xA6\xFE",
  /* U+4F64 */	"\xCA\x5F", "\x00\x00", "\x00\x00", "\xCA\x61",
  /* U+4F68 */	"\x00\x00", "\xA8\xD8", "\xCB\xBF", "\xCB\xCB",
  /* U+4F6C */	"\xA8\xD0", "\x00\x00", "\xCB\xCC", "\xA8\xCB",
  /* U+4F70 */	"\xA8\xD5", "\x00\x00", "\x96\xEA", "\xA8\xCE",
  /* U+4F74 */	"\xCB\xB9", "\xA8\xD6", "\xCB\xB8", "\xCB\xBC",
  /* U+4F78 */	"\xCB\xC3", "\xCB\xC1", "\xA8\xDE", "\xA8\xD9",
  /* U+4F7C */	"\xCB\xB3", "\xCB\xB5", "\xA8\xDB", "\xA8\xCF",
  /* U+4F80 */	"\xCB\xB6", "\xCB\xC2", "\xCB\xC9", "\xA8\xD4",
  /* U+4F84 */	"\xCB\xBB", "\xCB\xB4", "\xA8\xD3", "\xCB\xB7",
  /* U+4F88 */	"\xA8\xD7", "\xCB\xBA", "\x92\x6F", "\xA8\xD2",
  /* U+4F8C */	"\x00\x00", "\xA8\xCD", "\x00\x00", "\xA8\xDC",
  /* U+4F90 */	"\xCB\xC4", "\xA8\xDD", "\xCB\xC8", "\x00\x00",
  /* U+4F94 */	"\xCB\xC6", "\xCB\xCA", "\xA8\xDA", "\xCB\xBE",
  /* U+4F98 */	"\xCB\xB2", "\x00\x00", "\xCB\xC0", "\xA8\xD1",
  /* U+4F9C */	"\xCB\xC5", "\xA8\xCC", "\xCB\xC7", "\x00\x00",
  /* U+4FA0 */	"\x00\x00", "\x00\x00", "\x92\xA3", "\x00\x00",
  /* U+4FA4 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+4FA8 */	"\x89\x50", "\x00\x00", "\x00\x00", "\xFA\x57",
  /* U+4FAC */	"\x00\x00", "\x00\x00", "\xAB\x56", "\xAB\x4A",
  /* U+4FB0 */	"\x98\x66", "\x00\x00", "\xCD\xE0", "\xCD\xE8",
  /* U+4FB4 */	"\x00\x00", "\xAB\x49", "\xAB\x51", "\xAB\x5D",
  /* U+4FB8 */	"\x00\x00", "\xCD\xEE", "\xCD\xEC", "\xCD\xE7",
  /* U+4FBC */	"\x00\x00", "\x89\xD6", "\x00\x00", "\xAB\x4B",
  /* U+4FC0 */	"\xCD\xED", "\xCD\xE3", "\xAB\x59", "\xAB\x50",
  /* U+4FC4 */	"\xAB\x58", "\xCD\xDE", "\x00\x00", "\xCD\xEA",
  /* U+4FC8 */	"\x98\xB2", "\xCD\xE1", "\xAB\x54", "\xCD\xE2",
  /* U+4FCC */	"\x92\xAB", "\xCD\xDD", "\xAB\x5B", "\xAB\x4E",
  /* U+4FD0 */	"\xAB\x57", "\xAB\x4D", "\x00\x00", "\xCD\xDF",
  /* U+4FD4 */	"\xCD\xE4", "\x00\x00", "\xCD\xEB", "\xAB\x55",
  /* U+4FD8 */	"\xAB\x52", "\xCD\xE6", "\xAB\x5A", "\xCD\xE9",
  /* U+4FDC */	"\xCD\xE5", "\xAB\x4F", "\xAB\x5C", "\xAB\x53",
  /* U+4FE0 */	"\xAB\x4C", "\xAB\x48", "\x00\x00", "\x00\x00",
  /* U+4FE4 */	"\x96\xDE", "\x92\xAC", "\x00\x00", "\x00\x00",
  /* U+4FE8 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+4FEC */	"\xCD\xEF", "\x00\x00", "\xAD\xD7", "\xAD\xC1",
  /* U+4FF0 */	"\x8C\x70", "\xAD\xD1", "\x9F\x6E", "\xAD\xD6",
  /* U+4FF4 */	"\xD0\xD0", "\xD0\xCF", "\xD0\xD4", "\xD0\xD5",
  /* U+4FF8 */	"\xAD\xC4", "\x8E\xF2", "\xAD\xCD", "\x00\x00",
  /* U+4FFC */	"\x00\x00", "\x9F\x6C", "\xAD\xDA", "\x00\x00",
  /* U+5000 */	"\xAD\xCE", "\x00\x00", "\x00\x00", "\x89\xD8",
  /* U+5004 */	"\x00\x00", "\xD0\xC9", "\xAD\xC7", "\xD0\xCA",
  /* U+5008 */	"\xFA\x59", "\xAD\xDC", "\x00\x00", "\xAD\xD3",
  /* U+500C */	"\xAD\xBE", "\xAD\xBF", "\xD0\xDD", "\xB0\xBF",
  /* U+5010 */	"\x00\x00", "\xAD\xCC", "\xAD\xCB", "\xD0\xCB",
  /* U+5014 */	"\xAD\xCF", "\xD4\x5B", "\xAD\xC6", "\xD0\xD6",
  /* U+5018 */	"\xAD\xD5", "\xAD\xD4", "\xAD\xCA", "\xD0\xCE",
  /* U+501C */	"\xD0\xD7", "\x00\x00", "\xD0\xC8", "\xAD\xC9",
  /* U+5020 */	"\xD0\xD8", "\xAD\xD2", "\xD0\xCC", "\xAD\xC0",
  /* U+5024 */	"\x00\x00", "\xAD\xC3", "\xAD\xC2", "\xD0\xD9",
  /* U+5028 */	"\xAD\xD0", "\xAD\xC5", "\xAD\xD9", "\xAD\xDB",
  /* U+502C */	"\xD0\xD3", "\xAD\xD8", "\x92\xA8", "\xD0\xDB",
  /* U+5030 */	"\xD0\xCD", "\xD0\xDC", "\x00\x00", "\xD0\xD1",
  /* U+5034 */	"\x91\x63", "\xD0\xDA", "\x00\x00", "\xD0\xD2",
  /* U+5038 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x8C\x40",
  /* U+503C */	"\xAD\xC8", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+5040 */	"\xD4\x63", "\xD4\x57", "\x00\x00", "\xB0\xB3",
  /* U+5044 */	"\x00\x00", "\xD4\x5C", "\xD4\x62", "\xB0\xB2",
  /* U+5048 */	"\xD4\x55", "\xB0\xB6", "\xD4\x59", "\xD4\x52",
  /* U+504C */	"\xB0\xB4", "\xD4\x56", "\xB0\xB9", "\xB0\xBE",
  /* U+5050 */	"\x00\x00", "\xD4\x67", "\x00\x00", "\xD4\x51",
  /* U+5054 */	"\x00\x00", "\xB0\xBA", "\x9F\x73", "\xD4\x66",
  /* U+5058 */	"\x92\xAD", "\x00\x00", "\xB0\xB5", "\xD4\x58",
  /* U+505C */	"\xB0\xB1", "\xD4\x53", "\xD4\x4F", "\xD4\x5D",
  /* U+5060 */	"\xD4\x50", "\xD4\x4E", "\xD4\x5A", "\xD4\x60",
  /* U+5064 */	"\xD4\x61", "\xB0\xB7", "\x9B\xE9", "\x00\x00",
  /* U+5068 */	"\xD8\x5B", "\xD4\x5E", "\xD4\x4D", "\xD4\x5F",
  /* U+506C */	"\x92\xA9", "\xB0\xC1", "\xD4\x64", "\xB0\xC0",
  /* U+5070 */	"\xD4\x4C", "\x00\x00", "\xD4\x54", "\xD4\x65",
  /* U+5074 */	"\xB0\xBC", "\xB0\xBB", "\xB0\xB8", "\xB0\xBD",
  /* U+5078 */	"\x00\x00", "\x00\x00", "\xB0\xAF", "\x00\x00",
  /* U+507C */	"\x00\x00", "\xB0\xB0", "\x00\x00", "\x00\x00",
  /* U+5080 */	"\xB3\xC8", "\x92\xAA", "\xD8\x5E", "\xD8\x57",
  /* U+5084 */	"\x00\x00", "\xB3\xC5", "\x00\x00", "\xD8\x5F",
  /* U+5088 */	"\x89\xD9", "\x00\x00", "\x00\x00", "\xD8\x55",
  /* U+508C */	"\xD8\x58", "\xB3\xC4", "\xD8\x59", "\x00\x00",
  /* U+5090 */	"\xFD\x56", "\xB3\xC7", "\xD8\x5D", "\x00\x00",
  /* U+5094 */	"\xD8\x53", "\xD8\x52", "\xB3\xC9", "\x00\x00",
  /* U+5098 */	"\xB3\xCA", "\xB3\xC6", "\xB3\xCB", "\xD8\x51",
  /* U+509C */	"\xD8\x5C", "\xD8\x5A", "\xD8\x54", "\x00\x00",
  /* U+50A0 */	"\x00\x00", "\x00\x00", "\xB3\xC3", "\xD8\x56",
  /* U+50A4 */	"\x00\x00", "\x00\x00", "\x9F\xA8", "\x00\x00",
  /* U+50A8 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+50AC */	"\xB6\xCA", "\xB6\xC4", "\xDC\xB7", "\xB6\xCD",
  /* U+50B0 */	"\xDC\xBD", "\xDC\xC0", "\xB6\xC6", "\xB6\xC7",
  /* U+50B4 */	"\xDC\xBA", "\xB6\xC5", "\xDC\xC3", "\xB6\xCB",
  /* U+50B8 */	"\xDC\xC4", "\x00\x00", "\xDC\xBF", "\xB6\xCC",
  /* U+50BC */	"\x8C\x71", "\xDC\xB4", "\xB6\xC9", "\xDC\xB5",
  /* U+50C0 */	"\x00\x00", "\xDC\xBE", "\xDC\xBC", "\x00\x00",
  /* U+50C4 */	"\xDC\xB8", "\xB6\xC8", "\xDC\xB6", "\xB6\xCE",
  /* U+50C8 */	"\xDC\xBB", "\xDC\xC2", "\xDC\xB9", "\xDC\xC1",
  /* U+50CC */	"\x00\x00", "\x92\xA1", "\xB9\xB6", "\xB9\xB3",
  /* U+50D0 */	"\x90\xE3", "\xB9\xB4", "\x00\x00", "\xE0\xF9",
  /* U+50D4 */	"\xE0\xF1", "\xB9\xB2", "\xB9\xAF", "\xE0\xF2",
  /* U+50D8 */	"\x00\x00", "\xA0\xA6", "\xB9\xB1", "\xE0\xF5",
  /* U+50DC */	"\x00\x00", "\xE0\xF7", "\x00\x00", "\x94\xAB",
  /* U+50E0 */	"\xE0\xFE", "\xFC\x72", "\x00\x00", "\xE0\xFD",
  /* U+50E4 */	"\xE0\xF8", "\xB9\xAE", "\xE0\xF0", "\xB9\xAC",
  /* U+50E8 */	"\xE0\xF3", "\xB9\xB7", "\xE0\xF6", "\x00\x00",
  /* U+50EC */	"\xE0\xFA", "\xB9\xB0", "\xB9\xAD", "\xE0\xFC",
  /* U+50F0 */	"\xE0\xFB", "\xB9\xB5", "\x00\x00", "\xE0\xF4",
  /* U+50F4 */	"\x97\xC4", "\xBB\xF8", "\xE4\xEC", "\x00\x00",
  /* U+50F8 */	"\xE4\xE9", "\xBB\xF9", "\x00\x00", "\xBB\xF7",
  /* U+50FC */	"\x92\xAE", "\xE4\xF0", "\xE4\xED", "\xE4\xE6",
  /* U+5100 */	"\xBB\xF6", "\xFA\x67", "\xBB\xFA", "\xE4\xE7",
  /* U+5104 */	"\xBB\xF5", "\xBB\xFD", "\xE4\xEA", "\xE4\xEB",
  /* U+5108 */	"\xBB\xFB", "\xBB\xFC", "\xE4\xF1", "\xE4\xEE",
  /* U+510C */	"\xE4\xEF", "\x92\xA2", "\xFA\x69", "\x00\x00",
  /* U+5110 */	"\xBE\xAA", "\xE8\xF8", "\xBE\xA7", "\xE8\xF5",
  /* U+5114 */	"\xBE\xA9", "\xBE\xAB", "\x00\x00", "\xE8\xF6",
  /* U+5118 */	"\xBE\xA8", "\x00\x00", "\xE8\xF7", "\x00\x00",
  /* U+511C */	"\xE8\xF4", "\x00\x00", "\x00\x00", "\xC0\x76",
  /* U+5120 */	"\xEC\xBD", "\xC0\x77", "\xEC\xBB", "\x00\x00",
  /* U+5124 */	"\xEC\xBC", "\xEC\xBA", "\xEC\xB9", "\x00\x00",
  /* U+5128 */	"\x00\x00", "\xEC\xBE", "\xC0\x75", "\x92\x68",
  /* U+512C */	"\x00\x00", "\xEF\xB8", "\xEF\xB9", "\x00\x00",
  /* U+5130 */	"\xE4\xE8", "\xEF\xB7", "\xC0\x78", "\xC3\x5F",
  /* U+5134 */	"\xF1\xEB", "\xF1\xEC", "\x00\x00", "\xC4\xD7",
  /* U+5138 */	"\xC4\xD8", "\xF5\xC1", "\xF5\xC0", "\xC5\x6C",
  /* U+513C */	"\xC5\x6B", "\xF7\xD0", "\x00\x00", "\xA4\x49",
  /* U+5140 */	"\xA4\x61", "\xA4\xB9", "\x00\x00", "\xA4\xB8",
  /* U+5144 */	"\xA5\x53", "\xA5\x52", "\xA5\xFC", "\xA5\xFB",
  /* U+5148 */	"\xA5\xFD", "\xA5\xFA", "\x00\x00", "\xA7\x4A",
  /* U+514C */	"\xA7\x49", "\xA7\x4B", "\x00\x00", "\x00\x00",
  /* U+5150 */	"\x00\x00", "\x00\x00", "\xA8\xE0", "\x00\x00",
  /* U+5154 */	"\xA8\xDF", "\xA8\xE1", "\x89\x51", "\xAB\x5E",
  /* U+5158 */	"\x00\x00", "\xA2\x59", "\xD0\xDE", "\xA2\x5A",
  /* U+515C */	"\xB0\xC2", "\xA2\x5C", "\xA2\x5B", "\xD8\x60",
  /* U+5160 */	"\xFA\x6F", "\xA2\x5D", "\xB9\xB8", "\xA2\x5E",
  /* U+5164 */	"\x00\x00", "\xA4\x4A", "\x00\x00", "\xA4\xBA",
  /* U+5168 */	"\xA5\xFE", "\xA8\xE2", "\xFA\x71", "\xA4\x4B",
  /* U+516C */	"\xA4\xBD", "\xA4\xBB", "\xA4\xBC", "\x00\x00",
  /* U+5170 */	"\x00\x00", "\xA6\x40", "\x00\x00", "\x00\x00",
  /* U+5174 */	"\x89\x52", "\xA7\x4C", "\xA8\xE4", "\xA8\xE3",
  /* U+5178 */	"\xA8\xE5", "\x94\x5A", "\x00\x00", "\x00\x00",
  /* U+517C */	"\xAD\xDD", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+5180 */	"\xBE\xAC", "\x00\x00", "\xC6\xC3", "\x00\x00",
  /* U+5184 */	"\x00\x00", "\x00\x00", "\x89\xDD", "\xC9\x4E",
  /* U+5188 */	"\xC8\xA2", "\xA5\x54", "\xA5\x55", "\x00\x00",
  /* U+518C */	"\x00\x00", "\xA6\x41", "\x00\x00", "\xCA\x6A",
  /* U+5190 */	"\x00\x00", "\xAB\x60", "\xAB\x5F", "\xD0\xE0",
  /* U+5194 */	"\xD0\xDF", "\xB0\xC3", "\xC6\xC4", "\xA4\xBE",
  /* U+5198 */	"\xC9\x55", "\x00\x00", "\x9E\x52", "\x00\x00",
  /* U+519C */	"\x89\x53", "\x00\x00", "\xCB\xCD", "\x00\x00",
  /* U+51A0 */	"\xAB\x61", "\x00\x00", "\xAD\xE0", "\x00\x00",
  /* U+51A4 */	"\xAD\xDE", "\xAD\xDF", "\x00\x00", "\x9E\x55",
  /* U+51A8 */	"\x92\xBA", "\x00\x00", "\xBE\xAD", "\xC6\xC5",
  /* U+51AC */	"\xA5\x56", "\x00\x00", "\x8C\x5B", "\x00\x00",
  /* U+51B0 */	"\xA6\x42", "\xC9\xBC", "\xFA\x7D", "\xFA\xA8",
  /* U+51B4 */	"\x9A\x68", "\xFA\x47", "\xA7\x4D", "\xA7\x4E",
  /* U+51B8 */	"\xFA\x7E", "\xCA\x6B", "\x00\x00", "\x00\x00",
  /* U+51BC */	"\xCB\xCE", "\xA8\xE6", "\xCB\xCF", "\x00\x00",
  /* U+51C0 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x92\xBB",
  /* U+51C4 */	"\xD0\xE2", "\xD0\xE3", "\xAD\xE3", "\xFD\xB6",
  /* U+51C8 */	"\xD0\xE4", "\xFA\xA2", "\xD0\xE1", "\xAD\xE4",
  /* U+51CC */	"\xAD\xE2", "\xAD\xE1", "\xD0\xE5", "\xFA\xA3",
  /* U+51D0 */	"\xD4\x68", "\xFA\xA4", "\x9B\xB4", "\xFA\xA6",
  /* U+51D4 */	"\xD8\x61", "\x00\x00", "\x00\x00", "\xDC\xC5",
  /* U+51D8 */	"\xE1\x40", "\x00\x00", "\x00\x00", "\x89\xDF",
  /* U+51DC */	"\xBB\xFE", "\xBE\xAE", "\xE8\xF9", "\xFD\xDB",
  /* U+51E0 */	"\xA4\x4C", "\xA4\x5A", "\xFA\xA9", "\x00\x00",
  /* U+51E4 */	"\x89\x54", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+51E8 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+51EC */	"\x00\x00", "\xFA\xAB", "\x00\x00", "\x00\x00",
  /* U+51F0 */	"\xB0\xC4", "\xB3\xCD", "\x00\x00", "\xB9\xB9",
  /* U+51F4 */	"\xFC\x7A", "\xC9\x42", "\xA4\xBF", "\x00\x00",
  /* U+51F8 */	"\xA5\x59", "\xA5\x57", "\xA5\x58", "\x00\x00",
  /* U+51FC */	"\x89\xE0", "\xA8\xE7", "\x9F\x4F", "\x00\x00",
  /* U+5200 */	"\xA4\x4D", "\xA4\x4E", "\xC8\x7D", "\xA4\x62",
  /* U+5204 */	"\x00\x00", "\x89\xE1", "\xA4\xC0", "\xA4\xC1",
  /* U+5208 */	"\xA4\xC2", "\xC9\xBE", "\xA5\x5A", "\xFA\xB0",
  /* U+520C */	"\xC9\x6B", "\x00\x00", "\xA6\x46", "\x00\x00",
  /* U+5210 */	"\xC9\xBF", "\xA6\x44", "\xA6\x45", "\xC9\xBD",
  /* U+5214 */	"\x00\x00", "\x00\x00", "\xA6\x47", "\xA6\x43",
  /* U+5218 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+521C */	"\xCA\x6C", "\xAA\xEC", "\xCA\x6D", "\x9F\xCD",
  /* U+5220 */	"\xA0\xE7", "\xCA\x6E", "\x00\x00", "\x00\x00",
  /* U+5224 */	"\xA7\x50", "\xA7\x4F", "\xFA\xB1", "\x89\xA6",
  /* U+5228 */	"\xA7\x53", "\xA7\x51", "\xA7\x52", "\x00\x00",
  /* U+522C */	"\x00\x00", "\x00\x00", "\xA8\xED", "\x00\x00",
  /* U+5230 */	"\xA8\xEC", "\xCB\xD4", "\xCB\xD1", "\xCB\xD2",
  /* U+5234 */	"\x9E\xFA", "\xCB\xD0", "\xA8\xEE", "\xA8\xEA",
  /* U+5238 */	"\xA8\xE9", "\x00\x00", "\xA8\xEB", "\xA8\xE8",
  /* U+523C */	"\xFA\xB2", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+5240 */	"\x00\x00", "\xA8\xEF", "\x00\x00", "\xAB\x63",
  /* U+5244 */	"\xCD\xF0", "\x00\x00", "\xCB\xD3", "\xAB\x68",
  /* U+5248 */	"\x00\x00", "\xCD\xF1", "\xAB\x64", "\xAB\x67",
  /* U+524C */	"\xAB\x66", "\xAB\x65", "\xAB\x62", "\x00\x00",
  /* U+5250 */	"\x00\x00", "\x00\x00", "\xD0\xE8", "\x00\x00",
  /* U+5254 */	"\xAD\xE7", "\xD0\xEB", "\xAD\xE5", "\xFA\xB4",
  /* U+5258 */	"\x00\x00", "\x92\xC4", "\xD0\xE7", "\xAD\xE8",
  /* U+525C */	"\xAD\xE6", "\xAD\xE9", "\xD0\xE9", "\xD0\xEA",
  /* U+5260 */	"\x9F\x6F", "\xD0\xE6", "\xD0\xEC", "\x00\x00",
  /* U+5264 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+5268 */	"\x8B\xB0", "\xB3\xD1", "\xB0\xC5", "\xD4\x69",
  /* U+526C */	"\xD4\x6B", "\xD4\x6A", "\xD4\x6C", "\xB0\xC6",
  /* U+5270 */	"\x00\x00", "\x00\x00", "\xB3\xCE", "\x9F\xAC",
  /* U+5274 */	"\xB3\xCF", "\xB3\xD0", "\x00\x00", "\xB6\xD0",
  /* U+5278 */	"\xDC\xC7", "\x89\xE3", "\xDC\xC6", "\xDC\xC8",
  /* U+527C */	"\xDC\xC9", "\xB6\xD1", "\x00\x00", "\xB6\xCF",
  /* U+5280 */	"\xE1\x41", "\xE1\x42", "\xB9\xBB", "\xB9\xBA",
  /* U+5284 */	"\xE3\x5A", "\x00\x00", "\x00\x00", "\xBC\x40",
  /* U+5288 */	"\xBC\x41", "\xBC\x42", "\xBC\x44", "\xE4\xF2",
  /* U+528C */	"\xE4\xF3", "\xBC\x43", "\x00\x00", "\x9B\xD3",
  /* U+5290 */	"\x89\xE4", "\xBE\xAF", "\x00\x00", "\xBE\xB0",
  /* U+5294 */	"\xFA\xB5", "\x00\x00", "\xF1\xED", "\xF5\xC3",
  /* U+5298 */	"\xF5\xC2", "\xF7\xD1", "\x9F\xD5", "\xA4\x4F",
  /* U+529C */	"\x00\x00", "\x00\x00", "\x00\x00", "\xA5\x5C",
  /* U+52A0 */	"\xA5\x5B", "\x89\x55", "\x00\x00", "\xA6\x48",
  /* U+52A4 */	"\x92\xC5", "\x00\x00", "\xC9\xC0", "\x00\x00",
  /* U+52A8 */	"\x89\x56", "\xA7\x55", "\xA7\x56", "\xA7\x54",
  /* U+52AC */	"\xA7\x57", "\xCA\x6F", "\xCA\x70", "\x00\x00",
  /* U+52B0 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+52B4 */	"\x00\x00", "\xFA\xB3", "\x00\x00", "\x00\x00",
  /* U+52B8 */	"\x00\x00", "\xFA\xB6", "\x00\x00", "\xA8\xF1",
  /* U+52BC */	"\xCB\xD5", "\x00\x00", "\xA8\xF0", "\x00\x00",
  /* U+52C0 */	"\xCD\xF2", "\xAB\x6C", "\xCD\xF3", "\xAB\x6B",
  /* U+52C4 */	"\x00\x00", "\xFA\xB7", "\x00\x00", "\xAB\x69",
  /* U+52C8 */	"\x00\x00", "\xAB\x6A", "\x00\x00", "\x00\x00",
  /* U+52CC */	"\x9E\xDC", "\xD0\xED", "\x00\x00", "\x00\x00",
  /* U+52D0 */	"\xFB\xC4", "\x9F\x71", "\xB0\xC7", "\xD4\x6E",
  /* U+52D4 */	"\x00\x00", "\xB0\xCA", "\xD4\x6D", "\xB1\xE5",
  /* U+52D8 */	"\xB0\xC9", "\xB0\xC8", "\x00\x00", "\xB3\xD4",
  /* U+52DC */	"\x00\x00", "\xB3\xD3", "\xB3\xD2", "\xB6\xD2",
  /* U+52E0 */	"\xFA\xBA", "\x92\xC7", "\xB6\xD5", "\xB6\xD6",
  /* U+52E4 */	"\xB6\xD4", "\x00\x00", "\xB6\xD3", "\x00\x00",
  /* U+52E8 */	"\x00\x00", "\xE1\x43", "\x00\x00", "\xE1\x44",
  /* U+52EC */	"\x00\x00", "\x00\x00", "\x00\x00", "\xE4\xF5",
  /* U+52F0 */	"\xBC\x45", "\xE4\xF4", "\x00\x00", "\xBE\xB1",
  /* U+52F4 */	"\xEC\xBF", "\xC0\x79", "\x00\x00", "\xF1\xEE",
  /* U+52F8 */	"\xC4\x55", "\xC6\xC6", "\xA4\x63", "\xA4\xC3",
  /* U+52FC */	"\xC9\x56", "\x00\x00", "\xA4\xC4", "\xA4\xC5",
  /* U+5300 */	"\x00\x00", "\x9A\x4C", "\x00\x00", "\x00\x00",
  /* U+5304 */	"\x00\x00", "\xA5\x5D", "\xA5\x5E", "\x00\x00",
  /* U+5308 */	"\xA6\x49", "\xCA\x71", "\xCB\xD6", "\xCB\xD7",
  /* U+530C */	"\x00\x00", "\xAB\x6D", "\xD0\xEE", "\xB0\xCC",
  /* U+5310 */	"\xB0\xCB", "\xD8\x63", "\xD8\x62", "\x00\x00",
  /* U+5314 */	"\x00\x00", "\xA4\x50", "\xA4\xC6", "\xA5\x5F",
  /* U+5318 */	"\x00\x00", "\xB0\xCD", "\xC9\x43", "\x00\x00",
  /* U+531C */	"\xC9\x6C", "\xA5\x60", "\x00\x00", "\xC9\xC2",
  /* U+5320 */	"\xA6\x4B", "\xA6\x4A", "\xC9\xC1", "\xA7\x58",
  /* U+5324 */	"\x8C\x68", "\x00\x00", "\x00\x00", "\x89\xE5",
  /* U+5328 */	"\x00\x00", "\x00\x00", "\xAD\xEA", "\x00\x00",
  /* U+532C */	"\x9F\x7D", "\xD4\x6F", "\x00\x00", "\xB6\xD7",
  /* U+5330 */	"\xE1\x45", "\xB9\xBC", "\xA0\xA9", "\xFA\xC4",
  /* U+5334 */	"\xE8\xFA", "\x00\x00", "\x00\x00", "\xF3\xFD",
  /* U+5338 */	"\xC6\xC7", "\xA4\xC7", "\x00\x00", "\x89\x57",
  /* U+533C */	"\xCB\xD8", "\xCD\xF4", "\xB0\xD0", "\xB0\xCE",
  /* U+5340 */	"\xB0\xCF", "\xA4\x51", "\xFA\xAA", "\xA4\x64",
  /* U+5344 */	"\xA2\xCD", "\xA4\xCA", "\x00\x00", "\xA4\xC9",
  /* U+5348 */	"\xA4\xC8", "\xA5\x63", "\xA5\x62", "\x00\x00",
  /* U+534C */	"\xC9\x6D", "\xC9\xC3", "\x89\x58", "\x00\x00",
  /* U+5350 */	"\x00\x00", "\xA8\xF5", "\xA8\xF2", "\xA8\xF4",
  /* U+5354 */	"\xA8\xF3", "\x00\x00", "\x00\x00", "\xAB\x6E",
  /* U+5358 */	"\x00\x00", "\x00\x00", "\xB3\xD5", "\x00\x00",
  /* U+535C */	"\xA4\x52", "\x8B\xE3", "\xA4\xCB", "\x8B\x61",
  /* U+5360 */	"\xA5\x65", "\xA5\x64", "\x00\x00", "\xCA\x72",
  /* U+5364 */	"\x9A\xF1", "\x00\x00", "\xA8\xF6", "\x9E\xB7",
  /* U+5368 */	"\x00\x00", "\xC6\xC8", "\x00\x00", "\x00\x00",
  /* U+536C */	"\xC9\x57", "\xFA\xD1", "\xA5\x67", "\xA5\x66",
  /* U+5370 */	"\xA6\x4C", "\xA6\x4D", "\xCA\x73", "\xA7\x59",
  /* U+5374 */	"\xFA\xD2", "\xA7\x5A", "\x00\x00", "\xA8\xF7",
  /* U+5378 */	"\xA8\xF8", "\xA8\xF9", "\x00\x00", "\xAB\x6F",
  /* U+537C */	"\xCD\xF5", "\x9E\xBA", "\xFA\xD4", "\xAD\xEB",
  /* U+5380 */	"\x00\x00", "\x00\x00", "\xC9\x44", "\x00\x00",
  /* U+5384 */	"\xA4\xCC", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+5388 */	"\x00\x00", "\x00\x00", "\xC9\xC4", "\x00\x00",
  /* U+538C */	"\x00\x00", "\x00\x00", "\xCA\x74", "\xCA\x75",
  /* U+5390 */	"\x00\x00", "\x00\x00", "\xCB\xD9", "\xFA\xD9",
  /* U+5394 */	"\xCB\xDA", "\x00\x00", "\xCD\xF7", "\xCD\xF6",
  /* U+5398 */	"\xCD\xF9", "\xCD\xF8", "\xAB\x70", "\x00\x00",
  /* U+539C */	"\xD4\x70", "\xAD\xED", "\xD0\xEF", "\xAD\xEC",
  /* U+53A0 */	"\xFA\xDB", "\x00\x00", "\x9C\xE0", "\x00\x00",
  /* U+53A4 */	"\xD8\x64", "\xB3\xD6", "\xFB\xF7", "\xD8\x65",
  /* U+53A8 */	"\xFB\xFA", "\x89\xE7", "\xA0\x7A", "\xFA\xDC",
  /* U+53AC */	"\xE1\x46", "\xB9\xBD", "\xFA\xDD", "\x00\x00",
  /* U+53B0 */	"\x89\xE9", "\x00\x00", "\xBC\x46", "\x00\x00",
  /* U+53B4 */	"\xF1\xEF", "\x00\x00", "\xC6\xC9", "\x00\x00",
  /* U+53B8 */	"\x00\x00", "\xC9\x58", "\x00\x00", "\xA5\x68",
  /* U+53BC */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+53C0 */	"\x00\x00", "\xFA\xE2", "\x89\xEB", "\xB0\xD1",
  /* U+53C4 */	"\x00\x00", "\xFA\xE3", "\x00\x00", "\x00\x00",
  /* U+53C8 */	"\xA4\x53", "\xA4\x65", "\xA4\xCE", "\xA4\xCD",
  /* U+53CC */	"\x90\xC8", "\xA4\xCF", "\x00\x00", "\x00\x00",
  /* U+53D0 */	"\x92\xDA", "\x89\x59", "\x9C\xF5", "\x00\x00",
  /* U+53D4 */	"\xA8\xFB", "\x00\x00", "\xA8\xFA", "\xA8\xFC",
  /* U+53D8 */	"\x89\x5A", "\xFA\xE7", "\x9F\xA2", "\xAB\x71",
  /* U+53DC */	"\x00\x00", "\x00\x00", "\x00\x00", "\xAD\xEE",
  /* U+53E0 */	"\xFA\xEA", "\xE8\xFB", "\xC2\x4F", "\xA4\x66",
  /* U+53E4 */	"\xA5\x6A", "\xA5\x79", "\xA5\x74", "\x00\x00",
  /* U+53E8 */	"\xA5\x6F", "\xA5\x6E", "\xA5\x75", "\xA5\x73",
  /* U+53EC */	"\xA5\x6C", "\xA5\x7A", "\xA5\x6D", "\xA5\x69",
  /* U+53F0 */	"\xA5\x78", "\xA5\x77", "\xA5\x76", "\xA5\x6B",
  /* U+53F4 */	"\x00\x00", "\xA5\x72", "\xFA\xED", "\x8F\xAD",
  /* U+53F8 */	"\xA5\x71", "\x00\x00", "\x00\x00", "\xA5\x7B",
  /* U+53FC */	"\xA5\x70", "\x00\x00", "\xFB\x59", "\x00\x00",
  /* U+5400 */	"\x00\x00", "\xA6\x53", "\x00\x00", "\xA6\x59",
  /* U+5404 */	"\xA6\x55", "\x00\x00", "\xA6\x5B", "\xC9\xC5",
  /* U+5408 */	"\xA6\x58", "\xA6\x4E", "\xA6\x51", "\xA6\x54",
  /* U+540C */	"\xA6\x50", "\xA6\x57", "\xA6\x5A", "\xA6\x4F",
  /* U+5410 */	"\xA6\x52", "\xA6\x56", "\xA6\x5C", "\xFA\xEF",
  /* U+5414 */	"\x96\xEF", "\x00\x00", "\x9D\xEC", "\x00\x00",
  /* U+5418 */	"\xCA\x7E", "\xCA\x7B", "\x9D\xCA", "\xA7\x67",
  /* U+541C */	"\xCA\x7C", "\xA7\x5B", "\xA7\x5D", "\xA7\x75",
  /* U+5420 */	"\xA7\x70", "\xFD\x6D", "\x00\x00", "\x89\xEC",
  /* U+5424 */	"\xCA\xA5", "\xCA\x7D", "\xA7\x5F", "\xA7\x61",
  /* U+5428 */	"\xCA\xA4", "\xA7\x68", "\xCA\x78", "\xA7\x74",
  /* U+542C */	"\xA7\x76", "\xA7\x5C", "\xA7\x6D", "\xFB\x44",
  /* U+5430 */	"\xCA\x76", "\xA7\x73", "\x9D\xE2", "\xA7\x64",
  /* U+5434 */	"\x8C\x75", "\xA7\x6E", "\xA7\x6F", "\xCA\x77",
  /* U+5438 */	"\xA7\x6C", "\xA7\x6A", "\x00\x00", "\xA7\x6B",
  /* U+543C */	"\xA7\x71", "\xCA\xA1", "\xA7\x5E", "\x00\x00",
  /* U+5440 */	"\xA7\x72", "\xCA\xA3", "\xA7\x66", "\xA7\x63",
  /* U+5444 */	"\x00\x00", "\xCA\x7A", "\xA7\x62", "\xCA\xA6",
  /* U+5448 */	"\xA7\x65", "\x00\x00", "\xA7\x69", "\x9E\xC0",
  /* U+544C */	"\x00\x00", "\x9E\x56", "\xA7\x60", "\xCA\xA2",
  /* U+5450 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+5454 */	"\xCA\x79", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+5458 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+545C */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+5460 */	"\xCB\xEB", "\xCB\xEA", "\xA9\x4F", "\xCB\xED",
  /* U+5464 */	"\xCB\xEF", "\xCB\xE4", "\xCB\xE7", "\xCB\xEE",
  /* U+5468 */	"\xA9\x50", "\x9F\x79", "\x9A\xC7", "\xCB\xE1",
  /* U+546C */	"\xCB\xE5", "\xFA\xF4", "\x00\x00", "\xCB\xE9",
  /* U+5470 */	"\xCE\x49", "\xA9\x4B", "\xCE\x4D", "\xA8\xFD",
  /* U+5474 */	"\xCB\xE6", "\xA8\xFE", "\xA9\x4C", "\xA9\x45",
  /* U+5478 */	"\xA9\x41", "\x00\x00", "\xCB\xE2", "\xA9\x44",
  /* U+547C */	"\xA9\x49", "\xA9\x52", "\xCB\xE3", "\xCB\xDC",
  /* U+5480 */	"\xA9\x43", "\xCB\xDD", "\xCB\xDF", "\x00\x00",
  /* U+5484 */	"\xA9\x46", "\x98\xA1", "\xA9\x48", "\xCB\xDB",
  /* U+5488 */	"\xCB\xE0", "\x00\x00", "\x00\x00", "\xA9\x51",
  /* U+548C */	"\xA9\x4D", "\xCB\xE8", "\xA9\x53", "\xFA\xF8",
  /* U+5490 */	"\xA9\x4A", "\xCB\xDE", "\xA9\x47", "\x89\xF0",
  /* U+5494 */	"\x9E\x47", "\xA9\x42", "\xA9\x40", "\x9D\xF7",
  /* U+5498 */	"\xCB\xEC", "\x00\x00", "\xA9\x4E", "\x00\x00",
  /* U+549C */	"\x9F\xD3", "\x00\x00", "\x9A\xCA", "\x00\x00",
  /* U+54A0 */	"\xCE\x48", "\xCD\xFB", "\xCE\x4B", "\x89\xF1",
  /* U+54A4 */	"\xFA\xF9", "\xCD\xFD", "\xAB\x78", "\xAB\xA8",
  /* U+54A8 */	"\xAB\x74", "\xAB\xA7", "\xAB\x7D", "\xAB\xA4",
  /* U+54AC */	"\xAB\x72", "\xCD\xFC", "\xCE\x43", "\xAB\xA3",
  /* U+54B0 */	"\xCE\x4F", "\xAB\xA5", "\x8E\x5A", "\xAB\x79",
  /* U+54B4 */	"\x89\xF2", "\x00\x00", "\xCE\x45", "\xCE\x42",
  /* U+54B8 */	"\xAB\x77", "\x89\xF3", "\xCD\xFA", "\xAB\xA6",
  /* U+54BC */	"\xCE\x4A", "\xAB\x7C", "\xCE\x4C", "\xAB\xA9",
  /* U+54C0 */	"\xAB\x73", "\xAB\x7E", "\xAB\x7B", "\xCE\x40",
  /* U+54C4 */	"\xAB\xA1", "\xCE\x46", "\xCE\x47", "\xAB\x7A",
  /* U+54C8 */	"\xAB\xA2", "\xAB\x76", "\x00\x00", "\x92\x5D",
  /* U+54CC */	"\x8B\x51", "\x92\xE0", "\xAB\x75", "\xCD\xFE",
  /* U+54D0 */	"\x89\xF4", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+54D4 */	"\x00\x00", "\x00\x00", "\xCE\x44", "\x00\x00",
  /* U+54D8 */	"\x00\x00", "\x00\x00", "\x9F\xD4", "\x00\x00",
  /* U+54DC */	"\x00\x00", "\x00\x00", "\xCE\x4E", "\x00\x00",
  /* U+54E0 */	"\xD1\x44", "\xAD\xFB", "\xD0\xF1", "\x8A\x79",
  /* U+54E4 */	"\xD0\xF6", "\xAD\xF4", "\xAE\x40", "\xD0\xF4",
  /* U+54E8 */	"\xAD\xEF", "\xAD\xF9", "\xAD\xFE", "\xD0\xFB",
  /* U+54EC */	"\x00\x00", "\xAD\xFA", "\xAD\xFD", "\x89\xF5",
  /* U+54F0 */	"\x00\x00", "\xD0\xFE", "\xAD\xF5", "\xD0\xF5",
  /* U+54F4 */	"\x00\x00", "\x00\x00", "\x00\x00", "\xD1\x42",
  /* U+54F8 */	"\xD1\x43", "\x00\x00", "\xAD\xF7", "\xD1\x41",
  /* U+54FC */	"\xAD\xF3", "\xAE\x43", "\x00\x00", "\xD0\xF8",
  /* U+5500 */	"\x00\x00", "\xAD\xF1", "\x97\xA7", "\xD1\x46",
  /* U+5504 */	"\xD0\xF9", "\xD0\xFD", "\xAD\xF6", "\xAE\x42",
  /* U+5508 */	"\xD0\xFA", "\xAD\xFC", "\xD1\x40", "\xD1\x47",
  /* U+550C */	"\xD4\xA1", "\x93\xBA", "\xD1\x45", "\xAE\x44",
  /* U+5510 */	"\xAD\xF0", "\xD0\xFC", "\xD0\xF3", "\x9E\x58",
  /* U+5514 */	"\xAD\xF8", "\x00\x00", "\x00\x00", "\xD0\xF2",
  /* U+5518 */	"\x89\xF6", "\x00\x00", "\xD0\xF7", "\x00\x00",
  /* U+551C */	"\x00\x00", "\x00\x00", "\x9E\x57", "\x00\x00",
  /* U+5520 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x89\xF7",
  /* U+5524 */	"\x00\x00", "\x8A\x41", "\xD0\xF0", "\xAE\x41",
  /* U+5528 */	"\x89\xF8", "\x00\x00", "\xD4\x77", "\xFA\xF1",
  /* U+552C */	"\xB0\xE4", "\xD4\xA7", "\xB0\xE2", "\xB0\xDF",
  /* U+5530 */	"\xD4\x7C", "\xB0\xDB", "\xD4\xA2", "\xB0\xE6",
  /* U+5534 */	"\xD4\x76", "\xD4\x7B", "\xD4\x7A", "\xAD\xF2",
  /* U+5538 */	"\xB0\xE1", "\xD4\xA5", "\x00\x00", "\xD4\xA8",
  /* U+553C */	"\xD4\x73", "\x00\x00", "\xB3\xE8", "\x89\xFA",
  /* U+5540 */	"\xD4\xA9", "\xB0\xE7", "\x00\x00", "\xB0\xD9",
  /* U+5544 */	"\xB0\xD6", "\xD4\x7E", "\xB0\xD3", "\xFB\x42",
  /* U+5548 */	"\xD4\xA6", "\xFA\xBF", "\xB0\xDA", "\xD4\xAA",
  /* U+554C */	"\x00\x00", "\xD4\x74", "\xD4\xA4", "\xB0\xDD",
  /* U+5550 */	"\xD4\x75", "\xD4\x78", "\xD4\x7D", "\xFB\xA3",
  /* U+5554 */	"\x00\x00", "\xB0\xDE", "\xB0\xDC", "\xB0\xE8",
  /* U+5558 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+555C */	"\xB0\xE3", "\xFA\xF7", "\xB0\xD7", "\xB1\xD2",
  /* U+5560 */	"\x00\x00", "\xB0\xD8", "\xD4\x79", "\xB0\xE5",
  /* U+5564 */	"\xB0\xE0", "\xD4\xA3", "\xB0\xD5", "\x00\x00",
  /* U+5568 */	"\x00\x00", "\x9E\x4E", "\xB0\xD4", "\x94\xDC",
  /* U+556C */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+5570 */	"\x00\x00", "\x95\xDA", "\x9D\xF8", "\x9F\x6A",
  /* U+5574 */	"\x00\x00", "\xD4\x71", "\xD4\x72", "\xD8\x6A",
  /* U+5578 */	"\x00\x00", "\x8A\xB7", "\x00\x00", "\xB3\xD7",
  /* U+557C */	"\xB3\xDA", "\xD8\x75", "\xB3\xEE", "\xD8\x78",
  /* U+5580 */	"\xB3\xD8", "\xD8\x71", "\xB3\xDE", "\xB3\xE4",
  /* U+5584 */	"\xB5\xBD", "\x00\x00", "\xFB\x46", "\xB3\xE2",
  /* U+5588 */	"\xD8\x6E", "\xB3\xEF", "\xB3\xDB", "\xB3\xE3",
  /* U+558C */	"\xD8\x76", "\xDC\xD7", "\xD8\x7B", "\xD8\x6F",
  /* U+5590 */	"\x8A\x46", "\xD8\x66", "\xD8\x73", "\xD8\x6D",
  /* U+5594 */	"\xB3\xE1", "\xD8\x79", "\x00\x00", "\x00\x00",
  /* U+5598 */	"\xB3\xDD", "\xB3\xF1", "\xB3\xEA", "\x00\x00",
  /* U+559C */	"\xB3\xDF", "\xB3\xDC", "\x00\x00", "\xB3\xE7",
  /* U+55A0 */	"\x00\x00", "\xD8\x7A", "\xD8\x6C", "\xD8\x72",
  /* U+55A4 */	"\xD8\x74", "\xD8\x68", "\xD8\x77", "\xB3\xD9",
  /* U+55A8 */	"\xD8\x67", "\xFB\x47", "\xB3\xE0", "\xB3\xF0",
  /* U+55AC */	"\xB3\xEC", "\xD8\x69", "\xB3\xE6", "\x00\x00",
  /* U+55B0 */	"\x91\x48", "\xB3\xED", "\xB3\xE9", "\xB3\xE5",
  /* U+55B4 */	"\x92\xDE", "\xD8\x70", "\x00\x00", "\x00\x00",
  /* U+55B8 */	"\x00\x00", "\x8B\x53", "\x9D\xF6", "\xB3\xEB",
  /* U+55BC */	"\x9B\xDA", "\x00\x00", "\x00\x00", "\xDC\xD5",
  /* U+55C0 */	"\xDC\xD1", "\x9D\x7E", "\xDC\xE0", "\xDC\xCA",
  /* U+55C4 */	"\xDC\xD3", "\xB6\xE5", "\xB6\xE6", "\xB6\xDE",
  /* U+55C8 */	"\xDC\xDC", "\xB6\xE8", "\xDC\xCF", "\xDC\xCE",
  /* U+55CC */	"\xDC\xCC", "\xDC\xDE", "\xB6\xDC", "\xDC\xD8",
  /* U+55D0 */	"\xDC\xCD", "\xB6\xDF", "\xDC\xD6", "\xB6\xDA",
  /* U+55D4 */	"\xDC\xD2", "\xDC\xD9", "\xDC\xDB", "\x89\xFD",
  /* U+55D8 */	"\x99\xE4", "\xDC\xDF", "\xB6\xE3", "\xDC\xCB",
  /* U+55DC */	"\xB6\xDD", "\xDC\xD0", "\x9E\x43", "\xB6\xD8",
  /* U+55E0 */	"\x00\x00", "\xB6\xE4", "\xDC\xDA", "\xB6\xE0",
  /* U+55E4 */	"\xB6\xE1", "\xB6\xE7", "\xB6\xDB", "\xA2\x5F",
  /* U+55E8 */	"\xB6\xD9", "\xDC\xD4", "\x9D\xE9", "\x00\x00",
  /* U+55EC */	"\x8F\x52", "\x00\x00", "\x00\x00", "\xB6\xE2",
  /* U+55F0 */	"\x9D\xF5", "\x9D\xF0", "\xDC\xDD", "\x00\x00",
  /* U+55F4 */	"\x00\x00", "\x99\xE7", "\xB9\xCD", "\xB9\xC8",
  /* U+55F8 */	"\x00\x00", "\xE1\x55", "\xE1\x51", "\x8B\xBD",
  /* U+55FC */	"\xE1\x4B", "\xB9\xC2", "\xB9\xBE", "\xE1\x54",
  /* U+5600 */	"\xB9\xBF", "\xE1\x4E", "\xE1\x50", "\x00\x00",
  /* U+5604 */	"\xE1\x53", "\x9D\xEF", "\xB9\xC4", "\x00\x00",
  /* U+5608 */	"\xB9\xCB", "\xB9\xC5", "\x00\x00", "\x00\x00",
  /* U+560C */	"\xE1\x49", "\xB9\xC6", "\xB9\xC7", "\xE1\x4C",
  /* U+5610 */	"\xB9\xCC", "\x9F\xB7", "\xE1\x4A", "\xE1\x4F",
  /* U+5614 */	"\xB9\xC3", "\xE1\x48", "\xB9\xC9", "\xB9\xC1",
  /* U+5618 */	"\x00\x00", "\x00\x00", "\x00\x00", "\xB9\xC0",
  /* U+561C */	"\xE1\x4D", "\xE1\x52", "\x9D\xD0", "\xB9\xCA",
  /* U+5620 */	"\x9F\xEB", "\x8D\xA9", "\x9D\xCF", "\x98\xE1",
  /* U+5624 */	"\x00\x00", "\x9D\xE5", "\x00\x00", "\xE1\x47",
  /* U+5628 */	"\x00\x00", "\xBC\x4D", "\xE5\x47", "\x00\x00",
  /* U+562C */	"\xE5\x44", "\x9D\xC8", "\xBC\x47", "\xBC\x53",
  /* U+5630 */	"\xBC\x54", "\x00\x00", "\xBC\x4A", "\xE5\x42",
  /* U+5634 */	"\xBC\x4C", "\xE4\xF9", "\xBC\x52", "\xFB\x4F",
  /* U+5638 */	"\xE5\x46", "\xBC\x49", "\xE5\x48", "\xBC\x48",
  /* U+563C */	"\x00\x00", "\xE5\x43", "\xE5\x45", "\xBC\x4B",
  /* U+5640 */	"\xE5\x41", "\xE4\xFA", "\xE4\xF7", "\x9D\xEB",
  /* U+5644 */	"\x00\x00", "\xD8\x6B", "\xE4\xFD", "\x00\x00",
  /* U+5648 */	"\xE4\xF6", "\xE4\xFC", "\xE4\xFB", "\x00\x00",
  /* U+564C */	"\xE4\xF8", "\xFB\x54", "\xBC\x4F", "\xFB\x55",
  /* U+5650 */	"\x9A\xA2", "\x00\x00", "\x8A\xD6", "\xBC\x4E",
  /* U+5654 */	"\x9A\x5F", "\x00\x00", "\x00\x00", "\xBC\x50",
  /* U+5658 */	"\xE4\xFE", "\xBE\xB2", "\xE5\x40", "\x00\x00",
  /* U+565C */	"\x00\x00", "\x9E\xF5", "\xE9\x45", "\x00\x00",
  /* U+5660 */	"\xE8\xFD", "\x8F\xB7", "\xBE\xBE", "\xE9\x42",
  /* U+5664 */	"\xBE\xB6", "\xBE\xBA", "\xE9\x41", "\x00\x00",
  /* U+5668 */	"\xBE\xB9", "\xBE\xB5", "\xBE\xB8", "\xBE\xB3",
  /* U+566C */	"\xBE\xBD", "\xE9\x43", "\xE8\xFE", "\xBE\xBC",
  /* U+5670 */	"\xE8\xFC", "\xBE\xBB", "\xE9\x44", "\xE9\x40",
  /* U+5674 */	"\xBC\x51", "\x00\x00", "\xBE\xBF", "\xE9\x46",
  /* U+5678 */	"\xBE\xB7", "\xBE\xB4", "\x9A\xD2", "\x9E\x6A",
  /* U+567C */	"\x9E\xE8", "\x00\x00", "\xEC\xC6", "\xEC\xC8",
  /* U+5680 */	"\xC0\x7B", "\xEC\xC9", "\xEC\xC7", "\xEC\xC5",
  /* U+5684 */	"\xEC\xC4", "\xC0\x7D", "\xEC\xC3", "\xC0\x7E",
  /* U+5688 */	"\x00\x00", "\x8B\xBF", "\x91\xC2", "\x9D\x62",
  /* U+568C */	"\xEC\xC1", "\xEC\xC2", "\xC0\x7A", "\xC0\xA1",
  /* U+5690 */	"\xC0\x7C", "\x00\x00", "\x92\x60", "\xEC\xC0",
  /* U+5694 */	"\x00\x00", "\xC2\x50", "\x00\x00", "\xEF\xBC",
  /* U+5698 */	"\xEF\xBA", "\xEF\xBF", "\xEF\xBD", "\x00\x00",
  /* U+569C */	"\xEF\xBB", "\xEF\xBE", "\x92\x5E", "\x91\xC1",
  /* U+56A0 */	"\x00\x00", "\x8A\xC5", "\x00\x00", "\x00\x00",
  /* U+56A4 */	"\x97\xA3", "\xC3\x60", "\xF1\xF2", "\xF1\xF3",
  /* U+56A8 */	"\xC4\x56", "\x00\x00", "\xF1\xF4", "\xF1\xF0",
  /* U+56AC */	"\xF1\xF5", "\xF1\xF1", "\xC2\x51", "\x8B\x6C",
  /* U+56B0 */	"\x00\x00", "\x8D\x7E", "\xF3\xFE", "\xF4\x41",
  /* U+56B4 */	"\xC4\x59", "\xF4\x40", "\xC4\x58", "\xC4\x57",
  /* U+56B8 */	"\x00\x00", "\x9C\x54", "\x00\x00", "\x00\x00",
  /* U+56BC */	"\xC4\x5A", "\xF5\xC5", "\xF5\xC6", "\x9D\xBD",
  /* U+56C0 */	"\xC4\xDA", "\xC4\xD9", "\xC4\xDB", "\xF5\xC4",
  /* U+56C4 */	"\x00\x00", "\xF6\xD8", "\xF6\xD7", "\x00\x00",
  /* U+56C8 */	"\xC5\x6D", "\xC5\x6F", "\xC5\x6E", "\xF6\xD9",
  /* U+56CC */	"\xC5\xC8", "\xF8\xA6", "\x00\x00", "\x00\x00",
  /* U+56D0 */	"\x00\x00", "\xC5\xF1", "\x00\x00", "\xF8\xA5",
  /* U+56D4 */	"\xF8\xEE", "\x00\x00", "\x9C\xC5", "\xC9\x49",
  /* U+56D8 */	"\x00\x00", "\x00\x00", "\xA5\x7D", "\xA5\x7C",
  /* U+56DC */	"\x00\x00", "\xA6\x5F", "\xA6\x5E", "\xC9\xC7",
  /* U+56E0 */	"\xA6\x5D", "\xC9\xC6", "\x89\x5B", "\x00\x00",
  /* U+56E4 */	"\xA7\x79", "\xCA\xA9", "\x00\x00", "\xCA\xA8",
  /* U+56E8 */	"\x00\x00", "\x00\x00", "\xA7\x77", "\xA7\x7A",
  /* U+56EC */	"\x00\x00", "\xFB\x5C", "\xCA\xA7", "\xFB\x5B",
  /* U+56F0 */	"\xA7\x78", "\xFB\x57", "\x00\x00", "\x00\x00",
  /* U+56F4 */	"\x00\x00", "\x00\x00", "\x00\x00", "\xCB\xF0",
  /* U+56F8 */	"\x00\x00", "\xCB\xF1", "\xA9\x54", "\x00\x00",
  /* U+56FC */	"\x00\x00", "\x98\xC7", "\x00\x00", "\xAB\xAA",
  /* U+5700 */	"\xFB\x5A", "\xD1\x48", "\xD1\x49", "\xAE\x45",
  /* U+5704 */	"\xAE\x46", "\x00\x00", "\x00\x00", "\xD4\xAC",
  /* U+5708 */	"\xB0\xE9", "\xB0\xEB", "\xD4\xAB", "\xB0\xEA",
  /* U+570C */	"\xD8\x7C", "\xB3\xF2", "\x00\x00", "\x00\x00",
  /* U+5710 */	"\x00\x00", "\x00\x00", "\xB6\xE9", "\xB6\xEA",
  /* U+5714 */	"\xDC\xE1", "\x9C\xEE", "\xB9\xCF", "\x00\x00",
  /* U+5718 */	"\xB9\xCE", "\x00\x00", "\xE5\x49", "\xE9\x48",
  /* U+571C */	"\xE9\x47", "\x92\xE2", "\xF9\x6B", "\xA4\x67",
  /* U+5720 */	"\xC9\x59", "\x00\x00", "\xC9\x6E", "\xC9\x6F",
  /* U+5724 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+5728 */	"\xA6\x62", "\xA6\x66", "\xC9\xC9", "\x00\x00",
  /* U+572C */	"\xA6\x64", "\xA6\x63", "\xC9\xC8", "\xA6\x65",
  /* U+5730 */	"\xA6\x61", "\x00\x00", "\x94\xA7", "\xA6\x60",
  /* U+5734 */	"\xC9\xCA", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+5738 */	"\x00\x00", "\x00\x00", "\x00\x00", "\xA7\xA6",
  /* U+573C */	"\x00\x00", "\x8C\xCC", "\xA7\xA3", "\x9B\xD4",
  /* U+5740 */	"\xA7\x7D", "\xCA\xAA", "\xFB\x64", "\xFB\x76",
  /* U+5744 */	"\x00\x00", "\xCA\xAB", "\xFB\x60", "\xA7\xA1",
  /* U+5748 */	"\x00\x00", "\xCA\xAD", "\xA7\x7B", "\xCA\xAE",
  /* U+574C */	"\xCA\xAC", "\xA7\x7E", "\xA7\xA2", "\xA7\xA5",
  /* U+5750 */	"\xA7\xA4", "\xA7\x7C", "\xCA\xAF", "\x00\x00",
  /* U+5754 */	"\x99\xE5", "\x00\x00", "\x00\x00", "\x9A\xC2",
  /* U+5758 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x91\xFB",
  /* U+575C */	"\x00\x00", "\x00\x00", "\x00\x00", "\xA0\x73",
  /* U+5760 */	"\x00\x00", "\xA9\x59", "\xCB\xFE", "\x00\x00",
  /* U+5764 */	"\xA9\x5B", "\x00\x00", "\xA9\x5A", "\x9F\x72",
  /* U+5768 */	"\xCC\x40", "\xA9\x58", "\xA9\x57", "\xCB\xF5",
  /* U+576C */	"\x00\x00", "\xCB\xF4", "\x00\x00", "\xCB\xF2",
  /* U+5770 */	"\xCB\xF7", "\xCB\xF6", "\xCB\xF3", "\xCB\xFC",
  /* U+5774 */	"\xCB\xFD", "\xCB\xFA", "\xCB\xF8", "\xA9\x56",
  /* U+5778 */	"\x00\x00", "\x00\x00", "\x9F\xCC", "\xCB\xFB",
  /* U+577C */	"\xA9\x5C", "\xCC\x41", "\x98\xA5", "\x92\xE8",
  /* U+5780 */	"\xCB\xF9", "\x00\x00", "\xAB\xAB", "\xA9\x55",
  /* U+5784 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+5788 */	"\x9B\xBC", "\x00\x00", "\x96\xF3", "\xAB\xAC",
  /* U+578C */	"\xCE\x54", "\x92\xE7", "\x00\x00", "\xCE\x5A",
  /* U+5790 */	"\xFC\x67", "\x00\x00", "\x00\x00", "\xAB\xB2",
  /* U+5794 */	"\xCE\x58", "\xCE\x5E", "\x00\x00", "\xCE\x55",
  /* U+5798 */	"\xCE\x59", "\xCE\x5B", "\xCE\x5D", "\xCE\x57",
  /* U+579C */	"\x8B\x7D", "\xCE\x56", "\xCE\x51", "\xCE\x52",
  /* U+57A0 */	"\xAB\xAD", "\x9B\xF4", "\xAB\xAF", "\xAB\xAE",
  /* U+57A4 */	"\xCE\x53", "\xCE\x5C", "\x00\x00", "\x9E\xF7",
  /* U+57A8 */	"\x00\x00", "\x00\x00", "\x9E\xC1", "\x00\x00",
  /* U+57AC */	"\x00\x00", "\x00\x00", "\xAB\xB1", "\x00\x00",
  /* U+57B0 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+57B4 */	"\x99\x6F", "\xCE\x50", "\xD1\x53", "\x00\x00",
  /* U+57B8 */	"\xD1\x52", "\xD1\x57", "\xD1\x4E", "\x96\xF1",
  /* U+57BC */	"\xD1\x51", "\xD1\x50", "\x8E\x41", "\xD1\x54",
  /* U+57C0 */	"\x00\x00", "\xD1\x58", "\xAE\x47", "\xAE\x4A",
  /* U+57C4 */	"\x95\x4A", "\x00\x00", "\xD1\x4F", "\xD1\x55",
  /* U+57C8 */	"\x97\xE6", "\x00\x00", "\x00\x00", "\xAE\x49",
  /* U+57CC */	"\xD1\x4A", "\x00\x00", "\xAB\xB0", "\xD4\xBA",
  /* U+57D0 */	"\xD1\x56", "\x00\x00", "\xD1\x4D", "\x00\x00",
  /* U+57D4 */	"\xAE\x48", "\xD1\x4C", "\x00\x00", "\x96\xF5",
  /* U+57D8 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+57DC */	"\xD4\xB1", "\x92\xE6", "\x9F\x42", "\xB0\xEC",
  /* U+57E0 */	"\xB0\xF0", "\xD4\xC1", "\xD4\xAF", "\xD4\xBD",
  /* U+57E4 */	"\xB0\xF1", "\xD4\xBF", "\xFB\x67", "\xD4\xC5",
  /* U+57E8 */	"\x00\x00", "\xD4\xC9", "\x00\x00", "\x00\x00",
  /* U+57EC */	"\xD4\xC0", "\xD4\xB4", "\xD4\xBC", "\x99\xA9",
  /* U+57F0 */	"\xD4\xCA", "\xD4\xC8", "\xD4\xBE", "\xD4\xB9",
  /* U+57F4 */	"\xD4\xB2", "\xD8\xA6", "\xD4\xB0", "\xB0\xF5",
  /* U+57F8 */	"\xD4\xB7", "\xB0\xF6", "\xB0\xF2", "\xD4\xAD",
  /* U+57FC */	"\xD4\xC3", "\xD4\xB5", "\xFA\xE6", "\x00\x00",
  /* U+5800 */	"\xD4\xB3", "\xD4\xC6", "\xB0\xF3", "\xFB\x69",
  /* U+5804 */	"\xD4\xCC", "\xB0\xED", "\xB0\xEF", "\xD4\xBB",
  /* U+5808 */	"\xD4\xB6", "\xAE\x4B", "\xB0\xEE", "\xD4\xB8",
  /* U+580C */	"\xD4\xC7", "\xD4\xCB", "\xD4\xC2", "\x00\x00",
  /* U+5810 */	"\xD4\xC4", "\x00\x00", "\x97\xE5", "\x00\x00",
  /* U+5814 */	"\xD4\xAE", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+5818 */	"\x00\x00", "\xD8\xA1", "\x00\x00", "\xD8\xAA",
  /* U+581C */	"\xD8\xA9", "\xB3\xFA", "\xD8\xA2", "\x00\x00",
  /* U+5820 */	"\xB3\xFB", "\xB3\xF9", "\x96\x7D", "\xD8\xA4",
  /* U+5824 */	"\xB3\xF6", "\xD8\xA8", "\xFB\x6C", "\xD8\xA3",
  /* U+5828 */	"\xD8\xA5", "\xD8\x7D", "\xB3\xF4", "\x00\x00",
  /* U+582C */	"\xD8\xB2", "\xD8\xB1", "\xD8\xAE", "\xB3\xF3",
  /* U+5830 */	"\xB3\xF7", "\xB3\xF8", "\xD1\x4B", "\xD8\xAB",
  /* U+5834 */	"\xB3\xF5", "\xB0\xF4", "\xD8\xAD", "\xD8\x7E",
  /* U+5838 */	"\xD8\xB0", "\xD8\xAF", "\x99\xA2", "\xD8\xB3",
  /* U+583C */	"\x00\x00", "\xDC\xEF", "\x00\x00", "\xD8\xAC",
  /* U+5840 */	"\x9A\xBB", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+5844 */	"\x9A\x65", "\x00\x00", "\x00\x00", "\x94\x4E",
  /* U+5848 */	"\xD8\xA7", "\xDC\xE7", "\xB6\xF4", "\xB6\xF7",
  /* U+584C */	"\xB6\xF2", "\xDC\xE6", "\xDC\xEA", "\xDC\xE5",
  /* U+5850 */	"\x00\x00", "\xB6\xEC", "\xB6\xF6", "\xDC\xE2",
  /* U+5854 */	"\xB6\xF0", "\xDC\xE9", "\x00\x00", "\xB6\xEE",
  /* U+5858 */	"\xB6\xED", "\xDC\xEC", "\xB6\xEF", "\xDC\xEE",
  /* U+585C */	"\xFB\x6E", "\xDC\xEB", "\xB6\xEB", "\x99\xDF",
  /* U+5860 */	"\x00\x00", "\x00\x00", "\xB6\xF5", "\xDC\xF0",
  /* U+5864 */	"\xDC\xE4", "\xDC\xED", "\x00\x00", "\x00\x00",
  /* U+5868 */	"\xDC\xE3", "\x98\xE3", "\x00\x00", "\xB6\xF1",
  /* U+586C */	"\x92\x54", "\xB6\xF3", "\x00\x00", "\xDC\xE8",
  /* U+5870 */	"\x00\x00", "\xDC\xF1", "\x96\x7B", "\x8A\xAF",
  /* U+5874 */	"\xE1\x5D", "\xB9\xD0", "\xE1\x63", "\x00\x00",
  /* U+5878 */	"\x00\x00", "\xB9\xD5", "\xE1\x5F", "\xE1\x66",
  /* U+587C */	"\xE1\x57", "\xB9\xD7", "\xB9\xD1", "\xE1\x5C",
  /* U+5880 */	"\xBC\x55", "\xE1\x5B", "\xE1\x64", "\xB9\xD2",
  /* U+5884 */	"\x00\x00", "\xB9\xD6", "\xE1\x5A", "\xE1\x60",
  /* U+5888 */	"\xE1\x65", "\xE1\x56", "\xB9\xD4", "\xE1\x5E",
  /* U+588C */	"\x00\x00", "\x00\x00", "\xE1\x62", "\xE1\x68",
  /* U+5890 */	"\xE1\x58", "\xE1\x61", "\x8C\x77", "\xB9\xD3",
  /* U+5894 */	"\xE1\x67", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+5898 */	"\xE1\x59", "\x8B\xAF", "\x9E\xBD", "\x00\x00",
  /* U+589C */	"\xBC\x59", "\xE5\x4B", "\xBC\x57", "\xBC\x56",
  /* U+58A0 */	"\xE5\x4D", "\xE5\x52", "\x00\x00", "\xE5\x4E",
  /* U+58A4 */	"\x00\x00", "\xE5\x51", "\xBC\x5C", "\x9E\xE6",
  /* U+58A8 */	"\xBE\xA5", "\xBC\x5B", "\xFB\x6F", "\xE5\x4A",
  /* U+58AC */	"\xE5\x50", "\x00\x00", "\xBC\x5A", "\xE5\x4F",
  /* U+58B0 */	"\x8E\xE1", "\xE5\x4C", "\x00\x00", "\xBC\x58",
  /* U+58B4 */	"\x00\x00", "\x9B\x7D", "\x9C\x7E", "\x00\x00",
  /* U+58B8 */	"\x00\x00", "\x00\x00", "\xE9\x4D", "\xF9\xD9",
  /* U+58BC */	"\xE9\x4F", "\xE9\x4A", "\xBE\xC1", "\xE9\x4C",
  /* U+58C0 */	"\x00\x00", "\xBE\xC0", "\xE9\x4E", "\x00\x00",
  /* U+58C4 */	"\x00\x00", "\xBE\xC3", "\xE9\x50", "\xBE\xC2",
  /* U+58C8 */	"\xE9\x49", "\xE9\x4B", "\x00\x00", "\x92\xEA",
  /* U+58CC */	"\x00\x00", "\x00\x00", "\xC0\xA5", "\xEC\xCC",
  /* U+58D0 */	"\x8C\x78", "\xC0\xA4", "\xEC\xCD", "\xC0\xA3",
  /* U+58D4 */	"\xEC\xCB", "\xC0\xA2", "\xEC\xCA", "\x00\x00",
  /* U+58D8 */	"\xC2\x53", "\xC2\x52", "\xF1\xF6", "\xF1\xF8",
  /* U+58DC */	"\xFB\x72", "\xF1\xF7", "\xC3\x61", "\xC3\x62",
  /* U+58E0 */	"\xFB\x71", "\x00\x00", "\xC3\x63", "\xF4\x42",
  /* U+58E4 */	"\xC4\x5B", "\x00\x00", "\x00\x00", "\xF7\xD3",
  /* U+58E8 */	"\xF7\xD2", "\xC5\xF2", "\x00\x00", "\xA4\x68",
  /* U+58EC */	"\xA4\xD0", "\x00\x00", "\x00\x00", "\xA7\xA7",
  /* U+58F0 */	"\x89\x5C", "\x00\x00", "\x98\xF0", "\x96\xF2",
  /* U+58F4 */	"\xCE\x5F", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+58F8 */	"\x00\x00", "\xB3\xFC", "\xB3\xFD", "\xFB\x74",
  /* U+58FC */	"\xDC\xF2", "\xB9\xD8", "\xE1\x69", "\xE5\x53",
  /* U+5900 */	"\x00\x00", "\x00\x00", "\x8B\xC1", "\xC9\x5A",
  /* U+5904 */	"\x89\x5D", "\x89\xDE", "\xCA\xB0", "\x89\x5E",
  /* U+5908 */	"\x00\x00", "\x00\x00", "\xC6\xCA", "\x00\x00",
  /* U+590C */	"\xCC\x42", "\xCE\x60", "\xD1\x59", "\xAE\x4C",
  /* U+5910 */	"\x00\x00", "\xFE\x42", "\xF1\xF9", "\x00\x00",
  /* U+5914 */	"\xC4\xDC", "\xA4\x69", "\xA5\x7E", "\xC9\x70",
  /* U+5918 */	"\x00\x00", "\xA6\x67", "\xA6\x68", "\x00\x00",
  /* U+591C */	"\xA9\x5D", "\x00\x00", "\x00\x00", "\xFB\x7B",
  /* U+5920 */	"\xB0\xF7", "\x00\x00", "\xB9\xDA", "\x00\x00",
  /* U+5924 */	"\xB9\xDB", "\xB9\xD9", "\x00\x00", "\xA4\x6A",
  /* U+5928 */	"\x00\x00", "\xA4\xD1", "\xA4\xD3", "\xA4\xD2",
  /* U+592C */	"\xC9\x5B", "\xA4\xD4", "\xA5\xA1", "\xC9\x71",
  /* U+5930 */	"\x00\x00", "\xA5\xA2", "\x89\x5F", "\x00\x00",
  /* U+5934 */	"\x89\x60", "\x00\x00", "\x00\x00", "\xA6\x69",
  /* U+5938 */	"\xA6\x6A", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+593C */	"\xC9\xCB", "\x00\x00", "\xA7\xA8", "\x00\x00",
  /* U+5940 */	"\xCA\xB1", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+5944 */	"\xA9\x61", "\xCC\x43", "\x00\x00", "\xA9\x5F",
  /* U+5948 */	"\xA9\x60", "\xA9\x5E", "\xD1\x5A", "\x00\x00",
  /* U+594C */	"\x00\x00", "\x00\x00", "\xAB\xB6", "\xAB\xB5",
  /* U+5950 */	"\xAB\xB7", "\xAB\xB4", "\x00\x00", "\xCE\x61",
  /* U+5954 */	"\xA9\x62", "\xAB\xB3", "\x00\x00", "\xAE\x4D",
  /* U+5958 */	"\xAE\x4E", "\x00\x00", "\xAE\x4F", "\x00\x00",
  /* U+595C */	"\xD4\xCD", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+5960 */	"\xB3\xFE", "\xD8\xB4", "\xB0\xF8", "\x00\x00",
  /* U+5964 */	"\x00\x00", "\x9B\xCD", "\x00\x00", "\xB6\xF8",
  /* U+5968 */	"\x00\x00", "\xB9\xDD", "\xB9\xDC", "\xE1\x6A",
  /* U+596C */	"\x00\x00", "\xBC\x5D", "\xBE\xC4", "\x00\x00",
  /* U+5970 */	"\xEF\xC0", "\xF6\xDA", "\xF7\xD4", "\xA4\x6B",
  /* U+5974 */	"\xA5\xA3", "\x9D\xD3", "\xA5\xA4", "\xC9\xD1",
  /* U+5978 */	"\xA6\x6C", "\xA6\x6F", "\x00\x00", "\xC9\xCF",
  /* U+597C */	"\xC9\xCD", "\xA6\x6E", "\xC9\xD0", "\xC9\xD2",
  /* U+5980 */	"\xC9\xCC", "\xA6\x71", "\xA6\x70", "\xA6\x6D",
  /* U+5984 */	"\xA6\x6B", "\xC9\xCE", "\x00\x00", "\x00\x00",
  /* U+5988 */	"\x00\x00", "\x98\x4C", "\xA7\xB3", "\x00\x00",
  /* U+598C */	"\x00\x00", "\xA7\xB0", "\xCA\xB6", "\xCA\xB9",
  /* U+5990 */	"\xCA\xB8", "\x00\x00", "\xA7\xAA", "\xA7\xB2",
  /* U+5994 */	"\x97\x52", "\x00\x00", "\xA7\xAF", "\xCA\xB5",
  /* U+5998 */	"\xCA\xB3", "\xA7\xAE", "\x95\xC3", "\x00\x00",
  /* U+599C */	"\x00\x00", "\xA7\xA9", "\xA7\xAC", "\x9B\xB6",
  /* U+59A0 */	"\xCA\xB4", "\xCA\xBB", "\xCA\xB7", "\xA7\xAD",
  /* U+59A4 */	"\xA7\xB1", "\xA7\xB4", "\xCA\xB2", "\xCA\xBA",
  /* U+59A8 */	"\xA7\xAB", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+59AC */	"\x9A\xB9", "\x00\x00", "\xA9\x67", "\xA9\x6F",
  /* U+59B0 */	"\x97\xB3", "\xCC\x4F", "\xCC\x48", "\xA9\x70",
  /* U+59B4 */	"\xCC\x53", "\xCC\x44", "\xCC\x4B", "\x9F\x74",
  /* U+59B8 */	"\x92\xF1", "\xA9\x66", "\xCC\x45", "\xA9\x64",
  /* U+59BC */	"\xCC\x4C", "\xCC\x50", "\xA9\x63", "\x00\x00",
  /* U+59C0 */	"\xCC\x51", "\xCC\x4A", "\x00\x00", "\xCC\x4D",
  /* U+59C4 */	"\x97\xDF", "\xA9\x72", "\xA9\x69", "\xCC\x54",
  /* U+59C8 */	"\xCC\x52", "\xFB\xA6", "\xA9\x6E", "\xA9\x6C",
  /* U+59CC */	"\xCC\x49", "\xA9\x6B", "\xCC\x47", "\xCC\x46",
  /* U+59D0 */	"\xA9\x6A", "\xA9\x68", "\xA9\x71", "\xA9\x6D",
  /* U+59D4 */	"\xA9\x65", "\x00\x00", "\xCC\x4E", "\x00\x00",
  /* U+59D8 */	"\xAB\xB9", "\xFB\xAB", "\xAB\xC0", "\xCE\x6F",
  /* U+59DC */	"\xAB\xB8", "\xCE\x67", "\xCE\x63", "\x00\x00",
  /* U+59E0 */	"\xCE\x73", "\xCE\x62", "\x00\x00", "\xAB\xBB",
  /* U+59E4 */	"\xCE\x6C", "\xAB\xBE", "\xAB\xC1", "\x00\x00",
  /* U+59E8 */	"\xAB\xBC", "\xCE\x70", "\xAB\xBF", "\x98\x77",
  /* U+59EC */	"\xAE\x56", "\xCE\x76", "\xCE\x64", "\x98\x54",
  /* U+59F0 */	"\x95\xC5", "\xCE\x66", "\xCE\x6D", "\xCE\x71",
  /* U+59F4 */	"\xCE\x75", "\xCE\x72", "\xCE\x6B", "\xCE\x6E",
  /* U+59F8 */	"\x9D\x55", "\xFB\xB2", "\xCE\x68", "\xAB\xC3",
  /* U+59FC */	"\xCE\x6A", "\xCE\x69", "\xCE\x74", "\xAB\xBA",
  /* U+5A00 */	"\xCE\x65", "\xAB\xC2", "\x95\x7E", "\xAB\xBD",
  /* U+5A04 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+5A08 */	"\x00\x00", "\xAE\x5C", "\xD1\x62", "\x97\x42",
  /* U+5A0C */	"\xAE\x5B", "\x94\xE6", "\x00\x00", "\xD1\x60",
  /* U+5A10 */	"\x00\x00", "\xAE\x50", "\x92\xF5", "\xAE\x55",
  /* U+5A14 */	"\x00\x00", "\xD1\x5F", "\xD1\x5C", "\xD1\x61",
  /* U+5A18 */	"\xAE\x51", "\xD1\x5B", "\x8C\xC5", "\xAE\x54",
  /* U+5A1C */	"\xAE\x52", "\x00\x00", "\xD1\x63", "\xAE\x53",
  /* U+5A20 */	"\xAE\x57", "\x92\xFD", "\x00\x00", "\xAE\x58",
  /* U+5A24 */	"\xFB\xA2", "\xAE\x5A", "\x00\x00", "\x9C\x51",
  /* U+5A28 */	"\x00\x00", "\xAE\x59", "\x94\xE9", "\x98\x5C",
  /* U+5A2C */	"\x92\xF0", "\xD1\x5D", "\xD1\x5E", "\x00\x00",
  /* U+5A30 */	"\x00\x00", "\x00\x00", "\x00\x00", "\xD1\x64",
  /* U+5A34 */	"\x00\x00", "\xD4\xD4", "\xB0\xF9", "\xD8\xC2",
  /* U+5A38 */	"\xD4\xD3", "\xD4\xE6", "\x00\x00", "\x00\x00",
  /* U+5A3C */	"\xB1\x40", "\x94\x4C", "\xD4\xE4", "\x00\x00",
  /* U+5A40 */	"\xB0\xFE", "\xB0\xFA", "\xD4\xED", "\xD4\xDD",
  /* U+5A44 */	"\xD4\xE0", "\x91\x6B", "\xB1\x43", "\xD4\xEA",
  /* U+5A48 */	"\xD4\xE2", "\xB0\xFB", "\xB1\x44", "\x00\x00",
  /* U+5A4C */	"\xD4\xE7", "\xD4\xE5", "\x00\x00", "\x00\x00",
  /* U+5A50 */	"\xD4\xD6", "\xD4\xEB", "\xD4\xDF", "\xD4\xDA",
  /* U+5A54 */	"\x8B\x78", "\xD4\xD0", "\xD4\xEC", "\xD4\xDC",
  /* U+5A58 */	"\xD4\xCF", "\x94\xE2", "\xB1\x42", "\xD4\xE1",
  /* U+5A5C */	"\xD4\xEE", "\xD4\xDE", "\xD4\xD2", "\xD4\xD7",
  /* U+5A60 */	"\xD4\xCE", "\x98\x4F", "\xB1\x41", "\xFB\xB5",
  /* U+5A64 */	"\xD4\xDB", "\xD4\xD8", "\xB0\xFC", "\xD4\xD1",
  /* U+5A68 */	"\x92\x71", "\xD4\xE9", "\xB0\xFD", "\x93\x65",
  /* U+5A6C */	"\xD4\xD9", "\xD4\xD5", "\x98\x5B", "\x00\x00",
  /* U+5A70 */	"\xD4\xE8", "\x98\x50", "\x00\x00", "\x00\x00",
  /* U+5A74 */	"\x00\x00", "\x00\x00", "\x00\x00", "\xB4\x40",
  /* U+5A78 */	"\xD8\xBB", "\x97\xBC", "\xD8\xB8", "\xD8\xC9",
  /* U+5A7C */	"\xD8\xBD", "\xD8\xCA", "\x92\xF3", "\xB4\x42",
  /* U+5A80 */	"\x00\x00", "\x93\x40", "\x98\x4D", "\xD8\xC6",
  /* U+5A84 */	"\xD8\xC3", "\x00\x00", "\x95\x72", "\x00\x00",
  /* U+5A88 */	"\xFD\xEF", "\x00\x00", "\xD8\xC4", "\xD8\xC7",
  /* U+5A8C */	"\xD8\xCB", "\x00\x00", "\xD4\xE3", "\xD8\xCD",
  /* U+5A90 */	"\xDD\x47", "\xFD\xC1", "\xB4\x43", "\xD8\xCE",
  /* U+5A94 */	"\xD8\xB6", "\xD8\xC0", "\xFB\xBA", "\xD8\xC5",
  /* U+5A98 */	"\x00\x00", "\x92\xEB", "\xB4\x41", "\xB4\x44",
  /* U+5A9C */	"\xD8\xCC", "\xD8\xCF", "\xD8\xBA", "\xD8\xB7",
  /* U+5AA0 */	"\xFC\x73", "\x97\xB7", "\xD8\xB9", "\x00\x00",
  /* U+5AA4 */	"\x00\x00", "\xD8\xBE", "\xD8\xBC", "\xB4\x45",
  /* U+5AA8 */	"\x00\x00", "\xD8\xC8", "\x00\x00", "\xFB\xB4",
  /* U+5AAC */	"\xD8\xBF", "\x00\x00", "\xD8\xC1", "\xD8\xB5",
  /* U+5AB0 */	"\xDC\xFA", "\xDC\xF8", "\xB7\x42", "\xB7\x40",
  /* U+5AB4 */	"\xDD\x43", "\xDC\xF9", "\xDD\x44", "\xDD\x40",
  /* U+5AB8 */	"\xDC\xF7", "\xDD\x46", "\xDC\xF6", "\xDC\xFD",
  /* U+5ABC */	"\xB6\xFE", "\xB6\xFD", "\xB6\xFC", "\xDC\xFB",
  /* U+5AC0 */	"\xDD\x41", "\xB6\xF9", "\xB7\x41", "\x90\xA7",
  /* U+5AC4 */	"\xDC\xF4", "\x00\x00", "\xDC\xFE", "\xDC\xF3",
  /* U+5AC8 */	"\xDC\xFC", "\xB6\xFA", "\xDD\x42", "\xDC\xF5",
  /* U+5ACC */	"\xB6\xFB", "\xDD\x45", "\x97\x41", "\x92\xF4",
  /* U+5AD0 */	"\x00\x00", "\x00\x00", "\x00\x00", "\xFB\xBC",
  /* U+5AD4 */	"\x00\x00", "\xE1\x6E", "\xB9\xE2", "\xB9\xE1",
  /* U+5AD8 */	"\xB9\xE3", "\xE1\x7A", "\xE1\x70", "\xE1\x76",
  /* U+5ADC */	"\xE1\x6B", "\xE1\x79", "\xE1\x78", "\xE1\x7C",
  /* U+5AE0 */	"\xE1\x75", "\xB9\xDE", "\xE1\x74", "\xB9\xE4",
  /* U+5AE4 */	"\x95\x77", "\xE1\x6D", "\xB9\xDF", "\x00\x00",
  /* U+5AE8 */	"\xE1\x7B", "\xB9\xE0", "\xE1\x6F", "\xE1\x72",
  /* U+5AEC */	"\xE1\x77", "\xE1\x71", "\xE1\x6C", "\x00\x00",
  /* U+5AF0 */	"\x9E\xE2", "\x00\x00", "\x8F\x78", "\xE1\x73",
  /* U+5AF4 */	"\xE5\x55", "\xBC\x61", "\xE5\x58", "\xE5\x57",
  /* U+5AF8 */	"\xE5\x5A", "\xE5\x5C", "\xF9\xDC", "\xBC\x5F",
  /* U+5AFC */	"\x00\x00", "\xE5\x56", "\x96\x72", "\xE5\x54",
  /* U+5B00 */	"\x00\x00", "\xE5\x5D", "\xE5\x5B", "\xE5\x59",
  /* U+5B04 */	"\x00\x00", "\xE5\x5F", "\x00\x00", "\xE5\x5E",
  /* U+5B08 */	"\xBC\x63", "\xBC\x5E", "\x00\x00", "\xBC\x60",
  /* U+5B0C */	"\xBC\x62", "\x9E\xB5", "\x00\x00", "\xE5\x60",
  /* U+5B10 */	"\xE9\x57", "\x96\x4B", "\x00\x00", "\xE9\x56",
  /* U+5B14 */	"\xE9\x55", "\x8C\xAC", "\xE9\x58", "\xE9\x51",
  /* U+5B18 */	"\x00\x00", "\xE9\x52", "\xE9\x5A", "\xE9\x53",
  /* U+5B1C */	"\x00\x00", "\xBE\xC5", "\xE9\x5C", "\xA0\xFA",
  /* U+5B20 */	"\xE9\x5B", "\xE9\x54", "\x00\x00", "\xEC\xD1",
  /* U+5B24 */	"\xC0\xA8", "\xEC\xCF", "\xEC\xD4", "\xEC\xD3",
  /* U+5B28 */	"\xE9\x59", "\x00\x00", "\xC0\xA7", "\x95\x75",
  /* U+5B2C */	"\xEC\xD2", "\xEC\xCE", "\xEC\xD6", "\xEC\xD5",
  /* U+5B30 */	"\xC0\xA6", "\x00\x00", "\xEC\xD0", "\x00\x00",
  /* U+5B34 */	"\xBE\xC6", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+5B38 */	"\xC2\x54", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+5B3C */	"\xEF\xC1", "\xF1\xFA", "\xF1\xFB", "\xF1\xFC",
  /* U+5B40 */	"\xC4\x5C", "\x90\xDA", "\x00\x00", "\xC4\x5D",
  /* U+5B44 */	"\x93\x67", "\xF4\x43", "\xFE\xA4", "\xF5\xC8",
  /* U+5B48 */	"\xF5\xC7", "\x00\x00", "\x90\xDF", "\xF6\xDB",
  /* U+5B4C */	"\xF6\xDC", "\xF7\xD5", "\xF8\xA7", "\x93\x54",
  /* U+5B50 */	"\xA4\x6C", "\xA4\x6D", "\x00\x00", "\xA4\x6E",
  /* U+5B54 */	"\xA4\xD5", "\xA5\xA5", "\xC9\xD3", "\xA6\x72",
  /* U+5B58 */	"\xA6\x73", "\x00\x00", "\xA7\xB7", "\xA7\xB8",
  /* U+5B5C */	"\xA7\xB6", "\xA7\xB5", "\x00\x00", "\xA9\x73",
  /* U+5B60 */	"\x00\x00", "\x00\x00", "\xCC\x55", "\xA9\x75",
  /* U+5B64 */	"\xA9\x74", "\xCC\x56", "\x89\x61", "\x00\x00",
  /* U+5B68 */	"\x8B\xB4", "\xAB\xC4", "\x00\x00", "\xAE\x5D",
  /* U+5B6C */	"\xD1\x65", "\x9D\xC0", "\xD4\xF0", "\x00\x00",
  /* U+5B70 */	"\xB1\x45", "\xB4\x47", "\xD4\xEF", "\xB4\x46",
  /* U+5B74 */	"\x8E\x48", "\xB9\xE5", "\xFB\xC5", "\xE1\x7D",
  /* U+5B78 */	"\xBE\xC7", "\x00\x00", "\xC0\xA9", "\xEC\xD7",
  /* U+5B7C */	"\xFB\xC7", "\xC4\x5E", "\x00\x00", "\xC5\x70",
  /* U+5B80 */	"\xC6\xCB", "\xC9\x72", "\xFA\x79", "\xA5\xA6",
  /* U+5B84 */	"\xC9\x73", "\xA6\x76", "\x00\x00", "\xA6\x74",
  /* U+5B88 */	"\xA6\x75", "\xA6\x77", "\x00\x00", "\xA7\xBA",
  /* U+5B8C */	"\xA7\xB9", "\x00\x00", "\xCA\xBC", "\xA7\xBB",
  /* U+5B90 */	"\x9E\x67", "\x00\x00", "\xCA\xBD", "\xCC\x57",
  /* U+5B94 */	"\x00\x00", "\xCC\x58", "\x8C\xD9", "\xA9\x76",
  /* U+5B98 */	"\xA9\x78", "\xA9\x7A", "\xA9\x77", "\xA9\x7B",
  /* U+5B9C */	"\xA9\x79", "\xFB\xD2", "\x89\x62", "\x89\x63",
  /* U+5BA0 */	"\x00\x00", "\x00\x00", "\xAB\xC8", "\xAB\xC5",
  /* U+5BA4 */	"\xAB\xC7", "\xAB\xC9", "\xAB\xC6", "\xD1\x66",
  /* U+5BA8 */	"\xCE\x77", "\x00\x00", "\xFC\x7D", "\x00\x00",
  /* U+5BAC */	"\xD1\x68", "\xD1\x67", "\xAE\x63", "\x00\x00",
  /* U+5BB0 */	"\xAE\x5F", "\x00\x00", "\x00\x00", "\xAE\x60",
  /* U+5BB4 */	"\xAE\x62", "\xAE\x64", "\xAE\x61", "\x00\x00",
  /* U+5BB8 */	"\xAE\x66", "\xAE\x65", "\x00\x00", "\x00\x00",
  /* U+5BBC */	"\x00\x00", "\x00\x00", "\x00\x00", "\xB1\x4A",
  /* U+5BC0 */	"\xD4\xF2", "\xD4\xF1", "\xB1\x49", "\x9F\x6B",
  /* U+5BC4 */	"\xB1\x48", "\xB1\x47", "\xB1\x4B", "\xB1\x46",
  /* U+5BC8 */	"\x00\x00", "\x00\x00", "\xD8\xD5", "\xD8\xD2",
  /* U+5BCC */	"\xB4\x49", "\xD8\xD1", "\xD8\xD6", "\x00\x00",
  /* U+5BD0 */	"\xB4\x4B", "\xD8\xD4", "\xB4\x48", "\xB4\x4A",
  /* U+5BD4 */	"\xD8\xD3", "\xFB\xCC", "\xDD\x48", "\xFE\xAE",
  /* U+5BD8 */	"\xDD\x49", "\xDD\x4A", "\x00\x00", "\x00\x00",
  /* U+5BDC */	"\x00\x00", "\x00\x00", "\xB9\xE6", "\xB9\xEE",
  /* U+5BE0 */	"\xE1\x7E", "\xB9\xE8", "\xB9\xEC", "\xE1\xA1",
  /* U+5BE4 */	"\xB9\xED", "\xB9\xE9", "\xB9\xEA", "\xB9\xE7",
  /* U+5BE8 */	"\xB9\xEB", "\xBC\x66", "\xD8\xD0", "\xBC\x67",
  /* U+5BEC */	"\xBC\x65", "\x00\x00", "\xBC\x64", "\xE9\x5D",
  /* U+5BF0 */	"\xBE\xC8", "\xEC\xD8", "\xEC\xD9", "\xFB\xD1",
  /* U+5BF4 */	"\x00\x00", "\xC3\x64", "\xC4\x5F", "\x00\x00",
  /* U+5BF8 */	"\xA4\x6F", "\x00\x00", "\xA6\x78", "\x00\x00",
  /* U+5BFC */	"\x00\x00", "\x00\x00", "\x00\x00", "\xFB\x75",
  /* U+5C00 */	"\x00\x00", "\xAB\xCA", "\x00\x00", "\xD1\x69",
  /* U+5C04 */	"\xAE\x67", "\xFB\xD4", "\x00\x00", "\xB1\x4E",
  /* U+5C08 */	"\xB1\x4D", "\xB1\x4C", "\xB4\x4C", "\xB4\x4D",
  /* U+5C0C */	"\xD8\xD7", "\xB9\xEF", "\xBE\xC9", "\xA4\x70",
  /* U+5C10 */	"\xC9\x5C", "\xA4\xD6", "\xC9\x74", "\xFB\xD6",
  /* U+5C14 */	"\xFB\xD8", "\xC9\xD4", "\xA6\x79", "\x00\x00",
  /* U+5C18 */	"\x00\x00", "\x00\x00", "\xA9\x7C", "\x00\x00",
  /* U+5C1C */	"\x8B\x5D", "\x00\x00", "\x93\x4C", "\xDD\x4B",
  /* U+5C20 */	"\x9A\xE2", "\x00\x00", "\xA4\x71", "\x8B\xC9",
  /* U+5C24 */	"\xA4\xD7", "\xC9\xD5", "\x00\x00", "\x00\x00",
  /* U+5C28 */	"\xCA\xBE", "\x00\x00", "\xCA\xBF", "\x00\x00",
  /* U+5C2C */	"\xA7\xBC", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+5C30 */	"\xD8\xD8", "\xB4\x4E", "\x00\x00", "\xDD\x4C",
  /* U+5C34 */	"\x00\x00", "\x00\x00", "\x00\x00", "\xC0\xAA",
  /* U+5C38 */	"\xA4\x72", "\xA4\xA8", "\xA4\xD8", "\xC9\x75",
  /* U+5C3C */	"\xA5\xA7", "\x00\x00", "\xA7\xC0", "\xA7\xBF",
  /* U+5C40 */	"\xA7\xBD", "\xA7\xBE", "\x00\x00", "\x00\x00",
  /* U+5C44 */	"\xCC\x59", "\xA9\x7E", "\xA9\xA1", "\xCC\x5A",
  /* U+5C48 */	"\xA9\x7D", "\xFB\xDB", "\x9F\xC9", "\xAB\xCE",
  /* U+5C4C */	"\xCE\x78", "\xAB\xCD", "\xAB\xCB", "\xAB\xCC",
  /* U+5C50 */	"\xAE\x6A", "\xAE\x68", "\x00\x00", "\x9F\x44",
  /* U+5C54 */	"\xD1\x6B", "\xAE\x69", "\xD1\x6A", "\x00\x00",
  /* U+5C58 */	"\xAE\x5E", "\xD4\xF3", "\x00\x00", "\x00\x00",
  /* U+5C5C */	"\xB1\x50", "\xB1\x51", "\x98\xED", "\x00\x00",
  /* U+5C60 */	"\xB1\x4F", "\x00\x00", "\xB9\xF0", "\xE1\xA2",
  /* U+5C64 */	"\xBC\x68", "\xBC\x69", "\x00\x00", "\xE5\x61",
  /* U+5C68 */	"\xC0\xAB", "\xEF\xC2", "\xEF\xC3", "\x00\x00",
  /* U+5C6C */	"\xC4\xDD", "\xF8\xA8", "\xC9\x4B", "\xA4\xD9",
  /* U+5C70 */	"\x00\x00", "\xA4\x73", "\x00\x00", "\xC9\x77",
  /* U+5C74 */	"\xC9\x76", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+5C78 */	"\x00\x00", "\xA6\x7A", "\xC9\xD7", "\xC9\xD8",
  /* U+5C7C */	"\xC9\xD6", "\x00\x00", "\xC9\xD9", "\x00\x00",
  /* U+5C80 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+5C84 */	"\x00\x00", "\xFB\xDD", "\xCA\xC7", "\x00\x00",
  /* U+5C88 */	"\xCA\xC2", "\xCA\xC4", "\xCA\xC6", "\xCA\xC3",
  /* U+5C8C */	"\xA7\xC4", "\xCA\xC0", "\x00\x00", "\xCA\xC1",
  /* U+5C90 */	"\xA7\xC1", "\xA7\xC2", "\xCA\xC5", "\xCA\xC8",
  /* U+5C94 */	"\xA7\xC3", "\xCA\xC9", "\x00\x00", "\x00\x00",
  /* U+5C98 */	"\x00\x00", "\x8D\xF2", "\x89\x64", "\x00\x00",
  /* U+5C9C */	"\xFD\xF2", "\xCC\x68", "\x93\x4D", "\xCC\x62",
  /* U+5CA0 */	"\xCC\x5D", "\xA9\xA3", "\xCC\x65", "\xCC\x63",
  /* U+5CA4 */	"\xCC\x5C", "\xCC\x69", "\xCC\x6C", "\xCC\x67",
  /* U+5CA8 */	"\xCC\x60", "\xA9\xA5", "\xCC\x66", "\xA9\xA6",
  /* U+5CAC */	"\xCC\x61", "\xCC\x64", "\xCC\x5B", "\xCC\x5F",
  /* U+5CB0 */	"\xCC\x6B", "\xA9\xA7", "\x00\x00", "\xA9\xA8",
  /* U+5CB4 */	"\x00\x00", "\xCC\x5E", "\xCC\x6A", "\xA9\xA2",
  /* U+5CB8 */	"\xA9\xA4", "\x00\x00", "\xFB\xE7", "\x00\x00",
  /* U+5CBC */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+5CC0 */	"\x00\x00", "\xA0\xF2", "\x98\x68", "\x00\x00",
  /* U+5CC4 */	"\x00\x00", "\x00\x00", "\xCE\xAB", "\xCE\xA4",
  /* U+5CC8 */	"\xCE\xAA", "\xCE\xA3", "\xCE\xA5", "\xCE\x7D",
  /* U+5CCC */	"\xCE\x7B", "\x00\x00", "\xCE\xAC", "\xCE\xA9",
  /* U+5CD0 */	"\xCE\x79", "\x9F\x58", "\xAB\xD0", "\xCE\xA7",
  /* U+5CD4 */	"\xCE\xA8", "\x00\x00", "\xCE\xA6", "\xCE\x7C",
  /* U+5CD8 */	"\xCE\x7A", "\xAB\xCF", "\xCE\xA2", "\xCE\x7E",
  /* U+5CDC */	"\x00\x00", "\x00\x00", "\xCE\xA1", "\xCE\xAD",
  /* U+5CE0 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+5CE4 */	"\x00\x00", "\x8D\x73", "\x00\x00", "\x00\x00",
  /* U+5CE8 */	"\xAE\x6F", "\xFB\xDE", "\xAE\x6E", "\x00\x00",
  /* U+5CEC */	"\xD1\x6C", "\xAE\x6B", "\xD1\x6E", "\xFB\xDF",
  /* U+5CF0 */	"\xAE\x70", "\xD1\x6F", "\x00\x00", "\x00\x00",
  /* U+5CF4 */	"\xAE\x73", "\x8C\x48", "\xAE\x71", "\xD1\x70",
  /* U+5CF8 */	"\xCE\xAE", "\xD1\x72", "\x00\x00", "\xAE\x6D",
  /* U+5CFC */	"\x00\x00", "\xAE\x6C", "\x00\x00", "\xD1\x6D",
  /* U+5D00 */	"\xD1\x71", "\xAE\x72", "\x00\x00", "\x00\x00",
  /* U+5D04 */	"\x00\x00", "\x00\x00", "\xB1\x53", "\xB1\x52",
  /* U+5D08 */	"\x00\x00", "\x00\x00", "\x00\x00", "\xD4\xF5",
  /* U+5D0C */	"\xD4\xF9", "\xD4\xFB", "\xB1\x54", "\xD4\xFE",
  /* U+5D10 */	"\xFB\xE3", "\xB1\x58", "\xD5\x41", "\x00\x00",
  /* U+5D14 */	"\xB1\x5A", "\x8D\xA8", "\xB1\x56", "\xB1\x5E",
  /* U+5D18 */	"\xFB\xE4", "\xB1\x5B", "\xD4\xF7", "\xB1\x55",
  /* U+5D1C */	"\x00\x00", "\xD4\xF6", "\xD4\xF4", "\xD5\x43",
  /* U+5D20 */	"\xD4\xF8", "\x00\x00", "\xB1\x57", "\xD5\x42",
  /* U+5D24 */	"\xB1\x5C", "\xD4\xFD", "\xD4\xFC", "\xB1\x5D",
  /* U+5D28 */	"\xD4\xFA", "\xB1\x59", "\x00\x00", "\x00\x00",
  /* U+5D2C */	"\x9C\x75", "\x00\x00", "\xD5\x44", "\x98\x78",
  /* U+5D30 */	"\xD5\x40", "\xD8\xE7", "\xD8\xEE", "\xD8\xE3",
  /* U+5D34 */	"\xB4\x51", "\xD8\xDF", "\xD8\xEF", "\xD8\xD9",
  /* U+5D38 */	"\xD8\xEC", "\xD8\xEA", "\xD8\xE4", "\x00\x00",
  /* U+5D3C */	"\xD8\xED", "\xD8\xE6", "\x8D\x60", "\xD8\xDE",
  /* U+5D40 */	"\xD8\xF0", "\xD8\xDC", "\xD8\xE9", "\xD8\xDA",
  /* U+5D44 */	"\x00\x00", "\xD8\xF1", "\xFB\xE5", "\xB4\x52",
  /* U+5D48 */	"\x8D\x61", "\xD8\xEB", "\xDD\x4F", "\xD8\xDD",
  /* U+5D4C */	"\xB4\x4F", "\x00\x00", "\xD8\xE1", "\x00\x00",
  /* U+5D50 */	"\xB4\x50", "\xD8\xE0", "\xD8\xE5", "\x00\x00",
  /* U+5D54 */	"\x00\x00", "\xD8\xE2", "\x8D\x62", "\xA0\xA1",
  /* U+5D58 */	"\x00\x00", "\xD8\xE8", "\x00\x00", "\x9C\x40",
  /* U+5D5C */	"\x00\x00", "\x00\x00", "\xDD\x53", "\x00\x00",
  /* U+5D60 */	"\x00\x00", "\x00\x00", "\xDD\x56", "\xDD\x4E",
  /* U+5D64 */	"\x00\x00", "\xDD\x50", "\x00\x00", "\xDD\x55",
  /* U+5D68 */	"\xDD\x54", "\xB7\x43", "\x00\x00", "\xD8\xDB",
  /* U+5D6C */	"\xDD\x52", "\x00\x00", "\x00\x00", "\xB7\x44",
  /* U+5D70 */	"\x98\xAD", "\xDD\x4D", "\xDD\x51", "\x00\x00",
  /* U+5D74 */	"\x9E\xEA", "\x00\x00", "\x00\x00", "\xE1\xA9",
  /* U+5D78 */	"\x00\x00", "\xE1\xB0", "\xE1\xA7", "\x8C\xD4",
  /* U+5D7C */	"\xE1\xAE", "\xE1\xA5", "\xE1\xAD", "\xE1\xB1",
  /* U+5D80 */	"\xE1\xA4", "\xE1\xA8", "\xE1\xA3", "\x00\x00",
  /* U+5D84 */	"\xB9\xF1", "\x9C\xEB", "\xE1\xA6", "\xB9\xF2",
  /* U+5D88 */	"\xE1\xAC", "\xE1\xAB", "\xE1\xAA", "\xFB\xE0",
  /* U+5D8C */	"\x00\x00", "\xE1\xAF", "\x9F\x51", "\x00\x00",
  /* U+5D90 */	"\x00\x00", "\x00\x00", "\xE5\x65", "\xE5\x67",
  /* U+5D94 */	"\xBC\x6B", "\xE5\x68", "\x00\x00", "\xE5\x63",
  /* U+5D98 */	"\x00\x00", "\xE5\x62", "\xE5\x6C", "\x00\x00",
  /* U+5D9C */	"\xE5\x6A", "\xBC\x6A", "\xE5\x6D", "\xE5\x64",
  /* U+5DA0 */	"\xE5\x69", "\xE5\x6B", "\xE5\x66", "\x00\x00",
  /* U+5DA4 */	"\x8D\x65", "\x00\x00", "\x00\x00", "\xE9\x61",
  /* U+5DA8 */	"\xE9\x66", "\xE9\x60", "\xE9\x65", "\x9C\xF1",
  /* U+5DAC */	"\xE9\x5E", "\xE9\x68", "\xE9\x64", "\xE9\x69",
  /* U+5DB0 */	"\xE9\x63", "\xE9\x5F", "\xE9\x67", "\x00\x00",
  /* U+5DB4 */	"\xE9\x6A", "\xE9\x62", "\xFC\x58", "\xEC\xDA",
  /* U+5DB8 */	"\xC0\xAF", "\x8D\x66", "\xC0\xAD", "\x00\x00",
  /* U+5DBC */	"\xC0\xAC", "\xC0\xAE", "\x00\x00", "\x00\x00",
  /* U+5DC0 */	"\xEF\xC4", "\x96\x54", "\xF1\x72", "\xF1\xFD",
  /* U+5DC4 */	"\x00\x00", "\x00\x00", "\xF4\x44", "\xF4\x45",
  /* U+5DC8 */	"\x00\x00", "\xC4\x60", "\x00\x00", "\xF5\xC9",
  /* U+5DCC */	"\x00\x00", "\xC4\xDE", "\x00\x00", "\xF5\xCA",
  /* U+5DD0 */	"\x00\x00", "\xF6\xDE", "\xC5\x72", "\x00\x00",
  /* U+5DD4 */	"\xC5\x71", "\xF6\xDD", "\xC5\xC9", "\xFB\xE8",
  /* U+5DD8 */	"\xF7\xD6", "\x00\x00", "\x00\x00", "\xC6\xCC",
  /* U+5DDC */	"\x00\x00", "\xA4\x74", "\xA6\x7B", "\xC9\xDA",
  /* U+5DE0 */	"\xCA\xCA", "\xA8\xB5", "\xB1\x5F", "\x00\x00",
  /* U+5DE4 */	"\x00\x00", "\xA4\x75", "\xA5\xAA", "\xA5\xA9",
  /* U+5DE8 */	"\xA5\xA8", "\x00\x00", "\x00\x00", "\xA7\xC5",
  /* U+5DEC */	"\x00\x00", "\x00\x00", "\xAE\x74", "\x00\x00",
  /* U+5DF0 */	"\xDD\x57", "\xA4\x76", "\xA4\x77", "\xA4\x78",
  /* U+5DF4 */	"\xA4\xDA", "\x9F\xCE", "\x00\x00", "\xAB\xD1",
  /* U+5DF8 */	"\x00\x00", "\xCE\xAF", "\x00\x00", "\x00\x00",
  /* U+5DFC */	"\x00\x00", "\xB4\x53", "\xA4\x79", "\xC9\x5D",
  /* U+5E00 */	"\x00\x00", "\x00\x00", "\xA5\xAB", "\xA5\xAC",
  /* U+5E04 */	"\xC9\x78", "\x00\x00", "\xA6\x7C", "\x00\x00",
  /* U+5E08 */	"\x00\x00", "\xFB\xFC", "\xCA\xCB", "\x9A\xE4",
  /* U+5E0C */	"\xA7\xC6", "\x00\x00", "\xCA\xCC", "\x00\x00",
  /* U+5E10 */	"\x00\x00", "\xA9\xAE", "\x9F\x75", "\x00\x00",
  /* U+5E14 */	"\xCC\x6E", "\xA9\xAC", "\xA9\xAB", "\xCC\x6D",
  /* U+5E18 */	"\xA9\xA9", "\xCC\x6F", "\xA9\xAA", "\xA9\xAD",
  /* U+5E1C */	"\x00\x00", "\xAB\xD2", "\x00\x00", "\xAB\xD4",
  /* U+5E20 */	"\xCE\xB3", "\xCE\xB0", "\xCE\xB1", "\xCE\xB2",
  /* U+5E24 */	"\xCE\xB4", "\xAB\xD3", "\x00\x00", "\x00\x00",
  /* U+5E28 */	"\xD1\x74", "\xD1\x73", "\x00\x00", "\xAE\x76",
  /* U+5E2C */	"\x00\x00", "\xAE\x75", "\xFB\xF1", "\x00\x00",
  /* U+5E30 */	"\x00\x00", "\x00\x00", "\x00\x00", "\xB1\x62",
  /* U+5E34 */	"\xD5\x46", "\x00\x00", "\xB1\x61", "\xB1\x63",
  /* U+5E38 */	"\xB1\x60", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+5E3C */	"\x00\x00", "\xB4\x55", "\xD5\x45", "\x00\x00",
  /* U+5E40 */	"\xB4\x56", "\xD8\xF3", "\x8D\x69", "\xB4\x57",
  /* U+5E44 */	"\xD8\xF2", "\xB4\x54", "\x00\x00", "\x00\x00",
  /* U+5E48 */	"\x93\x4F", "\x00\x00", "\xDD\x5A", "\xDD\x5C",
  /* U+5E4C */	"\xB7\x45", "\xDD\x5B", "\xDD\x59", "\xDD\x58",
  /* U+5E50 */	"\x00\x00", "\x00\x00", "\x00\x00", "\xE1\xB4",
  /* U+5E54 */	"\xB9\xF7", "\xB9\xF5", "\x00\x00", "\xB9\xF6",
  /* U+5E58 */	"\xE1\xB2", "\xE1\xB3", "\x00\x00", "\xB9\xF3",
  /* U+5E5C */	"\xE5\x71", "\xE5\x6F", "\x93\x4E", "\xBC\x6D",
  /* U+5E60 */	"\xE5\x70", "\xBC\x6E", "\xBC\x6C", "\xB9\xF4",
  /* U+5E64 */	"\x00\x00", "\x00\x00", "\xE9\x6D", "\xE9\x6B",
  /* U+5E68 */	"\xE9\x6C", "\xE5\x6E", "\xEC\xDC", "\xC0\xB0",
  /* U+5E6C */	"\xEC\xDB", "\xEF\xC5", "\xEF\xC6", "\xE9\x6E",
  /* U+5E70 */	"\xF1\xFE", "\x00\x00", "\xA4\x7A", "\xA5\xAD",
  /* U+5E74 */	"\xA6\x7E", "\xC9\xDB", "\xA6\x7D", "\x00\x00",
  /* U+5E78 */	"\xA9\xAF", "\xB7\x46", "\xFB\xF4", "\xA4\xDB",
  /* U+5E7C */	"\xA5\xAE", "\xAB\xD5", "\xB4\x58", "\xC6\xCE",
  /* U+5E80 */	"\xC9\x79", "\x00\x00", "\xC9\x7A", "\xFB\xC3",
  /* U+5E84 */	"\xC9\xDC", "\x00\x00", "\x89\x65", "\xA7\xC8",
  /* U+5E88 */	"\xCA\xD0", "\xCA\xCE", "\xA7\xC9", "\xCA\xCD",
  /* U+5E8C */	"\xCA\xCF", "\xCA\xD1", "\x00\x00", "\xA7\xC7",
  /* U+5E90 */	"\x00\x00", "\x00\x00", "\x8C\x7A", "\x00\x00",
  /* U+5E94 */	"\x00\x00", "\xA9\xB3", "\xA9\xB4", "\xA9\xB1",
  /* U+5E98 */	"\x00\x00", "\x8C\x7B", "\xA9\xB0", "\xCE\xB8",
  /* U+5E9C */	"\xA9\xB2", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+5EA0 */	"\xAB\xD6", "\x00\x00", "\xCE\xB7", "\xCE\xB9",
  /* U+5EA4 */	"\xCE\xB6", "\xCE\xBA", "\xAB\xD7", "\xAE\x79",
  /* U+5EA8 */	"\xD1\x75", "\x00\x00", "\xD1\x77", "\xAE\x77",
  /* U+5EAC */	"\xD1\x78", "\xAE\x78", "\xD1\x76", "\x00\x00",
  /* U+5EB0 */	"\xCE\xB5", "\xD5\x47", "\xD5\x4A", "\xD5\x4B",
  /* U+5EB4 */	"\xD5\x48", "\xB1\x67", "\xB1\x66", "\xB1\x64",
  /* U+5EB8 */	"\xB1\x65", "\xD5\x49", "\x00\x00", "\x00\x00",
  /* U+5EBC */	"\x00\x00", "\x8D\x6A", "\xB1\x68", "\x00\x00",
  /* U+5EC0 */	"\x00\x00", "\xB4\x5A", "\xB4\x5B", "\x00\x00",
  /* U+5EC4 */	"\xB4\x5C", "\xDD\x5D", "\xDD\x5F", "\xDD\x61",
  /* U+5EC8 */	"\xB7\x48", "\xB7\x47", "\xB4\x59", "\xDD\x60",
  /* U+5ECC */	"\xDD\x5E", "\x93\x53", "\xE1\xB8", "\x00\x00",
  /* U+5ED0 */	"\x9D\xFB", "\xE1\xB6", "\xE1\xBC", "\xB9\xF8",
  /* U+5ED4 */	"\xE1\xBD", "\xE1\xBA", "\xB9\xF9", "\xE1\xB7",
  /* U+5ED8 */	"\xE1\xB5", "\xE1\xBB", "\xBC\x70", "\xE5\x73",
  /* U+5EDC */	"\xE1\xB9", "\xBC\x72", "\xE5\x74", "\xBC\x71",
  /* U+5EE0 */	"\xBC\x74", "\xE5\x75", "\xBC\x6F", "\xBC\x73",
  /* U+5EE4 */	"\x00\x00", "\xE9\x73", "\xE9\x71", "\xE9\x70",
  /* U+5EE8 */	"\xE9\x72", "\xE9\x6F", "\x00\x00", "\x00\x00",
  /* U+5EEC */	"\xC3\x66", "\x00\x00", "\xF4\x46", "\xF4\x47",
  /* U+5EF0 */	"\x00\x00", "\xF5\xCB", "\xF6\xDF", "\xC6\x55",
  /* U+5EF4 */	"\xFB\xFD", "\x00\x00", "\xA9\xB5", "\xA7\xCA",
  /* U+5EF8 */	"\x90\x59", "\xFC\x40", "\xAB\xD8", "\xFC\x41",
  /* U+5EFC */	"\xFC\x43", "\x00\x00", "\xA4\x7B", "\xA4\xDC",
  /* U+5F00 */	"\x00\x00", "\xA5\xAF", "\xC9\xDD", "\x00\x00",
  /* U+5F04 */	"\xA7\xCB", "\xCA\xD2", "\x00\x00", "\xCE\xBB",
  /* U+5F08 */	"\xAB\xD9", "\x00\x00", "\xB9\xFA", "\xA4\x7C",
  /* U+5F0C */	"\x93\x61", "\xFC\x46", "\x93\x62", "\xA6\xA1",
  /* U+5F10 */	"\x00\x00", "\x00\x00", "\xB7\x49", "\xA4\x7D",
  /* U+5F14 */	"\xA4\xDD", "\xA4\xDE", "\x00\x00", "\xA5\xB1",
  /* U+5F18 */	"\xA5\xB0", "\x00\x00", "\xC9\xDE", "\xA6\xA2",
  /* U+5F1C */	"\x00\x00", "\xCA\xD3", "\x00\x00", "\xA7\xCC",
  /* U+5F20 */	"\x00\x00", "\x00\x00", "\xCC\x71", "\xCC\x72",
  /* U+5F24 */	"\xCC\x73", "\x8D\x6B", "\xA9\xB6", "\xA9\xB7",
  /* U+5F28 */	"\xCC\x70", "\xA9\xB8", "\x00\x00", "\x00\x00",
  /* U+5F2C */	"\x00\x00", "\xAB\xDA", "\xCE\xBC", "\x00\x00",
  /* U+5F30 */	"\xD1\x7A", "\xAE\x7A", "\x00\x00", "\xD1\x79",
  /* U+5F34 */	"\x00\x00", "\xB1\x69", "\xD5\x4C", "\xB1\x6A",
  /* U+5F38 */	"\xD5\x4D", "\x00\x00", "\xFC\x4C", "\x00\x00",
  /* U+5F3C */	"\xB4\x5D", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+5F40 */	"\xDD\x62", "\x00\x00", "\x00\x00", "\xE1\xBF",
  /* U+5F44 */	"\xE1\xBE", "\x00\x00", "\xB9\xFB", "\x00\x00",
  /* U+5F48 */	"\xBC\x75", "\xE5\x76", "\xBE\xCA", "\xE9\x74",
  /* U+5F4C */	"\xC0\xB1", "\x95\xB8", "\xC5\x73", "\xF7\xD8",
  /* U+5F50 */	"\xC6\xD0", "\x8B\xCA", "\x00\x00", "\x00\x00",
  /* U+5F54 */	"\xCC\x74", "\x00\x00", "\xCE\xBD", "\xB1\x6B",
  /* U+5F58 */	"\xD8\xF4", "\xB7\x4A", "\x00\x00", "\x00\x00",
  /* U+5F5C */	"\x98\x7A", "\xC2\x55", "\x00\x00", "\x00\x00",
  /* U+5F60 */	"\x00\x00", "\xC6\xD1", "\xA7\xCE", "\xFC\x51",
  /* U+5F64 */	"\xA7\xCD", "\xAB\xDB", "\x00\x00", "\xD1\x7B",
  /* U+5F68 */	"\x00\x00", "\xB1\x6D", "\xB3\x43", "\xB1\x6E",
  /* U+5F6C */	"\xB1\x6C", "\xB4\x5E", "\x00\x00", "\xE1\xC0",
  /* U+5F70 */	"\xB9\xFC", "\xBC\x76", "\xFC\x54", "\xC9\x4C",
  /* U+5F74 */	"\xC9\xDF", "\x00\x00", "\xCA\xD5", "\xA7\xCF",
  /* U+5F78 */	"\xCA\xD4", "\xA7\xD0", "\x00\x00", "\xFA\xAF",
  /* U+5F7C */	"\xA9\xBC", "\xCC\x77", "\xCC\x76", "\xA9\xBB",
  /* U+5F80 */	"\xA9\xB9", "\xA9\xBA", "\xCC\x75", "\x8D\x6C",
  /* U+5F84 */	"\x00\x00", "\xAB\xDD", "\xCE\xBE", "\xAB\xE0",
  /* U+5F88 */	"\xAB\xDC", "\xAB\xE2", "\xAB\xDE", "\xAB\xDF",
  /* U+5F8C */	"\xAB\xE1", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+5F90 */	"\xAE\x7D", "\xAE\x7C", "\xAE\x7B", "\x00\x00",
  /* U+5F94 */	"\x00\x00", "\x00\x00", "\xD5\x4F", "\xB1\x6F",
  /* U+5F98 */	"\xB1\x72", "\xB1\x70", "\x00\x00", "\xD5\x4E",
  /* U+5F9C */	"\xB1\x75", "\x00\x00", "\xB1\x71", "\xD5\x50",
  /* U+5FA0 */	"\xB1\x74", "\xB1\x73", "\x00\x00", "\x00\x00",
  /* U+5FA4 */	"\xFA\x61", "\xD8\xF6", "\xD8\xF5", "\xFC\x57",
  /* U+5FA8 */	"\xB4\x61", "\xB4\x5F", "\xB4\x60", "\xD8\xF7",
  /* U+5FAC */	"\xB7\x4B", "\xDD\x64", "\xB7\x4C", "\xDD\x63",
  /* U+5FB0 */	"\x00\x00", "\x9B\x70", "\xE5\x77", "\x00\x00",
  /* U+5FB4 */	"\x00\x00", "\xBC\x78", "\xE1\xC1", "\xBC\x77",
  /* U+5FB8 */	"\x00\x00", "\xB9\xFD", "\xA0\x51", "\xEC\xDE",
  /* U+5FBC */	"\xE9\x75", "\xC0\xB2", "\xEC\xDD", "\xF2\x40",
  /* U+5FC0 */	"\xF4\x48", "\xF4\x49", "\x8C\x7C", "\xA4\xDF",
  /* U+5FC4 */	"\x8B\xCB", "\xA5\xB2", "\x00\x00", "\x00\x00",
  /* U+5FC8 */	"\x00\x00", "\xC9\x7B", "\x00\x00", "\x00\x00",
  /* U+5FCC */	"\xA7\xD2", "\xA7\xD4", "\x00\x00", "\xC9\xE2",
  /* U+5FD0 */	"\xCA\xD8", "\xCA\xD7", "\xCA\xD6", "\x00\x00",
  /* U+5FD4 */	"\xC9\xE1", "\xC9\xE0", "\xA6\xA4", "\xA7\xD3",
  /* U+5FD8 */	"\xA7\xD1", "\xA6\xA3", "\x00\x00", "\x93\x6E",
  /* U+5FDC */	"\x00\x00", "\xA9\xBD", "\xCC\x78", "\xFC\xD5",
  /* U+5FE0 */	"\xA9\xBE", "\xCA\xDD", "\x00\x00", "\xCA\xDF",
  /* U+5FE4 */	"\xCA\xDE", "\xCC\x79", "\x00\x00", "\x00\x00",
  /* U+5FE8 */	"\xCA\xDA", "\x00\x00", "\xA7\xD8", "\xA7\xD6",
  /* U+5FEC */	"\x00\x00", "\xCA\xD9", "\xCA\xDB", "\xCA\xE1",
  /* U+5FF0 */	"\x00\x00", "\xA7\xD5", "\x00\x00", "\xCA\xDC",
  /* U+5FF4 */	"\xCA\xE5", "\xA9\xC0", "\x00\x00", "\xCA\xE2",
  /* U+5FF8 */	"\xA7\xD7", "\x00\x00", "\xCA\xE0", "\xCA\xE3",
  /* U+5FFC */	"\x00\x00", "\xA9\xBF", "\x00\x00", "\xA9\xC1",
  /* U+6000 */	"\xCA\xE4", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+6004 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+6008 */	"\x00\x00", "\xCC\xAF", "\xCC\xA2", "\xCC\x7E",
  /* U+600C */	"\xCC\xAE", "\xCC\xA9", "\xAB\xE7", "\xA9\xC2",
  /* U+6010 */	"\xCC\xAA", "\xCC\xAD", "\xAB\xE3", "\xCC\xAC",
  /* U+6014 */	"\xA9\xC3", "\xA9\xC8", "\xA9\xC6", "\xCC\xA3",
  /* U+6018 */	"\x00\x00", "\xCC\x7C", "\xCC\xA5", "\xA9\xCD",
  /* U+601C */	"\xCC\xB0", "\xAB\xE4", "\xCC\xA6", "\x00\x00",
  /* U+6020 */	"\xAB\xE5", "\xA9\xC9", "\xCC\xA8", "\xFC\xA9",
  /* U+6024 */	"\xCE\xCD", "\xAB\xE6", "\xCC\x7B", "\xA9\xCA",
  /* U+6028 */	"\xAB\xE8", "\xA9\xCB", "\xA9\xC7", "\xA9\xCC",
  /* U+602C */	"\xCC\xA7", "\xCC\x7A", "\xCC\xAB", "\xA9\xC4",
  /* U+6030 */	"\x00\x00", "\xFC\x61", "\xCC\x7D", "\xCC\xA4",
  /* U+6034 */	"\xCC\xA1", "\xA9\xC5", "\x00\x00", "\xCE\xBF",
  /* U+6038 */	"\x00\x00", "\xCE\xC0", "\x00\x00", "\x89\x66",
  /* U+603C */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+6040 */	"\xCE\xCA", "\xD1\xA1", "\xCE\xCB", "\xAB\xEE",
  /* U+6044 */	"\xCE\xCE", "\xCE\xC4", "\xAB\xED", "\xCE\xC6",
  /* U+6048 */	"\x00\x00", "\xCE\xC7", "\xFA\xCB", "\x00\x00",
  /* U+604C */	"\xCE\xC9", "\xAB\xE9", "\x00\x00", "\x00\x00",
  /* U+6050 */	"\xAE\xA3", "\x00\x00", "\xF9\xDA", "\xCE\xC5",
  /* U+6054 */	"\xCE\xC1", "\xAE\xA4", "\x00\x00", "\x00\x00",
  /* U+6058 */	"\xCE\xCF", "\xAE\x7E", "\xD1\x7D", "\xCE\xC8",
  /* U+605C */	"\x00\x00", "\xD1\x7C", "\xCE\xC3", "\xCE\xCC",
  /* U+6060 */	"\x00\x00", "\x00\x00", "\xAB\xEC", "\xAE\xA1",
  /* U+6064 */	"\xAB\xF2", "\xAE\xA2", "\xCE\xD0", "\xD1\x7E",
  /* U+6068 */	"\xAB\xEB", "\xAE\xA6", "\xAB\xF1", "\xAB\xF0",
  /* U+606C */	"\xAB\xEF", "\xAE\xA5", "\xCE\xD1", "\xAE\xA7",
  /* U+6070 */	"\xAB\xEA", "\x00\x00", "\xCE\xC2", "\x00\x00",
  /* U+6074 */	"\x00\x00", "\x93\x7A", "\x00\x00", "\xA0\xE0",
  /* U+6078 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+607C */	"\x00\x00", "\x00\x00", "\x93\x6B", "\xB1\x76",
  /* U+6080 */	"\xD1\xA4", "\xD1\xA6", "\x00\x00", "\xD1\xA8",
  /* U+6084 */	"\xAE\xA8", "\xAE\xAE", "\xD5\x53", "\xD1\xAC",
  /* U+6088 */	"\xD1\xA3", "\xB1\x78", "\xD5\x51", "\x00\x00",
  /* U+608C */	"\xAE\xAD", "\xAE\xAB", "\xD1\xAE", "\x00\x00",
  /* U+6090 */	"\xD5\x52", "\x00\x00", "\xD1\xA5", "\x00\x00",
  /* U+6094 */	"\xAE\xAC", "\xD1\xA9", "\xAE\xAF", "\xD1\xAB",
  /* U+6098 */	"\x00\x00", "\x00\x00", "\xAE\xAA", "\xD1\xAA",
  /* U+609C */	"\xD1\xAD", "\xD1\xA7", "\xFC\x6B", "\xAE\xA9",
  /* U+60A0 */	"\xB1\x79", "\x00\x00", "\xD1\xA2", "\xB1\x77",
  /* U+60A4 */	"\xA0\xDC", "\x00\x00", "\x00\x00", "\x94\x68",
  /* U+60A8 */	"\xB1\x7A", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+60AC */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+60B0 */	"\xD5\x55", "\xD5\x5E", "\xB4\x64", "\xFC\x6D",
  /* U+60B4 */	"\xB1\x7C", "\xB1\xA3", "\xB4\x65", "\xD5\x60",
  /* U+60B8 */	"\xB1\xAA", "\xD8\xF9", "\xD5\x56", "\xB1\xA2",
  /* U+60BC */	"\xB1\xA5", "\xB1\x7E", "\xD5\x54", "\xD5\x62",
  /* U+60C0 */	"\xD5\x65", "\xD9\x49", "\x00\x00", "\xD5\x63",
  /* U+60C4 */	"\xD8\xFD", "\xB1\xA1", "\xB1\xA8", "\xB1\xAC",
  /* U+60C8 */	"\xD5\x5D", "\xD8\xF8", "\xD5\x61", "\xB1\x7B",
  /* U+60CC */	"\xD8\xFA", "\xD5\x64", "\xD8\xFC", "\xD5\x59",
  /* U+60D0 */	"\x00\x00", "\xB4\x62", "\x00\x00", "\xD5\x57",
  /* U+60D4 */	"\xD5\x58", "\xB1\xA7", "\x00\x00", "\x8D\x71",
  /* U+60D8 */	"\xB1\xA6", "\xD5\x5B", "\xB1\xAB", "\xD5\x5F",
  /* U+60DC */	"\xB1\xA4", "\xD5\x5C", "\xFD\x64", "\xB1\xA9",
  /* U+60E0 */	"\xB4\x66", "\xB4\x63", "\xD8\xFB", "\x99\xBA",
  /* U+60E4 */	"\xD5\x5A", "\x00\x00", "\xB1\x7D", "\x9A\xD0",
  /* U+60E8 */	"\x9A\x61", "\xA0\xE5", "\x00\x00", "\x00\x00",
  /* U+60EC */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+60F0 */	"\xB4\x6B", "\xB4\x6F", "\xD9\x40", "\xB7\x51",
  /* U+60F4 */	"\xB4\x6D", "\xD9\x44", "\xB4\x71", "\xDD\x65",
  /* U+60F8 */	"\xD9\x46", "\xB7\x53", "\xB4\x69", "\xB4\x6C",
  /* U+60FC */	"\xD9\x47", "\xA0\x5B", "\xD9\x48", "\xD9\x4E",
  /* U+6100 */	"\xB4\x73", "\xB7\x54", "\x00\x00", "\xD9\x4A",
  /* U+6104 */	"\xD9\x4F", "\xD9\x43", "\xB7\x5E", "\x96\xAC",
  /* U+6108 */	"\xB7\x55", "\xB4\x72", "\xD9\x41", "\xD9\x50",
  /* U+610C */	"\x97\x40", "\xB7\x5D", "\xB4\x70", "\xB7\x4E",
  /* U+6110 */	"\xD9\x4D", "\x00\x00", "\xB4\x74", "\xD9\x45",
  /* U+6114 */	"\xD8\xFE", "\xB4\x6A", "\xD9\x42", "\x00\x00",
  /* U+6118 */	"\xD9\x4B", "\x9E\xF1", "\xB7\x4D", "\xB7\x52",
  /* U+611C */	"\xB4\x67", "\xD9\x4C", "\x00\x00", "\xB7\x50",
  /* U+6120 */	"\x00\x00", "\x00\x00", "\x8C\x4D", "\xB4\x68",
  /* U+6124 */	"\x00\x00", "\x00\x00", "\x00\x00", "\xB7\x5C",
  /* U+6128 */	"\xE1\xC3", "\xDD\x70", "\x00\x00", "\xDD\x68",
  /* U+612C */	"\xE1\xC2", "\x00\x00", "\xDD\x6C", "\xDD\x6E",
  /* U+6130 */	"\x9F\x7E", "\x00\x00", "\xDD\x6B", "\x00\x00",
  /* U+6134 */	"\xB7\x5B", "\x00\x00", "\xDD\x6A", "\xB7\x5F",
  /* U+6138 */	"\x00\x00", "\x00\x00", "\x00\x00", "\xE1\xD2",
  /* U+613C */	"\x00\x00", "\x8D\x72", "\xB7\x5A", "\xBA\x40",
  /* U+6140 */	"\xDD\x71", "\xE1\xC4", "\xFC\x76", "\x00\x00",
  /* U+6144 */	"\xB7\x58", "\xDD\x69", "\xDD\x6D", "\xB9\xFE",
  /* U+6148 */	"\xB7\x4F", "\xDD\x66", "\xDD\x67", "\xBA\x41",
  /* U+614C */	"\xB7\x57", "\xB7\x59", "\xB7\x56", "\xDD\x6F",
  /* U+6150 */	"\x96\xA9", "\x00\x00", "\xE1\xC8", "\xE1\xC9",
  /* U+6154 */	"\xE1\xCE", "\xBC\x7D", "\xE1\xD5", "\x00\x00",
  /* U+6158 */	"\xBA\x47", "\xA0\x6E", "\xBA\x46", "\xE1\xD0",
  /* U+615C */	"\xFC\xAA", "\xBC\x7C", "\xE1\xC5", "\xBA\x45",
  /* U+6160 */	"\xFB\xCD", "\xE1\xD4", "\xBA\x43", "\xBA\x44",
  /* U+6164 */	"\xFC\x74", "\xE1\xD1", "\xE5\xAA", "\xBC\x7A",
  /* U+6168 */	"\xB4\x6E", "\x00\x00", "\xE1\xD3", "\xBC\xA3",
  /* U+616C */	"\xE1\xCB", "\x00\x00", "\xBC\x7B", "\xA0\x74",
  /* U+6170 */	"\xBC\xA2", "\xE1\xC6", "\xE1\xCA", "\xE1\xC7",
  /* U+6174 */	"\xE1\xCD", "\xBA\x48", "\xBC\x79", "\xBA\x42",
  /* U+6178 */	"\x00\x00", "\xE5\x7A", "\xE1\xCF", "\x00\x00",
  /* U+617C */	"\xBC\xA1", "\xA0\x71", "\xBC\xA4", "\x00\x00",
  /* U+6180 */	"\xE1\xCC", "\xFC\x79", "\xBC\x7E", "\xE5\x79",
  /* U+6184 */	"\x00\x00", "\x00\x00", "\x00\x00", "\xFC\x7C",
  /* U+6188 */	"\x00\x00", "\xE5\x7E", "\xBE\xCE", "\xE5\x78",
  /* U+618C */	"\xE9\xA3", "\xE5\xA9", "\xBC\xA8", "\x00\x00",
  /* U+6190 */	"\xBC\xA6", "\xBE\xCC", "\xE5\xA6", "\xE5\xA2",
  /* U+6194 */	"\xBC\xAC", "\x9C\x50", "\xE9\x78", "\x00\x00",
  /* U+6198 */	"\x93\x79", "\x93\x78", "\xBC\xAA", "\xE5\xA1",
  /* U+619C */	"\xA0\xDD", "\xE9\x76", "\x00\x00", "\xE5\xA5",
  /* U+61A0 */	"\x00\x00", "\xE5\xA8", "\xE5\x7D", "\x00\x00",
  /* U+61A4 */	"\xBC\xAB", "\x00\x00", "\x00\x00", "\xBC\xA5",
  /* U+61A8 */	"\xE9\x77", "\xBE\xCD", "\xE5\xA7", "\xBC\xA7",
  /* U+61AC */	"\xBC\xA9", "\xE5\xA4", "\xBC\xAD", "\xE5\xA3",
  /* U+61B0 */	"\xE5\x7C", "\xE5\x7B", "\xBE\xCB", "\xE5\xAB",
  /* U+61B4 */	"\xE9\x7A", "\xEC\xE0", "\xBE\xD0", "\x8D\x75",
  /* U+61B8 */	"\xE9\xA2", "\x8D\x76", "\xE9\x7E", "\x00\x00",
  /* U+61BC */	"\xEC\xE1", "\x00\x00", "\xBE\xD1", "\xE9\xA1",
  /* U+61C0 */	"\x93\x74", "\xE9\x7C", "\xC0\xB4", "\xEC\xDF",
  /* U+61C4 */	"\x00\x00", "\xE9\x79", "\xE9\x7B", "\xC0\xB5",
  /* U+61C8 */	"\xBE\xD3", "\xC0\xB3", "\xBE\xD2", "\xC0\xB7",
  /* U+61CC */	"\xE9\x7D", "\xBE\xCF", "\x00\x00", "\x8D\x77",
  /* U+61D0 */	"\xFC\xA5", "\x00\x00", "\x00\x00", "\xFC\xA2",
  /* U+61D4 */	"\x00\x00", "\x00\x00", "\xEF\xCF", "\x00\x00",
  /* U+61D8 */	"\xEF\xC7", "\x00\x00", "\x90\xC3", "\x00\x00",
  /* U+61DC */	"\x00\x00", "\x00\x00", "\xEC\xE7", "\xEF\xC8",
  /* U+61E0 */	"\xEC\xE3", "\x00\x00", "\xA0\x79", "\xC2\x56",
  /* U+61E4 */	"\xEC\xE5", "\xEC\xE4", "\xC0\xB6", "\xEC\xE2",
  /* U+61E8 */	"\xEC\xE6", "\xEF\xD0", "\xEF\xCC", "\xEF\xCE",
  /* U+61EC */	"\x00\x00", "\xEF\xC9", "\xEF\xCA", "\x00\x00",
  /* U+61F0 */	"\xEF\xCD", "\xEF\xCB", "\xC3\x67", "\x00\x00",
  /* U+61F4 */	"\x00\x00", "\xC3\x6A", "\xC3\x69", "\xC3\x68",
  /* U+61F8 */	"\xC4\x61", "\xF4\x4A", "\xC4\x62", "\xF2\x41",
  /* U+61FC */	"\xC4\xDF", "\xF5\xCC", "\xC4\xE0", "\xC5\x74",
  /* U+6200 */	"\xC5\xCA", "\xF7\xD9", "\x00\x00", "\xF7\xDA",
  /* U+6204 */	"\xF7\xDB", "\x00\x00", "\x00\x00", "\xF9\xBA",
  /* U+6208 */	"\xA4\xE0", "\xC9\x7C", "\xA5\xB3", "\x00\x00",
  /* U+620C */	"\xA6\xA6", "\xA6\xA7", "\xA6\xA5", "\x00\x00",
  /* U+6210 */	"\xA6\xA8", "\xA7\xDA", "\xA7\xD9", "\x00\x00",
  /* U+6214 */	"\xCC\xB1", "\xA9\xCF", "\xA9\xCE", "\x00\x00",
  /* U+6218 */	"\x00\x00", "\xD1\xAF", "\xB1\xAD", "\xB1\xAE",
  /* U+621C */	"\x00\x00", "\x00\x00", "\x00\x00", "\xB4\x75",
  /* U+6220 */	"\xDD\x72", "\xB7\x60", "\xB7\x61", "\xDD\x74",
  /* U+6224 */	"\xDD\x76", "\xDD\x75", "\x00\x00", "\xE1\xD7",
  /* U+6228 */	"\x00\x00", "\xE1\xD6", "\xBA\x49", "\xE1\xD8",
  /* U+622C */	"\x8D\x79", "\xE5\xAC", "\xBC\xAE", "\x00\x00",
  /* U+6230 */	"\xBE\xD4", "\x00\x00", "\xC0\xB8", "\xC2\x57",
  /* U+6234 */	"\xC0\xB9", "\x00\x00", "\xA4\xE1", "\x8B\xFC",
  /* U+6238 */	"\x00\x00", "\xA0\x76", "\xCA\xE6", "\x00\x00",
  /* U+623C */	"\x00\x00", "\xCC\xB2", "\xA9\xD1", "\xA9\xD0",
  /* U+6240 */	"\xA9\xD2", "\xAB\xF3", "\xCE\xD2", "\xCE\xD3",
  /* U+6244 */	"\x00\x00", "\x00\x00", "\xD1\xB0", "\xAE\xB0",
  /* U+6248 */	"\xB1\xAF", "\xB4\x76", "\xD9\x51", "\xA4\xE2",
  /* U+624C */	"\x8B\xCD", "\xA4\x7E", "\xA4\xE3", "\x00\x00",
  /* U+6250 */	"\xC9\x7D", "\xA5\xB7", "\xA5\xB6", "\xA5\xB4",
  /* U+6254 */	"\xA5\xB5", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+6258 */	"\xA6\xAB", "\xC9\xE9", "\xC9\xEB", "\xA6\xAA",
  /* U+625C */	"\xC9\xE3", "\x00\x00", "\xC9\xE4", "\x00\x00",
  /* U+6260 */	"\xC9\xEA", "\xC9\xE6", "\xC9\xE8", "\xA6\xA9",
  /* U+6264 */	"\xC9\xE5", "\xC9\xEC", "\xC9\xE7", "\x00\x00",
  /* U+6268 */	"\x9F\x5A", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+626C */	"\x00\x00", "\xA7\xE1", "\xA7\xEA", "\xA7\xE8",
  /* U+6270 */	"\xCA\xF0", "\xCA\xED", "\xCA\xF5", "\xA7\xE6",
  /* U+6274 */	"\xCA\xF6", "\x00\x00", "\xA7\xDF", "\xCA\xF3",
  /* U+6278 */	"\x00\x00", "\xA7\xE5", "\xCA\xEF", "\xCA\xEE",
  /* U+627C */	"\xA7\xE3", "\xCA\xF4", "\xA7\xE4", "\xA9\xD3",
  /* U+6280 */	"\xA7\xDE", "\xCA\xF1", "\x9F\xF4", "\xCA\xE7",
  /* U+6284 */	"\xA7\xDB", "\x9F\xBA", "\xA7\xEE", "\xCA\xEC",
  /* U+6288 */	"\xCA\xF2", "\xA7\xE0", "\xA7\xE2", "\x00\x00",
  /* U+628C */	"\xCA\xE8", "\x00\x00", "\xCA\xE9", "\xCA\xEA",
  /* U+6290 */	"\x8D\x7A", "\xA7\xED", "\xA7\xE7", "\xA7\xEC",
  /* U+6294 */	"\xCA\xEB", "\xA7\xEB", "\xA7\xDD", "\xA7\xDC",
  /* U+6298 */	"\xA7\xE9", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+629C */	"\x00\x00", "\x9E\x45", "\x00\x00", "\x00\x00",
  /* U+62A0 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+62A4 */	"\x93\xB0", "\x00\x00", "\xA0\x75", "\x00\x00",
  /* U+62A8 */	"\xA9\xE1", "\xCC\xBE", "\xCC\xB7", "\xA9\xDC",
  /* U+62AC */	"\xA9\xEF", "\xCC\xB3", "\xCC\xBA", "\xCC\xBC",
  /* U+62B0 */	"\xCC\xBF", "\xA9\xEA", "\x00\x00", "\xCC\xBB",
  /* U+62B4 */	"\xCC\xB4", "\xA9\xE8", "\xCC\xB8", "\x00\x00",
  /* U+62B8 */	"\xCC\xC0", "\xA9\xD9", "\x00\x00", "\xCC\xBD",
  /* U+62BC */	"\xA9\xE3", "\xA9\xE2", "\xCC\xB6", "\xA9\xD7",
  /* U+62C0 */	"\x00\x00", "\x00\x00", "\xA9\xD8", "\x9B\x46",
  /* U+62C4 */	"\xA9\xD6", "\xFC\xAE", "\xA9\xEE", "\xA9\xE6",
  /* U+62C8 */	"\xA9\xE0", "\xA9\xD4", "\xCC\xB9", "\xA9\xDF",
  /* U+62CC */	"\xA9\xD5", "\xA9\xE7", "\xA9\xF0", "\xCE\xD4",
  /* U+62D0 */	"\xA9\xE4", "\xCC\xB5", "\xA9\xDA", "\xA9\xDD",
  /* U+62D4 */	"\xA9\xDE", "\xFC\xB0", "\xA9\xEC", "\xA9\xED",
  /* U+62D8 */	"\xA9\xEB", "\xA9\xE5", "\xA9\xE9", "\xA9\xDB",
  /* U+62DC */	"\xAB\xF4", "\x00\x00", "\x00\x00", "\xFA\x51",
  /* U+62E0 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+62E4 */	"\x00\x00", "\x8D\x7B", "\x00\x00", "\x00\x00",
  /* U+62E8 */	"\x00\x00", "\x00\x00", "\x00\x00", "\xCE\xDA",
  /* U+62EC */	"\xAC\x41", "\xAB\xF8", "\xAB\xFA", "\xAC\x40",
  /* U+62F0 */	"\xCE\xE6", "\xAB\xFD", "\xD1\xB1", "\xAE\xB1",
  /* U+62F4 */	"\xAC\x43", "\xCE\xD7", "\xCE\xDF", "\xAB\xFE",
  /* U+62F8 */	"\xCE\xDE", "\xCE\xDB", "\xCE\xE3", "\xCE\xE5",
  /* U+62FC */	"\xAB\xF7", "\xAB\xFB", "\xAC\x42", "\xAE\xB3",
  /* U+6300 */	"\xCE\xE0", "\xAB\xF9", "\xAC\x45", "\xCE\xD9",
  /* U+6304 */	"\x00\x00", "\x00\x00", "\x00\x00", "\xAB\xFC",
  /* U+6308 */	"\xAE\xB2", "\xAB\xF6", "\x00\x00", "\xCE\xD6",
  /* U+630C */	"\xCE\xDD", "\xCE\xD5", "\xCE\xD8", "\xCE\xDC",
  /* U+6310 */	"\xD1\xB2", "\xAC\x44", "\x00\x00", "\xCE\xE1",
  /* U+6314 */	"\xCE\xE2", "\xCE\xE4", "\xAB\xF5", "\x00\x00",
  /* U+6318 */	"\x8D\x7C", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+631C */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+6320 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+6324 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+6328 */	"\xAE\xC1", "\xD1\xBE", "\xAE\xBF", "\xAE\xC0",
  /* U+632C */	"\xD1\xB4", "\xD1\xC4", "\x9E\xD6", "\xAE\xB6",
  /* U+6330 */	"\x00\x00", "\x93\xAC", "\xD5\x66", "\xD1\xC6",
  /* U+6334 */	"\xD1\xC0", "\x9F\x5B", "\xD1\xB7", "\x93\xA9",
  /* U+6338 */	"\xD1\xC9", "\xD1\xBA", "\xAE\xBC", "\xD5\x7D",
  /* U+633C */	"\xD1\xBD", "\xAE\xBE", "\xAE\xB5", "\x00\x00",
  /* U+6340 */	"\xD1\xCB", "\xD1\xBF", "\xAE\xB8", "\xD1\xB8",
  /* U+6344 */	"\xD1\xB5", "\xD1\xB6", "\xAE\xB9", "\xD1\xC5",
  /* U+6348 */	"\xD1\xCC", "\xAE\xBB", "\xD1\xBC", "\xD1\xBB",
  /* U+634C */	"\xAE\xC3", "\xAE\xC2", "\xAE\xB4", "\xAE\xBA",
  /* U+6350 */	"\xAE\xBD", "\xD1\xC8", "\x00\x00", "\x00\x00",
  /* U+6354 */	"\xD1\xC2", "\xAE\xB7", "\xD1\xB3", "\xD1\xCA",
  /* U+6358 */	"\xD1\xC1", "\xD1\xC3", "\xD1\xC7", "\x00\x00",
  /* U+635C */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+6360 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+6364 */	"\xA0\x7C", "\xD5\x67", "\x00\x00", "\xB1\xB7",
  /* U+6368 */	"\xB1\xCB", "\xB1\xCA", "\x00\x00", "\xB1\xBF",
  /* U+636C */	"\xFC\xB2", "\xD5\x79", "\xD5\x75", "\xD5\x72",
  /* U+6370 */	"\xD5\xA6", "\xB1\xBA", "\xB1\xB2", "\x00\x00",
  /* U+6374 */	"\x00\x00", "\xD5\x77", "\xB4\xA8", "\xB1\xB6",
  /* U+6378 */	"\xD5\xA1", "\x8A\xC1", "\xB1\xCC", "\xB1\xC9",
  /* U+637C */	"\xD5\x7B", "\xD5\x6A", "\x00\x00", "\x9F\xB4",
  /* U+6380 */	"\xB1\xC8", "\xD5\xA3", "\xD5\x69", "\xB1\xBD",
  /* U+6384 */	"\xB1\xC1", "\xD5\xA2", "\x00\x00", "\xD5\x73",
  /* U+6388 */	"\xB1\xC2", "\xB1\xBC", "\xD5\x68", "\xFC\xAC",
  /* U+638C */	"\xB4\x78", "\xD5\xA5", "\xD5\x71", "\xB1\xC7",
  /* U+6390 */	"\xD5\x74", "\xD5\xA4", "\xB1\xC6", "\x00\x00",
  /* U+6394 */	"\xD9\x52", "\x00\x00", "\xB1\xB3", "\xD5\x6F",
  /* U+6398 */	"\xB1\xB8", "\xB1\xC3", "\x00\x00", "\xB1\xBE",
  /* U+639C */	"\xD5\x78", "\xD5\x6E", "\xD5\x6C", "\xD5\x7E",
  /* U+63A0 */	"\xB1\xB0", "\xB1\xC4", "\xB1\xB4", "\xB4\x77",
  /* U+63A4 */	"\xD5\x7C", "\xB1\xB5", "\x00\x00", "\xB1\xB1",
  /* U+63A8 */	"\xB1\xC0", "\xB1\xBB", "\xB1\xB9", "\xD5\x70",
  /* U+63AC */	"\xB1\xC5", "\xD5\x6D", "\xD5\x7A", "\xD5\x76",
  /* U+63B0 */	"\xD9\x54", "\xD9\x53", "\x00\x00", "\x00\x00",
  /* U+63B4 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+63B8 */	"\x00\x00", "\x9E\x4C", "\x00\x00", "\x00\x00",
  /* U+63BC */	"\x00\x00", "\xD5\x6B", "\xD9\x64", "\x00\x00",
  /* U+63C0 */	"\xB4\x7A", "\x8F\xC5", "\xD9\x6A", "\xD9\x59",
  /* U+63C4 */	"\xD9\x67", "\xDD\x77", "\xB4\x7D", "\xD9\x6B",
  /* U+63C8 */	"\xD9\x6E", "\xB4\x7C", "\xD9\x5C", "\xD9\x6D",
  /* U+63CC */	"\xD9\x6C", "\xB4\x7E", "\xD9\x55", "\xB4\x79",
  /* U+63D0 */	"\xB4\xA3", "\x93\xAD", "\xB4\xA1", "\xD9\x69",
  /* U+63D4 */	"\x00\x00", "\xD9\x5F", "\xB4\xA5", "\xD9\x70",
  /* U+63D8 */	"\xD9\x68", "\xD9\x71", "\xB4\xAD", "\xB4\xAB",
  /* U+63DC */	"\xD9\x66", "\xD9\x65", "\x9D\xC3", "\xD9\x63",
  /* U+63E0 */	"\xD9\x5D", "\xB4\xA4", "\x8D\xA2", "\xB4\xA2",
  /* U+63E4 */	"\xD1\xB9", "\xD9\x56", "\x9D\x4A", "\xDD\xB7",
  /* U+63E8 */	"\xD9\x57", "\xB4\x7B", "\xB4\xAA", "\xDD\x79",
  /* U+63EC */	"\x00\x00", "\xB4\xA6", "\xB4\xA7", "\xD9\x58",
  /* U+63F0 */	"\xD9\x6F", "\xDD\x78", "\xD9\x60", "\xD9\x5B",
  /* U+63F4 */	"\xB4\xA9", "\xD9\x61", "\xD9\x5E", "\x00\x00",
  /* U+63F8 */	"\xFC\xB6", "\xB4\xAE", "\x00\x00", "\x8D\xA3",
  /* U+63FC */	"\x9E\x4B", "\x00\x00", "\x9E\x4D", "\x00\x00",
  /* U+6400 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+6404 */	"\x00\x00", "\x00\x00", "\xB7\x70", "\x8D\xA4",
  /* U+6408 */	"\x00\x00", "\xDD\x7C", "\xDD\xB1", "\xDD\xB6",
  /* U+640C */	"\xDD\xAA", "\xB7\x6C", "\xDD\xBB", "\xB7\x69",
  /* U+6410 */	"\xDD\x7A", "\x00\x00", "\xDD\x7B", "\xB7\x62",
  /* U+6414 */	"\xB7\x6B", "\xDD\xA4", "\xB7\x6E", "\xB7\x6F",
  /* U+6418 */	"\xDD\xA5", "\x00\x00", "\xDD\xB2", "\xDD\xB8",
  /* U+641C */	"\xB7\x6A", "\x00\x00", "\xB7\x64", "\xDD\xA3",
  /* U+6420 */	"\xDD\x7D", "\xDD\xBA", "\xDD\xA8", "\xDD\xA9",
  /* U+6424 */	"\xDD\x7E", "\xDD\xB4", "\xDD\xAB", "\xDD\xB5",
  /* U+6428 */	"\xDD\xAD", "\x00\x00", "\xB7\x65", "\xE1\xD9",
  /* U+642C */	"\xB7\x68", "\xB7\x66", "\xDD\xB9", "\xDD\xB0",
  /* U+6430 */	"\xDD\xAC", "\x00\x00", "\x8A\xFD", "\xDD\xA1",
  /* U+6434 */	"\xBA\x53", "\xDD\xAF", "\xB7\x6D", "\xDD\xA7",
  /* U+6438 */	"\xFC\xB5", "\xDD\xA6", "\xFC\xC3", "\x93\xB2",
  /* U+643C */	"\x00\x00", "\xB7\x67", "\xB7\x63", "\xE1\xEE",
  /* U+6440 */	"\xDD\xB3", "\xDD\xAE", "\x00\x00", "\xDD\xA2",
  /* U+6444 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+6448 */	"\x00\x00", "\x00\x00", "\x00\x00", "\xE1\xE9",
  /* U+644C */	"\x00\x00", "\xE1\xDA", "\xE1\xE5", "\x00\x00",
  /* U+6450 */	"\xE1\xEC", "\xBA\x51", "\xB4\xAC", "\xE1\xEA",
  /* U+6454 */	"\xBA\x4C", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+6458 */	"\xBA\x4B", "\xE1\xF1", "\x8D\xA5", "\xE1\xDB",
  /* U+645C */	"\xE1\xE8", "\xE1\xDC", "\xE1\xE7", "\xBA\x4F",
  /* U+6460 */	"\xE1\xEB", "\xD9\x62", "\x00\x00", "\x00\x00",
  /* U+6464 */	"\x00\x00", "\xE1\xF2", "\xE1\xE3", "\xBA\x52",
  /* U+6468 */	"\xE5\xBA", "\xBC\xAF", "\x00\x00", "\xE1\xF0",
  /* U+646C */	"\xE1\xEF", "\xBA\x54", "\xE5\xAD", "\xBC\xB0",
  /* U+6470 */	"\xE5\xAE", "\x93\xA1", "\xE1\xDF", "\xE1\xE0",
  /* U+6474 */	"\xE1\xDD", "\xE1\xE2", "\xE1\xDE", "\xE1\xF3",
  /* U+6478 */	"\xBA\x4E", "\xBC\xB1", "\xBA\x50", "\xBA\x55",
  /* U+647C */	"\x8A\xC6", "\xE1\xE1", "\x00\x00", "\xE1\xED",
  /* U+6480 */	"\x00\x00", "\x00\x00", "\xE1\xE6", "\x00\x00",
  /* U+6484 */	"\x00\x00", "\xE5\xB1", "\x00\x00", "\xBA\x4A",
  /* U+6488 */	"\xBC\xB4", "\xE9\xAA", "\xE5\xB6", "\xE5\xB5",
  /* U+648C */	"\xE5\xB7", "\x8A\x5B", "\x00\x00", "\xE5\xB4",
  /* U+6490 */	"\xBC\xB5", "\x89\x4D", "\xBC\xBB", "\xBC\xB8",
  /* U+6494 */	"\x00\x00", "\xBC\xB9", "\xE5\xAF", "\xE5\xB2",
  /* U+6498 */	"\xE5\xBC", "\xBC\xC1", "\xBC\xBF", "\x00\x00",
  /* U+649C */	"\xE5\xB3", "\xD9\x5A", "\xBC\xB2", "\xE5\xB9",
  /* U+64A0 */	"\xE5\xB0", "\x00\x00", "\xBC\xC2", "\xE5\xB8",
  /* U+64A4 */	"\xBA\x4D", "\xBC\xB7", "\xE1\xE4", "\x00\x00",
  /* U+64A8 */	"\x00\x00", "\xBC\xBA", "\x00\x00", "\xBC\xBE",
  /* U+64AC */	"\xBC\xC0", "\xBC\xBD", "\xBC\xBC", "\xFE\xD4",
  /* U+64B0 */	"\xBC\xB6", "\xE5\xBB", "\xBC\xB3", "\xBC\xC3",
  /* U+64B4 */	"\x8A\x78", "\x00\x00", "\x93\xAB", "\x00\x00",
  /* U+64B8 */	"\x00\x00", "\x00\x00", "\x00\x00", "\xBE\xD8",
  /* U+64BC */	"\xBE\xD9", "\xE9\xA9", "\xBE\xE2", "\xBE\xDF",
  /* U+64C0 */	"\x8D\xA7", "\xBE\xD6", "\xBE\xDD", "\xE9\xAB",
  /* U+64C4 */	"\xBE\xDB", "\xBE\xD5", "\x00\x00", "\xBE\xDC",
  /* U+64C8 */	"\x00\x00", "\xE9\xA8", "\xC0\xBB", "\xBE\xD7",
  /* U+64CC */	"\x00\x00", "\xBE\xDE", "\xC0\xBA", "\xE9\xA7",
  /* U+64D0 */	"\xE9\xA6", "\x00\x00", "\xBE\xE0", "\x9F\x45",
  /* U+64D4 */	"\xBE\xE1", "\x00\x00", "\xE9\xA5", "\xE9\xA4",
  /* U+64D8 */	"\xC0\xBC", "\xE9\xAE", "\xBE\xDA", "\xE9\xAC",
  /* U+64DC */	"\x00\x00", "\x8A\x56", "\x00\x00", "\x00\x00",
  /* U+64E0 */	"\xC0\xBD", "\xFC\xBF", "\xC0\xC2", "\xEC\xEA",
  /* U+64E4 */	"\xEC\xEC", "\xFC\xC0", "\xC0\xBF", "\x8E\xE6",
  /* U+64E8 */	"\xEC\xED", "\xEC\xE9", "\x8A\xA4", "\xEC\xEB",
  /* U+64EC */	"\xC0\xC0", "\xC0\xC3", "\x00\x00", "\xEC\xE8",
  /* U+64F0 */	"\xC0\xBE", "\xC0\xC1", "\xC2\x59", "\xE9\xAD",
  /* U+64F4 */	"\xC2\x58", "\x00\x00", "\x00\x00", "\xC2\x5E",
  /* U+64F8 */	"\xEF\xD4", "\x00\x00", "\xC2\x5C", "\xC2\x5D",
  /* U+64FC */	"\xEF\xD7", "\xEF\xD3", "\xC2\x5A", "\xEF\xD1",
  /* U+6500 */	"\xC3\x6B", "\xEF\xD5", "\x00\x00", "\xEF\xD6",
  /* U+6504 */	"\xEF\xD2", "\x00\x00", "\xC2\x5B", "\xF2\x42",
  /* U+6508 */	"\x00\x00", "\xF2\x45", "\x89\x43", "\x00\x00",
  /* U+650C */	"\xF2\x46", "\xF2\x44", "\xF2\x47", "\xC3\x6C",
  /* U+6510 */	"\xF2\x43", "\x93\xF3", "\x00\x00", "\xF4\x4E",
  /* U+6514 */	"\xC4\x64", "\xF4\x4D", "\xF4\x4C", "\xF4\x4B",
  /* U+6518 */	"\xC4\x63", "\xC4\x65", "\x00\x00", "\xF5\xCD",
  /* U+651C */	"\xC4\xE2", "\xC4\xE1", "\xFC\xAB", "\x9E\xA2",
  /* U+6520 */	"\xF6\xE1", "\xF6\xE0", "\xF6\xE3", "\xC5\xCB",
  /* U+6524 */	"\xC5\x75", "\xF7\xDD", "\xF6\xE2", "\x00\x00",
  /* U+6528 */	"\x00\x00", "\xF7\xDC", "\xC5\xCD", "\xC5\xCC",
  /* U+652C */	"\xC5\xF3", "\xF8\xA9", "\xF8\xEF", "\xA4\xE4",
  /* U+6530 */	"\x9D\xC7", "\x00\x00", "\xD9\x72", "\xE9\xAF",
  /* U+6534 */	"\xC6\xD2", "\x8B\xCE", "\xA6\xAC", "\xCA\xF7",
  /* U+6538 */	"\xA7\xF1", "\xA7\xEF", "\x00\x00", "\xA7\xF0",
  /* U+653C */	"\x00\x00", "\xCC\xC1", "\xA9\xF1", "\xAC\x46",
  /* U+6540 */	"\x00\x00", "\xCE\xE7", "\x00\x00", "\xCE\xE8",
  /* U+6544 */	"\x00\x00", "\xAC\x47", "\xD1\xCE", "\x00\x00",
  /* U+6548 */	"\xAE\xC4", "\xAE\xC5", "\xD1\xCD", "\x00\x00",
  /* U+654C */	"\x00\x00", "\xFC\xC5", "\x00\x00", "\xB1\xD3",
  /* U+6550 */	"\x00\x00", "\xB1\xCF", "\x00\x00", "\xD5\xA7",
  /* U+6554 */	"\xB1\xD6", "\xB1\xD5", "\xB1\xCE", "\xB1\xD1",
  /* U+6558 */	"\xB1\xD4", "\xB1\xD0", "\x00\x00", "\x00\x00",
  /* U+655C */	"\xD9\x76", "\xB1\xCD", "\xB4\xAF", "\xFC\xCB",
  /* U+6560 */	"\x00\x00", "\x00\x00", "\xB4\xB1", "\xB4\xB2",
  /* U+6564 */	"\xD9\x75", "\xD9\x78", "\xB4\xB0", "\xD9\x73",
  /* U+6568 */	"\xD9\x77", "\x00\x00", "\xD9\x74", "\x93\xB3",
  /* U+656C */	"\xB7\x71", "\xFC\xCA", "\x00\x00", "\xDD\xBC",
  /* U+6570 */	"\x00\x00", "\x00\x00", "\xBA\x56", "\xE1\xF4",
  /* U+6574 */	"\xBE\xE3", "\xBC\xC4", "\xE5\xBD", "\xBC\xC5",
  /* U+6578 */	"\xBC\xC6", "\xE5\xBF", "\xE5\xBE", "\xE5\xC0",
  /* U+657C */	"\xE9\xB1", "\x00\x00", "\x00\x00", "\xE9\xB0",
  /* U+6580 */	"\xEC\xEF", "\xEC\xEE", "\xC0\xC4", "\xC0\xC5",
  /* U+6584 */	"\xF2\x48", "\xFC\xC9", "\x8D\xAC", "\xA4\xE5",
  /* U+6588 */	"\xFB\xC6", "\x89\x67", "\x00\x00", "\x8C\x7E",
  /* U+658C */	"\xD9\x79", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+6590 */	"\xB4\xB4", "\xB4\xB3", "\xDD\xBD", "\x00\x00",
  /* U+6594 */	"\xEF\xD8", "\xC4\xE3", "\xF7\xDE", "\xA4\xE6",
  /* U+6598 */	"\x00\x00", "\xAE\xC6", "\x00\x00", "\xB1\xD8",
  /* U+659C */	"\xB1\xD7", "\xD9\x7A", "\xD9\x7B", "\xB7\x72",
  /* U+65A0 */	"\xE1\xF5", "\xBA\x57", "\xE9\xB2", "\x00\x00",
  /* U+65A4 */	"\xA4\xE7", "\xA5\xB8", "\x00\x00", "\xA9\xF2",
  /* U+65A8 */	"\xCC\xC2", "\x00\x00", "\xCE\xE9", "\xAC\x48",
  /* U+65AC */	"\xB1\xD9", "\x00\x00", "\xD9\x7C", "\xB4\xB5",
  /* U+65B0 */	"\xB7\x73", "\x00\x00", "\xE5\xC1", "\xE5\xC2",
  /* U+65B4 */	"\x00\x00", "\xFC\xCD", "\xEC\xF0", "\xC2\x5F",
  /* U+65B8 */	"\xF8\xF0", "\xA4\xE8", "\x00\x00", "\xCC\xC3",
  /* U+65BC */	"\xA9\xF3", "\xAC\x49", "\x9C\xF3", "\xCE\xEA",
  /* U+65C0 */	"\x00\x00", "\xAE\xC7", "\xD1\xD2", "\xD1\xD0",
  /* U+65C4 */	"\xD1\xD1", "\xAE\xC8", "\xD1\xCF", "\x00\x00",
  /* U+65C8 */	"\x00\x00", "\x00\x00", "\x00\x00", "\xB1\xDB",
  /* U+65CC */	"\xB1\xDC", "\xD5\xA8", "\xB1\xDD", "\xB1\xDA",
  /* U+65D0 */	"\xD9\x7D", "\xFC\xD0", "\xD9\x7E", "\xDD\xBE",
  /* U+65D4 */	"\x95\xBB", "\x00\x00", "\xBA\x59", "\xBA\x58",
  /* U+65D8 */	"\x00\x00", "\x00\x00", "\xEC\xF1", "\xEF\xD9",
  /* U+65DC */	"\x00\x00", "\xF2\x4A", "\xF2\x49", "\xF4\x4F",
  /* U+65E0 */	"\xFC\xD3", "\xC9\x5E", "\xAC\x4A", "\xFC\xD4",
  /* U+65E4 */	"\x00\x00", "\xA4\xE9", "\xA5\xB9", "\x00\x00",
  /* U+65E8 */	"\xA6\xAE", "\xA6\xAD", "\x00\x00", "\x00\x00",
  /* U+65EC */	"\xA6\xAF", "\xA6\xB0", "\xC9\xEE", "\xC9\xED",
  /* U+65F0 */	"\xCA\xF8", "\xA7\xF2", "\xCA\xFB", "\xCA\xFA",
  /* U+65F4 */	"\xCA\xF9", "\xCA\xFC", "\x00\x00", "\x00\x00",
  /* U+65F8 */	"\x00\x00", "\x00\x00", "\xA9\xF4", "\xCC\xC9",
  /* U+65FC */	"\xCC\xC5", "\xCC\xCE", "\x00\x00", "\x8D\xAE",
  /* U+6600 */	"\xA9\xFB", "\x00\x00", "\xA9\xF9", "\xCC\xCA",
  /* U+6604 */	"\xCC\xC6", "\xCC\xCD", "\xA9\xF8", "\xAA\x40",
  /* U+6608 */	"\xCC\xC8", "\xCC\xC4", "\xA9\xFE", "\xCC\xCB",
  /* U+660C */	"\xA9\xF7", "\xCC\xCC", "\xA9\xFA", "\xA9\xFC",
  /* U+6610 */	"\xCC\xD0", "\xCC\xCF", "\xCC\xC7", "\xA9\xF6",
  /* U+6614 */	"\xA9\xF5", "\xA9\xFD", "\x00\x00", "\x00\x00",
  /* U+6618 */	"\xFC\xD7", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+661C */	"\xCE\xEF", "\xCE\xF5", "\x93\xDB", "\xAC\x50",
  /* U+6620 */	"\xAC\x4D", "\xCE\xEC", "\xCE\xF1", "\xFE\x63",
  /* U+6624 */	"\xAC\x53", "\xAC\x4B", "\xCE\xF0", "\xAC\x4E",
  /* U+6628 */	"\xAC\x51", "\x00\x00", "\x00\x00", "\xCE\xF3",
  /* U+662C */	"\x00\x00", "\xAC\x4C", "\xCE\xF8", "\xAC\x4F",
  /* U+6630 */	"\x93\xD5", "\xAC\x52", "\xCE\xED", "\xCE\xF2",
  /* U+6634 */	"\xCE\xF6", "\xCE\xEE", "\xCE\xEB", "\x00\x00",
  /* U+6638 */	"\x00\x00", "\xCE\xF7", "\xCE\xF4", "\x00\x00",
  /* U+663C */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+6640 */	"\x00\x00", "\xAE\xD0", "\xAE\xC9", "\xAE\xCC",
  /* U+6644 */	"\xFC\xDA", "\xAE\xCF", "\x00\x00", "\xD1\xD5",
  /* U+6648 */	"\x9B\x71", "\xAE\xCA", "\xD1\xD3", "\xFC\xDD",
  /* U+664C */	"\xAE\xCE", "\x00\x00", "\x00\x00", "\xAE\xCB",
  /* U+6650 */	"\x00\x00", "\xD1\xD6", "\xAE\xCD", "\x8D\xAF",
  /* U+6654 */	"\x00\x00", "\x00\x00", "\x00\x00", "\xFA\xF2",
  /* U+6658 */	"\x00\x00", "\xD5\xAC", "\xB1\xDF", "\xD5\xAB",
  /* U+665C */	"\xD5\xAD", "\xB1\xDE", "\xB1\xE3", "\xD1\xD4",
  /* U+6660 */	"\x00\x00", "\xD5\xAA", "\xD5\xAE", "\x93\xD8",
  /* U+6664 */	"\xB1\xE0", "\xD5\xA9", "\xB1\xE2", "\xFC\xDF",
  /* U+6668 */	"\xB1\xE1", "\x00\x00", "\xD9\xA7", "\x93\xD3",
  /* U+666C */	"\xD9\xA2", "\x00\x00", "\xB4\xB6", "\xB4\xBA",
  /* U+6670 */	"\xB4\xB7", "\xD9\xA5", "\xD9\xA8", "\xFC\xE1",
  /* U+6674 */	"\xB4\xB8", "\x00\x00", "\xB4\xB9", "\xB4\xBE",
  /* U+6678 */	"\xDD\xC7", "\xD9\xA6", "\xB4\xBC", "\xD9\xA3",
  /* U+667C */	"\xD9\xA1", "\x8E\x76", "\xB4\xBD", "\x00\x00",
  /* U+6680 */	"\xD9\xA4", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+6684 */	"\xB7\x79", "\xFC\x62", "\xDD\xBF", "\xB7\x76",
  /* U+6688 */	"\xB7\x77", "\xB7\x75", "\xDD\xC4", "\xDD\xC3",
  /* U+668C */	"\xDD\xC0", "\xB7\x7B", "\x93\xD1", "\x00\x00",
  /* U+6690 */	"\xDD\xC2", "\xB4\xBB", "\x8D\xB1", "\x00\x00",
  /* U+6694 */	"\xDD\xC6", "\xDD\xC1", "\xB7\x78", "\xB7\x74",
  /* U+6698 */	"\xB7\x7A", "\xDD\xC5", "\x98\x59", "\x00\x00",
  /* U+669C */	"\x00\x00", "\xBA\x5C", "\x00\x00", "\xE1\xF8",
  /* U+66A0 */	"\xE1\xF7", "\xE1\xF6", "\xBA\x5A", "\x00\x00",
  /* U+66A4 */	"\xFB\x52", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+66A8 */	"\xBA\x5B", "\xE5\xC5", "\xE5\xC8", "\xBC\xC8",
  /* U+66AC */	"\x00\x00", "\xFB\x53", "\xBC\xC7", "\xE5\xC9",
  /* U+66B0 */	"\xE5\xC4", "\xBC\xCA", "\xE5\xC6", "\xFB\x4D",
  /* U+66B4 */	"\xBC\xC9", "\xE5\xC3", "\x9C\xBF", "\xE5\xC7",
  /* U+66B8 */	"\xBE\xE9", "\xBE\xE6", "\xE9\xBB", "\xE9\xBA",
  /* U+66BC */	"\x00\x00", "\xE9\xB9", "\xE9\xB4", "\x9B\x72",
  /* U+66C0 */	"\xE9\xB5", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+66C4 */	"\xBE\xE7", "\x00\x00", "\xBE\xE4", "\xBE\xE8",
  /* U+66C8 */	"\xE9\xB3", "\xBE\xE5", "\xE9\xB6", "\xE9\xB7",
  /* U+66CC */	"\xE9\xBC", "\xFB\x50", "\x93\xBE", "\xE9\xB8",
  /* U+66D0 */	"\x00\x00", "\x00\x00", "\xEC\xF2", "\x00\x00",
  /* U+66D4 */	"\x00\x00", "\x00\x00", "\xC0\xC7", "\x00\x00",
  /* U+66D8 */	"\xEF\xDC", "\xC0\xC6", "\xEF\xDA", "\xEF\xDB",
  /* U+66DC */	"\xC2\x60", "\xC3\x6E", "\xF2\x4B", "\x00\x00",
  /* U+66E0 */	"\xC3\x6D", "\x00\x00", "\x00\x00", "\xF4\x51",
  /* U+66E4 */	"\xF4\x52", "\x00\x00", "\xC4\x66", "\x00\x00",
  /* U+66E8 */	"\xF4\x50", "\xC4\xE4", "\x00\x00", "\xF7\xDF",
  /* U+66EC */	"\xC5\xCE", "\xF8\xAA", "\xF8\xAB", "\x00\x00",
  /* U+66F0 */	"\xA4\xEA", "\x9D\xF1", "\xA6\xB1", "\xA6\xB2",
  /* U+66F4 */	"\xA7\xF3", "\x00\x00", "\xCC\xD1", "\xAC\x54",
  /* U+66F8 */	"\xAE\xD1", "\xB1\xE4", "\x00\x00", "\x00\x00",
  /* U+66FC */	"\xB0\xD2", "\x00\x00", "\xB4\xBF", "\xB4\xC0",
  /* U+6700 */	"\xB3\xCC", "\xD9\xA9", "\xFC\xEB", "\xB7\x7C",
  /* U+6704 */	"\xE1\xFA", "\xE1\xF9", "\x00\x00", "\x00\x00",
  /* U+6708 */	"\xA4\xEB", "\xA6\xB3", "\xCC\xD2", "\xAA\x42",
  /* U+670C */	"\xA0\xBB", "\xAA\x41", "\x9B\x7E", "\xCE\xF9",
  /* U+6710 */	"\xCE\xFA", "\x00\x00", "\xD1\xD7", "\xD1\xD8",
  /* U+6714 */	"\xAE\xD2", "\xAE\xD3", "\x8D\xB3", "\xAE\xD4",
  /* U+6718 */	"\xD5\xAF", "\x8C\x52", "\x00\x00", "\xB1\xE6",
  /* U+671C */	"\x00\x00", "\xB4\xC2", "\x9A\xE8", "\xB4\xC1",
  /* U+6720 */	"\xDD\xC8", "\xDF\x7A", "\xE1\xFB", "\xE9\xBD",
  /* U+6724 */	"\x00\x00", "\x8E\xDC", "\xC2\x61", "\xC4\x67",
  /* U+6728 */	"\xA4\xEC", "\x00\x00", "\xA5\xBC", "\xA5\xBD",
  /* U+672C */	"\xA5\xBB", "\xA5\xBE", "\xA5\xBA", "\x00\x00",
  /* U+6730 */	"\x00\x00", "\xA6\xB6", "\x00\x00", "\xC9\xF6",
  /* U+6734 */	"\xA6\xB5", "\xA6\xB7", "\x9C\xF9", "\x00\x00",
  /* U+6738 */	"\xC9\xF1", "\xC9\xF0", "\xC9\xF3", "\xC9\xF2",
  /* U+673C */	"\xC9\xF5", "\xA6\xB4", "\xC9\xEF", "\xC9\xF4",
  /* U+6740 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+6744 */	"\xFA\x50", "\xCA\xFD", "\xA7\xFD", "\xCA\xFE",
  /* U+6748 */	"\xCB\x43", "\xA7\xFC", "\x00\x00", "\xCB\x47",
  /* U+674C */	"\xCB\x42", "\xCB\x45", "\xA7\xF5", "\xA7\xF6",
  /* U+6750 */	"\xA7\xF7", "\xA7\xF8", "\x00\x00", "\xA8\x40",
  /* U+6754 */	"\x00\x00", "\xCB\x41", "\xA7\xFA", "\xA8\x41",
  /* U+6758 */	"\x00\x00", "\xCB\x40", "\xCB\x46", "\x00\x00",
  /* U+675C */	"\xA7\xF9", "\xCB\x44", "\xA7\xFB", "\xA7\xF4",
  /* U+6760 */	"\xA7\xFE", "\x98\xE7", "\xFC\xF3", "\x00\x00",
  /* U+6764 */	"\x00\x00", "\x00\x00", "\x00\x00", "\xFC\xF2",
  /* U+6768 */	"\x00\x00", "\x00\x00", "\xAA\x57", "\x8C\xCA",
  /* U+676C */	"\xCC\xD4", "\xAA\x43", "\x00\x00", "\xAA\x4D",
  /* U+6770 */	"\xAA\x4E", "\xAA\x46", "\xAA\x58", "\xAA\x48",
  /* U+6774 */	"\xCC\xDC", "\xAA\x53", "\xCC\xD7", "\xAA\x49",
  /* U+6778 */	"\xCC\xE6", "\xCC\xE7", "\xCC\xDF", "\xCC\xD8",
  /* U+677C */	"\xAA\x56", "\xCC\xE4", "\xAA\x51", "\xAA\x4F",
  /* U+6780 */	"\x00\x00", "\xCC\xE5", "\x00\x00", "\xCC\xE3",
  /* U+6784 */	"\xCC\xDB", "\xCC\xD3", "\xCC\xDA", "\xAA\x4A",
  /* U+6788 */	"\x00\x00", "\xAA\x50", "\x00\x00", "\xAA\x44",
  /* U+678C */	"\xCC\xDE", "\xCC\xDD", "\xCC\xD5", "\x93\xE5",
  /* U+6790 */	"\xAA\x52", "\xCC\xE1", "\xCC\xD6", "\xAA\x55",
  /* U+6794 */	"\xCC\xE8", "\xAA\x45", "\x00\x00", "\xAA\x4C",
  /* U+6798 */	"\xCC\xD9", "\xCC\xE2", "\xAA\x54", "\x00\x00",
  /* U+679C */	"\xAA\x47", "\xAA\x4B", "\x00\x00", "\xCC\xE0",
  /* U+67A0 */	"\x9A\x59", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+67A4 */	"\x8D\xB5", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+67A8 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+67AC */	"\xFD\x4D", "\x00\x00", "\xCF\x5B", "\xAC\x5C",
  /* U+67B0 */	"\xAC\x69", "\xFD\x5E", "\xCF\x56", "\xCF\x4C",
  /* U+67B4 */	"\xAC\x62", "\xCF\x4A", "\xAC\x5B", "\xCF\x45",
  /* U+67B8 */	"\xAC\x65", "\xCF\x52", "\xCE\xFE", "\xCF\x41",
  /* U+67BC */	"\x00\x00", "\x00\x00", "\x00\x00", "\x8F\x7D",
  /* U+67C0 */	"\xCF\x44", "\xCE\xFB", "\xCF\x51", "\xCF\x61",
  /* U+67C4 */	"\xAC\x60", "\xCF\x46", "\xCF\x58", "\x00\x00",
  /* U+67C8 */	"\xCE\xFD", "\xCF\x5F", "\xCF\x60", "\xCF\x63",
  /* U+67CC */	"\xCF\x5A", "\xCF\x4B", "\xCF\x53", "\xAC\x66",
  /* U+67D0 */	"\xAC\x59", "\xAC\x61", "\xAC\x6D", "\xAC\x56",
  /* U+67D4 */	"\xAC\x58", "\x00\x00", "\x95\x47", "\xFC\xF6",
  /* U+67D8 */	"\xCF\x43", "\xAC\x6A", "\xAC\x63", "\xCF\x5D",
  /* U+67DC */	"\xCF\x40", "\xAC\x6C", "\xAC\x67", "\xCF\x49",
  /* U+67E0 */	"\x00\x00", "\x00\x00", "\xAC\x6B", "\xCF\x50",
  /* U+67E4 */	"\xCF\x48", "\xAC\x64", "\xCF\x5C", "\xCF\x54",
  /* U+67E8 */	"\x00\x00", "\xAC\x5E", "\xCF\x62", "\xCF\x47",
  /* U+67EC */	"\xAC\x5A", "\xCF\x59", "\xCF\x4F", "\xAC\x5F",
  /* U+67F0 */	"\xCF\x55", "\xAC\x57", "\xCE\xFC", "\xAC\x68",
  /* U+67F4 */	"\xAE\xE3", "\xAC\x5D", "\xCF\x4E", "\xCF\x4D",
  /* U+67F8 */	"\xCF\x42", "\x92\x50", "\xCF\x5E", "\x00\x00",
  /* U+67FC */	"\xCF\x57", "\x00\x00", "\x89\x68", "\xAC\x55",
  /* U+6800 */	"\x8D\xB6", "\xFC\xFB", "\xA0\x7D", "\x98\xFC",
  /* U+6804 */	"\x89\x69", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+6808 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+680C */	"\x00\x00", "\xFE\x4F", "\x00\x00", "\x00\x00",
  /* U+6810 */	"\x92\x56", "\x00\x00", "\xD1\xEC", "\xAE\xEA",
  /* U+6814 */	"\xD1\xED", "\x00\x00", "\xD1\xE1", "\xAE\xDF",
  /* U+6818 */	"\xAE\xEB", "\x00\x00", "\xD1\xDA", "\xFA\xC9",
  /* U+681C */	"\xD1\xE3", "\xD1\xEB", "\x93\xE8", "\xD1\xD9",
  /* U+6820 */	"\xD1\xF4", "\xAE\xD5", "\xFC\xF8", "\x00\x00",
  /* U+6824 */	"\x00\x00", "\xD1\xF3", "\xD1\xEE", "\x00\x00",
  /* U+6828 */	"\xD1\xEF", "\xAE\xDD", "\xAE\xE8", "\xD1\xE5",
  /* U+682C */	"\x00\x00", "\xD1\xE6", "\xD1\xF0", "\xD1\xE7",
  /* U+6830 */	"\x00\x00", "\xD1\xE2", "\xD1\xDC", "\xD1\xDD",
  /* U+6834 */	"\xD1\xEA", "\xD1\xE4", "\x9C\xE3", "\xFD\xA9",
  /* U+6838 */	"\xAE\xD6", "\xAE\xDA", "\xD1\xF2", "\xD1\xDE",
  /* U+683C */	"\xAE\xE6", "\xAE\xE2", "\xFC\x44", "\x00\x00",
  /* U+6840 */	"\xAE\xE5", "\xAE\xEC", "\xAE\xDB", "\xAE\xE7",
  /* U+6844 */	"\xD1\xE9", "\xAE\xE9", "\xAE\xD8", "\x96\x40",
  /* U+6848 */	"\xAE\xD7", "\xD1\xDB", "\x8D\xB8", "\xD1\xDF",
  /* U+684C */	"\xAE\xE0", "\xD1\xF1", "\xD1\xE8", "\xD1\xE0",
  /* U+6850 */	"\xAE\xE4", "\xAE\xE1", "\x00\x00", "\xAE\xD9",
  /* U+6854 */	"\xAE\xDC", "\x9B\x4A", "\x8F\xB9", "\x00\x00",
  /* U+6858 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+685C */	"\x00\x00", "\xFC\xFE", "\x00\x00", "\x00\x00",
  /* U+6860 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+6864 */	"\x00\x00", "\x89\x6A", "\x00\x00", "\x00\x00",
  /* U+6868 */	"\x00\x00", "\x00\x00", "\x00\x00", "\xD5\xC4",
  /* U+686C */	"\x00\x00", "\xD5\xB4", "\xD5\xB5", "\xD5\xB9",
  /* U+6870 */	"\x00\x00", "\xD5\xC8", "\xD5\xC5", "\x00\x00",
  /* U+6874 */	"\xD5\xBE", "\xD5\xBD", "\xB1\xED", "\xD5\xC1",
  /* U+6878 */	"\xD5\xD0", "\xD5\xB0", "\x00\x00", "\xD5\xD1",
  /* U+687C */	"\xD5\xC3", "\xD5\xD5", "\xD5\xC9", "\xB1\xEC",
  /* U+6880 */	"\xD5\xC7", "\xB1\xE7", "\xB1\xFC", "\xB1\xF2",
  /* U+6884 */	"\x8D\xB9", "\xB1\xF6", "\xB1\xF5", "\xD5\xB1",
  /* U+6888 */	"\x91\x7E", "\xD5\xCE", "\xD5\xD4", "\xD5\xCC",
  /* U+688C */	"\xD5\xD3", "\x00\x00", "\x00\x00", "\xD5\xC0",
  /* U+6890 */	"\xD5\xB2", "\xD5\xD2", "\xD5\xC2", "\xB1\xEA",
  /* U+6894 */	"\xB1\xF7", "\x00\x00", "\xD5\xCB", "\xB1\xF0",
  /* U+6898 */	"\x93\xF4", "\x00\x00", "\x00\x00", "\xD5\xCA",
  /* U+689C */	"\xD5\xB3", "\xB1\xF8", "\x00\x00", "\xB1\xFA",
  /* U+68A0 */	"\xD5\xCD", "\xB1\xFB", "\xB1\xE9", "\xD5\xBA",
  /* U+68A4 */	"\xD5\xCF", "\x00\x00", "\xFB\x7C", "\xB1\xEF",
  /* U+68A8 */	"\xB1\xF9", "\xD5\xBC", "\xD5\xC6", "\xD5\xB7",
  /* U+68AC */	"\xD5\xBB", "\xB1\xF4", "\xD5\xB6", "\xB1\xE8",
  /* U+68B0 */	"\xB1\xF1", "\xB1\xEE", "\xD5\xBF", "\xAE\xDE",
  /* U+68B4 */	"\xD9\xC0", "\xB1\xEB", "\x93\xE7", "\x00\x00",
  /* U+68B8 */	"\x00\x00", "\x97\xEF", "\x00\x00", "\x00\x00",
  /* U+68BC */	"\x00\x00", "\xFE\x4A", "\x00\x00", "\x00\x00",
  /* U+68C0 */	"\x00\x00", "\x00\x00", "\x00\x00", "\xFD\x45",
  /* U+68C4 */	"\xB1\xF3", "\x96\xA5", "\xD9\xC3", "\xD9\xD9",
  /* U+68C8 */	"\xD9\xCE", "\xB4\xD6", "\xFE\xE0", "\xB4\xD1",
  /* U+68CC */	"\xD9\xBD", "\xB4\xD2", "\xD9\xCD", "\x00\x00",
  /* U+68D0 */	"\xD9\xC6", "\xD9\xD3", "\xB4\xCE", "\xD9\xAB",
  /* U+68D4 */	"\xD9\xD5", "\xB4\xC4", "\xD9\xB3", "\xB4\xC7",
  /* U+68D8 */	"\xB4\xC6", "\x00\x00", "\xB4\xD7", "\x00\x00",
  /* U+68DC */	"\xD9\xAD", "\xD9\xCF", "\xD9\xD0", "\xB4\xC9",
  /* U+68E0 */	"\xB4\xC5", "\xD9\xBB", "\x00\x00", "\xB4\xD0",
  /* U+68E4 */	"\xD9\xB6", "\x00\x00", "\xD9\xD1", "\xB4\xCC",
  /* U+68E8 */	"\xD9\xC9", "\xD9\xD6", "\xD9\xB0", "\xD9\xB5",
  /* U+68EC */	"\xD9\xAF", "\x00\x00", "\xB4\xCB", "\xD9\xC2",
  /* U+68F0 */	"\xDD\xDE", "\xD9\xB1", "\xB4\xCF", "\xD9\xBA",
  /* U+68F4 */	"\xD9\xD2", "\xB4\xCA", "\xD9\xB7", "\xD9\xB4",
  /* U+68F8 */	"\xD9\xC5", "\xB4\xCD", "\xB4\xC3", "\xB4\xD9",
  /* U+68FC */	"\xD9\xC8", "\xD9\xC7", "\x00\x00", "\x00\x00",
  /* U+6900 */	"\xFD\x48", "\xFD\x47", "\xFE\xF2", "\xFE\x6A",
  /* U+6904 */	"\xD9\xAC", "\xB4\xC8", "\xD9\xD4", "\xD9\xBC",
  /* U+6908 */	"\xD9\xBE", "\x8D\xBD", "\xD9\xCB", "\xD9\xCA",
  /* U+690C */	"\xD9\xAA", "\xB4\xD3", "\xB4\xD5", "\xD9\xB2",
  /* U+6910 */	"\xD9\xB9", "\xD9\xC1", "\xB4\xD4", "\xD9\xB8",
  /* U+6914 */	"\xD9\xC4", "\xD9\xD7", "\x00\x00", "\xD9\xCC",
  /* U+6918 */	"\x9B\xA1", "\x8C\xA2", "\x9A\xB7", "\x8E\xFC",
  /* U+691C */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+6920 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+6924 */	"\x00\x00", "\xD9\xD8", "\x00\x00", "\x00\x00",
  /* U+6928 */	"\x00\x00", "\x00\x00", "\xD9\xAE", "\x00\x00",
  /* U+692C */	"\x9F\xA1", "\x00\x00", "\x00\x00", "\xDD\xF2",
  /* U+6930 */	"\xB7\xA6", "\x00\x00", "\xDD\xF0", "\xDD\xDB",
  /* U+6934 */	"\xDD\xE0", "\xDD\xD9", "\xFD\x51", "\xDD\xEC",
  /* U+6938 */	"\xDD\xCB", "\xDD\xD2", "\x00\x00", "\xDD\xEA",
  /* U+693C */	"\xDD\xF4", "\xDD\xDC", "\xFA\xAD", "\xDD\xCF",
  /* U+6940 */	"\xDD\xE2", "\xDD\xE7", "\xDD\xD3", "\x8D\xBE",
  /* U+6944 */	"\xDD\xE4", "\xDD\xD0", "\x89\xA4", "\x00\x00",
  /* U+6948 */	"\xDD\xD7", "\xDD\xD8", "\xB7\xA8", "\xDD\xEB",
  /* U+694C */	"\xDD\xE9", "\x00\x00", "\xDD\xCC", "\xDD\xEE",
  /* U+6950 */	"\x00\x00", "\xDD\xEF", "\xDD\xF1", "\xB7\xAC",
  /* U+6954 */	"\xB7\xA4", "\x9A\xD9", "\xD5\xB8", "\xDD\xD4",
  /* U+6958 */	"\xDD\xE6", "\xDD\xD5", "\xB7\xA1", "\xB7\xB1",
  /* U+695C */	"\xDD\xED", "\xB7\xAF", "\xB7\xAB", "\xDD\xCA",
  /* U+6960 */	"\xB7\xA3", "\xFD\x4E", "\xDD\xCD", "\xB7\xB0",
  /* U+6964 */	"\x8D\xC0", "\xDD\xDD", "\xDD\xC9", "\x97\xF0",
  /* U+6968 */	"\xB7\xA9", "\xDD\xE1", "\xDD\xD1", "\xB7\xAA",
  /* U+696C */	"\xDD\xDA", "\xB7\x7E", "\xB4\xD8", "\xDD\xE3",
  /* U+6970 */	"\xD9\xBF", "\xDD\xCE", "\x93\xB4", "\xFD\x44",
  /* U+6974 */	"\xDD\xE8", "\xB7\xA5", "\xDD\xE5", "\xB7\xA2",
  /* U+6978 */	"\xDD\xDF", "\xB7\xAD", "\xDD\xD6", "\xDD\xF3",
  /* U+697C */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+6980 */	"\x9F\xA7", "\x00\x00", "\xB7\xA7", "\xDE\xC6",
  /* U+6984 */	"\x00\x00", "\x8D\xC2", "\xB7\xAE", "\x00\x00",
  /* U+6988 */	"\x00\x00", "\x00\x00", "\x99\xB6", "\x00\x00",
  /* U+698C */	"\x00\x00", "\xE2\x4A", "\xE2\x48", "\x00\x00",
  /* U+6990 */	"\xE2\x5E", "\xE2\x46", "\x00\x00", "\xE2\x58",
  /* U+6994 */	"\xB7\x7D", "\xBA\x5F", "\xE2\x42", "\xE2\x5D",
  /* U+6998 */	"\xFD\x52", "\xE2\x47", "\xE2\x55", "\xBA\x64",
  /* U+699C */	"\xBA\x5D", "\x00\x00", "\xE2\x5B", "\x8D\xC1",
  /* U+69A0 */	"\xE2\x40", "\xE2\x5A", "\x8E\x46", "\xBA\x6F",
  /* U+69A4 */	"\xE2\x51", "\xE2\x61", "\xBA\x6D", "\xE2\x49",
  /* U+69A8 */	"\xBA\x5E", "\xE2\x4B", "\xE2\x59", "\xBA\x67",
  /* U+69AC */	"\xE2\x44", "\xBA\x6B", "\xBA\x61", "\xE2\x4D",
  /* U+69B0 */	"\xE2\x43", "\xE1\xFC", "\xA0\xD1", "\xE2\x57",
  /* U+69B4 */	"\xBA\x68", "\xE2\x60", "\xE1\xFD", "\xBA\x65",
  /* U+69B8 */	"\x00\x00", "\xE2\x53", "\x00\x00", "\xBA\x66",
  /* U+69BC */	"\xE2\x45", "\xE2\x50", "\xE2\x4C", "\xE2\x4E",
  /* U+69C0 */	"\x9F\xCA", "\xBA\x60", "\xE2\x5F", "\xBA\x6E",
  /* U+69C4 */	"\xE2\x4F", "\x00\x00", "\xE2\x62", "\x00\x00",
  /* U+69C8 */	"\x00\x00", "\xE1\xFE", "\xE2\x54", "\xBA\x63",
  /* U+69CC */	"\xBA\x6C", "\xBA\x6A", "\xE2\x41", "\xE2\x56",
  /* U+69D0 */	"\xBA\x69", "\x92\xCF", "\x00\x00", "\xBA\x62",
  /* U+69D4 */	"\xE2\x52", "\x9C\xF4", "\x8D\xC4", "\x00\x00",
  /* U+69D8 */	"\x00\x00", "\xE2\x5C", "\x00\x00", "\x00\x00",
  /* U+69DC */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+69E0 */	"\x00\x00", "\xFD\x41", "\xE5\xD5", "\x00\x00",
  /* U+69E4 */	"\xE5\xD1", "\xE5\xCD", "\xE5\xE1", "\xE5\xDE",
  /* U+69E8 */	"\xBC\xCD", "\x9B\x4C", "\x00\x00", "\xE5\xE5",
  /* U+69EC */	"\xE5\xD4", "\xBC\xD8", "\xE5\xDB", "\x00\x00",
  /* U+69F0 */	"\x00\x00", "\xE5\xD0", "\xE5\xDA", "\xBC\xD5",
  /* U+69F4 */	"\xE5\xEE", "\x00\x00", "\xE5\xEB", "\xE5\xDD",
  /* U+69F8 */	"\xE5\xCE", "\xFD\x57", "\xFC\xEF", "\xE5\xE2",
  /* U+69FC */	"\xE5\xE4", "\xBC\xD1", "\xE5\xD8", "\xE5\xD3",
  /* U+6A00 */	"\xE5\xCA", "\xBC\xCE", "\xBC\xD6", "\x9C\xDE",
  /* U+6A04 */	"\xE5\xE7", "\xBC\xD7", "\xE5\xCB", "\xE5\xED",
  /* U+6A08 */	"\xE5\xE0", "\xE5\xE6", "\xBC\xD4", "\xFD\x42",
  /* U+6A0C */	"\x98\x6C", "\xE5\xE3", "\x00\x00", "\xE5\xEA",
  /* U+6A10 */	"\x00\x00", "\xBC\xD9", "\x00\x00", "\xBC\xD3",
  /* U+6A14 */	"\xE5\xDC", "\xE5\xCF", "\xE5\xEF", "\xE5\xCC",
  /* U+6A18 */	"\xE5\xE8", "\xBC\xD0", "\x97\xF9", "\xE5\xD6",
  /* U+6A1C */	"\x95\x58", "\xE5\xD7", "\xBC\xCF", "\xBC\xCC",
  /* U+6A20 */	"\xE5\xD2", "\xBC\xD2", "\x00\x00", "\xBC\xCB",
  /* U+6A24 */	"\x00\x00", "\xE5\xE9", "\xE5\xEC", "\xE5\xD9",
  /* U+6A28 */	"\xE9\xCA", "\x00\x00", "\x00\x00", "\x98\x5E",
  /* U+6A2C */	"\xFE\x7B", "\x94\xCD", "\x00\x00", "\x00\x00",
  /* U+6A30 */	"\x00\x00", "\x00\x00", "\xE9\xC2", "\x93\xEE",
  /* U+6A34 */	"\xE9\xBE", "\xBE\xF6", "\x00\x00", "\x00\x00",
  /* U+6A38 */	"\xBE\xEB", "\xBE\xF0", "\xBE\xEC", "\xE9\xCC",
  /* U+6A3C */	"\xE9\xD7", "\xBE\xEA", "\xE9\xC4", "\xE9\xCD",
  /* U+6A40 */	"\xE5\xDF", "\xE9\xCE", "\x00\x00", "\x8C\xA3",
  /* U+6A44 */	"\xBE\xF1", "\xFD\x5A", "\xE9\xDD", "\xBE\xF5",
  /* U+6A48 */	"\xBE\xF8", "\xE9\xC0", "\x00\x00", "\xBE\xF4",
  /* U+6A4C */	"\x93\xF5", "\xE9\xDB", "\xE9\xDC", "\xE9\xD2",
  /* U+6A50 */	"\xE9\xD1", "\xE9\xC9", "\x93\xEF", "\x8E\xEA",
  /* U+6A54 */	"\xE9\xD3", "\xE9\xDA", "\xE9\xD9", "\x8F\x5B",
  /* U+6A58 */	"\xBE\xEF", "\xBE\xED", "\xE9\xCB", "\xE9\xC8",
  /* U+6A5C */	"\x00\x00", "\xE9\xC5", "\xE9\xD8", "\xBE\xF7",
  /* U+6A60 */	"\xE9\xD6", "\xBE\xF3", "\xBE\xF2", "\x8C\x5E",
  /* U+6A64 */	"\xE9\xD0", "\x8D\xC6", "\xE9\xBF", "\xE9\xC1",
  /* U+6A68 */	"\xE9\xC3", "\xE9\xD5", "\xE9\xCF", "\xBE\xEE",
  /* U+6A6C */	"\x00\x00", "\xE9\xC6", "\x00\x00", "\xE9\xD4",
  /* U+6A70 */	"\x00\x00", "\x8D\xC8", "\x00\x00", "\x00\x00",
  /* U+6A74 */	"\x8D\xC7", "\x00\x00", "\xE9\xC7", "\x00\x00",
  /* U+6A78 */	"\x00\x00", "\x00\x00", "\x93\xF7", "\x00\x00",
  /* U+6A7C */	"\x00\x00", "\x00\x00", "\xC0\xCF", "\xED\x45",
  /* U+6A80 */	"\xC0\xC8", "\xEC\xF5", "\x8D\xC9", "\xED\x41",
  /* U+6A84 */	"\xC0\xCA", "\xED\x48", "\x00\x00", "\xEC\xFC",
  /* U+6A88 */	"\x00\x00", "\xEC\xF7", "\xFB\xF2", "\x00\x00",
  /* U+6A8C */	"\xED\x49", "\xEC\xF3", "\xEC\xFE", "\x96\x70",
  /* U+6A90 */	"\xC0\xD1", "\xED\x44", "\xED\x4A", "\xEC\xFD",
  /* U+6A94 */	"\xC0\xC9", "\xED\x40", "\xEC\xF4", "\xC0\xD0",
  /* U+6A98 */	"\x00\x00", "\x8D\xCB", "\xED\x47", "\xEC\xF9",
  /* U+6A9C */	"\xC0\xCC", "\xFD\x5C", "\xEC\xFB", "\xEC\xF8",
  /* U+6AA0 */	"\xC0\xD2", "\xEC\xFA", "\xC0\xCB", "\xC0\xCE",
  /* U+6AA4 */	"\xED\x43", "\xEC\xF6", "\xED\x46", "\x8F\x65",
  /* U+6AA8 */	"\xED\x42", "\x00\x00", "\x00\x00", "\x8D\xCD",
  /* U+6AAC */	"\xC2\x63", "\xEF\xE7", "\xC2\x68", "\xC2\x69",
  /* U+6AB0 */	"\x00\x00", "\x9D\xA8", "\x94\xF9", "\xC2\x62",
  /* U+6AB4 */	"\xEF\xE6", "\x8D\xCE", "\xEF\xE3", "\xEF\xE4",
  /* U+6AB8 */	"\xC2\x66", "\xEF\xDE", "\xEF\xE2", "\xC2\x65",
  /* U+6ABC */	"\x00\x00", "\xEF\xDF", "\x93\xEA", "\x00\x00",
  /* U+6AC0 */	"\x00\x00", "\x00\x00", "\xC2\x67", "\xC2\x64",
  /* U+6AC4 */	"\x00\x00", "\xEF\xDD", "\xEF\xE1", "\xEF\xE5",
  /* U+6AC8 */	"\xFD\x5F", "\x93\xF0", "\x9F\xB6", "\xF2\x51",
  /* U+6ACC */	"\xF2\x4E", "\xF2\x57", "\x00\x00", "\xF2\x56",
  /* U+6AD0 */	"\xF2\x54", "\xF2\x4F", "\x00\x00", "\xC3\x72",
  /* U+6AD4 */	"\x8D\xCF", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+6AD8 */	"\x97\x63", "\xF2\x50", "\xC3\x71", "\xC0\xCD",
  /* U+6ADC */	"\xF2\x53", "\xC3\x70", "\xF2\x58", "\xF2\x52",
  /* U+6AE0 */	"\xF2\x4D", "\xEF\xE0", "\x00\x00", "\x00\x00",
  /* U+6AE4 */	"\x00\x00", "\xC3\x6F", "\x00\x00", "\xF2\x4C",
  /* U+6AE8 */	"\xF4\x56", "\x00\x00", "\xF4\x55", "\xF2\x55",
  /* U+6AEC */	"\xC4\x68", "\x00\x00", "\xF4\x59", "\xF4\x5A",
  /* U+6AF0 */	"\xF4\x54", "\xF4\x58", "\x00\x00", "\xF4\x53",
  /* U+6AF4 */	"\x00\x00", "\x00\x00", "\x8D\xD0", "\x00\x00",
  /* U+6AF8 */	"\xF5\xD1", "\xF4\x57", "\xC4\xE7", "\xC4\xE5",
  /* U+6AFC */	"\xF5\xCF", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+6B00 */	"\xF5\xD2", "\x00\x00", "\xF5\xCE", "\xF5\xD0",
  /* U+6B04 */	"\xC4\xE6", "\x93\xF1", "\x00\x00", "\x00\x00",
  /* U+6B08 */	"\xF6\xE5", "\xF6\xE6", "\xC5\x76", "\xF6\xE4",
  /* U+6B0C */	"\x00\x00", "\x00\x00", "\x00\x00", "\xF7\xE2",
  /* U+6B10 */	"\xC5\xCF", "\xF7\xE0", "\xF7\xE1", "\xF8\xAC",
  /* U+6B14 */	"\x00\x00", "\x00\x00", "\xC6\x56", "\xF8\xF3",
  /* U+6B18 */	"\xF8\xF1", "\xF8\xF2", "\xF8\xF4", "\x00\x00",
  /* U+6B1C */	"\x00\x00", "\xFD\x62", "\xF9\xBB", "\x00\x00",
  /* U+6B20 */	"\xA4\xED", "\xA6\xB8", "\x00\x00", "\xAA\x59",
  /* U+6B24 */	"\x00\x00", "\xCC\xE9", "\x00\x00", "\x00\x00",
  /* U+6B28 */	"\xCF\x64", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+6B2C */	"\xD1\xF5", "\xD1\xF7", "\x00\x00", "\xD1\xF6",
  /* U+6B30 */	"\x00\x00", "\xD1\xF8", "\xB1\xFD", "\xD5\xD7",
  /* U+6B34 */	"\xD1\xF9", "\xFD\x65", "\xD5\xD6", "\xD5\xD8",
  /* U+6B38 */	"\xD5\xD9", "\xD9\xDA", "\xB4\xDB", "\xD9\xDB",
  /* U+6B3C */	"\xD9\xDD", "\xB4\xDC", "\xB4\xDA", "\xD9\xDC",
  /* U+6B40 */	"\x00\x00", "\xDD\xFA", "\xDD\xF8", "\xDD\xF7",
  /* U+6B44 */	"\x00\x00", "\xDD\xF6", "\xDD\xF5", "\xB7\xB2",
  /* U+6B48 */	"\xDD\xF9", "\xBA\x70", "\xE2\x63", "\xE2\x65",
  /* U+6B4C */	"\xBA\x71", "\xE2\x64", "\xBC\xDB", "\x00\x00",
  /* U+6B50 */	"\xBC\xDA", "\xE5\xF0", "\x9F\xDB", "\x00\x00",
  /* U+6B54 */	"\xE9\xDF", "\xE9\xDE", "\xE9\xE0", "\x93\xF8",
  /* U+6B58 */	"\x00\x00", "\xBE\xF9", "\x00\x00", "\xED\x4B",
  /* U+6B5C */	"\xC0\xD3", "\x00\x00", "\xEF\xE8", "\xC2\x6A",
  /* U+6B60 */	"\xF2\x59", "\xC5\x77", "\xA4\xEE", "\xA5\xBF",
  /* U+6B64 */	"\xA6\xB9", "\xA8\x42", "\xAA\x5A", "\xAA\x5B",
  /* U+6B68 */	"\x00\x00", "\x00\x00", "\xAC\x6E", "\x00\x00",
  /* U+6B6C */	"\x00\x00", "\xD1\xFA", "\x00\x00", "\x8B\xF7",
  /* U+6B70 */	"\x00\x00", "\x00\x00", "\xB7\xB3", "\x00\x00",
  /* U+6B74 */	"\xFD\x66", "\x00\x00", "\xE6\xD1", "\xBE\xFA",
  /* U+6B78 */	"\xC2\x6B", "\xA4\xEF", "\x8B\xCF", "\xA6\xBA",
  /* U+6B7C */	"\x00\x00", "\x00\x00", "\xCC\xEB", "\xAA\x5C",
  /* U+6B80 */	"\xCC\xEA", "\x8D\xD1", "\xCF\x65", "\xAC\x6F",
  /* U+6B84 */	"\xCF\x66", "\x00\x00", "\xAC\x70", "\x00\x00",
  /* U+6B88 */	"\xD1\xFC", "\xAE\xEE", "\xAE\xED", "\x00\x00",
  /* U+6B8C */	"\xD5\xDE", "\xD5\xDC", "\xD5\xDD", "\xD5\xDB",
  /* U+6B90 */	"\x00\x00", "\xD5\xDA", "\x00\x00", "\x00\x00",
  /* U+6B94 */	"\xD9\xDE", "\xD9\xE1", "\xB4\xDE", "\xD9\xDF",
  /* U+6B98 */	"\xB4\xDD", "\xD9\xE0", "\x00\x00", "\xDD\xFB",
  /* U+6B9C */	"\x00\x00", "\x00\x00", "\xE2\x66", "\xE2\x67",
  /* U+6BA0 */	"\xE2\x68", "\x00\x00", "\xE5\xF3", "\xE5\xF2",
  /* U+6BA4 */	"\xBC\xDC", "\xE5\xF1", "\xE5\xF4", "\xE9\xE1",
  /* U+6BA8 */	"\x00\x00", "\x00\x00", "\xE9\xE2", "\xE9\xE3",
  /* U+6BAC */	"\x00\x00", "\xED\x4C", "\xC0\xD4", "\xC2\x6C",
  /* U+6BB0 */	"\xF2\x5A", "\x00\x00", "\xC4\xE8", "\xC9\x5F",
  /* U+6BB4 */	"\x00\x00", "\xAC\x71", "\xCF\x67", "\xAE\xEF",
  /* U+6BB8 */	"\x00\x00", "\x00\x00", "\xB1\xFE", "\x00\x00",
  /* U+6BBC */	"\xB4\xDF", "\xD9\xE2", "\x00\x00", "\xB7\xB5",
  /* U+6BC0 */	"\xB7\xB4", "\x8D\xD2", "\x00\x00", "\xE2\x69",
  /* U+6BC4 */	"\xE2\x6A", "\xBC\xDD", "\xBC\xDE", "\xE9\xE5",
  /* U+6BC8 */	"\xE9\xE4", "\xEF\xE9", "\xF7\xE3", "\xA4\xF0",
  /* U+6BCC */	"\xC9\x60", "\xA5\xC0", "\x00\x00", "\xA8\x43",
  /* U+6BD0 */	"\xCB\x48", "\x00\x00", "\xAC\x72", "\xB7\xB6",
  /* U+6BD4 */	"\xA4\xF1", "\x00\x00", "\xCF\x68", "\xAC\x73",
  /* U+6BD8 */	"\xCF\x69", "\x00\x00", "\xC0\xD5", "\xA4\xF2",
  /* U+6BDC */	"\xFD\x71", "\x00\x00", "\xCC\xEC", "\x00\x00",
  /* U+6BE0 */	"\xCF\x6A", "\xFD\x6F", "\xD2\x42", "\xD2\x41",
  /* U+6BE4 */	"\xD1\xFE", "\x00\x00", "\xD1\xFD", "\xD2\x43",
  /* U+6BE8 */	"\xD2\x40", "\x00\x00", "\x8D\xD3", "\xB2\x40",
  /* U+6BEC */	"\xB2\x41", "\x00\x00", "\x00\x00", "\xB4\xE0",
  /* U+6BF0 */	"\xD9\xE3", "\x00\x00", "\xD9\xE4", "\xD9\xE5",
  /* U+6BF4 */	"\x00\x00", "\x00\x00", "\x00\x00", "\xDE\x41",
  /* U+6BF8 */	"\xDE\x42", "\xDE\x40", "\x9F\xE7", "\xDD\xFD",
  /* U+6BFC */	"\xDD\xFE", "\xB7\xB7", "\xE2\x6B", "\xE5\xF7",
  /* U+6C00 */	"\xE5\xF6", "\xE5\xF5", "\xE5\xF8", "\xE9\xE7",
  /* U+6C04 */	"\xE9\xE6", "\xBE\xFB", "\xE9\xE8", "\x00\x00",
  /* U+6C08 */	"\xC0\xD6", "\xED\x4D", "\x00\x00", "\xEF\xEA",
  /* U+6C0C */	"\xF2\x5B", "\xF6\xE7", "\x00\x00", "\xA4\xF3",
  /* U+6C10 */	"\xA5\xC2", "\xA5\xC1", "\x00\x00", "\xAA\x5D",
  /* U+6C14 */	"\xC9\x61", "\xC9\x7E", "\xA6\xBB", "\x00\x00",
  /* U+6C18 */	"\xC9\xF7", "\xCB\x49", "\xCB\x4A", "\xAA\x5E",
  /* U+6C1C */	"\x90\xBD", "\xCC\xED", "\x00\x00", "\xAC\x74",
  /* U+6C20 */	"\xCF\x6B", "\xCF\x6C", "\x00\x00", "\xAE\xF0",
  /* U+6C24 */	"\xAE\xF4", "\xD2\x44", "\xAE\xF3", "\xAE\xF1",
  /* U+6C28 */	"\xAE\xF2", "\x00\x00", "\xD5\xDF", "\xB2\x42",
  /* U+6C2C */	"\xB4\xE3", "\x00\x00", "\xB4\xE1", "\xB4\xE2",
  /* U+6C30 */	"\xD9\xE6", "\x9F\xD0", "\x00\x00", "\xBA\x72",
  /* U+6C34 */	"\xA4\xF4", "\x8B\xD0", "\xC9\xA1", "\xFD\x72",
  /* U+6C38 */	"\xA5\xC3", "\x9C\xAE", "\x8B\xD1", "\xC9\xA4",
  /* U+6C3C */	"\x00\x00", "\x8A\xDB", "\xA5\xC6", "\xC9\xA3",
  /* U+6C40 */	"\xA5\xC5", "\xA5\xC4", "\xA8\x44", "\xC9\xA2",
  /* U+6C44 */	"\x00\x00", "\x00\x00", "\xC9\xF8", "\x00\x00",
  /* U+6C48 */	"\x00\x00", "\xFA\xE4", "\xC9\xFC", "\xC9\xFE",
  /* U+6C4C */	"\xCA\x40", "\xA6\xC5", "\xA6\xC6", "\xC9\xFB",
  /* U+6C50 */	"\xA6\xC1", "\x00\x00", "\xC9\xF9", "\x00\x00",
  /* U+6C54 */	"\xC9\xFD", "\xA6\xC2", "\x00\x00", "\xA6\xBD",
  /* U+6C58 */	"\x95\xCE", "\xA6\xBE", "\xFD\x76", "\xA6\xC4",
  /* U+6C5C */	"\xC9\xFA", "\xA6\xBC", "\xA8\x45", "\xA6\xBF",
  /* U+6C60 */	"\xA6\xC0", "\xA6\xC3", "\x00\x00", "\x00\x00",
  /* U+6C64 */	"\x00\x00", "\xCB\x5B", "\xCB\x59", "\xCB\x4C",
  /* U+6C68 */	"\xA8\x51", "\xCB\x53", "\xA8\x4C", "\xCB\x4D",
  /* U+6C6C */	"\x00\x00", "\xCB\x55", "\xFB\x62", "\xCB\x52",
  /* U+6C70 */	"\xA8\x4F", "\xCB\x51", "\xA8\x56", "\xCB\x5A",
  /* U+6C74 */	"\xA8\x58", "\x8D\xD4", "\xA8\x5A", "\x00\x00",
  /* U+6C78 */	"\xCB\x4B", "\xFD\x78", "\xA8\x4D", "\xCB\x5C",
  /* U+6C7C */	"\x00\x00", "\xA8\x54", "\xA8\x57", "\x8E\xE3",
  /* U+6C80 */	"\xCD\x45", "\xA8\x47", "\xA8\x5E", "\xA8\x55",
  /* U+6C84 */	"\xCB\x4E", "\xA8\x4A", "\xA8\x59", "\xCB\x56",
  /* U+6C88 */	"\xA8\x48", "\xA8\x49", "\xCD\x43", "\xCB\x4F",
  /* U+6C8C */	"\xA8\x50", "\xA8\x5B", "\xCB\x5D", "\xCB\x50",
  /* U+6C90 */	"\xA8\x4E", "\x00\x00", "\xA8\x53", "\xCC\xEE",
  /* U+6C94 */	"\xA8\x5C", "\xCB\x57", "\xA8\x52", "\x00\x00",
  /* U+6C98 */	"\xA8\x5D", "\xA8\x46", "\xCB\x54", "\xA8\x4B",
  /* U+6C9C */	"\xCB\x58", "\xCD\x44", "\x00\x00", "\x90\x76",
  /* U+6CA0 */	"\x00\x00", "\x00\x00", "\x98\xC6", "\x00\x00",
  /* U+6CA4 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+6CA8 */	"\x00\x00", "\x00\x00", "\x8D\xD5", "\xAA\x6A",
  /* U+6CAC */	"\xAA\x7A", "\xCC\xF5", "\xAA\x71", "\x97\xD1",
  /* U+6CB0 */	"\xCD\x4B", "\xAA\x62", "\x9E\xB6", "\xAA\x65",
  /* U+6CB4 */	"\xCD\x42", "\x00\x00", "\xCC\xF3", "\xCC\xF7",
  /* U+6CB8 */	"\xAA\x6D", "\xAA\x6F", "\xCC\xFA", "\xAA\x76",
  /* U+6CBC */	"\xAA\x68", "\xAA\x66", "\xAA\x67", "\xAA\x75",
  /* U+6CC0 */	"\xCD\x47", "\xAA\x70", "\xCC\xF9", "\xCC\xFB",
  /* U+6CC4 */	"\xAA\x6E", "\xAA\x73", "\xCC\xFC", "\xCD\x4A",
  /* U+6CC8 */	"\x00\x00", "\xAC\x75", "\xAA\x79", "\xFA\xC7",
  /* U+6CCC */	"\xAA\x63", "\xCD\x49", "\xA0\x42", "\xCD\x4D",
  /* U+6CD0 */	"\xCC\xF8", "\xCD\x4F", "\xCD\x40", "\xAA\x6C",
  /* U+6CD4 */	"\xCC\xF4", "\xAA\x6B", "\xAA\x7D", "\xAA\x72",
  /* U+6CD8 */	"\x00\x00", "\xCC\xF2", "\xCF\x75", "\xAA\x78",
  /* U+6CDC */	"\xAA\x7C", "\xCD\x41", "\xCD\x46", "\x98\x73",
  /* U+6CE0 */	"\xAA\x7E", "\xAA\x77", "\xAA\x69", "\xAA\x5F",
  /* U+6CE4 */	"\x00\x00", "\xAA\x64", "\x00\x00", "\xCC\xF6",
  /* U+6CE8 */	"\xAA\x60", "\xCD\x4E", "\x9F\xFC", "\xCC\xF0",
  /* U+6CEC */	"\xCC\xEF", "\xCC\xFD", "\xCC\xF1", "\xAA\x7B",
  /* U+6CF0 */	"\xAE\xF5", "\xAA\x74", "\xCC\xFE", "\xAA\x61",
  /* U+6CF4 */	"\x00\x00", "\xAC\xA6", "\x00\x00", "\x00\x00",
  /* U+6CF8 */	"\x00\x00", "\xCD\x4C", "\x00\x00", "\x00\x00",
  /* U+6CFC */	"\x00\x00", "\x00\x00", "\x00\x00", "\x8C\xA5",
  /* U+6D00 */	"\xCF\x7C", "\xCF\xA1", "\x8D\xD7", "\xCF\xA4",
  /* U+6D04 */	"\xCF\x77", "\x92\xFB", "\x8D\xD8", "\xCF\xA7",
  /* U+6D08 */	"\xCF\xAA", "\xCF\xAC", "\xCF\x74", "\xAC\x76",
  /* U+6D0C */	"\xAC\x7B", "\xD2\x49", "\xAC\xAD", "\xCF\xA5",
  /* U+6D10 */	"\xCF\xAD", "\xCF\x7B", "\xCF\x73", "\x00\x00",
  /* U+6D14 */	"\x00\x00", "\x00\x00", "\xD2\x64", "\xAC\x7E",
  /* U+6D18 */	"\xCF\xA2", "\xCF\x78", "\xCF\x7A", "\xAC\xA5",
  /* U+6D1C */	"\x00\x00", "\xCF\x7D", "\xAC\x7D", "\xCF\x70",
  /* U+6D20 */	"\xCF\xA8", "\x00\x00", "\xCF\xAB", "\x00\x00",
  /* U+6D24 */	"\x94\x4F", "\xAC\x7A", "\x8D\xD9", "\xAC\xA8",
  /* U+6D28 */	"\xCF\x6D", "\xAC\xAA", "\xAC\x78", "\xAC\xAE",
  /* U+6D2C */	"\xCF\xA9", "\xCF\x6F", "\xAC\xAB", "\xD2\x5E",
  /* U+6D30 */	"\xCD\x48", "\xAC\x7C", "\xAC\x77", "\xCF\x76",
  /* U+6D34 */	"\xCF\x6E", "\xAC\xAC", "\xAC\xA4", "\xCF\xA3",
  /* U+6D38 */	"\xAC\xA9", "\xAC\xA7", "\xCF\x79", "\xAC\xA1",
  /* U+6D3C */	"\xCF\x71", "\xAC\xA2", "\xAC\xA3", "\xCF\x72",
  /* U+6D40 */	"\xCF\xA6", "\xAC\x79", "\xCF\x7E", "\x00\x00",
  /* U+6D44 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+6D48 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+6D4C */	"\x00\x00", "\x00\x00", "\x89\x6B", "\x00\x00",
  /* U+6D50 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+6D54 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x97\xCE",
  /* U+6D58 */	"\xD2\x4C", "\xAE\xFD", "\xAF\x43", "\xFA\xF3",
  /* U+6D5C */	"\xFD\xAE", "\x00\x00", "\xD2\x55", "\xD2\x5B",
  /* U+6D60 */	"\xD2\x57", "\xD2\x4A", "\xD2\x4D", "\xD2\x46",
  /* U+6D64 */	"\xD2\x47", "\xAF\x4A", "\xAE\xFA", "\xD2\x56",
  /* U+6D68 */	"\xD2\x5F", "\xAF\x45", "\xAE\xF6", "\x00\x00",
  /* U+6D6C */	"\xAF\x40", "\xD2\x4E", "\xAF\x42", "\xD2\x4F",
  /* U+6D70 */	"\xD2\x59", "\xFB\xAF", "\x92\xB7", "\x00\x00",
  /* U+6D74 */	"\xAF\x44", "\xD2\x68", "\xD2\x48", "\xAE\xFC",
  /* U+6D78 */	"\xAE\xFB", "\xAF\x48", "\xD2\x45", "\xD2\x66",
  /* U+6D7C */	"\xD2\x5A", "\xD2\x67", "\xD2\x61", "\xD2\x53",
  /* U+6D80 */	"\xD2\x62", "\x8D\xDA", "\xD2\x5C", "\xD2\x65",
  /* U+6D84 */	"\xD2\x63", "\xAF\x49", "\xD2\x54", "\xAE\xF9",
  /* U+6D88 */	"\xAE\xF8", "\xAF\x41", "\xAF\x47", "\xD2\x60",
  /* U+6D8C */	"\xAF\x46", "\xD2\x51", "\xB2\x43", "\x9C\x5A",
  /* U+6D90 */	"\xD2\x69", "\xD2\x50", "\xD2\x4B", "\xAE\xFE",
  /* U+6D94 */	"\xAF\x4B", "\xAE\xF7", "\xFD\xAD", "\xD2\x58",
  /* U+6D98 */	"\xD2\x5D", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+6D9C */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+6DA0 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+6DA4 */	"\x8D\xDC", "\x94\x44", "\x00\x00", "\x00\x00",
  /* U+6DA8 */	"\x00\x00", "\x00\x00", "\xB2\x65", "\xD5\xE1",
  /* U+6DAC */	"\xD5\xE5", "\x00\x00", "\xB2\x52", "\xB2\x50",
  /* U+6DB0 */	"\x00\x00", "\x8D\xDD", "\xB2\x47", "\xD5\xE3",
  /* U+6DB4 */	"\xD5\xE2", "\xB2\x5B", "\x00\x00", "\xD5\xE8",
  /* U+6DB8 */	"\xB2\x55", "\xA0\xD6", "\xD5\xFA", "\xD6\x47",
  /* U+6DBC */	"\xB2\x44", "\xD5\xF7", "\xD5\xF0", "\xB2\x67",
  /* U+6DC0 */	"\xD5\xE0", "\x00\x00", "\xD5\xFC", "\x00\x00",
  /* U+6DC4 */	"\xB2\x64", "\xB2\x58", "\xB2\x63", "\xB2\x4E",
  /* U+6DC8 */	"\xD5\xEC", "\xD5\xFE", "\xD5\xF6", "\xB2\x4F",
  /* U+6DCC */	"\xB2\x49", "\xD6\x45", "\x00\x00", "\xD5\xFD",
  /* U+6DD0 */	"\xD6\x40", "\xB2\x51", "\xB2\x59", "\xD6\x42",
  /* U+6DD4 */	"\xD5\xEA", "\xD5\xFB", "\xD5\xEF", "\xD6\x44",
  /* U+6DD8 */	"\xB2\x5E", "\xB2\x46", "\xB2\x5C", "\xD5\xF4",
  /* U+6DDC */	"\xD5\xF2", "\xD5\xF3", "\xB2\x53", "\xD5\xEE",
  /* U+6DE0 */	"\xD5\xED", "\xB2\x48", "\xD5\xE7", "\xD6\x46",
  /* U+6DE4 */	"\xB2\x4A", "\xD5\xF1", "\xB2\x68", "\x00\x00",
  /* U+6DE8 */	"\xB2\x62", "\xD5\xE6", "\xB2\x5F", "\xB2\x5D",
  /* U+6DEC */	"\xB2\x66", "\xD5\xF8", "\xB2\x61", "\xD2\x52",
  /* U+6DF0 */	"\xD5\xF9", "\xB2\x60", "\xD6\x41", "\xB2\x45",
  /* U+6DF4 */	"\xD5\xF5", "\xB2\x57", "\xD5\xE9", "\xB2\x56",
  /* U+6DF8 */	"\x00\x00", "\xB2\x54", "\xB2\x4C", "\xB2\x4B",
  /* U+6DFC */	"\xD9\xE7", "\xD6\x43", "\x8C\x41", "\x00\x00",
  /* U+6E00 */	"\xD5\xEB", "\x00\x00", "\x97\xD5", "\xD9\xFC",
  /* U+6E04 */	"\x94\x4A", "\xB2\x4D", "\x00\x00", "\x00\x00",
  /* U+6E08 */	"\x00\x00", "\x00\x00", "\x94\x4D", "\x00\x00",
  /* U+6E0C */	"\x00\x00", "\x00\x00", "\x00\x00", "\x97\xCB",
  /* U+6E10 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+6E14 */	"\x00\x00", "\x8D\xDE", "\x00\x00", "\x00\x00",
  /* U+6E18 */	"\x8D\xDF", "\xB5\x41", "\xB2\x5A", "\xB4\xEE",
  /* U+6E1C */	"\xD9\xF6", "\xB4\xFC", "\x00\x00", "\xD9\xEA",
  /* U+6E20 */	"\xB4\xEB", "\xB4\xE7", "\xDA\x49", "\xB4\xED",
  /* U+6E24 */	"\xB4\xF1", "\xB4\xEC", "\xB4\xF5", "\xDA\x4D",
  /* U+6E28 */	"\xDA\x44", "\x8D\xE0", "\xFE\xF9", "\xD9\xF1",
  /* U+6E2C */	"\xB4\xFA", "\xB4\xF4", "\xD9\xFD", "\xB4\xE4",
  /* U+6E30 */	"\xDA\x4A", "\xDA\x43", "\xB4\xE8", "\xD9\xF7",
  /* U+6E34 */	"\xB4\xF7", "\xDA\x55", "\xDA\x56", "\x00\x00",
  /* U+6E38 */	"\xB4\xE5", "\xDA\x48", "\xB4\xF9", "\xD9\xFB",
  /* U+6E3C */	"\xD9\xED", "\xD9\xEE", "\xB4\xFD", "\xD9\xF2",
  /* U+6E40 */	"\xD9\xF9", "\xD9\xF3", "\x00\x00", "\xB4\xFB",
  /* U+6E44 */	"\xB5\x44", "\xD9\xEF", "\xD9\xE8", "\xD9\xE9",
  /* U+6E48 */	"\x00\x00", "\xD9\xEB", "\xB4\xEA", "\xD9\xF8",
  /* U+6E4C */	"\x00\x00", "\xB4\xF8", "\xB5\x42", "\xFD\xC0",
  /* U+6E50 */	"\xFC\xF9", "\xD9\xFA", "\xDA\x53", "\xDA\x4B",
  /* U+6E54 */	"\xB4\xE6", "\xDA\x51", "\xB4\xF2", "\x00\x00",
  /* U+6E58 */	"\xB4\xF0", "\xFB\x7E", "\xDA\x57", "\xB4\xEF",
  /* U+6E5C */	"\xDA\x41", "\xD9\xF4", "\xD9\xFE", "\xB5\x47",
  /* U+6E60 */	"\xDA\x45", "\xDA\x42", "\xD9\xF0", "\xB5\x43",
  /* U+6E64 */	"\xDA\x4F", "\xDA\x4C", "\xDA\x54", "\xB4\xE9",
  /* U+6E68 */	"\xDA\x40", "\xB5\x46", "\x00\x00", "\xDA\x47",
  /* U+6E6C */	"\x00\x00", "\x00\x00", "\xB4\xF3", "\xB4\xF6",
  /* U+6E70 */	"\x00\x00", "\xDA\x46", "\xB5\x45", "\xD9\xF5",
  /* U+6E74 */	"\xD5\xE4", "\x00\x00", "\x92\xB3", "\xDA\x50",
  /* U+6E78 */	"\xDA\x4E", "\xDA\x52", "\x00\x00", "\x00\x00",
  /* U+6E7C */	"\xFD\xAF", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+6E80 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+6E84 */	"\x00\x00", "\x00\x00", "\x8D\xE1", "\x00\x00",
  /* U+6E88 */	"\xD9\xEC", "\xB5\x40", "\x00\x00", "\x95\xD3",
  /* U+6E8C */	"\x00\x00", "\xDE\x61", "\xDE\x60", "\xDE\x46",
  /* U+6E90 */	"\xB7\xBD", "\x00\x00", "\xDE\x5F", "\xDE\x49",
  /* U+6E94 */	"\xDE\x4A", "\x00\x00", "\xB7\xC7", "\xDE\x68",
  /* U+6E98 */	"\xB7\xC2", "\xDE\x5E", "\x89\xC1", "\xDE\x43",
  /* U+6E9C */	"\xB7\xC8", "\xB7\xBE", "\xDE\x52", "\xDE\x48",
  /* U+6EA0 */	"\xDE\x4B", "\xDE\x63", "\xB7\xB8", "\xDE\x6A",
  /* U+6EA4 */	"\xDE\x62", "\xB7\xC1", "\xDE\x57", "\xB7\xCC",
  /* U+6EA8 */	"\x00\x00", "\x00\x00", "\xB7\xCB", "\xB7\xC5",
  /* U+6EAC */	"\x00\x00", "\x00\x00", "\xDE\x69", "\xB7\xB9",
  /* U+6EB0 */	"\xDE\x55", "\xDE\x4C", "\xDE\x59", "\xDE\x65",
  /* U+6EB4 */	"\xB7\xCD", "\xFD\x68", "\xB7\xBB", "\xDE\x54",
  /* U+6EB8 */	"\x9C\xB7", "\xDE\x4D", "\xB7\xC4", "\x8D\xE3",
  /* U+6EBC */	"\xB7\xC3", "\xDE\x50", "\xDE\x5A", "\xDE\x64",
  /* U+6EC0 */	"\xDE\x47", "\xDE\x51", "\xB7\xBC", "\xDE\x5B",
  /* U+6EC4 */	"\xB7\xC9", "\xB7\xC0", "\xDE\x4E", "\xB7\xBF",
  /* U+6EC8 */	"\xDE\x45", "\xDE\x53", "\xDE\x67", "\xB4\xFE",
  /* U+6ECC */	"\xBA\xB0", "\xDE\x56", "\xE2\x6C", "\xDE\x58",
  /* U+6ED0 */	"\xDE\x66", "\xB7\xC6", "\xDE\x4F", "\xB7\xBA",
  /* U+6ED4 */	"\xB7\xCA", "\xBC\xF0", "\xDE\x44", "\x00\x00",
  /* U+6ED8 */	"\xDE\x5D", "\xFA\xC0", "\x8D\xE5", "\xFA\x64",
  /* U+6EDC */	"\xDE\x5C", "\x89\x47", "\x00\x00", "\x00\x00",
  /* U+6EE0 */	"\x00\x00", "\x00\x00", "\x8D\xE4", "\x00\x00",
  /* U+6EE4 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+6EE8 */	"\x8D\xE7", "\x8D\xE8", "\x00\x00", "\xE2\xAA",
  /* U+6EEC */	"\xBA\xAD", "\xE2\x7D", "\xE2\xA4", "\xBA\xA2",
  /* U+6EF0 */	"\x00\x00", "\xE2\x6E", "\xBA\xAF", "\x00\x00",
  /* U+6EF4 */	"\xBA\x77", "\xE2\x6D", "\xE2\xB0", "\xBA\xB1",
  /* U+6EF8 */	"\xE2\x71", "\xE2\xA3", "\xFD\xC7", "\xE2\x73",
  /* U+6EFC */	"\xE2\xB3", "\xE2\xAF", "\xBA\x75", "\xBA\xA1",
  /* U+6F00 */	"\xE6\x53", "\xBA\xAE", "\xBA\x7D", "\xE2\x6F",
  /* U+6F04 */	"\xFD\xB0", "\xE2\xAE", "\xBA\xA3", "\xE2\xAB",
  /* U+6F08 */	"\xE2\xB8", "\xE2\x75", "\xE2\x7E", "\x94\x45",
  /* U+6F0C */	"\x97\xD6", "\xE2\xB6", "\xE2\xAC", "\xBA\x7C",
  /* U+6F10 */	"\x00\x00", "\x00\x00", "\xE2\x7C", "\xBA\x76",
  /* U+6F14 */	"\xBA\x74", "\xBA\xA8", "\xFC\xC6", "\x98\x44",
  /* U+6F18 */	"\xE2\x7A", "\xE2\x77", "\xE2\x78", "\x00\x00",
  /* U+6F1C */	"\xE2\xB2", "\x00\x00", "\xE2\xB7", "\xE2\xB5",
  /* U+6F20 */	"\xBA\x7A", "\xE2\xB9", "\xBA\x7E", "\xBA\xA7",
  /* U+6F24 */	"\x8D\xE9", "\xE2\x70", "\xE5\xFA", "\xE2\x79",
  /* U+6F28 */	"\x00\x00", "\xBA\x78", "\xBA\xAC", "\xBA\xA9",
  /* U+6F2C */	"\xBA\x7B", "\xE2\xA5", "\xE2\x74", "\xBA\xAA",
  /* U+6F30 */	"\xE2\xA7", "\xBA\xA4", "\xBA\xA6", "\xBA\x73",
  /* U+6F34 */	"\x8D\xEA", "\xE2\xA9", "\xE2\xA1", "\xE2\x72",
  /* U+6F38 */	"\xBA\xA5", "\xE2\xB1", "\xE2\xB4", "\xE2\x7B",
  /* U+6F3C */	"\xE2\xA8", "\xFE\x50", "\xBA\x79", "\xBC\xDF",
  /* U+6F40 */	"\xE2\xA6", "\xE5\xF9", "\x00\x00", "\xE2\xAD",
  /* U+6F44 */	"\xFD\xCC", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+6F48 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+6F4C */	"\x00\x00", "\x00\x00", "\xE2\x76", "\xE6\x44",
  /* U+6F50 */	"\xE6\x4E", "\xBC\xE2", "\xE6\x4D", "\xE6\x59",
  /* U+6F54 */	"\xBC\xE4", "\xE6\x4B", "\x9D\xA7", "\xE6\x4F",
  /* U+6F58 */	"\xBC\xEF", "\x00\x00", "\xE6\x46", "\xBC\xE7",
  /* U+6F5C */	"\xFD\xCD", "\xE6\x52", "\xE9\xF0", "\xBC\xF3",
  /* U+6F60 */	"\xBC\xF2", "\xE6\x54", "\xE6\x43", "\xE6\x5E",
  /* U+6F64 */	"\xBC\xED", "\x00\x00", "\xBC\xE3", "\xE6\x57",
  /* U+6F68 */	"\x00\x00", "\xE6\x5B", "\xE6\x60", "\xE6\x55",
  /* U+6F6C */	"\xE6\x49", "\xBC\xE6", "\xBC\xE9", "\xBC\xF1",
  /* U+6F70 */	"\xBC\xEC", "\x00\x00", "\xE6\x4C", "\xE2\xA2",
  /* U+6F74 */	"\xFD\xCF", "\x00\x00", "\xE6\x48", "\xE6\x5F",
  /* U+6F78 */	"\xBC\xE8", "\x95\xD2", "\xBC\xEB", "\xE6\x61",
  /* U+6F7C */	"\xBC\xE0", "\xE6\x56", "\xE5\xFB", "\xE6\x5C",
  /* U+6F80 */	"\xC0\xDF", "\x8D\xED", "\xE6\x4A", "\x00\x00",
  /* U+6F84 */	"\xBC\xE1", "\xE6\x45", "\xBC\xE5", "\xE5\xFC",
  /* U+6F88 */	"\xBA\xAB", "\xE6\x41", "\xFC\xBA", "\xE6\x5A",
  /* U+6F8C */	"\xE6\x42", "\xE6\x40", "\xBC\xEA", "\x00\x00",
  /* U+6F90 */	"\xE6\x58", "\x00\x00", "\xE5\xFE", "\xE6\x51",
  /* U+6F94 */	"\xE6\x50", "\xE6\x5D", "\xE6\x47", "\xBC\xEE",
  /* U+6F98 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+6F9C */	"\x00\x00", "\xFD\xC5", "\xE9\xF3", "\xFD\xD2",
  /* U+6FA0 */	"\xBF\x49", "\xBE\xFE", "\xEA\x40", "\xE9\xEB",
  /* U+6FA4 */	"\xBF\x41", "\xE9\xF7", "\xBF\x48", "\xBF\x43",
  /* U+6FA8 */	"\xE9\xF5", "\xED\x4F", "\xE9\xFB", "\xEA\x42",
  /* U+6FAC */	"\xE9\xFA", "\xE9\xE9", "\xE9\xF8", "\xEA\x44",
  /* U+6FB0 */	"\xEA\x46", "\xBE\xFD", "\xEA\x45", "\xBF\x44",
  /* U+6FB4 */	"\xBF\x4A", "\x9C\xDC", "\xBF\x47", "\x00\x00",
  /* U+6FB8 */	"\xE9\xFE", "\xBF\x46", "\xE9\xF9", "\x95\xCF",
  /* U+6FBC */	"\xE9\xED", "\xE9\xF2", "\x8D\xEE", "\xE9\xFD",
  /* U+6FC0 */	"\xBF\x45", "\xBF\x42", "\xBE\xFC", "\xBF\x40",
  /* U+6FC4 */	"\xE9\xF1", "\x00\x00", "\xE5\xFD", "\xE9\xEC",
  /* U+6FC8 */	"\xE9\xEF", "\xEA\x41", "\xE9\xF4", "\xE9\xEA",
  /* U+6FCC */	"\xED\x4E", "\xEA\x43", "\xE9\xEE", "\xE9\xFC",
  /* U+6FD0 */	"\x00\x00", "\x00\x00", "\x00\x00", "\xFD\xD4",
  /* U+6FD4 */	"\xED\x51", "\xC0\xE3", "\x00\x00", "\x00\x00",
  /* U+6FD8 */	"\xC0\xD7", "\x96\xEC", "\x96\xEB", "\xC0\xDB",
  /* U+6FDC */	"\xED\x53", "\xED\x59", "\xED\x57", "\xC0\xD9",
  /* U+6FE0 */	"\xC0\xDA", "\xC0\xE1", "\xED\x5A", "\xED\x52",
  /* U+6FE4 */	"\xC0\xDC", "\x00\x00", "\xED\x56", "\xED\x55",
  /* U+6FE8 */	"\xED\x5B", "\xC0\xE2", "\x00\x00", "\xC0\xDD",
  /* U+6FEC */	"\xC0\xE0", "\xED\x54", "\xC0\xE4", "\xC0\xDE",
  /* U+6FF0 */	"\xC0\xE5", "\xC0\xD8", "\xED\x58", "\x00\x00",
  /* U+6FF4 */	"\xED\x50", "\x00\x00", "\x90\xB6", "\xEF\xF7",
  /* U+6FF8 */	"\xFD\xC3", "\x00\x00", "\xC2\x71", "\xEF\xF4",
  /* U+6FFC */	"\xEF\xF6", "\x00\x00", "\xC2\x6F", "\xEF\xF2",
  /* U+7000 */	"\xEF\xF3", "\xEF\xEE", "\x00\x00", "\x98\xAB",
  /* U+7004 */	"\xE9\xF6", "\xEF\xEF", "\xC2\x70", "\xEF\xEB",
  /* U+7008 */	"\x00\x00", "\xC2\x6D", "\xEF\xF8", "\xC2\x6E",
  /* U+700C */	"\xEF\xEC", "\xEF\xED", "\xEF\xF1", "\xC2\x73",
  /* U+7010 */	"\x00\x00", "\xC2\x72", "\x00\x00", "\x00\x00",
  /* U+7014 */	"\xEF\xF0", "\xC3\x78", "\xF2\x5F", "\xF2\x65",
  /* U+7018 */	"\xC3\x79", "\xF2\x5C", "\xC3\x76", "\xC3\x73",
  /* U+701C */	"\xF2\x67", "\xC3\x77", "\x96\xEE", "\xC3\x74",
  /* U+7020 */	"\xF2\x5E", "\xF2\x61", "\xF2\x62", "\xF2\x63",
  /* U+7024 */	"\xF2\x66", "\x00\x00", "\xEF\xF5", "\xF2\x5D",
  /* U+7028 */	"\xC3\x75", "\xF2\x64", "\xF2\x68", "\xF2\x60",
  /* U+702C */	"\x8D\xF4", "\x00\x00", "\x00\x00", "\xF4\x5D",
  /* U+7030 */	"\xC4\x6A", "\xF4\x60", "\xC4\x6B", "\xF4\x68",
  /* U+7034 */	"\xF4\x5F", "\xF4\x5C", "\x00\x00", "\xF4\x5E",
  /* U+7038 */	"\xF4\x62", "\xF4\x65", "\xF4\x64", "\xF4\x67",
  /* U+703C */	"\xF4\x5B", "\x00\x00", "\xC4\x69", "\xF4\x63",
  /* U+7040 */	"\xF4\x66", "\xF4\x69", "\xF4\x61", "\xF5\xD3",
  /* U+7044 */	"\xF5\xD4", "\xF5\xD8", "\xF5\xD9", "\x00\x00",
  /* U+7048 */	"\xF5\xD6", "\xF5\xD7", "\xF5\xD5", "\xFD\xE0",
  /* U+704C */	"\xC4\xE9", "\x8C\x67", "\x00\x00", "\x00\x00",
  /* U+7050 */	"\x8D\xF6", "\xC5\x78", "\xF6\xEB", "\x00\x00",
  /* U+7054 */	"\x8D\xF7", "\xF6\xE8", "\xF6\xE9", "\xF6\xEA",
  /* U+7058 */	"\xC5\x79", "\x00\x00", "\xF7\xE5", "\xF7\xE4",
  /* U+705C */	"\x8F\xFA", "\xF8\xAF", "\xC5\xF4", "\xF8\xAD",
  /* U+7060 */	"\xF8\xB0", "\xF8\xAE", "\xF8\xF5", "\xC6\x57",
  /* U+7064 */	"\xC6\x65", "\xF9\xA3", "\xF9\x6C", "\x97\xD0",
  /* U+7068 */	"\xF9\xA2", "\xF9\xD0", "\xF9\xD1", "\xA4\xF5",
  /* U+706C */	"\x8B\xD2", "\x00\x00", "\x00\x00", "\x8D\xF8",
  /* U+7070 */	"\xA6\xC7", "\xCA\x41", "\x00\x00", "\x00\x00",
  /* U+7074 */	"\xCB\x5E", "\x90\xD9", "\xA8\x5F", "\x8C\x47",
  /* U+7078 */	"\xA8\x62", "\xFA\xF0", "\xCB\x5F", "\x00\x00",
  /* U+707C */	"\xA8\x60", "\xA8\x61", "\xFD\xE1", "\x8D\xF9",
  /* U+7080 */	"\x00\x00", "\xFD\xE3", "\xCD\x58", "\xCD\x5A",
  /* U+7084 */	"\xCD\x55", "\xCD\x52", "\xCD\x54", "\x00\x00",
  /* U+7088 */	"\x00\x00", "\x8D\xFA", "\xAA\xA4", "\xFB\x63",
  /* U+708C */	"\x00\x00", "\x00\x00", "\xAA\xA2", "\x90\xA6",
  /* U+7090 */	"\x00\x00", "\xCD\x56", "\xAA\xA3", "\xCD\x53",
  /* U+7094 */	"\xCD\x50", "\xAA\xA1", "\xCD\x57", "\x00\x00",
  /* U+7098 */	"\xCD\x51", "\xAA\xA5", "\xCD\x59", "\x00\x00",
  /* U+709C */	"\x00\x00", "\x00\x00", "\x00\x00", "\xCF\xAF",
  /* U+70A0 */	"\x99\x70", "\xCF\xB3", "\x00\x00", "\x91\xEB",
  /* U+70A4 */	"\xAC\xB7", "\x97\x70", "\x98\x6F", "\xFD\xE2",
  /* U+70A8 */	"\x00\x00", "\xCF\xB6", "\x00\x00", "\xAC\xAF",
  /* U+70AC */	"\xAC\xB2", "\xAC\xB4", "\xAC\xB6", "\xAC\xB3",
  /* U+70B0 */	"\xCF\xB2", "\xCF\xB1", "\x00\x00", "\xAC\xB1",
  /* U+70B4 */	"\xCF\xB4", "\xCF\xB5", "\x00\x00", "\xCF\xAE",
  /* U+70B8 */	"\xAC\xB5", "\x98\xF2", "\xAC\xB0", "\x9A\xFC",
  /* U+70BC */	"\x89\x6C", "\xFD\xFD", "\xCF\xB0", "\x00\x00",
  /* U+70C0 */	"\x99\x5E", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+70C4 */	"\x95\xBD", "\xD2\x77", "\xD2\x78", "\xD2\x79",
  /* U+70C8 */	"\xAF\x50", "\x00\x00", "\xAF\x4C", "\xD2\x6E",
  /* U+70CC */	"\xFD\xE4", "\xD2\x76", "\xD2\x7B", "\xAF\x51",
  /* U+70D0 */	"\x91\xE6", "\xD2\x6C", "\xD2\x72", "\xD2\x6B",
  /* U+70D4 */	"\xD2\x75", "\xFD\xE5", "\xFD\xE6", "\xD2\x71",
  /* U+70D8 */	"\xAF\x4D", "\xAF\x4F", "\xD2\x7A", "\x00\x00",
  /* U+70DC */	"\xD2\x6A", "\xD2\x6D", "\xD2\x73", "\xFD\xE7",
  /* U+70E0 */	"\xD2\x74", "\xD2\x7C", "\xD2\x70", "\x00\x00",
  /* U+70E4 */	"\xAF\x4E", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+70E8 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+70EC */	"\x00\x00", "\x00\x00", "\x00\x00", "\xB2\x6D",
  /* U+70F0 */	"\xD6\x4E", "\x94\x54", "\x00\x00", "\xD6\x50",
  /* U+70F4 */	"\xD6\x4C", "\x99\xB8", "\xD6\x58", "\xD6\x4A",
  /* U+70F8 */	"\xD6\x57", "\xB2\x69", "\xD6\x48", "\xDA\x5B",
  /* U+70FC */	"\xD6\x52", "\xB2\x6C", "\x97\xE9", "\xD6\x53",
  /* U+7100 */	"\xD6\x56", "\x00\x00", "\xD6\x5A", "\x00\x00",
  /* U+7104 */	"\xD6\x4F", "\x93\x46", "\xD6\x54", "\x00\x00",
  /* U+7108 */	"\x00\x00", "\xB2\x6A", "\xB2\x6B", "\xD6\x59",
  /* U+710C */	"\xD6\x4D", "\xD6\x49", "\xD6\x5B", "\x00\x00",
  /* U+7110 */	"\xD6\x51", "\x00\x00", "\x00\x00", "\xD6\x55",
  /* U+7114 */	"\x00\x00", "\x00\x00", "\x00\x00", "\xD6\x4B",
  /* U+7118 */	"\x00\x00", "\xB5\x48", "\xB5\x49", "\xDA\x65",
  /* U+711C */	"\xB5\x4F", "\x98\x63", "\xDA\x59", "\xDA\x62",
  /* U+7120 */	"\xDA\x58", "\xB5\x4C", "\xDA\x60", "\xDA\x5E",
  /* U+7124 */	"\x00\x00", "\xDA\x5F", "\xB5\x4A", "\x00\x00",
  /* U+7128 */	"\xDA\x63", "\x95\xBC", "\x00\x00", "\xFD\xED",
  /* U+712C */	"\xFD\xF7", "\x00\x00", "\xDA\x5C", "\xDA\x5A",
  /* U+7130 */	"\xB5\x4B", "\xDA\x5D", "\xDA\x61", "\x98\x70",
  /* U+7134 */	"\x96\xF6", "\x8E\xA9", "\xB5\x4D", "\x00\x00",
  /* U+7138 */	"\x00\x00", "\x00\x00", "\xDA\x64", "\x94\x51",
  /* U+713C */	"\x00\x00", "\x00\x00", "\x8E\x43", "\x00\x00",
  /* U+7140 */	"\x8B\x5A", "\xDE\x70", "\xDE\x77", "\xDE\x79",
  /* U+7144 */	"\xDE\xA1", "\xFD\xEE", "\xB7\xDA", "\xDE\x6B",
  /* U+7148 */	"\x00\x00", "\xB7\xD2", "\xFD\xF0", "\xDE\x7A",
  /* U+714C */	"\xB7\xD7", "\xDE\xA2", "\xB7\xCE", "\xFD\xF4",
  /* U+7150 */	"\xDE\x7D", "\x9B\xF5", "\xDE\x6D", "\xDE\x7E",
  /* U+7154 */	"\xDE\x6C", "\x00\x00", "\xB7\xDC", "\x00\x00",
  /* U+7158 */	"\xDE\x78", "\xB7\xCF", "\xDE\xA3", "\x00\x00",
  /* U+715C */	"\xB7\xD4", "\xDE\x71", "\xB7\xD9", "\xDE\x7C",
  /* U+7160 */	"\xDE\x6F", "\xDE\x76", "\xDE\x72", "\xDE\x6E",
  /* U+7164 */	"\xB7\xD1", "\xB7\xD8", "\xB7\xD6", "\xB7\xD3",
  /* U+7168 */	"\xB7\xDB", "\xB7\xD0", "\xDE\x75", "\x97\x7E",
  /* U+716C */	"\xB7\xD5", "\x00\x00", "\xB5\x4E", "\x00\x00",
  /* U+7170 */	"\xDE\x7B", "\x9B\xD5", "\xDE\x73", "\x9A\xC3",
  /* U+7174 */	"\x00\x00", "\x97\xC8", "\xA0\xDB", "\x91\xD0",
  /* U+7178 */	"\xDE\x74", "\x00\x00", "\x9F\xE4", "\xE2\xC1",
  /* U+717C */	"\x8F\xDD", "\xBA\xB4", "\x91\xE9", "\x00\x00",
  /* U+7180 */	"\xE2\xBD", "\xE2\xC3", "\xE2\xBF", "\x00\x00",
  /* U+7184 */	"\xBA\xB6", "\xE2\xBE", "\xE2\xC2", "\xE2\xBA",
  /* U+7188 */	"\x98\xE0", "\xE2\xBC", "\xBA\xB5", "\x00\x00",
  /* U+718C */	"\x92\xCA", "\x00\x00", "\x98\x57", "\xE2\xC0",
  /* U+7190 */	"\xE2\xBB", "\x8C\x51", "\xBA\xB7", "\x00\x00",
  /* U+7194 */	"\xBA\xB2", "\x00\x00", "\xFD\xEB", "\xE2\xC4",
  /* U+7198 */	"\x9B\x49", "\xBA\xB3", "\xE6\x67", "\xE6\x64",
  /* U+719C */	"\xE6\x70", "\xE6\x6A", "\xE6\x6C", "\xBC\xF4",
  /* U+71A0 */	"\xE6\x66", "\xE6\x6E", "\x9D\x76", "\x9E\xAF",
  /* U+71A4 */	"\xE6\x6D", "\xE6\x6B", "\x00\x00", "\xE6\x71",
  /* U+71A8 */	"\xBC\xF7", "\xE6\x68", "\xE6\x6F", "\x00\x00",
  /* U+71AC */	"\xBC\xF5", "\x9C\xCC", "\x00\x00", "\xE6\x63",
  /* U+71B0 */	"\xE6\x65", "\xBC\xF6", "\xE6\x62", "\xE6\x72",
  /* U+71B4 */	"\xFD\xEA", "\xE6\x69", "\x00\x00", "\x8D\xF1",
  /* U+71B8 */	"\xEA\x4A", "\xBF\x51", "\xFD\xFB", "\x00\x00",
  /* U+71BC */	"\xEA\x55", "\xEA\x53", "\xBF\x4B", "\xEA\x49",
  /* U+71C0 */	"\xEA\x4C", "\xEA\x4D", "\xEA\x48", "\xBF\x55",
  /* U+71C4 */	"\xBF\x56", "\xEA\x47", "\xEA\x56", "\xEA\x51",
  /* U+71C8 */	"\xBF\x4F", "\xBF\x4C", "\xEA\x50", "\xEA\x4E",
  /* U+71CC */	"\x00\x00", "\x00\x00", "\xBF\x52", "\xEA\x52",
  /* U+71D0 */	"\xBF\x4D", "\x8E\x53", "\xBF\x4E", "\x00\x00",
  /* U+71D4 */	"\xEA\x4F", "\xBF\x50", "\xEA\x4B", "\x00\x00",
  /* U+71D8 */	"\xEA\x54", "\xBF\x53", "\xEA\x57", "\xEA\x58",
  /* U+71DC */	"\xBF\x54", "\xFA\xCF", "\x00\x00", "\xC0\xE7",
  /* U+71E0 */	"\xC0\xEE", "\xED\x5C", "\xED\x62", "\x00\x00",
  /* U+71E4 */	"\xED\x60", "\xC0\xEA", "\xC0\xE9", "\xC0\xE6",
  /* U+71E8 */	"\xED\x5E", "\x00\x00", "\x00\x00", "\x96\xF9",
  /* U+71EC */	"\xC0\xEC", "\xC0\xEB", "\xC0\xE8", "\x00\x00",
  /* U+71F0 */	"\xED\x61", "\xED\x5D", "\xED\x5F", "\x00\x00",
  /* U+71F4 */	"\xC0\xED", "\x98\xBF", "\x9E\x49", "\x00\x00",
  /* U+71F8 */	"\xC2\x77", "\xEF\xFB", "\x00\x00", "\xC2\x74",
  /* U+71FC */	"\xC2\x75", "\xEF\xFD", "\xC2\x76", "\xEF\xFA",
  /* U+7200 */	"\x8C\xA7", "\xEF\xF9", "\xF2\x6C", "\xEF\xFC",
  /* U+7204 */	"\x00\x00", "\xF2\x6D", "\xC3\x7A", "\xF2\x6B",
  /* U+7208 */	"\x00\x00", "\x9B\xCA", "\xF2\x6A", "\x00\x00",
  /* U+720C */	"\xF2\x69", "\xC3\x7B", "\xFD\xFE", "\x92\xDC",
  /* U+7210 */	"\xC4\x6C", "\x00\x00", "\x00\x00", "\xF4\x6A",
  /* U+7214 */	"\xF4\x6B", "\xFE\x41", "\x91\xCC", "\x91\xE2",
  /* U+7218 */	"\x00\x00", "\xF5\xDC", "\xF5\xDB", "\xC4\xEA",
  /* U+721C */	"\x00\x00", "\xF5\xDA", "\xF6\xEC", "\xF6\xED",
  /* U+7220 */	"\x00\x00", "\x00\x00", "\xF7\xE6", "\xF8\xB1",
  /* U+7224 */	"\xFE\x44", "\x00\x00", "\xF8\xF6", "\xF9\xBC",
  /* U+7228 */	"\xC6\x79", "\xF9\xC6", "\xA4\xF6", "\x8B\xD3",
  /* U+722C */	"\xAA\xA6", "\xAA\xA7", "\xFE\x47", "\x00\x00",
  /* U+7230 */	"\xAC\xB8", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+7234 */	"\x00\x00", "\xC0\xEF", "\xA4\xF7", "\x00\x00",
  /* U+7238 */	"\xAA\xA8", "\xAF\x52", "\xB7\xDD", "\xA4\xF8",
  /* U+723C */	"\x00\x00", "\xB2\x6E", "\xBA\xB8", "\xC9\x62",
  /* U+7240 */	"\xFE\x48", "\xCF\xB7", "\xD2\x7D", "\x00\x00",
  /* U+7244 */	"\xE2\xC5", "\x00\x00", "\xC0\xF0", "\xA4\xF9",
  /* U+7248 */	"\xAA\xA9", "\xCF\xB8", "\xCF\xB9", "\xDA\x66",
  /* U+724C */	"\xB5\x50", "\x00\x00", "\x00\x00", "\xDE\xA4",
  /* U+7250 */	"\x94\x55", "\x00\x00", "\xB7\xDE", "\xE2\xC6",
  /* U+7254 */	"\x00\x00", "\xFE\x4B", "\xBC\xF8", "\xFE\x4C",
  /* U+7258 */	"\xC3\x7C", "\xA4\xFA", "\xDA\x67", "\xA4\xFB",
  /* U+725C */	"\x8D\xBF", "\xA6\xC9", "\xCA\x42", "\xA6\xC8",
  /* U+7260 */	"\xA8\x65", "\xA8\x64", "\xA8\x63", "\xCB\x60",
  /* U+7264 */	"\x00\x00", "\x00\x00", "\x9E\x78", "\xAA\xAA",
  /* U+7268 */	"\x00\x00", "\xAA\xAB", "\xCD\x5B", "\x00\x00",
  /* U+726C */	"\xCF\xBA", "\x00\x00", "\xCF\xBD", "\xAC\xBA",
  /* U+7270 */	"\xCF\xBB", "\x00\x00", "\xAC\xB9", "\xCF\xBC",
  /* U+7274 */	"\xAC\xBB", "\x00\x00", "\xD2\xA2", "\xD2\xA1",
  /* U+7278 */	"\xD2\x7E", "\xAF\x53", "\x00\x00", "\xD6\x5D",
  /* U+727C */	"\xD6\x5E", "\xB2\x6F", "\xD6\x5C", "\xD6\x5F",
  /* U+7280 */	"\xB5\x52", "\xB2\x70", "\xFE\x51", "\x00\x00",
  /* U+7284 */	"\xB5\x51", "\xDA\x6B", "\xDA\x6A", "\x94\x56",
  /* U+7288 */	"\xDA\x68", "\xDA\x69", "\x00\x00", "\xDA\x6C",
  /* U+728C */	"\xDE\xA6", "\xDE\xA5", "\xDE\xA9", "\x9D\x61",
  /* U+7290 */	"\xDE\xA8", "\xDE\xA7", "\xBA\xB9", "\xE2\xC9",
  /* U+7294 */	"\x94\x57", "\xE2\xC8", "\xBA\xBA", "\xE2\xC7",
  /* U+7298 */	"\xE6\x73", "\x00\x00", "\xE6\x74", "\xBC\xF9",
  /* U+729C */	"\x00\x00", "\xEA\x59", "\xEA\x5A", "\x99\x66",
  /* U+72A0 */	"\x00\x00", "\xF2\x72", "\xC3\x7D", "\xF2\x71",
  /* U+72A4 */	"\xF2\x70", "\xF2\x6E", "\xF2\x6F", "\xC4\xEB",
  /* U+72A8 */	"\xF4\x6C", "\xF6\xEE", "\xF8\xF7", "\x00\x00",
  /* U+72AC */	"\xA4\xFC", "\x8B\xD5", "\xC9\xA5", "\xA5\xC7",
  /* U+72B0 */	"\xC9\xA6", "\x00\x00", "\xA0\x69", "\x00\x00",
  /* U+72B4 */	"\xCA\x43", "\xCA\x44", "\x00\x00", "\x00\x00",
  /* U+72B8 */	"\x00\x00", "\x00\x00", "\xCB\x66", "\x00\x00",
  /* U+72BC */	"\x00\x00", "\xCB\x62", "\x00\x00", "\xCB\x61",
  /* U+72C0 */	"\xAA\xAC", "\xCB\x65", "\xA8\x67", "\xCB\x63",
  /* U+72C4 */	"\xA8\x66", "\xCB\x67", "\xCB\x64", "\x00\x00",
  /* U+72C8 */	"\x00\x00", "\xCD\x5F", "\xCF\xBE", "\xCD\x5D",
  /* U+72CC */	"\xCD\x64", "\x98\xB4", "\xAA\xAD", "\x00\x00",
  /* U+72D0 */	"\xAA\xB0", "\xCD\x65", "\xCD\x61", "\x00\x00",
  /* U+72D4 */	"\xCD\x62", "\x00\x00", "\xCD\x5C", "\xAA\xAF",
  /* U+72D8 */	"\xCD\x5E", "\xAA\xAE", "\xCD\x63", "\x00\x00",
  /* U+72DC */	"\xCD\x60", "\x00\x00", "\x00\x00", "\xCF\xC2",
  /* U+72E0 */	"\xAC\xBD", "\xAC\xBE", "\xA0\x49", "\xCF\xC5",
  /* U+72E4 */	"\xCF\xBF", "\x00\x00", "\xCF\xC4", "\x00\x00",
  /* U+72E8 */	"\xCF\xC0", "\xAC\xBC", "\xCF\xC3", "\xCF\xC1",
  /* U+72EC */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+72F0 */	"\x00\x00", "\x00\x00", "\x00\x00", "\xD2\xA8",
  /* U+72F4 */	"\xD2\xA5", "\x00\x00", "\xD2\xA7", "\xAF\x58",
  /* U+72F8 */	"\xAF\x57", "\xAF\x55", "\xD2\xA4", "\xD2\xA9",
  /* U+72FC */	"\xAF\x54", "\xAF\x56", "\xD2\xA6", "\xD6\x67",
  /* U+7300 */	"\xD2\xA3", "\xD2\xAA", "\xA0\x4C", "\x00\x00",
  /* U+7304 */	"\x9E\x65", "\x00\x00", "\x00\x00", "\xD6\x62",
  /* U+7308 */	"\xD6\x66", "\x00\x00", "\xD6\x65", "\xDA\x6E",
  /* U+730C */	"\xDA\x79", "\x00\x00", "\x00\x00", "\xD6\x68",
  /* U+7310 */	"\x98\xB5", "\xD6\x63", "\xDA\x6D", "\xB2\x74",
  /* U+7314 */	"\x00\x00", "\x00\x00", "\xB2\x73", "\xD6\x61",
  /* U+7318 */	"\xD6\x64", "\xB2\x75", "\x00\x00", "\xB2\x72",
  /* U+731C */	"\xB2\x71", "\xD6\x60", "\xD6\x69", "\x00\x00",
  /* U+7320 */	"\x00\x00", "\x00\x00", "\xDA\x70", "\xDA\x77",
  /* U+7324 */	"\x00\x00", "\xB5\x54", "\xDA\x76", "\xDA\x73",
  /* U+7328 */	"\xFE\x58", "\xB5\x56", "\x99\x75", "\xFE\x53",
  /* U+732C */	"\xA0\x65", "\xDA\x75", "\xFE\x59", "\x00\x00",
  /* U+7330 */	"\xDA\x6F", "\xDA\x71", "\xDA\x74", "\xDA\x72",
  /* U+7334 */	"\xB5\x55", "\xDA\x78", "\xB5\x53", "\xB7\xDF",
  /* U+7338 */	"\x98\xB7", "\x98\xB8", "\xDE\xAD", "\xDE\xAC",
  /* U+733C */	"\xDE\xAA", "\x00\x00", "\xB7\xE2", "\xB7\xE1",
  /* U+7340 */	"\xDE\xAE", "\x98\xBA", "\xDE\xAB", "\xE2\xCA",
  /* U+7344 */	"\xBA\xBB", "\xB7\xE0", "\x00\x00", "\x00\x00",
  /* U+7348 */	"\x98\xBB", "\xDE\xB0", "\xDE\xAF", "\x00\x00",
  /* U+734C */	"\xE2\xCD", "\xE2\xCB", "\xBC\xFA", "\x9F\xBC",
  /* U+7350 */	"\xBA\xBC", "\xE2\xCC", "\xE6\x76", "\x00\x00",
  /* U+7354 */	"\x00\x00", "\x00\x00", "\x00\x00", "\xBC\xFB",
  /* U+7358 */	"\xE6\x75", "\xE6\x7E", "\xE6\x7D", "\xE6\x7B",
  /* U+735C */	"\x00\x00", "\xE6\x7A", "\xE6\x77", "\xE6\x78",
  /* U+7360 */	"\xE6\x79", "\xE6\x7C", "\xE6\xA1", "\x00\x00",
  /* U+7364 */	"\x00\x00", "\xEA\x5F", "\xEA\x5C", "\xEA\x5D",
  /* U+7368 */	"\xBF\x57", "\xEA\x5B", "\xEA\x61", "\xEA\x60",
  /* U+736C */	"\xEA\x5E", "\x00\x00", "\xED\x64", "\xED\x65",
  /* U+7370 */	"\xC0\xF1", "\xA0\x4A", "\xC0\xF2", "\xED\x63",
  /* U+7374 */	"\x9E\xC7", "\xC2\x79", "\xEF\xFE", "\xC2\x78",
  /* U+7378 */	"\xC3\x7E", "\x00\x00", "\xC3\xA1", "\xC4\x6D",
  /* U+737C */	"\xF4\x6E", "\xF4\x6D", "\xF5\xDD", "\xF6\xEF",
  /* U+7380 */	"\xC5\x7A", "\xF7\xE8", "\xF7\xE7", "\xF7\xE9",
  /* U+7384 */	"\xA5\xC8", "\xCF\xC6", "\xAF\x59", "\xB2\x76",
  /* U+7388 */	"\xD6\x6A", "\xA5\xC9", "\xC9\xA7", "\xA4\xFD",
  /* U+738C */	"\x8C\xA9", "\x00\x00", "\xCA\x45", "\x98\xAE",
  /* U+7390 */	"\x00\x00", "\x00\x00", "\xCB\x6C", "\xCB\x6A",
  /* U+7394 */	"\xCB\x6B", "\xCB\x68", "\xA8\x68", "\xCB\x69",
  /* U+7398 */	"\x92\xD6", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+739C */	"\xFA\xE1", "\xCD\x6D", "\x91\xD4", "\xAA\xB3",
  /* U+73A0 */	"\xCD\x6B", "\xCD\x67", "\xCD\x6A", "\x00\x00",
  /* U+73A4 */	"\xCD\x66", "\xAA\xB5", "\xCD\x69", "\xFA\xDE",
  /* U+73A8 */	"\xAA\xB2", "\xAA\xB1", "\xFE\x5B", "\xAA\xB4",
  /* U+73AC */	"\xCD\x6C", "\xCD\x68", "\x00\x00", "\x00\x00",
  /* U+73B0 */	"\x00\x00", "\x00\x00", "\xAC\xC2", "\xAC\xC5",
  /* U+73B4 */	"\xCF\xCE", "\xCF\xCD", "\xCF\xCC", "\xAC\xBF",
  /* U+73B8 */	"\xCF\xD5", "\xCF\xCB", "\x8C\x53", "\xAC\xC1",
  /* U+73BC */	"\xD2\xAF", "\x00\x00", "\xCF\xD2", "\xCF\xD0",
  /* U+73C0 */	"\xAC\xC4", "\x00\x00", "\xCF\xC8", "\xCF\xD3",
  /* U+73C4 */	"\x00\x00", "\xCF\xCA", "\xCF\xD4", "\xCF\xD1",
  /* U+73C8 */	"\xCF\xC9", "\xFE\x5E", "\xAC\xC0", "\xCF\xD6",
  /* U+73CC */	"\xCF\xC7", "\xAC\xC3", "\xFB\xD7", "\xFE\x5A",
  /* U+73D0 */	"\x94\xC5", "\x00\x00", "\xD2\xB4", "\xD2\xAB",
  /* U+73D4 */	"\xD2\xB6", "\xFA\xCA", "\xD2\xAE", "\xD2\xB9",
  /* U+73D8 */	"\xD2\xBA", "\xD2\xAC", "\xD2\xB8", "\xD2\xB5",
  /* U+73DC */	"\xD2\xB3", "\xD2\xB7", "\xAF\x5F", "\x00\x00",
  /* U+73E0 */	"\xAF\x5D", "\x98\xC1", "\x97\x5C", "\xD2\xB1",
  /* U+73E4 */	"\xFE\x74", "\xD2\xAD", "\x97\x73", "\xD2\xB0",
  /* U+73E8 */	"\xD2\xBB", "\xD2\xB2", "\xAF\x5E", "\xCF\xCF",
  /* U+73EC */	"\x00\x00", "\xAF\x5A", "\xAF\x5C", "\xFA\x46",
  /* U+73F0 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x97\x64",
  /* U+73F4 */	"\xD6\x78", "\xD6\x6D", "\xD6\x6B", "\xFE\x68",
  /* U+73F8 */	"\xD6\x6C", "\x96\x4E", "\xD6\x73", "\x97\x65",
  /* U+73FC */	"\xD6\x74", "\xD6\x70", "\xB2\x7B", "\xD6\x75",
  /* U+7400 */	"\xD6\x72", "\xD6\x6F", "\x8C\x5A", "\xB2\x79",
  /* U+7404 */	"\xD6\x6E", "\xB2\x77", "\xB2\x7A", "\xD6\x71",
  /* U+7408 */	"\xD6\x79", "\xAF\x5B", "\xB2\x78", "\xD6\x77",
  /* U+740C */	"\xD6\x76", "\xB2\x7C", "\x00\x00", "\x00\x00",
  /* U+7410 */	"\x00\x00", "\x89\xA1", "\x95\xFA", "\x00\x00",
  /* U+7414 */	"\x92\xD4", "\xFE\x69", "\xDA\x7E", "\xFB\x45",
  /* U+7418 */	"\x00\x00", "\x98\xC8", "\xDA\xA1", "\xB5\x60",
  /* U+741C */	"\x90\xEF", "\xDA\xA7", "\x98\xC9", "\x98\xCA",
  /* U+7420 */	"\xDA\xA9", "\xDA\xA2", "\xB5\x5A", "\xDA\xA6",
  /* U+7424 */	"\xDA\xA5", "\xB5\x5B", "\xB5\x61", "\x00\x00",
  /* U+7428 */	"\xB5\x62", "\xDA\xA8", "\xB5\x58", "\xDA\x7D",
  /* U+742C */	"\xDA\x7B", "\xDA\xA3", "\xDA\x7A", "\xB5\x5F",
  /* U+7430 */	"\xDA\x7C", "\xDA\xA4", "\xDA\xAA", "\xB5\x59",
  /* U+7434 */	"\xB5\x5E", "\xB5\x5C", "\xB5\x5D", "\x94\x6D",
  /* U+7438 */	"\x94\xB7", "\xFE\x6C", "\xB5\x57", "\x00\x00",
  /* U+743C */	"\x94\x6B", "\x00\x00", "\x00\x00", "\xB7\xE9",
  /* U+7440 */	"\xDE\xB7", "\xB7\xE8", "\xDE\xBB", "\x92\xFC",
  /* U+7444 */	"\xDE\xB1", "\x95\xEB", "\xDE\xBC", "\xFE\x73",
  /* U+7448 */	"\x97\x6E", "\xFE\x5F", "\xDE\xB2", "\xDE\xB3",
  /* U+744C */	"\x00\x00", "\xDE\xBD", "\xDE\xBA", "\xDE\xB8",
  /* U+7450 */	"\xDE\xB9", "\xDE\xB5", "\xDE\xB4", "\xFD\xBD",
  /* U+7454 */	"\xDE\xBE", "\xB7\xE5", "\x92\xD5", "\xDE\xB6",
  /* U+7458 */	"\x00\x00", "\xB7\xEA", "\xB7\xE4", "\xB7\xEB",
  /* U+745C */	"\xB7\xEC", "\xFE\xB9", "\xB7\xE7", "\xB7\xE6",
  /* U+7460 */	"\xFE\x71", "\x00\x00", "\xE2\xCE", "\xBA\xBE",
  /* U+7464 */	"\xBA\xBD", "\xFB\xBB", "\x00\x00", "\xE2\xD3",
  /* U+7468 */	"\x94\x7A", "\xBC\xFC", "\xBA\xBF", "\x95\xFB",
  /* U+746C */	"\xFE\x77", "\xBA\xC1", "\xE2\xD4", "\xB7\xE3",
  /* U+7470 */	"\xBA\xC0", "\xE2\xD0", "\xE2\xD2", "\xE2\xCF",
  /* U+7474 */	"\xFE\x79", "\xE2\xD1", "\xFE\x75", "\x00\x00",
  /* U+7478 */	"\x00\x00", "\xE6\xAB", "\x94\x5D", "\x00\x00",
  /* U+747C */	"\xE6\xAA", "\xE6\xA7", "\xBD\x40", "\xEA\x62",
  /* U+7480 */	"\xBD\x41", "\xE6\xA6", "\xFE\x7C", "\xBC\xFE",
  /* U+7484 */	"\x00\x00", "\xE6\xA8", "\xE6\xA5", "\xE6\xA2",
  /* U+7488 */	"\xE6\xA9", "\xE6\xA3", "\xE6\xA4", "\xBC\xFD",
  /* U+748C */	"\x93\x44", "\x8E\xA6", "\x00\x00", "\x00\x00",
  /* U+7490 */	"\xED\x69", "\x00\x00", "\xEA\x66", "\x00\x00",
  /* U+7494 */	"\xEA\x65", "\xEA\x67", "\x00\x00", "\xED\x66",
  /* U+7498 */	"\xBF\x5A", "\x92\xD3", "\xEA\x63", "\x94\xB8",
  /* U+749C */	"\xBF\x58", "\x00\x00", "\xBF\x5C", "\xBF\x5B",
  /* U+74A0 */	"\xEA\x64", "\xEA\x68", "\x00\x00", "\xBF\x59",
  /* U+74A4 */	"\xFC\x71", "\xED\x6D", "\xC0\xF5", "\xC2\x7A",
  /* U+74A8 */	"\xC0\xF6", "\xC0\xF3", "\xED\x6A", "\xED\x68",
  /* U+74AC */	"\x00\x00", "\xED\x6B", "\x00\x00", "\xED\x6E",
  /* U+74B0 */	"\xC0\xF4", "\xED\x6C", "\xED\x67", "\x00\x00",
  /* U+74B4 */	"\x97\x5E", "\xF0\x42", "\xF0\x45", "\xF2\x75",
  /* U+74B8 */	"\xF0\x40", "\x8C\xAD", "\xF4\x6F", "\xF0\x46",
  /* U+74BC */	"\x00\x00", "\xC3\xA2", "\xF0\x44", "\xC2\x7B",
  /* U+74C0 */	"\xF0\x41", "\xF0\x43", "\xF0\x47", "\xF2\x76",
  /* U+74C4 */	"\x00\x00", "\xF2\x74", "\x00\x00", "\x00\x00",
  /* U+74C8 */	"\xFE\xA7", "\x00\x00", "\xC3\xA3", "\xF2\x73",
  /* U+74CC */	"\x94\x6A", "\x00\x00", "\x00\x00", "\xC4\x6E",
  /* U+74D0 */	"\x93\xE3", "\x00\x00", "\x00\x00", "\x98\xCF",
  /* U+74D4 */	"\xC4\xED", "\xF6\xF1", "\xC4\xEC", "\xF6\xF3",
  /* U+74D8 */	"\xF6\xF0", "\xF6\xF2", "\xC5\xD0", "\xF8\xB2",
  /* U+74DC */	"\xA5\xCA", "\xCD\x6E", "\xD2\xBC", "\xD2\xBD",
  /* U+74E0 */	"\xB2\x7D", "\xDE\xBF", "\xBF\x5D", "\xC3\xA4",
  /* U+74E4 */	"\xC5\x7B", "\xF8\xB3", "\xA5\xCB", "\xA0\xD9",
  /* U+74E8 */	"\xCD\x6F", "\xA2\x60", "\x00\x00", "\x00\x00",
  /* U+74EC */	"\xCF\xD7", "\x00\x00", "\xCF\xD8", "\x00\x00",
  /* U+74F0 */	"\xA0\xBF", "\xA0\x4D", "\xA0\xB8", "\x00\x00",
  /* U+74F4 */	"\xD2\xBE", "\xD2\xBF", "\xB2\x7E", "\xB2\xA1",
  /* U+74F8 */	"\xA0\xCE", "\x00\x00", "\x00\x00", "\xDA\xAB",
  /* U+74FC */	"\x00\x00", "\xDE\xC2", "\xDE\xC1", "\xDE\xC0",
  /* U+7500 */	"\xE2\xD5", "\x00\x00", "\xE2\xD6", "\xE2\xD7",
  /* U+7504 */	"\xBA\xC2", "\xA0\xB7", "\x00\x00", "\xE6\xAD",
  /* U+7508 */	"\xE6\xAC", "\x00\x00", "\x00\x00", "\xEA\x69",
  /* U+750C */	"\xBF\x5E", "\xBF\x5F", "\xFE\xA9", "\xED\x72",
  /* U+7510 */	"\xED\x6F", "\xED\x70", "\xED\x71", "\xF0\x49",
  /* U+7514 */	"\xF0\x48", "\xC2\x7C", "\xF2\x77", "\xF5\xDE",
  /* U+7518 */	"\xA5\xCC", "\x89\xC3", "\xAC\xC6", "\x00\x00",
  /* U+751C */	"\xB2\xA2", "\xDE\xC3", "\xFE\xAB", "\xA5\xCD",
  /* U+7520 */	"\x00\x00", "\xD2\xC0", "\xB2\xA3", "\x00\x00",
  /* U+7524 */	"\x00\x00", "\xB5\x63", "\xB5\x64", "\x00\x00",
  /* U+7528 */	"\xA5\xCE", "\xA5\xCF", "\xCA\x46", "\xA8\x6A",
  /* U+752C */	"\xA8\x69", "\xAC\xC7", "\xCF\xD9", "\xDA\xAC",
  /* U+7530 */	"\xA5\xD0", "\xA5\xD1", "\xA5\xD2", "\xA5\xD3",
  /* U+7534 */	"\x9D\xF4", "\x89\x6D", "\x00\x00", "\xA8\x6B",
  /* U+7538 */	"\xA8\x6C", "\xCB\x6E", "\xCB\x6D", "\x9C\x7B",
  /* U+753C */	"\x00\x00", "\xAA\xB6", "\xCD\x72", "\xCD\x70",
  /* U+7540 */	"\xCD\x71", "\x00\x00", "\x98\xD2", "\x00\x00",
  /* U+7544 */	"\x00\x00", "\x00\x00", "\x9F\xA9", "\xCF\xDA",
  /* U+7548 */	"\xCF\xDB", "\x00\x00", "\xFE\xB2", "\xAC\xCB",
  /* U+754C */	"\xAC\xC9", "\xFE\xB1", "\xAC\xCA", "\xAC\xC8",
  /* U+7550 */	"\x00\x00", "\x97\xD9", "\x00\x00", "\xA0\xC4",
  /* U+7554 */	"\xAF\x60", "\x94\x76", "\x00\x00", "\x00\x00",
  /* U+7558 */	"\x00\x00", "\xAF\x64", "\xAF\x63", "\xD2\xC1",
  /* U+755C */	"\xAF\x62", "\xAF\x61", "\x00\x00", "\xD2\xC2",
  /* U+7560 */	"\x99\x78", "\x00\x00", "\xB2\xA6", "\xD6\x7B",
  /* U+7564 */	"\xD6\x7A", "\xB2\xA4", "\xB2\xA5", "\xFE\xB3",
  /* U+7568 */	"\x00\x00", "\x00\x00", "\xB5\x66", "\xB5\x65",
  /* U+756C */	"\xDA\xAE", "\x98\xD3", "\xFE\xB4", "\xDA\xAD",
  /* U+7570 */	"\xB2\xA7", "\x00\x00", "\x98\xD4", "\x00\x00",
  /* U+7574 */	"\x00\x00", "\x00\x00", "\xB7\xED", "\xDE\xC5",
  /* U+7578 */	"\xB7\xEE", "\xDE\xC4", "\x9F\xB9", "\x00\x00",
  /* U+757C */	"\x00\x00", "\xE2\xD8", "\xE6\xAE", "\xBD\x42",
  /* U+7580 */	"\xEA\x6A", "\x00\x00", "\x00\x00", "\x94\x71",
  /* U+7584 */	"\xED\x73", "\x00\x00", "\xC3\xA6", "\xC3\xA5",
  /* U+7588 */	"\x00\x00", "\x00\x00", "\xC5\x7C", "\xA5\xD4",
  /* U+758C */	"\xCD\x73", "\x98\xD5", "\xFE\xB8", "\xB2\xA8",
  /* U+7590 */	"\xE2\xD9", "\xBA\xC3", "\xC6\xD4", "\x00\x00",
  /* U+7594 */	"\xCB\x6F", "\xCB\x70", "\x00\x00", "\x00\x00",
  /* U+7598 */	"\xCD\x74", "\xAA\xB8", "\xAA\xB9", "\x00\x00",
  /* U+759C */	"\x00\x00", "\xAA\xB7", "\xFE\xBA", "\x00\x00",
  /* U+75A0 */	"\x00\x00", "\x00\x00", "\xAC\xCF", "\xAC\xD0",
  /* U+75A4 */	"\xAC\xCD", "\xAC\xCE", "\x00\x00", "\xCF\xDC",
  /* U+75A8 */	"\x00\x00", "\x00\x00", "\xCF\xDD", "\xAC\xCC",
  /* U+75AC */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+75B0 */	"\xD2\xC3", "\x9E\x5C", "\xAF\x68", "\xAF\x69",
  /* U+75B4 */	"\xFE\xBB", "\xB2\xAB", "\xD2\xC9", "\x00\x00",
  /* U+75B8 */	"\xAF\x6E", "\xAF\x6C", "\xD2\xCA", "\xD2\xC5",
  /* U+75BC */	"\xAF\x6B", "\xAF\x6A", "\xAF\x65", "\xD2\xC8",
  /* U+75C0 */	"\xD2\xC7", "\xD2\xC4", "\xAF\x6D", "\xA0\x44",
  /* U+75C4 */	"\xD2\xC6", "\xAF\x66", "\x00\x00", "\xAF\x67",
  /* U+75C8 */	"\x98\xD7", "\x00\x00", "\xB2\xAC", "\xD6\xA1",
  /* U+75CC */	"\xD6\xA2", "\xB2\xAD", "\xD6\x7C", "\xD6\x7E",
  /* U+75D0 */	"\xD6\xA4", "\xD6\xA3", "\xD6\x7D", "\x00\x00",
  /* U+75D4 */	"\xB2\xA9", "\xB2\xAA", "\x00\x00", "\xDA\xB6",
  /* U+75D8 */	"\xB5\x6B", "\xB5\x6A", "\xDA\xB0", "\xB5\x68",
  /* U+75DC */	"\x98\xD8", "\xDA\xB3", "\xB5\x6C", "\xDA\xB4",
  /* U+75E0 */	"\xB5\x6D", "\xDA\xB1", "\xB5\x67", "\xB5\x69",
  /* U+75E4 */	"\xDA\xB5", "\x00\x00", "\xDA\xB2", "\xDA\xAF",
  /* U+75E8 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+75EC */	"\x00\x00", "\xDE\xD2", "\x00\x00", "\xDE\xC7",
  /* U+75F0 */	"\xB7\xF0", "\xB7\xF3", "\xB7\xF2", "\xB7\xF7",
  /* U+75F4 */	"\xB7\xF6", "\xDE\xD3", "\xDE\xD1", "\xDE\xCA",
  /* U+75F8 */	"\xDE\xCE", "\xDE\xCD", "\xB7\xF4", "\xDE\xD0",
  /* U+75FC */	"\xDE\xCC", "\xDE\xD4", "\xDE\xCB", "\xB7\xF5",
  /* U+7600 */	"\xB7\xEF", "\xB7\xF1", "\xFE\xBC", "\xDE\xC9",
  /* U+7604 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x9F\xFE",
  /* U+7608 */	"\xE2\xDB", "\xBA\xC7", "\xE2\xDF", "\xBA\xC6",
  /* U+760C */	"\xE2\xDC", "\xBA\xC5", "\x00\x00", "\xDE\xC8",
  /* U+7610 */	"\xDE\xCF", "\xE2\xDE", "\x00\x00", "\xBA\xC8",
  /* U+7614 */	"\xE2\xE0", "\xE2\xDD", "\xE2\xDA", "\x00\x00",
  /* U+7618 */	"\x00\x00", "\xE6\xB1", "\xE6\xB5", "\xE6\xB7",
  /* U+761C */	"\xE6\xB3", "\xE6\xB2", "\xE6\xB0", "\xBD\x45",
  /* U+7620 */	"\xBD\x43", "\xBD\x48", "\xBD\x49", "\xE6\xB4",
  /* U+7624 */	"\xBD\x46", "\xE6\xAF", "\xBD\x47", "\xBA\xC4",
  /* U+7628 */	"\xE6\xB6", "\xBD\x44", "\x00\x00", "\x00\x00",
  /* U+762C */	"\xFE\xBD", "\xEA\x6C", "\x00\x00", "\xEA\x6B",
  /* U+7630 */	"\xEA\x73", "\xEA\x6D", "\xEA\x72", "\xEA\x6F",
  /* U+7634 */	"\xBF\x60", "\xEA\x71", "\x00\x00", "\x00\x00",
  /* U+7638 */	"\xBF\x61", "\x00\x00", "\xBF\x62", "\x9D\xDD",
  /* U+763C */	"\xEA\x70", "\xEA\x6E", "\x00\x00", "\x00\x00",
  /* U+7640 */	"\x9E\xE1", "\x00\x00", "\xC0\xF8", "\xED\x74",
  /* U+7644 */	"\x00\x00", "\x00\x00", "\xC0\xF7", "\xED\x77",
  /* U+7648 */	"\xED\x75", "\xED\x76", "\x00\x00", "\x00\x00",
  /* U+764C */	"\xC0\xF9", "\x98\xDA", "\x9D\xDF", "\xFE\xBF",
  /* U+7650 */	"\xF0\x4D", "\xFE\xBE", "\xC2\xA1", "\xF0\x4E",
  /* U+7654 */	"\x9E\xEB", "\x00\x00", "\xC2\x7D", "\xF0\x4F",
  /* U+7658 */	"\xC2\x7E", "\xF0\x4C", "\xF0\x50", "\x00\x00",
  /* U+765C */	"\xF0\x4A", "\x00\x00", "\x00\x00", "\xC3\xA7",
  /* U+7660 */	"\xF2\x78", "\xC3\xA8", "\xC4\x6F", "\x00\x00",
  /* U+7664 */	"\xF0\x4B", "\xC4\x70", "\x9E\x59", "\xA0\x5C",
  /* U+7668 */	"\x00\x00", "\xC4\xEE", "\xF5\xDF", "\x00\x00",
  /* U+766C */	"\xC5\x7E", "\xF6\xF4", "\xC5\x7D", "\xFE\xC0",
  /* U+7670 */	"\xF7\xEA", "\xC5\xF5", "\xC5\xF6", "\x94\x77",
  /* U+7674 */	"\x98\xDC", "\xF9\xCC", "\xFE\xC1", "\x00\x00",
  /* U+7678 */	"\xAC\xD1", "\xCF\xDE", "\x98\xDE", "\xB5\x6E",
  /* U+767C */	"\xB5\x6F", "\xA5\xD5", "\xA6\xCA", "\xCA\x47",
  /* U+7680 */	"\x00\x00", "\xCB\x71", "\xA8\x6D", "\x00\x00",
  /* U+7684 */	"\xAA\xBA", "\x00\x00", "\xAC\xD2", "\xAC\xD3",
  /* U+7688 */	"\xAC\xD4", "\xD6\xA6", "\xD2\xCB", "\xAF\x6F",
  /* U+768C */	"\x00\x00", "\x00\x00", "\xB2\xAE", "\xD6\xA5",
  /* U+7690 */	"\xFE\xC3", "\x00\x00", "\xDA\xB8", "\xB5\x71",
  /* U+7694 */	"\x00\x00", "\xDA\xB7", "\xB5\x70", "\x00\x00",
  /* U+7698 */	"\x00\x00", "\xDE\xD5", "\xBD\x4A", "\xE6\xBB",
  /* U+769C */	"\xE6\xB8", "\xE6\xB9", "\xE6\xBA", "\x00\x00",
  /* U+76A0 */	"\x00\x00", "\xFE\xC8", "\x00\x00", "\x00\x00",
  /* U+76A4 */	"\xED\x78", "\xFE\xC9", "\xF0\x51", "\x00\x00",
  /* U+76A8 */	"\x00\x00", "\x00\x00", "\xF4\x71", "\xF4\x70",
  /* U+76AC */	"\x00\x00", "\xF6\xF5", "\xA5\xD6", "\xCD\x75",
  /* U+76B0 */	"\xAF\x70", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+76B4 */	"\xB5\x72", "\xDE\xD6", "\x00\x00", "\xFE\xCA",
  /* U+76B8 */	"\xE2\xE1", "\x00\x00", "\xBD\x4B", "\xEA\x74",
  /* U+76BC */	"\x00\x00", "\xF0\x52", "\xF4\x72", "\xA5\xD7",
  /* U+76C0 */	"\x00\x00", "\x00\x00", "\xAA\xBB", "\xAC\xD7",
  /* U+76C4 */	"\xCF\xDF", "\xAC\xD8", "\xAC\xD6", "\x00\x00",
  /* U+76C8 */	"\xAC\xD5", "\xD2\xCC", "\xAF\x71", "\x00\x00",
  /* U+76CC */	"\xFE\xCB", "\xAF\x72", "\xAF\x73", "\x00\x00",
  /* U+76D0 */	"\x00\x00", "\x00\x00", "\xB2\xB0", "\xD6\xA7",
  /* U+76D4 */	"\xB2\xAF", "\x00\x00", "\x9F\xC2", "\x00\x00",
  /* U+76D8 */	"\x00\x00", "\x8C\x6B", "\xDA\xB9", "\xB2\xB1",
  /* U+76DC */	"\xB5\x73", "\xDE\xD7", "\xB7\xF8", "\xB7\xF9",
  /* U+76E0 */	"\x00\x00", "\xBA\xC9", "\x00\x00", "\xBA\xCA",
  /* U+76E4 */	"\xBD\x4C", "\xBF\x64", "\xEA\x75", "\xBF\x63",
  /* U+76E8 */	"\x00\x00", "\xED\x79", "\xC0\xFA", "\x00\x00",
  /* U+76EC */	"\xF0\x53", "\xF4\x73", "\xA5\xD8", "\xA8\x6E",
  /* U+76F0 */	"\xCD\x78", "\xCD\x77", "\xAA\xBC", "\xCD\x76",
  /* U+76F4 */	"\xAA\xBD", "\xCD\x79", "\x00\x00", "\xCF\xE5",
  /* U+76F8 */	"\xAC\xDB", "\xAC\xDA", "\xCF\xE7", "\xCF\xE6",
  /* U+76FC */	"\xAC\xDF", "\x00\x00", "\xAC\xDE", "\x00\x00",
  /* U+7700 */	"\x00\x00", "\xAC\xD9", "\x00\x00", "\xCF\xE1",
  /* U+7704 */	"\xCF\xE2", "\xCF\xE3", "\x00\x00", "\xAC\xE0",
  /* U+7708 */	"\xCF\xE0", "\xAC\xDC", "\xCF\xE4", "\xAC\xDD",
  /* U+770C */	"\x98\xC4", "\x00\x00", "\x94\xB0", "\x94\xB1",
  /* U+7710 */	"\xD2\xCF", "\xD2\xD3", "\xD2\xD1", "\xD2\xD0",
  /* U+7714 */	"\x00\x00", "\xD2\xD4", "\x00\x00", "\x00\x00",
  /* U+7718 */	"\x00\x00", "\xD2\xD5", "\xD2\xD6", "\xD2\xCE",
  /* U+771C */	"\x00\x00", "\xD2\xCD", "\xFE\xD1", "\xAF\x75",
  /* U+7720 */	"\xAF\x76", "\x00\x00", "\xD2\xD7", "\xD2\xD2",
  /* U+7724 */	"\xA0\xC1", "\xD6\xB0", "\xFE\xD2", "\xD2\xD8",
  /* U+7728 */	"\xAF\x77", "\xAF\x74", "\x00\x00", "\xA0\xCD",
  /* U+772C */	"\x00\x00", "\xD6\xAA", "\x00\x00", "\xD6\xA9",
  /* U+7730 */	"\x00\x00", "\xD6\xAB", "\xD6\xAC", "\xD6\xAE",
  /* U+7734 */	"\xD6\xAD", "\xD6\xB2", "\xB2\xB5", "\xB2\xB2",
  /* U+7738 */	"\xB2\xB6", "\xD6\xA8", "\xB2\xB7", "\xD6\xB1",
  /* U+773C */	"\xB2\xB4", "\xD6\xAF", "\xB2\xB3", "\x00\x00",
  /* U+7740 */	"\xFE\xD3", "\x00\x00", "\x00\x00", "\x98\xE5",
  /* U+7744 */	"\xDA\xBC", "\xDA\xBE", "\xDA\xBA", "\xDA\xBB",
  /* U+7748 */	"\x00\x00", "\x00\x00", "\xDA\xBF", "\xDA\xC1",
  /* U+774C */	"\xDA\xC2", "\xDA\xBD", "\xDA\xC0", "\xB5\x74",
  /* U+7750 */	"\x00\x00", "\x00\x00", "\xDE\xDB", "\x00\x00",
  /* U+7754 */	"\xDE\xE0", "\xDE\xD8", "\xDE\xDC", "\x00\x00",
  /* U+7758 */	"\xFE\xD6", "\xDE\xE1", "\xDE\xDD", "\xB7\xFA",
  /* U+775C */	"\xB8\x43", "\x00\x00", "\xB7\xFD", "\xDE\xD9",
  /* U+7760 */	"\xDE\xDA", "\xBA\xCE", "\xB8\x46", "\xB7\xFE",
  /* U+7764 */	"\x00\x00", "\xB8\x44", "\xB7\xFC", "\xDE\xDF",
  /* U+7768 */	"\xB8\x45", "\xDE\xDE", "\xB8\x41", "\xB7\xFB",
  /* U+776C */	"\xB8\x42", "\xDE\xE2", "\xE2\xE6", "\xE2\xE8",
  /* U+7770 */	"\x00\x00", "\x00\x00", "\x91\xE4", "\x00\x00",
  /* U+7774 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x8F\xC7",
  /* U+7778 */	"\x94\xAE", "\xB8\x40", "\x8A\x4F", "\x94\xB2",
  /* U+777C */	"\xE2\xE3", "\xBA\xCC", "\xE2\xE9", "\xBA\xCD",
  /* U+7780 */	"\xE2\xE7", "\xE2\xE2", "\xE2\xE5", "\xE2\xEA",
  /* U+7784 */	"\xBA\xCB", "\xE2\xE4", "\x00\x00", "\xBD\x4E",
  /* U+7788 */	"\xE6\xBF", "\xE6\xBE", "\x00\x00", "\xBD\x51",
  /* U+778C */	"\xBD\x4F", "\xE6\xBC", "\xBD\x4D", "\xE6\xBD",
  /* U+7790 */	"\x00\x00", "\xBD\x50", "\x00\x00", "\x8F\xD4",
  /* U+7794 */	"\x00\x00", "\xEA\x7D", "\x00\x00", "\xEA\xA1",
  /* U+7798 */	"\x98\xEA", "\xEA\x7E", "\xEA\x76", "\xEA\x7A",
  /* U+779C */	"\xEA\x79", "\xEA\x77", "\xBF\x66", "\xBF\x67",
  /* U+77A0 */	"\xBF\x65", "\xEA\x78", "\xEA\x7B", "\xEA\x7C",
  /* U+77A4 */	"\x00\x00", "\xBF\x68", "\x00\x00", "\xC1\x40",
  /* U+77A8 */	"\xED\xA3", "\x00\x00", "\xC0\xFC", "\xED\x7B",
  /* U+77AC */	"\xC0\xFE", "\xC1\x41", "\x00\x00", "\xFE\xD8",
  /* U+77B0 */	"\xC0\xFD", "\xED\xA2", "\xED\x7C", "\xC0\xFB",
  /* U+77B4 */	"\xED\xA1", "\xED\x7A", "\xED\x7E", "\xED\x7D",
  /* U+77B8 */	"\x00\x00", "\x9D\xE0", "\xF0\x55", "\xC2\xA4",
  /* U+77BC */	"\xC2\xA5", "\xC2\xA2", "\x98\xEE", "\xC2\xA3",
  /* U+77C0 */	"\x00\x00", "\x00\x00", "\xF0\x54", "\x95\xC4",
  /* U+77C4 */	"\xF2\x7B", "\xFC\xE8", "\x00\x00", "\xC3\xA9",
  /* U+77C8 */	"\x00\x00", "\xF2\x79", "\xF2\x7A", "\x98\xEF",
  /* U+77CC */	"\xF4\x74", "\xF4\x77", "\xF4\x75", "\xF4\x76",
  /* U+77D0 */	"\xF5\xE0", "\x00\x00", "\x00\x00", "\xC4\xEF",
  /* U+77D4 */	"\xF7\xEB", "\xF8\xB4", "\x00\x00", "\xC5\xF7",
  /* U+77D8 */	"\xF8\xF8", "\xF8\xF9", "\xC6\x66", "\xA5\xD9",
  /* U+77DC */	"\xAC\xE1", "\x8C\x6E", "\xDA\xC3", "\x00\x00",
  /* U+77E0 */	"\xDE\xE3", "\x00\x00", "\xA5\xDA", "\xA8\x6F",
  /* U+77E4 */	"\x00\x00", "\xAA\xBE", "\xFA\xD8", "\xCF\xE8",
  /* U+77E8 */	"\xCF\xE9", "\xAF\x78", "\x00\x00", "\x00\x00",
  /* U+77EC */	"\xDA\xC4", "\xB5\x75", "\xB8\x47", "\xC1\x42",
  /* U+77F0 */	"\xED\xA4", "\xF2\x7C", "\xF4\x78", "\xA5\xDB",
  /* U+77F4 */	"\xFE\xDC", "\x00\x00", "\x00\x00", "\xCD\xA1",
  /* U+77F8 */	"\xCD\x7A", "\xCD\x7C", "\xCD\x7E", "\xCD\x7D",
  /* U+77FC */	"\xCD\x7B", "\xAA\xBF", "\xA0\xAE", "\x00\x00",
  /* U+7800 */	"\x00\x00", "\x00\x00", "\xAC\xE2", "\xCF\xF2",
  /* U+7804 */	"\x00\x00", "\xCF\xED", "\xCF\xEA", "\x00\x00",
  /* U+7808 */	"\x9D\x4C", "\xCF\xF1", "\x00\x00", "\x00\x00",
  /* U+780C */	"\xAC\xE4", "\xAC\xE5", "\xCF\xF0", "\xCF\xEF",
  /* U+7810 */	"\xCF\xEE", "\xCF\xEB", "\xCF\xEC", "\xCF\xF3",
  /* U+7814 */	"\xAC\xE3", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+7818 */	"\x98\xF1", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+781C */	"\x98\xF3", "\xAF\x7C", "\x94\xC1", "\xAF\xA4",
  /* U+7820 */	"\xAF\xA3", "\xD2\xE1", "\xD2\xDB", "\xD2\xD9",
  /* U+7824 */	"\x00\x00", "\xAF\xA1", "\xD6\xB9", "\xAF\x7A",
  /* U+7828 */	"\xD2\xDE", "\xD2\xE2", "\xD2\xE4", "\xD2\xE0",
  /* U+782C */	"\xD2\xDA", "\xAF\xA2", "\xD2\xDF", "\xD2\xDD",
  /* U+7830 */	"\xAF\x79", "\xD2\xE5", "\xAF\xA5", "\xD2\xE3",
  /* U+7834 */	"\xAF\x7D", "\xD2\xDC", "\x00\x00", "\xAF\x7E",
  /* U+7838 */	"\xAF\x7B", "\x98\xF5", "\x00\x00", "\x00\x00",
  /* U+783C */	"\xFA\x4F", "\x96\xE2", "\x00\x00", "\x00\x00",
  /* U+7840 */	"\x00\x00", "\x00\x00", "\x94\x50", "\xB2\xB9",
  /* U+7844 */	"\x96\xA2", "\xD6\xBA", "\x00\x00", "\x98\xF6",
  /* U+7848 */	"\xD6\xB3", "\xD6\xB5", "\xD6\xB7", "\x96\xE5",
  /* U+784C */	"\xD6\xB8", "\xD6\xB6", "\xB2\xBA", "\x00\x00",
  /* U+7850 */	"\xD6\xBB", "\x98\xF7", "\xD6\xB4", "\xA0\x46",
  /* U+7854 */	"\x96\xE3", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+7858 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+785C */	"\xDA\xC8", "\xB5\x76", "\xDA\xD0", "\x00\x00",
  /* U+7860 */	"\xDA\xC5", "\x00\x00", "\xDA\xD1", "\x00\x00",
  /* U+7864 */	"\xDA\xC6", "\xDA\xC7", "\x98\xF8", "\x00\x00",
  /* U+7868 */	"\xDA\xCF", "\xDA\xCE", "\xDA\xCB", "\xB2\xB8",
  /* U+786C */	"\xB5\x77", "\xDA\xC9", "\xDA\xCC", "\xB5\x78",
  /* U+7870 */	"\xDA\xCD", "\xDA\xCA", "\x00\x00", "\x00\x00",
  /* U+7874 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+7878 */	"\x00\x00", "\xDE\xEE", "\x9E\xE4", "\xDE\xF2",
  /* U+787C */	"\xB8\x4E", "\x00\x00", "\xE2\xF0", "\xB8\x51",
  /* U+7880 */	"\xDE\xF0", "\xF9\xD6", "\x00\x00", "\xDE\xED",
  /* U+7884 */	"\xDE\xE8", "\xDE\xEA", "\xDE\xEB", "\xDE\xE4",
  /* U+7888 */	"\x94\xC3", "\xB8\x4D", "\x00\x00", "\x00\x00",
  /* U+788C */	"\xB8\x4C", "\x94\xC2", "\xB8\x48", "\xDE\xE7",
  /* U+7890 */	"\x00\x00", "\xB8\x4F", "\x00\x00", "\xB8\x50",
  /* U+7894 */	"\xDE\xE6", "\xDE\xE9", "\xDE\xF1", "\xB8\x4A",
  /* U+7898 */	"\xB8\x4B", "\xDE\xEF", "\xDE\xE5", "\x00\x00",
  /* U+789C */	"\x00\x00", "\x00\x00", "\xE2\xF2", "\xBA\xD0",
  /* U+78A0 */	"\xE2\xF4", "\xDE\xEC", "\xE2\xF6", "\xBA\xD4",
  /* U+78A4 */	"\xE2\xF7", "\xE2\xF3", "\x00\x00", "\xBA\xD1",
  /* U+78A8 */	"\xE2\xEF", "\xBA\xD3", "\xE2\xEC", "\xE2\xF1",
  /* U+78AC */	"\xE2\xF5", "\xE2\xEE", "\x00\x00", "\xFE\xE1",
  /* U+78B0 */	"\xB8\x49", "\xFE\xE9", "\xE2\xEB", "\xBA\xD2",
  /* U+78B4 */	"\xE2\xED", "\x00\x00", "\x96\xE4", "\x00\x00",
  /* U+78B8 */	"\x89\xAC", "\x96\xDB", "\xBD\x54", "\xE6\xC1",
  /* U+78BC */	"\xBD\x58", "\x00\x00", "\xBD\x56", "\x00\x00",
  /* U+78C0 */	"\x00\x00", "\xBA\xCF", "\x00\x00", "\xE6\xC8",
  /* U+78C4 */	"\xE6\xC9", "\xBD\x53", "\x00\x00", "\xFE\xE2",
  /* U+78C8 */	"\xE6\xC7", "\xE6\xCA", "\xBD\x55", "\xBD\x52",
  /* U+78CC */	"\xE6\xC3", "\xE6\xC0", "\xE6\xC5", "\xE6\xC2",
  /* U+78D0 */	"\xBD\x59", "\xE6\xC4", "\x94\xC4", "\xFE\xE3",
  /* U+78D4 */	"\xE6\xC6", "\xBD\x57", "\x00\x00", "\xFE\xE7",
  /* U+78D8 */	"\x9F\xFB", "\x00\x00", "\xBF\x6A", "\xEA\xA8",
  /* U+78DC */	"\x00\x00", "\xEA\xA2", "\xEA\xA6", "\xEA\xAC",
  /* U+78E0 */	"\xEA\xAD", "\xEA\xA9", "\xEA\xAA", "\xEA\xA7",
  /* U+78E4 */	"\x8C\x59", "\xEA\xA4", "\x00\x00", "\xBF\x6C",
  /* U+78E8 */	"\xBF\x69", "\xEA\xA3", "\xEA\xA5", "\x00\x00",
  /* U+78EC */	"\xBF\x6B", "\xEA\xAB", "\x93\xC9", "\xC1\x46",
  /* U+78F0 */	"\x94\xE8", "\xFB\x56", "\xED\xAA", "\xED\xA5",
  /* U+78F4 */	"\xC1\x45", "\x90\xC5", "\x00\x00", "\xC1\x43",
  /* U+78F8 */	"\x00\x00", "\xED\xAC", "\xC1\x44", "\xED\xA8",
  /* U+78FC */	"\xED\xA9", "\xED\xA6", "\xED\xAD", "\xF0\x56",
  /* U+7900 */	"\x00\x00", "\xC1\x47", "\xED\xA7", "\x00\x00",
  /* U+7904 */	"\xED\xAE", "\xED\xAB", "\xA0\xA8", "\x00\x00",
  /* U+7908 */	"\x00\x00", "\xF0\x5A", "\x00\x00", "\x00\x00",
  /* U+790C */	"\xF0\x57", "\x00\x00", "\xC2\xA6", "\x00\x00",
  /* U+7910 */	"\xF0\x5B", "\xF0\x5D", "\xF0\x5C", "\xF0\x58",
  /* U+7914 */	"\xF0\x59", "\x00\x00", "\x00\x00", "\xF2\xA3",
  /* U+7918 */	"\x00\x00", "\xC3\xAA", "\x00\x00", "\xF2\x7E",
  /* U+791C */	"\xF2\xA2", "\xF2\x7D", "\xF2\xA4", "\x00\x00",
  /* U+7920 */	"\x00\x00", "\xF2\xA1", "\x00\x00", "\xF4\x7A",
  /* U+7924 */	"\xF4\x7D", "\xF4\x79", "\xC4\x71", "\xF4\x7B",
  /* U+7928 */	"\xF4\x7C", "\xF4\x7E", "\xC4\x72", "\xC4\x74",
  /* U+792C */	"\xC4\x73", "\xF5\xE1", "\xFE\xE5", "\xF5\xE3",
  /* U+7930 */	"\x00\x00", "\xF5\xE2", "\x98\xFD", "\x98\xFB",
  /* U+7934 */	"\xFE\xE8", "\xF6\xF6", "\x8E\xBF", "\x00\x00",
  /* U+7938 */	"\xF8\xB5", "\xF8\xFA", "\xA5\xDC", "\x8B\xD8",
  /* U+793C */	"\xFE\xF7", "\xCB\x72", "\xAA\xC0", "\xCD\xA3",
  /* U+7940 */	"\xAA\xC1", "\xAA\xC2", "\xCD\xA2", "\x00\x00",
  /* U+7944 */	"\xCF\xF8", "\xCF\xF7", "\xAC\xE6", "\xAC\xE9",
  /* U+7948 */	"\xAC\xE8", "\xAC\xE7", "\xCF\xF4", "\xCF\xF6",
  /* U+794C */	"\xCF\xF5", "\x00\x00", "\x00\x00", "\xD2\xE8",
  /* U+7950 */	"\xAF\xA7", "\xD2\xEC", "\xD2\xEB", "\xD2\xEA",
  /* U+7954 */	"\xD2\xE6", "\xAF\xA6", "\xAF\xAA", "\xAF\xAD",
  /* U+7958 */	"\x8F\x68", "\x94\xC6", "\xAF\xAE", "\xD2\xE7",
  /* U+795C */	"\xD2\xE9", "\xAF\xAC", "\xAF\xAB", "\xAF\xA9",
  /* U+7960 */	"\xAF\xA8", "\xD6\xC2", "\x9D\xEA", "\xD6\xC0",
  /* U+7964 */	"\xD6\xBC", "\xB2\xBB", "\x00\x00", "\xD6\xBD",
  /* U+7968 */	"\xB2\xBC", "\xD6\xBE", "\xD6\xBF", "\xD6\xC1",
  /* U+796C */	"\x00\x00", "\xB2\xBD", "\x00\x00", "\x00\x00",
  /* U+7970 */	"\xDA\xD5", "\xFC\x69", "\xDA\xD4", "\xDA\xD3",
  /* U+7974 */	"\xDA\xD2", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+7978 */	"\x00\x00", "\xDE\xF6", "\xB8\x52", "\x00\x00",
  /* U+797C */	"\xDE\xF3", "\xDE\xF5", "\x9C\xDA", "\xB8\x53",
  /* U+7980 */	"\xFE\xF3", "\xB8\x54", "\xDE\xF4", "\x9C\x72",
  /* U+7984 */	"\x00\x00", "\x00\x00", "\xFE\xF0", "\x89\xC9",
  /* U+7988 */	"\xE3\x41", "\x00\x00", "\xE2\xF9", "\xE2\xFA",
  /* U+798C */	"\x00\x00", "\xBA\xD7", "\xBA\xD5", "\xBA\xD6",
  /* U+7990 */	"\xE3\x43", "\x99\x41", "\xE3\x42", "\xE2\xFE",
  /* U+7994 */	"\xE2\xFD", "\xE2\xFC", "\xE2\xFB", "\xE3\x40",
  /* U+7998 */	"\xE2\xF8", "\x99\x42", "\xE6\xCB", "\xE6\xD0",
  /* U+799C */	"\xE6\xCE", "\xFE\xF5", "\x00\x00", "\x91\xD7",
  /* U+79A0 */	"\xE6\xCD", "\xE6\xCC", "\xE6\xCF", "\x00\x00",
  /* U+79A4 */	"\xEA\xAE", "\x94\xCC", "\xBF\x6D", "\xC1\x48",
  /* U+79A8 */	"\xED\xB0", "\xFE\xF8", "\xC1\x49", "\xED\xAF",
  /* U+79AC */	"\xF0\x5F", "\xF0\x5E", "\xC2\xA7", "\x00\x00",
  /* U+79B0 */	"\xF2\xA5", "\xC3\xAB", "\xF4\xA1", "\xC5\xA1",
  /* U+79B4 */	"\xF6\xF7", "\x00\x00", "\xF8\xB7", "\xF8\xB6",
  /* U+79B8 */	"\xC9\xA8", "\xAC\xEA", "\xAC\xEB", "\xD6\xC3",
  /* U+79BC */	"\x00\x00", "\xB8\x56", "\xA5\xDD", "\xA8\x72",
  /* U+79C0 */	"\xA8\x71", "\xA8\x70", "\x00\x00", "\x00\x00",
  /* U+79C4 */	"\x97\xA8", "\xCD\xA4", "\xFE\xFC", "\x00\x00",
  /* U+79C8 */	"\xAA\xC4", "\xAA\xC3", "\x00\x00", "\xAC\xEE",
  /* U+79CC */	"\xFD\xBF", "\xCF\xFA", "\xCF\xFD", "\xCF\xFB",
  /* U+79D0 */	"\x00\x00", "\xAC\xEC", "\xAC\xED", "\x00\x00",
  /* U+79D4 */	"\xFE\xFE", "\xCF\xF9", "\xCF\xFC", "\x00\x00",
  /* U+79D8 */	"\xAF\xB5", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+79DC */	"\xD2\xF3", "\xD2\xF5", "\xD2\xF4", "\xAF\xB2",
  /* U+79E0 */	"\xD2\xEF", "\x00\x00", "\x96\xD1", "\xAF\xB0",
  /* U+79E4 */	"\xAF\xAF", "\x00\x00", "\xAF\xB3", "\xAF\xB1",
  /* U+79E8 */	"\x00\x00", "\xAF\xB4", "\xD2\xF2", "\xD2\xED",
  /* U+79EC */	"\xD2\xEE", "\xD2\xF1", "\xD2\xF0", "\x00\x00",
  /* U+79F0 */	"\x00\x00", "\x94\xD5", "\x00\x00", "\x00\x00",
  /* U+79F4 */	"\x94\xD0", "\x00\x00", "\xD6\xC6", "\xD6\xC7",
  /* U+79F8 */	"\xD6\xC5", "\x00\x00", "\xD6\xC4", "\xB2\xBE",
  /* U+79FC */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+7A00 */	"\xB5\x7D", "\x00\x00", "\xDA\xD6", "\xDA\xD8",
  /* U+7A04 */	"\xDA\xDA", "\xB5\x7C", "\x99\x44", "\x00\x00",
  /* U+7A08 */	"\xB5\x7A", "\x00\x00", "\xDA\xD7", "\xB5\x7B",
  /* U+7A0C */	"\xDA\xD9", "\xB5\x79", "\x00\x00", "\x00\x00",
  /* U+7A10 */	"\xDF\x41", "\xDE\xF7", "\xDE\xFA", "\xDE\xFE",
  /* U+7A14 */	"\xB8\x5A", "\xDE\xFC", "\x00\x00", "\xDE\xFB",
  /* U+7A18 */	"\xDE\xF8", "\xDE\xF9", "\xB8\x58", "\xDF\x40",
  /* U+7A1C */	"\xB8\x57", "\x00\x00", "\xB8\x5C", "\xB8\x5B",
  /* U+7A20 */	"\xB8\x59", "\x00\x00", "\xDE\xFD", "\x00\x00",
  /* U+7A24 */	"\x00\x00", "\x00\x00", "\xE3\x49", "\x00\x00",
  /* U+7A28 */	"\xE3\x48", "\x00\x00", "\x8C\x63", "\xE3\x44",
  /* U+7A2C */	"\x00\x00", "\xA0\xB3", "\xBA\xD8", "\xE3\x47",
  /* U+7A30 */	"\xE3\x46", "\xBA\xD9", "\x00\x00", "\x00\x00",
  /* U+7A34 */	"\x00\x00", "\x00\x00", "\x00\x00", "\xBD\x5E",
  /* U+7A38 */	"\x00\x00", "\xE6\xD2", "\x94\xCF", "\xBD\x5F",
  /* U+7A3C */	"\xBD\x5B", "\xBD\x5D", "\x9F\xFA", "\xBD\x5A",
  /* U+7A40 */	"\xBD\x5C", "\x00\x00", "\x00\x00", "\x91\xE5",
  /* U+7A44 */	"\xEA\xAF", "\x9C\x6A", "\xBF\x70", "\xEA\xB1",
  /* U+7A48 */	"\xEA\xB0", "\x8E\x49", "\xE3\x45", "\xBF\x72",
  /* U+7A4C */	"\xBF\x71", "\xBF\x6E", "\xBF\x6F", "\x00\x00",
  /* U+7A50 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+7A54 */	"\xED\xB5", "\x00\x00", "\xED\xB3", "\xC1\x4A",
  /* U+7A58 */	"\xED\xB4", "\x00\x00", "\xED\xB6", "\xED\xB2",
  /* U+7A5C */	"\xED\xB1", "\x00\x00", "\x00\x00", "\xF0\x60",
  /* U+7A60 */	"\xC2\xAA", "\xC2\xA8", "\xC2\xA9", "\x00\x00",
  /* U+7A64 */	"\x00\x00", "\x8E\x4C", "\x00\x00", "\xF2\xA6",
  /* U+7A68 */	"\xF2\xA7", "\xC3\xAD", "\x00\x00", "\xC3\xAC",
  /* U+7A6C */	"\xF4\xA3", "\xF4\xA4", "\xF4\xA2", "\x00\x00",
  /* U+7A70 */	"\xF6\xF8", "\xF6\xF9", "\x00\x00", "\x00\x00",
  /* U+7A74 */	"\xA5\xDE", "\xCA\x48", "\xA8\x73", "\x00\x00",
  /* U+7A78 */	"\xCD\xA5", "\xAA\xC6", "\xAA\xC5", "\xCD\xA6",
  /* U+7A7C */	"\x00\x00", "\x8E\x4D", "\xD0\x40", "\xAC\xEF",
  /* U+7A80 */	"\xCF\xFE", "\xAC\xF0", "\x00\x00", "\x9A\x73",
  /* U+7A84 */	"\xAF\xB6", "\xD2\xF8", "\xD2\xF6", "\xD2\xFC",
  /* U+7A88 */	"\xAF\xB7", "\xD2\xF7", "\xD2\xFB", "\xD2\xF9",
  /* U+7A8C */	"\xD2\xFA", "\x00\x00", "\x00\x00", "\xD6\xC8",
  /* U+7A90 */	"\xD6\xCA", "\x99\x47", "\xB2\xBF", "\x8C\xB1",
  /* U+7A94 */	"\xD6\xC9", "\xB2\xC0", "\xB5\xA2", "\xB5\xA1",
  /* U+7A98 */	"\xB5\x7E", "\xDA\xDB", "\x00\x00", "\x00\x00",
  /* U+7A9C */	"\x00\x00", "\x00\x00", "\xDF\x44", "\xB8\x5D",
  /* U+7AA0 */	"\xB8\x5E", "\x00\x00", "\xDF\x43", "\xDF\x42",
  /* U+7AA4 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+7AA8 */	"\xE3\x4A", "\xBA\xDB", "\xBA\xDA", "\xE3\x4B",
  /* U+7AAC */	"\xE3\x4C", "\x00\x00", "\xBD\x61", "\xBD\x60",
  /* U+7AB0 */	"\x8E\x50", "\xEA\xB5", "\xE6\xD3", "\xE6\xD5",
  /* U+7AB4 */	"\xE6\xD4", "\xEA\xB4", "\xEA\xB2", "\xEA\xB6",
  /* U+7AB8 */	"\xEA\xB3", "\x00\x00", "\xBF\x73", "\x8E\x4F",
  /* U+7ABC */	"\x99\x49", "\x00\x00", "\xED\xB7", "\xC1\x4B",
  /* U+7AC0 */	"\xED\xB8", "\xED\xB9", "\x8E\x51", "\x8E\x52",
  /* U+7AC4 */	"\xC2\xAB", "\xC2\xAC", "\x00\x00", "\xC4\x75",
  /* U+7AC8 */	"\x9A\xB2", "\x89\xA5", "\xC5\xD1", "\xA5\xDF",
  /* U+7ACC */	"\x00\x00", "\x00\x00", "\x00\x00", "\x99\x4C",
  /* U+7AD0 */	"\x00\x00", "\xD0\x41", "\x00\x00", "\x9F\xF8",
  /* U+7AD4 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+7AD8 */	"\xD2\xFD", "\xAF\xB8", "\x8E\x56", "\x99\x4D",
  /* U+7ADC */	"\x91\xCA", "\x8E\x57", "\x00\x00", "\xB3\xBA",
  /* U+7AE0 */	"\xB3\xB9", "\x00\x00", "\x94\xE1", "\xB5\xA4",
  /* U+7AE4 */	"\xDA\xDD", "\xB5\xA3", "\xDA\xDC", "\x90\x47",
  /* U+7AE8 */	"\x00\x00", "\x8F\xD8", "\x8E\x58", "\xDF\x45",
  /* U+7AEC */	"\x00\x00", "\xBA\xDC", "\xE3\x4D", "\xBA\xDD",
  /* U+7AF0 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+7AF4 */	"\x00\x00", "\x00\x00", "\xC4\x76", "\xF4\xA5",
  /* U+7AF8 */	"\x00\x00", "\xA6\xCB", "\xAA\xC7", "\xCD\xA7",
  /* U+7AFC */	"\x00\x00", "\xAC\xF2", "\x94\xEB", "\xAC\xF1",
  /* U+7B00 */	"\xD0\x42", "\xD0\x43", "\x00\x00", "\x00\x00",
  /* U+7B04 */	"\xD3\x40", "\xD3\x42", "\xAF\xB9", "\x00\x00",
  /* U+7B08 */	"\xD3\x44", "\xD3\x47", "\xD3\x45", "\x8E\x5C",
  /* U+7B0C */	"\x95\x53", "\x00\x00", "\xD3\x46", "\xD3\x43",
  /* U+7B10 */	"\xD2\xFE", "\xAF\xBA", "\xD3\x48", "\xD3\x41",
  /* U+7B14 */	"\x9F\xE5", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+7B18 */	"\xD6\xD3", "\xB2\xC6", "\xD6\xDC", "\xB2\xC3",
  /* U+7B1C */	"\x00\x00", "\xD6\xD5", "\xB2\xC7", "\x9F\x56",
  /* U+7B20 */	"\xB2\xC1", "\x00\x00", "\xD6\xD0", "\xD6\xDD",
  /* U+7B24 */	"\xD6\xD1", "\xD6\xCE", "\xB2\xC5", "\x95\x4F",
  /* U+7B28 */	"\xB2\xC2", "\x8E\x5E", "\xD6\xD4", "\xD6\xD7",
  /* U+7B2C */	"\xB2\xC4", "\xD6\xD8", "\xB2\xC8", "\xD6\xD9",
  /* U+7B30 */	"\xD6\xCF", "\xD6\xD6", "\xD6\xDA", "\xD6\xD2",
  /* U+7B34 */	"\xD6\xCD", "\xD6\xCB", "\x00\x00", "\x00\x00",
  /* U+7B38 */	"\xD6\xDB", "\x99\x6A", "\x00\x00", "\xDA\xDF",
  /* U+7B3C */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+7B40 */	"\xDA\xE4", "\x00\x00", "\x9C\x64", "\x9C\xD9",
  /* U+7B44 */	"\xDA\xE0", "\xDA\xE6", "\xB5\xA7", "\xD6\xCC",
  /* U+7B48 */	"\xDA\xE1", "\xB5\xA5", "\xDA\xDE", "\xB5\xAC",
  /* U+7B4C */	"\xDA\xE2", "\xB5\xAB", "\xDA\xE3", "\xB5\xAD",
  /* U+7B50 */	"\xB5\xA8", "\xB5\xAE", "\xB5\xA9", "\x00\x00",
  /* U+7B54 */	"\xB5\xAA", "\x8E\x5D", "\xB5\xA6", "\x00\x00",
  /* U+7B58 */	"\xDA\xE5", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+7B5C */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+7B60 */	"\xB8\x61", "\xDF\x50", "\x99\x50", "\xDF\x53",
  /* U+7B64 */	"\xDF\x47", "\xDF\x4C", "\xDF\x46", "\xB8\x63",
  /* U+7B68 */	"\x00\x00", "\xDF\x4A", "\x00\x00", "\x00\x00",
  /* U+7B6C */	"\x99\x51", "\xDF\x48", "\xB8\x62", "\x8E\x62",
  /* U+7B70 */	"\xDF\x4F", "\xDF\x4E", "\xDF\x4B", "\xDF\x4D",
  /* U+7B74 */	"\xDF\x49", "\xBA\xE1", "\xDF\x52", "\xB8\x5F",
  /* U+7B78 */	"\xDF\x51", "\x00\x00", "\x00\x00", "\x99\x52",
  /* U+7B7C */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+7B80 */	"\x00\x00", "\x00\x00", "\xE3\x5D", "\x00\x00",
  /* U+7B84 */	"\xBA\xE8", "\xE3\x58", "\x00\x00", "\xBA\xE7",
  /* U+7B88 */	"\xE3\x4E", "\x00\x00", "\xE3\x50", "\xBA\xE0",
  /* U+7B8C */	"\xE3\x55", "\xE3\x54", "\xE3\x57", "\xBA\xE5",
  /* U+7B90 */	"\xE3\x52", "\xE3\x51", "\x8E\x68", "\x00\x00",
  /* U+7B94 */	"\xBA\xE4", "\xBA\xDF", "\xE3\x53", "\xBA\xE2",
  /* U+7B98 */	"\xE3\x59", "\xE3\x5B", "\x00\x00", "\xE3\x56",
  /* U+7B9C */	"\xE3\x4F", "\xBA\xE3", "\x00\x00", "\x00\x00",
  /* U+7BA0 */	"\xBD\x69", "\xBA\xDE", "\x8E\x61", "\x9F\x59",
  /* U+7BA4 */	"\xE3\x5C", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+7BA8 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+7BAC */	"\xE6\xD9", "\xBD\x62", "\x00\x00", "\xE6\xDB",
  /* U+7BB0 */	"\x00\x00", "\xBD\x63", "\x8B\xB3", "\x00\x00",
  /* U+7BB4 */	"\xBD\x65", "\xE6\xDE", "\x00\x00", "\xE6\xD6",
  /* U+7BB8 */	"\xBA\xE6", "\xE6\xDC", "\x00\x00", "\x00\x00",
  /* U+7BBC */	"\x00\x00", "\x00\x00", "\xE6\xD8", "\x00\x00",
  /* U+7BC0 */	"\xB8\x60", "\xBD\x68", "\x00\x00", "\x00\x00",
  /* U+7BC4 */	"\xBD\x64", "\x00\x00", "\xBD\x66", "\xBD\x67",
  /* U+7BC8 */	"\x00\x00", "\xBF\x76", "\xE6\xDD", "\xE6\xD7",
  /* U+7BCC */	"\xBD\x6A", "\x00\x00", "\xE6\xDA", "\x9F\x5D",
  /* U+7BD0 */	"\x8E\x66", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+7BD4 */	"\xEA\xC0", "\xEA\xBB", "\x00\x00", "\x00\x00",
  /* U+7BD8 */	"\xEA\xC5", "\xBF\x74", "\xEA\xBD", "\xBF\x78",
  /* U+7BDC */	"\xEA\xC3", "\xEA\xBA", "\xEA\xB7", "\xEA\xC6",
  /* U+7BE0 */	"\xC1\x51", "\xBF\x79", "\xEA\xC2", "\xEA\xB8",
  /* U+7BE4 */	"\xBF\x77", "\xEA\xBC", "\xBF\x7B", "\xEA\xB9",
  /* U+7BE8 */	"\xEA\xBE", "\xBF\x7A", "\xEA\xC1", "\xEA\xC4",
  /* U+7BEC */	"\x8C\xB2", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+7BF0 */	"\xED\xCB", "\xED\xCC", "\xED\xBC", "\xED\xC3",
  /* U+7BF4 */	"\xED\xC1", "\x00\x00", "\x00\x00", "\xC1\x4F",
  /* U+7BF8 */	"\xED\xC8", "\xEA\xBF", "\x8E\x6E", "\xED\xBF",
  /* U+7BFC */	"\x9F\x64", "\xED\xC9", "\xC1\x4E", "\xED\xBE",
  /* U+7C00 */	"\xED\xBD", "\xED\xC7", "\xED\xC4", "\xED\xC6",
  /* U+7C04 */	"\x00\x00", "\xED\xBA", "\xED\xCA", "\xC1\x4C",
  /* U+7C08 */	"\x00\x00", "\xED\xC5", "\xED\xCE", "\xED\xC2",
  /* U+7C0C */	"\xC1\x50", "\xC1\x4D", "\xED\xC0", "\xED\xBB",
  /* U+7C10 */	"\xED\xCD", "\xBF\x75", "\x99\x53", "\x00\x00",
  /* U+7C14 */	"\x00\x00", "\xFA\xB8", "\x00\x00", "\x00\x00",
  /* U+7C18 */	"\x00\x00", "\xF0\x63", "\x00\x00", "\x99\x54",
  /* U+7C1C */	"\xF0\x61", "\xF0\x67", "\xC2\xB0", "\xF0\x65",
  /* U+7C20 */	"\xF0\x64", "\xC2\xB2", "\xF0\x6A", "\xC2\xB1",
  /* U+7C24 */	"\x00\x00", "\xF0\x6B", "\xF0\x68", "\xC2\xAE",
  /* U+7C28 */	"\xF0\x69", "\xF0\x62", "\xC2\xAF", "\xC2\xAD",
  /* U+7C2C */	"\xF2\xAB", "\xF0\x66", "\x00\x00", "\x00\x00",
  /* U+7C30 */	"\xF0\x6C", "\x00\x00", "\x00\x00", "\xF2\xA8",
  /* U+7C34 */	"\x00\x00", "\x8E\x70", "\x00\x00", "\xC3\xB2",
  /* U+7C38 */	"\xC3\xB0", "\xF2\xAA", "\x00\x00", "\xF2\xAC",
  /* U+7C3C */	"\xF2\xA9", "\xC3\xB1", "\xC3\xAE", "\xC3\xAF",
  /* U+7C40 */	"\xC3\xB3", "\x00\x00", "\x9F\x61", "\xC4\x78",
  /* U+7C44 */	"\x8E\x72", "\xF4\xAA", "\x00\x00", "\xF4\xA9",
  /* U+7C48 */	"\xF4\xA7", "\xF4\xA6", "\xF4\xA8", "\x00\x00",
  /* U+7C4C */	"\xC4\x77", "\xC4\x79", "\x00\x00", "\x00\x00",
  /* U+7C50 */	"\xC4\xF0", "\xA0\x6B", "\x00\x00", "\xF5\xE5",
  /* U+7C54 */	"\xF5\xE4", "\x00\x00", "\x9F\x40", "\xF6\xFA",
  /* U+7C58 */	"\x00\x00", "\xF6\xFC", "\xF6\xFE", "\xF6\xFD",
  /* U+7C5C */	"\xF6\xFB", "\x94\xED", "\x00\x00", "\xC5\xA3",
  /* U+7C60 */	"\xC5\xA2", "\x00\x00", "\x00\x00", "\xC5\xD3",
  /* U+7C64 */	"\xC5\xD2", "\xC5\xD4", "\xF7\xED", "\xF7\xEC",
  /* U+7C68 */	"\x00\x00", "\xF8\xFB", "\xF8\xB8", "\xF8\xFC",
  /* U+7C6C */	"\xC6\x58", "\x94\xEE", "\xC6\x59", "\xF9\x6D",
  /* U+7C70 */	"\x9F\xBD", "\x00\x00", "\xC6\x7E", "\xA6\xCC",
  /* U+7C74 */	"\x8E\x7B", "\xCD\xA8", "\x00\x00", "\x00\x00",
  /* U+7C78 */	"\xD0\x45", "\xD0\x46", "\xD0\x44", "\x99\x57",
  /* U+7C7C */	"\x94\xF7", "\xAC\xF3", "\x9F\x5F", "\xD0\x47",
  /* U+7C80 */	"\xD0\x48", "\xD0\x49", "\x00\x00", "\x8E\x73",
  /* U+7C84 */	"\xD3\x49", "\xD3\x4F", "\x9F\x62", "\x00\x00",
  /* U+7C88 */	"\xD3\x4D", "\xAF\xBB", "\xD3\x4B", "\x00\x00",
  /* U+7C8C */	"\xD3\x4C", "\xD3\x4E", "\x94\xF6", "\x00\x00",
  /* U+7C90 */	"\x00\x00", "\xD3\x4A", "\xB2\xC9", "\x00\x00",
  /* U+7C94 */	"\xD6\xDE", "\xB2\xCB", "\xD6\xE0", "\xB2\xCA",
  /* U+7C98 */	"\xD6\xDF", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+7C9C */	"\x99\x58", "\x00\x00", "\xDA\xE8", "\xB5\xAF",
  /* U+7CA0 */	"\x00\x00", "\xDA\xEA", "\xDA\xE7", "\xD6\xE1",
  /* U+7CA4 */	"\x00\x00", "\xB5\xB0", "\x8E\x75", "\xF9\xDB",
  /* U+7CA8 */	"\xDA\xE9", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+7CAC */	"\x90\x72", "\x00\x00", "\x94\xF8", "\xDF\x56",
  /* U+7CB0 */	"\x00\x00", "\xB8\x64", "\xDF\x54", "\xB8\x65",
  /* U+7CB4 */	"\xDF\x55", "\xB8\x66", "\x00\x00", "\x00\x00",
  /* U+7CB8 */	"\x99\x5A", "\xBA\xE9", "\xE3\x61", "\xE3\x5E",
  /* U+7CBC */	"\xE3\x60", "\xBA\xEA", "\xBA\xEB", "\xE3\x5F",
  /* U+7CC0 */	"\x00\x00", "\x00\x00", "\xA0\xB0", "\x8C\xB3",
  /* U+7CC4 */	"\x00\x00", "\xE6\xDF", "\x00\x00", "\x8E\x79",
  /* U+7CC8 */	"\xE6\xE0", "\x8E\x78", "\xBD\x6B", "\xE6\xE2",
  /* U+7CCC */	"\xE6\xE1", "\x94\xF3", "\xA2\x61", "\x00\x00",
  /* U+7CD0 */	"\xEA\xCA", "\xEA\xCB", "\xEA\xC7", "\x98\xAF",
  /* U+7CD4 */	"\xEA\xC8", "\xBF\x7C", "\xBF\x7D", "\xEA\xC9",
  /* U+7CD8 */	"\x00\x00", "\xC1\x57", "\xA0\xB2", "\x00\x00",
  /* U+7CDC */	"\xC1\x53", "\xC1\x58", "\xC1\x54", "\xC1\x56",
  /* U+7CE0 */	"\xC1\x52", "\x00\x00", "\xC1\x55", "\x00\x00",
  /* U+7CE4 */	"\x00\x00", "\x00\x00", "\x8E\x7A", "\xC2\xB3",
  /* U+7CE8 */	"\xED\xCF", "\x00\x00", "\xF2\xAE", "\x00\x00",
  /* U+7CEC */	"\xF2\xAD", "\x99\x5C", "\xF4\xAB", "\xC4\x7A",
  /* U+7CF0 */	"\xC4\x7B", "\xF7\x41", "\xF5\xE6", "\x8E\x7C",
  /* U+7CF4 */	"\xF7\x40", "\x8E\x7D", "\xF8\xFD", "\xF9\xA4",
  /* U+7CF8 */	"\xA6\xCD", "\x8B\xD9", "\x00\x00", "\xA8\x74",
  /* U+7CFC */	"\x89\xA2", "\xCD\xA9", "\xAA\xC8", "\x00\x00",
  /* U+7D00 */	"\xAC\xF6", "\xD0\x4C", "\xAC\xF4", "\xD0\x4A",
  /* U+7D04 */	"\xAC\xF9", "\xAC\xF5", "\xAC\xFA", "\xAC\xF8",
  /* U+7D08 */	"\xD0\x4B", "\xAC\xF7", "\xAF\xBF", "\xAF\xBE",
  /* U+7D0C */	"\xD3\x5A", "\xAF\xC7", "\xD3\x53", "\xD3\x59",
  /* U+7D10 */	"\xAF\xC3", "\xD3\x52", "\xD3\x58", "\xD3\x56",
  /* U+7D14 */	"\xAF\xC2", "\xAF\xC4", "\xD3\x55", "\xAF\xBD",
  /* U+7D18 */	"\xD3\x54", "\xAF\xC8", "\xAF\xC5", "\xAF\xC9",
  /* U+7D1C */	"\xAF\xC6", "\xD3\x51", "\xD3\x50", "\xD3\x57",
  /* U+7D20 */	"\xAF\xC0", "\xAF\xBC", "\xAF\xC1", "\x00\x00",
  /* U+7D24 */	"\x00\x00", "\x9E\xD7", "\x00\x00", "\x00\x00",
  /* U+7D28 */	"\xD6\xF0", "\xD6\xE9", "\x00\x00", "\xB5\xB5",
  /* U+7D2C */	"\xD6\xE8", "\x00\x00", "\xB2\xCF", "\xB2\xD6",
  /* U+7D30 */	"\xB2\xD3", "\xB2\xD9", "\xB2\xD8", "\xB2\xD4",
  /* U+7D34 */	"\x00\x00", "\xD6\xE2", "\xD6\xE5", "\x00\x00",
  /* U+7D38 */	"\xD6\xE4", "\xB2\xD0", "\xD6\xE6", "\xD6\xEF",
  /* U+7D3C */	"\xB2\xD1", "\xD6\xE3", "\xD6\xEC", "\xD6\xED",
  /* U+7D40 */	"\xB2\xD2", "\xD6\xEA", "\xB2\xD7", "\xB2\xCD",
  /* U+7D44 */	"\xB2\xD5", "\xD6\xE7", "\xB2\xCC", "\xD6\xEB",
  /* U+7D48 */	"\x00\x00", "\x00\x00", "\xD6\xEE", "\x00\x00",
  /* U+7D4C */	"\x00\x00", "\xA0\xB6", "\xDA\xFB", "\xDA\xF2",
  /* U+7D50 */	"\xB5\xB2", "\xDA\xF9", "\xDA\xF6", "\xDA\xEE",
  /* U+7D54 */	"\xDA\xF7", "\xB5\xB4", "\xDA\xEF", "\x00\x00",
  /* U+7D58 */	"\xDA\xEB", "\x00\x00", "\x9E\x42", "\xB8\x6C",
  /* U+7D5C */	"\xDA\xF4", "\x8E\xA4", "\xB5\xB1", "\xDA\xFA",
  /* U+7D60 */	"\x00\x00", "\xB5\xB8", "\xB5\xBA", "\xDA\xED",
  /* U+7D64 */	"\x00\x00", "\x00\x00", "\xB5\xB9", "\xDA\xF0",
  /* U+7D68 */	"\xB5\xB3", "\xDA\xF8", "\xDA\xF1", "\xDA\xF5",
  /* U+7D6C */	"\x00\x00", "\xDA\xF3", "\xB5\xB6", "\xDA\xEC",
  /* U+7D70 */	"\xB5\xBB", "\xB2\xCE", "\xB5\xB7", "\xB5\xBC",
  /* U+7D74 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+7D78 */	"\x00\x00", "\xB8\x68", "\xDF\x5D", "\xDF\x5F",
  /* U+7D7C */	"\xDF\x61", "\xDF\x65", "\x00\x00", "\xDF\x5B",
  /* U+7D80 */	"\xDF\x59", "\xB8\x6A", "\x00\x00", "\xDF\x60",
  /* U+7D84 */	"\xDF\x64", "\xDF\x5C", "\xDF\x58", "\x00\x00",
  /* U+7D88 */	"\xDF\x57", "\x8E\xA7", "\x00\x00", "\x8C\x76",
  /* U+7D8C */	"\xDF\x62", "\xDF\x5A", "\xDF\x5E", "\xB8\x6B",
  /* U+7D90 */	"\x00\x00", "\xB8\x69", "\xDF\x66", "\xB8\x67",
  /* U+7D94 */	"\xDF\x63", "\x00\x00", "\xE3\x72", "\x95\x42",
  /* U+7D98 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+7D9C */	"\xBA\xEE", "\xE3\x6A", "\xBD\x78", "\xE3\x74",
  /* U+7DA0 */	"\xBA\xF1", "\xE3\x78", "\xBA\xF7", "\xE3\x65",
  /* U+7DA4 */	"\x98\x7D", "\x00\x00", "\xE3\x75", "\xE3\x62",
  /* U+7DA8 */	"\x97\x55", "\xE3\x77", "\xE3\x66", "\x8E\xA8",
  /* U+7DAC */	"\xBA\xFE", "\xBA\xFB", "\xE3\x76", "\xE3\x70",
  /* U+7DB0 */	"\xBA\xED", "\xBA\xF5", "\xBA\xF4", "\x8E\xAA",
  /* U+7DB4 */	"\xBA\xF3", "\xBA\xF9", "\x00\x00", "\xE3\x63",
  /* U+7DB8 */	"\xBA\xFA", "\xE3\x71", "\xBA\xF6", "\xBA\xEC",
  /* U+7DBC */	"\xE3\x73", "\xBA\xEF", "\xBA\xF0", "\xBA\xF8",
  /* U+7DC0 */	"\xE3\x68", "\xE3\x67", "\xE3\x64", "\x00\x00",
  /* U+7DC4 */	"\xE3\x6C", "\xE3\x69", "\xE3\x6D", "\xBA\xFD",
  /* U+7DC8 */	"\x00\x00", "\xE3\x79", "\xBA\xF2", "\xE3\x6E",
  /* U+7DCC */	"\xE3\x6F", "\x89\xA3", "\xE3\x6B", "\x99\x60",
  /* U+7DD0 */	"\x99\x62", "\x00\x00", "\xBA\xFC", "\x94\xFC",
  /* U+7DD4 */	"\x99\x61", "\x00\x00", "\x00\x00", "\xE6\xE7",
  /* U+7DD8 */	"\xBD\x70", "\xBD\x79", "\xBD\x75", "\xE6\xE4",
  /* U+7DDC */	"\x94\xFA", "\xBD\x72", "\xBD\x76", "\xE6\xF0",
  /* U+7DE0 */	"\xBD\x6C", "\xE6\xE8", "\x00\x00", "\xBD\x74",
  /* U+7DE4 */	"\x8E\xAE", "\x8E\xB2", "\xE6\xEB", "\xE6\xE6",
  /* U+7DE8 */	"\xBD\x73", "\xBD\x77", "\xE6\xE5", "\x00\x00",
  /* U+7DEC */	"\xBD\x71", "\x00\x00", "\xE6\xEF", "\xBD\x6E",
  /* U+7DF0 */	"\xE6\xEE", "\xE6\xED", "\xBD\x7A", "\xE5\x72",
  /* U+7DF4 */	"\xBD\x6D", "\x8E\xB0", "\xE6\xEC", "\xE6\xE3",
  /* U+7DF8 */	"\x00\x00", "\xBD\x7B", "\xE6\xEA", "\xBD\x6F",
  /* U+7DFC */	"\x00\x00", "\x99\x63", "\x97\xAA", "\x00\x00",
  /* U+7E00 */	"\x00\x00", "\x00\x00", "\x00\x00", "\xE6\xE9",
  /* U+7E04 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x94\xFB",
  /* U+7E08 */	"\xBF\xA2", "\xBF\xA7", "\xBF\x7E", "\xEA\xD8",
  /* U+7E0C */	"\xEA\xCF", "\xEA\xDB", "\xEA\xD3", "\xEA\xD9",
  /* U+7E10 */	"\xBF\xA8", "\xBF\xA1", "\xEA\xCC", "\xEA\xD2",
  /* U+7E14 */	"\xEA\xDC", "\xEA\xD5", "\xEA\xDA", "\xEA\xCE",
  /* U+7E18 */	"\x00\x00", "\x00\x00", "\xEA\xD6", "\xBF\xA3",
  /* U+7E1C */	"\xEA\xD4", "\xBF\xA6", "\xBF\xA5", "\xEA\xD0",
  /* U+7E20 */	"\xEA\xD1", "\xEA\xCD", "\xEA\xD7", "\xBF\xA4",
  /* U+7E24 */	"\xEA\xDE", "\xEA\xDD", "\x00\x00", "\x8E\xBB",
  /* U+7E28 */	"\x00\x00", "\xED\xDA", "\xED\xD6", "\xC1\x5F",
  /* U+7E2C */	"\x00\x00", "\xED\xD0", "\xC1\x59", "\xC1\x69",
  /* U+7E30 */	"\xED\xDC", "\xC1\x61", "\xC1\x5D", "\xED\xD3",
  /* U+7E34 */	"\xC1\x64", "\xC1\x67", "\xED\xDE", "\xC1\x5C",
  /* U+7E38 */	"\xED\xD5", "\xC1\x65", "\xED\xE0", "\xED\xDD",
  /* U+7E3C */	"\xED\xD1", "\xC1\x60", "\xC1\x5A", "\xC1\x68",
  /* U+7E40 */	"\xED\xD8", "\xC1\x63", "\xED\xD2", "\xC1\x5E",
  /* U+7E44 */	"\xED\xDF", "\xC1\x62", "\xC1\x5B", "\xED\xD9",
  /* U+7E48 */	"\xC1\x66", "\xED\xD7", "\x00\x00", "\x00\x00",
  /* U+7E4C */	"\xED\xDB", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+7E50 */	"\xF0\x6E", "\xF0\x74", "\xC2\xB9", "\xF0\x77",
  /* U+7E54 */	"\xC2\xB4", "\xC2\xB5", "\xF0\x6F", "\xF0\x76",
  /* U+7E58 */	"\xF0\x71", "\xC2\xBA", "\xC2\xB7", "\x8C\xDC",
  /* U+7E5C */	"\xF0\x6D", "\x00\x00", "\xC2\xB6", "\xF0\x73",
  /* U+7E60 */	"\xF0\x75", "\xC2\xB8", "\xF0\x72", "\xF0\x70",
  /* U+7E64 */	"\x00\x00", "\x98\x76", "\x00\x00", "\x8E\xA1",
  /* U+7E68 */	"\xF2\xB8", "\xC3\xB7", "\xC3\xB8", "\xC3\xB4",
  /* U+7E6C */	"\x8C\xB4", "\xC3\xB5", "\x8E\xB7", "\xF2\xB4",
  /* U+7E70 */	"\xF2\xB2", "\x00\x00", "\xF2\xB6", "\xC3\xBA",
  /* U+7E74 */	"\xF2\xB7", "\xF2\xB0", "\xF2\xAF", "\xF2\xB3",
  /* U+7E78 */	"\xF2\xB1", "\xC3\xB6", "\xF2\xB5", "\xF4\xAC",
  /* U+7E7C */	"\xC4\x7E", "\xC4\x7D", "\xF4\xAD", "\x9D\xA6",
  /* U+7E80 */	"\xF4\xAF", "\xF4\xAE", "\xC4\xA1", "\x00\x00",
  /* U+7E84 */	"\x00\x00", "\x00\x00", "\xF5\xEB", "\xF5\xE8",
  /* U+7E88 */	"\xF5\xE9", "\x00\x00", "\xF5\xE7", "\xF5\xEA",
  /* U+7E8C */	"\xC4\xF2", "\xF5\xEC", "\x9E\xB0", "\xC4\xF1",
  /* U+7E90 */	"\x00\x00", "\xF7\x42", "\x8E\xB8", "\xC5\xD5",
  /* U+7E94 */	"\xC5\xD7", "\xF7\xEE", "\xC5\xD6", "\xF8\xB9",
  /* U+7E98 */	"\xF9\x40", "\xF9\x42", "\xF8\xFE", "\xF9\x41",
  /* U+7E9C */	"\xC6\x6C", "\x00\x00", "\x00\x00", "\x9D\x70",
  /* U+7EA0 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+7EA4 */	"\x89\x6E", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+7EA8 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+7EAC */	"\x89\x6F", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+7EB0 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+7EB4 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+7EB8 */	"\x00\x00", "\x00\x00", "\x89\x70", "\x00\x00",
  /* U+7EBC */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+7EC0 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+7EC4 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x89\x71",
  /* U+7EC8 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+7ECC */	"\x00\x00", "\x00\x00", "\x00\x00", "\x89\x72",
  /* U+7ED0 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+7ED4 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+7ED8 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+7EDC */	"\x00\x00", "\x00\x00", "\x00\x00", "\x89\x73",
  /* U+7EE0 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+7EE4 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+7EE8 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+7EEC */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+7EF0 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+7EF4 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+7EF8 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+7EFC */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+7F00 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+7F04 */	"\x00\x00", "\x00\x00", "\x89\x74", "\x00\x00",
  /* U+7F08 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+7F0C */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+7F10 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+7F14 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+7F18 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+7F1C */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+7F20 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+7F24 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+7F28 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+7F2C */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+7F30 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+7F34 */	"\x00\x00", "\x00\x00", "\xA6\xCE", "\x89\x75",
  /* U+7F38 */	"\xAC\xFB", "\xD2\x6F", "\xAF\xCA", "\x00\x00",
  /* U+7F3C */	"\x00\x00", "\xB2\xDA", "\xDA\xFC", "\xDA\xFD",
  /* U+7F40 */	"\x8E\xBC", "\x8E\xBD", "\x00\x00", "\xEA\xDF",
  /* U+7F44 */	"\xC1\x6A", "\xED\xE1", "\x00\x00", "\x8E\xBE",
  /* U+7F48 */	"\xC2\xBB", "\x9D\xD1", "\xF2\xBA", "\xF2\xB9",
  /* U+7F4C */	"\xC4\xA2", "\xF5\xED", "\x94\xFD", "\xF7\x43",
  /* U+7F50 */	"\xC5\xF8", "\xCA\x49", "\x8B\xD7", "\x8B\xDA",
  /* U+7F54 */	"\xAA\xC9", "\xA8\x75", "\x00\x00", "\x00\x00",
  /* U+7F58 */	"\xD0\x4D", "\x00\x00", "\x00\x00", "\xD3\x60",
  /* U+7F5C */	"\xD3\x5B", "\xD3\x5F", "\xD3\x5D", "\xAF\xCB",
  /* U+7F60 */	"\xD3\x5E", "\xD3\x5C", "\x00\x00", "\xD6\xF1",
  /* U+7F64 */	"\x00\x00", "\xDA\xFE", "\xDB\x40", "\xDF\x69",
  /* U+7F68 */	"\xDF\x6A", "\xB8\x6E", "\xB8\x6F", "\xDF\x68",
  /* U+7F6C */	"\xDF\x6B", "\xDF\x67", "\xB8\x6D", "\x00\x00",
  /* U+7F70 */	"\xBB\x40", "\xA0\xE2", "\xB8\x70", "\xE3\x7A",
  /* U+7F74 */	"\x00\x00", "\xBD\x7C", "\xE6\xF1", "\xBD\x7D",
  /* U+7F78 */	"\x9F\xE9", "\xBF\xA9", "\xEA\xE2", "\xEA\xE0",
  /* U+7F7C */	"\xEA\xE1", "\xED\xE4", "\xED\xE3", "\xED\xE2",
  /* U+7F80 */	"\x00\x00", "\x00\x00", "\x00\x00", "\xF2\xBB",
  /* U+7F84 */	"\x00\x00", "\xC3\xB9", "\xF2\xBC", "\xF7\x44",
  /* U+7F88 */	"\xC5\xF9", "\xF8\xBA", "\xA6\xCF", "\xAA\xCB",
  /* U+7F8C */	"\xAA\xCA", "\xD0\x4F", "\xAC\xFC", "\xFD\xA8",
  /* U+7F90 */	"\x00\x00", "\xD0\x4E", "\xD3\x62", "\x8A\xE7",
  /* U+7F94 */	"\xAF\xCC", "\xD6\xF2", "\xD3\x61", "\x8E\xC2",
  /* U+7F98 */	"\x00\x00", "\x00\x00", "\xB2\xDC", "\xD6\xF5",
  /* U+7F9C */	"\xD6\xF3", "\xD6\xF4", "\xB2\xDB", "\x00\x00",
  /* U+7FA0 */	"\xDB\x42", "\xDB\x43", "\xDB\x41", "\x8E\xC4",
  /* U+7FA4 */	"\xB8\x73", "\xDF\x6D", "\xDF\x6C", "\xDF\x6E",
  /* U+7FA8 */	"\xB8\x72", "\xB8\x71", "\x00\x00", "\x00\x00",
  /* U+7FAC */	"\xE6\xF2", "\xE6\xF4", "\x99\x64", "\xBD\x7E",
  /* U+7FB0 */	"\xE6\xF3", "\xEA\xE3", "\xBF\xAA", "\xF0\x79",
  /* U+7FB4 */	"\x99\x65", "\xF0\x78", "\xC3\xBB", "\xF2\xBD",
  /* U+7FB8 */	"\xC3\xBD", "\xC3\xBC", "\xF4\xB0", "\xF5\xEE",
  /* U+7FBC */	"\xC4\xF3", "\xA6\xD0", "\xD0\x50", "\xAC\xFD",
  /* U+7FC0 */	"\xD3\x65", "\xAF\xCE", "\xD3\x64", "\xD3\x63",
  /* U+7FC4 */	"\x00\x00", "\xAF\xCD", "\x00\x00", "\xD6\xFB",
  /* U+7FC8 */	"\x00\x00", "\xD6\xFD", "\xD6\xF6", "\xD6\xF7",
  /* U+7FCC */	"\xB2\xDD", "\xD6\xF8", "\xB2\xDE", "\xD6\xFC",
  /* U+7FD0 */	"\xD6\xF9", "\xD6\xFA", "\xB2\xDF", "\x00\x00",
  /* U+7FD4 */	"\xB5\xBE", "\xB5\xBF", "\x00\x00", "\xDB\x44",
  /* U+7FD8 */	"\x00\x00", "\x00\x00", "\x00\x00", "\xDF\x6F",
  /* U+7FDC */	"\xDF\x70", "\x95\x4E", "\xE3\x7E", "\xBB\x43",
  /* U+7FE0 */	"\xBB\x41", "\xBB\x42", "\xE3\x7B", "\xE3\x7C",
  /* U+7FE4 */	"\x00\x00", "\xE3\x7D", "\xE6\xF9", "\x98\xB3",
  /* U+7FE8 */	"\xE6\xFA", "\xBD\xA1", "\xE6\xF7", "\xE6\xF6",
  /* U+7FEC */	"\xE6\xF8", "\xE6\xF5", "\xBF\xAD", "\xEA\xE4",
  /* U+7FF0 */	"\xBF\xAB", "\xBF\xAC", "\xED\xE6", "\xC1\x6B",
  /* U+7FF4 */	"\xED\xE5", "\xEF\xA8", "\x00\x00", "\xF0\x7A",
  /* U+7FF8 */	"\xF0\x7B", "\xC2\xBC", "\x8E\xCB", "\xC2\xBD",
  /* U+7FFC */	"\xC1\x6C", "\xF2\xBE", "\xF2\xBF", "\xF4\xB1",
  /* U+8000 */	"\xC4\xA3", "\xA6\xD1", "\x8B\xDF", "\xA6\xD2",
  /* U+8004 */	"\xAC\xFE", "\xAA\xCC", "\xAF\xCF", "\xD0\x51",
  /* U+8008 */	"\x8E\xCE", "\x00\x00", "\x00\x00", "\xB5\xC0",
  /* U+800C */	"\xA6\xD3", "\xAD\x41", "\xD0\x52", "\xD0\x53",
  /* U+8010 */	"\xAD\x40", "\xAD\x42", "\xA6\xD4", "\x00\x00",
  /* U+8014 */	"\xD0\x54", "\xAF\xD1", "\xD3\x66", "\xAF\xD3",
  /* U+8018 */	"\xAF\xD0", "\xAF\xD2", "\x00\x00", "\xD7\x41",
  /* U+801C */	"\xB2\xE0", "\x8E\xCF", "\xD7\x40", "\xD6\xFE",
  /* U+8020 */	"\x99\x68", "\xDF\x71", "\x00\x00", "\x00\x00",
  /* U+8024 */	"\xE3\xA1", "\x99\x69", "\xBD\xA2", "\x00\x00",
  /* U+8028 */	"\xBF\xAE", "\xEA\xE6", "\xEA\xE5", "\x00\x00",
  /* U+802C */	"\xED\xE7", "\x00\x00", "\x99\x6B", "\x8E\xD1",
  /* U+8030 */	"\xF5\xEF", "\x99\x6C", "\x00\x00", "\xA6\xD5",
  /* U+8034 */	"\xCB\x73", "\xCD\xAA", "\xAD\x43", "\xD0\x55",
  /* U+8038 */	"\x00\x00", "\xD3\x68", "\x00\x00", "\x8E\xD4",
  /* U+803C */	"\x8E\xD5", "\xAF\xD4", "\xD3\x67", "\xAF\xD5",
  /* U+8040 */	"\x00\x00", "\x00\x00", "\x00\x00", "\xD7\x43",
  /* U+8044 */	"\x00\x00", "\x00\x00", "\xB2\xE2", "\xD7\x42",
  /* U+8048 */	"\xD7\x44", "\x00\x00", "\xB2\xE1", "\x00\x00",
  /* U+804C */	"\x00\x00", "\x00\x00", "\x00\x00", "\xDB\x46",
  /* U+8050 */	"\xDB\x47", "\xDB\x45", "\xB5\xC1", "\x00\x00",
  /* U+8054 */	"\x99\x6D", "\x00\x00", "\xB8\x74", "\x00\x00",
  /* U+8058 */	"\xB8\x75", "\x00\x00", "\xBB\x45", "\xA0\xBE",
  /* U+805C */	"\xE3\xA3", "\xE3\xA2", "\xBB\x44", "\x00\x00",
  /* U+8060 */	"\x00\x00", "\x8E\xD6", "\xA0\xBC", "\xA0\xB5",
  /* U+8064 */	"\xE6\xFB", "\x00\x00", "\xA0\xB4", "\xE6\xFC",
  /* U+8068 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+806C */	"\xEA\xE7", "\x00\x00", "\x00\x00", "\xC1\x70",
  /* U+8070 */	"\xC1\x6F", "\xC1\x6D", "\xC1\x6E", "\xC1\x71",
  /* U+8074 */	"\x00\x00", "\xF0\x7C", "\xC2\xBF", "\xC2\xBE",
  /* U+8078 */	"\xF2\xC0", "\xF4\xB2", "\x00\x00", "\x00\x00",
  /* U+807C */	"\x00\x00", "\xC5\xA5", "\xC5\xA4", "\xA6\xD6",
  /* U+8080 */	"\x8B\xE0", "\x00\x00", "\xD1\xFB", "\x00\x00",
  /* U+8084 */	"\xB8\x77", "\xB5\xC2", "\xB8\x76", "\xBB\x46",
  /* U+8088 */	"\x00\x00", "\xA6\xD7", "\xC9\xA9", "\xA6\xD8",
  /* U+808C */	"\xA6\xD9", "\x00\x00", "\x00\x00", "\xCD\xAB",
  /* U+8090 */	"\xCB\x76", "\x00\x00", "\xCB\x77", "\xA8\x77",
  /* U+8094 */	"\x00\x00", "\xCB\x74", "\xA8\x76", "\x00\x00",
  /* U+8098 */	"\xA8\x79", "\xCB\x75", "\xA8\x7B", "\xA8\x7A",
  /* U+809C */	"\xCB\x78", "\xA8\x78", "\x00\x00", "\x89\xB5",
  /* U+80A0 */	"\x00\x00", "\xAA\xD1", "\xAA\xCF", "\xCD\xAD",
  /* U+80A4 */	"\x00\x00", "\xAA\xCE", "\x00\x00", "\x8E\xDD",
  /* U+80A8 */	"\x00\x00", "\xAA\xD3", "\xAA\xD5", "\xAA\xD2",
  /* U+80AC */	"\x00\x00", "\xCD\xB0", "\xCD\xAC", "\xAA\xD6",
  /* U+80B0 */	"\x00\x00", "\xAA\xD0", "\xA8\x7C", "\x00\x00",
  /* U+80B4 */	"\xAA\xD4", "\xCD\xAF", "\x9E\x5D", "\x99\x71",
  /* U+80B8 */	"\xCD\xAE", "\x00\x00", "\xAA\xCD", "\x00\x00",
  /* U+80BC */	"\x89\xAE", "\x9D\xE8", "\x00\x00", "\x00\x00",
  /* U+80C0 */	"\x00\x00", "\x00\x00", "\xD0\x5B", "\xAD\x47",
  /* U+80C4 */	"\xAD\x48", "\xD0\x5D", "\x95\x65", "\xD0\x57",
  /* U+80C8 */	"\xD0\x5A", "\xD0\x63", "\xD0\x61", "\x00\x00",
  /* U+80CC */	"\xAD\x49", "\xD0\x67", "\xAD\x4C", "\xD0\x64",
  /* U+80D0 */	"\xD0\x5C", "\xD0\x59", "\x00\x00", "\x00\x00",
  /* U+80D4 */	"\xDB\x49", "\xD0\x62", "\xAD\x44", "\xD0\x65",
  /* U+80D8 */	"\xD0\x56", "\xD0\x5F", "\xAD\x46", "\xAD\x4B",
  /* U+80DC */	"\xD0\x60", "\xAD\x4F", "\xAD\x4D", "\x00\x00",
  /* U+80E0 */	"\xD0\x58", "\xAD\x4A", "\x00\x00", "\xD0\x5E",
  /* U+80E4 */	"\xAD\x4E", "\xAD\x45", "\xD0\x66", "\x00\x00",
  /* U+80E8 */	"\x00\x00", "\x99\x72", "\x00\x00", "\x00\x00",
  /* U+80EC */	"\x8B\x5C", "\xAF\xDA", "\x00\x00", "\xAF\xE3",
  /* U+80F0 */	"\xAF\xD8", "\xAF\xD6", "\xD3\x6A", "\xAF\xDE",
  /* U+80F4 */	"\xAF\xDB", "\xD3\x6C", "\x89\xB1", "\x00\x00",
  /* U+80F8 */	"\xAF\xDD", "\xD3\x6B", "\xD3\x69", "\xD3\x6E",
  /* U+80FC */	"\xAF\xE2", "\xAF\xE0", "\xDB\x48", "\x00\x00",
  /* U+8100 */	"\xD3\x6F", "\xD3\x6D", "\xAF\xD7", "\xA0\xC0",
  /* U+8104 */	"\x00\x00", "\xAF\xD9", "\xAF\xDC", "\x8E\xDF",
  /* U+8108 */	"\xAF\xDF", "\x95\x66", "\xAF\xE1", "\x00\x00",
  /* U+810C */	"\x99\x74", "\x00\x00", "\x99\x76", "\x00\x00",
  /* U+8110 */	"\x00\x00", "\x00\x00", "\x99\x77", "\x00\x00",
  /* U+8114 */	"\x99\x79", "\xD7\x4E", "\xB2\xE4", "\x9D\xDA",
  /* U+8118 */	"\xD7\x45", "\xD7\x47", "\x8E\xE0", "\xD7\x48",
  /* U+811C */	"\x00\x00", "\xD7\x50", "\xD7\x4C", "\xD7\x4A",
  /* U+8120 */	"\x00\x00", "\xD7\x4D", "\xD7\x51", "\xB2\xE5",
  /* U+8124 */	"\xB2\xE9", "\xD7\x46", "\x00\x00", "\xD7\x4F",
  /* U+8128 */	"\x00\x00", "\xB2\xE7", "\x93\x5C", "\xB2\xE6",
  /* U+812C */	"\xD7\x4B", "\xD7\x49", "\x00\x00", "\xB2\xE3",
  /* U+8130 */	"\xB2\xE8", "\x00\x00", "\x9D\xE6", "\x00\x00",
  /* U+8134 */	"\x8B\x5F", "\x00\x00", "\x00\x00", "\x95\x63",
  /* U+8138 */	"\x00\x00", "\xB5\xC8", "\xDB\x51", "\x00\x00",
  /* U+813C */	"\x00\x00", "\xDB\x4F", "\xB5\xCA", "\x00\x00",
  /* U+8140 */	"\x00\x00", "\x00\x00", "\x95\x67", "\xDB\x4A",
  /* U+8144 */	"\xDF\xA1", "\x00\x00", "\xB5\xC9", "\xDB\x4E",
  /* U+8148 */	"\x9D\xE3", "\x00\x00", "\xDB\x4B", "\xB5\xC5",
  /* U+814C */	"\xB5\xCB", "\xDB\x50", "\xB5\xC7", "\xDB\x4D",
  /* U+8150 */	"\xBB\x47", "\xB5\xC6", "\xDB\x4C", "\xB5\xCC",
  /* U+8154 */	"\xB5\xC4", "\xB5\xC3", "\x99\x7C", "\x00\x00",
  /* U+8158 */	"\x00\x00", "\x99\x7D", "\x99\x7E", "\xDF\x77",
  /* U+815C */	"\xDF\x75", "\x00\x00", "\xDF\x7B", "\x00\x00",
  /* U+8160 */	"\xDF\x73", "\xDF\xA2", "\xDF\x78", "\x00\x00",
  /* U+8164 */	"\xDF\x72", "\xB8\x7B", "\xB8\xA3", "\xDF\x7D",
  /* U+8168 */	"\x00\x00", "\xDF\x76", "\x00\x00", "\xB8\x7E",
  /* U+816C */	"\x00\x00", "\x8B\x5B", "\xB8\x7C", "\xDF\x7E",
  /* U+8170 */	"\xB8\x79", "\xB8\x78", "\xDF\x79", "\xB8\x7D",
  /* U+8174 */	"\xB5\xCD", "\x00\x00", "\xDF\x7C", "\xDF\x74",
  /* U+8178 */	"\xB8\x7A", "\xB8\xA1", "\xB8\xA2", "\x00\x00",
  /* U+817C */	"\x99\xA3", "\x00\x00", "\x00\x00", "\xBB\x4C",
  /* U+8180 */	"\xBB\x48", "\x00\x00", "\xBB\x4D", "\xE3\xA6",
  /* U+8184 */	"\x99\xA4", "\x00\x00", "\xE3\xA5", "\xE3\xA7",
  /* U+8188 */	"\xBB\x4A", "\xE3\xA4", "\xBB\x4B", "\xE3\xAA",
  /* U+818C */	"\xE3\xA9", "\xE3\xA8", "\x00\x00", "\xBB\x49",
  /* U+8190 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x99\xA6",
  /* U+8194 */	"\x00\x00", "\xE7\x41", "\x00\x00", "\xE7\x44",
  /* U+8198 */	"\xBD\xA8", "\xE7\x43", "\xBD\xA7", "\xBD\xA3",
  /* U+819C */	"\xBD\xA4", "\xBD\xA5", "\xE7\x40", "\xE6\xFE",
  /* U+81A0 */	"\xBD\xA6", "\x00\x00", "\xE7\x42", "\xE6\xFD",
  /* U+81A4 */	"\x00\x00", "\x99\xA8", "\xEA\xE9", "\xEA\xF3",
  /* U+81A8 */	"\xBF\xB1", "\xBF\xB0", "\x8A\xBE", "\xEA\xED",
  /* U+81AC */	"\xEA\xEF", "\x00\x00", "\xEA\xEA", "\x00\x00",
  /* U+81B0 */	"\xEA\xEE", "\xEA\xE8", "\xEA\xF1", "\xBF\xAF",
  /* U+81B4 */	"\xEA\xF0", "\xEA\xEC", "\x9E\x61", "\xEA\xF2",
  /* U+81B8 */	"\x00\x00", "\xEA\xEB", "\xC1\x74", "\xED\xE8",
  /* U+81BC */	"\xED\xEE", "\xC1\x78", "\xC1\x7A", "\xC1\x77",
  /* U+81C0 */	"\xC1\x76", "\x99\xAA", "\xC1\x75", "\xC1\x73",
  /* U+81C4 */	"\xED\xE9", "\xED\xEC", "\xC1\x72", "\xED\xED",
  /* U+81C8 */	"\xA0\xC8", "\xC1\x79", "\xED\xEB", "\x00\x00",
  /* U+81CC */	"\xED\xEA", "\xC2\xC0", "\x00\x00", "\xC2\xC1",
  /* U+81D0 */	"\xF0\xA1", "\xF0\x7D", "\xF0\x7E", "\x00\x00",
  /* U+81D4 */	"\x00\x00", "\xF2\xC2", "\x00\x00", "\xF2\xC1",
  /* U+81D8 */	"\xC3\xBE", "\xF4\xB4", "\xC4\xA4", "\xF4\xB3",
  /* U+81DC */	"\x00\x00", "\xF5\xF0", "\xF7\x45", "\xC5\xA6",
  /* U+81E0 */	"\xF9\x43", "\xF9\x44", "\xC5\xD8", "\xA6\xDA",
  /* U+81E4 */	"\x99\xAB", "\xAA\xD7", "\xDB\x52", "\xBB\x4E",
  /* U+81E8 */	"\xC1\x7B", "\xED\xEF", "\xA6\xDB", "\x00\x00",
  /* U+81EC */	"\xAF\xE5", "\xAF\xE4", "\xDB\x53", "\xFE\xC4",
  /* U+81F0 */	"\x00\x00", "\x00\x00", "\xEA\xF4", "\xA6\xDC",
  /* U+81F4 */	"\xAD\x50", "\x00\x00", "\x98\xC2", "\xDB\x54",
  /* U+81F8 */	"\xDB\x55", "\xDB\x56", "\xBB\x4F", "\xBF\xB2",
  /* U+81FC */	"\xA6\xDD", "\x00\x00", "\xAA\xD8", "\xD0\x68",
  /* U+8200 */	"\xAF\xE6", "\xD3\x70", "\xB2\xEA", "\x00\x00",
  /* U+8204 */	"\xDB\x57", "\xB8\xA4", "\x00\x00", "\xBB\x50",
  /* U+8208 */	"\xBF\xB3", "\xC1\x7C", "\xC2\xC2", "\xF4\xB5",
  /* U+820C */	"\xA6\xDE", "\xAA\xD9", "\x00\x00", "\x00\x00",
  /* U+8210 */	"\xAF\xE7", "\xD7\x52", "\xB5\xCE", "\x00\x00",
  /* U+8214 */	"\xBB\x51", "\xE3\xAB", "\xE7\x45", "\x00\x00",
  /* U+8218 */	"\x8E\xE8", "\x00\x00", "\xA0\xBA", "\xA6\xDF",
  /* U+821C */	"\xB5\xCF", "\xDF\xA3", "\xBB\x52", "\xA6\xE0",
  /* U+8220 */	"\xCD\xB1", "\xD0\x69", "\xAD\x51", "\x00\x00",
  /* U+8224 */	"\x00\x00", "\xD3\x72", "\xFD\x77", "\x00\x00",
  /* U+8228 */	"\xAF\xEA", "\x8E\xEE", "\xAF\xE8", "\xAF\xE9",
  /* U+822C */	"\xAF\xEB", "\x9E\xBF", "\x00\x00", "\xD3\x71",
  /* U+8230 */	"\x00\x00", "\x00\x00", "\xD7\x57", "\xD7\x54",
  /* U+8234 */	"\xD7\x56", "\xB2\xEB", "\xB2\xED", "\xB2\xEC",
  /* U+8238 */	"\xD7\x53", "\xB2\xEE", "\xD7\x55", "\x00\x00",
  /* U+823C */	"\xDB\x58", "\xDB\x59", "\x89\xC2", "\xDB\x5A",
  /* U+8240 */	"\xDF\xA6", "\x00\x00", "\xDF\xA7", "\x00\x00",
  /* U+8244 */	"\xDF\xA5", "\xDF\xA8", "\x00\x00", "\xB8\xA5",
  /* U+8248 */	"\x00\x00", "\xDF\xA4", "\x00\x00", "\xBB\x53",
  /* U+824C */	"\x00\x00", "\x00\x00", "\xE7\x4A", "\xE7\x46",
  /* U+8250 */	"\xE7\x49", "\xE7\x4B", "\xE7\x48", "\xE7\x47",
  /* U+8254 */	"\x99\xAC", "\xEA\xF5", "\xEA\xF6", "\xEA\xF7",
  /* U+8258 */	"\xBF\xB4", "\xBF\xB5", "\xED\xF1", "\xED\xF0",
  /* U+825C */	"\xED\xF2", "\x00\x00", "\xF0\xA3", "\xF0\xA2",
  /* U+8260 */	"\x00\x00", "\xF2\xC4", "\x95\x6B", "\xF2\xC5",
  /* U+8264 */	"\xF2\xC3", "\x95\x6C", "\xC4\xA5", "\x00\x00",
  /* U+8268 */	"\xF4\xB6", "\xF4\xB7", "\x00\x00", "\xF7\x46",
  /* U+826C */	"\xF7\xEF", "\xF8\xBB", "\xA6\xE1", "\xA8\x7D",
  /* U+8270 */	"\x00\x00", "\xC1\x7D", "\xA6\xE2", "\x00\x00",
  /* U+8274 */	"\xD7\x58", "\xDB\x5B", "\x99\xAF", "\xC6\x41",
  /* U+8278 */	"\xCA\x4A", "\x99\x4A", "\x89\x76", "\x8F\x48",
  /* U+827C */	"\xCA\x4B", "\xCA\x4D", "\xA6\xE3", "\xCA\x4E",
  /* U+8280 */	"\xCA\x4C", "\x00\x00", "\x00\x00", "\xCB\xA2",
  /* U+8284 */	"\xCB\xA3", "\xCB\x7B", "\x00\x00", "\xFB\xEE",
  /* U+8288 */	"\x00\x00", "\x00\x00", "\xCB\xA1", "\xA8\xA1",
  /* U+828C */	"\x00\x00", "\xA8\xA2", "\xCB\x7C", "\xCB\x7A",
  /* U+8290 */	"\xCB\x79", "\xCB\x7D", "\xA8\x7E", "\xCB\x7E",
  /* U+8294 */	"\xD0\x6A", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+8298 */	"\xCD\xB6", "\xAA\xDC", "\xCD\xB5", "\xCD\xB7",
  /* U+829C */	"\x00\x00", "\xAA\xDB", "\xCD\xBC", "\xAA\xDF",
  /* U+82A0 */	"\xCD\xB2", "\xCD\xC0", "\xCD\xC6", "\xAA\xE6",
  /* U+82A4 */	"\xCD\xC3", "\xAA\xE3", "\x99\xAE", "\xCD\xB9",
  /* U+82A8 */	"\xCD\xBF", "\xCD\xC1", "\x8E\xFB", "\xCD\xB4",
  /* U+82AC */	"\xAA\xE2", "\xAA\xDD", "\xCD\xBA", "\xAA\xE4",
  /* U+82B0 */	"\xAA\xE7", "\xAA\xE1", "\x00\x00", "\xAA\xDA",
  /* U+82B4 */	"\xCD\xBE", "\xCD\xB8", "\xCD\xC5", "\xAA\xE9",
  /* U+82B8 */	"\xAA\xE5", "\xAA\xE0", "\xCD\xBD", "\xAF\xEC",
  /* U+82BC */	"\xCD\xBB", "\xAA\xDE", "\xAA\xE8", "\x8C\xD0",
  /* U+82C0 */	"\xCD\xB3", "\x00\x00", "\xCD\xC2", "\xCD\xC4",
  /* U+82C4 */	"\x8B\x52", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+82C8 */	"\x00\x00", "\x00\x00", "\x99\xB0", "\x00\x00",
  /* U+82CC */	"\x00\x00", "\x00\x00", "\x00\x00", "\x89\x77",
  /* U+82D0 */	"\x8F\x41", "\xAD\x62", "\xAD\x5C", "\xAD\x64",
  /* U+82D4 */	"\xAD\x61", "\xD0\x71", "\xD0\x74", "\xAD\x5D",
  /* U+82D8 */	"\x99\xB1", "\xD0\x6B", "\x00\x00", "\xAD\x56",
  /* U+82DC */	"\xAD\x60", "\x00\x00", "\xAD\x63", "\xAD\x65",
  /* U+82E0 */	"\xD0\xA2", "\xD0\x77", "\x8F\x49", "\xAD\x55",
  /* U+82E4 */	"\xD0\xA1", "\xAD\x59", "\xAD\x57", "\xAD\x52",
  /* U+82E8 */	"\xD0\x6F", "\x00\x00", "\xD0\x7E", "\xD0\x73",
  /* U+82EC */	"\xD0\x76", "\xD0\xA5", "\xFA\x4D", "\xAD\x66",
  /* U+82F0 */	"\xD0\x7D", "\xAD\x5E", "\xD0\x78", "\xD0\xA4",
  /* U+82F4 */	"\xD0\x75", "\xD0\x79", "\xD0\x7C", "\x9D\xE4",
  /* U+82F8 */	"\x8C\xB5", "\xD0\x6D", "\xD0\xA3", "\xD0\x7B",
  /* U+82FC */	"\xFB\xE9", "\x9B\x54", "\xD0\x6C", "\x99\xB2",
  /* U+8300 */	"\xD0\x70", "\xAD\x5F", "\xAD\x5A", "\xAD\x53",
  /* U+8304 */	"\xAD\x58", "\xAD\x54", "\xAD\x67", "\xD0\x6E",
  /* U+8308 */	"\xD3\xA5", "\xAD\x5B", "\x00\x00", "\x9E\x68",
  /* U+830C */	"\xD0\x7A", "\xCE\x41", "\x00\x00", "\x00\x00",
  /* U+8310 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+8314 */	"\x00\x00", "\x00\x00", "\xD3\xA8", "\xAF\xFA",
  /* U+8318 */	"\x8F\x4A", "\xD3\x76", "\x8F\x42", "\xD3\xA3",
  /* U+831C */	"\xD3\x7D", "\x8F\x51", "\xD3\xB2", "\x00\x00",
  /* U+8320 */	"\xD3\xAA", "\x00\x00", "\xD3\x7E", "\x00\x00",
  /* U+8324 */	"\xD3\xA9", "\xD3\x78", "\xD3\x7C", "\xD3\xB5",
  /* U+8328 */	"\xAF\xFD", "\xD3\xAD", "\xD3\xA4", "\xAF\xED",
  /* U+832C */	"\xD3\xB3", "\xD3\x74", "\x00\x00", "\xD3\xAC",
  /* U+8330 */	"\x00\x00", "\xAF\xFC", "\xAF\xF7", "\xD3\x73",
  /* U+8334 */	"\xAF\xF5", "\xAF\xF4", "\xAF\xF9", "\xD3\xAB",
  /* U+8338 */	"\xAF\xF1", "\xAF\xF8", "\xD0\x72", "\xDB\x5C",
  /* U+833C */	"\xD3\xA6", "\x98\x46", "\x00\x00", "\xD3\x7A",
  /* U+8340 */	"\xAF\xFB", "\xD3\x7B", "\xD3\xA1", "\xAF\xFE",
  /* U+8344 */	"\xD3\x75", "\xD3\xAF", "\x00\x00", "\xD3\xAE",
  /* U+8348 */	"\xD3\xB6", "\xAF\xF3", "\xAF\xF0", "\xD3\xB4",
  /* U+834C */	"\xD3\xB0", "\xD3\xA7", "\xD3\xA2", "\xAF\xF6",
  /* U+8350 */	"\xAF\xF2", "\xD3\x77", "\xAF\xEE", "\xD3\xB1",
  /* U+8354 */	"\xAF\xEF", "\x00\x00", "\xD3\x79", "\x99\xB4",
  /* U+8358 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+835C */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+8360 */	"\x00\x00", "\x00\x00", "\x8E\xF5", "\xFD\x55",
  /* U+8364 */	"\x00\x00", "\x00\x00", "\x9C\xCD", "\x00\x00",
  /* U+8368 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+836C */	"\x00\x00", "\x00\x00", "\x00\x00", "\x89\x78",
  /* U+8370 */	"\x00\x00", "\x00\x00", "\x00\x00", "\xD7\x5E",
  /* U+8374 */	"\xD7\x60", "\xD7\x65", "\xD7\x79", "\xB2\xFC",
  /* U+8378 */	"\xB2\xF2", "\x00\x00", "\xD7\x5D", "\xB2\xFD",
  /* U+837C */	"\xB2\xFE", "\xD7\x68", "\xD7\x6F", "\xD7\x75",
  /* U+8380 */	"\x00\x00", "\xD7\x62", "\x00\x00", "\xD7\x69",
  /* U+8384 */	"\x00\x00", "\x8F\x53", "\xB3\x40", "\xD7\x77",
  /* U+8388 */	"\xD7\x72", "\xB2\xFA", "\xB2\xF8", "\xD7\x6E",
  /* U+838C */	"\xD7\x6A", "\xD7\x5C", "\xB2\xEF", "\xD7\x61",
  /* U+8390 */	"\xD7\x59", "\x8F\x6F", "\xB2\xF7", "\xB2\xF9",
  /* U+8394 */	"\xD7\x66", "\xD7\x63", "\xB2\xF4", "\xD7\x73",
  /* U+8398 */	"\xB2\xF1", "\xD7\x64", "\xD7\x7A", "\xD7\x6C",
  /* U+839C */	"\x8E\x63", "\xD7\x6B", "\xB2\xF0", "\x00\x00",
  /* U+83A0 */	"\xB2\xFB", "\x00\x00", "\xB2\xF3", "\xD7\x5A",
  /* U+83A4 */	"\xD7\x5F", "\xD7\x70", "\xD7\x76", "\xB3\x41",
  /* U+83A8 */	"\xD7\x5B", "\xD7\x67", "\xD7\x6D", "\xB2\xF6",
  /* U+83AC */	"\x8F\x56", "\x00\x00", "\xD7\x78", "\xD7\x71",
  /* U+83B0 */	"\xD7\x74", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+83B4 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+83B8 */	"\x00\x00", "\xFE\x76", "\x00\x00", "\x00\x00",
  /* U+83BC */	"\x00\x00", "\xB2\xF5", "\x9F\xC6", "\xDB\x6C",
  /* U+83C0 */	"\xDB\x60", "\xB5\xD7", "\xDB\x7D", "\xDB\xA7",
  /* U+83C4 */	"\xDB\xAA", "\xB5\xD5", "\xDB\x68", "\xDB\xA3",
  /* U+83C8 */	"\xDB\x69", "\xDB\x77", "\xB5\xE2", "\xDB\x73",
  /* U+83CC */	"\xB5\xDF", "\xFA\xAC", "\xDB\x74", "\xDB\x5D",
  /* U+83D0 */	"\x00\x00", "\xDB\xA4", "\x00\x00", "\x8F\x58",
  /* U+83D4 */	"\xB5\xE8", "\xDB\xA1", "\xDB\x75", "\xDB\xAC",
  /* U+83D8 */	"\xDB\x70", "\xDF\xC8", "\x00\x00", "\xDB\xAF",
  /* U+83DC */	"\xB5\xE6", "\xDB\x6E", "\xDB\x7A", "\xB5\xE9",
  /* U+83E0 */	"\xB5\xD4", "\xDB\x72", "\xDB\xAD", "\xDB\x6B",
  /* U+83E4 */	"\xDB\x64", "\xDB\x6F", "\x00\x00", "\xDB\x63",
  /* U+83E8 */	"\xDB\x61", "\xB5\xD0", "\xDB\xA5", "\xDB\x6A",
  /* U+83EC */	"\xDB\xA8", "\x98\x48", "\xDB\xA9", "\xB5\xD8",
  /* U+83F0 */	"\xB5\xDD", "\xB5\xD9", "\xB5\xE1", "\xDB\x7E",
  /* U+83F4 */	"\xB5\xDA", "\xDB\x76", "\xDB\x66", "\x00\x00",
  /* U+83F8 */	"\xB5\xD2", "\xDB\x5E", "\xDB\xA2", "\xDB\xAB",
  /* U+83FC */	"\xDB\x65", "\xB5\xE0", "\xDB\xB0", "\xDB\x71",
  /* U+8400 */	"\x00\x00", "\xDB\x6D", "\x00\x00", "\xB5\xD1",
  /* U+8404 */	"\xB5\xE5", "\x99\xB7", "\xDB\x7C", "\xB5\xE7",
  /* U+8408 */	"\x00\x00", "\xDB\x78", "\xB5\xDC", "\xB5\xD6",
  /* U+840C */	"\xB5\xDE", "\xB5\xD3", "\xB5\xE4", "\xDB\x79",
  /* U+8410 */	"\xDB\x67", "\xDB\x7B", "\xDB\x62", "\xDB\xA6",
  /* U+8414 */	"\x96\x65", "\x00\x00", "\xFA\x6C", "\x00\x00",
  /* U+8418 */	"\x9D\xE7", "\x00\x00", "\x00\x00", "\xDB\xAE",
  /* U+841C */	"\x9E\x62", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+8420 */	"\x96\xCC", "\x8E\x67", "\x00\x00", "\xDB\x5F",
  /* U+8424 */	"\xFC\x75", "\x00\x00", "\x98\x7E", "\x00\x00",
  /* U+8428 */	"\x00\x00", "\xDF\xC7", "\x00\x00", "\xDF\xDD",
  /* U+842C */	"\xB8\x55", "\xDF\xCC", "\xFD\xB9", "\xDF\xCA",
  /* U+8430 */	"\xDF\xB5", "\xB8\xA9", "\xDF\xC5", "\xDF\xD9",
  /* U+8434 */	"\xDF\xC1", "\xB8\xB1", "\xDF\xD8", "\xDF\xBF",
  /* U+8438 */	"\xB5\xE3", "\xDF\xCF", "\xDF\xC0", "\xDF\xD6",
  /* U+843C */	"\xB8\xB0", "\xB8\xA8", "\x97\xFC", "\xDF\xAA",
  /* U+8440 */	"\xDF\xB2", "\x00\x00", "\xDF\xCB", "\xDF\xC3",
  /* U+8444 */	"\xDF\xDC", "\xDF\xC6", "\xB8\xB6", "\xDF\xD7",
  /* U+8448 */	"\x98\xF9", "\xB8\xAD", "\x8F\x66", "\xDF\xC9",
  /* U+844C */	"\xDF\xD1", "\xDF\xB6", "\xDF\xD0", "\x00\x00",
  /* U+8450 */	"\xDF\xE1", "\xDF\xB1", "\xDF\xD2", "\x95\x6E",
  /* U+8454 */	"\xDF\xDF", "\x92\x45", "\xDF\xAB", "\xB5\xDB",
  /* U+8458 */	"\x8F\x60", "\xDF\xB9", "\xDF\xB8", "\xB8\xAF",
  /* U+845C */	"\x9E\xD1", "\xDF\xBC", "\xDF\xBE", "\xDF\xCD",
  /* U+8460 */	"\xDF\xDE", "\xB8\xB2", "\xFE\xCD", "\xB8\xB3",
  /* U+8464 */	"\x99\xB9", "\xDF\xB0", "\xB8\xAB", "\xDF\xB4",
  /* U+8468 */	"\xDF\xDA", "\xB8\xB4", "\x00\x00", "\xB8\xAC",
  /* U+846C */	"\xB8\xAE", "\xB8\xB5", "\xDF\xE0", "\xDF\xD3",
  /* U+8470 */	"\xDF\xCE", "\x8F\x62", "\x97\x4C", "\xDF\xBB",
  /* U+8474 */	"\xDF\xBA", "\xB8\xAA", "\xDF\xAC", "\xB8\xA7",
  /* U+8478 */	"\xDF\xC4", "\xDF\xAD", "\xDF\xC2", "\x00\x00",
  /* U+847C */	"\x00\x00", "\xDF\xB7", "\xDF\xDB", "\x91\xC7",
  /* U+8480 */	"\x95\x5F", "\x00\x00", "\xB8\xA6", "\x00\x00",
  /* U+8484 */	"\x00\x00", "\x00\x00", "\xDF\xB3", "\x00\x00",
  /* U+8488 */	"\x99\xBB", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+848C */	"\x00\x00", "\xDF\xAF", "\xDF\xD5", "\xDF\xAE",
  /* U+8490 */	"\xBB\x60", "\xE3\xD3", "\x8E\x6D", "\x8F\x71",
  /* U+8494 */	"\xE3\xC2", "\x00\x00", "\x94\xCB", "\xE3\xAC",
  /* U+8498 */	"\xE3\xCA", "\xBB\x58", "\xE3\xBB", "\xE3\xC5",
  /* U+849C */	"\xBB\x5B", "\xE3\xBE", "\xBB\x59", "\xE3\xAF",
  /* U+84A0 */	"\xE3\xCD", "\xE3\xAE", "\xE3\xC1", "\x95\xB1",
  /* U+84A4 */	"\xE3\xAD", "\x00\x00", "\x00\x00", "\xE3\xBF",
  /* U+84A8 */	"\xE3\xC8", "\xE3\xC6", "\xE3\xBA", "\xE3\xB5",
  /* U+84AC */	"\xE3\xB3", "\x9A\xF2", "\xE3\xB4", "\xE3\xC7",
  /* U+84B0 */	"\xE3\xD2", "\xE3\xBC", "\xBB\x5A", "\x00\x00",
  /* U+84B4 */	"\xE3\xB7", "\x00\x00", "\xE3\xCB", "\x00\x00",
  /* U+84B8 */	"\xBB\x5D", "\xE3\xB6", "\xE3\xB0", "\xE3\xC0",
  /* U+84BC */	"\xBB\x61", "\x96\xC3", "\x99\xBD", "\xBB\x55",
  /* U+84C0 */	"\xBB\x5E", "\xE3\xB8", "\xE3\xB2", "\x00\x00",
  /* U+84C4 */	"\xBB\x57", "\xDF\xD4", "\xBB\x56", "\xE3\xC3",
  /* U+84C8 */	"\x00\x00", "\xBB\x54", "\xBB\x63", "\xBB\x5C",
  /* U+84CC */	"\xE3\xC4", "\xE3\xB9", "\xE3\xB1", "\xE3\xCC",
  /* U+84D0 */	"\xE3\xBD", "\xBB\x62", "\xE3\xD0", "\xBB\x5F",
  /* U+84D4 */	"\xE3\xCF", "\x00\x00", "\xE3\xC9", "\xE3\xCE",
  /* U+84D8 */	"\x00\x00", "\x00\x00", "\xA0\xCF", "\xE3\xD1",
  /* U+84DC */	"\x00\x00", "\x00\x00", "\x8F\x6D", "\x00\x00",
  /* U+84E0 */	"\x00\x00", "\x99\xBE", "\x8E\xF4", "\x00\x00",
  /* U+84E4 */	"\x8F\x72", "\x95\xE4", "\x00\x00", "\xE7\x73",
  /* U+84E8 */	"\xE7\x74", "\xE7\x67", "\xE7\x66", "\xE7\x62",
  /* U+84EC */	"\xBD\xB4", "\x00\x00", "\xBD\xAC", "\xE7\x76",
  /* U+84F0 */	"\xE7\x75", "\xDF\xA9", "\xE7\x5F", "\xE7\x63",
  /* U+84F4 */	"\xE7\x5D", "\x00\x00", "\xE7\x70", "\xE7\x61",
  /* U+84F8 */	"\x99\xBF", "\xE7\x77", "\xE7\x5A", "\xE7\x58",
  /* U+84FC */	"\xE7\x64", "\xE7\x6E", "\xE7\x69", "\xBD\xB6",
  /* U+8500 */	"\xE7\x4F", "\x00\x00", "\xE7\x6D", "\x92\x42",
  /* U+8504 */	"\x00\x00", "\xFB\xA5", "\xBD\xB7", "\xDF\xBD",
  /* U+8508 */	"\xE7\x5B", "\xE7\x52", "\xE7\x55", "\xE7\x7B",
  /* U+850C */	"\xE7\x5C", "\xE7\x53", "\xE7\x51", "\xE7\x4E",
  /* U+8510 */	"\x99\xC0", "\xBD\xB0", "\xE7\x65", "\xBD\xAF",
  /* U+8514 */	"\xBD\xB3", "\xE7\x60", "\xE7\x68", "\xBD\xA9",
  /* U+8518 */	"\xE7\x78", "\xE7\x7C", "\xBD\xAB", "\x00\x00",
  /* U+851C */	"\xE7\x57", "\xE7\x6B", "\xE7\x6F", "\xE7\x54",
  /* U+8520 */	"\xE7\x79", "\xBD\xB2", "\x00\x00", "\xBD\xB1",
  /* U+8524 */	"\xE7\x4C", "\xBD\xB5", "\xE7\x72", "\xE7\x56",
  /* U+8528 */	"\xE7\x6A", "\xE7\x50", "\xE7\x5E", "\xE7\x59",
  /* U+852C */	"\xBD\xAD", "\xBD\xAE", "\xE7\x6C", "\xE7\x7D",
  /* U+8530 */	"\xE7\x7A", "\xE7\x71", "\x00\x00", "\xFD\xB4",
  /* U+8534 */	"\x8F\x77", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+8538 */	"\x99\xC1", "\x00\x00", "\x00\x00", "\xE7\x4D",
  /* U+853C */	"\x00\x00", "\xBD\xAA", "\xEB\x49", "\x00\x00",
  /* U+8540 */	"\xEB\x40", "\xEB\x43", "\xFA\xB9", "\xBF\xBB",
  /* U+8544 */	"\xEB\x45", "\xEA\xF9", "\xEB\x41", "\xEB\x47",
  /* U+8548 */	"\xBF\xB8", "\xBF\xBC", "\xBF\xB6", "\x8F\x40",
  /* U+854C */	"\xFA\x44", "\xEA\xFB", "\xEB\x4C", "\x00\x00",
  /* U+8550 */	"\x00\x00", "\xEB\x46", "\x99\xC2", "\xEA\xFC",
  /* U+8554 */	"\xEB\x55", "\xEB\x4F", "\xEA\xF8", "\xEE\x46",
  /* U+8558 */	"\xEA\xFE", "\xBF\xB7", "\x8F\x5C", "\xEB\x4A",
  /* U+855C */	"\x00\x00", "\xEB\x54", "\xBF\xBF", "\x8C\xBD",
  /* U+8560 */	"\xEB\x51", "\xEA\xFD", "\xEB\x44", "\xEB\x48",
  /* U+8564 */	"\xEB\x42", "\xEB\x56", "\xEB\x53", "\xEB\x50",
  /* U+8568 */	"\xBF\xB9", "\xBF\xBA", "\xBF\xBE", "\xEA\xFA",
  /* U+856C */	"\xEB\x57", "\xBF\xBD", "\xEB\x4D", "\x99\xC4",
  /* U+8570 */	"\x99\xC5", "\xEB\x4B", "\x00\x00", "\x8F\x7B",
  /* U+8574 */	"\x00\x00", "\xEB\x4E", "\xEE\x53", "\xEE\x40",
  /* U+8578 */	"\xEE\x45", "\xEE\x52", "\xEE\x44", "\xED\xFB",
  /* U+857C */	"\xEE\x41", "\x00\x00", "\xC1\xA2", "\x00\x00",
  /* U+8580 */	"\xED\xF4", "\xEE\x4D", "\xEE\x4F", "\xED\xF3",
  /* U+8584 */	"\xC1\xA1", "\xEE\x51", "\xEE\x49", "\xC1\xA8",
  /* U+8588 */	"\xEE\x50", "\xEE\x42", "\xC1\xAA", "\xED\xF9",
  /* U+858C */	"\xEB\x52", "\xEE\x4A", "\xEE\x47", "\xED\xF5",
  /* U+8590 */	"\xEE\x55", "\xC1\xA4", "\x00\x00", "\x00\x00",
  /* U+8594 */	"\xC1\xA5", "\xED\xF7", "\xEE\x48", "\x8C\xB6",
  /* U+8598 */	"\xEE\x54", "\xEE\x4B", "\xED\xFD", "\xC1\xA7",
  /* U+859C */	"\xC1\xA3", "\xEE\x4C", "\xED\xFE", "\xEE\x56",
  /* U+85A0 */	"\xED\xF8", "\xEE\x43", "\xEE\x4E", "\xED\xFA",
  /* U+85A4 */	"\xED\xFC", "\x00\x00", "\xC2\xCB", "\xED\xF6",
  /* U+85A8 */	"\xC1\xA9", "\xC2\xC4", "\xC1\x7E", "\x00\x00",
  /* U+85AC */	"\x00\x00", "\x00\x00", "\x00\x00", "\xC1\xA6",
  /* U+85B0 */	"\xC2\xC8", "\xF0\xB3", "\x00\x00", "\xF0\xA9",
  /* U+85B4 */	"\xF0\xA4", "\xF0\xAA", "\xF0\xB4", "\xF0\xB8",
  /* U+85B8 */	"\xF0\xB7", "\xC2\xCA", "\xC2\xC9", "\x00\x00",
  /* U+85BC */	"\x00\x00", "\xF0\xAB", "\xF0\xB9", "\xF0\xAE",
  /* U+85C0 */	"\xF0\xA6", "\x8F\xA3", "\xF0\xA8", "\xF0\xA7",
  /* U+85C4 */	"\xF0\xAD", "\xF0\xB2", "\xF0\xA5", "\xF0\xAC",
  /* U+85C8 */	"\xF0\xB1", "\xC2\xC7", "\x00\x00", "\xF0\xAF",
  /* U+85CC */	"\x00\x00", "\xC2\xC5", "\xF0\xB0", "\xC2\xC3",
  /* U+85D0 */	"\xC2\xC6", "\xF2\xD5", "\xF0\xB5", "\x00\x00",
  /* U+85D4 */	"\x00\x00", "\xC3\xC2", "\x8C\xCE", "\xF2\xCD",
  /* U+85D8 */	"\xF2\xD1", "\xF2\xC9", "\xF2\xCC", "\x00\x00",
  /* U+85DC */	"\xF2\xD4", "\xC3\xC0", "\xF2\xD9", "\xF2\xD2",
  /* U+85E0 */	"\x99\xC6", "\xF2\xCA", "\xF2\xDA", "\xF2\xD3",
  /* U+85E4 */	"\xC3\xC3", "\xC3\xC4", "\xF2\xD7", "\x00\x00",
  /* U+85E8 */	"\xF2\xCB", "\xC3\xBF", "\xC3\xC1", "\xF2\xC6",
  /* U+85EC */	"\xF2\xCE", "\xF2\xC8", "\x96\xCD", "\xF2\xD8",
  /* U+85F0 */	"\xF2\xD6", "\xF2\xC7", "\xF2\xCF", "\x00\x00",
  /* U+85F4 */	"\x00\x00", "\x00\x00", "\xF4\xBE", "\xC3\xC5",
  /* U+85F8 */	"\xF2\xD0", "\xC4\xA7", "\xC4\xA9", "\xC4\xA6",
  /* U+85FC */	"\x96\xC7", "\xF4\xC3", "\xF4\xBB", "\xF4\xB9",
  /* U+8600 */	"\xF4\xBD", "\xF4\xBA", "\x8F\xA5", "\x00\x00",
  /* U+8604 */	"\xF4\xBF", "\xF4\xC1", "\xC4\xAA", "\xC4\xAC",
  /* U+8608 */	"\x00\x00", "\xF4\xC0", "\xC4\xAD", "\xC4\xAB",
  /* U+860C */	"\xF4\xC2", "\xFA\xBB", "\x00\x00", "\x8C\x61",
  /* U+8610 */	"\x95\x70", "\xC4\xA8", "\x00\x00", "\x00\x00",
  /* U+8614 */	"\x93\x68", "\x00\x00", "\x8F\x7E", "\xC4\xF4",
  /* U+8618 */	"\xF5\xF1", "\xF5\xF7", "\xC4\xF6", "\xF4\xBC",
  /* U+861C */	"\xF5\xF6", "\x00\x00", "\xF5\xFD", "\xF5\xF4",
  /* U+8620 */	"\xF5\xFB", "\xF5\xFA", "\xF4\xB8", "\xF5\xF5",
  /* U+8624 */	"\xF0\xB6", "\xF5\xFE", "\xF5\xF3", "\xF5\xF8",
  /* U+8628 */	"\x8F\xAA", "\xF5\xFC", "\xF5\xF2", "\x00\x00",
  /* U+862C */	"\xF7\x4A", "\xC4\xF5", "\xF5\xF9", "\xA0\x50",
  /* U+8630 */	"\x00\x00", "\xF7\xF4", "\xF7\x4B", "\xF7\x49",
  /* U+8634 */	"\xF7\x47", "\xF7\x48", "\xF7\x4C", "\x00\x00",
  /* U+8638 */	"\xC5\xD9", "\xF7\xF2", "\xF7\xF0", "\xF7\xF5",
  /* U+863C */	"\xF7\xF3", "\x00\x00", "\xF7\xF6", "\xC5\xDA",
  /* U+8640 */	"\xF7\xF1", "\x00\x00", "\x90\xD3", "\xF8\xBC",
  /* U+8644 */	"\x00\x00", "\x95\x56", "\xF9\x45", "\xF9\x46",
  /* U+8648 */	"\xF9\x47", "\x00\x00", "\x00\x00", "\xF9\xC7",
  /* U+864C */	"\xF9\xBD", "\xCA\x4F", "\xAA\xEA", "\x00\x00",
  /* U+8650 */	"\xAD\x68", "\x00\x00", "\xD3\xB8", "\xD3\xB7",
  /* U+8654 */	"\xB0\x40", "\xB3\x42", "\xD7\x7C", "\x00\x00",
  /* U+8658 */	"\x00\x00", "\xD7\x7B", "\x00\x00", "\xB5\xEA",
  /* U+865C */	"\xB8\xB8", "\x00\x00", "\xB8\xB7", "\xB8\xB9",
  /* U+8660 */	"\x00\x00", "\xE3\xD4", "\xE7\x7E", "\xEB\x58",
  /* U+8664 */	"\xEB\x5A", "\xEB\x59", "\x00\x00", "\xC1\xAB",
  /* U+8668 */	"\xEE\x57", "\xF0\xBA", "\xF9\xA5", "\xA6\xE4",
  /* U+866C */	"\x8F\xB8", "\xCD\xC9", "\xCD\xCA", "\xCD\xC8",
  /* U+8670 */	"\xCD\xC7", "\xAA\xEB", "\x99\xC8", "\xD0\xA9",
  /* U+8674 */	"\xD0\xA7", "\x00\x00", "\x00\x00", "\xD0\xA6",
  /* U+8678 */	"\x00\x00", "\xAD\x69", "\xAD\x6B", "\xAD\x6A",
  /* U+867C */	"\xD0\xA8", "\x00\x00", "\x8F\xAF", "\x00\x00",
  /* U+8680 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+8684 */	"\x00\x00", "\xD3\xC4", "\xD3\xC1", "\xD3\xBF",
  /* U+8688 */	"\x00\x00", "\x00\x00", "\xB0\x41", "\xD3\xC2",
  /* U+868C */	"\xB0\x46", "\xD3\xBC", "\xD3\xCB", "\x00\x00",
  /* U+8690 */	"\xD3\xCD", "\xD3\xBD", "\x99\xC9", "\xB0\x43",
  /* U+8694 */	"\xD3\xCE", "\xD3\xC9", "\xD3\xBB", "\xD3\xC0",
  /* U+8698 */	"\xD3\xCA", "\xD3\xC6", "\xD3\xC3", "\x00\x00",
  /* U+869C */	"\xB0\x48", "\xD3\xCC", "\xD3\xBE", "\x00\x00",
  /* U+86A0 */	"\x95\x79", "\xD3\xC7", "\xD3\xB9", "\xB0\x47",
  /* U+86A4 */	"\xB0\x44", "\xD3\xC5", "\x00\x00", "\xD3\xC8",
  /* U+86A8 */	"\xD3\xBA", "\xB0\x45", "\xB0\x42", "\x00\x00",
  /* U+86AC */	"\x00\x00", "\x9F\x49", "\x00\x00", "\xB3\x4C",
  /* U+86B0 */	"\xD7\xA5", "\xB3\x4B", "\x99\xCA", "\xD7\xA8",
  /* U+86B4 */	"\xD7\xAB", "\xB3\x48", "\xB3\x46", "\xD7\x7E",
  /* U+86B8 */	"\xD7\xA9", "\xD7\xA7", "\xD7\xA4", "\xD7\xAC",
  /* U+86BC */	"\xD7\xAD", "\xD7\xAF", "\xD7\xB0", "\xD7\x7D",
  /* U+86C0 */	"\xB3\x45", "\xD7\xA2", "\xD7\xA1", "\xD7\xAE",
  /* U+86C4 */	"\xB3\x47", "\xD7\xA3", "\xB3\x49", "\xB3\x44",
  /* U+86C8 */	"\xD7\xA6", "\xB3\x4D", "\x00\x00", "\xB3\x4A",
  /* U+86CC */	"\xD7\xAA", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+86D0 */	"\xB5\xF1", "\xDB\xBF", "\x00\x00", "\xDB\xB4",
  /* U+86D4 */	"\xB5\xEE", "\x00\x00", "\xDF\xE7", "\xDB\xBD",
  /* U+86D8 */	"\xDB\xB1", "\xB5\xEC", "\xDB\xB6", "\xB5\xEF",
  /* U+86DC */	"\xDB\xBA", "\xDB\xB8", "\xB5\xF2", "\xB5\xEB",
  /* U+86E0 */	"\x00\x00", "\x00\x00", "\xDB\xB2", "\xDB\xB5",
  /* U+86E4 */	"\xB5\xF0", "\x00\x00", "\xDB\xB3", "\x00\x00",
  /* U+86E8 */	"\xDB\xBE", "\xDB\xBC", "\xDB\xB7", "\xDB\xB9",
  /* U+86EC */	"\xDB\xBB", "\xB5\xED", "\x00\x00", "\x99\xCB",
  /* U+86F0 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+86F4 */	"\x00\x00", "\xDF\xE8", "\xDF\xEE", "\xDF\xE4",
  /* U+86F8 */	"\xDF\xEA", "\xB8\xBA", "\xDF\xE6", "\xB8\xC0",
  /* U+86FC */	"\x00\x00", "\x00\x00", "\xB8\xBF", "\x00\x00",
  /* U+8700 */	"\xB8\xBE", "\xDF\xED", "\xB8\xC1", "\xB8\xC2",
  /* U+8704 */	"\xDF\xE3", "\xDF\xF0", "\xB8\xC3", "\xB8\xBD",
  /* U+8708 */	"\xB8\xBC", "\xDF\xEC", "\xB8\xC4", "\xDF\xE2",
  /* U+870C */	"\xDF\xE5", "\xDF\xEF", "\xDF\xEB", "\x00\x00",
  /* U+8710 */	"\x00\x00", "\xE3\xF4", "\xE3\xE9", "\xB8\xBB",
  /* U+8714 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+8718 */	"\xBB\x6A", "\xE3\xDD", "\xE3\xF2", "\xE3\xDE",
  /* U+871C */	"\xBB\x65", "\x00\x00", "\xE3\xDB", "\x00\x00",
  /* U+8720 */	"\xE3\xE4", "\xE3\xDC", "\xBB\x67", "\xE3\xD6",
  /* U+8724 */	"\xE3\xF1", "\xBB\x68", "\xE3\xEE", "\xE3\xEF",
  /* U+8728 */	"\xE3\xD7", "\xBB\x6D", "\xE3\xE6", "\x00\x00",
  /* U+872C */	"\xE3\xE0", "\xE3\xE7", "\xE3\xDA", "\x00\x00",
  /* U+8730 */	"\xE3\xF3", "\xE3\xEB", "\xE3\xE5", "\xE3\xD5",
  /* U+8734 */	"\xBB\x69", "\xE3\xEC", "\x00\x00", "\xBB\x6C",
  /* U+8738 */	"\xE3\xF0", "\x00\x00", "\xE3\xEA", "\xBB\x66",
  /* U+873C */	"\xE3\xE8", "\x00\x00", "\xE3\xE2", "\xBB\x64",
  /* U+8740 */	"\xE3\xD9", "\xE3\xE1", "\xE3\xED", "\xE3\xDF",
  /* U+8744 */	"\x00\x00", "\x00\x00", "\xE3\xE3", "\x00\x00",
  /* U+8748 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+874C */	"\xBD\xC1", "\xDF\xE9", "\xE7\xB2", "\xE7\xBB",
  /* U+8750 */	"\xE7\xB1", "\xE7\xAD", "\xE7\xAA", "\xBD\xC2",
  /* U+8754 */	"\xE7\xA8", "\xBB\x6B", "\xE7\xA1", "\xBD\xC0",
  /* U+8758 */	"\xE7\xA7", "\xBD\xBF", "\xE7\xAC", "\xE7\xA9",
  /* U+875C */	"\xE7\xB9", "\xE7\xB4", "\xE7\xAE", "\xE7\xB3",
  /* U+8760 */	"\xBD\xBB", "\xE7\xAB", "\xE7\xBE", "\xE7\xA2",
  /* U+8764 */	"\xE7\xA3", "\xE7\xBA", "\xBD\xBC", "\xE7\xBF",
  /* U+8768 */	"\xBD\xBE", "\xE7\xC0", "\xE7\xB0", "\xE3\xD8",
  /* U+876C */	"\xE7\xB6", "\xE7\xAF", "\xE7\xB8", "\xE7\xB5",
  /* U+8770 */	"\x9D\xD5", "\x8F\xB0", "\x00\x00", "\xE7\xA6",
  /* U+8774 */	"\xBD\xB9", "\xE7\xBD", "\xBD\xBA", "\xE7\xA4",
  /* U+8778 */	"\xBD\xBD", "\xEB\x64", "\xE7\xB7", "\xE7\xBC",
  /* U+877C */	"\x00\x00", "\xFA\x7A", "\x00\x00", "\x00\x00",
  /* U+8780 */	"\x00\x00", "\xEB\x61", "\xBD\xB8", "\xBF\xC0",
  /* U+8784 */	"\xEB\x6B", "\xEB\x67", "\x9E\x5F", "\xEB\x65",
  /* U+8788 */	"\xEB\x60", "\xEB\x6F", "\x00\x00", "\x99\xCD",
  /* U+878C */	"\xA0\xC9", "\xBF\xC4", "\x00\x00", "\xEB\x5C",
  /* U+8790 */	"\xEB\x68", "\xEB\x69", "\xEB\x5F", "\xEB\x5E",
  /* U+8794 */	"\xEB\x6C", "\x00\x00", "\xEB\x62", "\xEB\x5D",
  /* U+8798 */	"\xEB\x63", "\x00\x00", "\xEB\x6E", "\xEB\x5B",
  /* U+879C */	"\xEB\x6D", "\xEB\x6A", "\xBF\xC2", "\xBF\xC1",
  /* U+87A0 */	"\x00\x00", "\x00\x00", "\xBF\xC3", "\xEB\x66",
  /* U+87A4 */	"\xF0\xCB", "\x9A\xDB", "\x00\x00", "\x00\x00",
  /* U+87A8 */	"\x00\x00", "\xA0\xC6", "\xEE\x59", "\xC1\xB1",
  /* U+87AC */	"\xEE\x5D", "\xEE\x5A", "\xEE\x61", "\xEE\x67",
  /* U+87B0 */	"\xEE\x5C", "\x8F\xB4", "\xEE\x70", "\xC1\xAE",
  /* U+87B4 */	"\xEE\x6A", "\xEE\x5F", "\xEE\x6B", "\xEE\x66",
  /* U+87B8 */	"\xEE\x6D", "\xEE\x5E", "\xC1\xB3", "\xC1\xB2",
  /* U+87BC */	"\xEE\x60", "\xEE\x6E", "\xEE\x58", "\xEE\x6C",
  /* U+87C0 */	"\xC1\xAC", "\xA0\xD7", "\xEE\x64", "\xEE\x63",
  /* U+87C4 */	"\xEE\x68", "\xEE\x5B", "\xC1\xB0", "\x00\x00",
  /* U+87C8 */	"\xC1\xB4", "\xEE\x62", "\xEE\x69", "\xC1\xB5",
  /* U+87CC */	"\xEE\x65", "\x00\x00", "\xA0\xC7", "\x00\x00",
  /* U+87D0 */	"\x00\x00", "\xC1\xAD", "\xC1\xAF", "\xF0\xC7",
  /* U+87D4 */	"\xF0\xC5", "\x00\x00", "\xA0\x43", "\xF0\xCC",
  /* U+87D8 */	"\xF0\xC9", "\xF0\xCD", "\x8F\xB5", "\xF0\xBE",
  /* U+87DC */	"\xF0\xC6", "\xF0\xD1", "\xEE\x6F", "\xF0\xC2",
  /* U+87E0 */	"\xC2\xCF", "\xE7\xA5", "\xF0\xBD", "\xF0\xCA",
  /* U+87E4 */	"\xF0\xC4", "\xF0\xC1", "\xF0\xBC", "\xF0\xBB",
  /* U+87E8 */	"\xF0\xD0", "\x00\x00", "\xF0\xC0", "\xF0\xBF",
  /* U+87EC */	"\xC2\xCD", "\xF0\xC8", "\x8F\xB2", "\xC2\xCC",
  /* U+87F0 */	"\x00\x00", "\x00\x00", "\xC2\xCE", "\xF0\xC3",
  /* U+87F4 */	"\xF0\xCF", "\xA0\x61", "\xF2\xDE", "\xF2\xDF",
  /* U+87F8 */	"\x00\x00", "\xC3\xC9", "\xF2\xDC", "\xC3\xC6",
  /* U+87FC */	"\xF2\xE4", "\x00\x00", "\xC3\xCA", "\xF2\xE6",
  /* U+8800 */	"\xF2\xDB", "\xF0\xCE", "\xF2\xE8", "\xF2\xDD",
  /* U+8804 */	"\x9E\x5E", "\xC3\xC7", "\xF2\xE3", "\x00\x00",
  /* U+8808 */	"\xF2\xE5", "\xF2\xE0", "\xF2\xE7", "\xF2\xE2",
  /* U+880C */	"\xF2\xE1", "\xC3\xC8", "\x00\x00", "\x8F\xB6",
  /* U+8810 */	"\xF4\xC5", "\xF4\xC6", "\x00\x00", "\xF4\xC8",
  /* U+8814 */	"\xC4\xAE", "\xC4\xAF", "\xF4\xC9", "\xF4\xC7",
  /* U+8818 */	"\x9F\xE8", "\xF4\xC4", "\x00\x00", "\xF6\x42",
  /* U+881C */	"\xF6\x45", "\xF6\x41", "\x00\x00", "\xC4\xFA",
  /* U+8820 */	"\xF6\x43", "\xC4\xF9", "\xC4\xF8", "\xC4\xF7",
  /* U+8824 */	"\xF6\x44", "\xF7\x51", "\xF7\x4F", "\x9C\xB2",
  /* U+8828 */	"\xF7\x4E", "\xF6\x40", "\xF7\x50", "\xF6\x46",
  /* U+882C */	"\xF7\x4D", "\x95\x7C", "\xF7\xF9", "\xF7\xD7",
  /* U+8830 */	"\xF7\xF7", "\xC5\xDB", "\xF7\xF8", "\xF7\xFA",
  /* U+8834 */	"\x00\x00", "\xF8\xBF", "\xC5\xFA", "\xF8\xBE",
  /* U+8838 */	"\xF8\xBD", "\xC5\xFB", "\x00\x00", "\xC6\x5A",
  /* U+883C */	"\xF9\x6E", "\xF9\xA7", "\xF9\xA6", "\xF9\xA8",
  /* U+8840 */	"\xA6\xE5", "\xD0\xAA", "\x9F\xC7", "\xD3\xCF",
  /* U+8844 */	"\xD3\xD0", "\x8F\xBB", "\x8F\xBC", "\x00\x00",
  /* U+8848 */	"\xDB\xC0", "\x00\x00", "\xF6\x47", "\xF8\xC0",
  /* U+884C */	"\xA6\xE6", "\xAD\x6C", "\xD0\xAB", "\x8F\xEC",
  /* U+8850 */	"\x00\x00", "\x00\x00", "\xD7\xB1", "\xB3\x4E",
  /* U+8854 */	"\x00\x00", "\xDB\xC2", "\xDB\xC1", "\xB5\xF3",
  /* U+8858 */	"\x00\x00", "\xB8\xC5", "\xE7\xC1", "\xBD\xC3",
  /* U+885C */	"\x00\x00", "\xBD\xC4", "\x8F\xC0", "\x00\x00",
  /* U+8860 */	"\x93\x6A", "\xBF\xC5", "\xC5\xFC", "\xA6\xE7",
  /* U+8864 */	"\x8B\xE4", "\x9C\x7C", "\x00\x00", "\xD0\xAC",
  /* U+8868 */	"\xAA\xED", "\xD0\xAE", "\xD0\xAD", "\xAD\x6D",
  /* U+886C */	"\x00\x00", "\xD3\xD1", "\x95\xA1", "\xD3\xD8",
  /* U+8870 */	"\xB0\x49", "\xD3\xD6", "\xD3\xD4", "\x00\x00",
  /* U+8874 */	"\xD3\xDB", "\xD3\xD2", "\xD3\xD3", "\xB0\x4A",
  /* U+8878 */	"\x00\x00", "\xB0\x4E", "\x00\x00", "\x00\x00",
  /* U+887C */	"\xD3\xDC", "\xB0\x4D", "\xD3\xDA", "\xD3\xD7",
  /* U+8880 */	"\xD3\xD5", "\xB0\x4B", "\xB0\x4C", "\xD3\xD9",
  /* U+8884 */	"\xFE\xEC", "\x00\x00", "\x00\x00", "\x95\xA3",
  /* U+8888 */	"\xB3\x50", "\xD7\xB2", "\x00\x00", "\xB3\x55",
  /* U+888C */	"\xD7\xC2", "\xB3\x54", "\xD7\xC4", "\x8C\x45",
  /* U+8890 */	"\x8C\xB8", "\xD7\xB8", "\xB3\x52", "\xD7\xC3",
  /* U+8894 */	"\x00\x00", "\xD7\xB3", "\xB3\x53", "\xD7\xBF",
  /* U+8898 */	"\xD7\xBB", "\xD7\xBD", "\xD7\xB7", "\xD7\xBE",
  /* U+889C */	"\x8F\xC1", "\x00\x00", "\xB3\x4F", "\xD7\xBA",
  /* U+88A0 */	"\xA0\x52", "\xD7\xB9", "\xD7\xB5", "\x00\x00",
  /* U+88A4 */	"\xD7\xC0", "\x00\x00", "\x00\x00", "\xD7\xBC",
  /* U+88A8 */	"\xD7\xB4", "\x00\x00", "\xD7\xB6", "\xB3\x51",
  /* U+88AC */	"\xD7\xC1", "\x00\x00", "\x99\xD0", "\x00\x00",
  /* U+88B0 */	"\x00\x00", "\xB5\xF6", "\xDB\xCD", "\x00\x00",
  /* U+88B4 */	"\x8F\xC3", "\x8F\xC4", "\xDB\xC9", "\xDB\xCB",
  /* U+88B8 */	"\xDB\xC6", "\xDB\xC5", "\xDB\xC3", "\x00\x00",
  /* U+88BC */	"\xDB\xCA", "\xDB\xCC", "\xDB\xC8", "\x95\xA4",
  /* U+88C0 */	"\xDB\xC7", "\xB5\xF4", "\xB5\xF5", "\x00\x00",
  /* U+88C4 */	"\x00\x00", "\x8F\xC6", "\x00\x00", "\x9E\x60",
  /* U+88C8 */	"\x00\x00", "\xDB\xCF", "\xB8\xCD", "\xDF\xF2",
  /* U+88CC */	"\xDF\xF8", "\xDF\xF3", "\xDF\xF4", "\xF9\xD8",
  /* U+88D0 */	"\xDF\xF9", "\x00\x00", "\xB8\xCF", "\x00\x00",
  /* U+88D4 */	"\xB8\xC7", "\xB8\xCE", "\xDF\xF1", "\xDB\xC4",
  /* U+88D8 */	"\xB8\xCA", "\xB8\xC8", "\xDF\xF7", "\xDF\xF6",
  /* U+88DC */	"\xB8\xC9", "\xB8\xCB", "\xDF\xF5", "\xB8\xC6",
  /* U+88E0 */	"\x00\x00", "\xB8\xCC", "\x00\x00", "\x00\x00",
  /* U+88E4 */	"\x00\x00", "\x00\x00", "\x95\xA5", "\xE3\xF6",
  /* U+88E8 */	"\xBB\x74", "\x00\x00", "\x00\x00", "\xE4\x42",
  /* U+88EC */	"\xE4\x41", "\x00\x00", "\xE3\xFB", "\xBB\x76",
  /* U+88F0 */	"\xE4\x40", "\xE3\xF7", "\xE3\xF8", "\xBB\x6E",
  /* U+88F4 */	"\xBB\x70", "\x9C\xB3", "\xE3\xFD", "\xE3\xF5",
  /* U+88F8 */	"\xBB\x72", "\xBB\x71", "\xE3\xF9", "\xE3\xFE",
  /* U+88FC */	"\xE3\xFC", "\xBB\x73", "\xE3\xFA", "\x99\xD1",
  /* U+8900 */	"\xFE\xF1", "\xDB\xCE", "\xBB\x6F", "\x00\x00",
  /* U+8904 */	"\x00\x00", "\xE7\xC2", "\xE7\xC9", "\xBD\xC6",
  /* U+8908 */	"\x00\x00", "\xE7\xCD", "\xBD\xCA", "\xE7\xC5",
  /* U+890C */	"\xE7\xC3", "\x00\x00", "\xE7\xCC", "\x00\x00",
  /* U+8910 */	"\xBD\xC5", "\xE7\xCB", "\xBD\xC7", "\xBD\xC8",
  /* U+8914 */	"\xE7\xC4", "\xBD\xC9", "\xE7\xCA", "\xE7\xC6",
  /* U+8918 */	"\xE7\xC7", "\xE7\xC8", "\xBB\x75", "\x00\x00",
  /* U+891C */	"\x00\x00", "\x00\x00", "\xEB\x70", "\xEB\x7C",
  /* U+8920 */	"\x00\x00", "\xBF\xCA", "\xEB\x77", "\xEB\x79",
  /* U+8924 */	"\x99\xD2", "\xBF\xC8", "\xEB\x71", "\xEB\x75",
  /* U+8928 */	"\x00\x00", "\xEB\x78", "\xBF\xC6", "\xBF\xC9",
  /* U+892C */	"\xEB\x7B", "\xEB\x73", "\xEB\x74", "\xEB\x7A",
  /* U+8930 */	"\xEB\x72", "\xEB\x76", "\xBF\xC7", "\xEE\x72",
  /* U+8934 */	"\x00\x00", "\xEE\x71", "\xC1\xB7", "\xEE\x77",
  /* U+8938 */	"\xC1\xB9", "\x00\x00", "\x00\x00", "\xC1\xB6",
  /* U+893C */	"\xEE\x73", "\xC1\xBA", "\xEE\x74", "\x00\x00",
  /* U+8940 */	"\x00\x00", "\xEE\x75", "\xEE\x78", "\x9C\xC2",
  /* U+8944 */	"\xC1\xB8", "\x00\x00", "\xF0\xD6", "\x99\xD3",
  /* U+8948 */	"\x00\x00", "\xF0\xD9", "\x00\x00", "\xF0\xD3",
  /* U+894C */	"\xF0\xD5", "\x95\xA7", "\x00\x00", "\xF0\xD4",
  /* U+8950 */	"\xF0\xD7", "\xF0\xD8", "\xEE\x76", "\xF0\xD2",
  /* U+8954 */	"\x95\xA9", "\x00\x00", "\xC3\xCD", "\xF2\xEC",
  /* U+8958 */	"\xF2\xEF", "\xF2\xF1", "\xF2\xEA", "\xF2\xEB",
  /* U+895C */	"\xF2\xEE", "\xF2\xF0", "\xC3\xCE", "\xC3\xCC",
  /* U+8960 */	"\xC3\xCB", "\xF2\xED", "\xF2\xE9", "\xF4\xCA",
  /* U+8964 */	"\xC4\xB0", "\x95\xA6", "\xF4\xCB", "\x00\x00",
  /* U+8968 */	"\x00\x00", "\xF6\x49", "\xC4\xFB", "\xF6\x4B",
  /* U+896C */	"\xC4\xFC", "\xF6\x48", "\xF6\x4A", "\xC5\xA8",
  /* U+8970 */	"\x00\x00", "\xF7\x52", "\xC5\xA7", "\xF7\xFD",
  /* U+8974 */	"\xF7\xFC", "\x00\x00", "\xF7\xFB", "\x9C\x5D",
  /* U+8978 */	"\x00\x00", "\xF9\x48", "\xF9\x49", "\xF9\x4B",
  /* U+897C */	"\xF9\x4A", "\x00\x00", "\xCA\x50", "\xA6\xE8",
  /* U+8980 */	"\x98\xE2", "\xAD\x6E", "\xD7\xC5", "\xB5\xF7",
  /* U+8984 */	"\x00\x00", "\xDF\xFA", "\xC2\xD0", "\x8F\xC9",
  /* U+8988 */	"\xF2\xF2", "\xA0\xC2", "\x8F\xCA", "\xA8\xA3",
  /* U+898C */	"\x00\x00", "\x00\x00", "\x00\x00", "\xB3\x57",
  /* U+8990 */	"\x00\x00", "\x99\xD4", "\x00\x00", "\xB3\x56",
  /* U+8994 */	"\xA0\xB9", "\xDB\xD0", "\xB5\xF8", "\xDB\xD2",
  /* U+8998 */	"\xDB\xD1", "\x00\x00", "\x00\x00", "\xDF\xFB",
  /* U+899C */	"\xB8\xD0", "\xE4\x43", "\xE4\x46", "\xE4\x45",
  /* U+89A0 */	"\x00\x00", "\xE4\x44", "\xE7\xCE", "\xE7\xD0",
  /* U+89A4 */	"\xE7\xCF", "\x9B\x58", "\xBF\xCC", "\x8F\xCD",
  /* U+89A8 */	"\x00\x00", "\xA0\xD4", "\xBF\xCB", "\x00\x00",
  /* U+89AC */	"\xC1\xBB", "\xEE\x79", "\xEE\x7B", "\xEE\x7A",
  /* U+89B0 */	"\x00\x00", "\x00\x00", "\xC2\xD1", "\x00\x00",
  /* U+89B4 */	"\x00\x00", "\x00\x00", "\xF2\xF4", "\xF2\xF3",
  /* U+89B8 */	"\x00\x00", "\xF4\xCC", "\xC4\xB1", "\x00\x00",
  /* U+89BC */	"\x8F\xCE", "\xC4\xFD", "\xF7\x54", "\xF7\x53",
  /* U+89C0 */	"\xC6\x5B", "\x8B\xE5", "\x00\x00", "\x00\x00",
  /* U+89C4 */	"\x00\x00", "\x00\x00", "\x89\x79", "\x00\x00",
  /* U+89C8 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+89CC */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+89D0 */	"\x00\x00", "\x00\x00", "\xA8\xA4", "\xD0\xAF",
  /* U+89D4 */	"\xAD\x6F", "\xD7\xC8", "\xD7\xC6", "\x00\x00",
  /* U+89D8 */	"\x00\x00", "\xD7\xC7", "\xDB\xD4", "\xDB\xD5",
  /* U+89DC */	"\xE0\x43", "\xDB\xD3", "\x00\x00", "\xDF\xFC",
  /* U+89E0 */	"\xE0\x41", "\xE0\x40", "\xE0\x42", "\xB8\xD1",
  /* U+89E4 */	"\xDF\xFE", "\xDF\xFD", "\xE0\x44", "\x8F\xD0",
  /* U+89E8 */	"\xE4\x49", "\xE4\x47", "\x00\x00", "\xE4\x48",
  /* U+89EC */	"\xE7\xD3", "\xE7\xD1", "\x00\x00", "\x00\x00",
  /* U+89F0 */	"\xE7\xD2", "\xEB\x7D", "\xEE\x7C", "\xEE\x7D",
  /* U+89F4 */	"\xC2\xD2", "\x00\x00", "\xF2\xF5", "\xF4\xCD",
  /* U+89F8 */	"\xC4\xB2", "\x00\x00", "\xF6\x4C", "\xF7\x55",
  /* U+89FC */	"\xC5\xA9", "\x00\x00", "\xF7\xFE", "\xF9\x4C",
  /* U+8A00 */	"\xA8\xA5", "\x00\x00", "\xAD\x71", "\xAD\x72",
  /* U+8A04 */	"\xD0\xB0", "\x00\x00", "\x00\x00", "\xD0\xB1",
  /* U+8A08 */	"\xAD\x70", "\x00\x00", "\xB0\x54", "\x00\x00",
  /* U+8A0C */	"\xB0\x52", "\x00\x00", "\xB0\x51", "\xB0\x58",
  /* U+8A10 */	"\xB0\x50", "\xB0\x59", "\xD3\xDD", "\xB0\x56",
  /* U+8A14 */	"\x00\x00", "\xB0\x53", "\xB0\x57", "\xB0\x55",
  /* U+8A18 */	"\xB0\x4F", "\x00\x00", "\x00\x00", "\xB3\x5F",
  /* U+8A1C */	"\x95\xB6", "\xB3\x59", "\xD7\xCC", "\xB3\x5E",
  /* U+8A20 */	"\x00\x00", "\x00\x00", "\xB3\x60", "\xB3\x5A",
  /* U+8A24 */	"\x00\x00", "\xB3\x5B", "\x00\x00", "\xD7\xCA",
  /* U+8A28 */	"\x00\x00", "\x99\xD6", "\xB3\x58", "\x95\xE5",
  /* U+8A2C */	"\xD7\xCB", "\xB3\x5D", "\x00\x00", "\x00\x00",
  /* U+8A30 */	"\xD7\xC9", "\xB3\x5C", "\x00\x00", "\x00\x00",
  /* U+8A34 */	"\xB6\x44", "\x00\x00", "\xB6\x46", "\x00\x00",
  /* U+8A38 */	"\x99\xD7", "\xDB\xD8", "\xB6\x45", "\xB5\xF9",
  /* U+8A3C */	"\xB5\xFD", "\x95\xB5", "\xB8\xE4", "\xE0\x49",
  /* U+8A40 */	"\xDB\xDA", "\xB5\xFE", "\x00\x00", "\x00\x00",
  /* U+8A44 */	"\xDB\xDD", "\xDB\xDE", "\xB6\x43", "\x00\x00",
  /* U+8A48 */	"\xDB\xE0", "\xA0\xCA", "\xDB\xE2", "\x00\x00",
  /* U+8A4C */	"\xDB\xE3", "\xDB\xD7", "\xDB\xD6", "\xDB\xE4",
  /* U+8A50 */	"\xB6\x42", "\xDB\xE1", "\xDB\xDF", "\x00\x00",
  /* U+8A54 */	"\xB6\x40", "\xB5\xFB", "\xB6\x47", "\xDB\xDB",
  /* U+8A58 */	"\xDB\xDC", "\xDB\xD9", "\x00\x00", "\xB6\x41",
  /* U+8A5C */	"\x00\x00", "\x00\x00", "\xB5\xFC", "\x00\x00",
  /* U+8A60 */	"\xB5\xFA", "\xE0\x48", "\xB8\xDF", "\xB8\xDA",
  /* U+8A64 */	"\x00\x00", "\x00\x00", "\xB8\xD5", "\x9F\xFD",
  /* U+8A68 */	"\xB8\xE5", "\xB8\xD6", "\x00\x00", "\xB8\xD2",
  /* U+8A6C */	"\xB8\xE1", "\xB8\xDE", "\xB8\xE0", "\x00\x00",
  /* U+8A70 */	"\xB8\xD7", "\xB8\xDC", "\xB8\xD3", "\xB8\xD4",
  /* U+8A74 */	"\xE0\x50", "\xE0\x4D", "\xE0\x45", "\xE0\x4A",
  /* U+8A78 */	"\x00\x00", "\xB8\xE2", "\xE0\x51", "\xB8\xE3",
  /* U+8A7C */	"\xB8\xD9", "\x00\x00", "\xA0\x58", "\xE0\x47",
  /* U+8A80 */	"\x00\x00", "\xE0\x4F", "\xE0\x4B", "\xE0\x4E",
  /* U+8A84 */	"\xE0\x4C", "\xB8\xDD", "\xE0\x46", "\xB8\xD8",
  /* U+8A88 */	"\x00\x00", "\x00\x00", "\x00\x00", "\xE4\x4C",
  /* U+8A8C */	"\xBB\x78", "\xBB\x7B", "\x00\x00", "\xE4\x4E",
  /* U+8A90 */	"\x8F\xD6", "\xBB\xA5", "\xE4\x4D", "\xBB\x7D",
  /* U+8A94 */	"\x99\xD8", "\xBD\xCF", "\xE4\x4F", "\x00\x00",
  /* U+8A98 */	"\xBB\xA4", "\xE4\x4B", "\xBB\xA6", "\x00\x00",
  /* U+8A9C */	"\x8F\xD3", "\x00\x00", "\xBB\x79", "\x00\x00",
  /* U+8AA0 */	"\xB8\xDB", "\xBB\x7C", "\x00\x00", "\xBB\x7A",
  /* U+8AA4 */	"\xBB\x7E", "\xBB\xA2", "\xBB\x77", "\xBB\xA7",
  /* U+8AA8 */	"\xBB\xA3", "\x8F\xE5", "\xBB\xA1", "\xE4\x4A",
  /* U+8AAC */	"\x00\x00", "\x00\x00", "\x00\x00", "\x8F\xE9",
  /* U+8AB0 */	"\xBD\xD6", "\x00\x00", "\xBD\xD2", "\x00\x00",
  /* U+8AB4 */	"\x99\xD9", "\x00\x00", "\xBD\xD9", "\x00\x00",
  /* U+8AB8 */	"\xE7\xD6", "\xBD\xDA", "\xE7\xE2", "\xE7\xDB",
  /* U+8ABC */	"\xBD\xCB", "\xE7\xE3", "\xE7\xDD", "\xBD\xD5",
  /* U+8AC0 */	"\xE7\xDE", "\x00\x00", "\xBD\xD4", "\xE7\xE1",
  /* U+8AC4 */	"\xBD\xCE", "\xE7\xDF", "\xE7\xD5", "\xBD\xCD",
  /* U+8AC8 */	"\xEB\xAA", "\xBD\xD3", "\x00\x00", "\xBD\xD0",
  /* U+8ACC */	"\x00\x00", "\xBD\xD8", "\x00\x00", "\xE7\xD4",
  /* U+8AD0 */	"\x00\x00", "\xE7\xD8", "\xBD\xCC", "\xE7\xD7",
  /* U+8AD4 */	"\xE7\xD9", "\xE7\xDA", "\xBD\xD7", "\xE7\xDC",
  /* U+8AD8 */	"\xE7\xE0", "\xE7\xE4", "\x92\x7C", "\xBD\xDB",
  /* U+8ADC */	"\xBF\xD2", "\xEB\xA5", "\xEB\xAB", "\xEB\xA8",
  /* U+8AE0 */	"\xEB\x7E", "\xEB\xAC", "\xEB\xA1", "\x00\x00",
  /* U+8AE4 */	"\xEB\xA7", "\x00\x00", "\xBF\xCD", "\xBF\xD3",
  /* U+8AE8 */	"\xEB\xAD", "\x00\x00", "\x9C\x45", "\xBF\xCF",
  /* U+8AEC */	"\x00\x00", "\xBF\xD9", "\xBF\xD4", "\xEB\xAF",
  /* U+8AF0 */	"\xEB\xA9", "\xBF\xD0", "\xEB\xA2", "\xBF\xDA",
  /* U+8AF4 */	"\xEB\xA3", "\xEB\xA4", "\xBF\xDB", "\xBF\xD8",
  /* U+8AF8 */	"\xBD\xD1", "\x00\x00", "\xBF\xCE", "\xEB\xB0",
  /* U+8AFC */	"\xBF\xDC", "\x00\x00", "\xBF\xD5", "\xEB\xAE",
  /* U+8B00 */	"\xBF\xD1", "\xBF\xD6", "\xBF\xD7", "\x00\x00",
  /* U+8B04 */	"\xC1\xC3", "\xEE\xA4", "\xEE\xAD", "\xEE\xAA",
  /* U+8B08 */	"\xEE\xAC", "\x00\x00", "\xC1\xC0", "\xEE\xA5",
  /* U+8B0C */	"\x8F\xDE", "\xEE\xAB", "\xC1\xBC", "\xEE\xA7",
  /* U+8B10 */	"\xC1\xC4", "\xEE\xA3", "\xEE\xA8", "\xEE\xAF",
  /* U+8B14 */	"\xEB\xA6", "\xEE\xA9", "\xEE\xA2", "\xC1\xBD",
  /* U+8B18 */	"\xEE\xA1", "\xC1\xBE", "\xEE\xB0", "\xC1\xBF",
  /* U+8B1C */	"\xEE\xAE", "\xC1\xC2", "\xEE\x7E", "\x8F\xDF",
  /* U+8B20 */	"\xC1\xC1", "\x00\x00", "\xEE\xA6", "\xF0\xDC",
  /* U+8B24 */	"\xF0\xEA", "\xF0\xE5", "\xF0\xE7", "\xF0\xDB",
  /* U+8B28 */	"\xC2\xD3", "\x00\x00", "\xF0\xDA", "\xC2\xD6",
  /* U+8B2C */	"\xC2\xD5", "\xA0\x4B", "\xF0\xE9", "\xF0\xE1",
  /* U+8B30 */	"\xF0\xDE", "\xF0\xE4", "\x00\x00", "\xF0\xDD",
  /* U+8B34 */	"\x00\x00", "\xF0\xDF", "\xF0\xE8", "\xF0\xE6",
  /* U+8B38 */	"\x00\x00", "\xC2\xD4", "\xF0\xED", "\xF0\xEB",
  /* U+8B3C */	"\xF0\xE2", "\xF0\xEC", "\xF0\xE3", "\x8F\xE2",
  /* U+8B40 */	"\xF2\xF9", "\xC3\xCF", "\xF3\x41", "\xA0\xCC",
  /* U+8B44 */	"\x00\x00", "\xF6\x4F", "\xC3\xD6", "\xF0\xE0",
  /* U+8B48 */	"\xF2\xF7", "\xC3\xD2", "\xF2\xF8", "\xF2\xFD",
  /* U+8B4C */	"\x8F\xE3", "\x8F\xE4", "\xC3\xD4", "\xC3\xD5",
  /* U+8B50 */	"\xF2\xF6", "\xF3\x40", "\xF3\x42", "\xF2\xFA",
  /* U+8B54 */	"\xF2\xFC", "\xF2\xFE", "\xF2\xFB", "\xF3\x43",
  /* U+8B58 */	"\xC3\xD1", "\xC3\xD7", "\xC3\xD3", "\x00\x00",
  /* U+8B5C */	"\xC3\xD0", "\xF4\xD0", "\x9B\xC4", "\xC4\xB7",
  /* U+8B60 */	"\xF4\xCE", "\x00\x00", "\x9B\xFC", "\xF4\xD2",
  /* U+8B64 */	"\x00\x00", "\xF4\xD3", "\xC4\xB5", "\xF4\xD4",
  /* U+8B68 */	"\xF4\xD1", "\x96\x4C", "\xF4\xCF", "\xC4\xB8",
  /* U+8B6C */	"\xC4\xB4", "\xF4\xD5", "\x00\x00", "\xC4\xB6",
  /* U+8B70 */	"\xC4\xB3", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+8B74 */	"\xC4\xFE", "\x00\x00", "\x00\x00", "\xC5\x40",
  /* U+8B78 */	"\xF6\x4E", "\xF6\x4D", "\xF6\x50", "\xF6\x51",
  /* U+8B7C */	"\x00\x00", "\xC5\x41", "\xF7\x56", "\xF7\x5B",
  /* U+8B80 */	"\xC5\xAA", "\x9A\xF6", "\xF7\x58", "\x8C\xAE",
  /* U+8B84 */	"\xF7\x57", "\xF7\x5A", "\xF7\x59", "\x00\x00",
  /* U+8B88 */	"\xF8\x43", "\x00\x00", "\xC5\xDC", "\xF8\x42",
  /* U+8B8C */	"\xF8\x40", "\x00\x00", "\xF8\x41", "\x00\x00",
  /* U+8B90 */	"\x8F\xE7", "\x00\x00", "\xC5\xFE", "\xC5\xFD",
  /* U+8B94 */	"\xF8\xC1", "\xF8\xC2", "\xC6\x40", "\x00\x00",
  /* U+8B98 */	"\xF9\x4D", "\xF9\x4E", "\xC6\x67", "\x8F\xE8",
  /* U+8B9C */	"\xC6\x6D", "\x00\x00", "\xF9\xA9", "\xF9\xC8",
  /* U+8BA0 */	"\x8B\xE7", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+8BA4 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+8BA8 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+8BAC */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+8BB0 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+8BB4 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+8BB8 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+8BBC */	"\x00\x00", "\x00\x00", "\x89\x7A", "\x00\x00",
  /* U+8BC0 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+8BC4 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+8BC8 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+8BCC */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+8BD0 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+8BD4 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+8BD8 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+8BDC */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+8BE0 */	"\x00\x00", "\x00\x00", "\x89\x7B", "\x00\x00",
  /* U+8BE4 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+8BE8 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+8BEC */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+8BF0 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+8BF4 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+8BF8 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+8BFC */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+8C00 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+8C04 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+8C08 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+8C0C */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+8C10 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+8C14 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+8C18 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+8C1C */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+8C20 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+8C24 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+8C28 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+8C2C */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+8C30 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+8C34 */	"\x00\x00", "\x00\x00", "\x00\x00", "\xA8\xA6",
  /* U+8C38 */	"\x00\x00", "\xD7\xCD", "\x00\x00", "\xD7\xCE",
  /* U+8C3C */	"\xE0\x52", "\xE4\x50", "\xE7\xE5", "\xC1\xC6",
  /* U+8C40 */	"\x00\x00", "\xC1\xC5", "\xF0\xEE", "\xF3\x44",
  /* U+8C44 */	"\x00\x00", "\xF8\x44", "\xA8\xA7", "\xD3\xDE",
  /* U+8C48 */	"\xB0\x5A", "\xB3\x61", "\xE0\x54", "\xE0\x53",
  /* U+8C4C */	"\xBD\xDC", "\xE7\xE6", "\xBD\xDD", "\xEE\xB1",
  /* U+8C50 */	"\xC2\xD7", "\x99\xDA", "\x00\x00", "\x00\x00",
  /* U+8C54 */	"\xC6\x76", "\xA8\xA8", "\xCD\xCB", "\xD3\xDF",
  /* U+8C58 */	"\x00\x00", "\x00\x00", "\xB3\x62", "\x00\x00",
  /* U+8C5C */	"\xD7\xCF", "\xD7\xD0", "\x00\x00", "\xDB\xE5",
  /* U+8C60 */	"\x00\x00", "\xB6\x48", "\xB8\xE6", "\x00\x00",
  /* U+8C64 */	"\xE0\x56", "\xE0\x55", "\xE0\x57", "\x00\x00",
  /* U+8C68 */	"\xE4\x51", "\xE4\x52", "\xBB\xA8", "\xBF\xDD",
  /* U+8C6C */	"\xBD\xDE", "\xBF\xDE", "\x00\x00", "\xEE\xB5",
  /* U+8C70 */	"\xEE\xB2", "\xEE\xB4", "\xEE\xB3", "\xC1\xC7",
  /* U+8C74 */	"\x00\x00", "\xF0\xEF", "\xF3\x46", "\xF3\x45",
  /* U+8C78 */	"\xCB\xA4", "\xB0\x5C", "\xB0\x5B", "\xD3\xE0",
  /* U+8C7C */	"\x00\x00", "\xD7\xD1", "\x00\x00", "\x00\x00",
  /* U+8C80 */	"\xDB\xE7", "\xDB\xE6", "\xB6\x49", "\x00\x00",
  /* U+8C84 */	"\xE0\x59", "\xE0\x5A", "\xE0\x58", "\x00\x00",
  /* U+8C88 */	"\x00\x00", "\xB8\xE8", "\xB8\xE7", "\x00\x00",
  /* U+8C8C */	"\xBB\xAA", "\xBB\xA9", "\x00\x00", "\xE7\xE7",
  /* U+8C90 */	"\xEB\xB3", "\xEB\xB1", "\xEB\xB2", "\xBF\xDF",
  /* U+8C94 */	"\xEE\xB7", "\xEE\xB6", "\x00\x00", "\xF0\xF2",
  /* U+8C98 */	"\xF0\xF1", "\xF0\xF0", "\xF3\x47", "\x8F\xED",
  /* U+8C9C */	"\xF9\xAA", "\xA8\xA9", "\xAD\x73", "\x95\xC0",
  /* U+8CA0 */	"\xAD\x74", "\xB0\x5D", "\xB0\x5E", "\xD3\xE2",
  /* U+8CA4 */	"\xD3\xE1", "\xD7\xD2", "\x00\x00", "\xB3\x68",
  /* U+8CA8 */	"\xB3\x66", "\xB3\x63", "\xB3\x67", "\xB3\x65",
  /* U+8CAC */	"\xB3\x64", "\xA0\xCB", "\x00\x00", "\xB6\x4A",
  /* U+8CB0 */	"\xDB\xEA", "\x00\x00", "\xB8\xED", "\xB6\x4C",
  /* U+8CB4 */	"\xB6\x51", "\xDB\xEC", "\xB6\x53", "\xB6\x52",
  /* U+8CB8 */	"\xB6\x55", "\xDB\xEB", "\xDB\xE8", "\xB6\x4F",
  /* U+8CBC */	"\xB6\x4B", "\xB6\x4D", "\xDB\xE9", "\xB6\x54",
  /* U+8CC0 */	"\xB6\x50", "\xB6\x4E", "\xB8\xEF", "\xB8\xEE",
  /* U+8CC4 */	"\xB8\xEC", "\xB8\xF0", "\x00\x00", "\xB8\xEA",
  /* U+8CC8 */	"\xB8\xEB", "\x00\x00", "\xB8\xE9", "\x00\x00",
  /* U+8CCC */	"\xE0\x5B", "\x9E\x48", "\x00\x00", "\xE4\x54",
  /* U+8CD0 */	"\x00\x00", "\xBB\xAC", "\xBB\xAD", "\xBB\xAB",
  /* U+8CD4 */	"\x99\xDB", "\xE4\x53", "\x8F\xF3", "\xE4\x55",
  /* U+8CD8 */	"\x00\x00", "\xE7\xEA", "\xE7\xEC", "\x8F\xF9",
  /* U+8CDC */	"\xBD\xE7", "\xE7\xED", "\xBD\xE0", "\xE7\xE9",
  /* U+8CE0 */	"\xBD\xDF", "\xBD\xE9", "\xBD\xE5", "\xBD\xE6",
  /* U+8CE4 */	"\xBD\xE2", "\xE7\xE8", "\xBD\xE1", "\xE7\xEE",
  /* U+8CE8 */	"\xE7\xEB", "\x95\xC1", "\xBD\xE8", "\xA0\x4E",
  /* U+8CEC */	"\xBD\xE3", "\xBD\xE4", "\xEB\xB5", "\x00\x00",
  /* U+8CF0 */	"\xEB\xB7", "\xEB\xB6", "\x99\xDC", "\xEB\xB8",
  /* U+8CF4 */	"\xBF\xE0", "\xEB\xB4", "\x00\x00", "\xA0\x64",
  /* U+8CF8 */	"\xC1\xCB", "\xEE\xB8", "\xC1\xC8", "\xC1\xCC",
  /* U+8CFC */	"\xC1\xCA", "\xC1\xC9", "\xF0\xF3", "\x00\x00",
  /* U+8D00 */	"\xF0\xF6", "\x00\x00", "\xF0\xF5", "\x8F\xF7",
  /* U+8D04 */	"\xF0\xF4", "\xC2\xD8", "\xF3\x48", "\xF3\x49",
  /* U+8D08 */	"\xC3\xD8", "\xF3\x4A", "\xC3\xD9", "\x89\xB0",
  /* U+8D0C */	"\xA0\x48", "\xC4\xBA", "\x00\x00", "\xC4\xB9",
  /* U+8D10 */	"\xF6\x52", "\x8F\xFB", "\x8F\xF6", "\xC5\x42",
  /* U+8D14 */	"\xF6\x53", "\xF7\x5C", "\xC5\xAB", "\xC5\xAC",
  /* U+8D18 */	"\x9D\xDC", "\xF8\x45", "\x00\x00", "\xC6\x42",
  /* U+8D1C */	"\x99\xDD", "\x8B\xE8", "\x00\x00", "\x00\x00",
  /* U+8D20 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+8D24 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+8D28 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+8D2C */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+8D30 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+8D34 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+8D38 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+8D3C */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+8D40 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+8D44 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+8D48 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+8D4C */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+8D50 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+8D54 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+8D58 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+8D5C */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+8D60 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+8D64 */	"\xA8\xAA", "\x00\x00", "\xB3\x6A", "\xB3\x69",
  /* U+8D68 */	"\xE0\x5C", "\xE0\x5D", "\x00\x00", "\xBB\xAE",
  /* U+8D6C */	"\xEB\xB9", "\xBD\xEA", "\xEB\xBA", "\xEE\xB9",
  /* U+8D70 */	"\xA8\xAB", "\x00\x00", "\xD0\xB2", "\xAD\x76",
  /* U+8D74 */	"\xAD\x75", "\x00\x00", "\xD3\xE3", "\xB0\x5F",
  /* U+8D78 */	"\xD3\xE4", "\xD7\xD5", "\x92\xC1", "\xD7\xD4",
  /* U+8D7C */	"\x00\x00", "\xD7\xD3", "\x00\x00", "\x00\x00",
  /* U+8D80 */	"\xDB\xEE", "\xB6\x58", "\x9F\xD6", "\x00\x00",
  /* U+8D84 */	"\xDB\xED", "\xB6\x57", "\x00\x00", "\x00\x00",
  /* U+8D88 */	"\x00\x00", "\xDB\xEF", "\xB6\x56", "\x00\x00",
  /* U+8D8C */	"\xE0\x5F", "\xE0\x62", "\xE0\x60", "\xE0\x61",
  /* U+8D90 */	"\xE0\x65", "\xE0\x5E", "\xE0\x66", "\xE0\x63",
  /* U+8D94 */	"\xE0\x64", "\xBB\xB0", "\xE4\x56", "\x00\x00",
  /* U+8D98 */	"\x00\x00", "\xBB\xAF", "\x00\x00", "\xE7\xF2",
  /* U+8D9C */	"\xE7\xF0", "\x00\x00", "\x00\x00", "\xBD\xEB",
  /* U+8DA0 */	"\xE7\xEF", "\xE7\xF1", "\x00\x00", "\xBD\xEC",
  /* U+8DA4 */	"\x00\x00", "\xEB\xBB", "\xA0\xD2", "\xEB\xBC",
  /* U+8DA8 */	"\xC1\xCD", "\x90\x40", "\xF3\x4C", "\xF3\x4E",
  /* U+8DAC */	"\xF3\x4B", "\xF3\x4D", "\xF4\xD6", "\xF6\x54",
  /* U+8DB0 */	"\x00\x00", "\x00\x00", "\xF9\x6F", "\xA8\xAC",
  /* U+8DB4 */	"\xAD\x77", "\xD3\xE5", "\xD3\xE7", "\xD3\xE6",
  /* U+8DB8 */	"\x00\x00", "\xD7\xD8", "\xB3\x6C", "\x00\x00",
  /* U+8DBC */	"\xD7\xD6", "\x00\x00", "\xB3\x6B", "\xD7\xD9",
  /* U+8DC0 */	"\x8A\xC4", "\xD7\xDA", "\xD7\xD7", "\x99\xE0",
  /* U+8DC4 */	"\x00\x00", "\xDB\xFB", "\xB6\x60", "\xDB\xF3",
  /* U+8DC8 */	"\xDB\xF9", "\x00\x00", "\x00\x00", "\xB6\x5B",
  /* U+8DCC */	"\xB6\x5E", "\xDB\xF2", "\xB6\x59", "\xDB\xF6",
  /* U+8DD0 */	"\xE0\x6C", "\xB6\x5D", "\x00\x00", "\xDB\xF1",
  /* U+8DD4 */	"\x9F\xF0", "\xDB\xF7", "\xDB\xF4", "\xDB\xFA",
  /* U+8DD8 */	"\xDB\xF0", "\xDB\xF8", "\xB6\x5C", "\xB6\x5F",
  /* U+8DDC */	"\xDB\xF5", "\xB6\x5A", "\x00\x00", "\xB8\xF2",
  /* U+8DE0 */	"\xE0\x68", "\xB8\xF1", "\xE0\x6F", "\xE0\x6E",
  /* U+8DE4 */	"\xB8\xF8", "\x00\x00", "\xB8\xF9", "\xE0\x70",
  /* U+8DE8 */	"\xB8\xF3", "\xE0\x6D", "\xB8\xF7", "\xE0\x72",
  /* U+8DEC */	"\xE0\x69", "\x00\x00", "\xE0\x6B", "\xB8\xF4",
  /* U+8DF0 */	"\xE0\x67", "\xE0\x6A", "\xE0\x71", "\xB8\xF5",
  /* U+8DF4 */	"\xE0\x73", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+8DF8 */	"\x00\x00", "\x00\x00", "\xB8\xF6", "\x00\x00",
  /* U+8DFC */	"\xBB\xB1", "\xE4\x5B", "\xE4\x61", "\xE4\x59",
  /* U+8E00 */	"\xE4\x62", "\x9F\xF3", "\xE4\x58", "\xE4\x5D",
  /* U+8E04 */	"\xE4\x63", "\xE4\x60", "\xE4\x5F", "\xE4\x5E",
  /* U+8E08 */	"\x00\x00", "\xE4\x57", "\xE4\x5C", "\x00\x00",
  /* U+8E0C */	"\x00\x00", "\xE4\x5A", "\x9D\xBF", "\xBD\xF1",
  /* U+8E10 */	"\xBD\xEE", "\xE7\xFB", "\xE8\x41", "\xE8\x43",
  /* U+8E14 */	"\xE8\x40", "\xE7\xF8", "\xE7\xFA", "\xE8\x45",
  /* U+8E18 */	"\xE8\x42", "\xE7\xFC", "\xE8\x46", "\xE7\xF9",
  /* U+8E1C */	"\xE8\x44", "\xBD\xEF", "\xBD\xF5", "\xBD\xF3",
  /* U+8E20 */	"\xE7\xF3", "\xBD\xF4", "\xBD\xF0", "\xE7\xF4",
  /* U+8E24 */	"\xE7\xF6", "\xE7\xF5", "\xE7\xFD", "\xE7\xFE",
  /* U+8E28 */	"\x9F\xF6", "\xBD\xF2", "\x95\xC8", "\xBD\xED",
  /* U+8E2C */	"\x00\x00", "\x9E\x5A", "\xE7\xF7", "\x00\x00",
  /* U+8E30 */	"\xEB\xC6", "\xBF\xE2", "\x00\x00", "\xEB\xBD",
  /* U+8E34 */	"\xBF\xE3", "\xBF\xE6", "\xEB\xC2", "\x00\x00",
  /* U+8E38 */	"\xEB\xBF", "\xBF\xE5", "\x99\xE3", "\x00\x00",
  /* U+8E3C */	"\xEB\xC3", "\xEB\xC4", "\xEB\xBE", "\xEB\xC7",
  /* U+8E40 */	"\xEB\xC0", "\xEB\xC5", "\xBF\xE4", "\x00\x00",
  /* U+8E44 */	"\xBF\xE1", "\xEB\xC1", "\x8A\x4A", "\xEE\xBF",
  /* U+8E48 */	"\xC1\xD0", "\xC1\xCE", "\xC1\xD1", "\xC1\xCF",
  /* U+8E4C */	"\xEE\xBE", "\xEE\xBB", "\xEE\xBA", "\x9F\xF1",
  /* U+8E50 */	"\xEE\xBD", "\x00\x00", "\x00\x00", "\xEE\xBC",
  /* U+8E54 */	"\xF1\x45", "\xC2\xDE", "\xF0\xFB", "\xF0\xFA",
  /* U+8E58 */	"\x00\x00", "\xC2\xD9", "\xF1\x41", "\xF1\x40",
  /* U+8E5C */	"\xF0\xF7", "\xF1\x43", "\xF0\xFC", "\xC2\xDD",
  /* U+8E60 */	"\xF0\xF9", "\xF1\x42", "\xF0\xF8", "\xC2\xDA",
  /* U+8E64 */	"\xC2\xDC", "\xF0\xFD", "\xC2\xDB", "\xF0\xFE",
  /* U+8E68 */	"\x8A\xA7", "\xF1\x44", "\xF3\x52", "\x00\x00",
  /* U+8E6C */	"\xC3\xDE", "\xF3\x4F", "\x00\x00", "\xF3\x53",
  /* U+8E70 */	"\x00\x00", "\x99\xE6", "\xC3\xDB", "\xF3\x51",
  /* U+8E74 */	"\xC3\xE0", "\x9F\xF7", "\xC3\xDD", "\x9F\xED",
  /* U+8E78 */	"\xF3\x50", "\x00\x00", "\xC3\xDF", "\xF3\x54",
  /* U+8E7C */	"\xC3\xDA", "\x00\x00", "\x8A\x5C", "\x00\x00",
  /* U+8E80 */	"\x9D\xAE", "\xC4\xBC", "\xC4\xBE", "\x00\x00",
  /* U+8E84 */	"\xF4\xD9", "\xC4\xBD", "\xF4\xD7", "\xC3\xDC",
  /* U+8E88 */	"\xF4\xD8", "\xC4\xBB", "\xC5\x43", "\xC5\x45",
  /* U+8E8C */	"\xF6\x56", "\xC5\x44", "\xF6\x55", "\x00\x00",
  /* U+8E90 */	"\xF7\x61", "\xC5\xAD", "\xF7\x60", "\xC5\xAE",
  /* U+8E94 */	"\xF7\x5E", "\xF7\x5D", "\xF7\x62", "\xF7\x63",
  /* U+8E98 */	"\xF8\x46", "\x00\x00", "\xF7\x5F", "\x00\x00",
  /* U+8E9C */	"\x00\x00", "\xF8\xC6", "\xF8\xC3", "\xF8\xC4",
  /* U+8EA0 */	"\xF8\xC5", "\xC6\x5C", "\x00\x00", "\xF9\x51",
  /* U+8EA4 */	"\xF9\x50", "\xF9\x4F", "\xF9\x70", "\x95\xC9",
  /* U+8EA8 */	"\xF9\xBE", "\xF9\xAB", "\xC6\x6E", "\xA8\xAD",
  /* U+8EAC */	"\xB0\x60", "\x90\x48", "\x00\x00", "\x00\x00",
  /* U+8EB0 */	"\x99\xE8", "\x00\x00", "\xB8\xFA", "\x00\x00",
  /* U+8EB4 */	"\x00\x00", "\x00\x00", "\x90\x49", "\x00\x00",
  /* U+8EB8 */	"\x00\x00", "\x8C\xBA", "\xBD\xF6", "\x00\x00",
  /* U+8EBC */	"\x90\xB1", "\xEB\xC8", "\x00\x00", "\x00\x00",
  /* U+8EC0 */	"\xC2\xDF", "\x00\x00", "\xF3\x55", "\x90\x4A",
  /* U+8EC4 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+8EC8 */	"\x00\x00", "\xF9\xAC", "\xA8\xAE", "\xAA\xEE",
  /* U+8ECC */	"\xAD\x79", "\xAD\x78", "\x99\xEA", "\xB0\x63",
  /* U+8ED0 */	"\x00\x00", "\xD3\xE8", "\xB0\x61", "\xD3\xE9",
  /* U+8ED4 */	"\xB0\x62", "\x00\x00", "\x00\x00", "\xD7\xDF",
  /* U+8ED8 */	"\xD7\xDB", "\x00\x00", "\x9B\xD1", "\xB3\x6D",
  /* U+8EDC */	"\xD7\xDE", "\xD7\xDD", "\xD7\xDC", "\xB3\x6E",
  /* U+8EE0 */	"\xD7\xE0", "\xD7\xE1", "\x99\xEB", "\x00\x00",
  /* U+8EE4 */	"\x99\xEC", "\xDC\x43", "\xDC\x41", "\xDC\x45",
  /* U+8EE8 */	"\xDC\x46", "\xDC\x4C", "\x00\x00", "\xDC\x48",
  /* U+8EEC */	"\xDC\x4A", "\x99\xED", "\xDC\x42", "\xDB\xFC",
  /* U+8EF0 */	"\x00\x00", "\xDC\x49", "\x99\xEE", "\x00\x00",
  /* U+8EF4 */	"\xDC\x4B", "\xDC\x44", "\xDC\x47", "\xDB\xFD",
  /* U+8EF8 */	"\xB6\x62", "\xDC\x40", "\xDB\xFE", "\xB6\x61",
  /* U+8EFC */	"\xB6\x63", "\x00\x00", "\xB8\xFD", "\xE0\x75",
  /* U+8F00 */	"\xE0\x77", "\xE0\x76", "\xE0\x7B", "\xB8\xFB",
  /* U+8F04 */	"\x00\x00", "\xE0\x78", "\xE0\x74", "\xE0\x79",
  /* U+8F08 */	"\xE0\x7A", "\xB8\xFC", "\xB8\xFE", "\xE0\x7C",
  /* U+8F0C */	"\x00\x00", "\xE4\x67", "\xE4\x66", "\x00\x00",
  /* U+8F10 */	"\xE4\x64", "\xE4\x65", "\xBB\xB3", "\xBB\xB5",
  /* U+8F14 */	"\xBB\xB2", "\xBB\xB4", "\xE8\x4D", "\xE8\x4E",
  /* U+8F18 */	"\xE8\x49", "\x90\x4C", "\xE8\x4A", "\xBD\xF8",
  /* U+8F1C */	"\xBD\xFD", "\xBD\xF7", "\xBD\xFE", "\xBD\xF9",
  /* U+8F20 */	"\xE8\x4B", "\x00\x00", "\x00\x00", "\xE8\x4C",
  /* U+8F24 */	"\xE8\x48", "\xBE\x40", "\xBD\xFB", "\x00\x00",
  /* U+8F28 */	"\x00\x00", "\xBD\xFA", "\xBD\xFC", "\x00\x00",
  /* U+8F2C */	"\xE8\x47", "\x90\x4D", "\xEB\xCA", "\xBF\xE8",
  /* U+8F30 */	"\x95\xCB", "\x00\x00", "\xEB\xCC", "\xBF\xEA",
  /* U+8F34 */	"\xEB\xCF", "\xEB\xCB", "\xEB\xC9", "\xEB\xCE",
  /* U+8F38 */	"\xBF\xE9", "\xEB\xCD", "\x00\x00", "\xBF\xE7",
  /* U+8F3C */	"\x00\x00", "\x00\x00", "\xC1\xD3", "\xC1\xD6",
  /* U+8F40 */	"\xEE\xC1", "\x97\xE2", "\xC1\xD4", "\xEE\xC0",
  /* U+8F44 */	"\xC1\xD2", "\xC1\xD5", "\xF1\x46", "\xF1\x47",
  /* U+8F48 */	"\xF1\x48", "\xC2\xE0", "\x95\xCC", "\xF1\x49",
  /* U+8F4C */	"\x00\x00", "\xC2\xE1", "\xC3\xE2", "\xF3\x58",
  /* U+8F50 */	"\xF3\x59", "\xF3\x57", "\xF3\x56", "\xF3\x5A",
  /* U+8F54 */	"\xC3\xE1", "\xF4\xDD", "\xF4\xDB", "\xF4\xDC",
  /* U+8F58 */	"\xF4\xDE", "\xF4\xDA", "\xF4\xDF", "\xF6\x58",
  /* U+8F5C */	"\x9F\x78", "\xF6\x59", "\xF6\x57", "\xC5\x46",
  /* U+8F60 */	"\xF7\x64", "\xC5\xAF", "\xF7\x65", "\xF8\x48",
  /* U+8F64 */	"\xF8\x47", "\x00\x00", "\x89\x7C", "\x89\x7D",
  /* U+8F68 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+8F6C */	"\x00\x00", "\x00\x00", "\x89\x7E", "\x00\x00",
  /* U+8F70 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+8F74 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+8F78 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+8F7C */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+8F80 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+8F84 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+8F88 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+8F8C */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+8F90 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x99\x5D",
  /* U+8F94 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+8F98 */	"\x00\x00", "\x00\x00", "\x00\x00", "\xA8\xAF",
  /* U+8F9C */	"\xB6\x64", "\x00\x00", "\x00\x00", "\xB9\x40",
  /* U+8FA0 */	"\x9B\x5A", "\x00\x00", "\x00\x00", "\xBB\xB6",
  /* U+8FA4 */	"\x00\x00", "\x90\x50", "\xBF\xEC", "\x8C\x4F",
  /* U+8FA8 */	"\xBF\xEB", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+8FAC */	"\x00\x00", "\xC3\xE3", "\xC4\x7C", "\xC5\x47",
  /* U+8FB0 */	"\xA8\xB0", "\xB0\x64", "\xB9\x41", "\x90\x54",
  /* U+8FB4 */	"\xF3\x5B", "\xC6\xD6", "\x9A\xA8", "\x99\xEF",
  /* U+8FB8 */	"\xFE\xEB", "\x9D\xA3", "\x9D\xA1", "\x99\x43",
  /* U+8FBC */	"\x99\x45", "\x00\x00", "\x9D\x7D", "\xCB\xA6",
  /* U+8FC0 */	"\x00\x00", "\x99\xF0", "\xA8\xB1", "\x00\x00",
  /* U+8FC4 */	"\xA8\xB4", "\xA8\xB3", "\xA8\xB2", "\x00\x00",
  /* U+8FC8 */	"\x00\x00", "\xCB\xA5", "\x99\xF1", "\xCD\xCD",
  /* U+8FCC */	"\x99\xF2", "\xCD\xCF", "\xAA\xEF", "\x8C\xBC",
  /* U+8FD0 */	"\x9D\x60", "\xAA\xF1", "\xCD\xCC", "\xCD\xCE",
  /* U+8FD4 */	"\xAA\xF0", "\xCD\xD1", "\xCD\xD0", "\xCD\xD2",
  /* U+8FD8 */	"\x00\x00", "\x00\x00", "\xA0\xA3", "\x00\x00",
  /* U+8FDC */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+8FE0 */	"\xD0\xB6", "\xD0\xB4", "\xAD\x7C", "\xD0\xB3",
  /* U+8FE4 */	"\xAD\xA3", "\xAD\x7E", "\xAD\x7B", "\x00\x00",
  /* U+8FE8 */	"\xAD\xA4", "\x00\x00", "\xAD\x7D", "\xAD\xA2",
  /* U+8FEC */	"\x00\x00", "\xAD\xA1", "\xD0\xB5", "\x00\x00",
  /* U+8FF0 */	"\xAD\x7A", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+8FF4 */	"\xB0\x6A", "\xD3\xEB", "\xD3\xF1", "\xB0\x67",
  /* U+8FF8 */	"\xB0\x6E", "\x90\x5B", "\xB0\x69", "\xD3\xEE",
  /* U+8FFC */	"\xD3\xF0", "\xB0\x6C", "\xD3\xEA", "\xD3\xED",
  /* U+9000 */	"\xB0\x68", "\xB0\x65", "\xD3\xEC", "\xB0\x6B",
  /* U+9004 */	"\xD3\xEF", "\xB0\x6D", "\xB0\x66", "\x00\x00",
  /* U+9008 */	"\x9E\xDB", "\x00\x00", "\x00\x00", "\xD7\xE3",
  /* U+900C */	"\xD7\xE6", "\xB3\x70", "\x00\x00", "\xB3\x7A",
  /* U+9010 */	"\xB3\x76", "\xD7\xE4", "\x9D\x79", "\x00\x00",
  /* U+9014 */	"\xB3\x7E", "\xB3\x77", "\xB3\x7C", "\xB3\x72",
  /* U+9018 */	"\x00\x00", "\xB3\x6F", "\xB3\x71", "\xB3\x7D",
  /* U+901C */	"\xD7\xE5", "\xB3\x75", "\xB3\x78", "\xB3\x74",
  /* U+9020 */	"\xB3\x79", "\xD7\xE7", "\xB3\x7B", "\xB3\x73",
  /* U+9024 */	"\xD7\xE2", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+9028 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+902C */	"\x00\x00", "\xDC\x4D", "\xB6\x65", "\xDC\x4F",
  /* U+9030 */	"\x00\x00", "\xB6\x67", "\xB6\x69", "\x99\xF3",
  /* U+9034 */	"\xDC\x4E", "\xB6\x66", "\xB6\x6A", "\x90\x62",
  /* U+9038 */	"\xB6\x68", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+903C */	"\xB9\x47", "\xE0\xA3", "\xB9\x4F", "\xE0\x7E",
  /* U+9040 */	"\x00\x00", "\xB9\x50", "\xB9\x45", "\x00\x00",
  /* U+9044 */	"\xE0\xA1", "\x00\x00", "\x00\x00", "\xB9\x4A",
  /* U+9048 */	"\x00\x00", "\xE0\xA2", "\xB9\x43", "\xB9\x42",
  /* U+904C */	"\x9F\x55", "\xB9\x4D", "\xB9\x4C", "\xB9\x4B",
  /* U+9050 */	"\xB9\x49", "\xB9\x4E", "\xE0\x7D", "\xB9\x44",
  /* U+9054 */	"\xB9\x46", "\xB9\x48", "\x9B\xF9", "\x00\x00",
  /* U+9058 */	"\xBB\xB8", "\xBB\xBB", "\x00\x00", "\xBB\xBF",
  /* U+905C */	"\xBB\xB9", "\xBB\xBE", "\xBB\xBC", "\x00\x00",
  /* U+9060 */	"\xBB\xB7", "\x90\x65", "\xBB\xBD", "\xBB\xBA",
  /* U+9064 */	"\x96\xE0", "\x00\x00", "\x00\x00", "\xE8\x52",
  /* U+9068 */	"\xBE\x43", "\xBE\x41", "\x00\x00", "\xE8\x53",
  /* U+906C */	"\x98\xBE", "\xBE\x44", "\xBE\x42", "\xE8\x51",
  /* U+9070 */	"\xE8\x50", "\x00\x00", "\xBF\xF0", "\xE8\x4F",
  /* U+9074 */	"\xBF\xEE", "\xBF\xED", "\xEB\xD0", "\xBE\x45",
  /* U+9078 */	"\xBF\xEF", "\xEB\xD1", "\xBF\xF2", "\xEB\xD2",
  /* U+907C */	"\xBF\xF1", "\xC1\xD8", "\xEE\xC3", "\xC1\xD7",
  /* U+9080 */	"\xC1\xDC", "\xC1\xDA", "\xC1\xDB", "\xC2\xE3",
  /* U+9084 */	"\xC1\xD9", "\xEE\xC2", "\xEB\xD3", "\xC2\xE2",
  /* U+9088 */	"\xC2\xE4", "\x00\x00", "\xC3\xE4", "\xC3\xE5",
  /* U+908C */	"\x00\x00", "\xF4\xE0", "\x00\x00", "\xC5\xDE",
  /* U+9090 */	"\xC5\xDD", "\xA8\xB6", "\x00\x00", "\x00\x00",
  /* U+9094 */	"\xCA\x55", "\xB0\x6F", "\x00\x00", "\xCA\x52",
  /* U+9098 */	"\xCA\x53", "\xCA\x51", "\x00\x00", "\xCA\x54",
  /* U+909C */	"\x00\x00", "\x00\x00", "\xCB\xAA", "\xCB\xA7",
  /* U+90A0 */	"\xCB\xAC", "\xCB\xA8", "\xA8\xB7", "\xA8\xBA",
  /* U+90A4 */	"\x00\x00", "\xCB\xA9", "\xA8\xB9", "\xCB\xAB",
  /* U+90A8 */	"\x90\x68", "\x00\x00", "\xA8\xB8", "\x00\x00",
  /* U+90AC */	"\x00\x00", "\x00\x00", "\x90\x6C", "\xCD\xD5",
  /* U+90B0 */	"\xCD\xD7", "\xAA\xF4", "\xCD\xD3", "\xCD\xD6",
  /* U+90B4 */	"\xCD\xD4", "\xAA\xF2", "\xAA\xF5", "\x00\x00",
  /* U+90B8 */	"\xAA\xF3", "\x00\x00", "\x00\x00", "\x95\xD8",
  /* U+90BC */	"\x00\x00", "\xD0\xB8", "\xD0\xBC", "\xD0\xB9",
  /* U+90C0 */	"\x00\x00", "\xAD\xA7", "\x00\x00", "\xAD\xA8",
  /* U+90C4 */	"\x90\x6A", "\xD0\xBB", "\x00\x00", "\xD0\xBD",
  /* U+90C8 */	"\xD0\xBF", "\x00\x00", "\xAD\xA5", "\xD0\xBE",
  /* U+90CC */	"\x00\x00", "\x00\x00", "\xAD\xA6", "\x00\x00",
  /* U+90D0 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+90D4 */	"\xD7\xEE", "\xD0\xBA", "\xD3\xF2", "\xD3\xFB",
  /* U+90D8 */	"\xD3\xF9", "\xD3\xF4", "\xD3\xF5", "\xD3\xFA",
  /* U+90DC */	"\xD3\xFC", "\xB0\x71", "\x00\x00", "\xD3\xF7",
  /* U+90E0 */	"\xD3\xF3", "\xB0\x70", "\xB0\x72", "\xD3\xF6",
  /* U+90E4 */	"\xD3\xFD", "\xD3\xF8", "\x00\x00", "\x00\x00",
  /* U+90E8 */	"\xB3\xA1", "\xD7\xF1", "\xD7\xE9", "\xD7\xEF",
  /* U+90EC */	"\xD7\xF0", "\xB3\xA2", "\x00\x00", "\xD7\xE8",
  /* U+90F0 */	"\xD7\xEA", "\xD0\xB7", "\xD7\xEC", "\xD7\xED",
  /* U+90F4 */	"\xD7\xEB", "\xB6\x6C", "\x00\x00", "\x00\x00",
  /* U+90F8 */	"\x00\x00", "\xDC\x56", "\xEB\xD4", "\xDC\x57",
  /* U+90FC */	"\xDC\x54", "\xB3\xA3", "\xB6\x6E", "\xDC\x53",
  /* U+9100 */	"\xDC\x59", "\xDC\x58", "\xB6\x6B", "\xDC\x5C",
  /* U+9104 */	"\xDC\x52", "\xDC\x5B", "\xDC\x50", "\xDC\x5A",
  /* U+9108 */	"\xDC\x55", "\xB6\x6D", "\x00\x00", "\xE0\xAA",
  /* U+910C */	"\x00\x00", "\xE0\xA5", "\xE0\xAB", "\xE0\xA6",
  /* U+9110 */	"\xE0\xA4", "\xE0\xA7", "\xB9\x51", "\x00\x00",
  /* U+9114 */	"\xE0\xA9", "\x00\x00", "\xE0\xA8", "\xB9\x52",
  /* U+9118 */	"\xBB\xC1", "\xBB\xC0", "\xE4\x6E", "\xE4\x71",
  /* U+911C */	"\xE4\x69", "\xE4\x6D", "\xBB\xC2", "\xE4\x6C",
  /* U+9120 */	"\xE4\x6A", "\xE4\x70", "\xE4\x6B", "\xE4\x68",
  /* U+9124 */	"\xE4\x6F", "\x00\x00", "\xE8\x59", "\xBE\x48",
  /* U+9128 */	"\xF1\x4A", "\xE8\x56", "\xE8\x57", "\xE8\x55",
  /* U+912C */	"\xDC\x51", "\xBE\x47", "\xE8\x5A", "\xE8\x54",
  /* U+9130 */	"\xBE\x46", "\xBE\x49", "\xE8\x58", "\xEB\xD5",
  /* U+9134 */	"\xBF\xF3", "\xEB\xD6", "\xEB\xD7", "\x00\x00",
  /* U+9138 */	"\xEE\xC4", "\xC1\xDD", "\xF1\x4B", "\xF1\x4C",
  /* U+913C */	"\x00\x00", "\x00\x00", "\xF1\x4D", "\xF3\x5D",
  /* U+9140 */	"\xF3\x5C", "\xF4\xE2", "\x00\x00", "\xF4\xE1",
  /* U+9144 */	"\xF6\x5B", "\xF6\x5C", "\xF6\x5A", "\xF7\x66",
  /* U+9148 */	"\xC5\xB0", "\xA8\xBB", "\xAD\xAA", "\xAD\xA9",
  /* U+914C */	"\xB0\x75", "\xB0\x74", "\xD4\x40", "\xD4\x41",
  /* U+9150 */	"\xD3\xFE", "\x9F\xB2", "\xB0\x73", "\xD7\xF5",
  /* U+9154 */	"\x00\x00", "\xD7\xF6", "\xD7\xF2", "\xB3\xA4",
  /* U+9158 */	"\xD7\xF3", "\x9F\xAE", "\xD7\xF4", "\x00\x00",
  /* U+915C */	"\x9F\xB0", "\x00\x00", "\x89\xAD", "\xDC\x5F",
  /* U+9160 */	"\xDC\x61", "\xDC\x5D", "\xDC\x60", "\xB6\x6F",
  /* U+9164 */	"\xDC\x5E", "\xB6\x70", "\x00\x00", "\x90\x6E",
  /* U+9168 */	"\xDD\x73", "\xB9\x55", "\xB9\x54", "\x00\x00",
  /* U+916C */	"\xB9\x53", "\x00\x00", "\xE0\xAC", "\xE0\xAD",
  /* U+9170 */	"\x9E\x71", "\x00\x00", "\xE4\x73", "\xE4\x75",
  /* U+9174 */	"\xBB\xC6", "\xBB\xC3", "\x9E\x4A", "\xBB\xC5",
  /* U+9178 */	"\xBB\xC4", "\xE4\x74", "\xE4\x72", "\x00\x00",
  /* U+917C */	"\x9F\xDC", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+9180 */	"\xE8\x61", "\xE8\x5E", "\xE8\x5F", "\xBE\x4D",
  /* U+9184 */	"\xE8\x60", "\xE8\x5B", "\xE8\x5C", "\xBE\x4A",
  /* U+9188 */	"\x00\x00", "\xBE\x4B", "\xE8\x5D", "\xBE\x4C",
  /* U+918C */	"\x89\xAB", "\xEB\xDB", "\x9F\xB8", "\xEB\xDC",
  /* U+9190 */	"\xEB\xD9", "\xEB\xDA", "\xBF\xF4", "\xEB\xD8",
  /* U+9194 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+9198 */	"\x00\x00", "\xEE\xC8", "\xEE\xC5", "\xEE\xC7",
  /* U+919C */	"\xC1\xE0", "\xEE\xCB", "\xC1\xDF", "\xEE\xC9",
  /* U+91A0 */	"\xEE\xCC", "\xEE\xCA", "\xEE\xC6", "\xC1\xDE",
  /* U+91A4 */	"\x00\x00", "\xF1\x4F", "\x00\x00", "\xF1\x50",
  /* U+91A8 */	"\xF1\x4E", "\x90\x70", "\xF1\x52", "\xC2\xE5",
  /* U+91AC */	"\xC2\xE6", "\xF3\x5F", "\xC3\xE7", "\xF1\x51",
  /* U+91B0 */	"\xF3\x5E", "\xC3\xE6", "\xF4\xE5", "\xF4\xE6",
  /* U+91B4 */	"\xC4\xBF", "\xF4\xE4", "\x8B\x63", "\xF4\xE3",
  /* U+91B8 */	"\x00\x00", "\xF6\x5D", "\xC5\x48", "\x95\xDC",
  /* U+91BC */	"\xF8\x49", "\xF8\xC8", "\xF8\xC7", "\x00\x00",
  /* U+91C0 */	"\xC6\x43", "\xC6\x5D", "\xF8\xC9", "\xF9\x71",
  /* U+91C4 */	"\x90\x71", "\xC6\x6F", "\xA8\xBC", "\xAA\xF6",
  /* U+91C8 */	"\x00\x00", "\xB9\x56", "\x00\x00", "\xC4\xC0",
  /* U+91CC */	"\xA8\xBD", "\xAD\xAB", "\xB3\xA5", "\xB6\x71",
  /* U+91D0 */	"\xC2\xE7", "\xAA\xF7", "\x00\x00", "\xD0\xC1",
  /* U+91D4 */	"\xD0\xC0", "\xD4\x42", "\xFC\x5E", "\xB0\x78",
  /* U+91D8 */	"\xB0\x76", "\xB0\x7A", "\xD4\x44", "\x00\x00",
  /* U+91DC */	"\xB0\x79", "\xB0\x77", "\x00\x00", "\x89\x49",
  /* U+91E0 */	"\x00\x00", "\x00\x00", "\xD4\x43", "\xB3\xA8",
  /* U+91E4 */	"\xD7\xFC", "\x96\x5B", "\xB3\xA7", "\xB3\xA9",
  /* U+91E8 */	"\xD8\x42", "\xB3\xAB", "\xD7\xFE", "\xD8\x40",
  /* U+91EC */	"\xD7\xF7", "\xB3\xAA", "\xD8\x43", "\x00\x00",
  /* U+91F0 */	"\x00\x00", "\xD7\xF9", "\x00\x00", "\xD7\xFA",
  /* U+91F4 */	"\xD7\xF8", "\xB3\xA6", "\x8C\x50", "\xD8\x41",
  /* U+91F8 */	"\xD7\xFB", "\xD7\xFD", "\x94\xA6", "\x00\x00",
  /* U+91FC */	"\x00\x00", "\xDC\x6D", "\x8F\xD5", "\xDC\x6C",
  /* U+9200 */	"\xDC\x6A", "\xDC\x62", "\xDC\x71", "\xDC\x65",
  /* U+9204 */	"\xDC\x6F", "\xDC\x76", "\xDC\x6E", "\xB6\x79",
  /* U+9208 */	"\x9E\x73", "\xB6\x75", "\xDC\x63", "\x00\x00",
  /* U+920C */	"\xDC\x69", "\xB6\x77", "\x90\x75", "\xDC\x68",
  /* U+9210 */	"\xB6\x78", "\xB6\x7A", "\xDC\x6B", "\x99\xF7",
  /* U+9214 */	"\xB6\x72", "\xB6\x73", "\xDC\x77", "\xDC\x75",
  /* U+9218 */	"\x00\x00", "\xDC\x74", "\xDC\x66", "\x00\x00",
  /* U+921C */	"\xDC\x72", "\x00\x00", "\xB6\x76", "\x00\x00",
  /* U+9220 */	"\x00\x00", "\x8C\xBF", "\x00\x00", "\xB6\x74",
  /* U+9224 */	"\xDC\x73", "\xDC\x64", "\xDC\x67", "\xDC\x70",
  /* U+9228 */	"\x99\xF9", "\x00\x00", "\x96\x63", "\x95\xB9",
  /* U+922C */	"\x00\x00", "\xE4\xBA", "\xE0\xB7", "\x00\x00",
  /* U+9230 */	"\xE0\xB0", "\xE0\xC3", "\xE0\xCC", "\xE0\xB3",
  /* U+9234 */	"\xB9\x61", "\x94\xD4", "\xE0\xC0", "\xB9\x57",
  /* U+9238 */	"\xB9\x59", "\xB9\x65", "\xE0\xB1", "\x00\x00",
  /* U+923C */	"\xFC\xFA", "\xB9\x5A", "\xB9\x5C", "\xB9\x66",
  /* U+9240 */	"\xB9\x5B", "\x90\x77", "\x00\x00", "\x00\x00",
  /* U+9244 */	"\x90\xAB", "\xB9\x64", "\xE0\xB9", "\x00\x00",
  /* U+9248 */	"\xE0\xAE", "\xB9\x62", "\xE0\xB8", "\xB9\x5E",
  /* U+924C */	"\xE0\xCA", "\xB9\x63", "\xE0\xC8", "\xE0\xBC",
  /* U+9250 */	"\xE0\xC6", "\xB9\x60", "\xE0\xAF", "\xE0\xC9",
  /* U+9254 */	"\xE0\xC4", "\x9D\x4D", "\xE0\xCB", "\xB9\x58",
  /* U+9258 */	"\x99\xFA", "\x00\x00", "\xB9\x67", "\xB9\x5D",
  /* U+925C */	"\x00\x00", "\x92\xE3", "\xE0\xB5", "\x97\xBB",
  /* U+9260 */	"\xE0\xBD", "\xE0\xC1", "\x90\x78", "\xE0\xC5",
  /* U+9264 */	"\xB9\x5F", "\xE0\xB4", "\xE0\xB2", "\xE0\xBE",
  /* U+9268 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x99\xFB",
  /* U+926C */	"\xE0\xBB", "\xE0\xBA", "\x97\xE0", "\xE0\xBF",
  /* U+9270 */	"\xE0\xC2", "\x00\x00", "\xE0\xC7", "\x00\x00",
  /* U+9274 */	"\x00\x00", "\x00\x00", "\xE4\x78", "\x96\xDC",
  /* U+9278 */	"\xBB\xC7", "\xE4\xA4", "\xE4\x7A", "\xBB\xCC",
  /* U+927C */	"\xBB\xD0", "\xE4\xAD", "\xE4\xB5", "\xE4\xA6",
  /* U+9280 */	"\xBB\xC8", "\x9C\xA8", "\xE4\xAA", "\xE0\xB6",
  /* U+9284 */	"\x97\x72", "\xBB\xC9", "\xE4\xB1", "\xE4\xB6",
  /* U+9288 */	"\xE4\xAE", "\x94\x40", "\xE4\xB0", "\xE4\xB9",
  /* U+928C */	"\xE4\xB2", "\xE4\x7E", "\xE4\xA9", "\x92\xF2",
  /* U+9290 */	"\x00\x00", "\xBB\xD1", "\x00\x00", "\xBB\xCD",
  /* U+9294 */	"\xE4\x7C", "\xE4\xAB", "\xBB\xCB", "\xE4\xA5",
  /* U+9298 */	"\xBB\xCA", "\xE4\xB3", "\xE4\xA2", "\xE4\x79",
  /* U+929C */	"\xBB\xCE", "\xE4\xB8", "\x00\x00", "\x00\x00",
  /* U+92A0 */	"\xE4\x7B", "\xE4\xAF", "\xE4\xAC", "\xE4\xA7",
  /* U+92A4 */	"\xE4\x77", "\xE4\x76", "\xE4\xA1", "\xE4\xB4",
  /* U+92A8 */	"\xBB\xCF", "\xE4\xB7", "\xE4\x7D", "\xE4\xA3",
  /* U+92AC */	"\xBE\x52", "\x00\x00", "\x99\xFD", "\x00\x00",
  /* U+92B0 */	"\x00\x00", "\x99\xFC", "\xBE\x5A", "\xBE\x55",
  /* U+92B4 */	"\xE8\xA4", "\xE8\xA1", "\xE8\x67", "\xBE\x50",
  /* U+92B8 */	"\x00\x00", "\xF9\xD7", "\x96\x4A", "\xBE\x4F",
  /* U+92BC */	"\xBE\x56", "\x00\x00", "\x96\xD8", "\x99\xFE",
  /* U+92C0 */	"\xE8\x65", "\xBE\x54", "\xE8\x71", "\xE8\x63",
  /* U+92C4 */	"\xE8\x64", "\xBE\x4E", "\xE8\xA3", "\xBE\x58",
  /* U+92C8 */	"\xE8\x74", "\xE8\x79", "\xE8\x73", "\xEB\xEE",
  /* U+92CC */	"\xE8\x6F", "\xE8\x77", "\xE8\x75", "\xE8\x68",
  /* U+92D0 */	"\xE8\x62", "\xE8\x7D", "\xBE\x57", "\xE8\x7E",
  /* U+92D4 */	"\x90\x4B", "\xE8\x78", "\x00\x00", "\xE8\x6D",
  /* U+92D8 */	"\xE8\x6B", "\xE8\x66", "\x00\x00", "\xFA\x41",
  /* U+92DC */	"\x00\x00", "\xE8\x6E", "\xE8\x7B", "\xE8\x6A",
  /* U+92E0 */	"\xE8\x7A", "\xE8\xA2", "\x00\x00", "\x9A\x40",
  /* U+92E4 */	"\xBE\x53", "\x97\x5B", "\xE8\x76", "\xE8\x7C",
  /* U+92E8 */	"\xE8\x72", "\xE8\x6C", "\xBE\x51", "\x9A\x41",
  /* U+92EC */	"\x91\xDD", "\x00\x00", "\xE4\xA8", "\xE8\x70",
  /* U+92F0 */	"\xBE\x59", "\xE8\x69", "\x93\xFC", "\x9A\x42",
  /* U+92F4 */	"\x9A\x43", "\x00\x00", "\x96\x59", "\xEB\xF4",
  /* U+92F8 */	"\xBF\xF7", "\xEB\xF3", "\xEB\xF0", "\xEC\x44",
  /* U+92FC */	"\xBF\xFB", "\x9A\x44", "\xEC\x41", "\xEB\xF8",
  /* U+9300 */	"\xEC\x43", "\xEB\xE9", "\xEB\xF6", "\x90\x51",
  /* U+9304 */	"\xBF\xFD", "\x00\x00", "\xEB\xE1", "\x94\xBF",
  /* U+9308 */	"\xEB\xDF", "\xEC\x42", "\x00\x00", "\xEC\x40",
  /* U+930C */	"\xEB\xFE", "\xEB\xED", "\xEB\xEC", "\xEB\xE2",
  /* U+9310 */	"\xC0\x40", "\x00\x00", "\xEB\xE8", "\xEB\xF2",
  /* U+9314 */	"\xEB\xFD", "\xC0\x43", "\xEC\x45", "\x00\x00",
  /* U+9318 */	"\xC1\xE8", "\xC0\x45", "\xBF\xFE", "\xEB\xE6",
  /* U+931C */	"\x00\x00", "\xEB\xEF", "\xEB\xDE", "\xEB\xE0",
  /* U+9320 */	"\xBF\xF5", "\xC0\x42", "\xBF\xFA", "\xEB\xE7",
  /* U+9324 */	"\xEB\xF7", "\xEB\xF1", "\xC0\x41", "\xEB\xDD",
  /* U+9328 */	"\xC1\xE3", "\xEB\xF9", "\xEB\xFC", "\xBF\xFC",
  /* U+932C */	"\x90\xA2", "\xEB\xEB", "\xC0\x44", "\xBF\xF9",
  /* U+9330 */	"\x9C\xAB", "\x97\x76", "\x00\x00", "\xBF\xF8",
  /* U+9334 */	"\xEB\xF5", "\xEB\xFB", "\xBF\xF6", "\x00\x00",
  /* U+9338 */	"\xEB\xE4", "\xEB\xFA", "\x00\x00", "\x00\x00",
  /* U+933C */	"\xEB\xE5", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+9340 */	"\xFC\x55", "\xFE\x45", "\x94\xA8", "\x9A\x45",
  /* U+9344 */	"\xFA\x4B", "\x9D\xE1", "\xEB\xEA", "\xEE\xD2",
  /* U+9348 */	"\x96\xD9", "\xEE\xD7", "\xC1\xE5", "\xC1\xE7",
  /* U+934C */	"\xEE\xDD", "\xC1\xE1", "\xEE\xEC", "\xEE\xE3",
  /* U+9350 */	"\xEE\xD8", "\xEE\xD9", "\xEE\xE2", "\x00\x00",
  /* U+9354 */	"\xC1\xEE", "\xEE\xE1", "\xEE\xD1", "\xEE\xE0",
  /* U+9358 */	"\xEE\xD4", "\xEE\xED", "\xC1\xED", "\xC1\xEB",
  /* U+935C */	"\xEE\xD5", "\x00\x00", "\xEE\xE8", "\x97\x74",
  /* U+9360 */	"\xEE\xDA", "\xEE\xE7", "\xFD\xF5", "\xEE\xE9",
  /* U+9364 */	"\xEE\xD0", "\xC1\xE6", "\x92\xE5", "\xEE\xEA",
  /* U+9368 */	"\x96\x45", "\x91\xDA", "\xEE\xDE", "\x90\xA3",
  /* U+936C */	"\xC1\xEA", "\xEE\xDB", "\xA0\x5F", "\x00\x00",
  /* U+9370 */	"\xC1\xEC", "\xEE\xE4", "\x00\x00", "\x90\xAF",
  /* U+9374 */	"\x97\xBF", "\xC1\xE4", "\xEE\xD6", "\xEE\xE5",
  /* U+9378 */	"\x91\x4C", "\xEE\xDF", "\xEB\xE3", "\xEE\xE6",
  /* U+937C */	"\xEE\xD3", "\x96\x7A", "\xC1\xE9", "\x00\x00",
  /* U+9380 */	"\xEE\xEB", "\x91\xDE", "\xC1\xE2", "\xEE\xCE",
  /* U+9384 */	"\x9A\x46", "\xFE\xB0", "\x97\x79", "\x94\x6C",
  /* U+9388 */	"\xF1\x60", "\xF1\x59", "\xC2\xE9", "\x00\x00",
  /* U+938C */	"\xF1\x54", "\xF1\x63", "\xF1\x5B", "\xEE\xDC",
  /* U+9390 */	"\x98\x58", "\xF1\x65", "\xF1\x55", "\x00\x00",
  /* U+9394 */	"\xC2\xE8", "\xF1\x5F", "\xC2\xEA", "\xC2\xF2",
  /* U+9398 */	"\xC2\xF0", "\xF1\x61", "\xC2\xF1", "\xF1\x57",
  /* U+939C */	"\x92\x66", "\xF1\x58", "\xF1\x5D", "\xF1\x62",
  /* U+93A0 */	"\x93\xFB", "\xEE\xCD", "\xC2\xEB", "\xF1\x6A",
  /* U+93A4 */	"\xF1\x67", "\xF1\x6B", "\xF1\x5E", "\xF1\x5A",
  /* U+93A8 */	"\xF1\x68", "\xF3\x6A", "\xF1\x5C", "\x00\x00",
  /* U+93AC */	"\xC2\xEE", "\x9A\x47", "\xC2\xED", "\xEE\xCF",
  /* U+93B0 */	"\xC2\xEF", "\xF1\x64", "\xF1\x66", "\xC2\xEC",
  /* U+93B4 */	"\xF1\x69", "\xF1\x53", "\x00\x00", "\xF1\x56",
  /* U+93B8 */	"\x97\x49", "\x00\x00", "\x00\x00", "\x97\x48",
  /* U+93BC */	"\x00\x00", "\x93\x4A", "\x00\x00", "\x9C\xE2",
  /* U+93C0 */	"\xF3\x73", "\x00\x00", "\xF3\x63", "\xC3\xEB",
  /* U+93C4 */	"\xF3\x71", "\x00\x00", "\x92\x64", "\xF3\x61",
  /* U+93C8 */	"\xC3\xEC", "\x00\x00", "\xF3\x6C", "\x91\xDF",
  /* U+93CC */	"\xF3\x68", "\xC3\xF1", "\xF3\x72", "\xF3\x62",
  /* U+93D0 */	"\xF3\x65", "\xC3\xE9", "\xF3\x74", "\xFB\x79",
  /* U+93D4 */	"\xF3\x6D", "\xF3\x70", "\xC3\xEF", "\xC3\xF4",
  /* U+93D8 */	"\xC3\xF2", "\xF3\x69", "\xF3\x64", "\x96\xD7",
  /* U+93DC */	"\xC3\xED", "\xC3\xEE", "\xF3\x60", "\xC3\xEA",
  /* U+93E0 */	"\x93\x43", "\xC3\xE8", "\xC3\xF0", "\xF3\x6F",
  /* U+93E4 */	"\xC3\xF3", "\x00\x00", "\xF3\x6B", "\xF3\x75",
  /* U+93E8 */	"\xC3\xF5", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+93EC */	"\xF3\x67", "\x00\x00", "\xF3\x6E", "\x00\x00",
  /* U+93F0 */	"\xFD\xCB", "\xFE\x7A", "\x00\x00", "\x91\xDB",
  /* U+93F4 */	"\x8C\x6A", "\xF4\xF3", "\xF5\x42", "\xF4\xF5",
  /* U+93F8 */	"\xF4\xFC", "\xF3\x66", "\xF4\xFA", "\xF4\xE9",
  /* U+93FC */	"\xF5\x40", "\xC4\xC3", "\xF4\xED", "\xF4\xFE",
  /* U+9400 */	"\xF4\xF4", "\x97\xAF", "\x00\x00", "\xC4\xC2",
  /* U+9404 */	"\x95\xDD", "\x00\x00", "\xF5\x44", "\xF4\xF6",
  /* U+9408 */	"\x93\x48", "\xF4\xFB", "\xF4\xFD", "\xF4\xE7",
  /* U+940C */	"\xF5\x41", "\xF4\xF2", "\xF4\xF7", "\xF4\xEB",
  /* U+9410 */	"\xF4\xEF", "\xF5\x43", "\xF4\xF9", "\xF4\xE8",
  /* U+9414 */	"\xF4\xEC", "\xF4\xEE", "\xF4\xF8", "\x9A\x4B",
  /* U+9418 */	"\xC4\xC1", "\xF4\xF1", "\x00\x00", "\xFC\x45",
  /* U+941C */	"\x00\x00", "\x9A\x4D", "\x00\x00", "\x00\x00",
  /* U+9420 */	"\xF4\xEA", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+9424 */	"\x91\xBC", "\x90\xE2", "\x90\xB4", "\x95\xE1",
  /* U+9428 */	"\xF4\xF0", "\xF6\x61", "\xF6\x66", "\xC5\x4F",
  /* U+942C */	"\xF6\x68", "\x9A\x4E", "\xC5\x49", "\x00\x00",
  /* U+9430 */	"\xF6\x64", "\xF6\x6A", "\xC5\x4E", "\xC5\x4A",
  /* U+9434 */	"\x00\x00", "\xC5\x4B", "\xF6\x60", "\xF6\x67",
  /* U+9438 */	"\xC5\x4D", "\xF6\x65", "\xC5\x4C", "\xF6\x5F",
  /* U+943C */	"\xF6\x63", "\xF6\x62", "\x9A\x4F", "\xF6\x5E",
  /* U+9440 */	"\xF6\x69", "\x00\x00", "\xFE\x40", "\xFE\x43",
  /* U+9444 */	"\xC5\xB1", "\xF7\x6D", "\xF7\x70", "\xF7\x6C",
  /* U+9448 */	"\xF7\x6E", "\xF7\x6F", "\xF7\x69", "\xF7\x6A",
  /* U+944C */	"\xF7\x67", "\x96\xDD", "\x00\x00", "\xF7\x6B",
  /* U+9450 */	"\xF7\x68", "\xC5\xB2", "\xC5\xB3", "\x00\x00",
  /* U+9454 */	"\x9A\x51", "\xF8\x4B", "\x00\x00", "\xF8\x4D",
  /* U+9458 */	"\x96\xA7", "\x00\x00", "\x00\x00", "\x90\xB0",
  /* U+945C */	"\x00\x00", "\xF8\x4C", "\xF8\x4E", "\x00\x00",
  /* U+9460 */	"\xC5\xE0", "\x00\x00", "\xF8\x4A", "\xC5\xDF",
  /* U+9464 */	"\xC5\xE1", "\x9C\x4E", "\x00\x00", "\x94\x43",
  /* U+9468 */	"\xF8\xCB", "\xF8\xCC", "\xC6\x44", "\xF8\xCA",
  /* U+946C */	"\x8E\xBA", "\xF9\x53", "\xF9\x52", "\xF9\x54",
  /* U+9470 */	"\xC6\x5F", "\xF9\x55", "\xC6\x5E", "\xF9\x56",
  /* U+9474 */	"\xF9\x72", "\xF9\x75", "\xF9\x74", "\xC6\x68",
  /* U+9478 */	"\xF9\x73", "\x9A\x52", "\x00\x00", "\xFC\xC1",
  /* U+947C */	"\xC6\x72", "\xC6\x70", "\xC6\x71", "\xC6\x77",
  /* U+9480 */	"\xF9\xC0", "\xF9\xC1", "\xF9\xBF", "\xF9\xC9",
  /* U+9484 */	"\x00\x00", "\x8B\xE9", "\x00\x00", "\x00\x00",
  /* U+9488 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+948C */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+9490 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+9494 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+9498 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+949C */	"\x00\x00", "\x00\x00", "\x00\x00", "\x9C\xAF",
  /* U+94A0 */	"\x00\x00", "\x00\x00", "\x8B\xFD", "\x00\x00",
  /* U+94A4 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+94A8 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+94AC */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+94B0 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+94B4 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+94B8 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+94BC */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+94C0 */	"\x00\x00", "\x9A\xBC", "\x00\x00", "\x9A\xB8",
  /* U+94C4 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+94C8 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+94CC */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+94D0 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+94D4 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+94D8 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+94DC */	"\x9A\xAE", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+94E0 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+94E4 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+94E8 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+94EC */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+94F0 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+94F4 */	"\x00\x00", "\x00\x00", "\x9A\xA7", "\x00\x00",
  /* U+94F8 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+94FC */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+9500 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+9504 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+9508 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+950C */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+9510 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+9514 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+9518 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+951C */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+9520 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+9524 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+9528 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+952C */	"\x00\x00", "\x9A\x53", "\x00\x00", "\x00\x00",
  /* U+9530 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+9534 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+9538 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+953C */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+9540 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+9544 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x9D\x74",
  /* U+9548 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+954C */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+9550 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+9554 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+9558 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+955C */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+9560 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+9564 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+9568 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+956C */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+9570 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+9574 */	"\x00\x00", "\x00\x00", "\x00\x00", "\xAA\xF8",
  /* U+9578 */	"\x8B\xEA", "\x00\x00", "\xD8\x44", "\xDC\x78",
  /* U+957C */	"\xE8\xA5", "\xF3\x76", "\x00\x00", "\x8B\xEB",
  /* U+9580 */	"\xAA\xF9", "\x00\x00", "\xAD\xAC", "\xB0\x7B",
  /* U+9584 */	"\x00\x00", "\x90\xB2", "\xD8\x45", "\x00\x00",
  /* U+9588 */	"\xD8\x46", "\xB3\xAC", "\x00\x00", "\xB6\x7D",
  /* U+958C */	"\xDC\x7A", "\xDC\x79", "\xB6\xA3", "\xB6\x7C",
  /* U+9590 */	"\xDC\x7B", "\xB6\x7E", "\xB6\xA2", "\xB6\xA1",
  /* U+9594 */	"\xB6\x7B", "\x00\x00", "\x95\xE9", "\x95\xE8",
  /* U+9598 */	"\xB9\x68", "\x95\xE6", "\x00\x00", "\xE0\xD0",
  /* U+959C */	"\xE0\xCE", "\x00\x00", "\xE0\xCF", "\xE0\xCD",
  /* U+95A0 */	"\x90\xB5", "\xBB\xD2", "\x9A\x54", "\xBB\xD5",
  /* U+95A4 */	"\xBB\xD7", "\xBB\xD6", "\x90\xB3", "\x95\xE7",
  /* U+95A8 */	"\xBB\xD3", "\xBB\xD4", "\x8B\x50", "\xE8\xA7",
  /* U+95AC */	"\xE8\xA6", "\xBE\x5B", "\xE8\xA8", "\x00\x00",
  /* U+95B0 */	"\xE8\xA9", "\xBE\x5C", "\x00\x00", "\x00\x00",
  /* U+95B4 */	"\x00\x00", "\xEC\x4D", "\xEC\x4B", "\xEE\xF3",
  /* U+95B8 */	"\x00\x00", "\xEC\x49", "\xEC\x4A", "\xC0\x46",
  /* U+95BC */	"\xEC\x46", "\xEC\x4E", "\xEC\x48", "\xEC\x4C",
  /* U+95C0 */	"\xEE\xEF", "\x00\x00", "\x00\x00", "\xEE\xF1",
  /* U+95C4 */	"\x00\x00", "\xEE\xF2", "\xC1\xF3", "\xEE\xEE",
  /* U+95C8 */	"\xC1\xF2", "\xEE\xF0", "\xC1\xEF", "\xC1\xF0",
  /* U+95CC */	"\xC1\xF1", "\xEC\x47", "\x00\x00", "\x00\x00",
  /* U+95D0 */	"\xC2\xF5", "\xF1\x6E", "\xF1\x6C", "\xF1\x6D",
  /* U+95D4 */	"\xC2\xF3", "\xC2\xF6", "\xC2\xF4", "\x00\x00",
  /* U+95D8 */	"\x00\x00", "\x00\x00", "\xF3\x77", "\xF3\x78",
  /* U+95DC */	"\xC3\xF6", "\x00\x00", "\xF5\x45", "\xF5\x47",
  /* U+95E0 */	"\xF5\x46", "\xC4\xC4", "\xC5\x50", "\xF6\x6D",
  /* U+95E4 */	"\xF6\x6C", "\xF6\x6B", "\x00\x00", "\x00\x00",
  /* U+95E8 */	"\x8B\xEC", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+95EC */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+95F0 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+95F4 */	"\x9A\x56", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+95F8 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+95FC */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+9600 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+9604 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+9608 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+960C */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+9610 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+9614 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+9618 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+961C */	"\xAA\xFA", "\x8B\xFB", "\xC9\xAA", "\x00\x00",
  /* U+9620 */	"\xCA\x58", "\xA6\xE9", "\xCA\x56", "\xCA\x59",
  /* U+9624 */	"\xCA\x57", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+9628 */	"\xCB\xAE", "\x00\x00", "\xA8\xC1", "\x00\x00",
  /* U+962C */	"\xA8\xC2", "\xCB\xB0", "\xA8\xBF", "\xCB\xAF",
  /* U+9630 */	"\xCB\xAD", "\xA8\xC0", "\xA8\xBE", "\x9A\x57",
  /* U+9634 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+9638 */	"\xA0\xAA", "\xCD\xD8", "\xCD\xDB", "\xAA\xFD",
  /* U+963C */	"\xCD\xDA", "\xCD\xD9", "\x00\x00", "\xAA\xFC",
  /* U+9640 */	"\xAA\xFB", "\x9F\xA6", "\xAB\x40", "\xCD\xDC",
  /* U+9644 */	"\xAA\xFE", "\x99\xCC", "\x00\x00", "\x00\x00",
  /* U+9648 */	"\x00\x00", "\x00\x00", "\xD0\xC6", "\xAD\xAE",
  /* U+964C */	"\xAD\xAF", "\xAD\xB0", "\xD0\xC7", "\xD0\xC3",
  /* U+9650 */	"\xAD\xAD", "\xD0\xC4", "\x00\x00", "\xD0\xC5",
  /* U+9654 */	"\xD0\xC2", "\x00\x00", "\x9C\x59", "\x00\x00",
  /* U+9658 */	"\xB0\xA4", "\x00\x00", "\x00\x00", "\xB0\xA1",
  /* U+965C */	"\xD4\x45", "\xB0\xA2", "\xB0\xA5", "\xD4\x46",
  /* U+9660 */	"\x00\x00", "\xB0\x7E", "\xB0\x7C", "\xB0\x7D",
  /* U+9664 */	"\xB0\xA3", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+9668 */	"\x00\x00", "\x99\xB5", "\xB3\xAD", "\xD8\x49",
  /* U+966C */	"\xB3\xB5", "\xD8\x48", "\x00\x00", "\xD8\x4B",
  /* U+9670 */	"\xB3\xB1", "\xD8\x4A", "\xB6\xAB", "\xB3\xAF",
  /* U+9674 */	"\xB3\xB2", "\xB3\xAE", "\xB3\xB3", "\xB3\xB4",
  /* U+9678 */	"\xB3\xB0", "\x00\x00", "\x00\x00", "\x90\xBE",
  /* U+967C */	"\xD8\x47", "\xB6\xA7", "\xDC\x7D", "\x00\x00",
  /* U+9680 */	"\xDC\xA3", "\x9F\xAF", "\x00\x00", "\xDC\xA2",
  /* U+9684 */	"\xB6\xAC", "\xB6\xA8", "\xB6\xA9", "\xDC\x7C",
  /* U+9688 */	"\xDC\x7E", "\xDC\xA1", "\xB6\xA4", "\xB6\xA6",
  /* U+968C */	"\x00\x00", "\xB6\xAA", "\xB6\xA5", "\x95\xF2",
  /* U+9690 */	"\x00\x00", "\xE0\xD3", "\xE0\xD1", "\xE0\xD2",
  /* U+9694 */	"\xB9\x6A", "\xB9\x6B", "\x90\xBF", "\xE0\xD4",
  /* U+9698 */	"\xB9\x69", "\xBB\xD8", "\x00\x00", "\xBB\xDA",
  /* U+969C */	"\xBB\xD9", "\x00\x00", "\xE4\xBB", "\x00\x00",
  /* U+96A0 */	"\x00\x00", "\xE4\xBC", "\xE8\xAB", "\x90\xC1",
  /* U+96A4 */	"\xE8\xAA", "\xFE\xE4", "\x00\x00", "\xC0\x47",
  /* U+96A8 */	"\xC0\x48", "\xEC\x4F", "\xC0\x49", "\x00\x00",
  /* U+96AC */	"\xEE\xF6", "\x00\x00", "\xEE\xF4", "\x00\x00",
  /* U+96B0 */	"\xEE\xF5", "\xC1\xF4", "\x00\x00", "\xF1\x6F",
  /* U+96B4 */	"\xC3\xF7", "\x00\x00", "\x90\xC4", "\x00\x00",
  /* U+96B8 */	"\xC1\xF5", "\xAB\x41", "\x00\x00", "\xB0\xA6",
  /* U+96BC */	"\xD4\x47", "\x90\xC7", "\x00\x00", "\xD8\x4C",
  /* U+96C0 */	"\xB3\xB6", "\xB6\xAD", "\xDC\xA4", "\xDC\xA6",
  /* U+96C4 */	"\xB6\xAF", "\xB6\xAE", "\xB6\xB0", "\xB6\xB1",
  /* U+96C8 */	"\xDC\xA5", "\xB9\x6E", "\xB9\x6F", "\xB9\x6D",
  /* U+96CC */	"\xBB\xDB", "\xB9\x6C", "\xE0\xD5", "\x00\x00",
  /* U+96D0 */	"\x00\x00", "\x00\x00", "\xBB\xDC", "\xE8\xAC",
  /* U+96D4 */	"\xEC\x50", "\xC0\x4A", "\xC1\xF6", "\xF1\x70",
  /* U+96D8 */	"\xF1\x74", "\xC2\xF9", "\xF1\x71", "\xC2\xFA",
  /* U+96DC */	"\xC2\xF8", "\xF1\x75", "\xC2\xFB", "\xF1\x73",
  /* U+96E0 */	"\x00\x00", "\xF3\x79", "\xC2\xF7", "\xC3\xF8",
  /* U+96E4 */	"\x00\x00", "\xF8\xCD", "\x00\x00", "\x00\x00",
  /* U+96E8 */	"\xAB\x42", "\xB3\xB8", "\xB3\xB7", "\x00\x00",
  /* U+96EC */	"\x00\x00", "\x00\x00", "\x00\x00", "\xB6\xB2",
  /* U+96F0 */	"\xDC\xA8", "\xDC\xA7", "\xB6\xB3", "\x00\x00",
  /* U+96F4 */	"\x92\xE4", "\xE0\xD9", "\xB9\x73", "\xB9\x70",
  /* U+96F8 */	"\xE0\xD8", "\xB9\x72", "\xE0\xD6", "\xB9\x71",
  /* U+96FC */	"\x00\x00", "\xE0\xD7", "\x00\x00", "\xE4\xBD",
  /* U+9700 */	"\xBB\xDD", "\x00\x00", "\xE8\xAF", "\x9F\x52",
  /* U+9704 */	"\xBE\x5D", "\xE8\xAD", "\xBE\x5E", "\xBE\x5F",
  /* U+9708 */	"\xE8\xAE", "\xBE\x60", "\x00\x00", "\xEC\x51",
  /* U+970C */	"\x00\x00", "\xC0\x4E", "\xC0\x4B", "\xC0\x50",
  /* U+9710 */	"\xEC\x53", "\xC0\x4C", "\xEC\x52", "\xC0\x4F",
  /* U+9714 */	"\x00\x00", "\x00\x00", "\xC0\x4D", "\x00\x00",
  /* U+9718 */	"\xEE\xF9", "\xEE\xFB", "\x00\x00", "\x90\xDB",
  /* U+971C */	"\xC1\xF7", "\xEE\xFA", "\xC1\xF8", "\xEE\xF8",
  /* U+9720 */	"\xEE\xF7", "\xA0\x66", "\xF1\x77", "\xF1\x76",
  /* U+9724 */	"\xC2\xFC", "\xF1\x78", "\xF3\x7E", "\xC3\xFA",
  /* U+9728 */	"\xF3\x7D", "\xF3\x7A", "\xC3\xF9", "\xF3\x7B",
  /* U+972C */	"\xF3\x7C", "\x00\x00", "\xF5\x48", "\xF5\x49",
  /* U+9730 */	"\xC4\xC5", "\x90\xD2", "\xC5\x53", "\x00\x00",
  /* U+9734 */	"\x00\x00", "\xF6\x6E", "\x90\xD4", "\x00\x00",
  /* U+9738 */	"\xC5\x51", "\xC5\x52", "\xF6\x6F", "\x00\x00",
  /* U+973C */	"\x00\x00", "\xC5\xB4", "\xC5\xB5", "\xF7\x71",
  /* U+9740 */	"\x9A\x5B", "\x95\xFD", "\xC6\x45", "\xF8\xCF",
  /* U+9744 */	"\xC6\x47", "\x00\x00", "\xF8\xCE", "\xF8\xD0",
  /* U+9748 */	"\xC6\x46", "\xF9\x57", "\x00\x00", "\xF9\xAD",
  /* U+974C */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+9750 */	"\x00\x00", "\x8B\xC4", "\xAB\x43", "\x00\x00",
  /* U+9754 */	"\x00\x00", "\x8C\x66", "\xB9\x74", "\x90\xDE",
  /* U+9758 */	"\xE4\xBE", "\x00\x00", "\xE8\xB0", "\xC0\x51",
  /* U+975C */	"\xC0\x52", "\x9C\xE4", "\xAB\x44", "\x90\xE1",
  /* U+9760 */	"\xBE\x61", "\xC3\xFB", "\xAD\xB1", "\x00\x00",
  /* U+9764 */	"\x00\x00", "\x00\x00", "\xC0\x53", "\x00\x00",
  /* U+9768 */	"\xC5\xE2", "\xAD\xB2", "\xD8\x4D", "\x00\x00",
  /* U+976C */	"\xDC\xA9", "\x9E\x46", "\xDC\xAB", "\x00\x00",
  /* U+9770 */	"\xDC\xAA", "\x96\x51", "\xE0\xDD", "\xE0\xDA",
  /* U+9774 */	"\xB9\x75", "\x00\x00", "\xB9\x76", "\xE0\xDB",
  /* U+9778 */	"\xE0\xDC", "\x00\x00", "\xE4\xC0", "\xE4\xC5",
  /* U+977C */	"\xBB\xDE", "\xE4\xBF", "\xE4\xC1", "\xE4\xC8",
  /* U+9780 */	"\xE4\xC3", "\xE4\xC7", "\xE4\xC4", "\xE4\xC2",
  /* U+9784 */	"\xE4\xC6", "\xBB\xDF", "\x00\x00", "\xFB\x58",
  /* U+9788 */	"\xE8\xB3", "\x90\xE6", "\xE8\xB1", "\xBE\x63",
  /* U+978C */	"\x00\x00", "\xBE\x62", "\xE8\xB2", "\xBE\x64",
  /* U+9790 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+9794 */	"\xEC\x56", "\x00\x00", "\x00\x00", "\xEC\x55",
  /* U+9798 */	"\xC0\x54", "\xEC\x54", "\xEE\xFC", "\x96\x50",
  /* U+979C */	"\xEE\xFE", "\xEF\x41", "\xEF\x40", "\x90\xE7",
  /* U+97A0 */	"\xC1\xF9", "\xEE\xFD", "\xF1\xA1", "\xC2\xFD",
  /* U+97A4 */	"\xF1\x7D", "\xF1\xA2", "\xC2\xFE", "\x00\x00",
  /* U+97A8 */	"\xF1\x7B", "\x00\x00", "\xF1\x7E", "\xF1\x7C",
  /* U+97AC */	"\xF1\x79", "\xC3\x40", "\xF1\x7A", "\x00\x00",
  /* U+97B0 */	"\x00\x00", "\x90\xE8", "\x9A\x5D", "\xF3\xA1",
  /* U+97B4 */	"\x9F\x7A", "\x00\x00", "\xF3\xA3", "\xF3\xA2",
  /* U+97B8 */	"\x9B\x5C", "\xF5\x4A", "\x9F\x7C", "\xF5\x4B",
  /* U+97BC */	"\x00\x00", "\xFC\x52", "\x90\xE9", "\xF6\x70",
  /* U+97C0 */	"\x90\xEA", "\xC5\xB7", "\x9A\x5E", "\xC5\xB6",
  /* U+97C4 */	"\xF8\x4F", "\xF8\x50", "\xC6\x48", "\xF8\xD1",
  /* U+97C8 */	"\x9F\x76", "\xC6\x69", "\x00\x00", "\xAD\xB3",
  /* U+97CC */	"\xB6\xB4", "\xE4\xCA", "\xE4\xC9", "\xE8\xB5",
  /* U+97D0 */	"\xE8\xB4", "\x00\x00", "\x90\xEB", "\xC1\xFA",
  /* U+97D4 */	"\xEF\x43", "\xEF\x42", "\xF1\xA5", "\xF1\xA3",
  /* U+97D8 */	"\xF1\xA6", "\xF1\xA4", "\x00\x00", "\x00\x00",
  /* U+97DC */	"\xC3\xFC", "\xF3\xA4", "\xF3\xA5", "\xF3\xA6",
  /* U+97E0 */	"\x90\xEC", "\xF6\x71", "\x00\x00", "\xF7\x72",
  /* U+97E4 */	"\x00\x00", "\xF8\xD2", "\x8B\xEE", "\x00\x00",
  /* U+97E8 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+97EC */	"\x00\x00", "\xAD\xB4", "\x90\xEE", "\x00\x00",
  /* U+97F0 */	"\xEC\x57", "\xEF\x44", "\x91\xC6", "\xAD\xB5",
  /* U+97F4 */	"\x00\x00", "\x90\xF2", "\xBB\xE0", "\x00\x00",
  /* U+97F8 */	"\xEC\x58", "\xC3\x41", "\xF1\xA7", "\xC3\xFD",
  /* U+97FC */	"\x00\x00", "\xF5\x4C", "\xF5\x4D", "\xC5\x54",
  /* U+9800 */	"\xF8\x51", "\xAD\xB6", "\xB3\xBB", "\xB3\xBC",
  /* U+9804 */	"\xD8\x4E", "\xB6\xB5", "\xB6\xB6", "\xDC\xAC",
  /* U+9808 */	"\xB6\xB7", "\x00\x00", "\xB9\x7A", "\x00\x00",
  /* U+980C */	"\xB9\x7C", "\xE0\xDF", "\xE0\xE0", "\xE0\xDE",
  /* U+9810 */	"\xB9\x77", "\xB9\x78", "\xB9\x7B", "\xB9\x79",
  /* U+9814 */	"\xFC\xBC", "\x8A\x74", "\xE4\xCB", "\xBB\xE1",
  /* U+9818 */	"\xBB\xE2", "\x00\x00", "\x00\x00", "\xE8\xBC",
  /* U+981C */	"\xBE\x67", "\xE8\xB7", "\xE8\xB6", "\x96\x57",
  /* U+9820 */	"\xE8\xBB", "\xBE\x65", "\x00\x00", "\x9C\xEF",
  /* U+9824 */	"\xC0\x5B", "\x00\x00", "\xE8\xB8", "\xE8\xBD",
  /* U+9828 */	"\xE8\xBA", "\xE8\xB9", "\x00\x00", "\xBE\x66",
  /* U+982C */	"\x00\x00", "\xC0\x59", "\x9F\xDF", "\xEC\x5A",
  /* U+9830 */	"\xC0\x55", "\x00\x00", "\xEC\x5B", "\x90\xF7",
  /* U+9834 */	"\x90\xF6", "\xEC\x59", "\x00\x00", "\xC0\x58",
  /* U+9838 */	"\xC0\x56", "\xC0\x5A", "\x00\x00", "\xC0\x57",
  /* U+983C */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+9840 */	"\x00\x00", "\xEF\x45", "\x00\x00", "\xEF\x4A",
  /* U+9844 */	"\xEF\x46", "\xEF\x49", "\xC1\xFB", "\x9B\x5E",
  /* U+9848 */	"\xED\xD4", "\xEF\x48", "\xEF\x47", "\x90\xF8",
  /* U+984C */	"\xC3\x44", "\xC3\x42", "\xC3\x45", "\xC3\x43",
  /* U+9850 */	"\xF1\xA8", "\xF1\xA9", "\xF1\xAA", "\xC3\x46",
  /* U+9854 */	"\x00\x00", "\x00\x00", "\x00\x00", "\xF3\xAA",
  /* U+9858 */	"\xC4\x40", "\xF3\xA8", "\x00\x00", "\xC4\x41",
  /* U+985C */	"\xF3\xA7", "\xF3\xA9", "\xC3\xFE", "\xF5\x51",
  /* U+9860 */	"\xF5\x4E", "\x00\x00", "\xF5\x4F", "\xF5\x50",
  /* U+9864 */	"\xF6\x72", "\xC5\x56", "\x90\xF9", "\xC5\x55",
  /* U+9868 */	"\x8C\xC9", "\xF7\x74", "\xF7\x73", "\xC5\xB8",
  /* U+986C */	"\xFA\x6A", "\x00\x00", "\x00\x00", "\xC5\xE3",
  /* U+9870 */	"\xC6\x49", "\xC6\x60", "\xF9\x58", "\xF9\xAE",
  /* U+9874 */	"\xF9\xAF", "\x8B\xEF", "\x00\x00", "\x00\x00",
  /* U+9878 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+987C */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+9880 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+9884 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+9888 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+988C */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+9890 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+9894 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+9898 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+989C */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+98A0 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+98A4 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+98A8 */	"\xAD\xB7", "\xDC\xAD", "\x00\x00", "\x00\x00",
  /* U+98AC */	"\xE0\xE1", "\xE4\xCC", "\xE4\xCD", "\xBB\xE3",
  /* U+98B0 */	"\x00\x00", "\xBB\xE4", "\xE8\xBE", "\xBE\x68",
  /* U+98B4 */	"\x9F\xE0", "\x00\x00", "\xC1\xFC", "\x91\x42",
  /* U+98B8 */	"\xF1\xAB", "\x9A\x62", "\xC3\x47", "\xF3\xAD",
  /* U+98BC */	"\xC4\x42", "\xF3\xAC", "\xF3\xAE", "\xF3\xAB",
  /* U+98C0 */	"\xF6\x75", "\xF5\x52", "\xF5\x53", "\x95\x69",
  /* U+98C4 */	"\xC4\xC6", "\x00\x00", "\xF6\x74", "\x91\x44",
  /* U+98C8 */	"\x91\x43", "\xF6\x73", "\x91\x41", "\xF7\x75",
  /* U+98CC */	"\xF9\xB0", "\x00\x00", "\x8B\xF0", "\x00\x00",
  /* U+98D0 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+98D4 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+98D8 */	"\x00\x00", "\x00\x00", "\x00\x00", "\xAD\xB8",
  /* U+98DC */	"\x96\x60", "\x00\x00", "\x8B\xF1", "\xAD\xB9",
  /* U+98E0 */	"\x99\xF6", "\x91\x49", "\xB0\xA7", "\xD4\x48",
  /* U+98E4 */	"\x00\x00", "\xD8\x4F", "\x91\x4A", "\xB6\xB8",
  /* U+98E8 */	"\x00\x00", "\xB6\xBB", "\xB6\xB9", "\xDC\xAE",
  /* U+98EC */	"\x91\x4B", "\xB6\xBD", "\x00\x00", "\xB6\xBA",
  /* U+98F0 */	"\x00\x00", "\x9A\x64", "\xB6\xBC", "\x00\x00",
  /* U+98F4 */	"\xB9\x7E", "\x8A\xBF", "\xE0\xE2", "\x00\x00",
  /* U+98F8 */	"\x00\x00", "\xE0\xE3", "\xE8\xC0", "\x00\x00",
  /* U+98FC */	"\xB9\x7D", "\xB9\xA1", "\xB9\xA2", "\x00\x00",
  /* U+9900 */	"\xE4\xCF", "\x00\x00", "\xE4\xCE", "\xBB\xE5",
  /* U+9904 */	"\x00\x00", "\xBB\xE6", "\x00\x00", "\xE4\xD0",
  /* U+9908 */	"\xE8\xBF", "\xBB\xE8", "\xBE\x69", "\x00\x00",
  /* U+990C */	"\xBB\xE7", "\x00\x00", "\x9A\x66", "\x00\x00",
  /* U+9910 */	"\xC0\x5C", "\xE8\xC1", "\xBE\x6B", "\xBE\x6A",
  /* U+9914 */	"\xE8\xC2", "\xE8\xC5", "\xE8\xC3", "\xE8\xC4",
  /* U+9918 */	"\xBE\x6C", "\x9A\x67", "\xC0\x61", "\xC0\x5F",
  /* U+991C */	"\x9A\x69", "\x00\x00", "\xC0\x5E", "\xEC\x5D",
  /* U+9920 */	"\x00\x00", "\xC0\x60", "\x00\x00", "\x00\x00",
  /* U+9924 */	"\xEC\x5C", "\xEF\x4B", "\x00\x00", "\xEC\x5E",
  /* U+9928 */	"\xC0\x5D", "\xEC\x5F", "\xEF\x4E", "\xEF\x4C",
  /* U+992C */	"\xEF\x4D", "\xEF\x52", "\xC3\x4B", "\xEF\x51",
  /* U+9930 */	"\xEF\x54", "\xEF\x53", "\xEF\x50", "\xEF\x4F",
  /* U+9934 */	"\x00\x00", "\xC1\xFD", "\x00\x00", "\x9A\x6A",
  /* U+9938 */	"\x96\x52", "\x91\x4D", "\xF1\xAE", "\x96\x66",
  /* U+993C */	"\xF1\xAD", "\xC3\x4A", "\xC3\x48", "\xC3\x49",
  /* U+9940 */	"\x9F\x7B", "\xF1\xAC", "\x9A\x6B", "\xF3\xB1",
  /* U+9944 */	"\x00\x00", "\xC4\x43", "\x00\x00", "\xF3\xB0",
  /* U+9948 */	"\xF3\xAF", "\xC4\x44", "\xA0\x6C", "\xF5\x58",
  /* U+994C */	"\xF5\x57", "\x96\x67", "\xF5\x55", "\x00\x00",
  /* U+9950 */	"\xF5\x54", "\xC4\xC8", "\xC4\xC7", "\xF5\x59",
  /* U+9954 */	"\xF7\x76", "\xC5\xB9", "\xF6\x77", "\xC5\x57",
  /* U+9958 */	"\xF6\x76", "\xF5\x56", "\x00\x00", "\xF7\x77",
  /* U+995C */	"\xC5\xE4", "\x9A\x6C", "\xC6\x61", "\xF9\x59",
  /* U+9960 */	"\x00\x00", "\xF9\xB1", "\x9A\x6D", "\x8B\xF2",
  /* U+9964 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+9968 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+996C */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+9970 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+9974 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+9978 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+997C */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+9980 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+9984 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+9988 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+998C */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+9990 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+9994 */	"\x00\x00", "\x00\x00", "\xAD\xBA", "\xD8\x50",
  /* U+9998 */	"\xEF\x55", "\xAD\xBB", "\x00\x00", "\x96\x6A",
  /* U+999C */	"\xE4\xD2", "\xE4\xD1", "\xEC\x60", "\x00\x00",
  /* U+99A0 */	"\x00\x00", "\xEF\x57", "\x00\x00", "\xEF\x56",
  /* U+99A4 */	"\xFC\xEA", "\xC3\x4C", "\xF3\xB2", "\xF3\xB3",
  /* U+99A8 */	"\xC4\xC9", "\x00\x00", "\x96\x6C", "\xF9\xB2",
  /* U+99AC */	"\xB0\xA8", "\xB6\xBF", "\xB6\xBE", "\xE0\xE4",
  /* U+99B0 */	"\xE0\xE6", "\xB9\xA4", "\xE0\xE5", "\xB9\xA3",
  /* U+99B4 */	"\xB9\xA5", "\xE0\xE7", "\x00\x00", "\x00\x00",
  /* U+99B8 */	"\x91\xC4", "\xE4\xD4", "\xE4\xD6", "\xE4\xD5",
  /* U+99BC */	"\x96\x77", "\xE4\xD8", "\x00\x00", "\x00\x00",
  /* U+99C0 */	"\x00\x00", "\xBB\xE9", "\xE4\xD7", "\xE4\xD3",
  /* U+99C4 */	"\x99\xF4", "\x9A\x6F", "\x00\x00", "\xE4\xD9",
  /* U+99C8 */	"\x00\x00", "\xE8\xCC", "\x00\x00", "\xE8\xCF",
  /* U+99CC */	"\xE8\xD1", "\xE8\xC7", "\xE8\xCB", "\xE8\xC8",
  /* U+99D0 */	"\xBE\x6E", "\xBE\x71", "\xBE\x73", "\xE8\xC9",
  /* U+99D4 */	"\xE8\xCA", "\xBE\x72", "\xE8\xCD", "\xE8\xD0",
  /* U+99D8 */	"\xE8\xCE", "\xBE\x74", "\x9F\xAB", "\xBE\x70",
  /* U+99DC */	"\xE8\xC6", "\xBE\x6D", "\x00\x00", "\xBE\x6F",
  /* U+99E0 */	"\x8C\xBE", "\x8E\xC1", "\xC0\x63", "\xEC\x66",
  /* U+99E4 */	"\xEC\x64", "\xEC\x63", "\x95\x55", "\xEC\x69",
  /* U+99E8 */	"\x00\x00", "\xEC\x68", "\xEC\x67", "\x00\x00",
  /* U+99EC */	"\xEC\x62", "\xC0\x62", "\xEC\x61", "\x00\x00",
  /* U+99F0 */	"\xEC\x65", "\xC0\x64", "\x00\x00", "\x00\x00",
  /* U+99F4 */	"\xEF\x5A", "\x91\x52", "\xEF\x5E", "\xEF\x5B",
  /* U+99F8 */	"\xEF\x5D", "\xEF\x5C", "\xEF\x59", "\xEF\x5F",
  /* U+99FC */	"\xEF\x62", "\xEF\x60", "\xEF\x61", "\xC2\x40",
  /* U+9A00 */	"\x00\x00", "\xC1\xFE", "\xEF\x58", "\xEF\x63",
  /* U+9A04 */	"\xF1\xB3", "\xF1\xB6", "\xF1\xB8", "\xF1\xB7",
  /* U+9A08 */	"\x00\x00", "\xF1\xB1", "\xF1\xB5", "\xF1\xB0",
  /* U+9A0C */	"\x91\x53", "\xF1\xB2", "\xC3\x4D", "\xF1\xAF",
  /* U+9A10 */	"\x91\x55", "\xF1\xB4", "\x00\x00", "\x00\x00",
  /* U+9A14 */	"\xF3\xC0", "\xF3\xB5", "\xC4\x45", "\x00\x00",
  /* U+9A18 */	"\x00\x00", "\xC4\x46", "\xF3\xB4", "\xF3\xB9",
  /* U+9A1C */	"\xF3\xBF", "\xF3\xB7", "\xF3\xBE", "\x95\x5D",
  /* U+9A20 */	"\xF3\xBB", "\x96\x71", "\xF3\xBA", "\xF3\xBD",
  /* U+9A24 */	"\xF3\xB8", "\xF3\xB6", "\x9C\x6D", "\xF3\xBC",
  /* U+9A28 */	"\x00\x00", "\xF5\x60", "\xF5\x5E", "\xC4\xCA",
  /* U+9A2C */	"\xF5\x5D", "\xF5\x63", "\xF5\x61", "\x96\x73",
  /* U+9A30 */	"\xC4\xCB", "\xF5\x5C", "\xF5\x5A", "\x00\x00",
  /* U+9A34 */	"\xF5\x5B", "\xC4\xCD", "\xF5\x5F", "\xC4\xCC",
  /* U+9A38 */	"\xF5\x62", "\xF6\x78", "\xF6\x7E", "\x91\x54",
  /* U+9A3C */	"\x9A\x71", "\xF6\x79", "\xC5\x5B", "\xF6\xA1",
  /* U+9A40 */	"\xC5\x5A", "\xF6\x7D", "\xF6\x7C", "\xC5\x59",
  /* U+9A44 */	"\xF6\x7B", "\xC5\x58", "\xF6\x7A", "\x00\x00",
  /* U+9A48 */	"\xF7\x7D", "\xF7\xA1", "\xF7\x7E", "\x00\x00",
  /* U+9A4C */	"\xF7\x7B", "\xC5\xBB", "\xF7\x78", "\xF7\x7C",
  /* U+9A50 */	"\xF7\xA3", "\x00\x00", "\xF7\xA2", "\xF7\x79",
  /* U+9A54 */	"\xF7\x7A", "\xC5\xBA", "\xF8\x52", "\xC5\xE7",
  /* U+9A58 */	"\x91\x56", "\xF8\x53", "\xC5\xE5", "\xC5\xE6",
  /* U+9A5C */	"\x96\x6D", "\x00\x00", "\xF8\xD3", "\xC6\x4A",
  /* U+9A60 */	"\xF9\x76", "\x00\x00", "\xC6\x6A", "\x95\x57",
  /* U+9A64 */	"\xF9\xB3", "\xC6\x6B", "\xF9\xB4", "\xF9\xB5",
  /* U+9A68 */	"\xF9\xC3", "\xF9\xC2", "\xC6\x7A", "\xF9\xCD",
  /* U+9A6C */	"\x89\xC6", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+9A70 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+9A74 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+9A78 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+9A7C */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+9A80 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+9A84 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+9A88 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+9A8C */	"\x00\x00", "\x00\x00", "\x00\x00", "\x89\xC7",
  /* U+9A90 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+9A94 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+9A98 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+9A9C */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+9AA0 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+9AA4 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+9AA8 */	"\xB0\xA9", "\x00\x00", "\x00\x00", "\xE0\xE9",
  /* U+9AAC */	"\x00\x00", "\xE0\xE8", "\x00\x00", "\xBB\xEA",
  /* U+9AB0 */	"\xBB\xEB", "\xE4\xDA", "\x8A\x6A", "\xE8\xD2",
  /* U+9AB4 */	"\xEC\x6C", "\x00\x00", "\x8B\x57", "\xBE\x75",
  /* U+9AB8 */	"\xC0\x65", "\xEC\x6A", "\x9F\xE1", "\xEC\x6D",
  /* U+9ABC */	"\xC0\x66", "\x9B\x5F", "\xEF\x64", "\xEC\x6B",
  /* U+9AC0 */	"\xF1\xB9", "\xC3\x4E", "\xF3\xC1", "\x00\x00",
  /* U+9AC4 */	"\x00\x00", "\x00\x00", "\xF5\x66", "\xF5\x64",
  /* U+9AC8 */	"\x00\x00", "\x00\x00", "\xF5\x65", "\x00\x00",
  /* U+9ACC */	"\x00\x00", "\xF6\xA2", "\x00\x00", "\xC5\x5C",
  /* U+9AD0 */	"\xF7\xA4", "\xC5\xEA", "\xC5\xBC", "\xC5\xE8",
  /* U+9AD4 */	"\xC5\xE9", "\xF8\xD4", "\xC6\x62", "\xA0\x5D",
  /* U+9AD8 */	"\xB0\xAA", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+9ADC */	"\xF1\xBA", "\x00\x00", "\x00\x00", "\xD4\x49",
  /* U+9AE0 */	"\x91\x5B", "\xB9\xA6", "\x91\x5C", "\xE4\xDB",
  /* U+9AE4 */	"\x00\x00", "\x00\x00", "\xBB\xEC", "\xE4\xDC",
  /* U+9AE8 */	"\x00\x00", "\x00\x00", "\x00\x00", "\xE8\xD4",
  /* U+9AEC */	"\xE8\xD3", "\xC0\x68", "\xBE\x76", "\xBE\x77",
  /* U+9AF0 */	"\x00\x00", "\xE8\xD7", "\xE8\xD6", "\xE8\xD5",
  /* U+9AF4 */	"\x91\x5E", "\x00\x00", "\xEC\x6E", "\xEC\x71",
  /* U+9AF8 */	"\x00\x00", "\xEC\x70", "\xEC\x6F", "\xC0\x67",
  /* U+9AFC */	"\xEF\x68", "\xEF\x66", "\xEF\x65", "\x9F\x5C",
  /* U+9B00 */	"\x00\x00", "\xEF\x67", "\x9F\x57", "\xC3\x4F",
  /* U+9B04 */	"\xF1\xBC", "\xF1\xBD", "\xC3\x50", "\x00\x00",
  /* U+9B08 */	"\xF1\xBB", "\x9F\x65", "\xF3\xC3", "\xF3\xC2",
  /* U+9B0C */	"\xF3\xC5", "\xC4\x47", "\xF3\xC4", "\x9A\x72",
  /* U+9B10 */	"\xF5\x67", "\xF5\x69", "\xF5\x68", "\x00\x00",
  /* U+9B14 */	"\x91\x60", "\xF6\xA3", "\xF6\xA6", "\xF6\xA4",
  /* U+9B18 */	"\xF6\xA5", "\xF7\xA5", "\xC5\xBD", "\x00\x00",
  /* U+9B1C */	"\x00\x00", "\x00\x00", "\xF8\x54", "\xF8\x55",
  /* U+9B20 */	"\xF8\x56", "\x00\x00", "\xC6\x4B", "\xC6\x63",
  /* U+9B24 */	"\xF9\xB6", "\xB0\xAB", "\x00\x00", "\xBE\x78",
  /* U+9B28 */	"\xC0\x69", "\xF1\xBE", "\x9F\x5E", "\xF7\xA6",
  /* U+9B2C */	"\x00\x00", "\x91\x61", "\xF9\xC4", "\xD4\x4A",
  /* U+9B30 */	"\x00\x00", "\xC6\x7B", "\xB0\xAC", "\xEC\x72",
  /* U+9B34 */	"\x91\x64", "\xF1\xBF", "\x00\x00", "\xF3\xC6",
  /* U+9B38 */	"\x00\x00", "\x9F\x41", "\xF6\xA7", "\xF7\xA7",
  /* U+9B3C */	"\xB0\xAD", "\x00\x00", "\xE4\xDD", "\xE4\xDE",
  /* U+9B40 */	"\x91\x69", "\xBB\xED", "\xBB\xEE", "\xE8\xD9",
  /* U+9B44 */	"\xBE\x7A", "\xBE\x79", "\xE8\xD8", "\x00\x00",
  /* U+9B48 */	"\xEF\x69", "\x00\x00", "\xF1\xC0", "\xF1\xC2",
  /* U+9B4C */	"\xF1\xC1", "\xC3\x53", "\xC3\x52", "\xC3\x51",
  /* U+9B50 */	"\x91\x68", "\xC5\x5E", "\xF6\xA8", "\x00\x00",
  /* U+9B54 */	"\xC5\x5D", "\xF7\xA9", "\xF7\xA8", "\x00\x00",
  /* U+9B58 */	"\xC6\x4C", "\xF8\xD5", "\xB3\xBD", "\xE0\xEA",
  /* U+9B5C */	"\x00\x00", "\x00\x00", "\x00\x00", "\xE4\xE1",
  /* U+9B60 */	"\xE4\xDF", "\xE4\xE0", "\x00\x00", "\x00\x00",
  /* U+9B64 */	"\xE8\xE2", "\x00\x00", "\xE8\xDD", "\xE8\xDA",
  /* U+9B68 */	"\xE8\xE1", "\x9A\x74", "\x00\x00", "\x00\x00",
  /* U+9B6C */	"\xE8\xE3", "\x00\x00", "\x00\x00", "\xBE\x7C",
  /* U+9B70 */	"\xE8\xE0", "\xE8\xDC", "\x00\x00", "\x00\x00",
  /* U+9B74 */	"\xE8\xDB", "\xE8\xDF", "\xE8\xDE", "\xBE\x7B",
  /* U+9B78 */	"\x00\x00", "\x00\x00", "\xEC\x7D", "\xEC\x78",
  /* U+9B7C */	"\xEC\x76", "\xEC\xA1", "\xEC\x77", "\x96\xB2",
  /* U+9B80 */	"\xEC\x73", "\x9A\x75", "\xEC\x79", "\xFD\xA5",
  /* U+9B84 */	"\x00\x00", "\xEC\x74", "\xEF\x72", "\xEC\x75",
  /* U+9B88 */	"\xEC\xA2", "\x00\x00", "\x00\x00", "\x9E\xE9",
  /* U+9B8C */	"\x00\x00", "\x8B\xBA", "\x91\x6D", "\xA0\x60",
  /* U+9B90 */	"\xEC\x7C", "\xC0\x6A", "\xEC\x7B", "\xEC\x7A",
  /* U+9B94 */	"\x00\x00", "\xEC\x7E", "\x00\x00", "\x9F\xDE",
  /* U+9B98 */	"\x00\x00", "\x00\x00", "\xEF\x6A", "\xEF\x6D",
  /* U+9B9C */	"\x00\x00", "\x9F\xC3", "\xEF\x6C", "\x96\xB5",
  /* U+9BA0 */	"\xEF\x74", "\xEF\x6F", "\xEF\x73", "\x00\x00",
  /* U+9BA4 */	"\xEF\x71", "\xEF\x70", "\xEF\x6E", "\x00\x00",
  /* U+9BA8 */	"\xEF\x6B", "\x00\x00", "\xC2\x43", "\xC2\x42",
  /* U+9BAC */	"\x00\x00", "\xC2\x44", "\xC2\x41", "\xEF\x75",
  /* U+9BB0 */	"\xA0\x67", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+9BB4 */	"\x00\x00", "\xF1\xC8", "\xF1\xCB", "\x00\x00",
  /* U+9BB8 */	"\xF1\xC9", "\xF1\xCD", "\x00\x00", "\x00\x00",
  /* U+9BBC */	"\x00\x00", "\xF1\xCE", "\x00\x00", "\xF1\xC6",
  /* U+9BC0 */	"\xC3\x58", "\xF1\xC7", "\x00\x00", "\xF1\xC5",
  /* U+9BC4 */	"\xF1\xCC", "\x00\x00", "\xF1\xC4", "\xF1\xC3",
  /* U+9BC8 */	"\xC3\x57", "\xC3\x55", "\xC3\x54", "\x00\x00",
  /* U+9BCC */	"\x00\x00", "\x00\x00", "\x00\x00", "\x96\xB3",
  /* U+9BD0 */	"\x00\x00", "\x00\x00", "\x00\x00", "\xF1\xCA",
  /* U+9BD4 */	"\xF3\xCF", "\xF3\xD5", "\xC4\x4A", "\xF3\xD0",
  /* U+9BD8 */	"\x00\x00", "\xF3\xD3", "\xF3\xD7", "\xC4\x4B",
  /* U+9BDC */	"\xF3\xD2", "\x9A\x76", "\xF3\xCA", "\x00\x00",
  /* U+9BE0 */	"\xF3\xC9", "\xF3\xD6", "\xF3\xCD", "\x00\x00",
  /* U+9BE4 */	"\xF3\xCB", "\xF3\xD4", "\xF3\xCC", "\xC4\x49",
  /* U+9BE8 */	"\xC4\x48", "\x95\xD5", "\xF3\xC7", "\xF3\xC8",
  /* U+9BEC */	"\xF3\xD1", "\x9E\xCA", "\x00\x00", "\x00\x00",
  /* U+9BF0 */	"\xF3\xCE", "\x9A\x77", "\x00\x00", "\x00\x00",
  /* U+9BF4 */	"\x9A\x78", "\x00\x00", "\x00\x00", "\xF5\x6C",
  /* U+9BF8 */	"\xF5\x6F", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+9BFC */	"\x00\x00", "\xC3\x56", "\x00\x00", "\x91\x70",
  /* U+9C00 */	"\x00\x00", "\x00\x00", "\x91\x6F", "\x00\x00",
  /* U+9C04 */	"\x00\x00", "\xF5\x6D", "\xF5\x73", "\xF5\x71",
  /* U+9C08 */	"\xF5\x6B", "\xF5\x76", "\x9F\xA3", "\xF5\x6A",
  /* U+9C0C */	"\x91\x71", "\xC4\xCF", "\xF5\x72", "\x00\x00",
  /* U+9C10 */	"\x96\xB1", "\x00\x00", "\xF5\x6E", "\xC4\xCE",
  /* U+9C14 */	"\xF5\x75", "\x9F\x63", "\x00\x00", "\xF5\x74",
  /* U+9C18 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x9F\x67",
  /* U+9C1C */	"\xF6\xAB", "\xF6\xAA", "\x00\x00", "\x8B\xB9",
  /* U+9C20 */	"\x9A\x7A", "\xF6\xB1", "\x00\x00", "\xF6\xAD",
  /* U+9C24 */	"\xF6\xB0", "\xC5\x60", "\x8B\x56", "\x00\x00",
  /* U+9C28 */	"\xF6\xAE", "\xF6\xAF", "\x00\x00", "\xF6\xA9",
  /* U+9C2C */	"\xF6\xAC", "\xC5\x5F", "\x00\x00", "\x9A\xDA",
  /* U+9C30 */	"\x00\x00", "\xC5\xBF", "\xF7\xB4", "\xF7\xAF",
  /* U+9C34 */	"\xF7\xB3", "\x96\xB0", "\xF7\xB6", "\xF7\xB2",
  /* U+9C38 */	"\x00\x00", "\xF7\xAE", "\x9A\x7E", "\xC5\xC1",
  /* U+9C3C */	"\xF7\xB1", "\xF7\xB5", "\xC5\xC0", "\xF7\xAC",
  /* U+9C40 */	"\xF5\x70", "\xF7\xB0", "\x00\x00", "\x00\x00",
  /* U+9C44 */	"\xF7\xAD", "\x9D\xDE", "\xF7\xAA", "\x00\x00",
  /* U+9C48 */	"\xF7\xAB", "\xC5\xBE", "\xF8\x5A", "\xF8\x5C",
  /* U+9C4C */	"\xF8\x5F", "\xF8\x5B", "\xF8\x60", "\x96\xAD",
  /* U+9C50 */	"\xF8\x59", "\x00\x00", "\xF8\x57", "\x96\xAE",
  /* U+9C54 */	"\xC5\xEB", "\xF8\x5D", "\xC5\xED", "\xC5\xEC",
  /* U+9C58 */	"\xF8\x58", "\xF8\x5E", "\x00\x00", "\x00\x00",
  /* U+9C5C */	"\x00\x00", "\x9E\xA1", "\xF8\xDA", "\xC6\x4D",
  /* U+9C60 */	"\xF8\xDB", "\x00\x00", "\xF8\xD9", "\xF8\xD6",
  /* U+9C64 */	"\x00\x00", "\x00\x00", "\xF8\xD8", "\xF8\xD7",
  /* U+9C68 */	"\xF9\x5A", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+9C6C */	"\x00\x00", "\xF9\x5C", "\xF9\x5B", "\x00\x00",
  /* U+9C70 */	"\x00\x00", "\xF9\x79", "\x9E\x50", "\xF9\x78",
  /* U+9C74 */	"\xF9\x77", "\xF9\x7A", "\x00\x00", "\xC6\x73",
  /* U+9C78 */	"\xC6\x74", "\xF9\xCA", "\xF9\xCE", "\x96\xAF",
  /* U+9C7C */	"\x8B\xF4", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+9C80 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+9C84 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+9C88 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+9C8C */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+9C90 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+9C94 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+9C98 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+9C9C */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+9CA0 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+9CA4 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+9CA8 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+9CAC */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+9CB0 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+9CB4 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+9CB8 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+9CBC */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+9CC0 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+9CC4 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+9CC8 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+9CCC */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+9CD0 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+9CD4 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+9CD8 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+9CDC */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+9CE0 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+9CE4 */	"\x00\x00", "\xB3\xBE", "\xDC\xAF", "\xE0\xED",
  /* U+9CE8 */	"\x00\x00", "\xB9\xA7", "\xE0\xEB", "\x00\x00",
  /* U+9CEC */	"\x00\x00", "\xE0\xEC", "\x00\x00", "\x00\x00",
  /* U+9CF0 */	"\x00\x00", "\xE4\xE2", "\xE4\xE3", "\xBB\xF1",
  /* U+9CF4 */	"\xBB\xEF", "\xE4\xE4", "\xBB\xF0", "\xE8\xE8",
  /* U+9CF8 */	"\x00\x00", "\xE8\xEB", "\xE8\xE5", "\xE8\xEC",
  /* U+9CFC */	"\xE8\xE4", "\xE8\xE6", "\x00\x00", "\xE8\xE7",
  /* U+9D00 */	"\xE8\xEA", "\x00\x00", "\x9F\xA4", "\xBE\xA1",
  /* U+9D04 */	"\xE8\xEF", "\xE8\xEE", "\xBE\x7D", "\xE8\xE9",
  /* U+9D08 */	"\xE8\xED", "\xBE\x7E", "\x00\x00", "\x00\x00",
  /* U+9D0C */	"\x96\xBD", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+9D10 */	"\xEC\xAC", "\x00\x00", "\xC0\x6F", "\x00\x00",
  /* U+9D14 */	"\xEC\xA7", "\xC0\x6B", "\x96\xF4", "\xEC\xA4",
  /* U+9D18 */	"\xEC\xAA", "\xEC\xAD", "\x00\x00", "\xC0\x70",
  /* U+9D1C */	"\x00\x00", "\xEC\xA9", "\xEC\xA6", "\xEC\xAE",
  /* U+9D20 */	"\xEC\xA5", "\x96\xB8", "\xEC\xAB", "\xC0\x6C",
  /* U+9D24 */	"\x00\x00", "\xEC\xA3", "\xC0\x6D", "\x00\x00",
  /* U+9D28 */	"\xC0\x6E", "\xEC\xA8", "\x00\x00", "\x00\x00",
  /* U+9D2C */	"\x00\x00", "\xEF\xA9", "\xEF\x7A", "\xEF\x7B",
  /* U+9D30 */	"\xEF\x7E", "\xEF\x7C", "\x00\x00", "\xEF\x76",
  /* U+9D34 */	"\xFA\xA1", "\x00\x00", "\xEF\x79", "\xEF\xA5",
  /* U+9D38 */	"\xEF\x7D", "\x91\xA7", "\x00\x00", "\xC2\x45",
  /* U+9D3C */	"\x00\x00", "\xEF\xA7", "\xEF\xA4", "\xC2\x46",
  /* U+9D40 */	"\xEF\xA6", "\xEF\x77", "\xEF\xA2", "\xEF\xA3",
  /* U+9D44 */	"\xA0\x5E", "\xEF\xA1", "\x00\x00", "\x00\x00",
  /* U+9D48 */	"\x00\x00", "\x9A\x7D", "\xF1\xD2", "\xF1\xD4",
  /* U+9D4C */	"\xF1\xD7", "\x00\x00", "\x89\x48", "\xF1\xD1",
  /* U+9D50 */	"\x9E\xB1", "\xC3\x59", "\xF1\xD9", "\xF1\xD0",
  /* U+9D54 */	"\xF1\xDA", "\x00\x00", "\xF1\xD6", "\xF1\xD8",
  /* U+9D58 */	"\xF1\xDC", "\xF1\xD5", "\xF1\xDD", "\xF1\xD3",
  /* U+9D5C */	"\xF1\xCF", "\xC3\x5A", "\x9D\xDB", "\xF1\xDB",
  /* U+9D60 */	"\xC3\x5B", "\xC4\x4D", "\x00\x00", "\x00\x00",
  /* U+9D64 */	"\x00\x00", "\x00\x00", "\x00\x00", "\xEF\x78",
  /* U+9D68 */	"\xF3\xF1", "\xF3\xE8", "\xC4\x4F", "\xF3\xE4",
  /* U+9D6C */	"\xC4\x50", "\x95\xBF", "\x8A\x73", "\xF3\xED",
  /* U+9D70 */	"\xF3\xE7", "\xF3\xDD", "\xC4\x4E", "\xF3\xEA",
  /* U+9D74 */	"\xF3\xE5", "\xF3\xE6", "\x00\x00", "\xF3\xD8",
  /* U+9D78 */	"\xF3\xDF", "\xF3\xEE", "\x00\x00", "\xF3\xEB",
  /* U+9D7C */	"\x9E\xFE", "\xF3\xE3", "\x91\x7A", "\xF3\xEF",
  /* U+9D80 */	"\xF3\xDE", "\xF3\xD9", "\xF3\xEC", "\x91\x7B",
  /* U+9D84 */	"\xF3\xDB", "\xF3\xE9", "\xF3\xE0", "\xF3\xF0",
  /* U+9D88 */	"\xF3\xDC", "\xC4\x4C", "\xF3\xDA", "\xF3\xE1",
  /* U+9D8C */	"\xF3\xE2", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+9D90 */	"\xF5\x7D", "\x00\x00", "\xF5\x7B", "\x9A\xA3",
  /* U+9D94 */	"\xF5\xA2", "\x00\x00", "\xF5\xAE", "\xF5\xA5",
  /* U+9D98 */	"\xF5\x7C", "\xF5\x78", "\xF5\xA7", "\xF5\x7E",
  /* U+9D9C */	"\xF5\xA3", "\xF5\x7A", "\xF5\xAA", "\xF5\x77",
  /* U+9DA0 */	"\xF5\xA1", "\xF5\xA6", "\xF5\xA8", "\xF5\xAB",
  /* U+9DA4 */	"\xF5\x79", "\x96\xC2", "\xF5\xAF", "\xF5\xB0",
  /* U+9DA8 */	"\xF5\xA9", "\xF5\xAD", "\xF5\xA4", "\x9F\x77",
  /* U+9DAC */	"\xF6\xC1", "\xF6\xC4", "\x00\x00", "\xC5\x61",
  /* U+9DB0 */	"\x00\x00", "\xF6\xC3", "\xF6\xC8", "\xF6\xC6",
  /* U+9DB4 */	"\xC5\x62", "\xF6\xBD", "\xF6\xB3", "\xF6\xB2",
  /* U+9DB8 */	"\xC5\x64", "\xF6\xBF", "\xF6\xC0", "\xF6\xBC",
  /* U+9DBC */	"\xF6\xB4", "\x9A\xA4", "\xF6\xB9", "\xF5\xAC",
  /* U+9DC0 */	"\x9A\xA5", "\xF6\xB5", "\xC5\x63", "\xF6\xBB",
  /* U+9DC4 */	"\x91\xA1", "\xF6\xBA", "\x00\x00", "\xF6\xB6",
  /* U+9DC8 */	"\xF6\xC2", "\x89\xB8", "\xF6\xB7", "\xF7\xBB",
  /* U+9DCC */	"\xF6\xC5", "\xF6\xC7", "\xF6\xBE", "\xF6\xB8",
  /* U+9DD0 */	"\xF7\xBC", "\xF7\xBE", "\xF7\xB8", "\xC5\xC2",
  /* U+9DD4 */	"\x91\x73", "\xF7\xC5", "\xF7\xC3", "\xC5\xC3",
  /* U+9DD8 */	"\xF7\xC2", "\xF7\xC1", "\xF7\xBA", "\xF7\xB7",
  /* U+9DDC */	"\xF7\xBD", "\xF7\xC6", "\xF7\xB9", "\xF7\xBF",
  /* U+9DE0 */	"\x00\x00", "\xF8\x69", "\xF8\x6E", "\xF8\x64",
  /* U+9DE4 */	"\xF8\x67", "\xC5\xEE", "\xF8\x6B", "\x00\x00",
  /* U+9DE8 */	"\xF8\x72", "\xF7\xC0", "\x00\x00", "\xF8\x65",
  /* U+9DEC */	"\xF8\x6F", "\xF8\x73", "\xF8\x6A", "\xF8\x63",
  /* U+9DF0 */	"\xF8\x6D", "\x00\x00", "\xF8\x6C", "\xF8\x71",
  /* U+9DF4 */	"\xF8\x70", "\xF7\xC4", "\xF8\x68", "\xF8\x62",
  /* U+9DF8 */	"\xF8\x66", "\xC6\x4E", "\xC6\x4F", "\xF8\x61",
  /* U+9DFC */	"\x9A\xA6", "\xF8\xE6", "\xF8\xDD", "\xF8\xE5",
  /* U+9E00 */	"\xF8\xE2", "\xF8\xE3", "\xF8\xDC", "\xF8\xDF",
  /* U+9E04 */	"\xF8\xE7", "\xF8\xE1", "\xF8\xE0", "\xF8\xDE",
  /* U+9E08 */	"\x00\x00", "\xF8\xE4", "\x89\xBD", "\xF9\x5D",
  /* U+9E0C */	"\x89\xB9", "\xF9\x5E", "\x91\x7D", "\xF9\x60",
  /* U+9E10 */	"\xF9\x5F", "\xF9\x62", "\xF9\x61", "\xF9\x7C",
  /* U+9E14 */	"\xF9\x7B", "\xF9\xB7", "\x00\x00", "\xF9\xB8",
  /* U+9E18 */	"\x96\xBB", "\xF9\xC5", "\xC6\x78", "\xC6\x7C",
  /* U+9E1C */	"\x9F\xF2", "\xF9\xCF", "\xC6\x7D", "\x8B\xF5",
  /* U+9E20 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+9E24 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+9E28 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+9E2C */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+9E30 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+9E34 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+9E38 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+9E3C */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+9E40 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+9E44 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+9E48 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+9E4C */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+9E50 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+9E54 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+9E58 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+9E5C */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+9E60 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+9E64 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+9E68 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+9E6C */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+9E70 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+9E74 */	"\x00\x00", "\xB3\xBF", "\x00\x00", "\x00\x00",
  /* U+9E78 */	"\x00\x00", "\xC4\xD0", "\xF6\xC9", "\x9A\xA9",
  /* U+9E7C */	"\xC6\x50", "\xC6\x51", "\x00\x00", "\xB3\xC0",
  /* U+9E80 */	"\xE0\xEE", "\x9F\x54", "\xB9\xA8", "\xE8\xF0",
  /* U+9E84 */	"\x9F\xE3", "\x9E\xED", "\xEC\xB0", "\xEC\xB1",
  /* U+9E88 */	"\xEC\xAF", "\xEF\xAB", "\xEF\xAA", "\xC2\x47",
  /* U+9E8C */	"\xF1\xDF", "\xEF\xAC", "\xF1\xDE", "\x00\x00",
  /* U+9E90 */	"\x91\xAA", "\xF3\xF3", "\xC4\x51", "\xC4\x53",
  /* U+9E94 */	"\xF3\xF2", "\x91\xAB", "\xA0\x70", "\xC4\x52",
  /* U+9E98 */	"\x9F\x6D", "\xF5\xB1", "\xF5\xB3", "\xF5\xB2",
  /* U+9E9C */	"\xF6\xCA", "\xC5\x65", "\x91\xAC", "\xC5\xEF",
  /* U+9EA0 */	"\xF8\xE8", "\xF9\x63", "\x91\xAD", "\x00\x00",
  /* U+9EA4 */	"\xF9\xD2", "\xB3\xC1", "\xA0\xFD", "\xE4\xE5",
  /* U+9EA8 */	"\x9F\xE2", "\xBE\xA2", "\x91\xAF", "\x9E\x41",
  /* U+9EAC */	"\x9A\xAA", "\xEC\xB3", "\xEC\xB2", "\x91\xB0",
  /* U+9EB0 */	"\xEF\xAD", "\x9A\xAB", "\x00\x00", "\x00\x00",
  /* U+9EB4 */	"\xC4\x54", "\xC4\xD1", "\xF7\xC7", "\xF9\xCB",
  /* U+9EB8 */	"\x00\x00", "\x00\x00", "\x00\x00", "\xB3\xC2",
  /* U+9EBC */	"\xBB\xF2", "\x9A\xAC", "\xBE\xA3", "\x9A\x4A",
  /* U+9EC0 */	"\xF3\xF4", "\x91\xB2", "\xF8\x74", "\xB6\xC0",
  /* U+9EC4 */	"\x8B\xF6", "\x00\x00", "\x9A\xAD", "\x89\xB6",
  /* U+9EC8 */	"\xEF\xAE", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+9ECC */	"\xC6\x64", "\xB6\xC1", "\xBE\xA4", "\xC2\x48",
  /* U+9ED0 */	"\xF8\x75", "\xB6\xC2", "\x00\x00", "\xE8\xF1",
  /* U+9ED4 */	"\xC0\x72", "\xEC\xB4", "\xEC\xB5", "\x00\x00",
  /* U+9ED8 */	"\xC0\x71", "\x00\x00", "\xEF\xAF", "\xC2\x4C",
  /* U+9EDC */	"\xC2\x4A", "\xC2\x4B", "\xC2\x49", "\xF1\xE0",
  /* U+9EE0 */	"\xC3\x5C", "\x00\x00", "\x9A\xAF", "\x00\x00",
  /* U+9EE4 */	"\xF5\xB5", "\xF5\xB4", "\xF5\xB7", "\xF5\xB6",
  /* U+9EE8 */	"\xC4\xD2", "\x00\x00", "\x00\x00", "\xF6\xCB",
  /* U+9EEC */	"\x00\x00", "\xF6\xCD", "\xF6\xCC", "\xC5\x66",
  /* U+9EF0 */	"\xF7\xC8", "\x9A\xB0", "\xF8\x76", "\xF8\x77",
  /* U+9EF4 */	"\xC5\xF0", "\xF9\x64", "\xF9\x7D", "\xC6\x75",
  /* U+9EF8 */	"\x9A\xB1", "\xDC\xB0", "\xEC\xB6", "\xEF\xB0",
  /* U+9EFC */	"\xF3\xF5", "\xE0\xEF", "\x9A\xA1", "\xEF\xB1",
  /* U+9F00 */	"\xF1\xE2", "\xF1\xE1", "\x91\xB9", "\x00\x00",
  /* U+9F04 */	"\x00\x00", "\x00\x00", "\xF8\x78", "\xC6\x52",
  /* U+9F08 */	"\x91\xBA", "\xF9\x65", "\xF9\x7E", "\x00\x00",
  /* U+9F0C */	"\x00\x00", "\x00\x00", "\xB9\xA9", "\xE8\xF2",
  /* U+9F10 */	"\xE8\xF3", "\x00\x00", "\xEC\xB7", "\xB9\xAA",
  /* U+9F14 */	"\x00\x00", "\xC3\x5D", "\xF1\xE3", "\x91\xBE",
  /* U+9F18 */	"\xF6\xCF", "\xC5\x67", "\xF6\xD0", "\xF6\xCE",
  /* U+9F1C */	"\xF8\x79", "\x00\x00", "\xF8\xE9", "\x00\x00",
  /* U+9F20 */	"\xB9\xAB", "\x00\x00", "\xEF\xB4", "\xEF\xB3",
  /* U+9F24 */	"\xEF\xB2", "\xF1\xE4", "\xA0\x41", "\x8B\xB7",
  /* U+9F28 */	"\xF1\xE8", "\xF1\xE7", "\xF1\xE6", "\xF1\xE5",
  /* U+9F2C */	"\xC3\x5E", "\xF3\xF6", "\xF5\xB9", "\xC4\xD3",
  /* U+9F30 */	"\xF5\xB8", "\xF6\xD1", "\xF7\xCB", "\xF7\xCA",
  /* U+9F34 */	"\xC5\xC4", "\xF7\xC9", "\xF8\x7C", "\xF8\x7B",
  /* U+9F38 */	"\xF8\x7A", "\x91\xC0", "\x00\x00", "\xBB\xF3",
  /* U+9F3C */	"\x00\x00", "\xEC\xB8", "\xC2\x4D", "\x00\x00",
  /* U+9F40 */	"\xF3\xF7", "\xF3\xF8", "\xF7\xCC", "\xF8\x7D",
  /* U+9F44 */	"\x9A\xB3", "\x91\xC3", "\xF8\xEA", "\xF9\x66",
  /* U+9F48 */	"\xF9\xB9", "\xF9\xD4", "\xBB\xF4", "\xC2\x4E",
  /* U+9F4C */	"\xF1\xE9", "\xF3\xF9", "\xF6\xD2", "\xF8\x7E",
  /* U+9F50 */	"\xA0\xFC", "\x00\x00", "\xBE\xA6", "\x9F\xEE",
  /* U+9F54 */	"\xEF\xB5", "\xF1\xEA", "\xF3\xFA", "\xF3\xFB",
  /* U+9F58 */	"\xF3\xFC", "\xF5\xBE", "\x9F\x69", "\xF5\xBA",
  /* U+9F5C */	"\xC5\x68", "\xF5\xBD", "\xF5\xBC", "\xC4\xD4",
  /* U+9F60 */	"\xF5\xBB", "\xC4\xD6", "\x91\xC8", "\xC4\xD5",
  /* U+9F64 */	"\xF6\xD4", "\xF6\xD3", "\xC5\x69", "\xC5\x6A",
  /* U+9F68 */	"\x00\x00", "\x91\xC9", "\xC5\xC6", "\xF7\xCD",
  /* U+9F6C */	"\xC5\xC5", "\x00\x00", "\xF8\xA3", "\xF8\xA4",
  /* U+9F70 */	"\xF8\xA2", "\xF8\xA1", "\xC6\x54", "\x00\x00",
  /* U+9F74 */	"\xF8\xEB", "\xF8\xEC", "\xF8\xED", "\xC6\x53",
  /* U+9F78 */	"\xF9\x67", "\xF9\x6A", "\xF9\x69", "\xF9\x68",
  /* U+9F7C */	"\x00\x00", "\x00\x00", "\xF9\xD3", "\x8D\xE6",
  /* U+9F80 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+9F84 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+9F88 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+9F8C */	"\x00\x00", "\xC0\x73", "\x91\xCB", "\x00\x00",
  /* U+9F90 */	"\xC3\x65", "\xF5\xBF", "\xF6\xD5", "\x00\x00",
  /* U+9F94 */	"\xC5\xC7", "\xF7\xCE", "\x00\x00", "\x00\x00",
  /* U+9F98 */	"\xF9\xD5", "\x89\xC8", "\x00\x00", "\x00\x00",
  /* U+9F9C */	"\xC0\x74", "\x00\x00", "\x00\x00", "\x8D\xAA",
  /* U+9FA0 */	"\xEF\xB6", "\x00\x00", "\xF7\xCF", "\x00\x00",
  /* U+9FA4 */	"\xF9\xA1", "\x9F\xDD",

  /* U+E000 */	"\xFA\x40", "\xFA\x41", "\xFA\x42", "\xFA\x43",
  /* U+E004 */	"\xFA\x44", "\xFA\x45", "\xFA\x46", "\xFA\x47",
  /* U+E008 */	"\xFA\x48", "\xFA\x49", "\xFA\x4A", "\xFA\x4B",
  /* U+E00C */	"\xFA\x4C", "\xFA\x4D", "\xFA\x4E", "\xFA\x4F",
  /* U+E010 */	"\xFA\x50", "\xFA\x51", "\xFA\x52", "\xFA\x53",
  /* U+E014 */	"\xFA\x54", "\xFA\x55", "\xFA\x56", "\xFA\x57",
  /* U+E018 */	"\xFA\x58", "\xFA\x59", "\xFA\x5A", "\xFA\x5B",
  /* U+E01C */	"\xFA\x5C", "\xFA\x5D", "\xFA\x5E", "\xAD\xC5",
  /* U+E020 */	"\xFA\x60", "\xFA\x61", "\xFA\x62", "\xFA\x63",
  /* U+E024 */	"\xFA\x64", "\xFA\x65", "\xB0\xB0", "\xFA\x67",
  /* U+E028 */	"\xFA\x68", "\xFA\x69", "\xFA\x6A", "\xFA\x6B",
  /* U+E02C */	"\xFA\x6C", "\xFA\x6D", "\xFA\x6E", "\xFA\x6F",
  /* U+E030 */	"\xFA\x70", "\xFA\x71", "\xFA\x72", "\xFA\x73",
  /* U+E034 */	"\xFA\x74", "\xFA\x75", "\xFA\x76", "\xFA\x77",
  /* U+E038 */	"\xFA\x78", "\xFA\x79", "\xFA\x7A", "\xFA\x7B",
  /* U+E03C */	"\xFA\x7C", "\xFA\x7D", "\xFA\x7E", "\xFA\xA1",
  /* U+E040 */	"\xFA\xA2", "\xFA\xA3", "\xFA\xA4", "\xFA\xA5",
  /* U+E044 */	"\xFA\xA6", "\xFA\xA7", "\xFA\xA8", "\xFA\xA9",
  /* U+E048 */	"\xFA\xAA", "\xFA\xAB", "\xFA\xAC", "\xFA\xAD",
  /* U+E04C */	"\xFA\xAE", "\xFA\xAF", "\xFA\xB0", "\xFA\xB1",
  /* U+E050 */	"\xFA\xB2", "\xFA\xB3", "\xFA\xB4", "\xFA\xB5",
  /* U+E054 */	"\xFA\xB6", "\xFA\xB7", "\xFA\xB8", "\xFA\xB9",
  /* U+E058 */	"\xFA\xBA", "\xFA\xBB", "\xFA\xBC", "\xA5\x5D",
  /* U+E05C */	"\xFA\xBE", "\xFA\xBF", "\xFA\xC0", "\xFA\xC1",
  /* U+E060 */	"\xFA\xC2", "\xFA\xC3", "\xFA\xC4", "\xA2\xCD",
  /* U+E064 */	"\xFA\xC6", "\xFA\xC7", "\xFA\xC8", "\xFA\xC9",
  /* U+E068 */	"\xFA\xCA", "\xFA\xCB", "\xFA\xCC", "\xFA\xCD",
  /* U+E06C */	"\xFA\xCE", "\xFA\xCF", "\xFA\xD0", "\xFA\xD1",
  /* U+E070 */	"\xFA\xD2", "\xFA\xD3", "\xFA\xD4", "\xAD\xEB",
  /* U+E074 */	"\xFA\xD6", "\xFA\xD7", "\xFA\xD8", "\xFA\xD9",
  /* U+E078 */	"\xFA\xDA", "\xFA\xDB", "\xFA\xDC", "\xFA\xDD",
  /* U+E07C */	"\xFA\xDE", "\xFA\xDF", "\xFA\xE0", "\xFA\xE1",
  /* U+E080 */	"\xFA\xE2", "\xFA\xE3", "\xFA\xE4", "\xFA\xE5",
  /* U+E084 */	"\xFA\xE6", "\xFA\xE7", "\xFA\xE8", "\xFA\xE9",
  /* U+E088 */	"\xFA\xEA", "\xFA\xEB", "\xFA\xEC", "\xFA\xED",
  /* U+E08C */	"\xFA\xEE", "\xFA\xEF", "\xFA\xF0", "\xFA\xF1",
  /* U+E090 */	"\xFA\xF2", "\xFA\xF3", "\xFA\xF4", "\xFA\xF5",
  /* U+E094 */	"\xFA\xF6", "\xFA\xF7", "\xFA\xF8", "\xFA\xF9",
  /* U+E098 */	"\xFA\xFA", "\xFA\xFB", "\xFA\xFC", "\xFA\xFD",
  /* U+E09C */	"\xFA\xFE", "\xFB\x40", "\xFB\x41", "\xFB\x42",
  /* U+E0A0 */	"\xFB\x43", "\xFB\x44", "\xFB\x45", "\xFB\x46",
  /* U+E0A4 */	"\xFB\x47", "\x9D\xEF", "\xFB\x49", "\xFB\x4A",
  /* U+E0A8 */	"\xFB\x4B", "\xFB\x4C", "\xFB\x4D", "\xFB\x4E",
  /* U+E0AC */	"\xFB\x4F", "\xFB\x50", "\xFB\x51", "\xFB\x52",
  /* U+E0B0 */	"\xFB\x53", "\xFB\x54", "\xFB\x55", "\xFB\x56",
  /* U+E0B4 */	"\xFB\x57", "\xFB\x58", "\xFB\x59", "\xFB\x5A",
  /* U+E0B8 */	"\xFB\x5B", "\xFB\x5C", "\xFB\x5D", "\xFB\x5E",
  /* U+E0BC */	"\xFB\x5F", "\xFB\x60", "\xFB\x61", "\xFB\x62",
  /* U+E0C0 */	"\xFB\x63", "\xFB\x64", "\xFB\x65", "\xFB\x66",
  /* U+E0C4 */	"\xFB\x67", "\xFB\x68", "\xFB\x69", "\xFB\x6A",
  /* U+E0C8 */	"\xFB\x6B", "\xFB\x6C", "\xFB\x6D", "\xFB\x6E",
  /* U+E0CC */	"\xFB\x6F", "\xFB\x70", "\xFB\x71", "\xFB\x72",
  /* U+E0D0 */	"\xFB\x73", "\xFB\x74", "\xFB\x75", "\xFB\x76",
  /* U+E0D4 */	"\xFB\x77", "\xFB\x78", "\xFB\x79", "\xFB\x7A",
  /* U+E0D8 */	"\xFB\x7B", "\xFB\x7C", "\xFB\x7D", "\xFB\x7E",
  /* U+E0DC */	"\xFB\xA1", "\xFB\xA2", "\xFB\xA3", "\xFB\xA4",
  /* U+E0E0 */	"\xFB\xA5", "\xFB\xA6", "\xFB\xA7", "\xFB\xA8",
  /* U+E0E4 */	"\xFB\xA9", "\xFB\xAA", "\xFB\xAB", "\xFB\xAC",
  /* U+E0E8 */	"\xFB\xAD", "\xFB\xAE", "\xFB\xAF", "\xFB\xB0",
  /* U+E0EC */	"\xFB\xB1", "\xFB\xB2", "\xFB\xB3", "\xFB\xB4",
  /* U+E0F0 */	"\xFB\xB5", "\xFB\xB6", "\xFB\xB7", "\xB4\x40",
  /* U+E0F4 */	"\xFB\xB9", "\xFB\xBA", "\xFB\xBB", "\xFB\xBC",
  /* U+E0F8 */	"\xFB\xBD", "\xFB\xBE", "\xFB\xBF", "\xFB\xC0",
  /* U+E0FC */	"\xFB\xC1", "\xFB\xC2", "\xFB\xC3", "\xFB\xC4",
  /* U+E100 */	"\xFB\xC5", "\xFB\xC6", "\xFB\xC7", "\xFB\xC8",
  /* U+E104 */	"\xFB\xC9", "\xFB\xCA", "\xFB\xCB", "\xFB\xCC",
  /* U+E108 */	"\xFB\xCD", "\xFB\xCE", "\xFB\xCF", "\xFB\xD0",
  /* U+E10C */	"\xFB\xD1", "\xFB\xD2", "\xFB\xD3", "\xFB\xD4",
  /* U+E110 */	"\xFB\xD5", "\xFB\xD6", "\xFB\xD7", "\xFB\xD8",
  /* U+E114 */	"\xFB\xD9", "\xFB\xDA", "\xFB\xDB", "\xFB\xDC",
  /* U+E118 */	"\xFB\xDD", "\xFB\xDE", "\xFB\xDF", "\xFB\xE0",
  /* U+E11C */	"\xFB\xE1", "\xFB\xE2", "\xFB\xE3", "\xFB\xE4",
  /* U+E120 */	"\xFB\xE5", "\xFB\xE6", "\xFB\xE7", "\xFB\xE8",
  /* U+E124 */	"\xFB\xE9", "\xFB\xEA", "\xFB\xEB", "\xFB\xEC",
  /* U+E128 */	"\xFB\xED", "\xFB\xEE", "\xFB\xEF", "\xFB\xF0",
  /* U+E12C */	"\xFB\xF1", "\xFB\xF2", "\xC9\xDB", "\xFB\xF4",
  /* U+E130 */	"\xFB\xF5", "\xFB\xF6", "\xFB\xF7", "\xFB\xF8",
  /* U+E134 */	"\x9D\xFB", "\xFB\xFA", "\xFB\xFB", "\xFB\xFC",
  /* U+E138 */	"\xFB\xFD", "\xFB\xFE", "\xFC\x40", "\xFC\x41",
  /* U+E13C */	"\xFC\x42", "\xFC\x43", "\xFC\x44", "\xFC\x45",
  /* U+E140 */	"\xFC\x46", "\xFC\x47", "\xFC\x48", "\xFC\x49",
  /* U+E144 */	"\xFC\x4A", "\xFC\x4B", "\xFC\x4C", "\xFC\x4D",
  /* U+E148 */	"\xFC\x4E", "\xD8\xF4", "\xFC\x50", "\xFC\x51",
  /* U+E14C */	"\xFC\x52", "\xFC\x53", "\xFC\x54", "\xFC\x55",
  /* U+E150 */	"\xFC\x56", "\xFC\x57", "\xFC\x58", "\xFC\x59",
  /* U+E154 */	"\xFC\x5A", "\xFC\x5B", "\xFC\x5C", "\xFC\x5D",
  /* U+E158 */	"\xFC\x5E", "\xFC\x5F", "\xFC\x60", "\xFC\x61",
  /* U+E15C */	"\xFC\x62", "\xFC\x63", "\xFC\x64", "\xFC\x65",
  /* U+E160 */	"\xFC\x66", "\xFC\x67", "\xFC\x68", "\xFC\x69",
  /* U+E164 */	"\xFC\x6A", "\xFC\x6B", "\xA0\xDC", "\xFC\x6D",
  /* U+E168 */	"\xFC\x6E", "\xFC\x6F", "\xFC\x70", "\xFC\x71",
  /* U+E16C */	"\xFC\x72", "\xFC\x73", "\xFC\x74", "\xFC\x75",
  /* U+E170 */	"\xFC\x76", "\xFC\x77", "\xFC\x78", "\xFC\x79",
  /* U+E174 */	"\xFC\x7A", "\xFC\x7B", "\xFC\x7C", "\xFC\x7D",
  /* U+E178 */	"\xFC\x7E", "\xFC\xA1", "\xFC\xA2", "\xFC\xA3",
  /* U+E17C */	"\xFC\xA4", "\xFC\xA5", "\xFC\xA6", "\xFC\xA7",
  /* U+E180 */	"\xFC\xA8", "\xFC\xA9", "\xFC\xAA", "\xFC\xAB",
  /* U+E184 */	"\xFC\xAC", "\xFC\xAD", "\xFC\xAE", "\xFC\xAF",
  /* U+E188 */	"\xFC\xB0", "\xFC\xB1", "\xFC\xB2", "\xFC\xB3",
  /* U+E18C */	"\xFC\xB4", "\xFC\xB5", "\xFC\xB6", "\xFC\xB7",
  /* U+E190 */	"\xFC\xB8", "\xBC\xB5", "\xFC\xBA", "\xFC\xBB",
  /* U+E194 */	"\xFC\xBC", "\xFC\xBD", "\xFC\xBE", "\xFC\xBF",
  /* U+E198 */	"\xFC\xC0", "\xFC\xC1", "\xFC\xC2", "\xFC\xC3",
  /* U+E19C */	"\xFC\xC4", "\xFC\xC5", "\xFC\xC6", "\xFC\xC7",
  /* U+E1A0 */	"\xFC\xC8", "\xFC\xC9", "\xFC\xCA", "\xFC\xCB",
  /* U+E1A4 */	"\xFC\xCC", "\xFC\xCD", "\xFC\xCE", "\xFC\xCF",
  /* U+E1A8 */	"\xFC\xD0", "\xFC\xD1", "\xFC\xD2", "\xFC\xD3",
  /* U+E1AC */	"\xFC\xD4", "\xFC\xD5", "\xFC\xD6", "\xFC\xD7",
  /* U+E1B0 */	"\xFC\xD8", "\xFC\xD9", "\xFC\xDA", "\xFC\xDB",
  /* U+E1B4 */	"\xFC\xDC", "\xFC\xDD", "\xFC\xDE", "\xFC\xDF",
  /* U+E1B8 */	"\xFC\xE0", "\xFC\xE1", "\xB4\xB8", "\xFC\xE3",
  /* U+E1BC */	"\xFC\xE4", "\xFC\xE5", "\xFC\xE6", "\xFC\xE7",
  /* U+E1C0 */	"\xFC\xE8", "\xFC\xE9", "\xFC\xEA", "\xFC\xEB",
  /* U+E1C4 */	"\xFC\xEC", "\xFC\xED", "\xFC\xEE", "\xFC\xEF",
  /* U+E1C8 */	"\xFC\xF0", "\xA7\xFB", "\xFC\xF2", "\xFC\xF3",
  /* U+E1CC */	"\xFC\xF4", "\xFC\xF5", "\xFC\xF6", "\xFC\xF7",
  /* U+E1D0 */	"\xFC\xF8", "\xFC\xF9", "\xFC\xFA", "\xFC\xFB",
  /* U+E1D4 */	"\xFC\xFC", "\xFC\xFD", "\xFC\xFE", "\xFD\x40",
  /* U+E1D8 */	"\xFD\x41", "\xFD\x42", "\xFD\x43", "\xFD\x44",
  /* U+E1DC */	"\xFD\x45", "\xFD\x46", "\xFD\x47", "\xFD\x48",
  /* U+E1E0 */	"\xFD\x49", "\xFD\x4A", "\xFD\x4B", "\xFD\x4C",
  /* U+E1E4 */	"\xFD\x4D", "\xFD\x4E", "\xFD\x4F", "\xFD\x50",
  /* U+E1E8 */	"\xFD\x51", "\xFD\x52", "\xFD\x53", "\xFD\x54",
  /* U+E1EC */	"\xFD\x55", "\xFD\x56", "\xFD\x57", "\xFD\x58",
  /* U+E1F0 */	"\xFD\x59", "\xFD\x5A", "\xFD\x5B", "\xFD\x5C",
  /* U+E1F4 */	"\xFD\x5D", "\xFD\x5E", "\xFD\x5F", "\xFD\x60",
  /* U+E1F8 */	"\xFD\x61", "\xFD\x62", "\xFD\x63", "\xFD\x64",
  /* U+E1FC */	"\xFD\x65", "\xFD\x66", "\xFD\x67", "\xFD\x68",
  /* U+E200 */	"\xFD\x69", "\xFD\x6A", "\xFD\x6B", "\xFD\x6C",
  /* U+E204 */	"\xFD\x6D", "\xFD\x6E", "\xFD\x6F", "\xFD\x70",
  /* U+E208 */	"\xFD\x71", "\xFD\x72", "\xFD\x73", "\xFD\x74",
  /* U+E20C */	"\xFD\x75", "\xFD\x76", "\xFD\x77", "\xFD\x78",
  /* U+E210 */	"\xFD\x79", "\xFD\x7A", "\xFD\x7B", "\xFD\x7C",
  /* U+E214 */	"\xFD\x7D", "\xFD\x7E", "\xFD\xA1", "\xFD\xA2",
  /* U+E218 */	"\xFD\xA3", "\xFD\xA4", "\xFD\xA5", "\xFD\xA6",
  /* U+E21C */	"\xFD\xA7", "\xFD\xA8", "\xFD\xA9", "\xFD\xAA",
  /* U+E220 */	"\xFD\xAB", "\xFD\xAC", "\xFD\xAD", "\xFD\xAE",
  /* U+E224 */	"\xFD\xAF", "\xFD\xB0", "\xFD\xB1", "\xFD\xB2",
  /* U+E228 */	"\xFD\xB3", "\xFD\xB4", "\xFD\xB5", "\xFD\xB6",
  /* U+E22C */	"\xCB\x58", "\xB4\xFC", "\xFD\xB9", "\xFD\xBA",
  /* U+E230 */	"\xB4\xE4", "\xFD\xBC", "\xFD\xBD", "\xFD\xBE",
  /* U+E234 */	"\xFD\xBF", "\xFD\xC0", "\xFD\xC1", "\xFD\xC2",
  /* U+E238 */	"\xFD\xC3", "\xFD\xC4", "\xFD\xC5", "\xFD\xC6",
  /* U+E23C */	"\xFD\xC7", "\xFD\xC8", "\xFD\xC9", "\xFD\xCA",
  /* U+E240 */	"\xFD\xCB", "\xFD\xCC", "\xFD\xCD", "\xFD\xCE",
  /* U+E244 */	"\xFD\xCF", "\xFD\xD0", "\xFD\xD1", "\xFD\xD2",
  /* U+E248 */	"\xFD\xD3", "\xFD\xD4", "\xFD\xD5", "\xFD\xD6",
  /* U+E24C */	"\xFD\xD7", "\xFD\xD8", "\xFD\xD9", "\xFD\xDA",
  /* U+E250 */	"\xFD\xDB", "\xFD\xDC", "\xFD\xDD", "\xFD\xDE",
  /* U+E254 */	"\xFD\xDF", "\xFD\xE0", "\xFD\xE1", "\xFD\xE2",
  /* U+E258 */	"\xFD\xE3", "\xFD\xE4", "\xFD\xE5", "\xFD\xE6",
  /* U+E25C */	"\xFD\xE7", "\xFD\xE8", "\xFD\xE9", "\xFD\xEA",
  /* U+E260 */	"\xFD\xEB", "\xFD\xEC", "\xFD\xED", "\xFD\xEE",
  /* U+E264 */	"\xFD\xEF", "\xFD\xF0", "\xB5\x4E", "\xFD\xF2",
  /* U+E268 */	"\xFD\xF3", "\xFD\xF4", "\xFD\xF5", "\xFD\xF6",
  /* U+E26C */	"\xFD\xF7", "\xFD\xF8", "\xFD\xF9", "\xFD\xFA",
  /* U+E270 */	"\xFD\xFB", "\xFD\xFC", "\xFD\xFD", "\xFD\xFE",
  /* U+E274 */	"\xFE\x40", "\xFE\x41", "\xFE\x42", "\xFE\x43",
  /* U+E278 */	"\xFE\x44", "\xFE\x45", "\xFE\x46", "\xFE\x47",
  /* U+E27C */	"\xFE\x48", "\xFE\x49", "\xFE\x4A", "\xFE\x4B",
  /* U+E280 */	"\xFE\x4C", "\xFE\x4D", "\xFE\x4E", "\xFE\x4F",
  /* U+E284 */	"\xFE\x50", "\xFE\x51", "\x99\x75", "\xFE\x53",
  /* U+E288 */	"\xFE\x54", "\xFE\x55", "\xFE\x56", "\xFE\x57",
  /* U+E28C */	"\xFE\x58", "\xFE\x59", "\xFE\x5A", "\xFE\x5B",
  /* U+E290 */	"\xFE\x5C", "\xFE\x5D", "\xFE\x5E", "\xFE\x5F",
  /* U+E294 */	"\xFE\x60", "\xFE\x61", "\xFE\x62", "\xFE\x63",
  /* U+E298 */	"\xFE\x64", "\xFE\x65", "\xFE\x66", "\xFE\x67",
  /* U+E29C */	"\xFE\x68", "\xFE\x69", "\xFE\x6A", "\xFE\x6B",
  /* U+E2A0 */	"\xFE\x6C", "\xFE\x6D", "\xFE\x6E", "\xB7\xEC",
  /* U+E2A4 */	"\xFE\x70", "\xFE\x71", "\xFE\x72", "\xFE\x73",
  /* U+E2A8 */	"\xFE\x74", "\xFE\x75", "\xFE\x76", "\xFE\x77",
  /* U+E2AC */	"\xFE\x78", "\xFE\x79", "\xFE\x7A", "\xFE\x7B",
  /* U+E2B0 */	"\xFE\x7C", "\xFE\x7D", "\xFE\x7E", "\xFE\xA1",
  /* U+E2B4 */	"\xFE\xA2", "\xFE\xA3", "\xFE\xA4", "\xFE\xA5",
  /* U+E2B8 */	"\xFE\xA6", "\xFE\xA7", "\xFE\xA8", "\xFE\xA9",
  /* U+E2BC */	"\xA2\x60", "\xFE\xAB", "\xFE\xAC", "\xFE\xAD",
  /* U+E2C0 */	"\xFE\xAE", "\xFE\xAF", "\xFE\xB0", "\xFE\xB1",
  /* U+E2C4 */	"\xFE\xB2", "\xFE\xB3", "\xFE\xB4", "\xFE\xB5",
  /* U+E2C8 */	"\xFE\xB6", "\xFE\xB7", "\xFE\xB8", "\xFE\xB9",
  /* U+E2CC */	"\xFE\xBA", "\xFE\xBB", "\xFE\xBC", "\xFE\xBD",
  /* U+E2D0 */	"\xFE\xBE", "\xFE\xBF", "\xFE\xC0", "\xFE\xC1",
  /* U+E2D4 */	"\xFE\xC2", "\xFE\xC3", "\xFE\xC4", "\xFE\xC5",
  /* U+E2D8 */	"\xFE\xC6", "\xFE\xC7", "\xFE\xC8", "\xFE\xC9",
  /* U+E2DC */	"\xFE\xCA", "\xFE\xCB", "\xFE\xCC", "\xFE\xCD",
  /* U+E2E0 */	"\xFE\xCE", "\xFE\xCF", "\xFE\xD0", "\xFE\xD1",
  /* U+E2E4 */	"\xFE\xD2", "\xFE\xD3", "\xFE\xD4", "\xFE\xD5",
  /* U+E2E8 */	"\xFE\xD6", "\xFE\xD7", "\xFE\xD8", "\xFE\xD9",
  /* U+E2EC */	"\xFE\xDA", "\xFE\xDB", "\xFE\xDC", "\xCF\xF1",
  /* U+E2F0 */	"\xFE\xDE", "\xFE\xDF", "\xFE\xE0", "\xFE\xE1",
  /* U+E2F4 */	"\xFE\xE2", "\xFE\xE3", "\xFE\xE4", "\xFE\xE5",
  /* U+E2F8 */	"\xFE\xE6", "\xFE\xE7", "\xFE\xE8", "\xFE\xE9",
  /* U+E2FC */	"\xFE\xEA", "\xFE\xEB", "\xFE\xEC", "\xFE\xED",
  /* U+E300 */	"\xFE\xEE", "\xFE\xEF", "\xFE\xF0", "\xFE\xF1",
  /* U+E304 */	"\xFE\xF2", "\xFE\xF3", "\xFE\xF4", "\xFE\xF5",
  /* U+E308 */	"\xFE\xF6", "\xFE\xF7", "\xFE\xF8", "\xFE\xF9",
  /* U+E30C */	"\xFE\xFA", "\xFE\xFB", "\xFE\xFC", "\xFE\xFD",
  /* U+E310 */	"\xFE\xFE", "\x8E\x40", "\x8E\x41", "\x8E\x42",
  /* U+E314 */	"\x8E\x43", "\x8E\x44", "\x8E\x45", "\x8E\x46",
  /* U+E318 */	"\x8E\x47", "\x8E\x48", "\x8E\x49", "\x8E\x4A",
  /* U+E31C */	"\x8E\x4B", "\x8E\x4C", "\x8E\x4D", "\x8E\x4E",
  /* U+E320 */	"\x8E\x4F", "\x8E\x50", "\x8E\x51", "\x8E\x52",
  /* U+E324 */	"\x8E\x53", "\x8E\x54", "\x8E\x55", "\x8E\x56",
  /* U+E328 */	"\x8E\x57", "\x8E\x58", "\x8E\x59", "\x8E\x5A",
  /* U+E32C */	"\x8E\x5B", "\x8E\x5C", "\x8E\x5D", "\x8E\x5E",
  /* U+E330 */	"\x8E\x5F", "\x8E\x60", "\x8E\x61", "\x8E\x62",
  /* U+E334 */	"\x8E\x63", "\x8E\x64", "\x8E\x65", "\x8E\x66",
  /* U+E338 */	"\x8E\x67", "\x8E\x68", "\xBA\xE6", "\x8E\x6A",
  /* U+E33C */	"\x8E\x6B", "\x8E\x6C", "\x8E\x6D", "\x8E\x6E",
  /* U+E340 */	"\xED\xCA", "\x8E\x70", "\x8E\x71", "\x8E\x72",
  /* U+E344 */	"\x8E\x73", "\x8E\x74", "\x8E\x75", "\x8E\x76",
  /* U+E348 */	"\x8E\x77", "\x8E\x78", "\x8E\x79", "\x8E\x7A",
  /* U+E34C */	"\x8E\x7B", "\x8E\x7C", "\x8E\x7D", "\xA2\x61",
  /* U+E350 */	"\x8E\xA1", "\x8E\xA2", "\x8E\xA3", "\x8E\xA4",
  /* U+E354 */	"\x8E\xA5", "\x8E\xA6", "\x8E\xA7", "\x8E\xA8",
  /* U+E358 */	"\x8E\xA9", "\x8E\xAA", "\xBA\xFC", "\x8E\xAC",
  /* U+E35C */	"\x8E\xAD", "\x8E\xAE", "\x8E\xAF", "\x8E\xB0",
  /* U+E360 */	"\x8E\xB1", "\x8E\xB2", "\x8E\xB3", "\xBF\xA6",
  /* U+E364 */	"\x8E\xB5", "\x8E\xB6", "\x8E\xB7", "\x8E\xB8",
  /* U+E368 */	"\x8E\xB9", "\x8E\xBA", "\x8E\xBB", "\x8E\xBC",
  /* U+E36C */	"\x8E\xBD", "\x8E\xBE", "\x8E\xBF", "\x8E\xC0",
  /* U+E370 */	"\x8E\xC1", "\x8E\xC2", "\x8E\xC3", "\x8E\xC4",
  /* U+E374 */	"\x8E\xC5", "\x8E\xC6", "\x8E\xC7", "\x8E\xC8",
  /* U+E378 */	"\x8E\xC9", "\x8E\xCA", "\x8E\xCB", "\x8E\xCC",
  /* U+E37C */	"\xAA\xCC", "\x8E\xCE", "\x8E\xCF", "\xBF\xAE",
  /* U+E380 */	"\x8E\xD1", "\x8E\xD2", "\x8E\xD3", "\x8E\xD4",
  /* U+E384 */	"\x8E\xD5", "\x8E\xD6", "\x8E\xD7", "\x8E\xD8",
  /* U+E388 */	"\x8E\xD9", "\x8E\xDA", "\x8E\xDB", "\x8E\xDC",
  /* U+E38C */	"\x8E\xDD", "\x8E\xDE", "\x8E\xDF", "\x8E\xE0",
  /* U+E390 */	"\x8E\xE1", "\x8E\xE2", "\x8E\xE3", "\x8E\xE4",
  /* U+E394 */	"\x8E\xE5", "\x8E\xE6", "\x8E\xE7", "\x8E\xE8",
  /* U+E398 */	"\x8E\xE9", "\x8E\xEA", "\x8E\xEB", "\x8E\xEC",
  /* U+E39C */	"\x8E\xED", "\x8E\xEE", "\x8E\xEF", "\x8E\xF0",
  /* U+E3A0 */	"\x8E\xF1", "\x8E\xF2", "\x8E\xF3", "\x8E\xF4",
  /* U+E3A4 */	"\x8E\xF5", "\x8E\xF6", "\x8E\xF7", "\x8E\xF8",
  /* U+E3A8 */	"\x8E\xF9", "\x8E\xFA", "\x8E\xFB", "\x8E\xFC",
  /* U+E3AC */	"\x8E\xFD", "\x8E\xFE", "\x8F\x40", "\x8F\x41",
  /* U+E3B0 */	"\x8F\x42", "\x8F\x43", "\x8F\x44", "\x8F\x45",
  /* U+E3B4 */	"\x8F\x46", "\x8F\x47", "\x8F\x48", "\x8F\x49",
  /* U+E3B8 */	"\x8F\x4A", "\x8F\x4B", "\x8F\x4C", "\x8F\x4D",
  /* U+E3BC */	"\x8F\x4E", "\x8F\x4F", "\x8F\x50", "\x8F\x51",
  /* U+E3C0 */	"\x8F\x52", "\x8F\x53", "\x8F\x54", "\x8F\x55",
  /* U+E3C4 */	"\x8F\x56", "\xB5\xD7", "\x8F\x58", "\x8F\x59",
  /* U+E3C8 */	"\x8F\x5A", "\x8F\x5B", "\x8F\x5C", "\x8F\x5D",
  /* U+E3CC */	"\x8F\x5E", "\x8F\x5F", "\x8F\x60", "\x8F\x61",
  /* U+E3D0 */	"\x8F\x62", "\x8F\x63", "\x8F\x64", "\x8F\x65",
  /* U+E3D4 */	"\x8F\x66", "\x8F\x67", "\x8F\x68", "\xE3\xC8",
  /* U+E3D8 */	"\x8F\x6A", "\x8F\x6B", "\x8F\x6C", "\x8F\x6D",
  /* U+E3DC */	"\xDB\x79", "\x8F\x6F", "\x8F\x70", "\x8F\x71",
  /* U+E3E0 */	"\x8F\x72", "\x8F\x73", "\x8F\x74", "\x8F\x75",
  /* U+E3E4 */	"\x8F\x76", "\x8F\x77", "\x8F\x78", "\x8F\x79",
  /* U+E3E8 */	"\x8F\x7A", "\x8F\x7B", "\x8F\x7C", "\x8F\x7D",
  /* U+E3EC */	"\x8F\x7E", "\x8F\xA1", "\x8F\xA2", "\x8F\xA3",
  /* U+E3F0 */	"\x8F\xA4", "\x8F\xA5", "\x8F\xA6", "\x8F\xA7",
  /* U+E3F4 */	"\x8F\xA8", "\x8F\xA9", "\x8F\xAA", "\x8F\xAB",
  /* U+E3F8 */	"\x8F\xAC", "\x8F\xAD", "\x8F\xAE", "\x8F\xAF",
  /* U+E3FC */	"\x8F\xB0", "\x8F\xB1", "\x8F\xB2", "\x8F\xB3",
  /* U+E400 */	"\x8F\xB4", "\x8F\xB5", "\x8F\xB6", "\x8F\xB7",
  /* U+E404 */	"\x8F\xB8", "\x8F\xB9", "\x8F\xBA", "\x8F\xBB",
  /* U+E408 */	"\x8F\xBC", "\x8F\xBD", "\x8F\xBE", "\x8F\xBF",
  /* U+E40C */	"\x8F\xC0", "\x8F\xC1", "\x8F\xC2", "\x8F\xC3",
  /* U+E410 */	"\x8F\xC4", "\x8F\xC5", "\x8F\xC6", "\x8F\xC7",
  /* U+E414 */	"\x8F\xC8", "\x8F\xC9", "\x8F\xCA", "\xBF\xCC",
  /* U+E418 */	"\xA0\xD4", "\x8F\xCD", "\x8F\xCE", "\x8F\xCF",
  /* U+E41C */	"\x8F\xD0", "\x8F\xD1", "\x8F\xD2", "\x8F\xD3",
  /* U+E420 */	"\x8F\xD4", "\x8F\xD5", "\x8F\xD6", "\x8F\xD7",
  /* U+E424 */	"\x8F\xD8", "\x8F\xD9", "\x8F\xDA", "\x8F\xDB",
  /* U+E428 */	"\x8F\xDC", "\x8F\xDD", "\x8F\xDE", "\x8F\xDF",
  /* U+E42C */	"\x8F\xE0", "\x8F\xE1", "\x8F\xE2", "\x8F\xE3",
  /* U+E430 */	"\x8F\xE4", "\x8F\xE5", "\x8F\xE6", "\x8F\xE7",
  /* U+E434 */	"\x8F\xE8", "\x8F\xE9", "\x8F\xEA", "\x8F\xEB",
  /* U+E438 */	"\x8F\xEC", "\x8F\xED", "\x8F\xEE", "\x8F\xEF",
  /* U+E43C */	"\x8F\xF0", "\x8F\xF1", "\x8F\xF2", "\x8F\xF3",
  /* U+E440 */	"\x8F\xF4", "\x8F\xF5", "\x8F\xF6", "\x8F\xF7",
  /* U+E444 */	"\x8F\xF8", "\x8F\xF9", "\x8F\xFA", "\x8F\xFB",
  /* U+E448 */	"\x8F\xFC", "\x8F\xFD", "\xB0\x5F", "\x90\x40",
  /* U+E44C */	"\x90\x41", "\x90\x42", "\x90\x43", "\x90\x44",
  /* U+E450 */	"\x90\x45", "\x90\x46", "\x90\x47", "\x90\x48",
  /* U+E454 */	"\x90\x49", "\x90\x4A", "\x90\x4B", "\x90\x4C",
  /* U+E458 */	"\x90\x4D", "\x90\x4E", "\x90\x4F", "\x90\x50",
  /* U+E45C */	"\x90\x51", "\x90\x52", "\x90\x53", "\x90\x54",
  /* U+E460 */	"\x90\x55", "\x90\x56", "\x90\x57", "\x90\x58",
  /* U+E464 */	"\x90\x59", "\x90\x5A", "\x90\x5B", "\x90\x5C",
  /* U+E468 */	"\x90\x5D", "\x90\x5E", "\x90\x5F", "\x90\x60",
  /* U+E46C */	"\x90\x61", "\x90\x62", "\x90\x63", "\x90\x64",
  /* U+E470 */	"\x90\x65", "\x90\x66", "\x90\x67", "\x90\x68",
  /* U+E474 */	"\x90\x69", "\x90\x6A", "\x90\x6B", "\x90\x6C",
  /* U+E478 */	"\xB3\xA3", "\x90\x6E", "\x90\x6F", "\x90\x70",
  /* U+E47C */	"\x90\x71", "\x90\x72", "\x90\x73", "\x90\x74",
  /* U+E480 */	"\x90\x75", "\x90\x76", "\x90\x77", "\x90\x78",
  /* U+E484 */	"\x90\x79", "\xF9\xD7", "\x90\x7B", "\x90\x7C",
  /* U+E488 */	"\x90\x7D", "\x90\x7E", "\x90\xA1", "\x90\xA2",
  /* U+E48C */	"\x90\xA3", "\x90\xA4", "\x90\xA5", "\x90\xA6",
  /* U+E490 */	"\x90\xA7", "\x90\xA8", "\x90\xA9", "\x90\xAA",
  /* U+E494 */	"\x90\xAB", "\x90\xAC", "\x90\xAD", "\x90\xAE",
  /* U+E498 */	"\x90\xAF", "\x90\xB0", "\x90\xB1", "\x90\xB2",
  /* U+E49C */	"\x90\xB3", "\x90\xB4", "\x90\xB5", "\x90\xB6",
  /* U+E4A0 */	"\x90\xB7", "\x90\xB8", "\x90\xB9", "\x90\xBA",
  /* U+E4A4 */	"\x90\xBB", "\x90\xBC", "\x90\xBD", "\x90\xBE",
  /* U+E4A8 */	"\x90\xBF", "\x90\xC0", "\x90\xC1", "\x90\xC2",
  /* U+E4AC */	"\x90\xC3", "\x90\xC4", "\x90\xC5", "\x90\xC6",
  /* U+E4B0 */	"\x90\xC7", "\x90\xC8", "\x90\xC9", "\x90\xCA",
  /* U+E4B4 */	"\x90\xCB", "\x90\xCC", "\x90\xCD", "\x90\xCE",
  /* U+E4B8 */	"\x90\xCF", "\x90\xD0", "\x90\xD1", "\x90\xD2",
  /* U+E4BC */	"\x90\xD3", "\x90\xD4", "\x90\xD5", "\x90\xD6",
  /* U+E4C0 */	"\x90\xD7", "\x90\xD8", "\x90\xD9", "\x90\xDA",
  /* U+E4C4 */	"\x90\xDB", "\xC0\x52", "\x90\xDD", "\x90\xDE",
  /* U+E4C8 */	"\x90\xDF", "\x90\xE0", "\x90\xE1", "\x90\xE2",
  /* U+E4CC */	"\x90\xE3", "\x90\xE4", "\x90\xE5", "\x90\xE6",
  /* U+E4D0 */	"\x90\xE7", "\x90\xE8", "\x90\xE9", "\x90\xEA",
  /* U+E4D4 */	"\x90\xEB", "\x90\xEC", "\x90\xED", "\x90\xEE",
  /* U+E4D8 */	"\x90\xEF", "\x90\xF0", "\xC5\x54", "\x90\xF2",
  /* U+E4DC */	"\x90\xF3", "\x90\xF4", "\x90\xF5", "\x90\xF6",
  /* U+E4E0 */	"\x90\xF7", "\x90\xF8", "\x90\xF9", "\x90\xFA",
  /* U+E4E4 */	"\x90\xFB", "\x90\xFC", "\x90\xFD", "\x90\xFE",
  /* U+E4E8 */	"\x91\x40", "\x91\x41", "\x91\x42", "\x91\x43",
  /* U+E4EC */	"\x91\x44", "\x91\x45", "\x91\x46", "\x91\x47",
  /* U+E4F0 */	"\x91\x48", "\x91\x49", "\x91\x4A", "\x91\x4B",
  /* U+E4F4 */	"\x91\x4C", "\x91\x4D", "\x91\x4E", "\x91\x4F",
  /* U+E4F8 */	"\x91\x50", "\x91\x51", "\x91\x52", "\x91\x53",
  /* U+E4FC */	"\x91\x54", "\x91\x55", "\x91\x56", "\x91\x57",
  /* U+E500 */	"\x91\x58", "\x91\x59", "\x91\x5A", "\x91\x5B",
  /* U+E504 */	"\x91\x5C", "\x91\x5D", "\x91\x5E", "\x91\x5F",
  /* U+E508 */	"\x91\x60", "\x91\x61", "\x91\x62", "\x91\x63",
  /* U+E50C */	"\x91\x64", "\x91\x65", "\x91\x66", "\x91\x67",
  /* U+E510 */	"\x91\x68", "\x91\x69", "\x91\x6A", "\x91\x6B",
  /* U+E514 */	"\x91\x6C", "\x91\x6D", "\x91\x6E", "\x91\x6F",
  /* U+E518 */	"\x91\x70", "\x91\x71", "\x91\x72", "\x91\x73",
  /* U+E51C */	"\x91\x74", "\x91\x75", "\x91\x76", "\x91\x77",
  /* U+E520 */	"\x91\x78", "\x91\x79", "\x91\x7A", "\x91\x7B",
  /* U+E524 */	"\x91\x7C", "\x91\x7D", "\x91\x7E", "\x91\xA1",
  /* U+E528 */	"\x91\xA2", "\x91\xA3", "\x91\xA4", "\x91\xA5",
  /* U+E52C */	"\x91\xA6", "\x91\xA7", "\x91\xA8", "\x91\xA9",
  /* U+E530 */	"\x91\xAA", "\x91\xAB", "\x91\xAC", "\x91\xAD",
  /* U+E534 */	"\x91\xAE", "\x91\xAF", "\x91\xB0", "\x91\xB1",
  /* U+E538 */	"\x91\xB2", "\x91\xB3", "\x91\xB4", "\x91\xB5",
  /* U+E53C */	"\x91\xB6", "\x91\xB7", "\x91\xB8", "\x91\xB9",
  /* U+E540 */	"\x91\xBA", "\x91\xBB", "\x91\xBC", "\x91\xBD",
  /* U+E544 */	"\x91\xBE", "\xF1\xE3", "\x91\xC0", "\x91\xC1",
  /* U+E548 */	"\x91\xC2", "\x91\xC3", "\x91\xC4", "\x91\xC5",
  /* U+E54C */	"\x91\xC6", "\x91\xC7", "\x91\xC8", "\x91\xC9",
  /* U+E550 */	"\x91\xCA", "\x91\xCB", "\x91\xCC", "\x91\xCD",
  /* U+E554 */	"\x91\xCE", "\x91\xCF", "\x91\xD0", "\x91\xD1",
  /* U+E558 */	"\x91\xD2", "\x91\xD3", "\x91\xD4", "\x91\xD5",
  /* U+E55C */	"\x91\xD6", "\x91\xD7", "\x91\xD8", "\x91\xD9",
  /* U+E560 */	"\x91\xDA", "\x91\xDB", "\x91\xDC", "\x91\xDD",
  /* U+E564 */	"\x91\xDE", "\x91\xDF", "\x91\xE0", "\x91\xE1",
  /* U+E568 */	"\x91\xE2", "\x91\xE3", "\x91\xE4", "\x91\xE5",
  /* U+E56C */	"\x91\xE6", "\x91\xE7", "\x91\xE8", "\x91\xE9",
  /* U+E570 */	"\x91\xEA", "\x91\xEB", "\x91\xEC", "\x91\xED",
  /* U+E574 */	"\x91\xEE", "\x91\xEF", "\x91\xF0", "\x91\xF1",
  /* U+E578 */	"\x91\xF2", "\x91\xF3", "\x91\xF4", "\x91\xF5",
  /* U+E57C */	"\x91\xF6", "\x91\xF7", "\x91\xF8", "\x91\xF9",
  /* U+E580 */	"\x91\xFA", "\x91\xFB", "\x91\xFC", "\x91\xFD",
  /* U+E584 */	"\x91\xFE", "\x92\x40", "\x92\x41", "\x92\x42",
  /* U+E588 */	"\x92\x43", "\x92\x42", "\x92\x45", "\x92\x46",
  /* U+E58C */	"\x92\x47", "\x92\x48", "\x92\x49", "\x92\x4A",
  /* U+E590 */	"\x92\x4B", "\x92\x4C", "\x92\x4D", "\x92\x4E",
  /* U+E594 */	"\x92\x4F", "\x92\x50", "\x92\x51", "\x92\x52",
  /* U+E598 */	"\x92\x53", "\x92\x54", "\x92\x55", "\x92\x56",
  /* U+E59C */	"\x92\x57", "\x92\x58", "\x92\x59", "\x92\x5A",
  /* U+E5A0 */	"\x92\x5B", "\x92\x5C", "\x92\x5D", "\x92\x5E",
  /* U+E5A4 */	"\x92\x5F", "\x92\x60", "\x92\x61", "\x92\x62",
  /* U+E5A8 */	"\x92\x63", "\x92\x64", "\x92\x65", "\x92\x66",
  /* U+E5AC */	"\x92\x67", "\x92\x68", "\x92\x69", "\x92\x6A",
  /* U+E5B0 */	"\x92\x6B", "\x92\x6C", "\x92\x6D", "\x92\x6E",
  /* U+E5B4 */	"\x92\x6F", "\x92\x70", "\x92\x71", "\x92\x72",
  /* U+E5B8 */	"\x92\x73", "\x92\x74", "\x92\x75", "\x92\x76",
  /* U+E5BC */	"\x92\x77", "\x92\x78", "\x92\x79", "\x92\x7A",
  /* U+E5C0 */	"\x92\x7B", "\x92\x7C", "\x92\x7D", "\x92\x7E",
  /* U+E5C4 */	"\x92\xA1", "\x92\xA2", "\x92\xA3", "\x92\xA4",
  /* U+E5C8 */	"\x92\xA5", "\x92\xA6", "\x92\xA7", "\x92\xA8",
  /* U+E5CC */	"\x92\xA9", "\x92\xAA", "\x92\xAB", "\x92\xAC",
  /* U+E5D0 */	"\x92\xAD", "\x92\xAE", "\xA2\x59", "\xA2\x5A",
  /* U+E5D4 */	"\xA2\x5C", "\xA2\x5B", "\x92\xB3", "\x92\xB4",
  /* U+E5D8 */	"\x92\xB5", "\x92\xB6", "\x92\xB7", "\x92\xB8",
  /* U+E5DC */	"\x92\xB9", "\x92\xBA", "\x92\xBB", "\x92\xBC",
  /* U+E5E0 */	"\x92\xBD", "\x92\xBE", "\x92\xBF", "\x92\xC0",
  /* U+E5E4 */	"\x92\xC1", "\x92\xC2", "\x92\xC3", "\x92\xC4",
  /* U+E5E8 */	"\x92\xC5", "\x92\xC6", "\x92\xC7", "\xA0\x5F",
  /* U+E5EC */	"\x92\xC9", "\x92\xCA", "\x92\xCB", "\x92\xCC",
  /* U+E5F0 */	"\x92\xCD", "\x92\xCE", "\x92\xCF", "\x92\xD0",
  /* U+E5F4 */	"\xE6\xAB", "\x92\xD2", "\x92\xD3", "\x92\xD4",
  /* U+E5F8 */	"\x92\xD5", "\x92\xD6", "\x92\xD7", "\x92\xD8",
  /* U+E5FC */	"\x92\xD9", "\x92\xDA", "\x92\xDB", "\x92\xDC",
  /* U+E600 */	"\x92\xDD", "\x92\xDE", "\x92\xDF", "\x92\xE0",
  /* U+E604 */	"\x92\xE1", "\x92\xE2", "\x92\xE3", "\x92\xE4",
  /* U+E608 */	"\x92\xE5", "\x92\xE6", "\x92\xE7", "\x92\xE8",
  /* U+E60C */	"\x92\xE9", "\x92\xEA", "\x92\xEB", "\x92\xEC",
  /* U+E610 */	"\x92\xED", "\x92\xEE", "\x92\xEF", "\x92\xF0",
  /* U+E614 */	"\x92\xF1", "\x92\xF2", "\x92\xF3", "\x92\xF4",
  /* U+E618 */	"\x92\xF5", "\x92\xF6", "\x92\xF7", "\x92\xF8",
  /* U+E61C */	"\x92\xF9", "\x92\xFA", "\x92\xFB", "\x92\xFC",
  /* U+E620 */	"\x92\xFD", "\x92\xFE", "\x93\x40", "\x93\x41",
  /* U+E624 */	"\x93\x42", "\x93\x43", "\x93\x44", "\x93\x45",
  /* U+E628 */	"\x93\x46", "\x93\x47", "\x93\x48", "\x93\x49",
  /* U+E62C */	"\x93\x4A", "\x93\x4B", "\x93\x4C", "\x93\x4D",
  /* U+E630 */	"\x93\x4E", "\x93\x4F", "\x93\x50", "\x93\x51",
  /* U+E634 */	"\x93\x52", "\x93\x53", "\x93\x54", "\x93\x55",
  /* U+E638 */	"\x93\x56", "\x93\x57", "\x93\x58", "\x93\x59",
  /* U+E63C */	"\x93\x5A", "\x93\x5B", "\x93\x5C", "\x93\x5D",
  /* U+E640 */	"\x93\x5E", "\x93\x5F", "\x93\x60", "\x93\x61",
  /* U+E644 */	"\x93\x62", "\x93\x63", "\x93\x64", "\x93\x65",
  /* U+E648 */	"\x93\x66", "\x93\x67", "\x93\x68", "\x93\x69",
  /* U+E64C */	"\x93\x6A", "\x93\x6B", "\x93\x6C", "\x93\x6D",
  /* U+E650 */	"\x93\x6E", "\x93\x6F", "\x93\x70", "\x93\x71",
  /* U+E654 */	"\x93\x72", "\x93\x73", "\x93\x74", "\x93\x75",
  /* U+E658 */	"\x93\x76", "\x93\x77", "\x93\x78", "\x93\x79",
  /* U+E65C */	"\x93\x7A", "\x93\x7B", "\x93\x7C", "\x93\x7D",
  /* U+E660 */	"\x93\x7E", "\x93\xA1", "\x93\xA2", "\x93\xA3",
  /* U+E664 */	"\x93\xA4", "\x93\xA5", "\x93\xA6", "\x93\xA7",
  /* U+E668 */	"\x93\xA8", "\x93\xA9", "\x93\xAA", "\x93\xAB",
  /* U+E66C */	"\x93\xAC", "\x93\xAD", "\x93\xAE", "\x93\xAF",
  /* U+E670 */	"\x93\xB0", "\x93\xB1", "\x93\xB2", "\x93\xB3",
  /* U+E674 */	"\x93\xB4", "\x93\xB5", "\x93\xB6", "\x93\xB7",
  /* U+E678 */	"\x93\xB8", "\x93\xB9", "\x93\xBA", "\x93\xBB",
  /* U+E67C */	"\x93\xBC", "\x93\xBD", "\x93\xBE", "\x93\xBF",
  /* U+E680 */	"\x93\xC0", "\x93\xC1", "\x93\xC2", "\x93\xC3",
  /* U+E684 */	"\x93\xC4", "\x93\xC5", "\x93\xC6", "\x93\xC7",
  /* U+E688 */	"\x93\xC8", "\x93\xC9", "\x93\xCA", "\x93\xCB",
  /* U+E68C */	"\x93\xCC", "\x93\xCD", "\x93\xCE", "\x93\xCF",
  /* U+E690 */	"\x93\xD0", "\x93\xD1", "\x93\xD2", "\x93\xD3",
  /* U+E694 */	"\x93\xD4", "\x93\xD5", "\x93\xD6", "\x93\xD7",
  /* U+E698 */	"\x93\xD8", "\x93\xD9", "\x93\xDA", "\x93\xDB",
  /* U+E69C */	"\x93\xDC", "\x93\xDD", "\x93\xDE", "\x93\xDF",
  /* U+E6A0 */	"\x93\xE0", "\x93\xE1", "\x93\xE2", "\x93\xE3",
  /* U+E6A4 */	"\x93\xE4", "\x93\xE5", "\x93\xE6", "\x93\xE7",
  /* U+E6A8 */	"\x93\xE8", "\x93\xE9", "\x93\xEA", "\x93\xEB",
  /* U+E6AC */	"\x93\xEC", "\x93\xED", "\x93\xEE", "\x93\xEF",
  /* U+E6B0 */	"\x93\xF0", "\x93\xF1", "\x93\xF2", "\x93\xF3",
  /* U+E6B4 */	"\x93\xF4", "\x93\xF5", "\x93\xF6", "\x93\xF7",
  /* U+E6B8 */	"\x93\xF8", "\x93\xF9", "\x93\xFA", "\x93\xFB",
  /* U+E6BC */	"\x93\xFC", "\x93\xFD", "\x93\xFE", "\x94\x40",
  /* U+E6C0 */	"\x94\x41", "\x94\x42", "\x94\x43", "\x94\x44",
  /* U+E6C4 */	"\x94\x45", "\x94\x46", "\xD2\x56", "\x94\x48",
  /* U+E6C8 */	"\x94\x49", "\x94\x4A", "\x94\x4B", "\x94\x4C",
  /* U+E6CC */	"\x94\x4D", "\x94\x4E", "\x94\x4F", "\x94\x50",
  /* U+E6D0 */	"\x94\x51", "\x94\x52", "\x94\x53", "\x94\x54",
  /* U+E6D4 */	"\x94\x55", "\x94\x56", "\x94\x57", "\x94\x58",
  /* U+E6D8 */	"\x94\x59", "\x94\x5A", "\x94\x5B", "\x94\x5C",
  /* U+E6DC */	"\x94\x5D", "\x94\x5E", "\x94\x5F", "\x94\x60",
  /* U+E6E0 */	"\x94\x61", "\x94\x62", "\x94\x63", "\x94\x64",
  /* U+E6E4 */	"\x94\x65", "\x94\x66", "\x94\x67", "\x94\x68",
  /* U+E6E8 */	"\x94\x69", "\x94\x6A", "\x94\x6B", "\x94\x6C",
  /* U+E6EC */	"\x94\x6D", "\x94\x6E", "\x94\x6F", "\x94\x70",
  /* U+E6F0 */	"\x94\x71", "\x94\x72", "\x94\x73", "\x94\x74",
  /* U+E6F4 */	"\x94\x75", "\x94\x76", "\x94\x77", "\x94\x78",
  /* U+E6F8 */	"\x94\x79", "\x94\x7A", "\x94\x7B", "\x94\x7C",
  /* U+E6FC */	"\x94\x7D", "\x94\x7E", "\x94\xA1", "\x94\xA2",
  /* U+E700 */	"\x94\xA3", "\x94\xA4", "\x94\xA5", "\x94\xA6",
  /* U+E704 */	"\x94\xA7", "\x94\xA8", "\x94\xA9", "\x94\xAA",
  /* U+E708 */	"\x94\xAB", "\x94\xAC", "\x94\xAD", "\x94\xAE",
  /* U+E70C */	"\x94\xAF", "\x94\xB0", "\x94\xB1", "\x94\xB2",
  /* U+E710 */	"\x94\xB3", "\x94\xB4", "\x94\xB5", "\x94\xB6",
  /* U+E714 */	"\x94\xB7", "\x94\xB8", "\x94\xB9", "\x94\xBA",
  /* U+E718 */	"\x94\xBB", "\x94\xBC", "\x94\xBD", "\x94\xBE",
  /* U+E71C */	"\x94\xBF", "\x94\xC0", "\x94\xC1", "\x94\xC2",
  /* U+E720 */	"\x94\xC3", "\x94\xC4", "\x94\xC5", "\x94\xC6",
  /* U+E724 */	"\x94\xC7", "\x94\xC8", "\x94\xC9", "\xE6\xD0",
  /* U+E728 */	"\x94\xCB", "\x94\xCC", "\x94\xCD", "\x94\xCE",
  /* U+E72C */	"\x94\xCF", "\x94\xD0", "\x94\xD1", "\x94\xD2",
  /* U+E730 */	"\x94\xD3", "\x94\xD4", "\x94\xD5", "\x94\xD6",
  /* U+E734 */	"\x94\xD7", "\x94\xD8", "\x94\xD9", "\x94\xDA",
  /* U+E738 */	"\x94\xDB", "\x94\xDC", "\x94\xDD", "\x94\xDE",
  /* U+E73C */	"\x94\xDF", "\x94\xE0", "\x94\xE1", "\x94\xE2",
  /* U+E740 */	"\x94\xE3", "\x94\xE4", "\x94\xE5", "\x94\xE6",
  /* U+E744 */	"\x94\xE7", "\x94\xE8", "\x94\xE9", "\x94\xEA",
  /* U+E748 */	"\x94\xEB", "\x94\xEC", "\x94\xED", "\x94\xEE",
  /* U+E74C */	"\x94\xEF", "\x94\xF0", "\x94\xF1", "\x94\xF2",
  /* U+E750 */	"\x94\xF3", "\x94\xF4", "\x94\xF5", "\x94\xF6",
  /* U+E754 */	"\x94\xF7", "\x94\xF8", "\x94\xF9", "\x94\xFA",
  /* U+E758 */	"\x94\xFB", "\x94\xFC", "\x94\xFD", "\x94\xFE",
  /* U+E75C */	"\x95\x40", "\x95\x41", "\x95\x42", "\x95\x43",
  /* U+E760 */	"\x95\x44", "\x95\x45", "\x95\x46", "\x95\x47",
  /* U+E764 */	"\x95\x48", "\x95\x49", "\x95\x4A", "\x95\x4B",
  /* U+E768 */	"\x95\x4C", "\x95\x4D", "\x95\x4E", "\x95\x4F",
  /* U+E76C */	"\x95\x50", "\x95\x51", "\x95\x52", "\x95\x53",
  /* U+E770 */	"\x95\x54", "\x95\x55", "\x95\x56", "\x95\x57",
  /* U+E774 */	"\x95\x58", "\x95\x59", "\x95\x5A", "\x95\x5B",
  /* U+E778 */	"\x95\x5C", "\x95\x5D", "\x95\x5E", "\x95\x5F",
  /* U+E77C */	"\x95\x60", "\x95\x61", "\x95\x62", "\x95\x63",
  /* U+E780 */	"\x95\x64", "\x95\x65", "\x95\x66", "\x95\x67",
  /* U+E784 */	"\x95\x68", "\x95\x69", "\x95\x6A", "\x95\x6B",
  /* U+E788 */	"\x95\x6C", "\x95\x6D", "\x95\x6E", "\x95\x6F",
  /* U+E78C */	"\x95\x70", "\x95\x71", "\x95\x72", "\x95\x73",
  /* U+E790 */	"\x95\x74", "\x95\x75", "\x95\x76", "\x95\x77",
  /* U+E794 */	"\x95\x78", "\x95\x79", "\x95\x7A", "\x95\x7B",
  /* U+E798 */	"\x95\x7C", "\x95\x7D", "\x95\x7E", "\x95\xA1",
  /* U+E79C */	"\x95\xA2", "\x95\xA3", "\x95\xA4", "\x95\xA5",
  /* U+E7A0 */	"\x95\xA6", "\x95\xA7", "\x95\xA8", "\x95\xA9",
  /* U+E7A4 */	"\x95\xAA", "\x95\xAB", "\x95\xAC", "\x95\xAD",
  /* U+E7A8 */	"\x95\xAE", "\x95\xAF", "\x95\xB0", "\x95\xB1",
  /* U+E7AC */	"\x95\xB2", "\x95\xB3", "\x95\xB4", "\x95\xB5",
  /* U+E7B0 */	"\x95\xB6", "\x95\xB7", "\x95\xB8", "\x95\xB9",
  /* U+E7B4 */	"\x95\xBA", "\x95\xBB", "\x95\xBC", "\x95\xBD",
  /* U+E7B8 */	"\x95\xBE", "\x95\xBF", "\x95\xC0", "\x95\xC1",
  /* U+E7BC */	"\x95\xC2", "\x95\xC3", "\x95\xC4", "\x95\xC5",
  /* U+E7C0 */	"\x95\xC6", "\x95\xC7", "\x95\xC8", "\x95\xC9",
  /* U+E7C4 */	"\x95\xCA", "\x95\xCB", "\x95\xCC", "\x95\xCD",
  /* U+E7C8 */	"\x95\xCE", "\x95\xCF", "\x95\xD0", "\x95\xD1",
  /* U+E7CC */	"\x95\xD2", "\x95\xD3", "\x95\xD4", "\x95\xD5",
  /* U+E7D0 */	"\x95\xD6", "\x95\xD7", "\x95\xD8", "\xCA\x52",
  /* U+E7D4 */	"\x95\xDA", "\x95\xDB", "\x95\xDC", "\x95\xDD",
  /* U+E7D8 */	"\x95\xDE", "\x95\xDF", "\x95\xE0", "\x95\xE1",
  /* U+E7DC */	"\x95\xE2", "\x95\xE3", "\x95\xE4", "\x95\xE5",
  /* U+E7E0 */	"\x95\xE6", "\x95\xE7", "\x95\xE8", "\x95\xE9",
  /* U+E7E4 */	"\x95\xEA", "\x95\xEB", "\x95\xEC", "\x95\xED",
  /* U+E7E8 */	"\x95\xEE", "\x95\xEF", "\x95\xF0", "\x95\xF1",
  /* U+E7EC */	"\x95\xF2", "\x95\xF3", "\x95\xF4", "\x95\xF5",
  /* U+E7F0 */	"\x95\xF6", "\x95\xF7", "\x95\xF8", "\x95\xF9",
  /* U+E7F4 */	"\x95\xFA", "\x95\xFB", "\x95\xFC", "\x95\xFD",
  /* U+E7F8 */	"\x95\xFE", "\x96\x40", "\x96\x41", "\x96\x42",
  /* U+E7FC */	"\x96\x43", "\x9C\xE4", "\x96\x45", "\x96\x46",
  /* U+E800 */	"\x96\x47", "\x96\x48", "\x96\x49", "\x96\x4A",
  /* U+E804 */	"\x96\x4B", "\x96\x4C", "\x96\x4D", "\x96\x4E",
  /* U+E808 */	"\x96\x4F", "\x96\x50", "\x96\x51", "\x96\x52",
  /* U+E80C */	"\x96\x53", "\x96\x54", "\x96\x55", "\x96\x56",
  /* U+E810 */	"\x96\x57", "\x96\x58", "\x96\x59", "\x96\x5A",
  /* U+E814 */	"\x96\x5B", "\x96\x5C", "\x96\x5D", "\x96\x5E",
  /* U+E818 */	"\x96\x5F", "\x96\x60", "\x96\x61", "\x96\x62",
  /* U+E81C */	"\x96\x63", "\x96\x64", "\x96\x65", "\x96\x66",
  /* U+E820 */	"\x96\x67", "\x96\x68", "\x96\x69", "\x96\x6A",
  /* U+E824 */	"\x96\x6B", "\x96\x6C", "\x96\x6D", "\x96\x6E",
  /* U+E828 */	"\x96\x6F", "\x96\x70", "\x96\x71", "\x96\x72",
  /* U+E82C */	"\x96\x73", "\x96\x74", "\x96\x75", "\x96\x76",
  /* U+E830 */	"\x96\x77", "\x96\x78", "\x96\x79", "\x96\x7A",
  /* U+E834 */	"\x96\x7B", "\x96\x7C", "\x96\x7D", "\x96\x7E",
  /* U+E838 */	"\x96\xA1", "\x96\xA2", "\x96\xA3", "\x96\xA4",
  /* U+E83C */	"\x96\xA5", "\x96\xA6", "\x96\xA7", "\x96\xA8",
  /* U+E840 */	"\x96\xA9", "\x96\xAA", "\x96\xAB", "\x96\xAC",
  /* U+E844 */	"\x96\xAD", "\x96\xAE", "\x96\xAF", "\x96\xB0",
  /* U+E848 */	"\x96\xB1", "\x96\xB2", "\x96\xB3", "\x96\xB4",
  /* U+E84C */	"\x96\xB5", "\x96\xB6", "\x96\xB7", "\x96\xB8",
  /* U+E850 */	"\x96\xB9", "\x96\xBA", "\x96\xBB", "\x96\xBC",
  /* U+E854 */	"\x96\xBD", "\x96\xBE", "\x96\xBF", "\x96\xC0",
  /* U+E858 */	"\x96\xC1", "\x96\xC2", "\x96\xC3", "\x96\xC4",
  /* U+E85C */	"\x96\xC5", "\x96\xC6", "\x96\xC7", "\x96\xC8",
  /* U+E860 */	"\x96\xC9", "\x96\xCA", "\x96\xCB", "\x96\xCC",
  /* U+E864 */	"\x96\xCD", "\x96\xCE", "\x96\xCF", "\x96\xD0",
  /* U+E868 */	"\x96\xD1", "\x96\xD2", "\x96\xD3", "\x96\xD4",
  /* U+E86C */	"\x96\xD5", "\x96\xD6", "\x96\xD7", "\x96\xD8",
  /* U+E870 */	"\x96\xD9", "\x96\xDA", "\x96\xDB", "\x96\xDC",
  /* U+E874 */	"\x96\xDD", "\x96\xDE", "\x96\xDF", "\x96\xE0",
  /* U+E878 */	"\x96\xE1", "\x96\xE2", "\x96\xE3", "\x96\xE4",
  /* U+E87C */	"\x96\xE5", "\x96\xE6", "\x96\xE7", "\x96\xE8",
  /* U+E880 */	"\x96\xE9", "\x96\xEA", "\x96\xEB", "\x96\xEC",
  /* U+E884 */	"\x96\xEE", "\x96\xEE", "\x96\xEF", "\x96\xF0",
  /* U+E888 */	"\x96\xF1", "\x96\xF2", "\x96\xF3", "\x96\xF4",
  /* U+E88C */	"\x96\xF5", "\x96\xF6", "\x96\xF7", "\x96\xF8",
  /* U+E890 */	"\x96\xF9", "\x96\xFA", "\x96\xFB", "\xE9\x59",
  /* U+E894 */	"\x96\xFD", "\x96\xFE", "\x97\x40", "\x97\x41",
  /* U+E898 */	"\x97\x42", "\x97\x43", "\x97\x44", "\x97\x45",
  /* U+E89C */	"\x97\x46", "\x97\x47", "\x97\x48", "\x97\x49",
  /* U+E8A0 */	"\x97\x4A", "\x97\x4B", "\x97\x4C", "\x97\x4D",
  /* U+E8A4 */	"\x97\x4E", "\x97\x4F", "\x97\x50", "\x97\x51",
  /* U+E8A8 */	"\x97\x52", "\x97\x53", "\x97\x54", "\x97\x55",
  /* U+E8AC */	"\x97\x56", "\x97\x57", "\x97\x58", "\x97\x59",
  /* U+E8B0 */	"\x97\x5A", "\x97\x5B", "\x97\x5C", "\x97\x5D",
  /* U+E8B4 */	"\x97\x5E", "\x97\x5F", "\x97\x60", "\x97\x61",
  /* U+E8B8 */	"\x97\x62", "\x97\x63", "\x97\x64", "\x97\x65",
  /* U+E8BC */	"\x97\x66", "\x97\x67", "\x97\x68", "\x97\x69",
  /* U+E8C0 */	"\x97\x6A", "\x97\x6B", "\x97\x6C", "\x97\x6D",
  /* U+E8C4 */	"\x97\x6E", "\x97\x6F", "\x97\x70", "\x97\x71",
  /* U+E8C8 */	"\x97\x72", "\x97\x73", "\x97\x74", "\x97\x75",
  /* U+E8CC */	"\x97\x76", "\x97\x77", "\x97\x78", "\x97\x79",
  /* U+E8D0 */	"\x97\x7A", "\x97\x7B", "\x97\x7C", "\x97\x7D",
  /* U+E8D4 */	"\x97\x7E", "\x97\xA1", "\x97\xA2", "\x97\xA3",
  /* U+E8D8 */	"\x97\xA4", "\x97\xA5", "\x97\xA6", "\x97\xA7",
  /* U+E8DC */	"\x97\xA8", "\x97\xA9", "\x97\xAA", "\x97\xAB",
  /* U+E8E0 */	"\x97\xAC", "\x97\xAD", "\x97\xAE", "\x97\xAF",
  /* U+E8E4 */	"\x97\xB0", "\x97\xB1", "\x97\xB2", "\x97\xB3",
  /* U+E8E8 */	"\x97\xB4", "\x97\xB5", "\x97\xB6", "\x97\xB7",
  /* U+E8EC */	"\x97\xB8", "\x97\xB9", "\x97\xBA", "\x97\xBB",
  /* U+E8F0 */	"\x97\xBC", "\x97\xBD", "\x97\xBE", "\x97\xBF",
  /* U+E8F4 */	"\x97\xC0", "\x97\xC1", "\x97\xC2", "\x97\xC3",
  /* U+E8F8 */	"\x97\xC4", "\x97\xC5", "\x97\xC6", "\x97\xC7",
  /* U+E8FC */	"\x97\xC8", "\x97\xC9", "\x97\xCA", "\x97\xCB",
  /* U+E900 */	"\x97\xCC", "\x97\xCD", "\x97\xCE", "\x97\xCF",
  /* U+E904 */	"\x97\xD0", "\x97\xD1", "\x97\xD2", "\x97\xD3",
  /* U+E908 */	"\x97\xD4", "\x97\xD5", "\x97\xD6", "\x97\xD7",
  /* U+E90C */	"\x97\xD8", "\x97\xD9", "\x97\xDA", "\x97\xDB",
  /* U+E910 */	"\x97\xDC", "\x97\xDD", "\x97\xDE", "\x97\xDF",
  /* U+E914 */	"\x97\xE0", "\x97\xE1", "\x97\xE2", "\x97\xE3",
  /* U+E918 */	"\x97\xE4", "\x97\xE5", "\x97\xE6", "\x97\xE7",
  /* U+E91C */	"\x97\xE8", "\x97\xE9", "\x97\xEA", "\x97\xEB",
  /* U+E920 */	"\x97\xEC", "\x97\xED", "\x97\xEE", "\x97\xEF",
  /* U+E924 */	"\x97\xF0", "\x97\xF1", "\x97\xF2", "\x97\xF3",
  /* U+E928 */	"\x97\xF4", "\x97\xF5", "\x97\xF6", "\x97\xF7",
  /* U+E92C */	"\x97\xF8", "\x97\xF9", "\x97\xFA", "\x97\xFB",
  /* U+E930 */	"\x97\xFC", "\x97\xFD", "\x97\xFE", "\x98\x40",
  /* U+E934 */	"\x98\x41", "\x98\x42", "\x98\x43", "\x98\x44",
  /* U+E938 */	"\x98\x45", "\x98\x46", "\x98\x47", "\x98\x48",
  /* U+E93C */	"\x98\x49", "\x98\x4A", "\x98\x4B", "\x98\x4C",
  /* U+E940 */	"\x98\x4D", "\x98\x4E", "\x98\x4F", "\x98\x50",
  /* U+E944 */	"\x98\x51", "\x98\x52", "\x98\x53", "\x98\x54",
  /* U+E948 */	"\x98\x55", "\x98\x56", "\x98\x57", "\x98\x58",
  /* U+E94C */	"\x98\x59", "\x98\x5A", "\x98\x5B", "\x98\x5C",
  /* U+E950 */	"\x98\x5D", "\x98\x5E", "\x98\x5F", "\x98\x60",
  /* U+E954 */	"\x98\x61", "\x98\x62", "\x98\x63", "\x98\x64",
  /* U+E958 */	"\x98\x65", "\x98\x66", "\x98\x67", "\x98\x68",
  /* U+E95C */	"\x98\x69", "\x98\x6A", "\x98\x6B", "\x98\x6C",
  /* U+E960 */	"\x98\x6D", "\x98\x6E", "\x98\x6F", "\x98\x70",
  /* U+E964 */	"\x98\x71", "\x98\x72", "\x98\x73", "\x98\x74",
  /* U+E968 */	"\x98\x75", "\x98\x76", "\x98\x77", "\x98\x78",
  /* U+E96C */	"\x98\x79", "\x98\x7A", "\x98\x7B", "\x98\x7C",
  /* U+E970 */	"\x98\x7D", "\x98\x7E", "\x98\xA1", "\x98\xA2",
  /* U+E974 */	"\x98\xA3", "\x98\xA4", "\x98\xA5", "\x98\xA6",
  /* U+E978 */	"\x98\xA7", "\x98\xA8", "\x98\xA9", "\x98\xAA",
  /* U+E97C */	"\x98\xAB", "\x98\xAC", "\x98\xAD", "\x98\xAE",
  /* U+E980 */	"\x98\xAF", "\x98\xB0", "\x98\xB1", "\x98\xB2",
  /* U+E984 */	"\x98\xB3", "\x98\xB4", "\x98\xB5", "\x98\xB6",
  /* U+E988 */	"\x98\xB7", "\x98\xB8", "\x98\xB9", "\x98\xBA",
  /* U+E98C */	"\x98\xBB", "\x98\xBC", "\x98\xBD", "\x98\xBE",
  /* U+E990 */	"\x98\xBF", "\x98\xC0", "\x98\xC1", "\x98\xC2",
  /* U+E994 */	"\x98\xC3", "\x98\xC4", "\x98\xC5", "\x98\xC6",
  /* U+E998 */	"\x98\xC7", "\x98\xC8", "\x98\xC9", "\x98\xCA",
  /* U+E99C */	"\x98\xCB", "\x98\xCC", "\x98\xCD", "\x98\xCE",
  /* U+E9A0 */	"\x98\xCF", "\x98\xD0", "\x98\xD1", "\x98\xD2",
  /* U+E9A4 */	"\x98\xD3", "\x98\xD4", "\x98\xD5", "\x98\xD6",
  /* U+E9A8 */	"\x98\xD7", "\x98\xD8", "\x98\xD9", "\x98\xDA",
  /* U+E9AC */	"\x98\xDB", "\x98\xDC", "\x98\xDD", "\x98\xDE",
  /* U+E9B0 */	"\x98\xDF", "\x98\xE0", "\x98\xE1", "\x98\xE2",
  /* U+E9B4 */	"\x98\xE3", "\x98\xE4", "\x98\xE5", "\x98\xE6",
  /* U+E9B8 */	"\x98\xE7", "\x98\xE8", "\x98\xE9", "\x98\xEA",
  /* U+E9BC */	"\x98\xEB", "\x98\xEC", "\x98\xED", "\x98\xEE",
  /* U+E9C0 */	"\x98\xEF", "\x98\xF0", "\x98\xF1", "\x98\xF2",
  /* U+E9C4 */	"\x98\xF3", "\x98\xF4", "\x98\xF5", "\x98\xF6",
  /* U+E9C8 */	"\x98\xF7", "\x98\xF8", "\x98\xF9", "\x98\xFA",
  /* U+E9CC */	"\x98\xFB", "\x98\xFC", "\x98\xFD", "\x98\xFE",
  /* U+E9D0 */	"\x99\x40", "\x99\x41", "\x99\x42", "\x99\x43",
  /* U+E9D4 */	"\x99\x44", "\x99\x45", "\x99\x46", "\x99\x47",
  /* U+E9D8 */	"\x99\x48", "\x99\x49", "\x99\x4A", "\x99\x4B",
  /* U+E9DC */	"\x99\x4C", "\x99\x4D", "\x99\x4E", "\x99\x4F",
  /* U+E9E0 */	"\x99\x50", "\x99\x51", "\x99\x52", "\x99\x53",
  /* U+E9E4 */	"\x99\x54", "\x99\x55", "\x99\x56", "\x99\x57",
  /* U+E9E8 */	"\x99\x58", "\x99\x59", "\x99\x5A", "\x99\x5B",
  /* U+E9EC */	"\x99\x5C", "\x99\x5D", "\x99\x5E", "\x99\x5F",
  /* U+E9F0 */	"\x99\x60", "\x99\x61", "\x99\x62", "\x99\x63",
  /* U+E9F4 */	"\x99\x64", "\x99\x65", "\x99\x66", "\x99\x67",
  /* U+E9F8 */	"\x99\x68", "\x99\x69", "\x99\x6A", "\x99\x6B",
  /* U+E9FC */	"\x99\x6C", "\x99\x6D", "\x99\x6E", "\x99\x6F",
  /* U+EA00 */	"\x99\x70", "\x99\x71", "\x99\x72", "\x99\x73",
  /* U+EA04 */	"\x99\x74", "\x99\x75", "\x99\x76", "\x99\x77",
  /* U+EA08 */	"\x99\x78", "\x99\x79", "\x99\x7A", "\x99\x7B",
  /* U+EA0C */	"\x99\x7C", "\x99\x7D", "\x99\x7E", "\x99\xA1",
  /* U+EA10 */	"\x99\xA2", "\x99\xA3", "\x99\xA4", "\x99\xA5",
  /* U+EA14 */	"\x99\xA6", "\x99\xA7", "\x99\xA8", "\x99\xA9",
  /* U+EA18 */	"\x99\xAA", "\x99\xAB", "\x99\xAC", "\x99\xAD",
  /* U+EA1C */	"\x99\xAE", "\x99\xAF", "\x99\xB0", "\x99\xB1",
  /* U+EA20 */	"\x99\xB2", "\x99\xB3", "\x99\xB4", "\x99\xB5",
  /* U+EA24 */	"\x99\xB6", "\x99\xB7", "\x99\xB8", "\x99\xB9",
  /* U+EA28 */	"\x99\xBA", "\x99\xBB", "\x99\xBC", "\x99\xBD",
  /* U+EA2C */	"\x99\xBE", "\x99\xBF", "\x99\xC0", "\x99\xC1",
  /* U+EA30 */	"\x99\xC2", "\x99\xC3", "\x99\xC4", "\x99\xC5",
  /* U+EA34 */	"\x99\xC6", "\x99\xC7", "\x99\xC8", "\x99\xC9",
  /* U+EA38 */	"\x99\xCA", "\x99\xCB", "\x99\xCC", "\x99\xCD",
  /* U+EA3C */	"\x99\xCE", "\x99\xCF", "\x99\xD0", "\x99\xD1",
  /* U+EA40 */	"\x99\xD2", "\x99\xD3", "\x99\xD4", "\x99\xD5",
  /* U+EA44 */	"\x99\xD6", "\x99\xD7", "\x99\xD8", "\x99\xD9",
  /* U+EA48 */	"\x99\xDA", "\x99\xDB", "\x99\xDC", "\x99\xDD",
  /* U+EA4C */	"\x99\xDE", "\x99\xDF", "\x99\xE0", "\x99\xE1",
  /* U+EA50 */	"\x99\xE2", "\x99\xE3", "\x99\xE4", "\x99\xE5",
  /* U+EA54 */	"\x99\xE6", "\x99\xE7", "\x99\xE8", "\x99\xE9",
  /* U+EA58 */	"\x99\xEA", "\x99\xEB", "\x99\xEC", "\x99\xED",
  /* U+EA5C */	"\x99\xEE", "\x99\xEF", "\x99\xF0", "\x99\xF1",
  /* U+EA60 */	"\x99\xF2", "\x99\xF3", "\x99\xF4", "\x99\xF5",
  /* U+EA64 */	"\x99\xF6", "\x99\xF7", "\x99\xF8", "\x99\xF9",
  /* U+EA68 */	"\x99\xFA", "\x99\xFB", "\x99\xFC", "\x99\xFD",
  /* U+EA6C */	"\x99\xFE", "\x9A\x40", "\x9A\x41", "\x9A\x42",
  /* U+EA70 */	"\x9A\x43", "\x9A\x44", "\x9A\x45", "\x9A\x46",
  /* U+EA74 */	"\x9A\x47", "\x9A\x48", "\x9A\x49", "\x9A\x4A",
  /* U+EA78 */	"\x9A\x4B", "\x9A\x4C", "\x9A\x4D", "\x9A\x4E",
  /* U+EA7C */	"\x9A\x4F", "\x9A\x50", "\x9A\x51", "\x9A\x52",
  /* U+EA80 */	"\x9A\x53", "\x9A\x54", "\x9A\x55", "\x9A\x56",
  /* U+EA84 */	"\x9A\x57", "\x9A\x58", "\x9A\x59", "\x9A\x5A",
  /* U+EA88 */	"\x9A\x5B", "\x9A\x5C", "\x9A\x5D", "\x9A\x5E",
  /* U+EA8C */	"\x9A\x5F", "\x9A\x60", "\x9A\x61", "\x9A\x62",
  /* U+EA90 */	"\x9A\x63", "\x9A\x64", "\x9A\x65", "\x9A\x66",
  /* U+EA94 */	"\x9A\x67", "\x9A\x68", "\x9A\x69", "\x9A\x6A",
  /* U+EA98 */	"\x9A\x6B", "\x9A\x6C", "\x9A\x6D", "\x9A\x6E",
  /* U+EA9C */	"\x9A\x6F", "\x9A\x70", "\x9A\x71", "\x9A\x72",
  /* U+EAA0 */	"\x9A\x73", "\x9A\x74", "\x9A\x75", "\x9A\x76",
  /* U+EAA4 */	"\x9A\x77", "\x9A\x78", "\x9A\x79", "\x9A\x7A",
  /* U+EAA8 */	"\x9A\x7B", "\x9A\x7C", "\x9A\x7D", "\x9A\x7E",
  /* U+EAAC */	"\x9A\xA1", "\x9A\xA2", "\x9A\xA3", "\x9A\xA4",
  /* U+EAB0 */	"\x9A\xA5", "\x9A\xA6", "\x9A\xA7", "\x9A\xA8",
  /* U+EAB4 */	"\x9A\xA9", "\x9A\xAA", "\x9A\xAB", "\x9A\xAC",
  /* U+EAB8 */	"\x9A\xAD", "\x9A\xAE", "\x9A\xAF", "\x9A\xB0",
  /* U+EABC */	"\x9A\xB1", "\x9A\xB2", "\x9A\xB3", "\x9A\xB4",
  /* U+EAC0 */	"\x9A\xB5", "\x9A\xB6", "\x9A\xB7", "\x9A\xB8",
  /* U+EAC4 */	"\x9A\xB9", "\x9A\xBA", "\x9A\xBB", "\x9A\xBC",
  /* U+EAC8 */	"\x9A\xBD", "\x9A\xBE", "\x9A\xBF", "\x9A\xC0",
  /* U+EACC */	"\x9A\xC1", "\x9A\xC2", "\x9A\xC3", "\x9A\xC4",
  /* U+EAD0 */	"\x9A\xC5", "\x9A\xC6", "\x9A\xC7", "\x9A\xC8",
  /* U+EAD4 */	"\x9A\xC9", "\x9A\xCA", "\x9A\xCB", "\x9A\xCC",
  /* U+EAD8 */	"\x9A\xCD", "\x9A\xCE", "\x9A\xCF", "\x9A\xD0",
  /* U+EADC */	"\x9A\xD1", "\x9A\xD2", "\x9A\xD3", "\x9A\xD4",
  /* U+EAE0 */	"\x9A\xD5", "\x9A\xD6", "\x9A\xD7", "\x9A\xD8",
  /* U+EAE4 */	"\x9A\xD9", "\x9A\xDA", "\x9A\xDB", "\x9A\xDC",
  /* U+EAE8 */	"\x9A\xDD", "\x9A\xDE", "\x9A\xDF", "\x9A\xE0",
  /* U+EAEC */	"\x9A\xE1", "\x9A\xE2", "\x9A\xE3", "\x9A\xE4",
  /* U+EAF0 */	"\x9A\xE5", "\x9A\xE6", "\x9A\xE7", "\x9A\xE8",
  /* U+EAF4 */	"\x9A\xE9", "\x9A\xEA", "\x9A\xEB", "\x9A\xEC",
  /* U+EAF8 */	"\x9A\xED", "\x9A\xEE", "\x9A\xEF", "\x9A\xF0",
  /* U+EAFC */	"\x9A\xF1", "\x9A\xF2", "\x9A\xF3", "\x9A\xF4",
  /* U+EB00 */	"\x9A\xF5", "\x9A\xF6", "\x9A\xF7", "\x9A\xF8",
  /* U+EB04 */	"\x9A\xF9", "\x9A\xFA", "\x9A\xFB", "\x9A\xFC",
  /* U+EB08 */	"\x9A\xFD", "\x9A\xFE", "\x9B\x40", "\x9B\x41",
  /* U+EB0C */	"\x9B\x42", "\x9B\x43", "\x9B\x44", "\x9B\x45",
  /* U+EB10 */	"\x9B\x46", "\x9B\x47", "\x9B\x48", "\x9B\x49",
  /* U+EB14 */	"\x9B\x4A", "\x9B\x4B", "\x9B\x4C", "\x9B\x4D",
  /* U+EB18 */	"\x9B\x4E", "\x9B\x4F", "\x9B\x50", "\x9B\x51",
  /* U+EB1C */	"\x9B\x52", "\x9B\x53", "\x9B\x54", "\x9B\x55",
  /* U+EB20 */	"\x9B\x56", "\x9B\x57", "\x9B\x58", "\x9B\x59",
  /* U+EB24 */	"\x9B\x5A", "\x9B\x5B", "\x9B\x5C", "\x9B\x5D",
  /* U+EB28 */	"\x9B\x5E", "\x9B\x5F", "\x9B\x60", "\x9B\x61",
  /* U+EB2C */	"\x9B\x62", "\x9B\x63", "\x9B\x64", "\x9B\x65",
  /* U+EB30 */	"\x9B\x66", "\x9B\x67", "\x9B\x68", "\x9B\x69",
  /* U+EB34 */	"\x9B\x6A", "\x9B\x6B", "\x9B\x6C", "\x9B\x6D",
  /* U+EB38 */	"\x9B\x6E", "\x9B\x6F", "\x9B\x70", "\x9B\x71",
  /* U+EB3C */	"\x9B\x72", "\x9B\x73", "\x9B\x74", "\x9B\x75",
  /* U+EB40 */	"\xEF\xF9", "\x9B\x77", "\xC5\xF7", "\x9B\x79",
  /* U+EB44 */	"\x9B\x7A", "\xF5\xE8", "\x9B\x7C", "\x9B\x7D",
  /* U+EB48 */	"\x9B\x7E", "\x9B\xA1", "\x9B\xA2", "\x9B\xA3",
  /* U+EB4C */	"\x9B\xA4", "\x9B\xA5", "\x9B\xA6", "\x9B\xA7",
  /* U+EB50 */	"\x9B\xA8", "\x9B\xA9", "\x9B\xAA", "\x9B\xAB",
  /* U+EB54 */	"\x9B\xAC", "\x9B\xAD", "\x9B\xAE", "\x9B\xAF",
  /* U+EB58 */	"\x9B\xB0", "\x9B\xB1", "\x9B\xB2", "\x9B\xB3",
  /* U+EB5C */	"\x9B\xB4", "\x9B\xB5", "\x9B\xB6", "\x9B\xB7",
  /* U+EB60 */	"\x9B\xB8", "\x9B\xB9", "\x9B\xBA", "\x9B\xBB",
  /* U+EB64 */	"\x9B\xBC", "\x9B\xBD", "\x9B\xBE", "\x9B\xBF",
  /* U+EB68 */	"\x9B\xC0", "\x9B\xC1", "\x9B\xC2", "\x9B\xC3",
  /* U+EB6C */	"\x9B\xC4", "\x9B\xC5", "\xE8\xCD", "\x9B\xC7",
  /* U+EB70 */	"\x9B\xC8", "\x9B\xC9", "\x9B\xCA", "\x9B\xCB",
  /* U+EB74 */	"\x9B\xCC", "\x9B\xCD", "\x9B\xCE", "\x9B\xCF",
  /* U+EB78 */	"\x9B\xD0", "\x9B\xD1", "\x9B\xD2", "\x9B\xD3",
  /* U+EB7C */	"\x9B\xD4", "\x9B\xD5", "\x9B\xD6", "\x9B\xD7",
  /* U+EB80 */	"\x9B\xD8", "\x9B\xD9", "\x9B\xDA", "\x9B\xDB",
  /* U+EB84 */	"\x9B\xDC", "\x9B\xDD", "\xD0\xC0", "\x9B\xDF",
  /* U+EB88 */	"\x9B\xE0", "\x9B\xE1", "\x9B\xE2", "\x9B\xE3",
  /* U+EB8C */	"\x9B\xE4", "\x9B\xE5", "\x9B\xE6", "\x9B\xE7",
  /* U+EB90 */	"\x9B\xE8", "\x9B\xE9", "\x9B\xEA", "\x9B\xEB",
  /* U+EB94 */	"\xFD\x64", "\x9B\xED", "\x9B\xEE", "\x9B\xEF",
  /* U+EB98 */	"\x9B\xF0", "\x9B\xF1", "\x9B\xF2", "\x9B\xF3",
  /* U+EB9C */	"\x9B\xF4", "\x9B\xF5", "\xBF\x47", "\x9B\xF7",
  /* U+EBA0 */	"\x9B\xF8", "\x9B\xF9", "\x9B\xFA", "\x9B\xFB",
  /* U+EBA4 */	"\x9B\xFC", "\x9B\xFD", "\x9B\xFE", "\x9C\x40",
  /* U+EBA8 */	"\x9C\x41", "\xEB\xC9", "\x9C\x43", "\x9C\x44",
  /* U+EBAC */	"\x9C\x45", "\x9C\x46", "\x9C\x47", "\x9C\x48",
  /* U+EBB0 */	"\x9C\x49", "\x9C\x4A", "\x9C\x4B", "\x9C\x4C",
  /* U+EBB4 */	"\x9C\x4D", "\x9C\x4E", "\x9C\x4F", "\x9C\x50",
  /* U+EBB8 */	"\x9C\x51", "\x9C\x52", "\xCD\xE7", "\x9C\x54",
  /* U+EBBC */	"\x9C\x55", "\x9C\x56", "\x9C\x57", "\x9C\x58",
  /* U+EBC0 */	"\x9C\x59", "\x9C\x5A", "\x9C\x5B", "\x9C\x5C",
  /* U+EBC4 */	"\x9C\x5D", "\x9C\x5E", "\x9C\x5F", "\x9C\x60",
  /* U+EBC8 */	"\x9C\x61", "\xC0\xE7", "\x9C\x63", "\x9C\x64",
  /* U+EBCC */	"\x9C\x65", "\x9C\x66", "\x9C\x67", "\xDC\x52",
  /* U+EBD0 */	"\x9C\x69", "\x9C\x6A", "\xF8\x6D", "\x9C\x6C",
  /* U+EBD4 */	"\x9C\x6D", "\x9C\x6E", "\x9C\x6F", "\x9C\x70",
  /* U+EBD8 */	"\x9C\x71", "\x9C\x72", "\x9C\x73", "\x9C\x74",
  /* U+EBDC */	"\x9C\x75", "\x9C\x76", "\xDB\x5D", "\x9C\x78",
  /* U+EBE0 */	"\x9C\x79", "\x9C\x7A", "\x9C\x7B", "\x9C\x7C",
  /* U+EBE4 */	"\x9C\x7D", "\x9C\x7E", "\x9C\xA1", "\x9C\xA2",
  /* U+EBE8 */	"\x9C\xA3", "\x9C\xA4", "\x9C\xA5", "\x9C\xA6",
  /* U+EBEC */	"\x9C\xA7", "\x9C\xA8", "\x9C\xA9", "\x9C\xAA",
  /* U+EBF0 */	"\x9C\xAB", "\x9C\xAC", "\x9C\xAD", "\x9C\xAE",
  /* U+EBF4 */	"\x9C\xAF", "\x9C\xB0", "\x9C\xB1", "\x9C\xB2",
  /* U+EBF8 */	"\x9C\xB3", "\x9C\xB4", "\x9C\xB5", "\x9C\xB6",
  /* U+EBFC */	"\x9C\xB7", "\x9C\xB8", "\x9C\xB9", "\x9C\xBA",
  /* U+EC00 */	"\x9C\xBB", "\xC9\x5C", "\xAF\xB0", "\x9C\xBE",
  /* U+EC04 */	"\x9C\xBF", "\x9C\xC0", "\x9C\xC1", "\x9C\xC2",
  /* U+EC08 */	"\x9C\xC3", "\x9C\xC4", "\x9C\xC5", "\x9C\xC6",
  /* U+EC0C */	"\x9C\xC7", "\x9C\xC8", "\x9C\xC9", "\x9C\xCA",
  /* U+EC10 */	"\x9C\xCB", "\x9C\xCC", "\x9C\xCD", "\x9C\xCE",
  /* U+EC14 */	"\x9C\xCF", "\xD4\xD1", "\x9C\xD1", "\x9C\xD2",
  /* U+EC18 */	"\x9C\xD3", "\x9C\xD4", "\x9C\xD5", "\x9C\xD6",
  /* U+EC1C */	"\x9C\xD7", "\x9C\xD8", "\x9C\xD9", "\x9C\xDA",
  /* U+EC20 */	"\x9C\xDB", "\x9C\xDC", "\x9C\xDD", "\x9C\xDE",
  /* U+EC24 */	"\x9C\xDF", "\x9C\xE0", "\x9C\xE1", "\x9C\xE2",
  /* U+EC28 */	"\x9C\xE3", "\x9C\xE4", "\x9C\xE5", "\x9C\xE6",
  /* U+EC2C */	"\x9C\xE7", "\x9C\xE8", "\x9C\xE9", "\x9C\xEA",
  /* U+EC30 */	"\x9C\xEB", "\x9C\xEC", "\x9C\xED", "\x9C\xEE",
  /* U+EC34 */	"\x9C\xEF", "\x9C\xF0", "\x9C\xF1", "\x9C\xF2",
  /* U+EC38 */	"\x9C\xF3", "\x9C\xF4", "\x9C\xF5", "\x9C\xF6",
  /* U+EC3C */	"\x9C\xF7", "\x9C\xF8", "\x9C\xF9", "\x9C\xFA",
  /* U+EC40 */	"\x9C\xFB", "\x9C\xFC", "\x9C\xFD", "\x9C\xFE",
  /* U+EC44 */	"\x9D\x40", "\x9D\x41", "\x9D\x42", "\x9D\x43",
  /* U+EC48 */	"\x9D\x44", "\x9D\x45", "\x9D\x46", "\x9D\x47",
  /* U+EC4C */	"\x9D\x48", "\x9D\x49", "\x9D\x4A", "\x9D\x4B",
  /* U+EC50 */	"\x9D\x4C", "\x9D\x4D", "\x9D\x4E", "\x9D\x4F",
  /* U+EC54 */	"\x9D\x50", "\x9D\x51", "\x9D\x52", "\x9D\x53",
  /* U+EC58 */	"\x9D\x54", "\x9D\x55", "\x9D\x56", "\xE0\x7C",
  /* U+EC5C */	"\x9D\x58", "\x9D\x59", "\xB5\xAE", "\x9D\x5B",
  /* U+EC60 */	"\x9D\x5C", "\x9D\x5D", "\x9D\x5E", "\x9D\x5F",
  /* U+EC64 */	"\x9D\x60", "\x9D\x61", "\x9D\x62", "\x9D\x63",
  /* U+EC68 */	"\x9D\x64", "\x9D\x65", "\x9D\x66", "\x9D\x67",
  /* U+EC6C */	"\x9D\x68", "\x9D\x69", "\x9D\x6A", "\x9D\x6B",
  /* U+EC70 */	"\x9D\x6C", "\x9D\x6D", "\x9D\x6E", "\x9D\x6F",
  /* U+EC74 */	"\x9D\x70", "\x9D\x71", "\x9D\x72", "\x9D\x73",
  /* U+EC78 */	"\x9D\x74", "\x9D\x75", "\x9D\x76", "\x9D\x77",
  /* U+EC7C */	"\x9D\x78", "\x9D\x79", "\x9D\x7A", "\x9D\x7B",
  /* U+EC80 */	"\x9D\x7C", "\x9D\x7D", "\x9D\x7E", "\x9D\xA1",
  /* U+EC84 */	"\x9D\xA2", "\x9D\xA3", "\x9D\xA4", "\x9D\xA5",
  /* U+EC88 */	"\x9D\xA6", "\x9D\xA7", "\x9D\xA8", "\x9D\xA9",
  /* U+EC8C */	"\x9D\xAA", "\x9D\xAB", "\x9D\xAC", "\x9D\xAD",
  /* U+EC90 */	"\x9D\xAE", "\x9D\xAF", "\x9D\xB0", "\x9D\xB1",
  /* U+EC94 */	"\x9D\xB2", "\x9D\xB3", "\x9D\xB4", "\x9D\xB5",
  /* U+EC98 */	"\x9D\xB6", "\x9D\xB7", "\x9D\xB8", "\x9D\xB9",
  /* U+EC9C */	"\x9D\xBA", "\x9D\xBB", "\x9D\xBC", "\x9D\xBD",
  /* U+ECA0 */	"\x9D\xBE", "\x9D\xBF", "\x9D\xC0", "\x9D\xC1",
  /* U+ECA4 */	"\x9D\xC2", "\x9D\xC3", "\xA9\xE4", "\x9D\xC5",
  /* U+ECA8 */	"\x9D\xC6", "\x9D\xC7", "\x9D\xC8", "\x9D\xC9",
  /* U+ECAC */	"\x9D\xCA", "\x9D\xCB", "\x9D\xCC", "\x9D\xCD",
  /* U+ECB0 */	"\x9D\xCE", "\x9D\xCF", "\x9D\xD0", "\x9D\xD1",
  /* U+ECB4 */	"\x9D\xD2", "\x9D\xD3", "\x9D\xD4", "\x9D\xD5",
  /* U+ECB8 */	"\x9D\xD6", "\x9D\xD7", "\x9D\xD8", "\x9D\xD9",
  /* U+ECBC */	"\x9D\xDA", "\x9D\xDB", "\x9D\xDC", "\x9D\xDD",
  /* U+ECC0 */	"\x9D\xDE", "\x9D\xDF", "\x9D\xE0", "\x9D\xE1",
  /* U+ECC4 */	"\x9D\xE2", "\x9D\xE3", "\x9D\xE4", "\x9D\xE5",
  /* U+ECC8 */	"\x9D\xE6", "\x9D\xE7", "\x9D\xE8", "\x9D\xE9",
  /* U+ECCC */	"\x9D\xEA", "\x9D\xEB", "\x9D\xEC", "\x9D\xED",
  /* U+ECD0 */	"\x9D\xEE", "\x9D\xEF", "\x9D\xF0", "\x9D\xF1",
  /* U+ECD4 */	"\x9D\xF2", "\x9D\xF3", "\x9D\xF4", "\x9D\xF5",
  /* U+ECD8 */	"\x9D\xF6", "\x9D\xF7", "\x9D\xF8", "\x9D\xF9",
  /* U+ECDC */	"\x9D\xFA", "\x9D\xFB", "\x9D\xFC", "\x9D\xFD",
  /* U+ECE0 */	"\x9D\xFE", "\x9E\x40", "\x9E\x41", "\x9E\x42",
  /* U+ECE4 */	"\x9E\x43", "\x9E\x44", "\x9E\x45", "\x9E\x46",
  /* U+ECE8 */	"\x9E\x47", "\x9E\x48", "\x9E\x49", "\x9E\x4A",
  /* U+ECEC */	"\x9E\x4B", "\x9E\x4C", "\x9E\x4D", "\x9E\x4E",
  /* U+ECF0 */	"\x9E\x4F", "\x9E\x50", "\x9E\x51", "\x9E\x52",
  /* U+ECF4 */	"\x9E\x53", "\x9E\x54", "\x9E\x55", "\x9E\x56",
  /* U+ECF8 */	"\x9E\x57", "\x9E\x58", "\x9E\x59", "\x9E\x5A",
  /* U+ECFC */	"\x9E\x5B", "\x9E\x5C", "\x9E\x5D", "\x9E\x5E",
  /* U+ED00 */	"\x9E\x5F", "\x9E\x60", "\x9E\x61", "\x9E\x62",
  /* U+ED04 */	"\x9E\x63", "\x9E\x64", "\x9E\x65", "\x9E\x66",
  /* U+ED08 */	"\x9E\x67", "\x9E\x68", "\x9E\x69", "\x9E\x6A",
  /* U+ED0C */	"\x9E\x6B", "\x9E\x6C", "\x9E\x6D", "\x9E\x6E",
  /* U+ED10 */	"\x9E\x6F", "\x9E\x70", "\x9E\x71", "\x9E\x72",
  /* U+ED14 */	"\x9E\x73", "\x9E\x74", "\x9E\x75", "\x9E\x76",
  /* U+ED18 */	"\x9E\x77", "\x9E\x78", "\x9E\x79", "\x9E\x7A",
  /* U+ED1C */	"\x9E\x7B", "\x9E\x7C", "\x9E\x7D", "\x9E\x7E",
  /* U+ED20 */	"\x9E\xA1", "\x9E\xA2", "\x9E\xA3", "\x9E\xA4",
  /* U+ED24 */	"\x9E\xA5", "\x9E\xA6", "\x9E\xA7", "\x9E\xA8",
  /* U+ED28 */	"\xAB\xEC", "\x9E\xAA", "\x9E\xAB", "\x9E\xAC",
  /* U+ED2C */	"\x9E\xAD", "\x9E\xAE", "\x9E\xAF", "\x9E\xB0",
  /* U+ED30 */	"\x9E\xB1", "\x9E\xB2", "\x9E\xB3", "\x9E\xB4",
  /* U+ED34 */	"\x9E\xB5", "\x9E\xB6", "\x9E\xB7", "\x9E\xB8",
  /* U+ED38 */	"\x9E\xB9", "\x9E\xBA", "\x9E\xBB", "\x9E\xBC",
  /* U+ED3C */	"\x9E\xBD", "\x9E\xBE", "\x9E\xBF", "\x9E\xC0",
  /* U+ED40 */	"\x9E\xC1", "\x9E\xC2", "\x9E\xC3", "\x9E\xC4",
  /* U+ED44 */	"\x9E\xC5", "\x9E\xC6", "\x9E\xC7", "\x9E\xC8",
  /* U+ED48 */	"\x9E\xC9", "\x9E\xCA", "\x9E\xCB", "\x9E\xCC",
  /* U+ED4C */	"\x9E\xCD", "\x9E\xCE", "\x9E\xCF", "\x9E\xD0",
  /* U+ED50 */	"\x9E\xD1", "\x9E\xD2", "\x9E\xD3", "\x9E\xD4",
  /* U+ED54 */	"\x9E\xD5", "\x9E\xD6", "\x9E\xD7", "\x9E\xD8",
  /* U+ED58 */	"\x9E\xD9", "\x9E\xDA", "\x9E\xDB", "\x9E\xDC",
  /* U+ED5C */	"\x9E\xDD", "\x9E\xDE", "\x9E\xDF", "\x9E\xE0",
  /* U+ED60 */	"\x9E\xE1", "\x9E\xE2", "\x9E\xE3", "\x9E\xE4",
  /* U+ED64 */	"\x9E\xE5", "\x9E\xE6", "\x9E\xE7", "\x9E\xE8",
  /* U+ED68 */	"\x9E\xE9", "\x9E\xEA", "\x9E\xEB", "\x9E\xEC",
  /* U+ED6C */	"\x9E\xED", "\x9E\xEE", "\xDE\xCD", "\x9E\xF0",
  /* U+ED70 */	"\x9E\xF1", "\x9E\xF2", "\x9E\xF3", "\x9E\xF4",
  /* U+ED74 */	"\x9E\xF5", "\x9E\xF6", "\x9E\xF7", "\x9E\xF8",
  /* U+ED78 */	"\x9E\xF9", "\x9E\xFA", "\x9E\xFB", "\x9E\xFC",
  /* U+ED7C */	"\xC9\xFC", "\x9E\xFE", "\x9F\x40", "\x9F\x41",
  /* U+ED80 */	"\x9F\x42", "\x9F\x43", "\x9F\x44", "\x9F\x45",
  /* U+ED84 */	"\x9F\x46", "\x9F\x47", "\x9F\x48", "\x9F\x49",
  /* U+ED88 */	"\x9F\x4A", "\x9F\x4B", "\x9F\x4C", "\x9F\x4D",
  /* U+ED8C */	"\x9F\x4E", "\x9F\x4F", "\x9F\x50", "\x9F\x51",
  /* U+ED90 */	"\x9F\x52", "\x9F\x53", "\x9F\x54", "\x9F\x55",
  /* U+ED94 */	"\x9F\x56", "\x9F\x57", "\x9F\x58", "\x9F\x59",
  /* U+ED98 */	"\x9F\x5A", "\x9F\x5B", "\x9F\x5C", "\x9F\x5D",
  /* U+ED9C */	"\x9F\x5E", "\x9F\x5F", "\xF9\xC4", "\x9F\x61",
  /* U+EDA0 */	"\x9F\x62", "\x9F\x63", "\x9F\x64", "\x9F\x65",
  /* U+EDA4 */	"\x91\xBE", "\x9F\x67", "\x9F\x68", "\x9F\x69",
  /* U+EDA8 */	"\x9F\x6A", "\x9F\x6B", "\x9F\x6C", "\x9F\x6D",
  /* U+EDAC */	"\x9F\x6E", "\x9F\x6F", "\x9F\x70", "\x9F\x71",
  /* U+EDB0 */	"\x9F\x72", "\x9F\x73", "\x9F\x74", "\x9F\x75",
  /* U+EDB4 */	"\x9F\x76", "\x9F\x77", "\x9F\x78", "\x9F\x79",
  /* U+EDB8 */	"\x9F\x7A", "\x9F\x7B", "\x9F\x7C", "\x9F\x7D",
  /* U+EDBC */	"\x9F\x7E", "\x9F\xA1", "\x9F\xA2", "\x9F\xA3",
  /* U+EDC0 */	"\x9F\xA4", "\x9F\xA5", "\x9F\xA6", "\x9F\xA7",
  /* U+EDC4 */	"\x9F\xA8", "\x9F\xA9", "\x9F\xAA", "\x9F\xAB",
  /* U+EDC8 */	"\x9F\xAC", "\x9F\xAD", "\x9F\xAE", "\x9F\xAF",
  /* U+EDCC */	"\x9F\xB0", "\x9F\xB1", "\x9F\xB2", "\x9F\xB3",
  /* U+EDD0 */	"\x9F\xB4", "\x9F\xB5", "\x9F\xB6", "\x9F\xB7",
  /* U+EDD4 */	"\x9F\xB8", "\x9F\xB9", "\x9F\xBA", "\x9F\xBB",
  /* U+EDD8 */	"\x9F\xBC", "\x9F\xBD", "\x9F\xBE", "\x9F\xBF",
  /* U+EDDC */	"\x9F\xC0", "\x9F\xC1", "\x9F\xC2", "\x9F\xC3",
  /* U+EDE0 */	"\x9F\xC4", "\x9F\xC5", "\x9F\xC6", "\x9F\xC7",
  /* U+EDE4 */	"\x9F\xC8", "\x9F\xC9", "\x9F\xCA", "\xB9\xB0",
  /* U+EDE8 */	"\x9F\xCC", "\x9F\xCD", "\x9F\xCE", "\x9F\xCF",
  /* U+EDEC */	"\x9F\xD0", "\x9F\xD1", "\x9F\xD2", "\x9F\xD3",
  /* U+EDF0 */	"\x9F\xD4", "\x9F\xD5", "\x9F\xD6", "\x9F\xD7",
  /* U+EDF4 */	"\x93\x61", "\x9F\xD9", "\x9F\xDA", "\x9F\xDB",
  /* U+EDF8 */	"\x9F\xDC", "\x9F\xDD", "\x9F\xDE", "\x9F\xDF",
  /* U+EDFC */	"\x9F\xE0", "\x9F\xE1", "\x9F\xE2", "\x9F\xE3",
  /* U+EE00 */	"\x9F\xE4", "\x9F\xE5", "\x9F\xE6", "\x9F\xE7",
  /* U+EE04 */	"\x9F\xE8", "\x9F\xE9", "\x9F\xEA", "\x9F\xEB",
  /* U+EE08 */	"\x9F\xEC", "\x9F\xED", "\x9F\xEE", "\x9F\xEF",
  /* U+EE0C */	"\x9F\xF0", "\x9F\xF1", "\x9F\xF2", "\x9F\xF3",
  /* U+EE10 */	"\x9F\xF4", "\x9F\xF5", "\x9F\xF6", "\x9F\xF7",
  /* U+EE14 */	"\x9F\xF8", "\x9F\xF9", "\x9F\xFA", "\x9F\xFB",
  /* U+EE18 */	"\x9F\xFC", "\x9F\xFD", "\x9F\xFE", "\xA0\x40",
  /* U+EE1C */	"\xA0\x41", "\xA0\x42", "\xA0\x43", "\xA0\x44",
  /* U+EE20 */	"\xA0\x45", "\xA0\x46", "\xA0\x47", "\xA0\x48",
  /* U+EE24 */	"\xA0\x49", "\xA0\x4A", "\xA0\x4B", "\xA0\x4C",
  /* U+EE28 */	"\xA0\x4D", "\xA0\x4E", "\xA0\x4F", "\xA0\x50",
  /* U+EE2C */	"\xA0\x51", "\xA0\x52", "\xA0\x53", "\xA0\x54",
  /* U+EE30 */	"\xA0\x55", "\xA0\x56", "\xA0\x57", "\xA0\x58",
  /* U+EE34 */	"\xA0\x59", "\xA0\x5A", "\xA0\x5B", "\xA0\x5C",
  /* U+EE38 */	"\xA0\x5D", "\xA0\x5E", "\xA0\x5F", "\xA0\x60",
  /* U+EE3C */	"\xA0\x61", "\xA0\x62", "\x8F\xB6", "\xA0\x64",
  /* U+EE40 */	"\xA0\x65", "\xA0\x66", "\xA0\x67", "\xA0\x68",
  /* U+EE44 */	"\xA0\x69", "\xA0\x6A", "\xA0\x6B", "\xA0\x6C",
  /* U+EE48 */	"\xA0\x6D", "\xA0\x6E", "\xA0\x6F", "\xA0\x70",
  /* U+EE4C */	"\xA0\x71", "\xA0\x72", "\xA0\x73", "\xA0\x74",
  /* U+EE50 */	"\xA0\x75", "\xA0\x76", "\xA9\xF0", "\xA0\x78",
  /* U+EE54 */	"\xA0\x79", "\xA0\x7A", "\xA0\x7B", "\xA0\x7C",
  /* U+EE58 */	"\xA0\x7D", "\xA0\x7E", "\xA0\xA1", "\xA0\xA2",
  /* U+EE5C */	"\xA0\xA3", "\xA0\xA4", "\xA0\xA5", "\xA0\xA6",
  /* U+EE60 */	"\xA0\xA7", "\xA0\xA8", "\xA0\xA9", "\xA0\xAA",
  /* U+EE64 */	"\xA0\xAB", "\xA0\xAC", "\xA0\xAD", "\xA0\xAE",
  /* U+EE68 */	"\xA0\xAF", "\xA0\xB0", "\xA0\xB1", "\xA0\xB2",
  /* U+EE6C */	"\xA0\xB3", "\xA0\xB4", "\xA0\xB5", "\xA0\xB6",
  /* U+EE70 */	"\xA0\xB7", "\xA0\xB8", "\xA0\xB9", "\xA0\xBA",
  /* U+EE74 */	"\xA0\xBB", "\xA0\xBC", "\xA0\xBD", "\xA0\xBE",
  /* U+EE78 */	"\xA0\xBF", "\xA0\xC0", "\xA0\xC1", "\xA0\xC2",
  /* U+EE7C */	"\xA0\xC3", "\xA0\xC4", "\xA0\xC5", "\xA0\xC6",
  /* U+EE80 */	"\xA0\xC7", "\xA0\xC8", "\xA0\xC9", "\xA0\xCA",
  /* U+EE84 */	"\xA0\xCB", "\xA0\xCC", "\xA0\xCD", "\xA0\xCE",
  /* U+EE88 */	"\xA0\xCF", "\xA0\xD0", "\xA0\xD1", "\xA0\xD2",
  /* U+EE8C */	"\xA0\xD3", "\xA0\xD4", "\x94\x7A", "\xA0\xD6",
  /* U+EE90 */	"\xA0\xD7", "\xA0\xD8", "\xA0\xD9", "\xA0\xDA",
  /* U+EE94 */	"\xA0\xDB", "\xA0\xDC", "\xA0\xDD", "\xA0\xDE",
  /* U+EE98 */	"\xDE\x72", "\xA0\xE0", "\xA0\xE1", "\xA0\xE2",
  /* U+EE9C */	"\xA0\xE3", "\x94\x55", "\xA0\xE5", "\xA0\xE6",
  /* U+EEA0 */	"\xA0\xE7", "\xA0\xE8", "\xA0\xE9", "\xA0\xEA",
  /* U+EEA4 */	"\xA0\xEB", "\xA0\xEC", "\xA0\xED", "\xA0\xEE",
  /* U+EEA8 */	"\xA0\xEF", "\xA0\xF0", "\xA0\xF1", "\xA0\xF2",
  /* U+EEAC */	"\xA0\xF3", "\xA0\xF4", "\xA0\xF5", "\xA0\xF6",
  /* U+EEB0 */	"\xA0\xF7", "\xA0\xF8", "\xA0\xF9", "\xA0\xFA",
  /* U+EEB4 */	"\xA0\xFB", "\xA0\xFC", "\xA0\xFD", "\xA0\xFE",
  /* U+EEB8 */	"\x81\x40", "\x81\x41", "\x81\x42", "\x81\x43",
  /* U+EEBC */	"\x81\x44", "\x81\x45", "\x81\x46", "\x81\x47",
  /* U+EEC0 */	"\x81\x48", "\x81\x49", "\x81\x4A", "\x81\x4B",
  /* U+EEC4 */	"\x81\x4C", "\x81\x4D", "\x81\x4E", "\x81\x4F",
  /* U+EEC8 */	"\x81\x50", "\x81\x51", "\x81\x52", "\x81\x53",
  /* U+EECC */	"\x81\x54", "\x81\x55", "\x81\x56", "\x81\x57",
  /* U+EED0 */	"\x81\x58", "\x81\x59", "\x81\x5A", "\x81\x5B",
  /* U+EED4 */	"\x81\x5C", "\x81\x5D", "\x81\x5E", "\x81\x5F",
  /* U+EED8 */	"\x81\x60", "\x81\x61", "\x81\x62", "\x81\x63",
  /* U+EEDC */	"\x81\x64", "\x81\x65", "\x81\x66", "\x81\x67",
  /* U+EEE0 */	"\x81\x68", "\x81\x69", "\x81\x6A", "\x81\x6B",
  /* U+EEE4 */	"\x81\x6C", "\x81\x6D", "\x81\x6E", "\x81\x6F",
  /* U+EEE8 */	"\x81\x70", "\x81\x71", "\x81\x72", "\x81\x73",
  /* U+EEEC */	"\x81\x74", "\x81\x75", "\x81\x76", "\x81\x77",
  /* U+EEF0 */	"\x81\x78", "\x81\x79", "\x81\x7A", "\x81\x7B",
  /* U+EEF4 */	"\x81\x7C", "\x81\x7D", "\x81\x7E", "\x81\xA1",
  /* U+EEF8 */	"\x81\xA2", "\x81\xA3", "\x81\xA4", "\x81\xA5",
  /* U+EEFC */	"\x81\xA6", "\x81\xA7", "\x81\xA8", "\x81\xA9",
  /* U+EF00 */	"\x81\xAA", "\x81\xAB", "\x81\xAC", "\x81\xAD",
  /* U+EF04 */	"\x81\xAE", "\x81\xAF", "\x81\xB0", "\x81\xB1",
  /* U+EF08 */	"\x81\xB2", "\x81\xB3", "\x81\xB4", "\x81\xB5",
  /* U+EF0C */	"\x81\xB6", "\x81\xB7", "\x81\xB8", "\x81\xB9",
  /* U+EF10 */	"\x81\xBA", "\x81\xBB", "\x81\xBC", "\x81\xBD",
  /* U+EF14 */	"\x81\xBE", "\x81\xBF", "\x81\xC0", "\x81\xC1",
  /* U+EF18 */	"\x81\xC2", "\x81\xC3", "\x81\xC4", "\x81\xC5",
  /* U+EF1C */	"\x81\xC6", "\x81\xC7", "\x81\xC8", "\x81\xC9",
  /* U+EF20 */	"\x81\xCA", "\x81\xCB", "\x81\xCC", "\x81\xCD",
  /* U+EF24 */	"\x81\xCE", "\x81\xCF", "\x81\xD0", "\x81\xD1",
  /* U+EF28 */	"\x81\xD2", "\x81\xD3", "\x81\xD4", "\x81\xD5",
  /* U+EF2C */	"\x81\xD6", "\x81\xD7", "\x81\xD8", "\x81\xD9",
  /* U+EF30 */	"\x81\xDA", "\x81\xDB", "\x81\xDC", "\x81\xDD",
  /* U+EF34 */	"\x81\xDE", "\x81\xDF", "\x81\xE0", "\x81\xE1",
  /* U+EF38 */	"\x81\xE2", "\x81\xE3", "\x81\xE4", "\x81\xE5",
  /* U+EF3C */	"\x81\xE6", "\x81\xE7", "\x81\xE8", "\x81\xE9",
  /* U+EF40 */	"\x81\xEA", "\x81\xEB", "\x81\xEC", "\x81\xED",
  /* U+EF44 */	"\x81\xEE", "\x81\xEF", "\x81\xF0", "\x81\xF1",
  /* U+EF48 */	"\x81\xF2", "\x81\xF3", "\x81\xF4", "\x81\xF5",
  /* U+EF4C */	"\x81\xF6", "\x81\xF7", "\x81\xF8", "\x81\xF9",
  /* U+EF50 */	"\x81\xFA", "\x81\xFB", "\x81\xFC", "\x81\xFD",
  /* U+EF54 */	"\x81\xFE", "\x82\x40", "\x82\x41", "\x82\x42",
  /* U+EF58 */	"\x82\x43", "\x82\x44", "\x82\x45", "\x82\x46",
  /* U+EF5C */	"\x82\x47", "\x82\x48", "\x82\x49", "\x82\x4A",
  /* U+EF60 */	"\x82\x4B", "\x82\x4C", "\x82\x4D", "\x82\x4E",
  /* U+EF64 */	"\x82\x4F", "\x82\x50", "\x82\x51", "\x82\x52",
  /* U+EF68 */	"\x82\x53", "\x82\x54", "\x82\x55", "\x82\x56",
  /* U+EF6C */	"\x82\x57", "\x82\x58", "\x82\x59", "\x82\x5A",
  /* U+EF70 */	"\x82\x5B", "\x82\x5C", "\x82\x5D", "\x82\x5E",
  /* U+EF74 */	"\x82\x5F", "\x82\x60", "\x82\x61", "\x82\x62",
  /* U+EF78 */	"\x82\x63", "\x82\x64", "\x82\x65", "\x82\x66",
  /* U+EF7C */	"\x82\x67", "\x82\x68", "\x82\x69", "\x82\x6A",
  /* U+EF80 */	"\x82\x6B", "\x82\x6C", "\x82\x6D", "\x82\x6E",
  /* U+EF84 */	"\x82\x6F", "\x82\x70", "\x82\x71", "\x82\x72",
  /* U+EF88 */	"\x82\x73", "\x82\x74", "\x82\x75", "\x82\x76",
  /* U+EF8C */	"\x82\x77", "\x82\x78", "\x82\x79", "\x82\x7A",
  /* U+EF90 */	"\x82\x7B", "\x82\x7C", "\x82\x7D", "\x82\x7E",
  /* U+EF94 */	"\x82\xA1", "\x82\xA2", "\x82\xA3", "\x82\xA4",
  /* U+EF98 */	"\x82\xA5", "\x82\xA6", "\x82\xA7", "\x82\xA8",
  /* U+EF9C */	"\x82\xA9", "\x82\xAA", "\x82\xAB", "\x82\xAC",
  /* U+EFA0 */	"\x82\xAD", "\x82\xAE", "\x82\xAF", "\x82\xB0",
  /* U+EFA4 */	"\x82\xB1", "\x82\xB2", "\x82\xB3", "\x82\xB4",
  /* U+EFA8 */	"\x82\xB5", "\x82\xB6", "\x82\xB7", "\x82\xB8",
  /* U+EFAC */	"\x82\xB9", "\x82\xBA", "\x82\xBB", "\x82\xBC",
  /* U+EFB0 */	"\x82\xBD", "\x82\xBE", "\x82\xBF", "\x82\xC0",
  /* U+EFB4 */	"\x82\xC1", "\x82\xC2", "\x82\xC3", "\x82\xC4",
  /* U+EFB8 */	"\x82\xC5", "\x82\xC6", "\x82\xC7", "\x82\xC8",
  /* U+EFBC */	"\x82\xC9", "\x82\xCA", "\x82\xCB", "\x82\xCC",
  /* U+EFC0 */	"\x82\xCD", "\x82\xCE", "\x82\xCF", "\x82\xD0",
  /* U+EFC4 */	"\x82\xD1", "\x82\xD2", "\x82\xD3", "\x82\xD4",
  /* U+EFC8 */	"\x82\xD5", "\x82\xD6", "\x82\xD7", "\x82\xD8",
  /* U+EFCC */	"\x82\xD9", "\x82\xDA", "\x82\xDB", "\x82\xDC",
  /* U+EFD0 */	"\x82\xDD", "\x82\xDE", "\x82\xDF", "\x82\xE0",
  /* U+EFD4 */	"\x82\xE1", "\x82\xE2", "\x82\xE3", "\x82\xE4",
  /* U+EFD8 */	"\x82\xE5", "\x82\xE6", "\x82\xE7", "\x82\xE8",
  /* U+EFDC */	"\x82\xE9", "\x82\xEA", "\x82\xEB", "\x82\xEC",
  /* U+EFE0 */	"\x82\xED", "\x82\xEE", "\x82\xEF", "\x82\xF0",
  /* U+EFE4 */	"\x82\xF1", "\x82\xF2", "\x82\xF3", "\x82\xF4",
  /* U+EFE8 */	"\x82\xF5", "\x82\xF6", "\x82\xF7", "\x82\xF8",
  /* U+EFEC */	"\x82\xF9", "\x82\xFA", "\x82\xFB", "\x82\xFC",
  /* U+EFF0 */	"\x82\xFD", "\x82\xFE", "\x83\x40", "\x83\x41",
  /* U+EFF4 */	"\x83\x42", "\x83\x43", "\x83\x44", "\x83\x45",
  /* U+EFF8 */	"\x83\x46", "\x83\x47", "\x83\x48", "\x83\x49",
  /* U+EFFC */	"\x83\x4A", "\x83\x4B", "\x83\x4C", "\x83\x4D",
  /* U+F000 */	"\x83\x4E", "\x83\x4F", "\x83\x50", "\x83\x51",
  /* U+F004 */	"\x83\x52", "\x83\x53", "\x83\x54", "\x83\x55",
  /* U+F008 */	"\x83\x56", "\x83\x57", "\x83\x58", "\x83\x59",
  /* U+F00C */	"\x83\x5A", "\x83\x5B", "\x83\x5C", "\x83\x5D",
  /* U+F010 */	"\x83\x5E", "\x83\x5F", "\x83\x60", "\x83\x61",
  /* U+F014 */	"\x83\x62", "\x83\x63", "\x83\x64", "\x83\x65",
  /* U+F018 */	"\x83\x66", "\x83\x67", "\x83\x68", "\x83\x69",
  /* U+F01C */	"\x83\x6A", "\x83\x6B", "\x83\x6C", "\x83\x6D",
  /* U+F020 */	"\x83\x6E", "\x83\x6F", "\x83\x70", "\x83\x71",
  /* U+F024 */	"\x83\x72", "\x83\x73", "\x83\x74", "\x83\x75",
  /* U+F028 */	"\x83\x76", "\x83\x77", "\x83\x78", "\x83\x79",
  /* U+F02C */	"\x83\x7A", "\x83\x7B", "\x83\x7C", "\x83\x7D",
  /* U+F030 */	"\x83\x7E", "\x83\xA1", "\x83\xA2", "\x83\xA3",
  /* U+F034 */	"\x83\xA4", "\x83\xA5", "\x83\xA6", "\x83\xA7",
  /* U+F038 */	"\x83\xA8", "\x83\xA9", "\x83\xAA", "\x83\xAB",
  /* U+F03C */	"\x83\xAC", "\x83\xAD", "\x83\xAE", "\x83\xAF",
  /* U+F040 */	"\x83\xB0", "\x83\xB1", "\x83\xB2", "\x83\xB3",
  /* U+F044 */	"\x83\xB4", "\x83\xB5", "\x83\xB6", "\x83\xB7",
  /* U+F048 */	"\x83\xB8", "\x83\xB9", "\x83\xBA", "\x83\xBB",
  /* U+F04C */	"\x83\xBC", "\x83\xBD", "\x83\xBE", "\x83\xBF",
  /* U+F050 */	"\x83\xC0", "\x83\xC1", "\x83\xC2", "\x83\xC3",
  /* U+F054 */	"\x83\xC4", "\x83\xC5", "\x83\xC6", "\x83\xC7",
  /* U+F058 */	"\x83\xC8", "\x83\xC9", "\x83\xCA", "\x83\xCB",
  /* U+F05C */	"\x83\xCC", "\x83\xCD", "\x83\xCE", "\x83\xCF",
  /* U+F060 */	"\x83\xD0", "\x83\xD1", "\x83\xD2", "\x83\xD3",
  /* U+F064 */	"\x83\xD4", "\x83\xD5", "\x83\xD6", "\x83\xD7",
  /* U+F068 */	"\x83\xD8", "\x83\xD9", "\x83\xDA", "\x83\xDB",
  /* U+F06C */	"\x83\xDC", "\x83\xDD", "\x83\xDE", "\x83\xDF",
  /* U+F070 */	"\x83\xE0", "\x83\xE1", "\x83\xE2", "\x83\xE3",
  /* U+F074 */	"\x83\xE4", "\x83\xE5", "\x83\xE6", "\x83\xE7",
  /* U+F078 */	"\x83\xE8", "\x83\xE9", "\x83\xEA", "\x83\xEB",
  /* U+F07C */	"\x83\xEC", "\x83\xED", "\x83\xEE", "\x83\xEF",
  /* U+F080 */	"\x83\xF0", "\x83\xF1", "\x83\xF2", "\x83\xF3",
  /* U+F084 */	"\x83\xF4", "\x83\xF5", "\x83\xF6", "\x83\xF7",
  /* U+F088 */	"\x83\xF8", "\x83\xF9", "\x83\xFA", "\x83\xFB",
  /* U+F08C */	"\x83\xFC", "\x83\xFD", "\x83\xFE", "\x84\x40",
  /* U+F090 */	"\x84\x41", "\x84\x42", "\x84\x43", "\x84\x44",
  /* U+F094 */	"\x84\x45", "\x84\x46", "\x84\x47", "\x84\x48",
  /* U+F098 */	"\x84\x49", "\x84\x4A", "\x84\x4B", "\x84\x4C",
  /* U+F09C */	"\x84\x4D", "\x84\x4E", "\x84\x4F", "\x84\x50",
  /* U+F0A0 */	"\x84\x51", "\x84\x52", "\x84\x53", "\x84\x54",
  /* U+F0A4 */	"\x84\x55", "\x84\x56", "\x84\x57", "\x84\x58",
  /* U+F0A8 */	"\x84\x59", "\x84\x5A", "\x84\x5B", "\x84\x5C",
  /* U+F0AC */	"\x84\x5D", "\x84\x5E", "\x84\x5F", "\x84\x60",
  /* U+F0B0 */	"\x84\x61", "\x84\x62", "\x84\x63", "\x84\x64",
  /* U+F0B4 */	"\x84\x65", "\x84\x66", "\x84\x67", "\x84\x68",
  /* U+F0B8 */	"\x84\x69", "\x84\x6A", "\x84\x6B", "\x84\x6C",
  /* U+F0BC */	"\x84\x6D", "\x84\x6E", "\x84\x6F", "\x84\x70",
  /* U+F0C0 */	"\x84\x71", "\x84\x72", "\x84\x73", "\x84\x74",
  /* U+F0C4 */	"\x84\x75", "\x84\x76", "\x84\x77", "\x84\x78",
  /* U+F0C8 */	"\x84\x79", "\x84\x7A", "\x84\x7B", "\x84\x7C",
  /* U+F0CC */	"\x84\x7D", "\x84\x7E", "\x84\xA1", "\x84\xA2",
  /* U+F0D0 */	"\x84\xA3", "\x84\xA4", "\x84\xA5", "\x84\xA6",
  /* U+F0D4 */	"\x84\xA7", "\x84\xA8", "\x84\xA9", "\x84\xAA",
  /* U+F0D8 */	"\x84\xAB", "\x84\xAC", "\x84\xAD", "\x84\xAE",
  /* U+F0DC */	"\x84\xAF", "\x84\xB0", "\x84\xB1", "\x84\xB2",
  /* U+F0E0 */	"\x84\xB3", "\x84\xB4", "\x84\xB5", "\x84\xB6",
  /* U+F0E4 */	"\x84\xB7", "\x84\xB8", "\x84\xB9", "\x84\xBA",
  /* U+F0E8 */	"\x84\xBB", "\x84\xBC", "\x84\xBD", "\x84\xBE",
  /* U+F0EC */	"\x84\xBF", "\x84\xC0", "\x84\xC1", "\x84\xC2",
  /* U+F0F0 */	"\x84\xC3", "\x84\xC4", "\x84\xC5", "\x84\xC6",
  /* U+F0F4 */	"\x84\xC7", "\x84\xC8", "\x84\xC9", "\x84\xCA",
  /* U+F0F8 */	"\x84\xCB", "\x84\xCC", "\x84\xCD", "\x84\xCE",
  /* U+F0FC */	"\x84\xCF", "\x84\xD0", "\x84\xD1", "\x84\xD2",
  /* U+F100 */	"\x84\xD3", "\x84\xD4", "\x84\xD5", "\x84\xD6",
  /* U+F104 */	"\x84\xD7", "\x84\xD8", "\x84\xD9", "\x84\xDA",
  /* U+F108 */	"\x84\xDB", "\x84\xDC", "\x84\xDD", "\x84\xDE",
  /* U+F10C */	"\x84\xDF", "\x84\xE0", "\x84\xE1", "\x84\xE2",
  /* U+F110 */	"\x84\xE3", "\x84\xE4", "\x84\xE5", "\x84\xE6",
  /* U+F114 */	"\x84\xE7", "\x84\xE8", "\x84\xE9", "\x84\xEA",
  /* U+F118 */	"\x84\xEB", "\x84\xEC", "\x84\xED", "\x84\xEE",
  /* U+F11C */	"\x84\xEF", "\x84\xF0", "\x84\xF1", "\x84\xF2",
  /* U+F120 */	"\x84\xF3", "\x84\xF4", "\x84\xF5", "\x84\xF6",
  /* U+F124 */	"\x84\xF7", "\x84\xF8", "\x84\xF9", "\x84\xFA",
  /* U+F128 */	"\x84\xFB", "\x84\xFC", "\x84\xFD", "\x84\xFE",
  /* U+F12C */	"\x85\x40", "\x85\x41", "\x85\x42", "\x85\x43",
  /* U+F130 */	"\x85\x44", "\x85\x45", "\x85\x46", "\x85\x47",
  /* U+F134 */	"\x85\x48", "\x85\x49", "\x85\x4A", "\x85\x4B",
  /* U+F138 */	"\x85\x4C", "\x85\x4D", "\x85\x4E", "\x85\x4F",
  /* U+F13C */	"\x85\x50", "\x85\x51", "\x85\x52", "\x85\x53",
  /* U+F140 */	"\x85\x54", "\x85\x55", "\x85\x56", "\x85\x57",
  /* U+F144 */	"\x85\x58", "\x85\x59", "\x85\x5A", "\x85\x5B",
  /* U+F148 */	"\x85\x5C", "\x85\x5D", "\x85\x5E", "\x85\x5F",
  /* U+F14C */	"\x85\x60", "\x85\x61", "\x85\x62", "\x85\x63",
  /* U+F150 */	"\x85\x64", "\x85\x65", "\x85\x66", "\x85\x67",
  /* U+F154 */	"\x85\x68", "\x85\x69", "\x85\x6A", "\x85\x6B",
  /* U+F158 */	"\x85\x6C", "\x85\x6D", "\x85\x6E", "\x85\x6F",
  /* U+F15C */	"\x85\x70", "\x85\x71", "\x85\x72", "\x85\x73",
  /* U+F160 */	"\x85\x74", "\x85\x75", "\x85\x76", "\x85\x77",
  /* U+F164 */	"\x85\x78", "\x85\x79", "\x85\x7A", "\x85\x7B",
  /* U+F168 */	"\x85\x7C", "\x85\x7D", "\x85\x7E", "\x85\xA1",
  /* U+F16C */	"\x85\xA2", "\x85\xA3", "\x85\xA4", "\x85\xA5",
  /* U+F170 */	"\x85\xA6", "\x85\xA7", "\x85\xA8", "\x85\xA9",
  /* U+F174 */	"\x85\xAA", "\x85\xAB", "\x85\xAC", "\x85\xAD",
  /* U+F178 */	"\x85\xAE", "\x85\xAF", "\x85\xB0", "\x85\xB1",
  /* U+F17C */	"\x85\xB2", "\x85\xB3", "\x85\xB4", "\x85\xB5",
  /* U+F180 */	"\x85\xB6", "\x85\xB7", "\x85\xB8", "\x85\xB9",
  /* U+F184 */	"\x85\xBA", "\x85\xBB", "\x85\xBC", "\x85\xBD",
  /* U+F188 */	"\x85\xBE", "\x85\xBF", "\x85\xC0", "\x85\xC1",
  /* U+F18C */	"\x85\xC2", "\x85\xC3", "\x85\xC4", "\x85\xC5",
  /* U+F190 */	"\x85\xC6", "\x85\xC7", "\x85\xC8", "\x85\xC9",
  /* U+F194 */	"\x85\xCA", "\x85\xCB", "\x85\xCC", "\x85\xCD",
  /* U+F198 */	"\x85\xCE", "\x85\xCF", "\x85\xD0", "\x85\xD1",
  /* U+F19C */	"\x85\xD2", "\x85\xD3", "\x85\xD4", "\x85\xD5",
  /* U+F1A0 */	"\x85\xD6", "\x85\xD7", "\x85\xD8", "\x85\xD9",
  /* U+F1A4 */	"\x85\xDA", "\x85\xDB", "\x85\xDC", "\x85\xDD",
  /* U+F1A8 */	"\x85\xDE", "\x85\xDF", "\x85\xE0", "\x85\xE1",
  /* U+F1AC */	"\x85\xE2", "\x85\xE3", "\x85\xE4", "\x85\xE5",
  /* U+F1B0 */	"\x85\xE6", "\x85\xE7", "\x85\xE8", "\x85\xE9",
  /* U+F1B4 */	"\x85\xEA", "\x85\xEB", "\x85\xEC", "\x85\xED",
  /* U+F1B8 */	"\x85\xEE", "\x85\xEF", "\x85\xF0", "\x85\xF1",
  /* U+F1BC */	"\x85\xF2", "\x85\xF3", "\x85\xF4", "\x85\xF5",
  /* U+F1C0 */	"\x85\xF6", "\x85\xF7", "\x85\xF8", "\x85\xF9",
  /* U+F1C4 */	"\x85\xFA", "\x85\xFB", "\x85\xFC", "\x85\xFD",
  /* U+F1C8 */	"\x85\xFE", "\x86\x40", "\x86\x41", "\x86\x42",
  /* U+F1CC */	"\x86\x43", "\x86\x44", "\x86\x45", "\x86\x46",
  /* U+F1D0 */	"\x86\x47", "\x86\x48", "\x86\x49", "\x86\x4A",
  /* U+F1D4 */	"\x86\x4B", "\x86\x4C", "\x86\x4D", "\x86\x4E",
  /* U+F1D8 */	"\x86\x4F", "\x86\x50", "\x86\x51", "\x86\x52",
  /* U+F1DC */	"\x86\x53", "\x86\x54", "\x86\x55", "\x86\x56",
  /* U+F1E0 */	"\x86\x57", "\x86\x58", "\x86\x59", "\x86\x5A",
  /* U+F1E4 */	"\x86\x5B", "\x86\x5C", "\x86\x5D", "\x86\x5E",
  /* U+F1E8 */	"\x86\x5F", "\x86\x60", "\x86\x61", "\x86\x62",
  /* U+F1EC */	"\x86\x63", "\x86\x64", "\x86\x65", "\x86\x66",
  /* U+F1F0 */	"\x86\x67", "\x86\x68", "\x86\x69", "\x86\x6A",
  /* U+F1F4 */	"\x86\x6B", "\x86\x6C", "\x86\x6D", "\x86\x6E",
  /* U+F1F8 */	"\x86\x6F", "\x86\x70", "\x86\x71", "\x86\x72",
  /* U+F1FC */	"\x86\x73", "\x86\x74", "\x86\x75", "\x86\x76",
  /* U+F200 */	"\x86\x77", "\x86\x78", "\x86\x79", "\x86\x7A",
  /* U+F204 */	"\x86\x7B", "\x86\x7C", "\x86\x7D", "\x86\x7E",
  /* U+F208 */	"\x86\xA1", "\x86\xA2", "\x86\xA3", "\x86\xA4",
  /* U+F20C */	"\x86\xA5", "\x86\xA6", "\x86\xA7", "\x86\xA8",
  /* U+F210 */	"\x86\xA9", "\x86\xAA", "\x86\xAB", "\x86\xAC",
  /* U+F214 */	"\x86\xAD", "\x86\xAE", "\x86\xAF", "\x86\xB0",
  /* U+F218 */	"\x86\xB1", "\x86\xB2", "\x86\xB3", "\x86\xB4",
  /* U+F21C */	"\x86\xB5", "\x86\xB6", "\x86\xB7", "\x86\xB8",
  /* U+F220 */	"\x86\xB9", "\x86\xBA", "\x86\xBB", "\x86\xBC",
  /* U+F224 */	"\x86\xBD", "\x86\xBE", "\x86\xBF", "\x86\xC0",
  /* U+F228 */	"\x86\xC1", "\x86\xC2", "\x86\xC3", "\x86\xC4",
  /* U+F22C */	"\x86\xC5", "\x86\xC6", "\x86\xC7", "\x86\xC8",
  /* U+F230 */	"\x86\xC9", "\x86\xCA", "\x86\xCB", "\x86\xCC",
  /* U+F234 */	"\x86\xCD", "\x86\xCE", "\x86\xCF", "\x86\xD0",
  /* U+F238 */	"\x86\xD1", "\x86\xD2", "\x86\xD3", "\x86\xD4",
  /* U+F23C */	"\x86\xD5", "\x86\xD6", "\x86\xD7", "\x86\xD8",
  /* U+F240 */	"\x86\xD9", "\x86\xDA", "\x86\xDB", "\x86\xDC",
  /* U+F244 */	"\x86\xDD", "\x86\xDE", "\x86\xDF", "\x86\xE0",
  /* U+F248 */	"\x86\xE1", "\x86\xE2", "\x86\xE3", "\x86\xE4",
  /* U+F24C */	"\x86\xE5", "\x86\xE6", "\x86\xE7", "\x86\xE8",
  /* U+F250 */	"\x86\xE9", "\x86\xEA", "\x86\xEB", "\x86\xEC",
  /* U+F254 */	"\x86\xED", "\x86\xEE", "\x86\xEF", "\x86\xF0",
  /* U+F258 */	"\x86\xF1", "\x86\xF2", "\x86\xF3", "\x86\xF4",
  /* U+F25C */	"\x86\xF5", "\x86\xF6", "\x86\xF7", "\x86\xF8",
  /* U+F260 */	"\x86\xF9", "\x86\xFA", "\x86\xFB", "\x86\xFC",
  /* U+F264 */	"\x86\xFD", "\x86\xFE", "\x87\x40", "\x87\x41",
  /* U+F268 */	"\x87\x42", "\x87\x43", "\x87\x44", "\x87\x45",
  /* U+F26C */	"\x87\x46", "\x87\x47", "\x87\x48", "\x87\x49",
  /* U+F270 */	"\x87\x4A", "\x87\x4B", "\x87\x4C", "\x87\x4D",
  /* U+F274 */	"\x87\x4E", "\x87\x4F", "\x87\x50", "\x87\x51",
  /* U+F278 */	"\x87\x52", "\x87\x53", "\x87\x54", "\x87\x55",
  /* U+F27C */	"\x87\x56", "\x87\x57", "\x87\x58", "\x87\x59",
  /* U+F280 */	"\x87\x5A", "\x87\x5B", "\x87\x5C", "\x87\x5D",
  /* U+F284 */	"\x87\x5E", "\x87\x5F", "\x87\x60", "\x87\x61",
  /* U+F288 */	"\x87\x62", "\x87\x63", "\x87\x64", "\x87\x65",
  /* U+F28C */	"\x87\x66", "\x87\x67", "\x87\x68", "\x87\x69",
  /* U+F290 */	"\x87\x6A", "\x87\x6B", "\x87\x6C", "\x87\x6D",
  /* U+F294 */	"\x87\x6E", "\x87\x6F", "\x87\x70", "\x87\x71",
  /* U+F298 */	"\x87\x72", "\x87\x73", "\x87\x74", "\x87\x75",
  /* U+F29C */	"\x87\x76", "\x87\x77", "\x87\x78", "\x87\x79",
  /* U+F2A0 */	"\x87\x7A", "\x87\x7B", "\x87\x7C", "\x87\x7D",
  /* U+F2A4 */	"\x87\x7E", "\x87\xA1", "\x87\xA2", "\x87\xA3",
  /* U+F2A8 */	"\x87\xA4", "\x87\xA5", "\x87\xA6", "\x87\xA7",
  /* U+F2AC */	"\x87\xA8", "\x87\xA9", "\x87\xAA", "\x87\xAB",
  /* U+F2B0 */	"\x87\xAC", "\x87\xAD", "\x87\xAE", "\x87\xAF",
  /* U+F2B4 */	"\x87\xB0", "\x87\xB1", "\x87\xB2", "\x87\xB3",
  /* U+F2B8 */	"\x87\xB4", "\x87\xB5", "\x87\xB6", "\x87\xB7",
  /* U+F2BC */	"\x87\xB8", "\x87\xB9", "\x87\xBA", "\x87\xBB",
  /* U+F2C0 */	"\x87\xBC", "\x87\xBD", "\x87\xBE", "\x87\xBF",
  /* U+F2C4 */	"\x87\xC0", "\x87\xC1", "\x87\xC2", "\x87\xC3",
  /* U+F2C8 */	"\x87\xC4", "\x87\xC5", "\x87\xC6", "\x87\xC7",
  /* U+F2CC */	"\x87\xC8", "\x87\xC9", "\x87\xCA", "\x87\xCB",
  /* U+F2D0 */	"\x87\xCC", "\x87\xCD", "\x87\xCE", "\x87\xCF",
  /* U+F2D4 */	"\x87\xD0", "\x87\xD1", "\x87\xD2", "\x87\xD3",
  /* U+F2D8 */	"\x87\xD4", "\x87\xD5", "\x87\xD6", "\x87\xD7",
  /* U+F2DC */	"\x87\xD8", "\x87\xD9", "\x87\xDA", "\x87\xDB",
  /* U+F2E0 */	"\x87\xDC", "\x87\xDD", "\x87\xDE", "\x87\xDF",
  /* U+F2E4 */	"\x87\xE0", "\x87\xE1", "\x87\xE2", "\x87\xE3",
  /* U+F2E8 */	"\x87\xE4", "\x87\xE5", "\x87\xE6", "\x87\xE7",
  /* U+F2EC */	"\x87\xE8", "\x87\xE9", "\x87\xEA", "\x87\xEB",
  /* U+F2F0 */	"\x87\xEC", "\x87\xED", "\x87\xEE", "\x87\xEF",
  /* U+F2F4 */	"\x87\xF0", "\x87\xF1", "\x87\xF2", "\x87\xF3",
  /* U+F2F8 */	"\x87\xF4", "\x87\xF5", "\x87\xF6", "\x87\xF7",
  /* U+F2FC */	"\x87\xF8", "\x87\xF9", "\x87\xFA", "\x87\xFB",
  /* U+F300 */	"\x87\xFC", "\x87\xFD", "\x87\xFE", "\x88\x40",
  /* U+F304 */	"\x88\x41", "\x88\x42", "\x88\x43", "\x88\x44",
  /* U+F308 */	"\x88\x45", "\x88\x46", "\x88\x47", "\x88\x48",
  /* U+F30C */	"\x88\x49", "\x88\x4A", "\x88\x4B", "\x88\x4C",
  /* U+F310 */	"\x88\x4D", "\x88\x4E", "\x88\x4F", "\x88\x50",
  /* U+F314 */	"\x88\x51", "\x88\x52", "\x88\x53", "\x88\x54",
  /* U+F318 */	"\x88\x55", "\x88\x56", "\x88\x57", "\x88\x58",
  /* U+F31C */	"\x88\x59", "\x88\x5A", "\x88\x5B", "\x88\x5C",
  /* U+F320 */	"\x88\x5D", "\x88\x5E", "\x88\x5F", "\x88\x60",
  /* U+F324 */	"\x88\x61", "\x88\x62", "\x88\x63", "\x88\x64",
  /* U+F328 */	"\x88\x65", "\x88\x66", "\x88\x67", "\x88\x68",
  /* U+F32C */	"\x88\x69", "\x88\x6A", "\x88\x6B", "\x88\x6C",
  /* U+F330 */	"\x88\x6D", "\x88\x6E", "\x88\x6F", "\x88\x70",
  /* U+F334 */	"\x88\x71", "\x88\x72", "\x88\x73", "\x88\x74",
  /* U+F338 */	"\x88\x75", "\x88\x76", "\x88\x77", "\x88\x78",
  /* U+F33C */	"\x88\x79", "\x88\x7A", "\x88\x7B", "\x88\x7C",
  /* U+F340 */	"\x88\x7D", "\x88\x7E", "\x88\xA1", "\x88\xA2",
  /* U+F344 */	"\x88\xA3", "\x88\xA4", "\x88\xA5", "\x88\xA6",
  /* U+F348 */	"\x88\xA7", "\x88\xA8", "\x88\xA9", "\x88\xAA",
  /* U+F34C */	"\x88\xAB", "\x88\xAC", "\x88\xAD", "\x88\xAE",
  /* U+F350 */	"\x88\xAF", "\x88\xB0", "\x88\xB1", "\x88\xB2",
  /* U+F354 */	"\x88\xB3", "\x88\xB4", "\x88\xB5", "\x88\xB6",
  /* U+F358 */	"\x88\xB7", "\x88\xB8", "\x88\xB9", "\x88\xBA",
  /* U+F35C */	"\x88\xBB", "\x88\xBC", "\x88\xBD", "\x88\xBE",
  /* U+F360 */	"\x88\xBF", "\x88\xC0", "\x88\xC1", "\x88\xC2",
  /* U+F364 */	"\x88\xC3", "\x88\xC4", "\x88\xC5", "\x88\xC6",
  /* U+F368 */	"\x88\xC7", "\x88\xC8", "\x88\xC9", "\x88\xCA",
  /* U+F36C */	"\x88\xCB", "\x88\xCC", "\x88\xCD", "\x88\xCE",
  /* U+F370 */	"\x88\xCF", "\x88\xD0", "\x88\xD1", "\x88\xD2",
  /* U+F374 */	"\x88\xD3", "\x88\xD4", "\x88\xD5", "\x88\xD6",
  /* U+F378 */	"\x88\xD7", "\x88\xD8", "\x88\xD9", "\x88\xDA",
  /* U+F37C */	"\x88\xDB", "\x88\xDC", "\x88\xDD", "\x88\xDE",
  /* U+F380 */	"\x88\xDF", "\x88\xE0", "\x88\xE1", "\x88\xE2",
  /* U+F384 */	"\x88\xE3", "\x88\xE4", "\x88\xE5", "\x88\xE6",
  /* U+F388 */	"\x88\xE7", "\x88\xE8", "\x88\xE9", "\x88\xEA",
  /* U+F38C */	"\x88\xEB", "\x88\xEC", "\x88\xED", "\x88\xEE",
  /* U+F390 */	"\x88\xEF", "\x88\xF0", "\x88\xF1", "\x88\xF2",
  /* U+F394 */	"\x88\xF3", "\x88\xF4", "\x88\xF5", "\x88\xF6",
  /* U+F398 */	"\x88\xF7", "\x88\xF8", "\x88\xF9", "\x88\xFA",
  /* U+F39C */	"\x88\xFB", "\x88\xFC", "\x88\xFD", "\x88\xFE",
  /* U+F3A0 */	"\x89\x40", "\x89\x41", "\x89\x42", "\x89\x43",
  /* U+F3A4 */	"\x89\x44", "\x89\x45", "\x89\x46", "\x89\x47",
  /* U+F3A8 */	"\x89\x48", "\x89\x49", "\x89\x4A", "\x89\x4B",
  /* U+F3AC */	"\x89\x4C", "\x89\x4D", "\x89\x4E", "\x89\x4F",
  /* U+F3B0 */	"\x89\x50", "\x89\x51", "\x89\x52", "\x89\x53",
  /* U+F3B4 */	"\x89\x54", "\x89\x55", "\x89\x56", "\x89\x57",
  /* U+F3B8 */	"\x89\x58", "\x89\x59", "\x89\x5A", "\x89\x5B",
  /* U+F3BC */	"\x89\x5C", "\x89\x5D", "\x89\x5E", "\x89\x5F",
  /* U+F3C0 */	"\x89\x60", "\x89\x61", "\x89\x62", "\x89\x63",
  /* U+F3C4 */	"\x89\x64", "\x89\x65", "\x89\x66", "\x89\x67",
  /* U+F3C8 */	"\x89\x68", "\x89\x69", "\x89\x6A", "\x89\x6B",
  /* U+F3CC */	"\x89\x6C", "\x89\x6D", "\x89\x6E", "\x89\x6F",
  /* U+F3D0 */	"\x89\x70", "\x89\x71", "\x89\x72", "\x89\x73",
  /* U+F3D4 */	"\x89\x74", "\x89\x75", "\x89\x76", "\x89\x77",
  /* U+F3D8 */	"\x89\x78", "\x89\x79", "\x89\x7A", "\x89\x7B",
  /* U+F3DC */	"\x89\x7C", "\x89\x7D", "\x89\x7E", "\x89\xA1",
  /* U+F3E0 */	"\x89\xA2", "\x89\xA3", "\x89\xA4", "\x89\xA5",
  /* U+F3E4 */	"\x89\xA6", "\x89\xA7", "\x89\xA8", "\x89\xA9",
  /* U+F3E8 */	"\x89\xAA", "\x89\xAB", "\x89\xAC", "\x89\xAD",
  /* U+F3EC */	"\x89\xAE", "\x89\xAF", "\x89\xB0", "\x89\xB1",
  /* U+F3F0 */	"\x89\xB2", "\x89\xB3", "\x89\xB4", "\x89\xB5",
  /* U+F3F4 */	"\x89\xB6", "\x89\xB7", "\x89\xB8", "\x89\xB9",
  /* U+F3F8 */	"\x89\xBA", "\x89\xBB", "\x89\xBC", "\x89\xBD",
  /* U+F3FC */	"\x89\xBE", "\x89\xBF", "\x89\xC0", "\x89\xC1",
  /* U+F400 */	"\x89\xC2", "\x89\xC3", "\x89\xC4", "\x89\xC5",
  /* U+F404 */	"\x89\xC6", "\x89\xC7", "\x89\xC8", "\x89\xC9",
  /* U+F408 */	"\x89\xCA", "\x89\xCB", "\x89\xCC", "\x89\xCD",
  /* U+F40C */	"\x89\xCE", "\x89\xCF", "\x89\xD0", "\x89\xD1",
  /* U+F410 */	"\x89\xD2", "\x89\xD3", "\x89\xD4", "\x89\xD5",
  /* U+F414 */	"\x89\xD6", "\x89\xD7", "\x89\xD8", "\x89\xD9",
  /* U+F418 */	"\x89\xDA", "\x89\xDB", "\x89\xDC", "\x89\xDD",
  /* U+F41C */	"\x89\xDE", "\x89\xDF", "\x89\xE0", "\x89\xE1",
  /* U+F420 */	"\x89\xE2", "\x89\xE3", "\x89\xE4", "\x89\xE5",
  /* U+F424 */	"\x89\xE6", "\x89\xE7", "\x89\xE8", "\x89\xE9",
  /* U+F428 */	"\x89\xEA", "\x89\xEB", "\x89\xEC", "\x89\xED",
  /* U+F42C */	"\x89\xEE", "\x89\xEF", "\x89\xF0", "\x89\xF1",
  /* U+F430 */	"\x89\xF2", "\x89\xF3", "\x89\xF4", "\x89\xF5",
  /* U+F434 */	"\x89\xF6", "\x89\xF7", "\x89\xF8", "\x89\xF9",
  /* U+F438 */	"\x89\xFA", "\x89\xFB", "\x89\xFC", "\x89\xFD",
  /* U+F43C */	"\x89\xFE", "\x8A\x40", "\x8A\x41", "\x8A\x42",
  /* U+F440 */	"\x8A\x43", "\x8A\x44", "\x8A\x45", "\x8A\x46",
  /* U+F444 */	"\x8A\x47", "\x8A\x48", "\x8A\x49", "\x8A\x4A",
  /* U+F448 */	"\x8A\x4B", "\x8A\x4C", "\x8A\x4D", "\x8A\x4E",
  /* U+F44C */	"\x8A\x4F", "\x8A\x50", "\x8A\x51", "\x8A\x52",
  /* U+F450 */	"\x8A\x53", "\x8A\x54", "\x8A\x55", "\x8A\x56",
  /* U+F454 */	"\x8A\x57", "\x8A\x58", "\x8A\x59", "\x8A\x5A",
  /* U+F458 */	"\x8A\x5B", "\x8A\x5C", "\x8A\x5D", "\x8A\x5E",
  /* U+F45C */	"\x8A\x5F", "\x8A\x60", "\x8A\x61", "\x8A\x62",
  /* U+F460 */	"\x8A\x63", "\x8A\x64", "\x8A\x65", "\x8A\x66",
  /* U+F464 */	"\x8A\x67", "\x8A\x68", "\x8A\x69", "\x8A\x6A",
  /* U+F468 */	"\x8A\x6B", "\x8A\x6C", "\x8A\x6D", "\x8A\x6E",
  /* U+F46C */	"\x8A\x6F", "\x8A\x70", "\x8A\x71", "\x8A\x72",
  /* U+F470 */	"\x8A\x73", "\x8A\x74", "\x8A\x75", "\x8A\x76",
  /* U+F474 */	"\x8A\x77", "\x8A\x78", "\x8A\x79", "\x8A\x7A",
  /* U+F478 */	"\x8A\x7B", "\x8A\x7C", "\x8A\x7D", "\x8A\x7E",
  /* U+F47C */	"\x8A\xA1", "\x8A\xA2", "\x8A\xA3", "\x8A\xA4",
  /* U+F480 */	"\x8A\xA5", "\x8A\xA6", "\x8A\xA7", "\x8A\xA8",
  /* U+F484 */	"\x8A\xA9", "\x8A\xAA", "\x8A\xAB", "\x8A\xAC",
  /* U+F488 */	"\x8A\xAD", "\x8A\xAE", "\x8A\xAF", "\x8A\xB0",
  /* U+F48C */	"\x8A\xB1", "\x8A\xB2", "\x8A\xB3", "\x8A\xB4",
  /* U+F490 */	"\x8A\xB5", "\x8A\xB6", "\x8A\xB7", "\x8A\xB8",
  /* U+F494 */	"\x8A\xB9", "\x8A\xBA", "\x8A\xBB", "\x8A\xBC",
  /* U+F498 */	"\x8A\xBD", "\x8A\xBE", "\x8A\xBF", "\x8A\xC0",
  /* U+F49C */	"\x8A\xC1", "\x8A\xC2", "\x8A\xC3", "\x8A\xC4",
  /* U+F4A0 */	"\x8A\xC5", "\x8A\xC6", "\x8A\xC7", "\x8A\xC8",
  /* U+F4A4 */	"\x8A\xC9", "\x8A\xCA", "\x8A\xCB", "\x8A\xCC",
  /* U+F4A8 */	"\x8A\xCD", "\x8A\xCE", "\x8A\xCF", "\x8A\xD0",
  /* U+F4AC */	"\x8A\xD1", "\x8A\xD2", "\x8A\xD3", "\x8A\xD4",
  /* U+F4B0 */	"\x8A\xD5", "\x8A\xD6", "\x8A\xD7", "\x8A\xD8",
  /* U+F4B4 */	"\x8A\xD9", "\x8A\xDA", "\x8A\xDB", "\x8A\xDC",
  /* U+F4B8 */	"\x8A\xDD", "\x8A\xDE", "\x8A\xDF", "\x8A\xE0",
  /* U+F4BC */	"\x8A\xE1", "\x8A\xE2", "\x8A\xE3", "\x8A\xE4",
  /* U+F4C0 */	"\x8A\xE5", "\x8A\xE6", "\x8A\xE7", "\x8A\xE8",
  /* U+F4C4 */	"\x8A\xE9", "\x8A\xEA", "\x8A\xEB", "\x8A\xEC",
  /* U+F4C8 */	"\x8A\xED", "\x8A\xEE", "\x8A\xEF", "\x8A\xF0",
  /* U+F4CC */	"\x8A\xF1", "\x8A\xF2", "\x8A\xF3", "\x8A\xF4",
  /* U+F4D0 */	"\x8A\xF5", "\x8A\xF6", "\x8A\xF7", "\x8A\xF8",
  /* U+F4D4 */	"\x8A\xF9", "\x8A\xFA", "\x8A\xFB", "\x8A\xFC",
  /* U+F4D8 */	"\x8A\xFD", "\x8A\xFE", "\x8B\x40", "\x8B\x41",
  /* U+F4DC */	"\x8B\x42", "\x8B\x43", "\x8B\x44", "\x8B\x45",
  /* U+F4E0 */	"\x8B\x46", "\x8B\x47", "\x8B\x48", "\x8B\x49",
  /* U+F4E4 */	"\x8B\x4A", "\x8B\x4B", "\x8B\x4C", "\x8B\x4D",
  /* U+F4E8 */	"\x8B\x4E", "\x8B\x4F", "\x8B\x50", "\x8B\x51",
  /* U+F4EC */	"\x8B\x52", "\x8B\x53", "\x8B\x54", "\x8B\x55",
  /* U+F4F0 */	"\x8B\x56", "\x8B\x57", "\x8B\x58", "\x8B\x59",
  /* U+F4F4 */	"\x8B\x5A", "\x8B\x5B", "\x8B\x5C", "\x8B\x5D",
  /* U+F4F8 */	"\x8B\x5E", "\x8B\x5F", "\x8B\x60", "\x8B\x61",
  /* U+F4FC */	"\x8B\x62", "\x8B\x63", "\x8B\x64", "\x8B\x65",
  /* U+F500 */	"\x8B\x66", "\x8B\x67", "\x8B\x68", "\x8B\x69",
  /* U+F504 */	"\x8B\x6A", "\x8B\x6B", "\x8B\x6C", "\x8B\x6D",
  /* U+F508 */	"\x8B\x6E", "\x8B\x6F", "\x8B\x70", "\x8B\x71",
  /* U+F50C */	"\x8B\x72", "\x8B\x73", "\x8B\x74", "\x8B\x75",
  /* U+F510 */	"\x8B\x76", "\x8B\x77", "\x8B\x78", "\x8B\x79",
  /* U+F514 */	"\x8B\x7A", "\x8B\x7B", "\x8B\x7C", "\x8B\x7D",
  /* U+F518 */	"\x8B\x7E", "\x8B\xA1", "\x8B\xA2", "\x8B\xA3",
  /* U+F51C */	"\x8B\xA4", "\x8B\xA5", "\x8B\xA6", "\x8B\xA7",
  /* U+F520 */	"\x8B\xA8", "\x8B\xA9", "\x8B\xAA", "\x8B\xAB",
  /* U+F524 */	"\x8B\xAC", "\x8B\xAD", "\x8B\xAE", "\x8B\xAF",
  /* U+F528 */	"\x8B\xB0", "\x8B\xB1", "\x8B\xB2", "\x8B\xB3",
  /* U+F52C */	"\x8B\xB4", "\x8B\xB5", "\x8B\xB6", "\x8B\xB7",
  /* U+F530 */	"\x8B\xB8", "\x8B\xB9", "\x8B\xBA", "\x8B\xBB",
  /* U+F534 */	"\x8B\xBC", "\x8B\xBD", "\x8B\xBE", "\x8B\xBF",
  /* U+F538 */	"\x8B\xC0", "\x8B\xC1", "\x8B\xC2", "\x8B\xC3",
  /* U+F53C */	"\x8B\xC4", "\x8B\xC5", "\x8B\xC6", "\x8B\xC7",
  /* U+F540 */	"\x8B\xC8", "\x8B\xC9", "\x8B\xCA", "\x8B\xCB",
  /* U+F544 */	"\x8B\xCC", "\x8B\xCD", "\x8B\xCE", "\x8B\xCF",
  /* U+F548 */	"\x8B\xD0", "\x8B\xD1", "\x8B\xD2", "\x8B\xD3",
  /* U+F54C */	"\x8B\xD4", "\x8B\xD5", "\x8B\xD6", "\x8B\xD7",
  /* U+F550 */	"\x8B\xD8", "\x8B\xD9", "\x8B\xDA", "\x8B\xDB",
  /* U+F554 */	"\x8B\xDC", "\x8B\xDD", "\x8B\xDE", "\x8B\xDF",
  /* U+F558 */	"\x8B\xE0", "\x8B\xE1", "\x8B\xE2", "\x8B\xE3",
  /* U+F55C */	"\x8B\xE4", "\x8B\xE5", "\x8B\xE6", "\x8B\xE7",
  /* U+F560 */	"\x8B\xE8", "\x8B\xE9", "\x8B\xEA", "\x8B\xEB",
  /* U+F564 */	"\x8B\xEC", "\x8B\xED", "\x8B\xEE", "\x8B\xEF",
  /* U+F568 */	"\x8B\xF0", "\x8B\xF1", "\x8B\xF2", "\x8B\xF3",
  /* U+F56C */	"\x8B\xF4", "\x8B\xF5", "\x8B\xF6", "\x8B\xF7",
  /* U+F570 */	"\x8B\xF8", "\x8B\xF9", "\x8B\xFA", "\x8B\xFB",
  /* U+F574 */	"\x8B\xFC", "\x8B\xFD", "\x8B\xFE", "\x8C\x40",
  /* U+F578 */	"\x8C\x41", "\x8C\x42", "\x8C\x43", "\x8C\x44",
  /* U+F57C */	"\x8C\x45", "\x8C\x46", "\x8C\x47", "\x8C\x48",
  /* U+F580 */	"\x8C\x49", "\x8C\x4A", "\x8C\x4B", "\x8C\x4C",
  /* U+F584 */	"\x8C\x4D", "\x8C\x4E", "\x8C\x4F", "\x8C\x50",
  /* U+F588 */	"\x8C\x51", "\x8C\x52", "\x8C\x53", "\x8C\x54",
  /* U+F58C */	"\x8C\x55", "\x8C\x56", "\x8C\x57", "\x8C\x58",
  /* U+F590 */	"\x8C\x59", "\x8C\x5A", "\x8C\x5B", "\x8C\x5C",
  /* U+F594 */	"\x8C\x5D", "\x8C\x5E", "\x8C\x5F", "\x8C\x60",
  /* U+F598 */	"\x8C\x61", "\x8C\x62", "\x8C\x63", "\x8C\x64",
  /* U+F59C */	"\x8C\x65", "\x8C\x66", "\x8C\x67", "\x8C\x68",
  /* U+F5A0 */	"\x8C\x69", "\x8C\x6A", "\x8C\x6B", "\x8C\x6C",
  /* U+F5A4 */	"\x8C\x6D", "\x8C\x6E", "\x8C\x6F", "\x8C\x70",
  /* U+F5A8 */	"\x8C\x71", "\x8C\x72", "\x8C\x73", "\x8C\x74",
  /* U+F5AC */	"\x8C\x75", "\x8C\x76", "\x8C\x77", "\x8C\x78",
  /* U+F5B0 */	"\x8C\x79", "\x8C\x7A", "\x8C\x7B", "\x8C\x7C",
  /* U+F5B4 */	"\x8C\x7D", "\x8C\x7E", "\x8C\xA1", "\x8C\xA2",
  /* U+F5B8 */	"\x8C\xA3", "\x8C\xA4", "\x8C\xA5", "\x8C\xA6",
  /* U+F5BC */	"\x8C\xA7", "\x8C\xA8", "\x8C\xA9", "\x8C\xAA",
  /* U+F5C0 */	"\x8C\xAB", "\x8C\xAC", "\x8C\xAD", "\x8C\xAE",
  /* U+F5C4 */	"\x8C\xAF", "\x8C\xB0", "\x8C\xB1", "\x8C\xB2",
  /* U+F5C8 */	"\x8C\xB3", "\x8C\xB4", "\x8C\xB5", "\x8C\xB6",
  /* U+F5CC */	"\x8C\xB7", "\x8C\xB8", "\x8C\xB9", "\x8C\xBA",
  /* U+F5D0 */	"\x8C\xBB", "\x8C\xBC", "\x8C\xBD", "\x8C\xBE",
  /* U+F5D4 */	"\x8C\xBF", "\x8C\xC0", "\x8C\xC1", "\x8C\xC2",
  /* U+F5D8 */	"\x8C\xC3", "\x8C\xC4", "\x8C\xC5", "\x8C\xC6",
  /* U+F5DC */	"\x8C\xC7", "\x8C\xC8", "\x8C\xC9", "\x8C\xCA",
  /* U+F5E0 */	"\x8C\xCB", "\x8C\xCC", "\x8C\xCD", "\x8C\xCE",
  /* U+F5E4 */	"\x8C\xCF", "\x8C\xD0", "\x8C\xD1", "\x8C\xD2",
  /* U+F5E8 */	"\x8C\xD3", "\x8C\xD4", "\x8C\xD5", "\x8C\xD6",
  /* U+F5EC */	"\x8C\xD7", "\x8C\xD8", "\x8C\xD9", "\x8C\xDA",
  /* U+F5F0 */	"\x8C\xDB", "\x8C\xDC", "\x8C\xDD", "\x8C\xDE",
  /* U+F5F4 */	"\x8C\xDF", "\x8C\xE0", "\x8C\xE1", "\x8C\xE2",
  /* U+F5F8 */	"\x8C\xE3", "\x8C\xE4", "\x8C\xE5", "\x8C\xE6",
  /* U+F5FC */	"\x8C\xE7", "\x8C\xE8", "\x8C\xE9", "\x8C\xEA",
  /* U+F600 */	"\x8C\xEB", "\x8C\xEC", "\x8C\xED", "\x8C\xEE",
  /* U+F604 */	"\x8C\xEF", "\x8C\xF0", "\x8C\xF1", "\x8C\xF2",
  /* U+F608 */	"\x8C\xF3", "\x8C\xF4", "\x8C\xF5", "\x8C\xF6",
  /* U+F60C */	"\x8C\xF7", "\x8C\xF8", "\x8C\xF9", "\x8C\xFA",
  /* U+F610 */	"\x8C\xFB", "\x8C\xFC", "\x8C\xFD", "\x8C\xFE",
  /* U+F614 */	"\x8D\x40", "\x8D\x41", "\x8D\x42", "\x8D\x43",
  /* U+F618 */	"\x8D\x44", "\x8D\x45", "\x8D\x46", "\x8D\x47",
  /* U+F61C */	"\x8D\x48", "\x8D\x49", "\x8D\x4A", "\x8D\x4B",
  /* U+F620 */	"\x8D\x4C", "\x8D\x4D", "\x8D\x4E", "\x8D\x4F",
  /* U+F624 */	"\x8D\x50", "\x8D\x51", "\x8D\x52", "\x8D\x53",
  /* U+F628 */	"\x8D\x54", "\x8D\x55", "\x8D\x56", "\x8D\x57",
  /* U+F62C */	"\x8D\x58", "\x8D\x59", "\x8D\x5A", "\x8D\x5B",
  /* U+F630 */	"\x8D\x5C", "\x8D\x5D", "\x8D\x5E", "\x8D\x5F",
  /* U+F634 */	"\x8D\x60", "\x8D\x61", "\x8D\x62", "\x8D\x63",
  /* U+F638 */	"\x8D\x64", "\x8D\x65", "\x8D\x66", "\x8D\x67",
  /* U+F63C */	"\x8D\x68", "\x8D\x69", "\x8D\x6A", "\x8D\x6B",
  /* U+F640 */	"\x8D\x6C", "\x8D\x6D", "\x8D\x6E", "\x8D\x6F",
  /* U+F644 */	"\x8D\x70", "\x8D\x71", "\x8D\x72", "\x8D\x73",
  /* U+F648 */	"\x8D\x74", "\x8D\x75", "\x8D\x76", "\x8D\x77",
  /* U+F64C */	"\x8D\x78", "\x8D\x79", "\x8D\x7A", "\x8D\x7B",
  /* U+F650 */	"\x8D\x7C", "\x8D\x7D", "\x8D\x7E", "\x8D\xA1",
  /* U+F654 */	"\x8D\xA2", "\x8D\xA3", "\x8D\xA4", "\x8D\xA5",
  /* U+F658 */	"\x8D\xA6", "\x8D\xA7", "\x8D\xA8", "\x8D\xA9",
  /* U+F65C */	"\x8D\xAA", "\x8D\xAB", "\x8D\xAC", "\x8D\xAD",
  /* U+F660 */	"\x8D\xAE", "\x8D\xAF", "\x8D\xB0", "\x8D\xB1",
  /* U+F664 */	"\x8D\xB2", "\x8D\xB3", "\x8D\xB4", "\x8D\xB5",
  /* U+F668 */	"\x8D\xB6", "\x8D\xB7", "\x8D\xB8", "\x8D\xB9",
  /* U+F66C */	"\x8D\xBA", "\x8D\xBB", "\x8D\xBC", "\x8D\xBD",
  /* U+F670 */	"\x8D\xBE", "\x8D\xBF", "\x8D\xC0", "\x8D\xC1",
  /* U+F674 */	"\x8D\xC2", "\x8D\xC3", "\x8D\xC4", "\x8D\xC5",
  /* U+F678 */	"\x8D\xC6", "\x8D\xC7", "\x8D\xC8", "\x8D\xC9",
  /* U+F67C */	"\x8D\xCA", "\x8D\xCB", "\x8D\xCC", "\x8D\xCD",
  /* U+F680 */	"\x8D\xCE", "\x8D\xCF", "\x8D\xD0", "\x8D\xD1",
  /* U+F684 */	"\x8D\xD2", "\x8D\xD3", "\x8D\xD4", "\x8D\xD5",
  /* U+F688 */	"\x8D\xD6", "\x8D\xD7", "\x8D\xD8", "\x8D\xD9",
  /* U+F68C */	"\x8D\xDA", "\x8D\xDB", "\x8D\xDC", "\x8D\xDD",
  /* U+F690 */	"\x8D\xDE", "\x8D\xDF", "\x8D\xE0", "\x8D\xE1",
  /* U+F694 */	"\x8D\xE2", "\x8D\xE3", "\x8D\xE4", "\x8D\xE5",
  /* U+F698 */	"\x8D\xE6", "\x8D\xE7", "\x8D\xE8", "\x8D\xE9",
  /* U+F69C */	"\x8D\xEA", "\x8D\xEB", "\x8D\xEC", "\x8D\xED",
  /* U+F6A0 */	"\x8D\xEE", "\x8D\xEF", "\x8D\xF0", "\x8D\xF1",
  /* U+F6A4 */	"\x8D\xF2", "\x8D\xF3", "\x8D\xF4", "\x8D\xF5",
  /* U+F6A8 */	"\x8D\xF6", "\x8D\xF7", "\x8D\xF8", "\x8D\xF9",
  /* U+F6AC */	"\x8D\xFA", "\x8D\xFB", "\x8D\xFC", "\x8D\xFD",
  /* U+F6B0 */	"\x8D\xFE", "\xC6\xA1", "\xC6\xA2", "\xC6\xA3",
  /* U+F6B4 */	"\xC6\xA4", "\xC6\xA5", "\xC6\xA6", "\xC6\xA7",
  /* U+F6B8 */	"\xC6\xA8", "\xC6\xA9", "\xC6\xAA", "\xC6\xAB",
  /* U+F6BC */	"\xC6\xAC", "\xC6\xAD", "\xC6\xAE", "\xC6\xAF",
  /* U+F6C0 */	"\xC6\xB0", "\xC6\xB1", "\xC6\xB2", "\xC6\xB3",
  /* U+F6C4 */	"\xC6\xB4", "\xC6\xB5", "\xC6\xB6", "\xC6\xB7",
  /* U+F6C8 */	"\xC6\xB8", "\xC6\xB9", "\xC6\xBA", "\xC6\xBB",
  /* U+F6CC */	"\xC6\xBC", "\xC6\xBD", "\xC6\xBE", "\xC6\xBF",
  /* U+F6D0 */	"\xC6\xC0", "\xC6\xC1", "\xC6\xC2", "\xC6\xC3",
  /* U+F6D4 */	"\xC6\xC4", "\xC6\xC5", "\xC6\xC6", "\xC6\xC7",
  /* U+F6D8 */	"\xC6\xC8", "\xC6\xC9", "\xC6\xCA", "\xC6\xCB",
  /* U+F6DC */	"\xC6\xCC", "\xC6\xCD", "\xC6\xCE", "\xC6\xCF",
  /* U+F6E0 */	"\xC6\xD0", "\xC6\xD1", "\xC6\xD2", "\xC6\xD3",
  /* U+F6E4 */	"\xC6\xD4", "\xC6\xD5", "\xC6\xD6", "\xC6\xD7",
  /* U+F6E8 */	"\xC6\xD8", "\xC6\xD9", "\xC6\xDA", "\xC6\xDB",
  /* U+F6EC */	"\xC6\xDC", "\xC6\xDD", "\xC6\xDE", "\xC6\xDF",
  /* U+F6F0 */	"\xC6\xE0", "\xC6\xE1", "\xC6\xE2", "\xC6\xE3",
  /* U+F6F4 */	"\xC6\xE4", "\xC6\xE5", "\xC6\xE6", "\xC6\xE7",
  /* U+F6F8 */	"\xC6\xE8", "\xC6\xE9", "\xC6\xEA", "\xC6\xEB",
  /* U+F6FC */	"\xC6\xEC", "\xC6\xED", "\xC6\xEE", "\xC6\xEF",
  /* U+F700 */	"\xC6\xF0", "\xC6\xF1", "\xC6\xF2", "\xC6\xF3",
  /* U+F704 */	"\xC6\xF4", "\xC6\xF5", "\xC6\xF6", "\xC6\xF7",
  /* U+F708 */	"\xC6\xF8", "\xC6\xF9", "\xC6\xFA", "\xC6\xFB",
  /* U+F70C */	"\xC6\xFC", "\xC6\xFD", "\xC6\xFE", "\xC7\x40",
  /* U+F710 */	"\xC7\x41", "\xC7\x42", "\xC7\x43", "\xC7\x44",
  /* U+F714 */	"\xC7\x45", "\xC7\x46", "\xC7\x47", "\xC7\x48",
  /* U+F718 */	"\xC7\x49", "\xC7\x4A", "\xC7\x4B", "\xC7\x4C",
  /* U+F71C */	"\xC7\x4D", "\xC7\x4E", "\xC7\x4F", "\xC7\x50",
  /* U+F720 */	"\xC7\x51", "\xC7\x52", "\xC7\x53", "\xC7\x54",
  /* U+F724 */	"\xC7\x55", "\xC7\x56", "\xC7\x57", "\xC7\x58",
  /* U+F728 */	"\xC7\x59", "\xC7\x5A", "\xC7\x5B", "\xC7\x5C",
  /* U+F72C */	"\xC7\x5D", "\xC7\x5E", "\xC7\x5F", "\xC7\x60",
  /* U+F730 */	"\xC7\x61", "\xC7\x62", "\xC7\x63", "\xC7\x64",
  /* U+F734 */	"\xC7\x65", "\xC7\x66", "\xC7\x67", "\xC7\x68",
  /* U+F738 */	"\xC7\x69", "\xC7\x6A", "\xC7\x6B", "\xC7\x6C",
  /* U+F73C */	"\xC7\x6D", "\xC7\x6E", "\xC7\x6F", "\xC7\x70",
  /* U+F740 */	"\xC7\x71", "\xC7\x72", "\xC7\x73", "\xC7\x74",
  /* U+F744 */	"\xC7\x75", "\xC7\x76", "\xC7\x77", "\xC7\x78",
  /* U+F748 */	"\xC7\x79", "\xC7\x7A", "\xC7\x7B", "\xC7\x7C",
  /* U+F74C */	"\xC7\x7D", "\xC7\x7E", "\xC7\xA1", "\xC7\xA2",
  /* U+F750 */	"\xC7\xA3", "\xC7\xA4", "\xC7\xA5", "\xC7\xA6",
  /* U+F754 */	"\xC7\xA7", "\xC7\xA8", "\xC7\xA9", "\xC7\xAA",
  /* U+F758 */	"\xC7\xAB", "\xC7\xAC", "\xC7\xAD", "\xC7\xAE",
  /* U+F75C */	"\xC7\xAF", "\xC7\xB0", "\xC7\xB1", "\xC7\xB2",
  /* U+F760 */	"\xC7\xB3", "\xC7\xB4", "\xC7\xB5", "\xC7\xB6",
  /* U+F764 */	"\xC7\xB7", "\xC7\xB8", "\xC7\xB9", "\xC7\xBA",
  /* U+F768 */	"\xC7\xBB", "\xC7\xBC", "\xC7\xBD", "\xC7\xBE",
  /* U+F76C */	"\xC7\xBF", "\xC7\xC0", "\xC7\xC1", "\xC7\xC2",
  /* U+F770 */	"\xC7\xC3", "\xC7\xC4", "\xC7\xC5", "\xC7\xC6",
  /* U+F774 */	"\xC7\xC7", "\xC7\xC8", "\xC7\xC9", "\xC7\xCA",
  /* U+F778 */	"\xC7\xCB", "\xC7\xCC", "\xC7\xCD", "\xC7\xCE",
  /* U+F77C */	"\xC7\xCF", "\xC7\xD0", "\xC7\xD1", "\xC7\xD2",
  /* U+F780 */	"\xC7\xD3", "\xC7\xD4", "\xC7\xD5", "\xC7\xD6",
  /* U+F784 */	"\xC7\xD7", "\xC7\xD8", "\xC7\xD9", "\xC7\xDA",
  /* U+F788 */	"\xC7\xDB", "\xC7\xDC", "\xC7\xDD", "\xC7\xDE",
  /* U+F78C */	"\xC7\xDF", "\xC7\xE0", "\xC7\xE1", "\xC7\xE2",
  /* U+F790 */	"\xC7\xE3", "\xC7\xE4", "\xC7\xE5", "\xC7\xE6",
  /* U+F794 */	"\xC7\xE7", "\xC7\xE8", "\xC7\xE9", "\xC7\xEA",
  /* U+F798 */	"\xC7\xEB", "\xC7\xEC", "\xC7\xED", "\xC7\xEE",
  /* U+F79C */	"\xC7\xEF", "\xC7\xF0", "\xC7\xF1", "\xC7\xF2",
  /* U+F7A0 */	"\xC7\xF3", "\xC7\xF4", "\xC7\xF5", "\xC7\xF6",
  /* U+F7A4 */	"\xC7\xF7", "\xC7\xF8", "\xC7\xF9", "\xC7\xFA",
  /* U+F7A8 */	"\xC7\xFB", "\xC7\xFC", "\xC7\xFD", "\xC7\xFE",
  /* U+F7AC */	"\xC8\x40", "\xC8\x41", "\xC8\x42", "\xC8\x43",
  /* U+F7B0 */	"\xC8\x44", "\xC8\x45", "\xC8\x46", "\xC8\x47",
  /* U+F7B4 */	"\xC8\x48", "\xC8\x49", "\xC8\x4A", "\xC8\x4B",
  /* U+F7B8 */	"\xC8\x4C", "\xC8\x4D", "\xC8\x4E", "\xC8\x4F",
  /* U+F7BC */	"\xC8\x50", "\xC8\x51", "\xC8\x52", "\xC8\x53",
  /* U+F7C0 */	"\xC8\x54", "\xC8\x55", "\xC8\x56", "\xC8\x57",
  /* U+F7C4 */	"\xC8\x58", "\xC8\x59", "\xC8\x5A", "\xC8\x5B",
  /* U+F7C8 */	"\xC8\x5C", "\xC8\x5D", "\xC8\x5E", "\xC8\x5F",
  /* U+F7CC */	"\xC8\x60", "\xC8\x61", "\xC8\x62", "\xC8\x63",
  /* U+F7D0 */	"\xC8\x64", "\xC8\x65", "\xC8\x66", "\xC8\x67",
  /* U+F7D4 */	"\xC8\x68", "\xC8\x69", "\xC8\x6A", "\xC8\x6B",
  /* U+F7D8 */	"\xC8\x6C", "\xC8\x6D", "\xC8\x6E", "\xC8\x6F",
  /* U+F7DC */	"\xC8\x70", "\xC8\x71", "\xC8\x72", "\xC8\x73",
  /* U+F7E0 */	"\xC8\x74", "\xC8\x75", "\xC8\x76", "\xC8\x77",
  /* U+F7E4 */	"\xC8\x78", "\xC8\x79", "\xC8\x7A", "\xC8\x7B",
  /* U+F7E8 */	"\xC8\x7C", "\xC8\x7D", "\xC8\x7E", "\xC8\xA1",
  /* U+F7EC */	"\xC8\xA2", "\xC8\xA3", "\xC8\xA4", "\xC8\xA5",
  /* U+F7F0 */	"\xC8\xA6", "\xC8\xA7", "\xC8\xA8", "\xC8\xA9",
  /* U+F7F4 */	"\xC8\xAA", "\xC8\xAB", "\xC8\xAC", "\xC8\xAD",
  /* U+F7F8 */	"\xC8\xAE", "\xC8\xAF", "\xC8\xB0", "\xC8\xB1",
  /* U+F7FC */	"\xC8\xB2", "\xC8\xB3", "\xC8\xB4", "\xC8\xB5",
  /* U+F800 */	"\xC8\xB6", "\xC8\xB7", "\xC8\xB8", "\xC8\xB9",
  /* U+F804 */	"\xC8\xBA", "\xC8\xBB", "\xC8\xBC", "\xC8\xBD",
  /* U+F808 */	"\xC8\xBE", "\xC8\xBF", "\xC8\xC0", "\xC8\xC1",
  /* U+F80C */	"\xC8\xC2", "\xC8\xC3", "\xC8\xC4", "\xC8\xC5",
  /* U+F810 */	"\xC8\xC6", "\xC8\xC7", "\xC8\xC8", "\xC8\xC9",
  /* U+F814 */	"\xC8\xCA", "\xC8\xCB", "\xC8\xCC", "\xC8\xCD",
  /* U+F818 */	"\xC8\xCE", "\xC8\xCF", "\xC8\xD0", "\xC8\xD1",
  /* U+F81C */	"\xC8\xD2", "\xC8\xD3", "\xC8\xD4", "\xC8\xD5",
  /* U+F820 */	"\xC8\xD6", "\xC8\xD7", "\xC8\xD8", "\xC8\xD9",
  /* U+F824 */	"\xC8\xDA", "\xC8\xDB", "\xC8\xDC", "\xC8\xDD",
  /* U+F828 */	"\xC8\xDE", "\xC8\xDF", "\xC8\xE0", "\xC8\xE1",
  /* U+F82C */	"\xC8\xE2", "\xC8\xE3", "\xC8\xE4", "\xC8\xE5",
  /* U+F830 */	"\xC8\xE6", "\xC8\xE7", "\xC8\xE8", "\xC8\xE9",
  /* U+F834 */	"\xC8\xEA", "\xC8\xEB", "\xC8\xEC", "\xC8\xED",
  /* U+F838 */	"\xC8\xEE", "\xC8\xEF", "\xC8\xF0", "\xC8\xF1",
  /* U+F83C */	"\xC8\xF2", "\xC8\xF3", "\xC8\xF4", "\xC8\xF5",
  /* U+F840 */	"\xC8\xF6", "\xC8\xF7", "\xC8\xF8", "\xC8\xF9",
  /* U+F844 */	"\xC8\xFA", "\xC8\xFB", "\xC8\xFC", "\xC8\xFD",
  /* U+F848 */	"\xC8\xFE",

  /* U+F8F8 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+F8FC */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+F900 */	"\xB0\x5A", "\xA7\xF3", "\xA8\xAE", "\xB8\xEB",
  /* U+F904 */	"\xB7\xC6", "\xA6\xEA", "\xA5\x79", "\x8B\xF8",
  /* U+F908 */	"\xC0\x74", "\xAB\xB4", "\xAA\xF7", "\xB3\xE2",
  /* U+F90C */	"\xA9\x60", "\xC3\x69", "\xC4\xEE", "\xC3\xB9",
  /* U+F910 */	"\xC5\xDA", "\xC1\xB3", "\xBB\x72", "\xC5\xDE",
  /* U+F914 */	"\xBC\xD6", "\xAC\xA5", "\xAF\x4F", "\xAF\x5F",
  /* U+F918 */	"\xB8\xA8", "\xB9\x54", "\xC0\x64", "\xB6\xC3",
  /* U+F91C */	"\xA7\x5A", "\xC4\xE6", "\xC4\xEA", "\xC4\xF5",
  /* U+F920 */	"\xC6\x7D", "\xB4\x50", "\xC0\xDD", "\xC2\xC5",
  /* U+F924 */	"\xC4\xB0", "\xA9\xD4", "\xC3\xBE", "\xC4\xFA",
  /* U+F928 */	"\xB4\x59", "\xAE\xD4", "\xAE\xF6", "\xAF\x54",
  /* U+F92C */	"\x00\x00", "\xA8\xD3", "\xA7\x4E", "\xB3\xD2",
  /* U+F930 */	"\xBE\xDB", "\xC3\x72", "\xC4\x6C", "\xBF\x63",
  /* U+F934 */	"\xA6\xD1", "\xC4\xAA", "\xB8\xB8", "\xB8\xF4",
  /* U+F938 */	"\xC5\x53", "\xBE\x7C", "\xC6\x4F", "\xB8\x4C",
  /* U+F93C */	"\xB8\x53", "\xBA\xF1", "\xDB\x77", "\xBF\xFD",
  /* U+F940 */	"\xB3\xC0", "\xBD\xD7", "\xC3\x62", "\xA7\xCB",
  /* U+F944 */	"\xC5\xA2", "\xC5\xA4", "\xA8\x63", "\xBD\x55",
  /* U+F948 */	"\xB8\xEF", "\xB9\x70", "\xC2\x53", "\xB9\xF0",
  /* U+F94C */	"\xBC\xD3", "\xB2\x5C", "\xBA\x7C", "\xB2\xD6",
  /* U+F950 */	"\xC1\x5C", "\xAD\xAE", "\xB0\xC7", "\xA6\xD8",
  /* U+F954 */	"\xBB\xFE", "\xAD\xE2", "\xB8\x57", "\xBA\xF0",
  /* U+F958 */	"\xB5\xD9", "\xB3\xAE", "\xC5\xAA", "\xCE\xD4",
  /* U+F95C */	"\xBC\xD6", "\xBF\xD5", "\xA4\xA6", "\xB9\xE7",
  /* U+F960 */	"\xAB\xE3", "\xB2\x76", "\xB2\xA7", "\xA5\x5F",
  /* U+F964 */	"\xED\xA8", "\xAB\x4B", "\xB4\x5F", "\xA4\xA3",
  /* U+F968 */	"\xAA\x63", "\xBC\xC6", "\xAF\xC1", "\xB0\xD1",
  /* U+F96C */	"\xB6\xEB", "\xAC\xD9", "\xB8\xAD", "\xBB\xA1",
  /* U+F970 */	"\xB1\xFE", "\xA8\xB0", "\xA8\x48", "\xAC\x42",
  /* U+F974 */	"\xAD\x59", "\xB1\xB0", "\xB2\xA4", "\xAB\x47",
  /* U+F978 */	"\xA8\xE2", "\x00\x00", "\xB1\xE7", "\xC2\xB3",
  /* U+F97C */	"\xA8\x7D", "\xBD\xCC", "\xB6\x71", "\xC0\x79",
  /* U+F980 */	"\xA7\x66", "\xA4\x6B", "\xC3\x66", "\xAE\xC8",
  /* U+F984 */	"\xC2\x6F", "\xC4\x72", "\xBE\x5B", "\xC6\x7A",
  /* U+F988 */	"\xC4\x52", "\xBE\xA4", "\xA4\x4F", "\xBE\xE4",
  /* U+F98C */	"\xBE\xFA", "\xF7\x65", "\xA6\x7E", "\xBC\xA6",
  /* U+F990 */	"\xC5\xCA", "\xBC\xBF", "\xBA\xA7", "\xB7\xD2",
  /* U+F994 */	"\xE6\xA3", "\x00\x00", "\xBD\x6D", "\xC1\x70",
  /* U+F998 */	"\xBD\xFB", "\xBD\xAC", "\xB3\x73", "\xC1\xE5",
  /* U+F99C */	"\xA6\x43", "\xA6\x48", "\xAB\x7C", "\xAF\x50",
  /* U+F9A0 */	"\xB5\xF5", "\xBB\xA1", "\xB7\x47", "\xA9\xC0",
  /* U+F9A4 */	"\xB1\xC9", "\xC0\xD4", "\xC3\xAE", "\xC2\x79",
  /* U+F9A8 */	"\xA5\x4F", "\xCB\xF1", "\xB9\xE7", "\xC0\xAD",
  /* U+F9AC */	"\xCC\xB0", "\xAC\xC2", "\xBC\xFC", "\xB2\xDC",
  /* U+F9B0 */	"\xB2\xE2", "\xB9\x61", "\xB9\x73", "\xC6\x46",
  /* U+F9B4 */	"\xBB\xE2", "\xA8\xD2", "\xC2\xA7", "\xC4\xBF",
  /* U+F9B8 */	"\xC1\xF5", "\xB4\x63", "\xA4\x46", "\xB9\xB1",
  /* U+F9BC */	"\xBC\x64", "\xA7\xBF", "\xAE\xC6", "\xBC\xD6",
  /* U+F9C0 */	"\xBF\x52", "\xC0\xF8", "\xE7\x64", "\xBF\xF1",
  /* U+F9C4 */	"\xC0\x73", "\xB7\x77", "\xA8\xBF", "\xBC\x42",
  /* U+F9C8 */	"\xCC\xD8", "\xAC\x68", "\xAC\x79", "\xB7\xC8",
  /* U+F9CC */	"\xAF\x5B", "\xAF\x64", "\xB2\xB8", "\xAF\xC3",
  /* U+F9D0 */	"\xC3\xFE", "\xA4\xBB", "\xBC\xAE", "\xB3\xB0",
  /* U+F9D4 */	"\xAD\xDB", "\xB1\x5B", "\xB2\x5F", "\xBD\xFC",
  /* U+F9D8 */	"\xAB\xDF", "\xB7\x58", "\xAE\xDF", "\xB2\x76",
  /* U+F9DC */	"\xB6\xA9", "\xA7\x51", "\xA6\x4F", "\xBC\x69",
  /* U+F9E0 */	"\xA9\xF6", "\xA7\xF5", "\xB1\xF9", "\xAA\x64",
  /* U+F9E4 */	"\xB2\x7A", "\xB5\x67", "\xBF\xA9", "\x00\x00",
  /* U+F9E8 */	"\xB8\xCC", "\xA8\xBD", "\xC2\xF7", "\xB0\xCE",
  /* U+F9EC */	"\xB7\xC4", "\xA7\x5B", "\xBF\x4D", "\xBF\x5A",
  /* U+F9F0 */	"\xC4\xA9", "\x00\x00", "\xC5\xEC", "\xC5\xEF",
  /* U+F9F4 */	"\xAA\x4C", "\xB2\x4F", "\xC1\x7B", "\xA5\xDF",
  /* U+F9F8 */	"\xB2\xC1", "\xB2\xC9", "\xAA\xAC", "\xAA\xA5",
  /* U+F9FC */	"\xC3\xD1", "\xA4\xB0", "\xAF\xF9", "\xA8\xEB",
  /* U+FA00 */	"\xA4\xC1", "\xAB\xD7", "\xA9\xDD", "\xBF\x7D",
  /* U+FA04 */	"\xA6\x76", "\xAC\x7D", "\xBC\xC9", "\xBF\xE7",
  /* U+FA08 */	"\xA6\xE6", "\xAD\xB0", "\xA8\xA3", "\xB9\xF8",
  /* U+FA0C */	"\xC9\x4A", "\xDD\xFC", "\x00\x00", "\x00\x00",
  /* U+FA10 */	"\xB6\xEF", "\x00\x00", "\xB4\xB8", "\x00\x00",
  /* U+FA14 */	"\x00\x00", "\xE8\xF9", "\xBD\xDE", "\xAF\x71",
  /* U+FA18 */	"\x00\x00", "\xAF\xAB", "\xB2\xBB", "\xBA\xD6",
  /* U+FA1C */	"\xB9\x74", "\xBA\xEB", "\xA6\xD0", "\x00\x00",
  /* U+FA20 */	"\x00\x00", "\x00\x00", "\xBD\xD1", "\x00\x00",
  /* U+FA24 */	"\x00\x00", "\xB6\x68", "\xB3\xA3", "\x00\x00",
  /* U+FA28 */	"\x00\x00", "\x00\x00", "\xB6\xBA", "\xB9\x7D",
  /* U+FA2C */	"\xC0\x5D", "\xC5\x62",

  /* U+FE30 */	"\xA1\x4A", "\xA1\x57", "\x00\x00", "\xA1\x59",
  /* U+FE34 */	"\xA1\x5B", "\xA1\x5F", "\xA1\x60", "\xA1\x63",
  /* U+FE38 */	"\xA1\x64", "\xA1\x67", "\xA1\x68", "\xA1\x6B",
  /* U+FE3C */	"\xA1\x6C", "\xA1\x6F", "\xA1\x70", "\xA1\x73",
  /* U+FE40 */	"\xA1\x74", "\xA1\x77", "\xA1\x78", "\xA1\x7B",
  /* U+FE44 */	"\xA1\x7C", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+FE48 */	"\x00\x00", "\xA1\xC6", "\xA1\xC7", "\xA1\xCA",
  /* U+FE4C */	"\xA1\xCB", "\xA1\xC8", "\xA1\xC9", "\xA1\x5C",
  /* U+FE50 */	"\xA1\x4D", "\xA1\x4E", "\xA1\x4F", "\x00\x00",
  /* U+FE54 */	"\xA1\x51", "\xA1\x52", "\xA1\x53", "\xA1\x54",
  /* U+FE58 */	"\x00\x00", "\xA1\x7D", "\xA1\x7E", "\xA1\xA1",
  /* U+FE5C */	"\xA1\xA2", "\xA1\xA3", "\xA1\xA4", "\xA1\xCC",
  /* U+FE60 */	"\xA1\xCD", "\xA1\xCE", "\xA1\xDE", "\xA1\xDF",
  /* U+FE64 */	"\xA1\xE0", "\xA1\xE1", "\xA1\xE2", "\x00\x00",
  /* U+FE68 */	"\xA2\x42", "\xA2\x4C", "\xA2\x4D", "\xA2\x4E",

  /* U+FF01 */		    "\xA1\x49", "\xC8\xD0", "\xA1\xAD",
  /* U+FF04 */	"\xA2\x43", "\xA2\x48", "\xA1\xAE", "\xC8\xCF",
  /* U+FF08 */	"\xA1\x5D", "\xA1\x5E", "\xA1\xAF", "\xA1\xCF",
  /* U+FF0C */	"\xA1\x41", "\xA1\xD0", "\xA1\x44", "\xA1\xFE",
  /* U+FF10 */	"\xA2\xAF", "\xA2\xB0", "\xA2\xB1", "\xA2\xB2",
  /* U+FF14 */	"\xA2\xB3", "\xA2\xB4", "\xA2\xB5", "\xA2\xB6",
  /* U+FF18 */	"\xA2\xB7", "\xA2\xB8", "\xA1\x47", "\xA1\x46",
  /* U+FF1C */	"\xA1\xD5", "\xA1\xD7", "\xA1\xD6", "\xA1\x48",
  /* U+FF20 */	"\xA2\x49", "\xA2\xCF", "\xA2\xD0", "\xA2\xD1",
  /* U+FF24 */	"\xA2\xD2", "\xA2\xD3", "\xA2\xD4", "\xA2\xD5",
  /* U+FF28 */	"\xA2\xD6", "\xA2\xD7", "\xA2\xD8", "\xA2\xD9",
  /* U+FF2C */	"\xA2\xDA", "\xA2\xDB", "\xA2\xDC", "\xA2\xDD",
  /* U+FF30 */	"\xA2\xDE", "\xA2\xDF", "\xA2\xE0", "\xA2\xE1",
  /* U+FF34 */	"\xA2\xE2", "\xA2\xE3", "\xA2\xE4", "\xA2\xE5",
  /* U+FF38 */	"\xA2\xE6", "\xA2\xE7", "\xA2\xE8", "\xC6\xE4",
  /* U+FF3C */	"\xA2\x40", "\xC6\xE5", "\xA1\x73", "\xA1\xC4",
  /* U+FF40 */	"\xA1\xA5", "\xA2\xE9", "\xA2\xEA", "\xA2\xEB",
  /* U+FF44 */	"\xA2\xEC", "\xA2\xED", "\xA2\xEE", "\xA2\xEF",
  /* U+FF48 */	"\xA2\xF0", "\xA2\xF1", "\xA2\xF2", "\xA2\xF3",
  /* U+FF4C */	"\xA2\xF4", "\xA2\xF5", "\xA2\xF6", "\xA2\xF7",
  /* U+FF50 */	"\xA2\xF8", "\xA2\xF9", "\xA2\xFA", "\xA2\xFB",
  /* U+FF54 */	"\xA2\xFC", "\xA2\xFD", "\xA2\xFE", "\xA3\x40",
  /* U+FF58 */	"\xA3\x41", "\xA3\x42", "\xA3\x43", "\xA1\x61",
  /* U+FF5C */	"\xA1\x55", "\xA1\x62", "\xA1\xE3", "\x00\x00",
  /* U+FF60 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+FF64 */	"\xA1\x4E", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+FF68 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+FF6C */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+FF70 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+FF74 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+FF78 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+FF7C */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+FF80 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+FF84 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+FF88 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+FF8C */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+FF90 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+FF94 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+FF98 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+FF9C */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+FFA0 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+FFA4 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+FFA8 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+FFAC */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+FFB0 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+FFB4 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+FFB8 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+FFBC */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+FFC0 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+FFC4 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+FFC8 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+FFCC */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+FFD0 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+FFD4 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+FFD8 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+FFDC */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+FFE0 */	"\xA2\x46", "\xA2\x47", "\xC8\xCD", "\xA1\xC3",
  /* U+FFE4 */	"\xC8\xCE", "\xA2\x44", "\x00\x00", "\x00\x00",
  /* U+FFE8 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+FFEC */	"\x00\x00", "\xF9\xFE",

  /* U+20021 */		    "\x9C\x71", "\x00\x00", "\x00\x00",
  /* U+20024 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+20028 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+2002C */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+20030 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+20034 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+20038 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+2003C */	"\x00\x00", "\x00\x00", "\x93\x75", "\x00\x00",
  /* U+20040 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+20044 */	"\x00\x00", "\x00\x00", "\x93\x76", "\x00\x00",
  /* U+20048 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+2004C */	"\x00\x00", "\x00\x00", "\x95\x48", "\x00\x00",
  /* U+20050 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+20054 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+20058 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+2005C */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+20060 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+20064 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+20068 */	"\x8E\xC6", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+2006C */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+20070 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+20074 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+20078 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+2007C */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+20080 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+20084 */	"\x00\x00", "\x00\x00", "\x8B\xC5", "\x8B\xFA",
  /* U+20088 */	"\x00\x00", "\x00\x00", "\xC8\x7C", "\x00\x00",
  /* U+2008C */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+20090 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+20094 */	"\x9A\xB4", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+20098 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+2009C */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+200A0 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+200A4 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+200A8 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+200AC */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+200B0 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+200B4 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+200B8 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+200BC */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+200C0 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+200C4 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+200C8 */	"\x00\x00", "\x00\x00", "\x88\x4E", "\x88\x4B",
  /* U+200CC */	"\xC8\x7A", "\x88\x48", "\x00\x00", "\x00\x00",
  /* U+200D0 */	"\x00\x00", "\x88\x47", "\x00\x00", "\x00\x00",
  /* U+200D4 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+200D8 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+200DC */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+200E0 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+200E4 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+200E8 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+200EC */	"\x00\x00", "\x00\x00", "\xA0\xF6", "\x00\x00",
  /* U+200F0 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+200F4 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+200F8 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+200FC */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+20100 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+20104 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+20108 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+2010C */	"\x88\x45", "\x00\x00", "\x88\x53", "\x00\x00",
  /* U+20110 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+20114 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+20118 */	"\xFC\xAD",

  /* U+201A9 */		    "\x8A\xAD", "\x00\x00", "\x92\x72",
  /* U+201AC */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+201B0 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+201B4 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+201B8 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+201BC */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+201C0 */	"\x00\x00", "\xFC\x47", "\x00\x00", "\x00\x00",
  /* U+201C4 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+201C8 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+201CC */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+201D0 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+201D4 */	"\x94\xDF", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+201D8 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+201DC */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+201E0 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+201E4 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+201E8 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+201EC */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+201F0 */	"\x00\x00", "\x00\x00", "\x9F\xD1", "\x00\x00",
  /* U+201F4 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+201F8 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+201FC */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+20200 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+20204 */	"\xFB\xCB", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+20208 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+2020C */	"\x92\x7D", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+20210 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+20214 */	"\x98\xA4", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+20218 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+2021C */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+20220 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+20224 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+20228 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+2022C */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+20230 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+20234 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+20238 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+2023C */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+20240 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+20244 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+20248 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+2024C */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+20250 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+20254 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+20258 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x94\xE7",
  /* U+2025C */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+20260 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+20264 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+20268 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+2026C */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+20270 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+20274 */	"\x90\xCB", "\x92\x7B", "\x00\x00", "\x00\x00",
  /* U+20278 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+2027C */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+20280 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+20284 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+20288 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+2028C */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+20290 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+20294 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+20298 */	"\x00\x00", "\x94\xD8", "\x00\x00", "\x00\x00",
  /* U+2029C */	"\x00\x00", "\x00\x00", "\xFC\x5F", "\x00\x00",
  /* U+202A0 */	"\xFA\x54", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+202A4 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+202A8 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+202AC */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+202B0 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+202B4 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x9A\xB5",
  /* U+202B8 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+202BC */	"\x00\x00", "\x00\x00", "\x00\x00", "\x96\xDA",
  /* U+202C0 */	"\x92\x79", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+202C4 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+202C8 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+202CC */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+202D0 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+202D4 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+202D8 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+202DC */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+202E0 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+202E4 */	"\x00\x00", "\xFA\x74", "\x00\x00", "\x00\x00",
  /* U+202E8 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+202EC */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+202F0 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+202F4 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+202F8 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+202FC */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+20300 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+20304 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+20308 */	"\x00\x00", "\x00\x00", "\x92\x75", "\x00\x00",
  /* U+2030C */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+20310 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+20314 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+20318 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+2031C */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+20320 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+20324 */	"\x00\x00", "\x8D\xFB", "\x00\x00", "\x00\x00",
  /* U+20328 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+2032C */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+20330 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+20334 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+20338 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+2033C */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+20340 */	"\x00\x00", "\x8A\x49", "\x00\x00", "\x00\x00",
  /* U+20344 */	"\x00\x00", "\x92\xDF", "\x9B\x7C", "\xFA\x63",
  /* U+20348 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+2034C */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+20350 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+20354 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+20358 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+2035C */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+20360 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+20364 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+20368 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+2036C */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+20370 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+20374 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+20378 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+2037C */	"\x00\x00", "\x00\x00", "\xFA\x60", "\x92\x6D",
  /* U+20380 */	"\xFA\x62", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+20384 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+20388 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+2038C */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+20390 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+20394 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+20398 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+2039C */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+203A0 */	"\x9A\xB6", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+203A4 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x97\x6B",
  /* U+203A8 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+203AC */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+203B0 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+203B4 */	"\x00\x00", "\xFD\x6A", "\x00\x00", "\x00\x00",
  /* U+203B8 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+203BC */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+203C0 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+203C4 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+203C8 */	"\x00\x00", "\xFD\x54", "\x00\x00", "\x92\x73",
  /* U+203CC */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+203D0 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+203D4 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+203D8 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+203DC */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+203E0 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+203E4 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+203E8 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+203EC */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+203F0 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+203F4 */	"\x00\x00", "\x97\xD8", "\x00\x00", "\x00\x00",
  /* U+203F8 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+203FC */	"\x9F\xBB", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+20400 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+20404 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+20408 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+2040C */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+20410 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x93\x42",
  /* U+20414 */	"\x92\x76", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+20418 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+2041C */	"\x00\x00", "\x00\x00", "\x00\x00", "\xFA\x65",
  /* U+20420 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+20424 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+20428 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+2042C */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+20430 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+20434 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+20438 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+2043C */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+20440 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+20444 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+20448 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+2044C */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+20450 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+20454 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+20458 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+2045C */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+20460 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+20464 */	"\x00\x00", "\x92\x6C", "\x00\x00", "\x00\x00",
  /* U+20468 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+2046C */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+20470 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+20474 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+20478 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+2047C */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+20480 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+20484 */	"\x00\x00", "\x00\x00", "\x00\x00", "\xFA\x6E",
  /* U+20488 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+2048C */	"\x00\x00", "\x00\x00", "\x9E\xE0", "\x00\x00",
  /* U+20490 */	"\x00\x00", "\x92\xC0", "\x92\xBF", "\x00\x00",
  /* U+20494 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+20498 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+2049C */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+204A0 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x92\xBE",
  /* U+204A4 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+204A8 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+204AC */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+204B0 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+204B4 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+204B8 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+204BC */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+204C0 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+204C4 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+204C8 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+204CC */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+204D0 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+204D4 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x9A\xBA",
  /* U+204D8 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+204DC */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+204E0 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+204E4 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+204E8 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+204EC */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+204F0 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+204F4 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+204F8 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+204FC */	"\x8A\xB3", "\x00\x00", "\x97\x75", "\x00\x00",
  /* U+20500 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+20504 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+20508 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+2050C */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+20510 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+20514 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+20518 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+2051C */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+20520 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+20524 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+20528 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+2052C */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+20530 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+20534 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+20538 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+2053C */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+20540 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+20544 */	"\x00\x00", "\x00\x00", "\x00\x00", "\xFA\x40",
  /* U+20548 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+2054C */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+20550 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+20554 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+20558 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+2055C */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+20560 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+20564 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+20568 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+2056C */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+20570 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+20574 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+20578 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+2057C */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+20580 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+20584 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+20588 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+2058C */	"\x00\x00", "\x00\x00", "\xFA\x76", "\x00\x00",
  /* U+20590 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+20594 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+20598 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+2059C */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+205A0 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+205A4 */	"\x00\x00", "\xFB\xD0", "\x00\x00", "\x00\x00",
  /* U+205A8 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+205AC */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+205B0 */	"\x00\x00", "\x00\x00", "\x00\x00", "\xFA\x7B",
  /* U+205B4 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+205B8 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+205BC */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+205C0 */	"\x00\x00", "\x00\x00", "\x00\x00", "\xFE\x6D",
  /* U+205C4 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+205C8 */	"\x00\x00", "\x00\x00", "\x9B\xB3", "\x00\x00",
  /* U+205CC */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+205D0 */	"\x89\xCC", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+205D4 */	"\x00\x00", "\x9A\xBE", "\x00\x00", "\x00\x00",
  /* U+205D8 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+205DC */	"\x00\x00", "\x00\x00", "\x00\x00", "\xFA\x42",
  /* U+205E0 */	"\x92\xBC", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+205E4 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+205E8 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x94\x5C",
  /* U+205EC */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+205F0 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+205F4 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+205F8 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+205FC */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+20600 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+20604 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+20608 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+2060C */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+20610 */	"\x00\x00", "\x9B\xB5", "\x00\x00", "\x00\x00",
  /* U+20614 */	"\x00\x00", "\x9A\xBF", "\x00\x00", "\x00\x00",
  /* U+20618 */	"\x00\x00", "\x98\xA7", "\x97\xA4", "\x00\x00",
  /* U+2061C */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+20620 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+20624 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+20628 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+2062C */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+20630 */	"\x90\xFD", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+20634 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+20638 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+2063C */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+20640 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+20644 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+20648 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+2064C */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+20650 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+20654 */	"\x00\x00", "\x00\x00", "\xFC\x7B", "\x00\x00",
  /* U+20658 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+2065C */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+20660 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+20664 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+20668 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+2066C */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+20670 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+20674 */	"\x00\x00", "\x00\x00", "\x9A\xC0",

  /* U+2070E */				"\x92\xC3", "\x00\x00",
  /* U+20710 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+20714 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+20718 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+2071C */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+20720 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+20724 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+20728 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+2072C */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+20730 */	"\x00\x00", "\x8A\xAA", "\x00\x00", "\x00\x00",
  /* U+20734 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+20738 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+2073C */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+20740 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+20744 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+20748 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+2074C */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+20750 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+20754 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+20758 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+2075C */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+20760 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+20764 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+20768 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+2076C */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+20770 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+20774 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+20778 */	"\x00\x00", "\x9B\xD0",

  /* U+2082C */	"\x95\x50", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+20830 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+20834 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+20838 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+2083C */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+20840 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+20844 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+20848 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+2084C */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+20850 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+20854 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+20858 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+2085C */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+20860 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+20864 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+20868 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+2086C */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+20870 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x92\xC6",
  /* U+20874 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+20878 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+2087C */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+20880 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+20884 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+20888 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+2088C */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+20890 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+20894 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+20898 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+2089C */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+208A0 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+208A4 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+208A8 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+208AC */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+208B0 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+208B4 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+208B8 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+208BC */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+208C0 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+208C4 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+208C8 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+208CC */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+208D0 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+208D4 */	"\x00\x00", "\x98\xA6", "\x00\x00", "\x00\x00",
  /* U+208D8 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+208DC */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+208E0 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+208E4 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+208E8 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+208EC */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+208F0 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+208F4 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+208F8 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+208FC */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+20900 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+20904 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+20908 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+2090C */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+20910 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+20914 */	"\x00\x00", "\x00\x00", "\x95\x46", "\x00\x00",
  /* U+20918 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+2091C */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+20920 */	"\x00\x00", "\x00\x00", "\x00\x00", "\xFD\x63",
  /* U+20924 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+20928 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+2092C */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+20930 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+20934 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+20938 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+2093C */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+20940 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+20944 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+20948 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+2094C */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+20950 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+20954 */	"\xFA\xC2", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+20958 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+2095C */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+20960 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+20964 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+20968 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+2096C */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+20970 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+20974 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+20978 */	"\x00\x00", "\x9E\xC3", "\x00\x00", "\x00\x00",
  /* U+2097C */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+20980 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+20984 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+20988 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+2098C */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+20990 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+20994 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+20998 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+2099C */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+209A0 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+209A4 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+209A8 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+209AC */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+209B0 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+209B4 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+209B8 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+209BC */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+209C0 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+209C4 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+209C8 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+209CC */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+209D0 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+209D4 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+209D8 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+209DC */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+209E0 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+209E4 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x89\xB2",
  /* U+209E8 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+209EC */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+209F0 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+209F4 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+209F8 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+209FC */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+20A00 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+20A04 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+20A08 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+20A0C */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+20A10 */	"\x00\x00", "\x9C\x66", "\x00\x00", "\x00\x00",
  /* U+20A14 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+20A18 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+20A1C */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+20A20 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+20A24 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+20A28 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+20A2C */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+20A30 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+20A34 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+20A38 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+20A3C */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+20A40 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+20A44 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+20A48 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+20A4C */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+20A50 */	"\x90\x53", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+20A54 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+20A58 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+20A5C */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+20A60 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+20A64 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+20A68 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+20A6C */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+20A70 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+20A74 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+20A78 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+20A7C */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+20A80 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+20A84 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+20A88 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+20A8C */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+20A90 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+20A94 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+20A98 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+20A9C */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+20AA0 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+20AA4 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+20AA8 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+20AAC */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+20AB0 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+20AB4 */	"\x97\xC1", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+20AB8 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+20ABC */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+20AC0 */	"\x00\x00", "\x00\x00", "\x9A\xC4", "\x00\x00",
  /* U+20AC4 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+20AC8 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+20ACC */	"\x00\x00", "\x9A\xC5", "\x00\x00", "\x00\x00",
  /* U+20AD0 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+20AD4 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+20AD8 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+20ADC */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+20AE0 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+20AE4 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+20AE8 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+20AEC */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+20AF0 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+20AF4 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+20AF8 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+20AFC */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+20B00 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+20B04 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+20B08 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+20B0C */	"\x00\x00", "\x8E\xEF",

  /* U+20B8F */					    "\xFA\xE9",
  /* U+20B90 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+20B94 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+20B98 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+20B9C */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+20BA0 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+20BA4 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+20BA8 */	"\x92\x62", "\x8A\xF7", "\x00\x00", "\x00\x00",
  /* U+20BAC */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+20BB0 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+20BB4 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+20BB8 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+20BBC */	"\x00\x00", "\x00\x00", "\x00\x00", "\x9A\xC6",
  /* U+20BC0 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+20BC4 */	"\x00\x00", "\x00\x00", "\x92\xE1", "\x00\x00",
  /* U+20BC8 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x9A\xC9",
  /* U+20BCC */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+20BD0 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+20BD4 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+20BD8 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+20BDC */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+20BE0 */	"\x00\x00", "\x00\x00", "\xFA\xC6", "\x00\x00",
  /* U+20BE4 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+20BE8 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x97\xA5",
  /* U+20BEC */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+20BF0 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+20BF4 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+20BF8 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x9A\xCB",
  /* U+20BFC */	"\x00\x00", "\x00\x00", "\x00\x00", "\xFA\x72",
  /* U+20C00 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+20C04 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+20C08 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x8A\x5E",
  /* U+20C0C */	"\x00\x00", "\x94\xE0", "\x00\x00", "\x00\x00",
  /* U+20C10 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+20C14 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+20C18 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+20C1C */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+20C20 */	"\x92\xCC", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+20C24 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+20C28 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+20C2C */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+20C30 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+20C34 */	"\x8A\xE5", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+20C38 */	"\x00\x00", "\x00\x00", "\xFE\x5C", "\x9A\xCC",
  /* U+20C3C */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+20C40 */	"\x00\x00", "\x9D\xF9", "\x8A\x43", "\x8A\xA6",
  /* U+20C44 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+20C48 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+20C4C */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+20C50 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x9A\xCD",
  /* U+20C54 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+20C58 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+20C5C */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+20C60 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+20C64 */	"\x00\x00", "\x9A\xCE", "\x00\x00", "\x00\x00",
  /* U+20C68 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+20C6C */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+20C70 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+20C74 */	"\x00\x00", "\x00\x00", "\x00\x00", "\xFA\xEE",
  /* U+20C78 */	"\x9B\xCC", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+20C7C */	"\x9A\xCF", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+20C80 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+20C84 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+20C88 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+20C8C */	"\x00\x00", "\x9A\xD1", "\x00\x00", "\x00\x00",
  /* U+20C90 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+20C94 */	"\x00\x00", "\x00\x00", "\x9D\xFA", "\x00\x00",
  /* U+20C98 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+20C9C */	"\x9D\x7C", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+20CA0 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+20CA4 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+20CA8 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+20CAC */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+20CB0 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+20CB4 */	"\x00\x00", "\x9A\xD3", "\x00\x00", "\x00\x00",
  /* U+20CB8 */	"\x97\xA6", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+20CBC */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+20CC0 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+20CC4 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+20CC8 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+20CCC */	"\x00\x00", "\x00\x00", "\x00\x00", "\x99\x5F",
  /* U+20CD0 */	"\x00\x00", "\x00\x00", "\x00\x00", "\xFB\xF6",
  /* U+20CD4 */	"\x9F\xC5", "\x8A\x59", "\x8B\x6B", "\x00\x00",
  /* U+20CD8 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+20CDC */	"\x00\x00", "\x9A\xD4", "\x00\x00", "\x00\x00",
  /* U+20CE0 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+20CE4 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+20CE8 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+20CEC */	"\x00\x00", "\x9A\xD5", "\x00\x00", "\x00\x00",
  /* U+20CF0 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+20CF4 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+20CF8 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+20CFC */	"\x00\x00", "\x00\x00", "\x00\x00", "\x97\xA2",
  /* U+20D00 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+20D04 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+20D08 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+20D0C */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+20D10 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+20D14 */	"\x00\x00", "\x8A\x44", "\x00\x00", "\x00\x00",
  /* U+20D18 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+20D1C */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+20D20 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+20D24 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+20D28 */	"\x9F\x4A", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+20D2C */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+20D30 */	"\x00\x00", "\x90\xA1", "\xFD\xA4", "\x00\x00",
  /* U+20D34 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+20D38 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+20D3C */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+20D40 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+20D44 */	"\x00\x00", "\x00\x00", "\x8A\x64", "\x8A\xF2",
  /* U+20D48 */	"\x8A\xF8", "\x9D\xD8", "\x00\x00", "\x00\x00",
  /* U+20D4C */	"\x94\xD6", "\xFA\xFE", "\xFB\xA7", "\x00\x00",
  /* U+20D50 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+20D54 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+20D58 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+20D5C */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+20D60 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+20D64 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+20D68 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+20D6C */	"\x00\x00", "\x00\x00", "\x00\x00", "\x9A\xD6",
  /* U+20D70 */	"\x00\x00", "\x9F\x4D", "\x00\x00", "\x00\x00",
  /* U+20D74 */	"\xFA\xF6", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+20D78 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+20D7C */	"\x8A\x57", "\x00\x00", "\x8B\x43", "\x8B\x44",
  /* U+20D80 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+20D84 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+20D88 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+20D8C */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+20D90 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+20D94 */	"\x00\x00", "\x00\x00", "\x8A\xB6", "\x00\x00",
  /* U+20D98 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+20D9C */	"\x8A\xC0", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+20DA0 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+20DA4 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x9E\x54",
  /* U+20DA8 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+20DAC */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+20DB0 */	"\x00\x00", "\x00\x00", "\x9A\xD7", "\x00\x00",
  /* U+20DB4 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+20DB8 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+20DBC */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+20DC0 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+20DC4 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+20DC8 */	"\x9A\xD8", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+20DCC */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+20DD0 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+20DD4 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+20DD8 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+20DDC */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+20DE0 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+20DE4 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+20DE8 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+20DEC */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+20DF0 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+20DF4 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+20DF8 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+20DFC */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+20E00 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+20E04 */	"\x9A\xDC", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+20E08 */	"\x00\x00", "\x8A\xCA", "\x9E\xA8", "\x00\x00",
  /* U+20E0C */	"\x00\x00", "\x92\x63", "\x9A\xDD", "\x8B\x65",
  /* U+20E10 */	"\x8B\x6F", "\x8B\x7E", "\x00\x00", "\x00\x00",
  /* U+20E14 */	"\x00\x00", "\x00\x00", "\x8F\x43", "\x00\x00",
  /* U+20E18 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+20E1C */	"\x00\x00", "\x92\xD0", "\x00\x00", "\x00\x00",
  /* U+20E20 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+20E24 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+20E28 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+20E2C */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+20E30 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+20E34 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+20E38 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+20E3C */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+20E40 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+20E44 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+20E48 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+20E4C */	"\x8A\xF4", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+20E50 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+20E54 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+20E58 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+20E5C */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+20E60 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+20E64 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+20E68 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+20E6C */	"\x00\x00", "\x9D\xBE", "\x00\x00", "\x00\x00",
  /* U+20E70 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x9A\xE1",
  /* U+20E74 */	"\x00\x00", "\xFC\xDE", "\x9D\xFD", "\x8B\x66",
  /* U+20E78 */	"\x8B\x70", "\x8B\x75", "\x8A\xE4", "\x8B\xA4",
  /* U+20E7C */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+20E80 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+20E84 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+20E88 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+20E8C */	"\x8A\xED", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+20E90 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+20E94 */	"\x00\x00", "\x00\x00", "\x8A\x5D", "\x00\x00",
  /* U+20E98 */	"\x8B\x48", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+20E9C */	"\x00\x00", "\x9D\xED", "\x00\x00", "\x00\x00",
  /* U+20EA0 */	"\x00\x00", "\x00\x00", "\x9E\x40", "\x00\x00",
  /* U+20EA4 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+20EA8 */	"\x00\x00", "\x00\x00", "\x8A\xEF", "\x8A\xF6",
  /* U+20EAC */	"\x9E\x76", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+20EB0 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+20EB4 */	"\x00\x00", "\x00\x00", "\x9E\xE3", "\x00\x00",
  /* U+20EB8 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+20EBC */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+20EC0 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+20EC4 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+20EC8 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+20ECC */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+20ED0 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+20ED4 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x9A\xDE",
  /* U+20ED8 */	"\x8D\xFE", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+20EDC */	"\x00\x00", "\xFA\xFC", "\x00\x00", "\x00\x00",
  /* U+20EE0 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+20EE4 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+20EE8 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+20EEC */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+20EF0 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+20EF4 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+20EF8 */	"\x9C\xB1", "\x9E\x77", "\x8B\x64", "\x8B\x67",
  /* U+20EFC */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+20F00 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+20F04 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+20F08 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+20F0C */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+20F10 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+20F14 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+20F18 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+20F1C */	"\x00\x00", "\x97\x4B", "\x00\x00", "\x00\x00",
  /* U+20F20 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+20F24 */	"\x00\x00", "\x00\x00", "\x96\x53", "\x00\x00",
  /* U+20F28 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+20F2C */	"\x00\x00", "\x9A\xE0", "\x8B\x4A", "\x00\x00",
  /* U+20F30 */	"\x8A\xF1", "\x8A\xD7", "\x00\x00", "\x00\x00",
  /* U+20F34 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+20F38 */	"\x00\x00", "\x00\x00", "\x00\x00", "\xA0\xAB",
  /* U+20F3C */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+20F40 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+20F44 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+20F48 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+20F4C */	"\x8A\xB5", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+20F50 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+20F54 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+20F58 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+20F5C */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+20F60 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+20F64 */	"\x8A\x5F", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+20F68 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+20F6C */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+20F70 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+20F74 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+20F78 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+20F7C */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+20F80 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+20F84 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+20F88 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+20F8C */	"\x00\x00", "\x8A\xEE", "\x00\x00", "\x00\x00",
  /* U+20F90 */	"\x9A\xDF", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+20F94 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+20F98 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+20F9C */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+20FA0 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+20FA4 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+20FA8 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+20FAC */	"\x00\x00", "\x8A\xFE", "\x00\x00", "\x00\x00",
  /* U+20FB0 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+20FB4 */	"\x8A\x58", "\x8B\xA3", "\x8B\xA7", "\x00\x00",
  /* U+20FB8 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+20FBC */	"\x9A\xE3", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+20FC0 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+20FC4 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+20FC8 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+20FCC */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+20FD0 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+20FD4 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+20FD8 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+20FDC */	"\x00\x00", "\x00\x00", "\x00\x00", "\x92\x61",
  /* U+20FE0 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+20FE4 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+20FE8 */	"\x00\x00", "\x00\x00", "\x9D\xD7", "\x9E\x7D",
  /* U+20FEC */	"\x9E\xA7", "\x9E\xAB", "\x00\x00", "\x00\x00",
  /* U+20FF0 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+20FF4 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+20FF8 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+20FFC */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+21000 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+21004 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+21008 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+2100C */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+21010 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+21014 */	"\x90\x42", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+21018 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+2101C */	"\x00\x00", "\x8B\x79", "\x8B\x7A", "\x00\x00",
  /* U+21020 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+21024 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+21028 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+2102C */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+21030 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+21034 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+21038 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+2103C */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+21040 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+21044 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+21048 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+2104C */	"\x00\x00", "\x00\x00", "\x00\x00", "\x9A\xE6",
  /* U+21050 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+21054 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+21058 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+2105C */	"\x9A\xE5", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+21060 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+21064 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+21068 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+2106C */	"\x00\x00", "\x00\x00", "\x00\x00", "\x8A\x7E",
  /* U+21070 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+21074 */	"\x00\x00", "\x9E\x44", "\x9A\xE7", "\x8A\x7C",
  /* U+21078 */	"\x8B\x71", "\x00\x00", "\x00\x00", "\x9A\xE9",
  /* U+2107C */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+21080 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+21084 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+21088 */	"\x9A\xEA", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+2108C */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+21090 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+21094 */	"\x00\x00", "\x00\x00", "\x9A\xEB", "\x00\x00",
  /* U+21098 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+2109C */	"\x00\x00", "\x8A\xBD", "\x00\x00", "\x00\x00",
  /* U+210A0 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+210A4 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+210A8 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+210AC */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+210B0 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+210B4 */	"\xFB\x4E", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+210B8 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+210BC */	"\x00\x00", "\x00\x00", "\x00\x00", "\x9A\xED",
  /* U+210C0 */	"\x8A\xF9", "\x9E\x63", "\x00\x00", "\x00\x00",
  /* U+210C4 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x8B\x49",
  /* U+210C8 */	"\x8A\xCE", "\x8B\x6E", "\x00\x00", "\x00\x00",
  /* U+210CC */	"\x00\x00", "\x00\x00", "\x00\x00", "\x8A\xE8",
  /* U+210D0 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x9A\xEE",
  /* U+210D4 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+210D8 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+210DC */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+210E0 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+210E4 */	"\x92\xCE", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+210E8 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+210EC */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+210F0 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+210F4 */	"\x8A\x5A", "\x8B\x7B", "\x8B\x7C", "\x00\x00",
  /* U+210F8 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+210FC */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+21100 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+21104 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+21108 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+2110C */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+21110 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+21114 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+21118 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+2111C */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+21120 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+21124 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+21128 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+2112C */	"\x00\x00", "\x00\x00", "\x00\x00", "\x9A\xEF",
  /* U+21130 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+21134 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+21138 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x9A\xF0",
  /* U+2113C */	"\x00\x00", "\x8A\xFA", "\x00\x00", "\x00\x00",
  /* U+21140 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+21144 */	"\x00\x00", "\x89\x41", "\x00\x00", "\x00\x00",
  /* U+21148 */	"\x8B\x72", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+2114C */	"\x00\x00", "\x00\x00", "\x00\x00", "\x8A\xF3",
  /* U+21150 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+21154 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+21158 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+2115C */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+21160 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+21164 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+21168 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+2116C */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+21170 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+21174 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+21178 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+2117C */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+21180 */	"\x8B\xA8", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+21184 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x9E\xAE",
  /* U+21188 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+2118C */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+21190 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+21194 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+21198 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+2119C */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+211A0 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+211A4 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+211A8 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+211AC */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+211B0 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+211B4 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+211B8 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+211BC */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+211C0 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+211C4 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+211C8 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+211CC */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+211D0 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+211D4 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+211D8 */	"\x00\x00", "\x9E\x72", "\x00\x00", "\x00\x00",
  /* U+211DC */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+211E0 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+211E4 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+211E8 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+211EC */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+211F0 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+211F4 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+211F8 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+211FC */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+21200 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+21204 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+21208 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+2120C */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+21210 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+21214 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+21218 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+2121C */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+21220 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+21224 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+21228 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+2122C */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+21230 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+21234 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+21238 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+2123C */	"\xFB\x73", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+21240 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+21244 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+21248 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+2124C */	"\x00\x00", "\x00\x00", "\x00\x00", "\xFB\x5F",
  /* U+21250 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+21254 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+21258 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+2125C */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+21260 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+21264 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+21268 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+2126C */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+21270 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+21274 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+21278 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+2127C */	"\x90\xBA", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+21280 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+21284 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+21288 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+2128C */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+21290 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+21294 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+21298 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+2129C */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+212A0 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+212A4 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+212A8 */	"\x91\xFE", "\x9E\xF6", "\x00\x00", "\x00\x00",
  /* U+212AC */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+212B0 */	"\x97\xED", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+212B4 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+212B8 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+212BC */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+212C0 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+212C4 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+212C8 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+212CC */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+212D0 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+212D4 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+212D8 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+212DC */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+212E0 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x9A\xF3",
  /* U+212E4 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+212E8 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+212EC */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+212F0 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+212F4 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+212F8 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+212FC */	"\x00\x00", "\x00\x00", "\xA0\xEE", "\x00\x00",
  /* U+21300 */	"\x00\x00", "\x00\x00", "\x96\x7C", "\x93\x45",
  /* U+21304 */	"\x98\x6E", "\xFA\x56", "\x00\x00", "\x00\x00",
  /* U+21308 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+2130C */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+21310 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+21314 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+21318 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+2131C */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+21320 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+21324 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+21328 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+2132C */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+21330 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+21334 */	"\x00\x00", "\x00\x00", "\x9A\xF5", "\x00\x00",
  /* U+21338 */	"\x00\x00", "\x00\x00", "\xFC\x4B", "\x00\x00",
  /* U+2133C */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+21340 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+21344 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+21348 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+2134C */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+21350 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+21354 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+21358 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+2135C */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+21360 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+21364 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+21368 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+2136C */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+21370 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+21374 */	"\x00\x00", "\x9A\xF4", "\xFE\xDE", "\x00\x00",
  /* U+21378 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+2137C */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+21380 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+21384 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+21388 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+2138C */	"\x00\x00", "\x00\x00", "\xFC\xB7", "\x00\x00",
  /* U+21390 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+21394 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+21398 */	"\x97\xF1", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+2139C */	"\x97\xC7", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+213A0 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+213A4 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+213A8 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+213AC */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+213B0 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+213B4 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+213B8 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+213BC */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+213C0 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+213C4 */	"\x00\x00", "\x9C\xCB", "\x92\x40", "\x00\x00",
  /* U+213C8 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+213CC */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+213D0 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+213D4 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+213D8 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+213DC */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+213E0 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+213E4 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+213E8 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+213EC */	"\x00\x00", "\x9C\xE8", "\x00\x00", "\x00\x00",
  /* U+213F0 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+213F4 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+213F8 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+213FC */	"\x00\x00", "\x00\x00", "\x91\xFD", "\x00\x00",
  /* U+21400 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+21404 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+21408 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+2140C */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+21410 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x97\x4E",
  /* U+21414 */	"\x00\x00", "\x00\x00", "\xFB\x68", "\x00\x00",
  /* U+21418 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+2141C */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+21420 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+21424 */	"\x97\x6C", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+21428 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+2142C */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+21430 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+21434 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+21438 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+2143C */	"\x00\x00", "\x00\x00", "\x00\x00", "\x8C\xC2",
  /* U+21440 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+21444 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+21448 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+2144C */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+21450 */	"\x00\x00", "\x00\x00", "\x97\xE8", "\x00\x00",
  /* U+21454 */	"\xFB\x6A", "\x8B\x74", "\x00\x00", "\x00\x00",
  /* U+21458 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+2145C */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+21460 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+21464 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+21468 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+2146C */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+21470 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+21474 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+21478 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+2147C */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+21480 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+21484 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+21488 */	"\x00\x00", "\x00\x00", "\x8E\xE7", "\x00\x00",
  /* U+2148C */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+21490 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+21494 */	"\x00\x00", "\x00\x00", "\x00\x00", "\xFD\xC8",
  /* U+21498 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+2149C */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+214A0 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+214A4 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+214A8 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+214AC */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+214B0 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+214B4 */	"\x00\x00", "\x00\x00", "\x92\x41", "\x00\x00",
  /* U+214B8 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+214BC */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+214C0 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+214C4 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+214C8 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+214CC */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+214D0 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+214D4 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+214D8 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+214DC */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+214E0 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+214E4 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+214E8 */	"\x96\xA1", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+214EC */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+214F0 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+214F4 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+214F8 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+214FC */	"\x00\x00", "\x8E\xF3", "\x00\x00", "\x00\x00",
  /* U+21500 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+21504 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+21508 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+2150C */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+21510 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+21514 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+21518 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+2151C */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+21520 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+21524 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+21528 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+2152C */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+21530 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+21534 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+21538 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+2153C */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+21540 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+21544 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+21548 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+2154C */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+21550 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+21554 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+21558 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+2155C */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+21560 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+21564 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+21568 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+2156C */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+21570 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+21574 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x9A\xF7",
  /* U+21578 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+2157C */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+21580 */	"\x00\x00", "\x00\x00", "\x8F\xA6", "\x00\x00",
  /* U+21584 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+21588 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+2158C */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+21590 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+21594 */	"\x00\x00", "\x00\x00", "\xFA\xD6", "\x00\x00",
  /* U+21598 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+2159C */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+215A0 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+215A4 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+215A8 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+215AC */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+215B0 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+215B4 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+215B8 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+215BC */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+215C0 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+215C4 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+215C8 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+215CC */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+215D0 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+215D4 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+215D8 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+215DC */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+215E0 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+215E4 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+215E8 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+215EC */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+215F0 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+215F4 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+215F8 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+215FC */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+21600 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+21604 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+21608 */	"\x00\x00", "\x00\x00", "\x9C\xC7", "\x00\x00",
  /* U+2160C */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+21610 */	"\x00\x00", "\x00\x00", "\x00\x00", "\xFA\xD7",
  /* U+21614 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+21618 */	"\x00\x00", "\x9A\xF8", "\x00\x00", "\x00\x00",
  /* U+2161C */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+21620 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+21624 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+21628 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+2162C */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+21630 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+21634 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+21638 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+2163C */	"\x00\x00", "\x00\x00", "\xFB\xA1", "\x00\x00",
  /* U+21640 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+21644 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+21648 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+2164C */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+21650 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+21654 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+21658 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+2165C */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+21660 */	"\x00\x00", "\x8E\xC5", "\x00\x00", "\x00\x00",
  /* U+21664 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+21668 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+2166C */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+21670 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+21674 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+21678 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+2167C */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+21680 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+21684 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+21688 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+2168C */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+21690 */	"\x00\x00", "\x00\x00", "\xFB\xA4", "\x00\x00",
  /* U+21694 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+21698 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+2169C */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+216A0 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+216A4 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+216A8 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+216AC */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+216B0 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+216B4 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+216B8 */	"\xFB\xC2", "\x00\x00", "\x9A\xC1", "\x00\x00",
  /* U+216BC */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+216C0 */	"\x91\xFA", "\xFE\xDB", "\x97\xAB", "\x00\x00",
  /* U+216C4 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+216C8 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+216CC */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+216D0 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x91\x47",
  /* U+216D4 */	"\x00\x00", "\xFB\xB1", "\x00\x00", "\x00\x00",
  /* U+216D8 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+216DC */	"\x00\x00", "\x00\x00", "\x00\x00", "\x8F\xEA",
  /* U+216E0 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+216E4 */	"\x00\x00", "\x00\x00", "\x94\xD2", "\xFE\x61",
  /* U+216E8 */	"\xFA\xCE", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+216EC */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+216F0 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+216F4 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+216F8 */	"\x00\x00", "\x00\x00", "\x92\xED", "\x91\xF3",
  /* U+216FC */	"\x93\xC6", "\x00\x00", "\x93\x5A", "\x00\x00",
  /* U+21700 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+21704 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+21708 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+2170C */	"\x00\x00", "\xFA\xFB", "\x00\x00", "\x00\x00",
  /* U+21710 */	"\x92\xEF", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+21714 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+21718 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+2171C */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+21720 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+21724 */	"\x00\x00", "\x00\x00", "\xFA\xC8", "\x00\x00",
  /* U+21728 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+2172C */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+21730 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+21734 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+21738 */	"\x00\x00", "\x00\x00", "\x98\x47", "\x93\x66",
  /* U+2173C */	"\x98\x55", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+21740 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+21744 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+21748 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+2174C */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+21750 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+21754 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x96\xE6",
  /* U+21758 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+2175C */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+21760 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+21764 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+21768 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+2176C */	"\x9F\x43", "\x9F\xAA", "\x94\xDA", "\x92\xEE",
  /* U+21770 */	"\xFC\xAF", "\xFB\xFB", "\x00\x00", "\x8E\xF9",
  /* U+21774 */	"\x91\xF6", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+21778 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+2177C */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+21780 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+21784 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+21788 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+2178C */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+21790 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+21794 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+21798 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+2179C */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+217A0 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+217A4 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+217A8 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x93\x64",
  /* U+217AC */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+217B0 */	"\x94\xF5", "\x9C\xB6", "\xFB\xAD", "\x98\x4E",
  /* U+217B4 */	"\x8F\x44", "\x96\xFD", "\x00\x00", "\x00\x00",
  /* U+217B8 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+217BC */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+217C0 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x9A\xF9",
  /* U+217C4 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x9A\xFA",
  /* U+217C8 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+217CC */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+217D0 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+217D4 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+217D8 */	"\x00\x00", "\x97\x69", "\x95\xD4", "\x98\x4B",
  /* U+217DC */	"\xFB\xAA", "\x00\x00", "\x00\x00", "\x98\x7C",
  /* U+217E0 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+217E4 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+217E8 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+217EC */	"\x00\x00", "\x00\x00", "\x00\x00", "\x91\xEA",
  /* U+217F0 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+217F4 */	"\x00\x00", "\x9D\xAF", "\x9D\xC5", "\x00\x00",
  /* U+217F8 */	"\x91\xF1", "\x8E\xB1", "\x97\xA9", "\xFB\xAC",
  /* U+217FC */	"\xFC\xB8", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+21800 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+21804 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+21808 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+2180C */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+21810 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+21814 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+21818 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+2181C */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+21820 */	"\x9C\xB9", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+21824 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+21828 */	"\xFB\xB0", "\xFC\xD2", "\x93\xCB", "\x00\x00",
  /* U+2182C */	"\x00\x00", "\x9A\xFD", "\x00\x00", "\x00\x00",
  /* U+21830 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+21834 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+21838 */	"\x00\x00", "\x91\xF4", "\x8B\xAC", "\xA0\x55",
  /* U+2183C */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+21840 */	"\x95\x74", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+21844 */	"\x00\x00", "\x95\xBE", "\x00\x00", "\x00\x00",
  /* U+21848 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+2184C */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+21850 */	"\x00\x00", "\x00\x00", "\x97\xAD", "\x00\x00",
  /* U+21854 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+21858 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+2185C */	"\x00\x00", "\x00\x00", "\x8E\xE9", "\x00\x00",
  /* U+21860 */	"\x00\x00", "\x92\xF8", "\x97\xBE", "\x91\x6C",
  /* U+21864 */	"\x94\xAA", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+21868 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+2186C */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+21870 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+21874 */	"\x00\x00", "\x00\x00", "\x00\x00", "\xFC\x63",
  /* U+21878 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x9D\xC6",
  /* U+2187C */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+21880 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x97\xB5",
  /* U+21884 */	"\x92\xB8", "\x91\xEF", "\x00\x00", "\x00\x00",
  /* U+21888 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+2188C */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+21890 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+21894 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+21898 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+2189C */	"\x00\x00", "\x00\x00", "\xFE\xA6", "\x97\x60",
  /* U+218A0 */	"\x93\x58", "\x95\x76", "\x8F\xAC", "\x00\x00",
  /* U+218A4 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+218A8 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+218AC */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+218B0 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+218B4 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+218B8 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+218BC */	"\x00\x00", "\x00\x00", "\x91\xEC", "\x97\xB4",
  /* U+218C0 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+218C4 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+218C8 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+218CC */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+218D0 */	"\x00\x00", "\x91\xF7", "\x00\x00", "\x00\x00",
  /* U+218D4 */	"\x00\x00", "\x00\x00", "\x97\x4A", "\xFB\x49",
  /* U+218D8 */	"\x95\x78", "\x93\xBC", "\x00\x00", "\x00\x00",
  /* U+218DC */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+218E0 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+218E4 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+218E8 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+218EC */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+218F0 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+218F4 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+218F8 */	"\x00\x00", "\x00\x00", "\x91\xD6", "\x00\x00",
  /* U+218FC */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+21900 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x93\x55",
  /* U+21904 */	"\x93\x56", "\x98\x51", "\x00\x00", "\x00\x00",
  /* U+21908 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+2190C */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+21910 */	"\x8F\xF8", "\xFB\xC0", "\x93\xF2", "\x00\x00",
  /* U+21914 */	"\x00\x00", "\x90\xD0", "\x00\x00", "\x00\x00",
  /* U+21918 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+2191C */	"\x9C\x44", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+21920 */	"\x00\x00", "\x00\x00", "\x92\x55", "\x00\x00",
  /* U+21924 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x93\x63",
  /* U+21928 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+2192C */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+21930 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+21934 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+21938 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x91\xA5",
  /* U+2193C */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+21940 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+21944 */	"\xA0\xED", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+21948 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+2194C */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+21950 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+21954 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+21958 */	"\xFD\x6B", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+2195C */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+21960 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+21964 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+21968 */	"\x00\x00", "\x00\x00", "\x9A\xFE", "\x00\x00",
  /* U+2196C */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+21970 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+21974 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+21978 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+2197C */	"\x93\x51", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+21980 */	"\x8C\x57", "\x00\x00", "\x00\x00", "\xFA\x78",
  /* U+21984 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+21988 */	"\xFE\xA8", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+2198C */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+21990 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+21994 */	"\x00\x00", "\x00\x00", "\x93\x50", "\x00\x00",
  /* U+21998 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+2199C */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+219A0 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+219A4 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+219A8 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+219AC */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+219B0 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+219B4 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+219B8 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+219BC */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+219C0 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+219C4 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+219C8 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+219CC */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+219D0 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+219D4 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+219D8 */	"\x00\x00", "\x00\x00", "\x00\x00", "\xFA\x4C",
  /* U+219DC */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+219E0 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+219E4 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+219E8 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+219EC */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+219F0 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x92\xF7",
  /* U+219F4 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+219F8 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+219FC */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+21A00 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+21A04 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+21A08 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+21A0C */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+21A10 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+21A14 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+21A18 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+21A1C */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+21A20 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+21A24 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+21A28 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+21A2C */	"\x00\x00", "\x9B\x40", "\x00\x00", "\x00\x00",
  /* U+21A30 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+21A34 */	"\xFB\xCE", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+21A38 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+21A3C */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+21A40 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+21A44 */	"\x00\x00", "\x9B\x41", "\x00\x00", "\x00\x00",
  /* U+21A48 */	"\x00\x00", "\x00\x00", "\x00\x00", "\xFE\xAD",

  /* U+21B44 */	"\xFB\xD5", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+21B48 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+21B4C */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+21B50 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+21B54 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+21B58 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+21B5C */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+21B60 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+21B64 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+21B68 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+21B6C */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+21B70 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+21B74 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+21B78 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+21B7C */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+21B80 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+21B84 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+21B88 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+21B8C */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+21B90 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+21B94 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+21B98 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+21B9C */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+21BA0 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+21BA4 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+21BA8 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+21BAC */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+21BB0 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+21BB4 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+21BB8 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+21BBC */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+21BC0 */	"\x00\x00", "\x8B\xC2", "\x9A\x7C", "\x00\x00",
  /* U+21BC4 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+21BC8 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+21BCC */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+21BD0 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+21BD4 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+21BD8 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+21BDC */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+21BE0 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+21BE4 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+21BE8 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+21BEC */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+21BF0 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+21BF4 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+21BF8 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+21BFC */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+21C00 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+21C04 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+21C08 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+21C0C */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+21C10 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+21C14 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+21C18 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+21C1C */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+21C20 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+21C24 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+21C28 */	"\x00\x00", "\x00\x00", "\x9B\x42", "\x00\x00",
  /* U+21C2C */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+21C30 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+21C34 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+21C38 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+21C3C */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+21C40 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+21C44 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+21C48 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+21C4C */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+21C50 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+21C54 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+21C58 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+21C5C */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+21C60 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+21C64 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+21C68 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+21C6C */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+21C70 */	"\x9B\x43", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+21C74 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+21C78 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+21C7C */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+21C80 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+21C84 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+21C88 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+21C8C */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+21C90 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+21C94 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+21C98 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+21C9C */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+21CA0 */	"\x00\x00", "\x00\x00", "\x9E\x79", "\x00\x00",
  /* U+21CA4 */	"\x00\x00", "\xFB\xD9", "\x00\x00", "\x00\x00",
  /* U+21CA8 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+21CAC */	"\x9B\x44",

  /* U+21D46 */				"\xA0\xA7", "\x00\x00",
  /* U+21D48 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+21D4C */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+21D50 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+21D54 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+21D58 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+21D5C */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+21D60 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+21D64 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+21D68 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+21D6C */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+21D70 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+21D74 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+21D78 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+21D7C */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+21D80 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+21D84 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+21D88 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+21D8C */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+21D90 */	"\x9B\xF3", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+21D94 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+21D98 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+21D9C */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+21DA0 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+21DA4 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+21DA8 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+21DAC */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+21DB0 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+21DB4 */	"\x00\x00", "\x00\x00", "\x8C\x79", "\x00\x00",
  /* U+21DB8 */	"\x00\x00", "\x00\x00", "\x93\x5E", "\x00\x00",
  /* U+21DBC */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+21DC0 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+21DC4 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+21DC8 */	"\x00\x00", "\x00\x00", "\x89\xCB", "\x00\x00",
  /* U+21DCC */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+21DD0 */	"\x00\x00", "\x9F\x53", "\x00\x00", "\x00\x00",
  /* U+21DD4 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+21DD8 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+21DDC */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+21DE0 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+21DE4 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+21DE8 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x93\xD7",
  /* U+21DEC */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+21DF0 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+21DF4 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+21DF8 */	"\x00\x00", "\xFB\xE1", "\x00\x00", "\x00\x00",
  /* U+21DFC */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+21E00 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+21E04 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+21E08 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+21E0C */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+21E10 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+21E14 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+21E18 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+21E1C */	"\xFE\xD0", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+21E20 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+21E24 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+21E28 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+21E2C */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+21E30 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+21E34 */	"\x00\x00", "\x00\x00", "\x00\x00", "\xFB\xE2",
  /* U+21E38 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+21E3C */	"\x00\x00", "\xFC\xE3", "\x00\x00", "\x00\x00",
  /* U+21E40 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+21E44 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+21E48 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+21E4C */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+21E50 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+21E54 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+21E58 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+21E5C */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+21E60 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+21E64 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+21E68 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+21E6C */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+21E70 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+21E74 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+21E78 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+21E7C */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+21E80 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+21E84 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+21E88 */	"\x00\x00", "\x90\x74", "\x00\x00", "\x00\x00",
  /* U+21E8C */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+21E90 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+21E94 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+21E98 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+21E9C */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+21EA0 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+21EA4 */	"\xFB\xE6", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+21EA8 */	"\x9B\xB7", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+21EAC */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+21EB0 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+21EB4 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+21EB8 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+21EBC */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+21EC0 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+21EC4 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+21EC8 */	"\x9B\x45", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+21ECC */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+21ED0 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+21ED4 */	"\x00\x00", "\x9B\x47", "\x00\x00", "\x00\x00",
  /* U+21ED8 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+21EDC */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+21EE0 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+21EE4 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+21EE8 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+21EEC */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+21EF0 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+21EF4 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+21EF8 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+21EFC */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+21F00 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+21F04 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+21F08 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+21F0C */	"\x00\x00", "\x00\x00", "\x00\x00", "\x9F\x50",
  /* U+21F10 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+21F14 */	"\x00\x00", "\x9B\x48", "\x00\x00", "\x00\x00",
  /* U+21F18 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+21F1C */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+21F20 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+21F24 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+21F28 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+21F2C */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+21F30 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+21F34 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+21F38 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+21F3C */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+21F40 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+21F44 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+21F48 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+21F4C */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+21F50 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+21F54 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+21F58 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+21F5C */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+21F60 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+21F64 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+21F68 */	"\x00\x00", "\x00\x00", "\xFC\x5B", "\x00\x00",
  /* U+21F6C */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+21F70 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+21F74 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+21F78 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+21F7C */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+21F80 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+21F84 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+21F88 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+21F8C */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+21F90 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+21F94 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+21F98 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+21F9C */	"\x00\x00", "\x00\x00", "\x98\xA9", "\x00\x00",
  /* U+21FA0 */	"\x00\x00", "\x9C\xFD", "\x00\x00", "\x00\x00",
  /* U+21FA4 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+21FA8 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+21FAC */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+21FB0 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+21FB4 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+21FB8 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+21FBC */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+21FC0 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+21FC4 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+21FC8 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+21FCC */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+21FD0 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+21FD4 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+21FD8 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+21FDC */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+21FE0 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+21FE4 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+21FE8 */	"\x88\x4C", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+21FEC */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+21FF0 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+21FF4 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+21FF8 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+21FFC */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+22000 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+22004 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+22008 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+2200C */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+22010 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+22014 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+22018 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+2201C */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+22020 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+22024 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+22028 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+2202C */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+22030 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+22034 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+22038 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+2203C */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+22040 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+22044 */	"\x00\x00", "\x9B\x4B", "\x00\x00", "\x00\x00",
  /* U+22048 */	"\x00\x00", "\xFB\xEC", "\x00\x00", "\x00\x00",
  /* U+2204C */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+22050 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+22054 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+22058 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+2205C */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+22060 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+22064 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+22068 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+2206C */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+22070 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+22074 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+22078 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+2207C */	"\x00\x00", "\x00\x00", "\x8C\x69", "\x00\x00",
  /* U+22080 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+22084 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+22088 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+2208C */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+22090 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+22094 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+22098 */	"\x00\x00", "\x00\x00", "\x9B\xA8", "\x00\x00",
  /* U+2209C */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+220A0 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+220A4 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+220A8 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+220AC */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+220B0 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+220B4 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+220B8 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+220BC */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+220C0 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+220C4 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x8A\xD5",
  /* U+220C8 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+220CC */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+220D0 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+220D4 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+220D8 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+220DC */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+220E0 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+220E4 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+220E8 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+220EC */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+220F0 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+220F4 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+220F8 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+220FC */	"\xFA\x73", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+22100 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+22104 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+22108 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+2210C */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+22110 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+22114 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+22118 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+2211C */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+22120 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+22124 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+22128 */	"\x00\x00", "\x00\x00", "\xFD\x59", "\x00\x00",
  /* U+2212C */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+22130 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+22134 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+22138 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+2213C */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+22140 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+22144 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+22148 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+2214C */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+22150 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+22154 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+22158 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x91\xA2",
  /* U+2215C */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+22160 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+22164 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+22168 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+2216C */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+22170 */	"\x00\x00", "\x00\x00", "\x00\x00", "\xFB\xED",
  /* U+22174 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+22178 */	"\x00\x00", "\x00\x00", "\x9C\xA9", "\x00\x00",
  /* U+2217C */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+22180 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+22184 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+22188 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+2218C */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+22190 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+22194 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+22198 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+2219C */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+221A0 */	"\x00\x00", "\x8A\xA8", "\x00\x00", "\x00\x00",
  /* U+221A4 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+221A8 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+221AC */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+221B0 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+221B4 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+221B8 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+221BC */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+221C0 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x9B\xC3",
  /* U+221C4 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+221C8 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+221CC */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+221D0 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+221D4 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+221D8 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+221DC */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+221E0 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+221E4 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+221E8 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+221EC */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+221F0 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+221F4 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+221F8 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+221FC */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+22200 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+22204 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+22208 */	"\x8A\xE1", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+2220C */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+22210 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+22214 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+22218 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+2221C */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+22220 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+22224 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+22228 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+2222C */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+22230 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+22234 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+22238 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+2223C */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+22240 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+22244 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+22248 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+2224C */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+22250 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+22254 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+22258 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+2225C */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+22260 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+22264 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+22268 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+2226C */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+22270 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+22274 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+22278 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+2227C */	"\x9B\x4E",

  /* U+22321 */		    "\x95\xD0", "\x00\x00", "\x00\x00",
  /* U+22324 */	"\x00\x00", "\x90\x5F",

  /* U+223BD */		    "\x97\xEE", "\x00\x00", "\x00\x00",
  /* U+223C0 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+223C4 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+223C8 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+223CC */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+223D0 */	"\xFC\x4E", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+223D4 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x9B\x4F",
  /* U+223D8 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+223DC */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+223E0 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+223E4 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+223E8 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+223EC */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+223F0 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+223F4 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+223F8 */	"\x00\x00", "\x00\x00", "\x9B\x50", "\x00\x00",
  /* U+223FC */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+22400 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+22404 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+22408 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+2240C */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+22410 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+22414 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+22418 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+2241C */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+22420 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+22424 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+22428 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+2242C */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+22430 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+22434 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+22438 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+2243C */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+22440 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+22444 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+22448 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+2244C */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+22450 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+22454 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+22458 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+2245C */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+22460 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+22464 */	"\x00\x00", "\x9E\xC6", "\x00\x00", "\x00\x00",
  /* U+22468 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+2246C */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+22470 */	"\x00\x00", "\xFC\x50", "\x00\x00", "\x00\x00",
  /* U+22474 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+22478 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+2247C */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+22480 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+22484 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+22488 */	"\x00\x00", "\x00\x00", "\x00\x00", "\xFD\x73",
  /* U+2248C */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+22490 */	"\x00\x00", "\xFD\xA7", "\x00\x00", "\x00\x00",
  /* U+22494 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+22498 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+2249C */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+224A0 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+224A4 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+224A8 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+224AC */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+224B0 */	"\x9D\xA2", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+224B4 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+224B8 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+224BC */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+224C0 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+224C4 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+224C8 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+224CC */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+224D0 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+224D4 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+224D8 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+224DC */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+224E0 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+224E4 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+224E8 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+224EC */	"\x00\x00", "\xFA\x58", "\x00\x00", "\x00\x00",
  /* U+224F0 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+224F4 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+224F8 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+224FC */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+22500 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+22504 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+22508 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+2250C */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+22510 */	"\x00\x00", "\x00\x00", "\x00\x00", "\xFA\x5E",
  /* U+22514 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+22518 */	"\x00\x00", "\x00\x00", "\x00\x00", "\xA0\x59",
  /* U+2251C */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+22520 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+22524 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+22528 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+2252C */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+22530 */	"\xFA\x75", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+22534 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+22538 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+2253C */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+22540 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+22544 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+22548 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+2254C */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+22550 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+22554 */	"\xFB\xBE", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+22558 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+2255C */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+22560 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+22564 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+22568 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+2256C */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+22570 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+22574 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+22578 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+2257C */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+22580 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+22584 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+22588 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+2258C */	"\x00\x00", "\x9C\xA2", "\x00\x00", "\x00\x00",
  /* U+22590 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+22594 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+22598 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+2259C */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+225A0 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+225A4 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+225A8 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+225AC */	"\x00\x00", "\x00\x00", "\x00\x00", "\x93\x70",
  /* U+225B0 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+225B4 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+225B8 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+225BC */	"\x00\x00", "\x00\x00", "\x93\x71", "\x00\x00",
  /* U+225C0 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+225C4 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+225C8 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+225CC */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+225D0 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+225D4 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+225D8 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+225DC */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+225E0 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+225E4 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+225E8 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+225EC */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+225F0 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+225F4 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+225F8 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+225FC */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+22600 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+22604 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+22608 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+2260C */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+22610 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+22614 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+22618 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x93\x77",
  /* U+2261C */	"\xFE\xEF", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+22620 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+22624 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+22628 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x93\x6D",
  /* U+2262C */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+22630 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+22634 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+22638 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+2263C */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+22640 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+22644 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+22648 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+2264C */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+22650 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+22654 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+22658 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+2265C */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+22660 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+22664 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+22668 */	"\xFC\x5D", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+2266C */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+22670 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+22674 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+22678 */	"\x00\x00", "\x00\x00", "\x90\xB8", "\x00\x00",
  /* U+2267C */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+22680 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+22684 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+22688 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+2268C */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+22690 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+22694 */	"\x00\x00", "\x00\x00", "\x8A\xFC", "\x00\x00",
  /* U+22698 */	"\xFB\x41", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+2269C */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+226A0 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+226A4 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+226A8 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+226AC */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+226B0 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+226B4 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+226B8 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+226BC */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+226C0 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+226C4 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+226C8 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+226CC */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+226D0 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+226D4 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+226D8 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+226DC */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+226E0 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+226E4 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+226E8 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+226EC */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+226F0 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+226F4 */	"\x9E\x6B", "\x94\xE3", "\x8E\xE2", "\x00\x00",
  /* U+226F8 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+226FC */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+22700 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+22704 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+22708 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+2270C */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+22710 */	"\x00\x00", "\x00\x00", "\x8C\x7D", "\x00\x00",
  /* U+22714 */	"\x8E\xD7", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+22718 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x9C\x4D",
  /* U+2271C */	"\x00\x00", "\x00\x00", "\x00\x00", "\x96\xA3",
  /* U+22720 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+22724 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+22728 */	"\x00\x00", "\x00\x00", "\x9B\x51", "\x00\x00",
  /* U+2272C */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+22730 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+22734 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+22738 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+2273C */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+22740 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+22744 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+22748 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+2274C */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+22750 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+22754 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+22758 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+2275C */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+22760 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+22764 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+22768 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+2276C */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+22770 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+22774 */	"\x00\x00", "\x8A\xC3", "\x00\x00", "\x00\x00",
  /* U+22778 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+2277C */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+22780 */	"\x00\x00", "\x96\xAA", "\x00\x00", "\x00\x00",
  /* U+22784 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+22788 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+2278C */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+22790 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+22794 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+22798 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+2279C */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+227A0 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+227A4 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+227A8 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+227AC */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+227B0 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+227B4 */	"\xFC\x68", "\x8B\x6D", "\x00\x00", "\x00\x00",
  /* U+227B8 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+227BC */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+227C0 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+227C4 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+227C8 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+227CC */	"\x00\x00", "\xFD\x67", "\x00\x00", "\x00\x00",
  /* U+227D0 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+227D4 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+227D8 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+227DC */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+227E0 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+227E4 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+227E8 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+227EC */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+227F0 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+227F4 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+227F8 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+227FC */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+22800 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x8A\xE9",
  /* U+22804 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+22808 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+2280C */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+22810 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+22814 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+22818 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+2281C */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+22820 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+22824 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+22828 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+2282C */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+22830 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+22834 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+22838 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+2283C */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+22840 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+22844 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+22848 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+2284C */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+22850 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+22854 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+22858 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+2285C */	"\x00\x00", "\x00\x00", "\x00\x00", "\xFC\xA1",
  /* U+22860 */	"\x93\x6C", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+22864 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+22868 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+2286C */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+22870 */	"\x00\x00", "\x9B\x52", "\x00\x00", "\x00\x00",
  /* U+22874 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+22878 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+2287C */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+22880 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+22884 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+22888 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+2288C */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+22890 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+22894 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+22898 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+2289C */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+228A0 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+228A4 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+228A8 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+228AC */	"\x00\x00", "\xFE\x70", "\x00\x00", "\x00\x00",
  /* U+228B0 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+228B4 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+228B8 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+228BC */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+228C0 */	"\x00\x00", "\xFC\xA8", "\x00\x00", "\x00\x00",
  /* U+228C4 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+228C8 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+228CC */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+228D0 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+228D4 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+228D8 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+228DC */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+228E0 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+228E4 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+228E8 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+228EC */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+228F0 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+228F4 */	"\x00\x00", "\x00\x00", "\x00\x00", "\xFC\xE9",
  /* U+228F8 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+228FC */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+22900 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+22904 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+22908 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+2290C */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+22910 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+22914 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+22918 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+2291C */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+22920 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+22924 */	"\x00\x00", "\x00\x00", "\x9C\xB4", "\x00\x00",
  /* U+22928 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+2292C */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+22930 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+22934 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+22938 */	"\x00\x00", "\x8A\xEA", "\x00\x00", "\x00\x00",
  /* U+2293C */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+22940 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+22944 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+22948 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+2294C */	"\x00\x00", "\x00\x00", "\x00\x00", "\x9B\x53",
  /* U+22950 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+22954 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+22958 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+2295C */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+22960 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+22964 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x9B\x55",
  /* U+22968 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x96\xAB",
  /* U+2296C */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+22970 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+22974 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+22978 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+2297C */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+22980 */	"\xFC\xA7", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+22984 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+22988 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+2298C */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+22990 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x9B\x56",

  /* U+22A66 */				"\x8A\xBC", "\x00\x00",
  /* U+22A68 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+22A6C */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+22A70 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+22A74 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+22A78 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+22A7C */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+22A80 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+22A84 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+22A88 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+22A8C */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+22A90 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+22A94 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+22A98 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+22A9C */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+22AA0 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+22AA4 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+22AA8 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+22AAC */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+22AB0 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+22AB4 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+22AB8 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+22ABC */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+22AC0 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+22AC4 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+22AC8 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+22ACC */	"\x00\x00", "\x00\x00", "\x00\x00", "\x8A\xCB",
  /* U+22AD0 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+22AD4 */	"\x00\x00", "\x9B\x57", "\x00\x00", "\x00\x00",
  /* U+22AD8 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+22ADC */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+22AE0 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+22AE4 */	"\x00\x00", "\x00\x00", "\x89\xCD", "\x00\x00",
  /* U+22AE8 */	"\x9B\x59", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+22AEC */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+22AF0 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+22AF4 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+22AF8 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+22AFC */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+22B00 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+22B04 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+22B08 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+22B0C */	"\x00\x00", "\x00\x00", "\x9B\x5B", "\x00\x00",
  /* U+22B10 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+22B14 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+22B18 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+22B1C */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+22B20 */	"\x00\x00", "\x00\x00", "\x93\xA5", "\x00\x00",
  /* U+22B24 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+22B28 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+22B2C */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+22B30 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+22B34 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+22B38 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+22B3C */	"\x00\x00", "\x00\x00", "\x00\x00", "\x9B\x5D",
  /* U+22B40 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x9E\x4F",
  /* U+22B44 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+22B48 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+22B4C */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+22B50 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+22B54 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+22B58 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+22B5C */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+22B60 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+22B64 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+22B68 */	"\x00\x00", "\x00\x00", "\x93\xA3", "\x00\x00",
  /* U+22B6C */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+22B70 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+22B74 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+22B78 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+22B7C */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+22B80 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+22B84 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+22B88 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+22B8C */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+22B90 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+22B94 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+22B98 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+22B9C */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+22BA0 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+22BA4 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+22BA8 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+22BAC */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+22BB0 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+22BB4 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+22BB8 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+22BBC */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+22BC0 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+22BC4 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+22BC8 */	"\x00\x00", "\x00\x00", "\x8A\x7B", "\x00\x00",
  /* U+22BCC */	"\x00\x00", "\x00\x00", "\x8B\x42", "\x00\x00",
  /* U+22BD0 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+22BD4 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+22BD8 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+22BDC */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+22BE0 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+22BE4 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+22BE8 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+22BEC */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+22BF0 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+22BF4 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+22BF8 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+22BFC */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+22C00 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+22C04 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+22C08 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+22C0C */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+22C10 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+22C14 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+22C18 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+22C1C */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+22C20 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+22C24 */	"\x00\x00", "\x00\x00", "\x97\x50", "\x8F\xB3",
  /* U+22C28 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+22C2C */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+22C30 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+22C34 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+22C38 */	"\x8A\x50", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+22C3C */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+22C40 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+22C44 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+22C48 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+22C4C */	"\x9B\x60", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+22C50 */	"\x00\x00", "\x8B\x45", "\x00\x00", "\x00\x00",
  /* U+22C54 */	"\x00\x00", "\x8B\x46", "\x00\x00", "\x00\x00",
  /* U+22C58 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+22C5C */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+22C60 */	"\x00\x00", "\x00\x00", "\x9D\xFE", "\x00\x00",
  /* U+22C64 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+22C68 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+22C6C */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+22C70 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+22C74 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+22C78 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+22C7C */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+22C80 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+22C84 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+22C88 */	"\x9B\x62", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+22C8C */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+22C90 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+22C94 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+22C98 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x93\x7B",
  /* U+22C9C */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+22CA0 */	"\x00\x00", "\x93\xB1", "\x00\x00", "\x00\x00",
  /* U+22CA4 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+22CA8 */	"\x00\x00", "\x8A\x60", "\x00\x00", "\x00\x00",
  /* U+22CAC */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+22CB0 */	"\x00\x00", "\x00\x00", "\x8A\xD8", "\x00\x00",
  /* U+22CB4 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x9B\x63",
  /* U+22CB8 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+22CBC */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+22CC0 */	"\x00\x00", "\x00\x00", "\x8A\x69", "\x00\x00",
  /* U+22CC4 */	"\x00\x00", "\x00\x00", "\x8A\x47", "\x00\x00",
  /* U+22CC8 */	"\x00\x00", "\x8A\xCC", "\x00\x00", "\x00\x00",
  /* U+22CCC */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+22CD0 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+22CD4 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+22CD8 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+22CDC */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+22CE0 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+22CE4 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+22CE8 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+22CEC */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+22CF0 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+22CF4 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+22CF8 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+22CFC */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+22D00 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+22D04 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x93\x7C",
  /* U+22D08 */	"\x9B\x65", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+22D0C */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+22D10 */	"\x00\x00", "\x00\x00", "\x9B\x66", "\x00\x00",
  /* U+22D14 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+22D18 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+22D1C */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+22D20 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+22D24 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+22D28 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+22D2C */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+22D30 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+22D34 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+22D38 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+22D3C */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+22D40 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+22D44 */	"\x8A\x72", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+22D48 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+22D4C */	"\x8A\x7A", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+22D50 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+22D54 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+22D58 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+22D5C */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+22D60 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+22D64 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x93\xAF",
  /* U+22D68 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+22D6C */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+22D70 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+22D74 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+22D78 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+22D7C */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+22D80 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+22D84 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+22D88 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+22D8C */	"\x00\x00", "\x8A\xB0", "\x00\x00", "\x00\x00",
  /* U+22D90 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+22D94 */	"\x00\x00", "\x9B\x68", "\x00\x00", "\x00\x00",
  /* U+22D98 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+22D9C */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+22DA0 */	"\x9E\xA3", "\x00\x00", "\x00\x00", "\xFA\xEC",
  /* U+22DA4 */	"\x8B\x77", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+22DA8 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+22DAC */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+22DB0 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+22DB4 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x9B\x67",
  /* U+22DB8 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+22DBC */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+22DC0 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+22DC4 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+22DC8 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+22DCC */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+22DD0 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+22DD4 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+22DD8 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+22DDC */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+22DE0 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+22DE4 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+22DE8 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+22DEC */	"\x00\x00", "\x00\x00", "\x8B\x59", "\x00\x00",
  /* U+22DF0 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+22DF4 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+22DF8 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+22DFC */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+22E00 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+22E04 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+22E08 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+22E0C */	"\x00\x00", "\xFC\xB1", "\x00\x00", "\x00\x00",
  /* U+22E10 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+22E14 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+22E18 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+22E1C */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+22E20 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+22E24 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+22E28 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+22E2C */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+22E30 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+22E34 */	"\x00\x00", "\x00\x00", "\xFC\xBB", "\x00\x00",
  /* U+22E38 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+22E3C */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+22E40 */	"\x00\x00", "\x00\x00", "\x9B\x69", "\x00\x00",
  /* U+22E44 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+22E48 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+22E4C */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+22E50 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+22E54 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+22E58 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+22E5C */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+22E60 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+22E64 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+22E68 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+22E6C */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+22E70 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+22E74 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+22E78 */	"\x93\xA8", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+22E7C */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+22E80 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+22E84 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+22E88 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x8A\xE0",
  /* U+22E8C */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+22E90 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+22E94 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+22E98 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+22E9C */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+22EA0 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+22EA4 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+22EA8 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+22EAC */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+22EB0 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x9E\x51",
  /* U+22EB4 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+22EB8 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+22EBC */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+22EC0 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+22EC4 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+22EC8 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+22ECC */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+22ED0 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+22ED4 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+22ED8 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+22EDC */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+22EE0 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+22EE4 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+22EE8 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+22EEC */	"\x00\x00", "\x00\x00", "\x00\x00", "\x8F\x5F",

  /* U+22F74 */	"\x9B\x6A", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+22F78 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+22F7C */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+22F80 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+22F84 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+22F88 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+22F8C */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+22F90 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+22F94 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+22F98 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+22F9C */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+22FA0 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+22FA4 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+22FA8 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+22FAC */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+22FB0 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+22FB4 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+22FB8 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+22FBC */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+22FC0 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+22FC4 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+22FC8 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+22FCC */	"\x9B\x6B", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+22FD0 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+22FD4 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+22FD8 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+22FDC */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+22FE0 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x97\xEC",
  /* U+22FE4 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+22FE8 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+22FEC */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+22FF0 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+22FF4 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+22FF8 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+22FFC */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+23000 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+23004 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+23008 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+2300C */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+23010 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+23014 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+23018 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+2301C */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+23020 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+23024 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+23028 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+2302C */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+23030 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x9B\x6C",
  /* U+23034 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+23038 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+2303C */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+23040 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+23044 */	"\xFE\x4E", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+23048 */	"\x00\x00", "\x00\x00", "\x00\x00", "\xFD\xC2",
  /* U+2304C */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+23050 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+23054 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+23058 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+2305C */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+23060 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+23064 */	"\x00\x00", "\x00\x00", "\x9B\x6D", "\x00\x00",
  /* U+23068 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+2306C */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+23070 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+23074 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+23078 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+2307C */	"\x00\x00", "\x91\x67", "\xFC\xCC", "\x00\x00",
  /* U+23080 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+23084 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+23088 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+2308C */	"\x00\x00", "\x00\x00", "\x93\xB6", "\x00\x00",
  /* U+23090 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+23094 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+23098 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+2309C */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+230A0 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+230A4 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+230A8 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+230AC */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+230B0 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+230B4 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x90\xE4",
  /* U+230B8 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+230BC */	"\x90\xE5", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+230C0 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+230C4 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+230C8 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+230CC */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+230D0 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+230D4 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+230D8 */	"\x00\x00", "\x00\x00", "\x9E\xF2", "\x00\x00",
  /* U+230DC */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+230E0 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+230E4 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+230E8 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+230EC */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+230F0 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+230F4 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+230F8 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+230FC */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+23100 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x93\xCA",
  /* U+23104 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+23108 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+2310C */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+23110 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+23114 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+23118 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+2311C */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+23120 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+23124 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+23128 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+2312C */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+23130 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+23134 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+23138 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+2313C */	"\x00\x00", "\x8B\xBC", "\x00\x00", "\x00\x00",
  /* U+23140 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+23144 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+23148 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+2314C */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+23150 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+23154 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+23158 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+2315C */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+23160 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+23164 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+23168 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+2316C */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+23170 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+23174 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+23178 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+2317C */	"\x00\x00", "\x8F\x46", "\x00\x00", "\x00\x00",
  /* U+23180 */	"\x00\x00", "\x00\x00", "\x93\xCF", "\x00\x00",
  /* U+23184 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+23188 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+2318C */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+23190 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+23194 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+23198 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+2319C */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+231A0 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+231A4 */	"\xFC\xDB", "\xFC\xDC", "\x00\x00", "\x00\x00",
  /* U+231A8 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+231AC */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+231B0 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x93\xC0",
  /* U+231B4 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+231B8 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+231BC */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+231C0 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+231C4 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+231C8 */	"\xFC\xE6", "\x96\xE7", "\x00\x00", "\x00\x00",
  /* U+231CC */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+231D0 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+231D4 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+231D8 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+231DC */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+231E0 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+231E4 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+231E8 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+231EC */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+231F0 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+231F4 */	"\x00\x00", "\x00\x00", "\x00\x00", "\xFC\xD8",
  /* U+231F8 */	"\xFC\xD9", "\xFD\xA6", "\x00\x00", "\x00\x00",
  /* U+231FC */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+23200 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+23204 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+23208 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+2320C */	"\x00\x00", "\x00\x00", "\x00\x00", "\x93\xCE",
  /* U+23210 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+23214 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+23218 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+2321C */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+23220 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+23224 */	"\x00\x00", "\x95\xF1", "\x00\x00", "\x00\x00",
  /* U+23228 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+2322C */	"\x00\x00", "\x00\x00", "\x00\x00", "\x9C\xE9",
  /* U+23230 */	"\x00\x00", "\xFC\xE4", "\x94\xAF", "\xFA\x77",
  /* U+23234 */	"\x93\xCC", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+23238 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+2323C */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+23240 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+23244 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+23248 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+2324C */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+23250 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+23254 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+23258 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+2325C */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+23260 */	"\x00\x00", "\x00\x00", "\x90\x5A", "\x00\x00",
  /* U+23264 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+23268 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+2326C */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+23270 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+23274 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+23278 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+2327C */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+23280 */	"\x00\x00", "\x8C\x54", "\x00\x00", "\x00\x00",
  /* U+23284 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+23288 */	"\x00\x00", "\x93\xBF", "\xFB\x51", "\x00\x00",
  /* U+2328C */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+23290 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+23294 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+23298 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+2329C */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+232A0 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+232A4 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+232A8 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x93\xB9",
  /* U+232AC */	"\xFE\xD7", "\x93\xB7", "\x00\x00", "\x00\x00",
  /* U+232B0 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+232B4 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+232B8 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+232BC */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+232C0 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+232C4 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+232C8 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+232CC */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+232D0 */	"\x00\x00", "\x00\x00", "\x93\xD9", "\x00\x00",
  /* U+232D4 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+232D8 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+232DC */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+232E0 */	"\x93\xBB", "\x93\xDA", "\x00\x00", "\x00\x00",
  /* U+232E4 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+232E8 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+232EC */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+232F0 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+232F4 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+232F8 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+232FC */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+23300 */	"\x98\xA3", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+23304 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+23308 */	"\x00\x00", "\x00\x00", "\x90\xD1", "\x00\x00",
  /* U+2330C */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+23310 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+23314 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+23318 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+2331C */	"\x00\x00", "\x00\x00", "\x00\x00", "\x9B\x6E",

  /* U+233B4 */	"\xFA\x70", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+233B8 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+233BC */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+233C0 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+233C4 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+233C8 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+233CC */	"\x9B\xEB", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+233D0 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+233D4 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+233D8 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+233DC */	"\x00\x00", "\x00\x00", "\x9B\x6F", "\x00\x00",
  /* U+233E0 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+233E4 */	"\x00\x00", "\x00\x00", "\xFC\xFC", "\x00\x00",
  /* U+233E8 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+233EC */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+233F0 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+233F4 */	"\x8B\x40", "\xA0\x7B", "\x00\x00", "\x00\x00",
  /* U+233F8 */	"\x00\x00", "\x8C\xA1", "\x97\xF7", "\x00\x00",
  /* U+233FC */	"\x00\x00", "\x00\x00", "\x93\xE2", "\x00\x00",
  /* U+23400 */	"\xFC\xD6", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+23404 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+23408 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+2340C */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+23410 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+23414 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+23418 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+2341C */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+23420 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+23424 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+23428 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+2342C */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+23430 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+23434 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+23438 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+2343C */	"\x00\x00", "\x00\x00", "\x00\x00", "\x95\x59",
  /* U+23440 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+23444 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+23448 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+2344C */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+23450 */	"\x93\xA6", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+23454 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+23458 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+2345C */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+23460 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+23464 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+23468 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+2346C */	"\x00\x00", "\x00\x00", "\x00\x00", "\xFD\x40",
  /* U+23470 */	"\x00\x00", "\x00\x00", "\x93\x5F", "\x00\x00",
  /* U+23474 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+23478 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+2347C */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+23480 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+23484 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+23488 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+2348C */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+23490 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+23494 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+23498 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+2349C */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+234A0 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+234A4 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+234A8 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+234AC */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+234B0 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+234B4 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+234B8 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+234BC */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+234C0 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+234C4 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+234C8 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+234CC */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+234D0 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+234D4 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+234D8 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+234DC */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+234E0 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+234E4 */	"\x00\x00", "\x97\xF2", "\x00\x00", "\x00\x00",
  /* U+234E8 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+234EC */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+234F0 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+234F4 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+234F8 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+234FC */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+23500 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+23504 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+23508 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+2350C */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+23510 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+23514 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+23518 */	"\x00\x00", "\x9C\x76", "\x00\x00", "\x00\x00",
  /* U+2351C */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+23520 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+23524 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+23528 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+2352C */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+23530 */	"\x8E\xF8", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+23534 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+23538 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+2353C */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+23540 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+23544 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+23548 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+2354C */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+23550 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+23554 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+23558 */	"\x00\x00", "\x00\x00", "\x8F\x47", "\x00\x00",
  /* U+2355C */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+23560 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+23564 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x9B\x74",
  /* U+23568 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+2356C */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+23570 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+23574 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+23578 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+2357C */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+23580 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+23584 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+23588 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+2358C */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+23590 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+23594 */	"\x00\x00", "\x92\xB4", "\x00\x00", "\x00\x00",
  /* U+23598 */	"\x00\x00", "\x91\xED", "\x00\x00", "\x00\x00",
  /* U+2359C */	"\x96\xD2", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+235A0 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+235A4 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+235A8 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+235AC */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+235B0 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+235B4 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+235B8 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+235BC */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+235C0 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+235C4 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+235C8 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+235CC */	"\x00\x00", "\xFD\x46", "\x8F\x4F", "\x95\x49",
  /* U+235D0 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+235D4 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+235D8 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+235DC */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+235E0 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+235E4 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+235E8 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+235EC */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+235F0 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x9B\x75",
  /* U+235F4 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+235F8 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+235FC */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+23600 */	"\xFA\x5C", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+23604 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+23608 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+2360C */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+23610 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+23614 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+23618 */	"\x00\x00", "\x00\x00", "\x9B\x79", "\x00\x00",
  /* U+2361C */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+23620 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+23624 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+23628 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+2362C */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+23630 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+23634 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+23638 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+2363C */	"\xFD\x4B", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+23640 */	"\x96\xD3", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+23644 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+23648 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+2364C */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+23650 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+23654 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+23658 */	"\x00\x00", "\xFD\x58", "\x00\x00", "\x00\x00",
  /* U+2365C */	"\x00\x00", "\x00\x00", "\x00\x00", "\x94\x5F",
  /* U+23660 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+23664 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+23668 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+2366C */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+23670 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+23674 */	"\x00\x00", "\x00\x00", "\x00\x00", "\xA0\xF5",
  /* U+23678 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+2367C */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+23680 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+23684 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+23688 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+2368C */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+23690 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+23694 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+23698 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+2369C */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+236A0 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+236A4 */	"\x00\x00", "\x00\x00", "\x92\x43", "\x00\x00",
  /* U+236A8 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+236AC */	"\x00\x00", "\x97\xFA", "\x00\x00", "\x00\x00",
  /* U+236B0 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+236B4 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+236B8 */	"\x00\x00", "\x00\x00", "\x9D\xD9", "\x00\x00",
  /* U+236BC */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+236C0 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+236C4 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+236C8 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+236CC */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+236D0 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+236D4 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+236D8 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+236DC */	"\x00\x00", "\x00\x00", "\x00\x00", "\x97\xF4",
  /* U+236E0 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+236E4 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+236E8 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+236EC */	"\x00\x00", "\x00\x00", "\x92\x4D", "\x00\x00",
  /* U+236F0 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+236F4 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+236F8 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+236FC */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+23700 */	"\x00\x00", "\x00\x00", "\x00\x00", "\xFD\x5B",
  /* U+23704 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+23708 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+2370C */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+23710 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+23714 */	"\x00\x00", "\x00\x00", "\x9B\x7A", "\x00\x00",
  /* U+23718 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+2371C */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+23720 */	"\x9E\xD5", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+23724 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+23728 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+2372C */	"\x00\x00", "\xFA\xAE", "\x00\x00", "\x9C\xC9",
  /* U+23730 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+23734 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+23738 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+2373C */	"\x00\x00", "\x00\x00", "\x00\x00", "\x92\x58",
  /* U+23740 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+23744 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+23748 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+2374C */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+23750 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+23754 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+23758 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+2375C */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+23760 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+23764 */	"\x00\x00", "\x00\x00", "\x8E\xC8", "\x00\x00",
  /* U+23768 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+2376C */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+23770 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+23774 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+23778 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+2377C */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+23780 */	"\x00\x00", "\x94\xB4", "\x00\x00", "\x00\x00",
  /* U+23784 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+23788 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+2378C */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+23790 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+23794 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+23798 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+2379C */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+237A0 */	"\x00\x00", "\x00\x00", "\x93\xE1", "\x00\x00",
  /* U+237A4 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+237A8 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+237AC */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+237B0 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+237B4 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+237B8 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+237BC */	"\x93\xDF", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+237C0 */	"\x00\x00", "\x00\x00", "\xFC\xF0", "\x00\x00",
  /* U+237C4 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+237C8 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+237CC */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+237D0 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+237D4 */	"\x00\x00", "\x93\xEC", "\x97\xF6", "\x96\xCF",
  /* U+237D8 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+237DC */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+237E0 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+237E4 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+237E8 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+237EC */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+237F0 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+237F4 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+237F8 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+237FC */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+23800 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+23804 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+23808 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+2380C */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+23810 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+23814 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+23818 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+2381C */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+23820 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+23824 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+23828 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+2382C */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+23830 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+23834 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+23838 */	"\x00\x00", "\x00\x00", "\x93\xDE",

  /* U+239C2 */				"\x8A\xCF",

  /* U+23AA7 */					    "\x9B\xA2",
  /* U+23AA8 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+23AAC */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+23AB0 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+23AB4 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+23AB8 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+23ABC */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+23AC0 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+23AC4 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+23AC8 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+23ACC */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+23AD0 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+23AD4 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+23AD8 */	"\x00\x00", "\x00\x00", "\x00\x00", "\xFD\x69",
  /* U+23ADC */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+23AE0 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+23AE4 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+23AE8 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+23AEC */	"\x00\x00", "\x00\x00", "\x93\x52", "\x00\x00",
  /* U+23AF0 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+23AF4 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+23AF8 */	"\x00\x00", "\x00\x00", "\x98\xA2", "\x00\x00",
  /* U+23AFC */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+23B00 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+23B04 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+23B08 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+23B0C */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+23B10 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+23B14 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+23B18 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+23B1C */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+23B20 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+23B24 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+23B28 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+23B2C */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+23B30 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+23B34 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+23B38 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+23B3C */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+23B40 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+23B44 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+23B48 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+23B4C */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+23B50 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+23B54 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+23B58 */	"\x00\x00", "\x00\x00", "\xFD\x6E",

  /* U+23C63 */					    "\x8C\xA4",
  /* U+23C64 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+23C68 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+23C6C */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+23C70 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+23C74 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+23C78 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+23C7C */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+23C80 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+23C84 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+23C88 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+23C8C */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+23C90 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+23C94 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+23C98 */	"\x00\x00", "\xFA\x7C", "\x93\xFA", "\x90\x7C",
  /* U+23C9C */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+23CA0 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+23CA4 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+23CA8 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+23CAC */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+23CB0 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+23CB4 */	"\x00\x00", "\x8F\x67", "\x00\x00", "\x9D\xB7",
  /* U+23CB8 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+23CBC */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+23CC0 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+23CC4 */	"\x00\x00", "\x00\x00", "\x00\x00", "\xA0\xE9",
  /* U+23CC8 */	"\xFA\x4E", "\xFD\xA1", "\x00\x00", "\x00\x00",
  /* U+23CCC */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+23CD0 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+23CD4 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+23CD8 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+23CDC */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+23CE0 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+23CE4 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+23CE8 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+23CEC */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+23CF0 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+23CF4 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+23CF8 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+23CFC */	"\x9E\x74", "\x9F\xBF", "\x9E\xCB", "\x9B\xB9",
  /* U+23D00 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+23D04 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+23D08 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+23D0C */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+23D10 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+23D14 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+23D18 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+23D1C */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+23D20 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+23D24 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+23D28 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+23D2C */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+23D30 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+23D34 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+23D38 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+23D3C */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+23D40 */	"\x9D\xD4", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+23D44 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+23D48 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+23D4C */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+23D50 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+23D54 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+23D58 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x97\xB9",
  /* U+23D5C */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+23D60 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+23D64 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+23D68 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+23D6C */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+23D70 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+23D74 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+23D78 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+23D7C */	"\x00\x00", "\x00\x00", "\x8E\xF1", "\x00\x00",
  /* U+23D80 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+23D84 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+23D88 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+23D8C */	"\x00\x00", "\x00\x00", "\x00\x00", "\x95\x7B",
  /* U+23D90 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+23D94 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+23D98 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+23D9C */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+23DA0 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+23DA4 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+23DA8 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+23DAC */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+23DB0 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+23DB4 */	"\x00\x00", "\x00\x00", "\x9E\xD2", "\x97\x53",
  /* U+23DB8 */	"\x96\xA4", "\x8F\xBE", "\x94\xD9", "\x90\x58",
  /* U+23DBC */	"\xFD\x79", "\xFD\x7B", "\x00\x00", "\x00\x00",
  /* U+23DC0 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+23DC4 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+23DC8 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+23DCC */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+23DD0 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+23DD4 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+23DD8 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+23DDC */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+23DE0 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x8E\xDA",
  /* U+23DE4 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+23DE8 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+23DEC */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+23DF0 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+23DF4 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+23DF8 */	"\x8E\xFA", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+23DFC */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+23E00 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+23E04 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+23E08 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+23E0C */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+23E10 */	"\x00\x00", "\x9B\xA5", "\x00\x00", "\x00\x00",
  /* U+23E14 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+23E18 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+23E1C */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+23E20 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+23E24 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+23E28 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+23E2C */	"\x9E\xD9", "\x97\xD4", "\x90\xBB", "\xFD\xBC",
  /* U+23E30 */	"\xFD\xC6", "\x92\x48", "\x00\x00", "\x00\x00",
  /* U+23E34 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+23E38 */	"\x00\x00", "\x92\xB5", "\x00\x00", "\x00\x00",
  /* U+23E3C */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+23E40 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+23E44 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+23E48 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+23E4C */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+23E50 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+23E54 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+23E58 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+23E5C */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+23E60 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+23E64 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+23E68 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+23E6C */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+23E70 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+23E74 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+23E78 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+23E7C */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+23E80 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+23E84 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+23E88 */	"\x9D\xC1", "\x92\xB9", "\x92\xA6", "\x8F\x4B",
  /* U+23E8C */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+23E90 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+23E94 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+23E98 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+23E9C */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+23EA0 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+23EA4 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+23EA8 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+23EAC */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+23EB0 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+23EB4 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+23EB8 */	"\x00\x00", "\x9B\xA6", "\x00\x00", "\x00\x00",
  /* U+23EBC */	"\x00\x00", "\x00\x00", "\x00\x00", "\x92\xB6",
  /* U+23EC0 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+23EC4 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+23EC8 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+23ECC */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+23ED0 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+23ED4 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x8E\x40",
  /* U+23ED8 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+23EDC */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+23EE0 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+23EE4 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+23EE8 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+23EEC */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+23EF0 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+23EF4 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x9E\xD8",
  /* U+23EF8 */	"\x94\x5E", "\x98\x5F", "\x94\xCE", "\x92\x4A",
  /* U+23EFC */	"\xFD\x70", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+23F00 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+23F04 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+23F08 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+23F0C */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+23F10 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+23F14 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+23F18 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+23F1C */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+23F20 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+23F24 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+23F28 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+23F2C */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+23F30 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+23F34 */	"\x00\x00", "\x94\x67", "\x00\x00", "\x00\x00",
  /* U+23F38 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+23F3C */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+23F40 */	"\x00\x00", "\x8D\xEC", "\x00\x00", "\x00\x00",
  /* U+23F44 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+23F48 */	"\x00\x00", "\x00\x00", "\x9B\xD8", "\x00\x00",
  /* U+23F4C */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+23F50 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+23F54 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+23F58 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+23F5C */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+23F60 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+23F64 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+23F68 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+23F6C */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+23F70 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+23F74 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+23F78 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+23F7C */	"\x00\x00", "\x00\x00", "\x00\x00", "\x94\x48",
  /* U+23F80 */	"\xFA\xC1", "\x9C\xF7", "\xFD\xBE", "\x00\x00",
  /* U+23F84 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+23F88 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+23F8C */	"\x00\x00", "\x00\x00", "\x00\x00", "\x8F\xDA",
  /* U+23F90 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+23F94 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+23F98 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+23F9C */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+23FA0 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+23FA4 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+23FA8 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+23FAC */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+23FB0 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+23FB4 */	"\xFD\xD9", "\x00\x00", "\x00\x00", "\xFC\x7E",
  /* U+23FB8 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+23FBC */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+23FC0 */	"\x93\xF9", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+23FC4 */	"\x00\x00", "\xFA\x43", "\x00\x00", "\x00\x00",
  /* U+23FC8 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+23FCC */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+23FD0 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+23FD4 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+23FD8 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+23FDC */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+23FE0 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+23FE4 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+23FE8 */	"\x00\x00", "\x00\x00", "\x00\x00", "\xFA\xEB",
  /* U+23FEC */	"\xFA\xC3", "\x97\xD3", "\x95\xF9", "\x9C\x48",
  /* U+23FF0 */	"\xFD\xD8", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+23FF4 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+23FF8 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+23FFC */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+24000 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+24004 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+24008 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+2400C */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+24010 */	"\x00\x00", "\xA0\xD8", "\x00\x00", "\x00\x00",
  /* U+24014 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+24018 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+2401C */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+24020 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+24024 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+24028 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+2402C */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+24030 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+24034 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+24038 */	"\x00\x00", "\xFD\xD7", "\xFB\x4A", "\x9B\xAF",
  /* U+2403C */	"\x94\x4B", "\xFD\xC9", "\x00\x00", "\x00\x00",
  /* U+24040 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+24044 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+24048 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+2404C */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+24050 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+24054 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x8E\xAC",
  /* U+24058 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+2405C */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+24060 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+24064 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+24068 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+2406C */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+24070 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+24074 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+24078 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+2407C */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+24080 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+24084 */	"\x00\x00", "\xFD\xB2", "\x00\x00", "\x00\x00",
  /* U+24088 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x92\x5A",
  /* U+2408C */	"\xFC\xBD", "\x92\xD9", "\x00\x00", "\x00\x00",
  /* U+24090 */	"\x00\x00", "\xFD\xD5", "\x00\x00", "\x00\x00",
  /* U+24094 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+24098 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+2409C */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+240A0 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+240A4 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+240A8 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+240AC */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+240B0 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+240B4 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+240B8 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+240BC */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+240C0 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+240C4 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+240C8 */	"\x00\x00", "\x92\xDD", "\x00\x00", "\x00\x00",
  /* U+240CC */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+240D0 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+240D4 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+240D8 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+240DC */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+240E0 */	"\x00\x00", "\x92\x59", "\x00\x00", "\x00\x00",
  /* U+240E4 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+240E8 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+240EC */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+240F0 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+240F4 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+240F8 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+240FC */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+24100 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+24104 */	"\x96\xBA", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+24108 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+2410C */	"\x00\x00", "\x00\x00", "\x00\x00", "\x92\x5B",
  /* U+24110 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+24114 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+24118 */	"\x00\x00", "\x9B\xAB", "\x00\x00", "\x00\x00",
  /* U+2411C */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+24120 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+24124 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+24128 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+2412C */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+24130 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+24134 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+24138 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+2413C */	"\x00\x00", "\x00\x00", "\x00\x00", "\xFD\xDA",
  /* U+24140 */	"\xFD\xDE", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+24144 */	"\xFD\xD3", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+24148 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+2414C */	"\x00\x00", "\x00\x00", "\x8C\x46", "\x00\x00",
  /* U+24150 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+24154 */	"\x00\x00", "\xFD\xD6", "\xFD\xDC", "\xFD\xDD",
  /* U+24158 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+2415C */	"\x90\xFE", "\x00\x00", "\x00\x00", "\xFE\xA1",
  /* U+24160 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+24164 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+24168 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+2416C */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+24170 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+24174 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x8B\xAD",
  /* U+24178 */	"\x00\x00", "\x00\x00", "\x9C\xD8", "\x00\x00",
  /* U+2417C */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+24180 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+24184 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+24188 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+2418C */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+24190 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+24194 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+24198 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+2419C */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+241A0 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x9E\x6D",
  /* U+241A4 */	"\xFD\x7C", "\xFB\x61", "\x00\x00", "\x00\x00",
  /* U+241A8 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+241AC */	"\x96\xF8", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+241B0 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+241B4 */	"\x00\x00", "\x96\xF0", "\x00\x00", "\x00\x00",
  /* U+241B8 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+241BC */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+241C0 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+241C4 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+241C8 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+241CC */	"\x00\x00", "\xFC\xF4", "\x00\x00", "\x00\x00",
  /* U+241D0 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+241D4 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+241D8 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+241DC */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+241E0 */	"\x00\x00", "\x00\x00", "\xFE\x60", "\x00\x00",
  /* U+241E4 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+241E8 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+241EC */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+241F0 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+241F4 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+241F8 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+241FC */	"\x98\x52", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+24200 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+24204 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+24208 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+2420C */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+24210 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+24214 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+24218 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x96\x4F",
  /* U+2421C */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+24220 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+24224 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+24228 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+2422C */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+24230 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+24234 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+24238 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+2423C */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+24240 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+24244 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+24248 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x91\x6E",
  /* U+2424C */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+24250 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+24254 */	"\x00\x00", "\x00\x00", "\x98\x6D", "\x00\x00",
  /* U+24258 */	"\x00\x00", "\x98\x64", "\x00\x00", "\x00\x00",
  /* U+2425C */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+24260 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+24264 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+24268 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+2426C */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+24270 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+24274 */	"\x00\x00", "\x00\x00", "\x94\x53", "\xFD\xEC",
  /* U+24278 */	"\xFB\x78", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+2427C */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+24280 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+24284 */	"\x95\xBA", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+24288 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+2428C */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+24290 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x98\x5D",
  /* U+24294 */	"\x00\x00", "\x92\xF9", "\x00\x00", "\x00\x00",
  /* U+24298 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+2429C */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+242A0 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+242A4 */	"\x00\x00", "\x98\x5A", "\x00\x00", "\x00\x00",
  /* U+242A8 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+242AC */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+242B0 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+242B4 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+242B8 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+242BC */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+242C0 */	"\x00\x00", "\xFD\xF6", "\x00\x00", "\x00\x00",
  /* U+242C4 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+242C8 */	"\x00\x00", "\x93\xD0", "\x98\x62", "\x00\x00",
  /* U+242CC */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+242D0 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+242D4 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+242D8 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+242DC */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+242E0 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+242E4 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+242E8 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+242EC */	"\x00\x00", "\x00\x00", "\x9B\xAD", "\x00\x00",
  /* U+242F0 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+242F4 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+242F8 */	"\x00\x00", "\x00\x00", "\x97\x4F", "\x00\x00",
  /* U+242FC */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+24300 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+24304 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+24308 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+2430C */	"\x00\x00", "\x9B\xAE", "\x00\x00", "\x00\x00",
  /* U+24310 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+24314 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+24318 */	"\x00\x00", "\x00\x00", "\x94\x52", "\x00\x00",
  /* U+2431C */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+24320 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+24324 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+24328 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+2432C */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+24330 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+24334 */	"\x9B\xB0", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+24338 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+2433C */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+24340 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+24344 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+24348 */	"\x91\xD2", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+2434C */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+24350 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+24354 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+24358 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+2435C */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+24360 */	"\x00\x00", "\x00\x00", "\x97\xEA", "\xFB\x6B",
  /* U+24364 */	"\x91\xB1", "\xFD\xF3", "\x00\x00", "\x00\x00",
  /* U+24368 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+2436C */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+24370 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+24374 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+24378 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+2437C */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+24380 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+24384 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+24388 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+2438C */	"\x92\xCB", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+24390 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+24394 */	"\x00\x00", "\x00\x00", "\x9B\xB1", "\x00\x00",
  /* U+24398 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+2439C */	"\xFC\xEC", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+243A0 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+243A4 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+243A8 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+243AC */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+243B0 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+243B4 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+243B8 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+243BC */	"\x00\x00", "\x98\x6B", "\x00\x00", "\x00\x00",
  /* U+243C0 */	"\x00\x00", "\x97\x51", "\x00\x00", "\x00\x00",
  /* U+243C4 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+243C8 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+243CC */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+243D0 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+243D4 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+243D8 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+243DC */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+243E0 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+243E4 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+243E8 */	"\x00\x00", "\x98\x71", "\x95\xEF", "\x00\x00",
  /* U+243EC */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+243F0 */	"\x00\x00", "\x00\x00", "\x9E\xF3", "\x00\x00",
  /* U+243F4 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+243F8 */	"\x91\xE8", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+243FC */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+24400 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+24404 */	"\x9B\xBA", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+24408 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+2440C */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+24410 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+24414 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+24418 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+2441C */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+24420 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+24424 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+24428 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+2442C */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+24430 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+24434 */	"\x00\x00", "\xFB\x4C", "\x92\x6A", "\x00\x00",
  /* U+24438 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+2443C */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+24440 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+24444 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+24448 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+2444C */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+24450 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+24454 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+24458 */	"\x00\x00", "\x00\x00", "\xFD\xF8", "\x98\x61",
  /* U+2445C */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+24460 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+24464 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+24468 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+2446C */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+24470 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x91\xE7",
  /* U+24474 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+24478 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+2447C */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+24480 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+24484 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x93\xED",
  /* U+24488 */	"\x97\x44", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+2448C */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+24490 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+24494 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+24498 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+2449C */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+244A0 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+244A4 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+244A8 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+244AC */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+244B0 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+244B4 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+244B8 */	"\x00\x00", "\x91\xE1", "\x00\x00", "\x00\x00",
  /* U+244BC */	"\xFB\xF5", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+244C0 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+244C4 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+244C8 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+244CC */	"\x00\x00", "\x00\x00", "\x98\x69", "\x00\x00",
  /* U+244D0 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x8A\x62",
  /* U+244D4 */	"\x00\x00", "\x00\x00", "\x9B\xBB", "\x00\x00",
  /* U+244D8 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+244DC */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+244E0 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+244E4 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+244E8 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+244EC */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+244F0 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+244F4 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+244F8 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+244FC */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+24500 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+24504 */	"\x00\x00", "\x8C\xA8", "\x00\x00", "\x00\x00",
  /* U+24508 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+2450C */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+24510 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+24514 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+24518 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+2451C */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+24520 */	"\x00\x00", "\x9C\x55", "\x00\x00", "\x00\x00",
  /* U+24524 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+24528 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+2452C */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+24530 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+24534 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+24538 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+2453C */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+24540 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+24544 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+24548 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+2454C */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+24550 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+24554 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+24558 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+2455C */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+24560 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+24564 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+24568 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+2456C */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+24570 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+24574 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+24578 */	"\x8E\x77", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+2457C */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+24580 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+24584 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+24588 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+2458C */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+24590 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+24594 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+24598 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+2459C */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+245A0 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+245A4 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+245A8 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+245AC */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+245B0 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+245B4 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+245B8 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+245BC */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+245C0 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+245C4 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+245C8 */	"\x8A\xB2", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+245CC */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+245D0 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+245D4 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+245D8 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+245DC */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+245E0 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+245E4 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+245E8 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+245EC */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+245F0 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+245F4 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+245F8 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+245FC */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+24600 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+24604 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+24608 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+2460C */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+24610 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+24614 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+24618 */	"\x9E\xBC", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+2461C */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+24620 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+24624 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+24628 */	"\x00\x00", "\x00\x00", "\x93\xE6", "\x00\x00",
  /* U+2462C */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+24630 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+24634 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+24638 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+2463C */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+24640 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+24644 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+24648 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+2464C */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+24650 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+24654 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+24658 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+2465C */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+24660 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+24664 */	"\x00\x00", "\x93\xA2", "\x00\x00", "\x00\x00",
  /* U+24668 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+2466C */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+24670 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+24674 */	"\x9B\xBD", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+24678 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+2467C */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+24680 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+24684 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+24688 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+2468C */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+24690 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+24694 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x94\xB3",
  /* U+24698 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+2469C */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+246A0 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+246A4 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+246A8 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+246AC */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+246B0 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+246B4 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+246B8 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+246BC */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+246C0 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+246C4 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+246C8 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+246CC */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+246D0 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+246D4 */	"\x93\x7D", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+246D8 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+246DC */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+246E0 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+246E4 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+246E8 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+246EC */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+246F0 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+246F4 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+246F8 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+246FC */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+24700 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+24704 */	"\x00\x00", "\x00\x00", "\x9E\x66", "\x00\x00",
  /* U+24708 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+2470C */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+24710 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+24714 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+24718 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+2471C */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+24720 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+24724 */	"\x00\x00", "\x94\x59", "\x00\x00", "\x00\x00",
  /* U+24728 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+2472C */	"\x00\x00", "\x00\x00", "\x00\x00", "\x9B\xBF",
  /* U+24730 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+24734 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+24738 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+2473C */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+24740 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+24744 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+24748 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+2474C */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+24750 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+24754 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+24758 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+2475C */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+24760 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+24764 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+24768 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+2476C */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+24770 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+24774 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+24778 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+2477C */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+24780 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+24784 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+24788 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+2478C */	"\x00\x00", "\x00\x00", "\x00\x00", "\x94\x58",
  /* U+24790 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+24794 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+24798 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+2479C */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+247A0 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+247A4 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+247A8 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+247AC */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+247B0 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+247B4 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+247B8 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+247BC */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+247C0 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+247C4 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+247C8 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+247CC */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+247D0 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+247D4 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+247D8 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+247DC */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+247E0 */	"\x9E\xA5", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+247E4 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+247E8 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+247EC */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+247F0 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+247F4 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+247F8 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+247FC */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+24800 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+24804 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+24808 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+2480C */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+24810 */	"\x00\x00", "\x00\x00", "\x9B\xC7", "\x00\x00",
  /* U+24814 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+24818 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+2481C */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+24820 */	"\x00\x00", "\x00\x00", "\x00\x00", "\xFE\x54",
  /* U+24824 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+24828 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+2482C */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+24830 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+24834 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+24838 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+2483C */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+24840 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+24844 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+24848 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+2484C */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+24850 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+24854 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+24858 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+2485C */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+24860 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+24864 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+24868 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+2486C */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+24870 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+24874 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+24878 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+2487C */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+24880 */	"\x00\x00", "\x00\x00", "\x8E\x74", "\x00\x00",
  /* U+24884 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+24888 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+2488C */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+24890 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+24894 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+24898 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+2489C */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+248A0 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+248A4 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+248A8 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+248AC */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+248B0 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+248B4 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+248B8 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+248BC */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+248C0 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+248C4 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+248C8 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+248CC */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+248D0 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+248D4 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+248D8 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+248DC */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+248E0 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+248E4 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+248E8 */	"\x00\x00", "\x8B\xD6", "\x00\x00", "\x00\x00",
  /* U+248EC */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+248F0 */	"\x94\xB6", "\xFD\x74", "\x98\xC0", "\x94\xA5",
  /* U+248F4 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+248F8 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x9B\xC8",
  /* U+248FC */	"\x00\x00", "\x00\x00", "\x00\x00", "\x95\xED",
  /* U+24900 */	"\xFD\x7E", "\xFB\xEB", "\x00\x00", "\x00\x00",
  /* U+24904 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+24908 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+2490C */	"\xFD\x7D", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+24910 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+24914 */	"\x00\x00", "\x00\x00", "\x97\x6F", "\x94\x61",
  /* U+24918 */	"\x00\x00", "\x9F\xC1", "\x00\x00", "\x00\x00",
  /* U+2491C */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+24920 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+24924 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+24928 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+2492C */	"\x00\x00", "\x00\x00", "\x00\x00", "\x95\xD7",
  /* U+24930 */	"\x00\x00", "\x00\x00", "\x00\x00", "\xFA\x52",
  /* U+24934 */	"\x9C\x58", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+24938 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+2493C */	"\x00\x00", "\x00\x00", "\x9F\x68", "\x9B\xE7",
  /* U+24940 */	"\xFC\xCE", "\x96\xE8", "\xFA\x49", "\x97\xA1",
  /* U+24944 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+24948 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+2494C */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+24950 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+24954 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+24958 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+2495C */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+24960 */	"\x00\x00", "\x00\x00", "\x95\x4D", "\x9E\xF8",
  /* U+24964 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+24968 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+2496C */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+24970 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+24974 */	"\xFE\x49", "\x91\xCE", "\x97\x71", "\x00\x00",
  /* U+24978 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x8C\xCF",
  /* U+2497C */	"\x00\x00", "\x00\x00", "\x00\x00", "\xFD\xB1",
  /* U+24980 */	"\x00\x00", "\x00\x00", "\xFC\x6E", "\x00\x00",
  /* U+24984 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+24988 */	"\x9C\xF2", "\x93\xB8", "\x90\x43", "\x97\x59",
  /* U+2498C */	"\x94\xD7", "\xFE\x66", "\x94\x7D", "\xFC\x6F",
  /* U+24990 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+24994 */	"\x92\x46", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+24998 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+2499C */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+249A0 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+249A4 */	"\xFA\x6D", "\x00\x00", "\x00\x00", "\x8E\xF7",
  /* U+249A8 */	"\x00\x00", "\xFB\xB7", "\x00\x00", "\x94\x7C",
  /* U+249AC */	"\x92\xCD", "\x97\xB2", "\x00\x00", "\x00\x00",
  /* U+249B0 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+249B4 */	"\x00\x00", "\x00\x00", "\x00\x00", "\xFE\x65",
  /* U+249B8 */	"\x96\x7E", "\x97\x58", "\x9B\x77", "\x91\xCF",
  /* U+249BC */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+249C0 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+249C4 */	"\x00\x00", "\x94\xA4", "\x00\x00", "\x00\x00",
  /* U+249C8 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+249CC */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+249D0 */	"\x9C\xAD", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+249D4 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+249D8 */	"\x00\x00", "\x00\x00", "\x8B\xAB", "\x00\x00",
  /* U+249DC */	"\x00\x00", "\x00\x00", "\x96\xD5", "\xFC\xB3",
  /* U+249E0 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x93\xAE",
  /* U+249E4 */	"\x00\x00", "\x97\x6D", "\x00\x00", "\x00\x00",
  /* U+249E8 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+249EC */	"\x94\x46", "\x95\xF7", "\x00\x00", "\x00\x00",
  /* U+249F0 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+249F4 */	"\x00\x00", "\x00\x00", "\x9C\x46", "\x95\x5B",
  /* U+249F8 */	"\x91\xD1", "\x94\xF4", "\x00\x00", "\xFE\x67",
  /* U+249FC */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+24A00 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+24A04 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+24A08 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+24A0C */	"\x00\x00", "\x00\x00", "\x92\xA5", "\x00\x00",
  /* U+24A10 */	"\x00\x00", "\x00\x00", "\xFE\xDF", "\x8C\xAB",
  /* U+24A14 */	"\x00\x00", "\x9B\xC9", "\x00\x00", "\x00\x00",
  /* U+24A18 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+24A1C */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+24A20 */	"\x00\x00", "\xFC\xED", "\xFD\xFA", "\xFC\xC8",
  /* U+24A24 */	"\xFE\x62", "\x91\xFC", "\xFE\x6B", "\xFD\xF9",
  /* U+24A28 */	"\xFC\xC7", "\x91\x4E", "\x9C\xB8", "\x00\x00",
  /* U+24A2C */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+24A30 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+24A34 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+24A38 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+24A3C */	"\x00\x00", "\x00\x00", "\x97\x67", "\x00\x00",
  /* U+24A40 */	"\x00\x00", "\x00\x00", "\x95\xEE", "\x00\x00",
  /* U+24A44 */	"\x00\x00", "\x9B\xB2", "\x00\x00", "\x00\x00",
  /* U+24A48 */	"\x00\x00", "\x00\x00", "\x94\x60", "\x00\x00",
  /* U+24A4C */	"\x00\x00", "\x00\x00", "\x94\xA2", "\x98\x75",
  /* U+24A50 */	"\x97\xAC", "\x91\xD3", "\x00\x00", "\x00\x00",
  /* U+24A54 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+24A58 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+24A5C */	"\x00\x00", "\x98\x7B", "\x00\x00", "\x00\x00",
  /* U+24A60 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+24A64 */	"\x00\x00", "\x8E\xEB", "\x97\x6A", "\x96\x5E",
  /* U+24A68 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+24A6C */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+24A70 */	"\x00\x00", "\x97\xEB", "\x00\x00", "\x00\x00",
  /* U+24A74 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x9F\xF9",
  /* U+24A78 */	"\x95\xF8", "\xFE\xA2", "\x8F\xE6", "\x00\x00",
  /* U+24A7C */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+24A80 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+24A84 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+24A88 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+24A8C */	"\xFE\x7E", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+24A90 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x9D\xA4",
  /* U+24A94 */	"\x97\x68", "\x8E\xEC", "\x94\xBD", "\x00\x00",
  /* U+24A98 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+24A9C */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+24AA0 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+24AA4 */	"\x94\x5B", "\x9C\xF6", "\xFA\xA7", "\x9B\xD9",
  /* U+24AA8 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+24AAC */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+24AB0 */	"\x00\x00", "\xFA\x5D", "\x96\x56", "\x97\x62",
  /* U+24AB4 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+24AB8 */	"\x00\x00", "\x00\x00", "\x94\xBA", "\xA0\x4F",
  /* U+24ABC */	"\x92\xD8", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+24AC0 */	"\x9B\xCB", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+24AC4 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x94\xBB",
  /* U+24AC8 */	"\x00\x00", "\x00\x00", "\x9D\x5F", "\x00\x00",
  /* U+24ACC */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+24AD0 */	"\x00\x00", "\x90\xCF", "\x00\x00", "\x00\x00",
  /* U+24AD4 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+24AD8 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+24ADC */	"\x00\x00", "\x00\x00", "\x00\x00", "\x94\x65",
  /* U+24AE0 */	"\x00\x00", "\x00\x00", "\x9F\x4C", "\x00\x00",
  /* U+24AE4 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+24AE8 */	"\x00\x00", "\x90\xD8",

  /* U+24B6E */				"\x9E\xBE",

  /* U+24BF5 */		    "\xFB\x6D", "\x00\x00", "\x00\x00",
  /* U+24BF8 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+24BFC */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+24C00 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+24C04 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+24C08 */	"\x00\x00", "\x95\xCA",

  /* U+24C9E */				"\x9D\xC2", "\x97\xF8",
  /* U+24CA0 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+24CA4 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+24CA8 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+24CAC */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+24CB0 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+24CB4 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+24CB8 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+24CBC */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+24CC0 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+24CC4 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+24CC8 */	"\x00\x00", "\x8F\xFC", "\x00\x00", "\x00\x00",
  /* U+24CCC */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+24CD0 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+24CD4 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+24CD8 */	"\x00\x00", "\x94\x73", "\x00\x00", "\x00\x00",
  /* U+24CDC */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+24CE0 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+24CE4 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+24CE8 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+24CEC */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+24CF0 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+24CF4 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+24CF8 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+24CFC */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+24D00 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+24D04 */	"\x00\x00", "\x00\x00", "\x94\x74", "\x00\x00",
  /* U+24D08 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+24D0C */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+24D10 */	"\x00\x00", "\x00\x00", "\x00\x00", "\xFE\xB7",

  /* U+24DB8 */	"\x8A\x4B", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+24DBC */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+24DC0 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+24DC4 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+24DC8 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+24DCC */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+24DD0 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+24DD4 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+24DD8 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+24DDC */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+24DE0 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+24DE4 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+24DE8 */	"\x00\x00", "\x00\x00", "\x8A\x55", "\x8B\x69",
  /* U+24DEC */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+24DF0 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+24DF4 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+24DF8 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+24DFC */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+24E00 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+24E04 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+24E08 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+24E0C */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+24E10 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+24E14 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+24E18 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+24E1C */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+24E20 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+24E24 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+24E28 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+24E2C */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+24E30 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+24E34 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+24E38 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x8A\xDC",
  /* U+24E3C */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+24E40 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+24E44 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+24E48 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+24E4C */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+24E50 */	"\x8B\x76", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+24E54 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+24E58 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+24E5C */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+24E60 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+24E64 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+24E68 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+24E6C */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+24E70 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+24E74 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+24E78 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+24E7C */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+24E80 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+24E84 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+24E88 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+24E8C */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+24E90 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+24E94 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+24E98 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+24E9C */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+24EA0 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+24EA4 */	"\x00\x00", "\x9B\xCE", "\x00\x00", "\x8A\x68",
  /* U+24EA8 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+24EAC */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+24EB0 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+24EB4 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+24EB8 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+24EBC */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+24EC0 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+24EC4 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+24EC8 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+24ECC */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+24ED0 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+24ED4 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+24ED8 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+24EDC */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+24EE0 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+24EE4 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+24EE8 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+24EEC */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+24EF0 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+24EF4 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+24EF8 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+24EFC */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+24F00 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+24F04 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+24F08 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+24F0C */	"\x00\x00", "\x00\x00", "\xA0\xF8", "\x00\x00",
  /* U+24F10 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+24F14 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+24F18 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+24F1C */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+24F20 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+24F24 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+24F28 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+24F2C */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+24F30 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+24F34 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+24F38 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+24F3C */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+24F40 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+24F44 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+24F48 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+24F4C */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+24F50 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+24F54 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+24F58 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+24F5C */	"\x98\xDF", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+24F60 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+24F64 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+24F68 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+24F6C */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+24F70 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+24F74 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+24F78 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+24F7C */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+24F80 */	"\x00\x00", "\x00\x00", "\xFE\xB5", "\x00\x00",
  /* U+24F84 */	"\x00\x00", "\x00\x00", "\x9B\xCF", "\x00\x00",
  /* U+24F88 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+24F8C */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+24F90 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+24F94 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x96\xFB",
  /* U+24F98 */	"\x00\x00", "\x00\x00", "\x9B\xFB", "\x00\x00",
  /* U+24F9C */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+24FA0 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+24FA4 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+24FA8 */	"\x00\x00", "\x9E\xCE", "\x00\x00", "\x00\x00",
  /* U+24FAC */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+24FB0 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+24FB4 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+24FB8 */	"\x8E\xE5", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+24FBC */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+24FC0 */	"\x00\x00", "\x00\x00", "\x9E\x7B", "\x00\x00",
  /* U+24FC4 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+24FC8 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+24FCC */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+24FD0 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+24FD4 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+24FD8 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+24FDC */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+24FE0 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+24FE4 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+24FE8 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+24FEC */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+24FF0 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+24FF4 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+24FF8 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+24FFC */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+25000 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+25004 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+25008 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+2500C */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+25010 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+25014 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+25018 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+2501C */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+25020 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+25024 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+25028 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+2502C */	"\x9B\xD2", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+25030 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+25034 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+25038 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+2503C */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+25040 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+25044 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+25048 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+2504C */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+25050 */	"\x00\x00", "\x00\x00", "\x8A\xA5", "\x00\x00",
  /* U+25054 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+25058 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+2505C */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+25060 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+25064 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+25068 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+2506C */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+25070 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+25074 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+25078 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+2507C */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+25080 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+25084 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+25088 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+2508C */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+25090 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+25094 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+25098 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+2509C */	"\x00\x00", "\xFE\xCE",

  /* U+2512B */					    "\x8A\x45",
  /* U+2512C */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+25130 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+25134 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+25138 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+2513C */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+25140 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+25144 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+25148 */	"\x9D\xFC", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+2514C */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+25150 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+25154 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+25158 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+2515C */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+25160 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+25164 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+25168 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+2516C */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+25170 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+25174 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+25178 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+2517C */	"\x00\x00", "\xFE\xCF", "\x8B\xA5", "\x00\x00",
  /* U+25180 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+25184 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+25188 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+2518C */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+25190 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+25194 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+25198 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+2519C */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+251A0 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+251A4 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+251A8 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+251AC */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+251B0 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+251B4 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+251B8 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+251BC */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+251C0 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+251C4 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+251C8 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+251CC */	"\x00\x00", "\x8C\x4A", "\x00\x00", "\x00\x00",
  /* U+251D0 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+251D4 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+251D8 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+251DC */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+251E0 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x8A\xEC",
  /* U+251E4 */	"\x00\x00", "\x00\x00", "\xFC\xE0", "\x94\xAD",
  /* U+251E8 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+251EC */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+251F0 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+251F4 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+251F8 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+251FC */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+25200 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+25204 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+25208 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+2520C */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+25210 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+25214 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+25218 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+2521C */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+25220 */	"\xFE\xD5", "\x94\xAC", "\x00\x00", "\x00\x00",
  /* U+25224 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+25228 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+2522C */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+25230 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+25234 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+25238 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+2523C */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+25240 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+25244 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+25248 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+2524C */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+25250 */	"\xFC\x5A", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+25254 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+25258 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+2525C */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+25260 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+25264 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+25268 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+2526C */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+25270 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+25274 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+25278 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+2527C */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+25280 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+25284 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+25288 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+2528C */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+25290 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+25294 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+25298 */	"\x00\x00", "\x9B\xD6", "\x00\x00", "\x00\x00",
  /* U+2529C */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+252A0 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+252A4 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+252A8 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+252AC */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+252B0 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+252B4 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+252B8 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+252BC */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+252C0 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+252C4 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x8A\x6F",
  /* U+252C8 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+252CC */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+252D0 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+252D4 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+252D8 */	"\x8B\xA9", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+252DC */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+252E0 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+252E4 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+252E8 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+252EC */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+252F0 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+252F4 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+252F8 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+252FC */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+25300 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+25304 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+25308 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+2530C */	"\x00\x00", "\x00\x00", "\x8E\x5F", "\x00\x00",
  /* U+25310 */	"\x00\x00", "\x9D\xCB", "\x00\x00", "\xFC\xE7",

  /* U+25419 */		    "\x9B\xD7", "\x00\x00", "\x00\x00",
  /* U+2541C */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+25420 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+25424 */	"\x00\x00", "\x93\xC8", "\x00\x00", "\x00\x00",
  /* U+25428 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+2542C */	"\x00\x00", "\x00\x00", "\x00\x00", "\x91\xF0",
  /* U+25430 */	"\x8F\xE0", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+25434 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+25438 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+2543C */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+25440 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+25444 */	"\x00\x00", "\x00\x00", "\x9B\xDB", "\x00\x00",
  /* U+25448 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+2544C */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+25450 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+25454 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+25458 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+2545C */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+25460 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+25464 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+25468 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+2546C */	"\x90\xED", "\x00\x00", "\x9B\xDC",

  /* U+25531 */		    "\xA0\xEC", "\x00\x00", "\x00\x00",
  /* U+25534 */	"\x00\x00", "\x98\xFA", "\x00\x00", "\x00\x00",
  /* U+25538 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+2553C */	"\x00\x00", "\x00\x00", "\x00\x00", "\x9B\xE0",
  /* U+25540 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+25544 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+25548 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+2554C */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+25550 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+25554 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+25558 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x93\xC7",
  /* U+2555C */	"\x92\x49", "\x96\xE1", "\x9B\xE2", "\x00\x00",
  /* U+25560 */	"\x00\x00", "\x00\x00", "\x9B\xE4", "\x00\x00",
  /* U+25564 */	"\x00\x00", "\x8F\xE1", "\x9B\xE5", "\x00\x00",
  /* U+25568 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+2556C */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+25570 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+25574 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+25578 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+2557C */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+25580 */	"\x00\x00", "\x94\xC0", "\x00\x00", "\x00\x00",
  /* U+25584 */	"\x93\xC3", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+25588 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+2558C */	"\x00\x00", "\x00\x00", "\x00\x00", "\x93\xC5",
  /* U+25590 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+25594 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+25598 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+2559C */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+255A0 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+255A4 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+255A8 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+255AC */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+255B0 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+255B4 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+255B8 */	"\x00\x00", "\x90\x79", "\x00\x00", "\x00\x00",
  /* U+255BC */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+255C0 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+255C4 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+255C8 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+255CC */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+255D0 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+255D4 */	"\x00\x00", "\x97\x7B", "\x00\x00", "\x00\x00",
  /* U+255D8 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x90\x7E",
  /* U+255DC */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+255E0 */	"\xFE\xE6", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+255E4 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+255E8 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+255EC */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+255F0 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+255F4 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+255F8 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+255FC */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+25600 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+25604 */	"\x00\x00", "\xFE\x46", "\x00\x00", "\x00\x00",
  /* U+25608 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+2560C */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+25610 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+25614 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+25618 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+2561C */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+25620 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+25624 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+25628 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+2562C */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+25630 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+25634 */	"\x00\x00", "\x9D\xB8", "\x00\x00", "\x00\x00",
  /* U+25638 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+2563C */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+25640 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+25644 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+25648 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+2564C */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+25650 */	"\x00\x00", "\x92\x70", "\x00\x00", "\x00\x00",
  /* U+25654 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+25658 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+2565C */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+25660 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+25664 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+25668 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+2566C */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+25670 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+25674 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+25678 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+2567C */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+25680 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x95\xA8",
  /* U+25684 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+25688 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+2568C */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+25690 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+25694 */	"\x00\x00", "\x8C\xB0", "\x00\x00", "\x00\x00",
  /* U+25698 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+2569C */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+256A0 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+256A4 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+256A8 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+256AC */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+256B0 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+256B4 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+256B8 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+256BC */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+256C0 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+256C4 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+256C8 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+256CC */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+256D0 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+256D4 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+256D8 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+256DC */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+256E0 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x94\xC8",
  /* U+256E4 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+256E8 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+256EC */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+256F0 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+256F4 */	"\x00\x00", "\x00\x00", "\x98\xB9", "\x00\x00",
  /* U+256F8 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+256FC */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+25700 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+25704 */	"\x00\x00", "\x00\x00", "\x91\x40", "\x00\x00",
  /* U+25708 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+2570C */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+25710 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+25714 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+25718 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+2571C */	"\x00\x00", "\xFC\xBE", "\x00\x00", "\x00\x00",
  /* U+25720 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+25724 */	"\x00\x00", "\x91\x57", "\x00\x00", "\x00\x00",
  /* U+25728 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+2572C */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+25730 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+25734 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+25738 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+2573C */	"\x00\x00", "\x8B\xB2", "\x00\x00", "\x00\x00",
  /* U+25740 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+25744 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+25748 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+2574C */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+25750 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+25754 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+25758 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+2575C */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+25760 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+25764 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+25768 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+2576C */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+25770 */	"\x00\x00", "\x00\x00", "\xFA\xDF", "\x00\x00",
  /* U+25774 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+25778 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+2577C */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+25780 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+25784 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+25788 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+2578C */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+25790 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+25794 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+25798 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+2579C */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+257A0 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+257A4 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+257A8 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+257AC */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+257B0 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+257B4 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+257B8 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+257BC */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+257C0 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+257C4 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x9B\xE6",
  /* U+257C8 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+257CC */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+257D0 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+257D4 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+257D8 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+257DC */	"\x00\x00", "\x00\x00", "\x00\x00", "\x96\x43",
  /* U+257E0 */	"\x8E\x44", "\x9C\x4F", "\x00\x00", "\x00\x00",
  /* U+257E4 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+257E8 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+257EC */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+257F0 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+257F4 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+257F8 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+257FC */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+25800 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+25804 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+25808 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+2580C */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+25810 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+25814 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+25818 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+2581C */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+25820 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+25824 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+25828 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+2582C */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+25830 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+25834 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+25838 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+2583C */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+25840 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+25844 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+25848 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+2584C */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+25850 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+25854 */	"\x00\x00", "\x00\x00", "\x00\x00", "\xFE\xF4",
  /* U+25858 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+2585C */	"\x00\x00", "\x9B\xE8", "\x00\x00", "\x00\x00",
  /* U+25860 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+25864 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+25868 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+2586C */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+25870 */	"\x00\x00", "\x00\x00", "\x93\xDC", "\x00\x00",
  /* U+25874 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+25878 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+2587C */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+25880 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+25884 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+25888 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+2588C */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+25890 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+25894 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+25898 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+2589C */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+258A0 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+258A4 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+258A8 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+258AC */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+258B0 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+258B4 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+258B8 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+258BC */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+258C0 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+258C4 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+258C8 */	"\x96\x6F", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+258CC */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+258D0 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+258D4 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+258D8 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+258DC */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+258E0 */	"\x00\x00", "\x8E\x4A", "\x00\x00", "\x00\x00",
  /* U+258E4 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+258E8 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+258EC */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+258F0 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+258F4 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+258F8 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+258FC */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+25900 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x9B\xED",
  /* U+25904 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+25908 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+2590C */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+25910 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+25914 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+25918 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+2591C */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+25920 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+25924 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+25928 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+2592C */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+25930 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+25934 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+25938 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+2593C */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+25940 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+25944 */	"\x00\x00", "\x00\x00", "\x92\xF6", "\x00\x00",
  /* U+25948 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+2594C */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+25950 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+25954 */	"\x00\x00", "\x00\x00", "\x9D\xB9", "\x00\x00",
  /* U+25958 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+2595C */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+25960 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+25964 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+25968 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+2596C */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+25970 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+25974 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+25978 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+2597C */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+25980 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+25984 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+25988 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+2598C */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+25990 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+25994 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+25998 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+2599C */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+259A0 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+259A4 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+259A8 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+259AC */	"\x8E\x4E", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+259B0 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+259B4 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+259B8 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+259BC */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+259C0 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+259C4 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+259C8 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+259CC */	"\xFB\xCF",

  /* U+25A95 */		    "\x9E\xC2", "\x00\x00", "\x00\x00",
  /* U+25A98 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+25A9C */	"\x94\xE5", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+25AA0 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+25AA4 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+25AA8 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+25AAC */	"\x00\x00", "\x00\x00", "\x9B\xF0", "\x94\xE4",
  /* U+25AB0 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+25AB4 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+25AB8 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+25ABC */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+25AC0 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+25AC4 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+25AC8 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+25ACC */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+25AD0 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+25AD4 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+25AD8 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+25ADC */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+25AE0 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+25AE4 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+25AE8 */	"\x00\x00", "\x95\x51",

  /* U+25B74 */	"\x8B\xBB", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+25B78 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+25B7C */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+25B80 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+25B84 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+25B88 */	"\x00\x00", "\x9B\xF1", "\x00\x00", "\x00\x00",
  /* U+25B8C */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+25B90 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+25B94 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+25B98 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+25B9C */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+25BA0 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+25BA4 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+25BA8 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+25BAC */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+25BB0 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x94\xF0",
  /* U+25BB4 */	"\x8E\x64", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+25BB8 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+25BBC */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+25BC0 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+25BC4 */	"\x00\x00", "\x00\x00", "\x94\xEA", "\x00\x00",
  /* U+25BC8 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+25BCC */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+25BD0 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+25BD4 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+25BD8 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+25BDC */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+25BE0 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+25BE4 */	"\x8F\x61", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+25BE8 */	"\x9B\x64", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+25BEC */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+25BF0 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+25BF4 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+25BF8 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+25BFC */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+25C00 */	"\x00\x00", "\x8E\x5B", "\x00\x00", "\x00\x00",
  /* U+25C04 */	"\x00\x00", "\x00\x00", "\x9B\xF2", "\x00\x00",
  /* U+25C08 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+25C0C */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+25C10 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+25C14 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+25C18 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+25C1C */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+25C20 */	"\x00\x00", "\x9F\xBE", "\x00\x00", "\x00\x00",
  /* U+25C24 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+25C28 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+25C2C */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+25C30 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+25C34 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+25C38 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+25C3C */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+25C40 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+25C44 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+25C48 */	"\x00\x00", "\x00\x00", "\x9D\xC9", "\x00\x00",
  /* U+25C4C */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+25C50 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+25C54 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+25C58 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+25C5C */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+25C60 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+25C64 */	"\x00\x00", "\x8E\x6C", "\x00\x00", "\x00\x00",
  /* U+25C68 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+25C6C */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+25C70 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+25C74 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+25C78 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+25C7C */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+25C80 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+25C84 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+25C88 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+25C8C */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+25C90 */	"\x00\x00", "\x8F\x73", "\x00\x00", "\x00\x00",
  /* U+25C94 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+25C98 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+25C9C */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+25CA0 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+25CA4 */	"\x8C\xAF", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+25CA8 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+25CAC */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+25CB0 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+25CB4 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+25CB8 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+25CBC */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+25CC0 */	"\x8F\x75", "\x8E\x71", "\x00\x00", "\x00\x00",
  /* U+25CC4 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+25CC8 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+25CCC */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+25CD0 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+25CD4 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+25CD8 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+25CDC */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+25CE0 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+25CE4 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+25CE8 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+25CEC */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+25CF0 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+25CF4 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+25CF8 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+25CFC */	"\x00\x00", "\x00\x00", "\x8E\x60", "\x00\x00",
  /* U+25D00 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+25D04 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+25D08 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+25D0C */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+25D10 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+25D14 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+25D18 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+25D1C */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+25D20 */	"\x8E\x6A", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+25D24 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+25D28 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+25D2C */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+25D30 */	"\x8C\x4C", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+25D34 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+25D38 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+25D3C */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+25D40 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x95\x52",

  /* U+25E0E */				"\x95\x54", "\x00\x00",
  /* U+25E10 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+25E14 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+25E18 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+25E1C */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+25E20 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+25E24 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+25E28 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+25E2C */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+25E30 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+25E34 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+25E38 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+25E3C */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+25E40 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+25E44 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+25E48 */	"\x00\x00", "\x8A\xD4", "\x00\x00", "\x00\x00",
  /* U+25E4C */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+25E50 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+25E54 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+25E58 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+25E5C */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+25E60 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+25E64 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+25E68 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+25E6C */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+25E70 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+25E74 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+25E78 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+25E7C */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+25E80 */	"\x00\x00", "\x9D\xBB", "\x95\x43", "\x92\xFE",
  /* U+25E84 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+25E88 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+25E8C */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+25E90 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+25E94 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+25E98 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+25E9C */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+25EA0 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+25EA4 */	"\x00\x00", "\x00\x00", "\x94\xF2", "\x00\x00",
  /* U+25EA8 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+25EAC */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+25EB0 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+25EB4 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+25EB8 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+25EBC */	"\x94\xF1", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+25EC0 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+25EC4 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+25EC8 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+25ECC */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+25ED0 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+25ED4 */	"\x00\x00", "\x00\x00", "\x00\x00", "\xA0\xEA",
  /* U+25ED8 */	"\x9D\xD2", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+25EDC */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+25EE0 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+25EE4 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+25EE8 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+25EEC */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+25EF0 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+25EF4 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+25EF8 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+25EFC */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+25F00 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+25F04 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+25F08 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+25F0C */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+25F10 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+25F14 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+25F18 */	"\x00\x00", "\x00\x00", "\xA0\xB1", "\x00\x00",
  /* U+25F1C */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+25F20 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+25F24 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+25F28 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+25F2C */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+25F30 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+25F34 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+25F38 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+25F3C */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+25F40 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+25F44 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+25F48 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x91\xF8",

  /* U+25FE1 */		    "\x94\x62", "\x9B\xA4", "\x00\x00",
  /* U+25FE4 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+25FE8 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+25FEC */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+25FF0 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+25FF4 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+25FF8 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+25FFC */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+26000 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+26004 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+26008 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+2600C */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+26010 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+26014 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+26018 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+2601C */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+26020 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+26024 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+26028 */	"\x00\x00", "\x8E\xAD", "\x00\x00", "\x00\x00",
  /* U+2602C */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+26030 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+26034 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+26038 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+2603C */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+26040 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+26044 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+26048 */	"\x9E\xAD", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+2604C */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+26050 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+26054 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+26058 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+2605C */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+26060 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+26064 */	"\x96\xD0", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+26068 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+2606C */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+26070 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+26074 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+26078 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+2607C */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+26080 */	"\x00\x00", "\x00\x00", "\x00\x00", "\xFE\xEE",
  /* U+26084 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+26088 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+2608C */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+26090 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+26094 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x8A\xB4",
  /* U+26098 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+2609C */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+260A0 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+260A4 */	"\x97\x57", "\x8A\x77", "\x00\x00", "\x00\x00",
  /* U+260A8 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+260AC */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+260B0 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+260B4 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+260B8 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+260BC */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+260C0 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+260C4 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+260C8 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+260CC */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+260D0 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+260D4 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+260D8 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+260DC */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+260E0 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+260E4 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+260E8 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+260EC */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+260F0 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+260F4 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+260F8 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+260FC */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+26100 */	"\x00\x00", "\x00\x00", "\x9B\xF7", "\x00\x00",
  /* U+26104 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+26108 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+2610C */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+26110 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+26114 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+26118 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+2611C */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+26120 */	"\x00\x00", "\x8E\xB5", "\x00\x00", "\x00\x00",
  /* U+26124 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+26128 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+2612C */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+26130 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+26134 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+26138 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+2613C */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+26140 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+26144 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+26148 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+2614C */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+26150 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+26154 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+26158 */	"\x00\x00", "\xA0\x6D", "\x8E\xB6", "\x97\x56",
  /* U+2615C */	"\x95\x40", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+26160 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+26164 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+26168 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+2616C */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+26170 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+26174 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+26178 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+2617C */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+26180 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+26184 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+26188 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+2618C */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+26190 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+26194 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+26198 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+2619C */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+261A0 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+261A4 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+261A8 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+261AC */	"\x00\x00", "\xA0\xF3", "\x94\xBE", "\x00\x00",
  /* U+261B0 */	"\x00\x00", "\x00\x00", "\x9B\xFA", "\x00\x00",
  /* U+261B4 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+261B8 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+261BC */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+261C0 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+261C4 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+261C8 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+261CC */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+261D0 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+261D4 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+261D8 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+261DC */	"\x00\x00", "\xFD\xDF", "\x00\x00", "\x00\x00",
  /* U+261E0 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+261E4 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+261E8 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+261EC */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+261F0 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+261F4 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+261F8 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+261FC */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+26200 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+26204 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+26208 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+2620C */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+26210 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+26214 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+26218 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+2621C */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+26220 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+26224 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+26228 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+2622C */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+26230 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+26234 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+26238 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+2623C */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+26240 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+26244 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+26248 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+2624C */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+26250 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+26254 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+26258 */	"\x9D\xBC", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+2625C */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+26260 */	"\x00\x00", "\x94\xFE", "\x00\x00", "\x00\x00",
  /* U+26264 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+26268 */	"\x00\x00", "\x00\x00", "\x8B\xDB", "\xA0\xFE",
  /* U+2626C */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+26270 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+26274 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+26278 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+2627C */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+26280 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+26284 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+26288 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+2628C */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+26290 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+26294 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+26298 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+2629C */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+262A0 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+262A4 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+262A8 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+262AC */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+262B0 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+262B4 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+262B8 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+262BC */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+262C0 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+262C4 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+262C8 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+262CC */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+262D0 */	"\x8E\xC0", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+262D4 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+262D8 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+262DC */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+262E0 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+262E4 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+262E8 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+262EC */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+262F0 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+262F4 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+262F8 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+262FC */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+26300 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+26304 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+26308 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+2630C */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+26310 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+26314 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+26318 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+2631C */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+26320 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+26324 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+26328 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+2632C */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+26330 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+26334 */	"\x00\x00", "\x9F\x47", "\x00\x00", "\x00\x00",
  /* U+26338 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+2633C */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+26340 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+26344 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+26348 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x8B\xDE",
  /* U+2634C */	"\xA0\xFB", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+26350 */	"\x00\x00", "\x8E\xC3", "\x00\x00", "\x00\x00",
  /* U+26354 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+26358 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+2635C */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+26360 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+26364 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+26368 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+2636C */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+26370 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+26374 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+26378 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+2637C */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+26380 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+26384 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+26388 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+2638C */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+26390 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+26394 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+26398 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+2639C */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+263A0 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+263A4 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+263A8 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+263AC */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+263B0 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+263B4 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+263B8 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+263BC */	"\x00\x00", "\x00\x00", "\x96\x49", "\x00\x00",
  /* U+263C0 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+263C4 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+263C8 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+263CC */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+263D0 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+263D4 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+263D8 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+263DC */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+263E0 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+263E4 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+263E8 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+263EC */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+263F0 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+263F4 */	"\x00\x00", "\xFE\xC2", "\x00\x00", "\x00\x00",
  /* U+263F8 */	"\x95\x4C", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+263FC */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+26400 */	"\x00\x00", "\x00\x00", "\x9B\xFD", "\x00\x00",
  /* U+26404 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+26408 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+2640C */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+26410 */	"\x90\xCC", "\x9C\x60", "\x95\x4B", "\x00\x00",
  /* U+26414 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+26418 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+2641C */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+26420 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+26424 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+26428 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+2642C */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+26430 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+26434 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+26438 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+2643C */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+26440 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+26444 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+26448 */	"\x00\x00", "\x00\x00", "\x9B\xFE", "\x00\x00",
  /* U+2644C */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+26450 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+26454 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+26458 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+2645C */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+26460 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+26464 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+26468 */	"\x00\x00", "\x9C\x70", "\x00\x00", "\x00\x00",
  /* U+2646C */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+26470 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+26474 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+26478 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+2647C */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+26480 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+26484 */	"\x9C\x43", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+26488 */	"\x9C\x47", "\x8E\xCC", "\x00\x00", "\x00\x00",
  /* U+2648C */	"\x00\x00", "\x8E\x54", "\x00\x00", "\x00\x00",
  /* U+26490 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+26494 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+26498 */	"\x8E\xE4", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+2649C */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+264A0 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+264A4 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+264A8 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+264AC */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+264B0 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+264B4 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+264B8 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+264BC */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+264C0 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+264C4 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+264C8 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+264CC */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+264D0 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+264D4 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+264D8 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+264DC */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+264E0 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+264E4 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+264E8 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+264EC */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+264F0 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+264F4 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+264F8 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+264FC */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+26500 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+26504 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+26508 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+2650C */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+26510 */	"\x00\x00", "\x00\x00", "\x9C\x49", "\x00\x00",
  /* U+26514 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+26518 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+2651C */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+26520 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+26524 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+26528 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+2652C */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+26530 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+26534 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+26538 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+2653C */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+26540 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+26544 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+26548 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+2654C */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+26550 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+26554 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+26558 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+2655C */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+26560 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+26564 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+26568 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+2656C */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+26570 */	"\x00\x00", "\x00\x00", "\x8B\x5E", "\x00\x00",
  /* U+26574 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+26578 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+2657C */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+26580 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+26584 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+26588 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+2658C */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+26590 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+26594 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+26598 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+2659C */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+265A0 */	"\x95\x5E", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+265A4 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+265A8 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+265AC */	"\x00\x00", "\x95\x5C", "\x00\x00", "\x00\x00",
  /* U+265B0 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+265B4 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+265B8 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+265BC */	"\x00\x00", "\x00\x00", "\x00\x00", "\x9C\x4B",
  /* U+265C0 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+265C4 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+265C8 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+265CC */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+265D0 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+265D4 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+265D8 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+265DC */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+265E0 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+265E4 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+265E8 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+265EC */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+265F0 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+265F4 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+265F8 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+265FC */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+26600 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+26604 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+26608 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+2660C */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+26610 */	"\x00\x00", "\x00\x00", "\x8B\xE1", "\x00\x00",
  /* U+26614 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+26618 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+2661C */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+26620 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+26624 */	"\x00\x00", "\x00\x00", "\x8E\xD9",

  /* U+266AF */					    "\x9D\xB4",
  /* U+266B0 */	"\x00\x00", "\x92\x5F", "\x00\x00", "\x00\x00",
  /* U+266B4 */	"\x00\x00", "\x9C\x4C", "\x00\x00", "\x00\x00",
  /* U+266B8 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+266BC */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+266C0 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+266C4 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+266C8 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+266CC */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+266D0 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+266D4 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+266D8 */	"\x00\x00", "\x00\x00", "\x8A\xA1", "\x00\x00",
  /* U+266DC */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+266E0 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+266E4 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+266E8 */	"\x8E\xDB", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+266EC */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+266F0 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+266F4 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+266F8 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+266FC */	"\x9C\x56", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+26700 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+26704 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+26708 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+2670C */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+26710 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+26714 */	"\x00\x00", "\x00\x00", "\x8A\xA2", "\x00\x00",
  /* U+26718 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+2671C */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+26720 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+26724 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+26728 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+2672C */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+26730 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+26734 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+26738 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+2673C */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+26740 */	"\x00\x00", "\x97\x54", "\x00\x00", "\x00\x00",
  /* U+26744 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+26748 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+2674C */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+26750 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+26754 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+26758 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+2675C */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+26760 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+26764 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+26768 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+2676C */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+26770 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+26774 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+26778 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+2677C */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+26780 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+26784 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+26788 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+2678C */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+26790 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+26794 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+26798 */	"\x00\x00", "\x9C\x5E", "\x00\x00", "\x00\x00",
  /* U+2679C */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+267A0 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+267A4 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+267A8 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+267AC */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+267B0 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x9E\xD4",
  /* U+267B4 */	"\x95\x68", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+267B8 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+267BC */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+267C0 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+267C4 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+267C8 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+267CC */	"\xA0\xC3", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+267D0 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+267D4 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+267D8 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+267DC */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+267E0 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+267E4 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+267E8 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+267EC */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+267F0 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+267F4 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+267F8 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+267FC */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+26800 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+26804 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+26808 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+2680C */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+26810 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+26814 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+26818 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+2681C */	"\x8A\xE6", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+26820 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+26824 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+26828 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+2682C */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+26830 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+26834 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+26838 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+2683C */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+26840 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+26844 */	"\x00\x00", "\x00\x00", "\xA0\xF7", "\x00\x00",
  /* U+26848 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+2684C */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+26850 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+26854 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+26858 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+2685C */	"\x00\x00", "\x00\x00", "\x9C\x61", "\x00\x00",
  /* U+26860 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+26864 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+26868 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+2686C */	"\x00\x00", "\x00\x00", "\x9C\x5F", "\x00\x00",
  /* U+26870 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+26874 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+26878 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+2687C */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+26880 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+26884 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+26888 */	"\xFC\x4D", "\x00\x00", "\x9E\x5B", "\x00\x00",
  /* U+2688C */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+26890 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x9E\x69",
  /* U+26894 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+26898 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+2689C */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+268A0 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+268A4 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+268A8 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+268AC */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+268B0 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+268B4 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+268B8 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+268BC */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+268C0 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+268C4 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x9C\x63",
  /* U+268C8 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+268CC */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+268D0 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+268D4 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+268D8 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+268DC */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+268E0 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+268E4 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+268E8 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+268EC */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+268F0 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+268F4 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+268F8 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+268FC */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+26900 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+26904 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+26908 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+2690C */	"\x00\x00", "\x00\x00", "\xFE\xC7", "\x00\x00",
  /* U+26910 */	"\x00\x00", "\xFE\xC6", "\x00\x00", "\x00\x00",
  /* U+26914 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+26918 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+2691C */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+26920 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+26924 */	"\x00\x00", "\x00\x00", "\x9C\x67", "\x00\x00",
  /* U+26928 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+2692C */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+26930 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+26934 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+26938 */	"\x00\x00", "\x9C\x69", "\x00\x00", "\x00\x00",
  /* U+2693C */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+26940 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+26944 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+26948 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+2694C */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+26950 */	"\x00\x00", "\x8B\xE2", "\x00\x00", "\x00\x00",
  /* U+26954 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+26958 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+2695C */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+26960 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+26964 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+26968 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+2696C */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+26970 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+26974 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+26978 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+2697C */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+26980 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+26984 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+26988 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+2698C */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+26990 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+26994 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+26998 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+2699C */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+269A0 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+269A4 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+269A8 */	"\x91\x65", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+269AC */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+269B0 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+269B4 */	"\x00\x00", "\x9C\xE7", "\x00\x00", "\x00\x00",
  /* U+269B8 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+269BC */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+269C0 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+269C4 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+269C8 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+269CC */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+269D0 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+269D4 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+269D8 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+269DC */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+269E0 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+269E4 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+269E8 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+269EC */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+269F0 */	"\x00\x00", "\x00\x00", "\x8A\x54", "\x00\x00",
  /* U+269F4 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+269F8 */	"\x00\x00", "\x00\x00", "\x9C\x6C", "\x00\x00",
  /* U+269FC */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+26A00 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+26A04 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+26A08 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+26A0C */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+26A10 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+26A14 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+26A18 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+26A1C */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+26A20 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+26A24 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+26A28 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+26A2C */	"\x00\x00", "\x9C\x6E", "\xFE\x5D", "\x00\x00",
  /* U+26A30 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+26A34 */	"\x9C\x73", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+26A38 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+26A3C */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+26A40 */	"\x00\x00", "\x00\x00", "\x95\x6A", "\x00\x00",
  /* U+26A44 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+26A48 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+26A4C */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+26A50 */	"\x00\x00", "\x95\x6D", "\x8E\xF0",

  /* U+26B05 */		    "\x8F\x4D", "\x00\x00", "\x00\x00",
  /* U+26B08 */	"\x00\x00", "\x00\x00", "\x8E\xF6", "\x00\x00",
  /* U+26B0C */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+26B10 */	"\x00\x00", "\x00\x00", "\x00\x00", "\xFA\xBC",
  /* U+26B14 */	"\x00\x00", "\x8C\xD5", "\x00\x00", "\x00\x00",
  /* U+26B18 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+26B1C */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+26B20 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+26B24 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+26B28 */	"\xFB\xDA", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+26B2C */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+26B30 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+26B34 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+26B38 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+26B3C */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+26B40 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+26B44 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+26B48 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+26B4C */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+26B50 */	"\x8B\x4C", "\xFD\x75", "\x9B\xDD", "\xFA\xF5",
  /* U+26B54 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+26B58 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x9C\x74",
  /* U+26B5C */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+26B60 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+26B64 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+26B68 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+26B6C */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+26B70 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+26B74 */	"\x00\x00", "\x95\x45", "\x00\x00", "\x00\x00",
  /* U+26B78 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+26B7C */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+26B80 */	"\x00\x00", "\x00\x00", "\x96\xC6", "\x00\x00",
  /* U+26B84 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+26B88 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+26B8C */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+26B90 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+26B94 */	"\x00\x00", "\x00\x00", "\x8F\x6A", "\x8F\x4E",
  /* U+26B98 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+26B9C */	"\x00\x00", "\x9C\x78", "\x00\x00", "\x00\x00",
  /* U+26BA0 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+26BA4 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+26BA8 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+26BAC */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+26BB0 */	"\x00\x00", "\x00\x00", "\x00\x00", "\xFA\x55",
  /* U+26BB4 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+26BB8 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+26BBC */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+26BC0 */	"\x97\xE4", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+26BC4 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+26BC8 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+26BCC */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+26BD0 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+26BD4 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+26BD8 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+26BDC */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+26BE0 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+26BE4 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+26BE8 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+26BEC */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+26BF0 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+26BF4 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x9C\x41",
  /* U+26BF8 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+26BFC */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+26C00 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+26C04 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+26C08 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+26C0C */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+26C10 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+26C14 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+26C18 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+26C1C */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+26C20 */	"\x00\x00", "\x92\x5C", "\x00\x00", "\x00\x00",
  /* U+26C24 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+26C28 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+26C2C */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+26C30 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+26C34 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+26C38 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+26C3C */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+26C40 */	"\x96\xFA", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+26C44 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+26C48 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+26C4C */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+26C50 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+26C54 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+26C58 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+26C5C */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+26C60 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+26C64 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+26C68 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+26C6C */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+26C70 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+26C74 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+26C78 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+26C7C */	"\x00\x00", "\x00\x00", "\xFB\x66", "\x8E\x65",
  /* U+26C80 */	"\x98\x49", "\xFB\xA8", "\x98\x42", "\x00\x00",
  /* U+26C84 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+26C88 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+26C8C */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+26C90 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+26C94 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+26C98 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+26C9C */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+26CA0 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+26CA4 */	"\x9C\x7A", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+26CA8 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+26CAC */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+26CB0 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+26CB4 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x97\xFB",
  /* U+26CB8 */	"\x90\xCA", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+26CBC */	"\x00\x00", "\x9C\x5B", "\x00\x00", "\x00\x00",
  /* U+26CC0 */	"\x97\x4D", "\x00\x00", "\x00\x00", "\x8E\xD3",
  /* U+26CC4 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+26CC8 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+26CCC */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+26CD0 */	"\x00\x00", "\x95\x61", "\x00\x00", "\x00\x00",
  /* U+26CD4 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+26CD8 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+26CDC */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+26CE0 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+26CE4 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+26CE8 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+26CEC */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+26CF0 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+26CF4 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+26CF8 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+26CFC */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+26D00 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+26D04 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+26D08 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+26D0C */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+26D10 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+26D14 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+26D18 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+26D1C */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+26D20 */	"\x00\x00", "\x00\x00", "\x9F\x4B", "\x9F\xB5",
  /* U+26D24 */	"\x93\xD2", "\xFD\xAA", "\x98\x40", "\x91\x46",
  /* U+26D28 */	"\x98\x67", "\xFA\x5A", "\xFB\xA9", "\x00\x00",
  /* U+26D2C */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+26D30 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+26D34 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+26D38 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+26D3C */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+26D40 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+26D44 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+26D48 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+26D4C */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+26D50 */	"\x00\x00", "\x98\x41", "\x00\x00", "\x00\x00",
  /* U+26D54 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+26D58 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+26D5C */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+26D60 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+26D64 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+26D68 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+26D6C */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+26D70 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+26D74 */	"\x8C\xD3", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+26D78 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+26D7C */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+26D80 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+26D84 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+26D88 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+26D8C */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+26D90 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+26D94 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+26D98 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+26D9C */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+26DA0 */	"\xFC\xFD", "\xFD\xAB", "\x91\xBD", "\x8F\x4C",
  /* U+26DA4 */	"\x96\xC9", "\x8F\x55", "\xFB\xAE", "\x95\x6F",
  /* U+26DA8 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+26DAC */	"\x00\x00", "\x00\x00", "\x9C\x7D", "\x00\x00",
  /* U+26DB0 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+26DB4 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+26DB8 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+26DBC */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+26DC0 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+26DC4 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+26DC8 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+26DCC */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+26DD0 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+26DD4 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+26DD8 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+26DDC */	"\xA0\xF0", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+26DE0 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+26DE4 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+26DE8 */	"\x00\x00", "\x00\x00", "\x94\x6F", "\xFD\xAC",
  /* U+26DEC */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+26DF0 */	"\x96\xCB", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+26DF4 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+26DF8 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+26DFC */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+26E00 */	"\x96\xCE", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+26E04 */	"\x00\x00", "\xA0\x56", "\x00\x00", "\x9C\xE1",
  /* U+26E08 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+26E0C */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+26E10 */	"\x00\x00", "\x00\x00", "\x96\xC4", "\x00\x00",
  /* U+26E14 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+26E18 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+26E1C */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+26E20 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+26E24 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+26E28 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+26E2C */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+26E30 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+26E34 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+26E38 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+26E3C */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+26E40 */	"\x00\x00", "\x00\x00", "\x8F\x5E", "\x8F\x6C",
  /* U+26E44 */	"\x8E\xA3", "\xFB\xB3", "\x00\x00", "\x00\x00",
  /* U+26E48 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+26E4C */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+26E50 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+26E54 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+26E58 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+26E5C */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+26E60 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+26E64 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+26E68 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+26E6C */	"\x00\x00", "\x00\x00", "\xFC\x53", "\x00\x00",
  /* U+26E70 */	"\x00\x00", "\x00\x00", "\xFD\xB3", "\x00\x00",
  /* U+26E74 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x8F\x6B",
  /* U+26E78 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+26E7C */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+26E80 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+26E84 */	"\x96\xCA", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+26E88 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+26E8C */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+26E90 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+26E94 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+26E98 */	"\x00\x00", "\x8F\x79", "\x00\x00", "\x00\x00",
  /* U+26E9C */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+26EA0 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+26EA4 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+26EA8 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+26EAC */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+26EB0 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+26EB4 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+26EB8 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+26EBC */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+26EC0 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+26EC4 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+26EC8 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+26ECC */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+26ED0 */	"\x9E\x6F", "\xA0\xC5", "\xFC\x78", "\x8E\x42",
  /* U+26ED4 */	"\x8F\x5A", "\x90\xC2", "\x8E\xA5", "\x90\x61",
  /* U+26ED8 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+26EDC */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+26EE0 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+26EE4 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+26EE8 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+26EEC */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+26EF0 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+26EF4 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+26EF8 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+26EFC */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+26F00 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+26F04 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+26F08 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+26F0C */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+26F10 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+26F14 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+26F18 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+26F1C */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+26F20 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+26F24 */	"\x00\x00", "\x00\x00", "\x92\x4F", "\x00\x00",
  /* U+26F28 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+26F2C */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+26F30 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+26F34 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+26F38 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+26F3C */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+26F40 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+26F44 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+26F48 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+26F4C */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+26F50 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+26F54 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+26F58 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+26F5C */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+26F60 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+26F64 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+26F68 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+26F6C */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+26F70 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x93\x73",
  /* U+26F74 */	"\xFD\xB5", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+26F78 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+26F7C */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+26F80 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+26F84 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+26F88 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+26F8C */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+26F90 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+26F94 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+26F98 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+26F9C */	"\x00\x00", "\x00\x00", "\x00\x00", "\xFE\xCC",
  /* U+26FA0 */	"\x00\x00", "\xFB\xBD", "\x00\x00", "\x00\x00",
  /* U+26FA4 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+26FA8 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+26FAC */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+26FB0 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+26FB4 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+26FB8 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+26FBC */	"\x00\x00", "\x00\x00", "\x8C\xD6", "\x00\x00",
  /* U+26FC0 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+26FC4 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+26FC8 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+26FCC */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+26FD0 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+26FD4 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+26FD8 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+26FDC */	"\x00\x00", "\x00\x00", "\x98\x43", "\x96\xC5",
  /* U+26FE0 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+26FE4 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+26FE8 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+26FEC */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+26FF0 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+26FF4 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+26FF8 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+26FFC */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+27000 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+27004 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+27008 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+2700C */	"\x00\x00", "\x00\x00", "\x89\xBC", "\x00\x00",
  /* U+27010 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+27014 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+27018 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+2701C */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+27020 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+27024 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+27028 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+2702C */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+27030 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+27034 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+27038 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+2703C */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+27040 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+27044 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+27048 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x9C\xA3",
  /* U+2704C */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+27050 */	"\x00\x00", "\x00\x00", "\x92\x4B", "\x98\x4A",
  /* U+27054 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+27058 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+2705C */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+27060 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+27064 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+27068 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+2706C */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+27070 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+27074 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+27078 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+2707C */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+27080 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+27084 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+27088 */	"\x8F\xA4", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+2708C */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+27090 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+27094 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+27098 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+2709C */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+270A0 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+270A4 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+270A8 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+270AC */	"\x00\x00", "\xA0\xF1", "\x9E\xFB", "\x9C\xD2",
  /* U+270B0 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+270B4 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+270B8 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+270BC */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+270C0 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+270C4 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+270C8 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+270CC */	"\x00\x00", "\x8F\xA7", "\x00\x00", "\x00\x00",
  /* U+270D0 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+270D4 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+270D8 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+270DC */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+270E0 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+270E4 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+270E8 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+270EC */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+270F0 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+270F4 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+270F8 */	"\xFC\x5C", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+270FC */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+27100 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+27104 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+27108 */	"\x00\x00", "\x98\x45", "\x00\x00", "\x00\x00",
  /* U+2710C */	"\x90\x46", "\x8C\xD1", "\x00\x00", "\x00\x00",
  /* U+27110 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+27114 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+27118 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+2711C */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+27120 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+27124 */	"\x00\x00", "\x00\x00", "\xFE\xFA", "\x95\x60",
  /* U+27128 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+2712C */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+27130 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+27134 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+27138 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+2713C */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+27140 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+27144 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+27148 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+2714C */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+27150 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+27154 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+27158 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+2715C */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+27160 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+27164 */	"\x9F\x48", "\x92\x47", "\x00\x00", "\x00\x00",
  /* U+27168 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+2716C */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+27170 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+27174 */	"\x00\x00", "\x90\xFB", "\x00\x00", "\x00\x00",
  /* U+27178 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+2717C */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+27180 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+27184 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+27188 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+2718C */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+27190 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+27194 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+27198 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+2719C */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+271A0 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+271A4 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+271A8 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+271AC */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+271B0 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+271B4 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+271B8 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+271BC */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+271C0 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+271C4 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+271C8 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+271CC */	"\x00\x00", "\x9C\xA4", "\x00\x00", "\x00\x00",
  /* U+271D0 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+271D4 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+271D8 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+271DC */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+271E0 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+271E4 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+271E8 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+271EC */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+271F0 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+271F4 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+271F8 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+271FC */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+27200 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+27204 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+27208 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+2720C */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+27210 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+27214 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+27218 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x95\x71",
  /* U+2721C */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+27220 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+27224 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+27228 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+2722C */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+27230 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+27234 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+27238 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+2723C */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+27240 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+27244 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+27248 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+2724C */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+27250 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+27254 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+27258 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+2725C */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+27260 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+27264 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+27268 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+2726C */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+27270 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+27274 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+27278 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+2727C */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+27280 */	"\x9C\xA6", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+27284 */	"\x00\x00", "\x9C\xA7", "\x00\x00", "\x00\x00",
  /* U+27288 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x9C\xAA",
  /* U+2728C */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+27290 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+27294 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+27298 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+2729C */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+272A0 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+272A4 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+272A8 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+272AC */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+272B0 */	"\x00\x00", "\x00\x00", "\x9E\xD3", "\x00\x00",
  /* U+272B4 */	"\x00\x00", "\x00\x00", "\x9E\x70", "\x00\x00",
  /* U+272B8 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+272BC */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+272C0 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+272C4 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+272C8 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+272CC */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+272D0 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+272D4 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+272D8 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+272DC */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+272E0 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+272E4 */	"\x00\x00", "\x00\x00", "\x9C\xAC",

  /* U+2739A */				"\x8F\xAE",

  /* U+27422 */				"\x95\x7D", "\x00\x00",
  /* U+27424 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+27428 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+2742C */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+27430 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+27434 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+27438 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+2743C */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+27440 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+27444 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+27448 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+2744C */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+27450 */	"\x9C\xB0", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+27454 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+27458 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+2745C */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+27460 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+27464 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+27468 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+2746C */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+27470 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+27474 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+27478 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+2747C */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+27480 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+27484 */	"\x97\xB6", "\x00\x00", "\xA0\xBD",

  /* U+27574 */	"\x8A\xDF", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+27578 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+2757C */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+27580 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+27584 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+27588 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+2758C */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+27590 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+27594 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+27598 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+2759C */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+275A0 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x9E\xAA",
  /* U+275A4 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+275A8 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+275AC */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+275B0 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+275B4 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+275B8 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+275BC */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+275C0 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+275C4 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+275C8 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+275CC */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+275D0 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+275D4 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+275D8 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+275DC */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+275E0 */	"\x8F\xBD", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+275E4 */	"\x8F\xBF", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+275E8 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+275EC */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+275F0 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+275F4 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+275F8 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+275FC */	"\x00\x00", "\x93\x69", "\x9B\xA7", "\x00\x00",
  /* U+27600 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+27604 */	"\x00\x00", "\x00\x00", "\x00\x00", "\xC8\xA4",
  /* U+27608 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+2760C */	"\xFE\xEA", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+27610 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+27614 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+27618 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+2761C */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+27620 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+27624 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+27628 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+2762C */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+27630 */	"\x00\x00", "\x00\x00", "\x9B\xE1", "\x00\x00",
  /* U+27634 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+27638 */	"\x00\x00", "\x8B\x41", "\x00\x00", "\x00\x00",
  /* U+2763C */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+27640 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+27644 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+27648 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+2764C */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+27650 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+27654 */	"\x00\x00", "\x9D\xB6", "\xA0\xEB", "\x9B\xA3",
  /* U+27658 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+2765C */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+27660 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+27664 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+27668 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+2766C */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+27670 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+27674 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+27678 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+2767C */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+27680 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+27684 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+27688 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+2768C */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+27690 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+27694 */	"\x8B\xA1", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+27698 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+2769C */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+276A0 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+276A4 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+276A8 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+276AC */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+276B0 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+276B4 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+276B8 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+276BC */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+276C0 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+276C4 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+276C8 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+276CC */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+276D0 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+276D4 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+276D8 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+276DC */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+276E0 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+276E4 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+276E8 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+276EC */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+276F0 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+276F4 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+276F8 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+276FC */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+27700 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+27704 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+27708 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+2770C */	"\x00\x00", "\x00\x00", "\x00\x00", "\x8F\xC8",
  /* U+27710 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+27714 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+27718 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+2771C */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+27720 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+27724 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+27728 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+2772C */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+27730 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+27734 */	"\x00\x00", "\x89\x4C", "\x98\x60", "\x00\x00",
  /* U+27738 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+2773C */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+27740 */	"\x00\x00", "\x94\xC7", "\x00\x00", "\x00\x00",
  /* U+27744 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+27748 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+2774C */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+27750 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+27754 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+27758 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+2775C */	"\x00\x00", "\x00\x00", "\x8B\x58", "\x00\x00",
  /* U+27760 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+27764 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+27768 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+2776C */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+27770 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+27774 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+27778 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+2777C */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+27780 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+27784 */	"\x95\xAB", "\x95\xAA", "\x00\x00", "\x00\x00",
  /* U+27788 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+2778C */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+27790 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+27794 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+27798 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+2779C */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+277A0 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+277A4 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+277A8 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+277AC */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+277B0 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+277B4 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+277B8 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+277BC */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+277C0 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+277C4 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+277C8 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+277CC */	"\x9C\xC3",

  /* U+27858 */	"\x9C\xC4", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+2785C */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+27860 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+27864 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+27868 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+2786C */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+27870 */	"\x93\xD6", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+27874 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+27878 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+2787C */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+27880 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+27884 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+27888 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+2788C */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+27890 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+27894 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+27898 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+2789C */	"\x00\x00", "\x9D\xAC", "\x00\x00", "\x00\x00",
  /* U+278A0 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+278A4 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+278A8 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+278AC */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+278B0 */	"\x00\x00", "\x00\x00", "\x8B\xE6", "\x00\x00",
  /* U+278B4 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+278B8 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+278BC */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+278C0 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+278C4 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+278C8 */	"\x8A\x71", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+278CC */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+278D0 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+278D4 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+278D8 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+278DC */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+278E0 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+278E4 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+278E8 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+278EC */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+278F0 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+278F4 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+278F8 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+278FC */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+27900 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+27904 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+27908 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+2790C */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+27910 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+27914 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+27918 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+2791C */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+27920 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+27924 */	"\x8F\xD1", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+27928 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+2792C */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+27930 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+27934 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+27938 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+2793C */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+27940 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+27944 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+27948 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+2794C */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+27950 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+27954 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+27958 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+2795C */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+27960 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+27964 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x99\xD5",
  /* U+27968 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+2796C */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+27970 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+27974 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+27978 */	"\x00\x00", "\x00\x00", "\x90\xF4", "\x00\x00",
  /* U+2797C */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+27980 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+27984 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+27988 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+2798C */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+27990 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+27994 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+27998 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+2799C */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+279A0 */	"\x8A\xA3", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+279A4 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+279A8 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+279AC */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+279B0 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+279B4 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+279B8 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+279BC */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+279C0 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+279C4 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+279C8 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+279CC */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+279D0 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+279D4 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+279D8 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+279DC */	"\x00\x00", "\x9C\xCE", "\x00\x00", "\x00\x00",
  /* U+279E0 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+279E4 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+279E8 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+279EC */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+279F0 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+279F4 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+279F8 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+279FC */	"\x00\x00", "\x9C\xD4", "\x00\x00", "\x00\x00",
  /* U+27A00 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+27A04 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+27A08 */	"\x00\x00", "\x00\x00", "\x9C\xD5", "\x00\x00",
  /* U+27A0C */	"\x00\x00", "\x00\x00", "\xFB\xC8", "\x00\x00",
  /* U+27A10 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+27A14 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+27A18 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+27A1C */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+27A20 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+27A24 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+27A28 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+27A2C */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+27A30 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+27A34 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+27A38 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+27A3C */	"\x00\x00", "\x00\x00", "\x9D\xB3", "\x00\x00",
  /* U+27A40 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+27A44 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+27A48 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+27A4C */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+27A50 */	"\x00\x00", "\x00\x00", "\x00\x00", "\xFC\x70",
  /* U+27A54 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+27A58 */	"\x00\x00", "\x8F\xD7", "\x00\x00", "\x00\x00",
  /* U+27A5C */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+27A60 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+27A64 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+27A68 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+27A6C */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+27A70 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+27A74 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+27A78 */	"\x00\x00", "\x9B\x73", "\x00\x00", "\x00\x00",
  /* U+27A7C */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+27A80 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+27A84 */	"\xFA\x5B", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+27A88 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+27A8C */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+27A90 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+27A94 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+27A98 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+27A9C */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+27AA0 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+27AA4 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+27AA8 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+27AAC */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+27AB0 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+27AB4 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+27AB8 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+27ABC */	"\x00\x00", "\x8F\xD2", "\x90\x64", "\x00\x00",
  /* U+27AC0 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+27AC4 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+27AC8 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+27ACC */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+27AD0 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+27AD4 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+27AD8 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+27ADC */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+27AE0 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+27AE4 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+27AE8 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+27AEC */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+27AF0 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+27AF4 */	"\x98\xB6", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+27AF8 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+27AFC */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+27B00 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+27B04 */	"\x00\x00", "\x00\x00", "\x96\x68", "\x00\x00",
  /* U+27B08 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x9C\xD6",
  /* U+27B0C */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+27B10 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+27B14 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+27B18 */	"\x98\xBD", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+27B1C */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+27B20 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+27B24 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+27B28 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+27B2C */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+27B30 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+27B34 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+27B38 */	"\x8F\xDC", "\xFE\xF6", "\x8F\xD9", "\x00\x00",
  /* U+27B3C */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+27B40 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+27B44 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+27B48 */	"\x95\x41",

  /* U+27BF4 */	"\x97\xF3", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+27BF8 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+27BFC */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+27C00 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+27C04 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+27C08 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+27C0C */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+27C10 */	"\x00\x00", "\x00\x00", "\x9B\xF8",

  /* U+27D2F */					    "\x9E\x6C",
  /* U+27D30 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+27D34 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+27D38 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+27D3C */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+27D40 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+27D44 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+27D48 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+27D4C */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+27D50 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x8F\xF2",
  /* U+27D54 */	"\x8F\xEE", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+27D58 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+27D5C */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+27D60 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+27D64 */	"\x00\x00", "\x00\x00", "\x9C\xD7", "\x00\x00",
  /* U+27D68 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+27D6C */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+27D70 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x9E\x6E",
  /* U+27D74 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+27D78 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+27D7C */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+27D80 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+27D84 */	"\x8A\x40", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+27D88 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+27D8C */	"\x00\x00", "\x00\x00", "\x00\x00", "\x8F\xEF",
  /* U+27D90 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+27D94 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+27D98 */	"\x8F\xF4", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+27D9C */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+27DA0 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+27DA4 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+27DA8 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+27DAC */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+27DB0 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+27DB4 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+27DB8 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+27DBC */	"\x00\x00", "\x8F\xF5", "\x00\x00", "\x00\x00",
  /* U+27DC0 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+27DC4 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+27DC8 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+27DCC */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+27DD0 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+27DD4 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+27DD8 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+27DDC */	"\x95\xC2", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+27DE0 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+27DE4 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+27DE8 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+27DEC */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+27DF0 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+27DF4 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+27DF8 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+27DFC */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+27E00 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+27E04 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+27E08 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+27E0C */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+27E10 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+27E14 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+27E18 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+27E1C */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+27E20 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+27E24 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+27E28 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+27E2C */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+27E30 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+27E34 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+27E38 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+27E3C */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+27E40 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+27E44 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+27E48 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+27E4C */	"\x00\x00", "\x98\x6A", "\x00\x00", "\x97\xCF",

  /* U+27FF9 */		    "\x9E\x7C", "\x00\x00", "\x00\x00",
  /* U+27FFC */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+28000 */	"\x00\x00", "\x00\x00", "\x90\x41", "\x00\x00",
  /* U+28004 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+28008 */	"\x00\x00", "\x9C\xDB", "\x00\x00", "\x00\x00",
  /* U+2800C */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+28010 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+28014 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+28018 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+2801C */	"\x00\x00", "\x00\x00", "\x94\x41", "\x00\x00",
  /* U+28020 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x9C\xE6",
  /* U+28024 */	"\x9D\xB0", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+28028 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+2802C */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+28030 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+28034 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+28038 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+2803C */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+28040 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+28044 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+28048 */	"\x9C\xEA", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+2804C */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+28050 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+28054 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+28058 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+2805C */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+28060 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+28064 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+28068 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+2806C */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+28070 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+28074 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+28078 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+2807C */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+28080 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x9C\xED",
  /* U+28084 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+28088 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+2808C */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+28090 */	"\x9C\xFA", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+28094 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+28098 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+2809C */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+280A0 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+280A4 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+280A8 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+280AC */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+280B0 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+280B4 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+280B8 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+280BC */	"\x00\x00", "\x8B\x62", "\x8A\x4E", "\x00\x00",
  /* U+280C0 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+280C4 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+280C8 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+280CC */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+280D0 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+280D4 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+280D8 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+280DC */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+280E0 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+280E4 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+280E8 */	"\x9C\xCA", "\x8A\x66", "\x00\x00", "\x00\x00",
  /* U+280EC */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+280F0 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+280F4 */	"\x9C\xFB", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+280F8 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+280FC */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+28100 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+28104 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+28108 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+2810C */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+28110 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+28114 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+28118 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+2811C */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+28120 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+28124 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+28128 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+2812C */	"\x00\x00", "\x00\x00", "\x9C\xFC", "\x00\x00",
  /* U+28130 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+28134 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+28138 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+2813C */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+28140 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+28144 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+28148 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+2814C */	"\x00\x00", "\x00\x00", "\x00\x00", "\x9C\xFE",
  /* U+28150 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+28154 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+28158 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+2815C */	"\x00\x00", "\x8A\x53", "\x00\x00", "\x00\x00",
  /* U+28160 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+28164 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+28168 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+2816C */	"\x00\x00", "\x00\x00", "\x00\x00", "\x9C\xE5",
  /* U+28170 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+28174 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+28178 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+2817C */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+28180 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+28184 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+28188 */	"\x00\x00", "\x9D\x40", "\x00\x00", "\x00\x00",
  /* U+2818C */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+28190 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+28194 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+28198 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+2819C */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+281A0 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+281A4 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+281A8 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+281AC */	"\x00\x00", "\x00\x00", "\x00\x00", "\x9D\x41",
  /* U+281B0 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+281B4 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+281B8 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+281BC */	"\x90\x45", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+281C0 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+281C4 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+281C8 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+281CC */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+281D0 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+281D4 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+281D8 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+281DC */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+281E0 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+281E4 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+281E8 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+281EC */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+281F0 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+281F4 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+281F8 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+281FC */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+28200 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+28204 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x8B\x73",
  /* U+28208 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+2820C */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+28210 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+28214 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+28218 */	"\x97\xCA", "\x00\x00", "\x9D\x42", "\x00\x00",
  /* U+2821C */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+28220 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+28224 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+28228 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+2822C */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+28230 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+28234 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+28238 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+2823C */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+28240 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+28244 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+28248 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+2824C */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+28250 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+28254 */	"\x00\x00", "\x00\x00", "\x8A\x61", "\x00\x00",
  /* U+28258 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+2825C */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+28260 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+28264 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+28268 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+2826C */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+28270 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+28274 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+28278 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+2827C */	"\x8B\xAE", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+28280 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+28284 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+28288 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+2828C */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+28290 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+28294 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+28298 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x8A\xD2",
  /* U+2829C */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+282A0 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+282A4 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+282A8 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+282AC */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+282B0 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+282B4 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+282B8 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+282BC */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+282C0 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+282C4 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+282C8 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+282CC */	"\x00\x00", "\x8B\xA2", "\x00\x00", "\x00\x00",
  /* U+282D0 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+282D4 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+282D8 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+282DC */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+282E0 */	"\x00\x00", "\x00\x00", "\x9D\xF2", "\x00\x00",
  /* U+282E4 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+282E8 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+282EC */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+282F0 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+282F4 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+282F8 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+282FC */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+28300 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+28304 */	"\x00\x00", "\x00\x00", "\x9D\x43", "\x00\x00",
  /* U+28308 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+2830C */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+28310 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+28314 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+28318 */	"\x9C\xDF", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+2831C */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+28320 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+28324 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+28328 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+2832C */	"\x00\x00", "\x00\x00", "\x00\x00", "\x9D\x44",
  /* U+28330 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+28334 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+28338 */	"\x00\x00", "\x00\x00", "\x8E\xCA", "\x00\x00",
  /* U+2833C */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+28340 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+28344 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+28348 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+2834C */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+28350 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+28354 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+28358 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+2835C */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+28360 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+28364 */	"\x00\x00", "\x90\x4E", "\x00\x00", "\x00\x00",
  /* U+28368 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+2836C */	"\x00\x00", "\x8E\xB3", "\x00\x00", "\x00\x00",
  /* U+28370 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+28374 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+28378 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+2837C */	"\x00\x00", "\x9F\xF5", "\x00\x00", "\x00\x00",
  /* U+28380 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+28384 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+28388 */	"\x00\x00", "\x00\x00", "\x9D\x45",

  /* U+28412 */				"\x90\x4F", "\x00\x00",
  /* U+28414 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+28418 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+2841C */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+28420 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+28424 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+28428 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+2842C */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+28430 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+28434 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+28438 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+2843C */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+28440 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+28444 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+28448 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+2844C */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+28450 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+28454 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+28458 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+2845C */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+28460 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+28464 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+28468 */	"\x9D\x47", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+2846C */	"\x89\xCA", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+28470 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x9C\xB5",
  /* U+28474 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+28478 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+2847C */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+28480 */	"\x00\x00", "\x00\x00", "\xFB\xFE", "\x00\x00",
  /* U+28484 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+28488 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+2848C */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+28490 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+28494 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+28498 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+2849C */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+284A0 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+284A4 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+284A8 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+284AC */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+284B0 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+284B4 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+284B8 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+284BC */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+284C0 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+284C4 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+284C8 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+284CC */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+284D0 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+284D4 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+284D8 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+284DC */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+284E0 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+284E4 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+284E8 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+284EC */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+284F0 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+284F4 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+284F8 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+284FC */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+28500 */	"\x00\x00", "\x90\x5E", "\x00\x00", "\x00\x00",
  /* U+28504 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+28508 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+2850C */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+28510 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+28514 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+28518 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+2851C */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+28520 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+28524 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+28528 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+2852C */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+28530 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+28534 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+28538 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+2853C */	"\x90\x63", "\x90\x57", "\x00\x00", "\x00\x00",
  /* U+28540 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+28544 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+28548 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+2854C */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+28550 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+28554 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+28558 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+2855C */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+28560 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+28564 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+28568 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+2856C */	"\x90\x66", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+28570 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+28574 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+28578 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+2857C */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+28580 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+28584 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+28588 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+2858C */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+28590 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+28594 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+28598 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+2859C */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+285A0 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+285A4 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+285A8 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+285AC */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+285B0 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+285B4 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+285B8 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+285BC */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+285C0 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+285C4 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+285C8 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+285CC */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+285D0 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+285D4 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+285D8 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+285DC */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+285E0 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+285E4 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+285E8 */	"\x9B\xC0", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+285EC */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+285F0 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+285F4 */	"\xFC\xE5", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+285F8 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+285FC */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+28600 */	"\x91\x62", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+28604 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+28608 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x90\x67",
  /* U+2860C */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+28610 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+28614 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+28618 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+2861C */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+28620 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+28624 */	"\x00\x00", "\x8F\xA1", "\x00\x00", "\x00\x00",
  /* U+28628 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+2862C */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+28630 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+28634 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+28638 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x8F\xA2",
  /* U+2863C */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+28640 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+28644 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+28648 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+2864C */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+28650 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+28654 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+28658 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+2865C */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+28660 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+28664 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+28668 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+2866C */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+28670 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+28674 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+28678 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+2867C */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+28680 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+28684 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+28688 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+2868C */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+28690 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+28694 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+28698 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+2869C */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+286A0 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+286A4 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+286A8 */	"\x00\x00", "\x00\x00", "\x9D\x48", "\xFA\xD3",
  /* U+286AC */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+286B0 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+286B4 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+286B8 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+286BC */	"\x90\x5D", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+286C0 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+286C4 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+286C8 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+286CC */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+286D0 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+286D4 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+286D8 */	"\x90\xB9", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+286DC */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+286E0 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+286E4 */	"\x00\x00", "\x00\x00", "\x90\x6B", "\x00\x00",
  /* U+286E8 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+286EC */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+286F0 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+286F4 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+286F8 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+286FC */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+28700 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+28704 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+28708 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+2870C */	"\x00\x00", "\x00\x00", "\x00\x00", "\x8C\x5C",
  /* U+28710 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x90\x69",

  /* U+28804 */	"\xFE\x57", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+28808 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+2880C */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+28810 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+28814 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+28818 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+2881C */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+28820 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+28824 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+28828 */	"\x00\x00", "\x00\x00", "\x00\x00", "\xFE\x55",

  /* U+28933 */					    "\x90\x73",
  /* U+28934 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+28938 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+2893C */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+28940 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+28944 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+28948 */	"\x9B\xEF", "\x9C\xF0", "\x00\x00", "\x00\x00",
  /* U+2894C */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+28950 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+28954 */	"\x00\x00", "\x00\x00", "\x9D\x4B", "\x00\x00",
  /* U+28958 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+2895C */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+28960 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+28964 */	"\xFE\xD9", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+28968 */	"\xFE\xDA", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+2896C */	"\x91\xE0", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+28970 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+28974 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+28978 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+2897C */	"\x00\x00", "\x00\x00", "\x91\xD8", "\x00\x00",
  /* U+28980 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+28984 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+28988 */	"\x00\x00", "\x96\x46", "\x00\x00", "\x00\x00",
  /* U+2898C */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+28990 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+28994 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+28998 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+2899C */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+289A0 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+289A4 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+289A8 */	"\x93\x60", "\x00\x00", "\xFA\x53", "\x9C\xD3",
  /* U+289AC */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+289B0 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+289B4 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+289B8 */	"\x9D\x4E", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+289BC */	"\xFB\x40", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+289C0 */	"\x8D\xE2", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+289C4 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+289C8 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+289CC */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+289D0 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+289D4 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+289D8 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+289DC */	"\x94\x42", "\x00\x00", "\x90\x56", "\x00\x00",
  /* U+289E0 */	"\x00\x00", "\x98\x65", "\x00\x00", "\x8C\x6C",
  /* U+289E4 */	"\xFA\x4A", "\x00\x00", "\x00\x00", "\x9D\x50",
  /* U+289E8 */	"\x9D\x52", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+289EC */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+289F0 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+289F4 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+289F8 */	"\x00\x00", "\x95\xAF", "\x97\x5A", "\x93\x49",
  /* U+289FC */	"\x97\x47", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+28A00 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+28A04 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+28A08 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+28A0C */	"\x00\x00", "\x00\x00", "\x00\x00", "\xA0\xF4",
  /* U+28A10 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+28A14 */	"\x00\x00", "\x00\x00", "\x97\x78", "\x00\x00",
  /* U+28A18 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+28A1C */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+28A20 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+28A24 */	"\x00\x00", "\x8F\xCF", "\x00\x00", "\x00\x00",
  /* U+28A28 */	"\x00\x00", "\xFC\x60", "\x00\x00", "\x00\x00",
  /* U+28A2C */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+28A30 */	"\x00\x00", "\x00\x00", "\x8C\x4E", "\x00\x00",
  /* U+28A34 */	"\x00\x00", "\x00\x00", "\xFC\x56", "\x00\x00",
  /* U+28A38 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+28A3C */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+28A40 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+28A44 */	"\x91\xDC", "\x96\x61", "\x92\xEC", "\x93\x5D",
  /* U+28A48 */	"\x8E\xDE", "\x96\xFE", "\xFD\x4F", "\x95\xDE",
  /* U+28A4C */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+28A50 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+28A54 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+28A58 */	"\x00\x00", "\x98\xB0", "\xA0\x40", "\x00\x00",
  /* U+28A5C */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+28A60 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+28A64 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+28A68 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+28A6C */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+28A70 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+28A74 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+28A78 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+28A7C */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+28A80 */	"\x00\x00", "\x97\xBD", "\x97\x7D", "\x97\xF5",
  /* U+28A84 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+28A88 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+28A8C */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+28A90 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+28A94 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+28A98 */	"\x00\x00", "\x00\x00", "\x9B\xAC", "\xFA\xDA",
  /* U+28A9C */	"\x92\xC2", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+28AA0 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+28AA4 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+28AA8 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+28AAC */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+28AB0 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+28AB4 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+28AB8 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+28ABC */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+28AC0 */	"\x97\xB1", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+28AC4 */	"\x00\x00", "\x00\x00", "\x90\x7B", "\x00\x00",
  /* U+28AC8 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x93\xFE",
  /* U+28ACC */	"\x94\x7B", "\x00\x00", "\x97\x77", "\x00\x00",
  /* U+28AD0 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+28AD4 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+28AD8 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+28ADC */	"\x00\x00", "\x00\x00", "\xFA\xBE", "\xFD\x43",
  /* U+28AE0 */	"\x90\xC6", "\x90\xA4", "\x90\xA8", "\x94\xA9",
  /* U+28AE4 */	"\x00\x00", "\x90\xA9", "\x00\x00", "\x00\x00",
  /* U+28AE8 */	"\x00\x00", "\x00\x00", "\x8C\x65", "\x00\x00",
  /* U+28AEC */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+28AF0 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+28AF4 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+28AF8 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+28AFC */	"\x95\xE0", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+28B00 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+28B04 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+28B08 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+28B0C */	"\x90\x7D", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+28B10 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x92\x65",
  /* U+28B14 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+28B18 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+28B1C */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+28B20 */	"\x00\x00", "\xFD\xBA", "\x93\xC4", "\x00\x00",
  /* U+28B24 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+28B28 */	"\x00\x00", "\x00\x00", "\x00\x00", "\xFE\xED",
  /* U+28B2C */	"\x9D\xAB", "\xA0\xE3", "\x00\x00", "\x96\x48",
  /* U+28B30 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+28B34 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+28B38 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+28B3C */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+28B40 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+28B44 */	"\x00\x00", "\x00\x00", "\x9D\x53", "\x00\x00",
  /* U+28B48 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+28B4C */	"\x8A\xA9", "\x00\x00", "\x9B\xC5", "\x00\x00",
  /* U+28B50 */	"\x96\x5D", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+28B54 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+28B58 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+28B5C */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+28B60 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x97\x5F",
  /* U+28B64 */	"\x96\x5F", "\x96\x6E", "\xFB\x5D", "\x00\x00",
  /* U+28B68 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+28B6C */	"\x9D\xB1", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+28B70 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+28B74 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+28B78 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+28B7C */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+28B80 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+28B84 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+28B88 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+28B8C */	"\x00\x00", "\x00\x00", "\x00\x00", "\xFE\xA3",
  /* U+28B90 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+28B94 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+28B98 */	"\x00\x00", "\x9D\xB2", "\x00\x00", "\x00\x00",
  /* U+28B9C */	"\x95\xAE", "\xFC\xA3", "\x00\x00", "\x00\x00",
  /* U+28BA0 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+28BA4 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+28BA8 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+28BAC */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+28BB0 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+28BB4 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+28BB8 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+28BBC */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+28BC0 */	"\x00\x00", "\x00\x00", "\xA0\xA2", "\x00\x00",
  /* U+28BC4 */	"\x00\x00", "\x96\x55", "\x00\x00", "\x00\x00",
  /* U+28BC8 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+28BCC */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+28BD0 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+28BD4 */	"\x9D\x54", "\x00\x00", "\x00\x00", "\x93\x41",
  /* U+28BD8 */	"\x00\x00", "\x95\xAD", "\x91\xD5", "\x00\x00",
  /* U+28BDC */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+28BE0 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+28BE4 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x97\x7A",
  /* U+28BE8 */	"\xFD\xFC", "\x8E\x47", "\x93\xFD", "\x90\xA5",
  /* U+28BEC */	"\x90\xAC", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+28BF0 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+28BF4 */	"\x00\x00", "\x95\xAC", "\x00\x00", "\x00\x00",
  /* U+28BF8 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+28BFC */	"\x00\x00", "\x00\x00", "\x00\x00", "\x90\xAE",
  /* U+28C00 */	"\x00\x00", "\x00\x00", "\x00\x00", "\xFE\xA5",
  /* U+28C04 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+28C08 */	"\x00\x00", "\x9D\x56", "\x00\x00", "\x00\x00",
  /* U+28C0C */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+28C10 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+28C14 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+28C18 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+28C1C */	"\x97\xE3", "\x95\xE2", "\x00\x00", "\x00\x00",
  /* U+28C20 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x94\x66",
  /* U+28C24 */	"\x00\x00", "\x00\x00", "\x96\x47", "\x00\x00",
  /* U+28C28 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x91\xB8",
  /* U+28C2C */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+28C30 */	"\x9C\xEC", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+28C34 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+28C38 */	"\x00\x00", "\x90\xAD", "\x00\x00", "\x95\xE3",

  /* U+28CCA */				"\x8B\x4F", "\x00\x00",
  /* U+28CCC */	"\x00\x00", "\x8A\xE3", "\x00\x00", "\x00\x00",
  /* U+28CD0 */	"\x00\x00", "\x00\x00", "\x8B\x4D", "\x00\x00",
  /* U+28CD4 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+28CD8 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+28CDC */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+28CE0 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+28CE4 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+28CE8 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+28CEC */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+28CF0 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+28CF4 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+28CF8 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+28CFC */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+28D00 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+28D04 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+28D08 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+28D0C */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+28D10 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+28D14 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+28D18 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+28D1C */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+28D20 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+28D24 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+28D28 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+28D2C */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+28D30 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+28D34 */	"\x95\xEA", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+28D38 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+28D3C */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+28D40 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+28D44 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+28D48 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+28D4C */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+28D50 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+28D54 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+28D58 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+28D5C */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+28D60 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+28D64 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+28D68 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+28D6C */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+28D70 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+28D74 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+28D78 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+28D7C */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+28D80 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+28D84 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+28D88 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+28D8C */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+28D90 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+28D94 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+28D98 */	"\x00\x00", "\x8B\x4E", "\x00\x00", "\x00\x00",
  /* U+28D9C */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+28DA0 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+28DA4 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+28DA8 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+28DAC */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+28DB0 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+28DB4 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+28DB8 */	"\x00\x00", "\x8C\xC1", "\x00\x00", "\x00\x00",
  /* U+28DBC */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+28DC0 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+28DC4 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+28DC8 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+28DCC */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+28DD0 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+28DD4 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+28DD8 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+28DDC */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+28DE0 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+28DE4 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+28DE8 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+28DEC */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+28DF0 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+28DF4 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+28DF8 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+28DFC */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+28E00 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+28E04 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+28E08 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+28E0C */	"\x00\x00", "\x00\x00", "\x00\x00", "\x8B\xED",
  /* U+28E10 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+28E14 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+28E18 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+28E1C */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+28E20 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+28E24 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+28E28 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+28E2C */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+28E30 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+28E34 */	"\x00\x00", "\x00\x00", "\x91\xD9", "\x00\x00",
  /* U+28E38 */	"\x00\x00", "\xA0\xA4", "\x00\x00", "\x00\x00",
  /* U+28E3C */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+28E40 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+28E44 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+28E48 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+28E4C */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+28E50 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+28E54 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+28E58 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+28E5C */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+28E60 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+28E64 */	"\x00\x00", "\x95\xF5", "\x95\xF4", "\x00\x00",
  /* U+28E68 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+28E6C */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+28E70 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+28E74 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+28E78 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+28E7C */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+28E80 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+28E84 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+28E88 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+28E8C */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+28E90 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+28E94 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x9F\xB3",
  /* U+28E98 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+28E9C */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+28EA0 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+28EA4 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+28EA8 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+28EAC */	"\xFE\xAF", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+28EB0 */	"\x00\x00", "\x00\x00", "\xFE\x72", "\x92\x7A",
  /* U+28EB4 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+28EB8 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+28EBC */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+28EC0 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+28EC4 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+28EC8 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+28ECC */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+28ED0 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+28ED4 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+28ED8 */	"\x00\x00", "\xFE\xAC", "\x00\x00", "\x00\x00",
  /* U+28EDC */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+28EE0 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+28EE4 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x95\xF3",

  /* U+28FC5 */		    "\x9D\x58",

  /* U+29088 */	"\x93\x72", "\x00\x00", "\x00\x00", "\x91\xC5",
  /* U+2908C */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+29090 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x96\x42",
  /* U+29094 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+29098 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+2909C */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+290A0 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+290A4 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+290A8 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+290AC */	"\x00\x00", "\x00\x00", "\x00\x00", "\x90\xCD",
  /* U+290B0 */	"\x95\xFE", "\x91\x59", "\x00\x00", "\x00\x00",
  /* U+290B4 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+290B8 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+290BC */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+290C0 */	"\x9C\x65", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+290C4 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+290C8 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+290CC */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+290D0 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+290D4 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+290D8 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+290DC */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+290E0 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+290E4 */	"\x97\xCC", "\x90\xCE", "\x00\x00", "\x00\x00",
  /* U+290E8 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+290EC */	"\x9D\x59", "\xFC\xF5", "\x00\x00", "\x00\x00",
  /* U+290F0 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+290F4 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+290F8 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+290FC */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+29100 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+29104 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+29108 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+2910C */	"\x00\x00", "\xFE\xFD", "\x00\x00", "\x00\x00",
  /* U+29110 */	"\x9D\x5B", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+29114 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+29118 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+2911C */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+29120 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+29124 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+29128 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+2912C */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+29130 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+29134 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+29138 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+2913C */	"\x9D\x5C", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+29140 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+29144 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+29148 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+2914C */	"\x00\x00", "\x93\x7E", "\x00\x00", "\x00\x00",
  /* U+29150 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+29154 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+29158 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x98\xAC",
  /* U+2915C */	"\x00\x00", "\x00\x00", "\x9D\x5E", "\x00\x00",
  /* U+29160 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+29164 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+29168 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+2916C */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+29170 */	"\xFD\xD0", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+29174 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+29178 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+2917C */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+29180 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+29184 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+29188 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+2918C */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+29190 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+29194 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+29198 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+2919C */	"\xFD\x60", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+291A0 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+291A4 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+291A8 */	"\x9C\xCF", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+291AC */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+291B0 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+291B4 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+291B8 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+291BC */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+291C0 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+291C4 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+291C8 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+291CC */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+291D0 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+291D4 */	"\x00\x00", "\x90\xDD", "\x00\x00", "\x00\x00",
  /* U+291D8 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+291DC */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+291E0 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+291E4 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+291E8 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x90\xE0",

  /* U+2941D */		    "\x90\xF3", "\x00\x00", "\x00\x00",
  /* U+29420 */	"\x98\xB1", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+29424 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+29428 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+2942C */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+29430 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x90\xF0",
  /* U+29434 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+29438 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+2943C */	"\x00\x00", "\x00\x00", "\x00\x00", "\x93\xBD",
  /* U+29440 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+29444 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+29448 */	"\x95\xB7",

  /* U+294D0 */	"\x9F\x46", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+294D4 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+294D8 */	"\x00\x00", "\x8E\x4B", "\x96\x58", "\x00\x00",
  /* U+294DC */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+294E0 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+294E4 */	"\x00\x00", "\x8A\x4C", "\x00\x00", "\x9D\x63",

  /* U+2959E */				"\x9E\xCF", "\x00\x00",
  /* U+295A0 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+295A4 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+295A8 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+295AC */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+295B0 */	"\x9D\x65", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+295B4 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+295B8 */	"\x9D\x66", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+295BC */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+295C0 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+295C4 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+295C8 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+295CC */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+295D0 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+295D4 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x96\x5A",
  /* U+295D8 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+295DC */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+295E0 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+295E4 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+295E8 */	"\x00\x00", "\x9D\x64", "\x00\x00", "\x00\x00",
  /* U+295EC */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+295F0 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+295F4 */	"\x8A\x6C",

  /* U+29720 */	"\x8A\xD9", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+29724 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+29728 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+2972C */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+29730 */	"\x00\x00", "\x00\x00", "\x9D\x67",

  /* U+297D4 */	"\x8A\x70", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+297D8 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+297DC */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+297E0 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+297E4 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+297E8 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+297EC */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+297F0 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+297F4 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+297F8 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+297FC */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+29800 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+29804 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+29808 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+2980C */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+29810 */	"\x8B\xF3", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+29814 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+29818 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+2981C */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+29820 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+29824 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+29828 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+2982C */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+29830 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+29834 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+29838 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+2983C */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+29840 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+29844 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+29848 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+2984C */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+29850 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+29854 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x91\x50",
  /* U+29858 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+2985C */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+29860 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+29864 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+29868 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+2986C */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+29870 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+29874 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+29878 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+2987C */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+29880 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+29884 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+29888 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+2988C */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+29890 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+29894 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+29898 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+2989C */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+298A0 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+298A4 */	"\x9C\xC1", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+298A8 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+298AC */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+298B0 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+298B4 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+298B8 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+298BC */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+298C0 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+298C4 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+298C8 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+298CC */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+298D0 */	"\x00\x00", "\x9D\x68", "\x00\x00", "\x00\x00",
  /* U+298D4 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+298D8 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+298DC */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+298E0 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+298E4 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+298E8 */	"\x00\x00", "\x00\x00", "\x93\xA7", "\x00\x00",
  /* U+298EC */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+298F0 */	"\x00\x00", "\x96\x74", "\x00\x00", "\x00\x00",
  /* U+298F4 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+298F8 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+298FC */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+29900 */	"\x00\x00", "\x00\x00", "\x00\x00", "\xA0\xEF",
  /* U+29904 */	"\x00\x00", "\x91\x51", "\x00\x00", "\x00\x00",
  /* U+29908 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+2990C */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+29910 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+29914 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+29918 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+2991C */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+29920 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+29924 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+29928 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+2992C */	"\x00\x00", "\x00\x00", "\x00\x00", "\x96\xC1",
  /* U+29930 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+29934 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+29938 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+2993C */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+29940 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+29944 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x8C\x64",
  /* U+29948 */	"\x96\x76", "\x9D\x69", "\x00\x00", "\x00\x00",
  /* U+2994C */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+29950 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+29954 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+29958 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+2995C */	"\x00\x00", "\xFC\xA4", "\x00\x00", "\x00\x00",
  /* U+29960 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+29964 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+29968 */	"\x00\x00", "\x00\x00", "\x9D\x6A", "\x00\x00",
  /* U+2996C */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+29970 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+29974 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+29978 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+2997C */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+29980 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+29984 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+29988 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+2998C */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+29990 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+29994 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+29998 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+2999C */	"\x00\x00", "\x92\x4E", "\x00\x00", "\x00\x00",
  /* U+299A0 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+299A4 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+299A8 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+299AC */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+299B0 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+299B4 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+299B8 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+299BC */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+299C0 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x9D\x6B",
  /* U+299C4 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+299C8 */	"\x00\x00", "\x9B\xC1", "\x00\x00", "\x00\x00",
  /* U+299CC */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+299D0 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+299D4 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+299D8 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+299DC */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+299E0 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+299E4 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+299E8 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+299EC */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+299F0 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+299F4 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+299F8 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+299FC */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+29A00 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+29A04 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+29A08 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+29A0C */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+29A10 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+29A14 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+29A18 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+29A1C */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+29A20 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+29A24 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+29A28 */	"\x9D\x6C", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+29A2C */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+29A30 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+29A34 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+29A38 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+29A3C */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+29A40 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+29A44 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+29A48 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+29A4C */	"\x00\x00", "\x8A\x65",

  /* U+29B05 */		    "\x91\x5D", "\x00\x00", "\x00\x00",
  /* U+29B08 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+29B0C */	"\x00\x00", "\x00\x00", "\x9D\x6D",

  /* U+29BD5 */		    "\x91\x5A",

  /* U+29C73 */					    "\x8C\x42",
  /* U+29C74 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+29C78 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+29C7C */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+29C80 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+29C84 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+29C88 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+29C8C */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+29C90 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+29C94 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+29C98 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+29C9C */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+29CA0 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+29CA4 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+29CA8 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+29CAC */	"\x00\x00", "\x9C\xC0",

  /* U+29D3E */				"\x91\x6A", "\x00\x00",
  /* U+29D40 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+29D44 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+29D48 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+29D4C */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+29D50 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+29D54 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+29D58 */	"\x00\x00", "\x00\x00", "\x9D\x6E", "\x00\x00",
  /* U+29D5C */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+29D60 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+29D64 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+29D68 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+29D6C */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+29D70 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+29D74 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+29D78 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+29D7C */	"\x9E\xA6", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+29D80 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+29D84 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+29D88 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+29D8C */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+29D90 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+29D94 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+29D98 */	"\x9D\xCD", "\x00\x00", "\x00\x00", "\x9D\x6F",
  /* U+29D9C */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+29DA0 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+29DA4 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+29DA8 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+29DAC */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+29DB0 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+29DB4 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+29DB8 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+29DBC */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+29DC0 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+29DC4 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+29DC8 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+29DCC */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+29DD0 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+29DD4 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+29DD8 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+29DDC */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+29DE0 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+29DE4 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+29DE8 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+29DEC */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+29DF0 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+29DF4 */	"\x00\x00", "\x00\x00", "\x89\xBB", "\x00\x00",
  /* U+29DF8 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+29DFC */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+29E00 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+29E04 */	"\x00\x00", "\x00\x00", "\x9E\xF9", "\x00\x00",
  /* U+29E08 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+29E0C */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+29E10 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+29E14 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+29E18 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+29E1C */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+29E20 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+29E24 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+29E28 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+29E2C */	"\x00\x00", "\x96\xB4", "\x00\x00", "\x00\x00",
  /* U+29E30 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+29E34 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+29E38 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+29E3C */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+29E40 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+29E44 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+29E48 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+29E4C */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+29E50 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+29E54 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+29E58 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+29E5C */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+29E60 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+29E64 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+29E68 */	"\x91\x72", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+29E6C */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+29E70 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+29E74 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+29E78 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+29E7C */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+29E80 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+29E84 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+29E88 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+29E8C */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+29E90 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+29E94 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+29E98 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+29E9C */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+29EA0 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+29EA4 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+29EA8 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+29EAC */	"\x9E\xC8", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+29EB0 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+29EB4 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+29EB8 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+29EBC */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+29EC0 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x8B\x55",
  /* U+29EC4 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+29EC8 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+29ECC */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+29ED0 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+29ED4 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+29ED8 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+29EDC */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+29EE0 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+29EE4 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+29EE8 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+29EEC */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+29EF0 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+29EF4 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+29EF8 */	"\x9D\x71", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+29EFC */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+29F00 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+29F04 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+29F08 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+29F0C */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+29F10 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+29F14 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+29F18 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+29F1C */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+29F20 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x9D\x72",
  /* U+29F24 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+29F28 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+29F2C */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+29F30 */	"\x9E\xCC",

  /* U+29FB7 */					    "\x91\x74",
  /* U+29FB8 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+29FBC */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+29FC0 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+29FC4 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+29FC8 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+29FCC */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+29FD0 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+29FD4 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+29FD8 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+29FDC */	"\x00\x00", "\x00\x00", "\x9E\xD0", "\x00\x00",
  /* U+29FE0 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+29FE4 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+29FE8 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+29FEC */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+29FF0 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+29FF4 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+29FF8 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+29FFC */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+2A000 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+2A004 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+2A008 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+2A00C */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+2A010 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+2A014 */	"\x90\x5C", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+2A018 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+2A01C */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+2A020 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+2A024 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+2A028 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+2A02C */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+2A030 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+2A034 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+2A038 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+2A03C */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+2A040 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+2A044 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+2A048 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+2A04C */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+2A050 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+2A054 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+2A058 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+2A05C */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+2A060 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+2A064 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+2A068 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+2A06C */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+2A070 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+2A074 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+2A078 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+2A07C */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+2A080 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+2A084 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x8E\xD2",
  /* U+2A088 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+2A08C */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+2A090 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+2A094 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+2A098 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+2A09C */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+2A0A0 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+2A0A4 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+2A0A8 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+2A0AC */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+2A0B0 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+2A0B4 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+2A0B8 */	"\x00\x00", "\x91\xA8", "\x00\x00", "\x00\x00",
  /* U+2A0BC */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+2A0C0 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+2A0C4 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+2A0C8 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+2A0CC */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+2A0D0 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+2A0D4 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+2A0D8 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+2A0DC */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+2A0E0 */	"\x00\x00", "\x91\x77", "\x00\x00", "\x00\x00",
  /* U+2A0E4 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+2A0E8 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+2A0EC */	"\x00\x00", "\x96\xBF", "\x00\x00", "\x00\x00",
  /* U+2A0F0 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x96\xC0",
  /* U+2A0F4 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+2A0F8 */	"\x8F\xB1", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+2A0FC */	"\x00\x00", "\x00\x00", "\x96\xB7", "\x00\x00",
  /* U+2A100 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+2A104 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x8C\x55",
  /* U+2A108 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+2A10C */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+2A110 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+2A114 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+2A118 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+2A11C */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+2A120 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x91\x78",
  /* U+2A124 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+2A128 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+2A12C */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+2A130 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x89\xBE",
  /* U+2A134 */	"\x91\x7C", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+2A138 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+2A13C */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+2A140 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+2A144 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+2A148 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+2A14C */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+2A150 */	"\xFB\x77", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+2A154 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+2A158 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+2A15C */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+2A160 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+2A164 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+2A168 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+2A16C */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+2A170 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+2A174 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+2A178 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+2A17C */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+2A180 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+2A184 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+2A188 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+2A18C */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+2A190 */	"\x00\x00", "\x00\x00", "\x91\x75", "\x91\xA3",
  /* U+2A194 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+2A198 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+2A19C */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+2A1A0 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+2A1A4 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+2A1A8 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x91\x76",
  /* U+2A1AC */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+2A1B0 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+2A1B4 */	"\x96\xBE", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+2A1B8 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+2A1BC */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+2A1C0 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+2A1C4 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+2A1C8 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+2A1CC */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+2A1D0 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+2A1D4 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+2A1D8 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+2A1DC */	"\x00\x00", "\x00\x00", "\x00\x00", "\x91\x79",
  /* U+2A1E0 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+2A1E4 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+2A1E8 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+2A1EC */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+2A1F0 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+2A1F4 */	"\x00\x00", "\x96\xB6", "\x00\x00", "\x00\x00",
  /* U+2A1F8 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+2A1FC */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+2A200 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+2A204 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+2A208 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+2A20C */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+2A210 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+2A214 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+2A218 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+2A21C */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+2A220 */	"\x91\xA4", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+2A224 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+2A228 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+2A22C */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+2A230 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x91\xA6",
  /* U+2A234 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+2A238 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+2A23C */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+2A240 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+2A244 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+2A248 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+2A24C */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+2A250 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+2A254 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+2A258 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+2A25C */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+2A260 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+2A264 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+2A268 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+2A26C */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+2A270 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+2A274 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+2A278 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+2A27C */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+2A280 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+2A284 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+2A288 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+2A28C */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+2A290 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x9D\x75",
  /* U+2A294 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+2A298 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+2A29C */	"\x00\x00", "\x00\x00", "\x00\x00", "\x90\x52",
  /* U+2A2A0 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+2A2A4 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+2A2A8 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+2A2AC */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+2A2B0 */	"\x00\x00", "\x00\x00", "\xA0\x45", "\x00\x00",
  /* U+2A2B4 */	"\x91\xA9", "\x00\x00", "\x98\xAA", "\x00\x00",
  /* U+2A2B8 */	"\x00\x00", "\x00\x00", "\x8C\x5F", "\x00\x00",
  /* U+2A2BC */	"\x00\x00", "\x8B\xAA", "\x00\x00", "\x00\x00",
  /* U+2A2C0 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+2A2C4 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+2A2C8 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+2A2CC */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+2A2D0 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+2A2D4 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+2A2D8 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+2A2DC */	"\x00\x00", "\x00\x00", "\x00\x00", "\x9C\xDD",
  /* U+2A2E0 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+2A2E4 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+2A2E8 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+2A2EC */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+2A2F0 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+2A2F4 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+2A2F8 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+2A2FC */	"\x00\x00", "\x00\x00", "\x00\x00", "\x9D\x77",

  /* U+2A3A9 */		    "\x89\x40",

  /* U+2A434 */	"\x9E\xEC", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+2A438 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+2A43C */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+2A440 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+2A444 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+2A448 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+2A44C */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+2A450 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+2A454 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+2A458 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x93\xAA",

  /* U+2A5C6 */				"\x94\x78", "\x00\x00",
  /* U+2A5C8 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x9D\x7A",
  /* U+2A5CC */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+2A5D0 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+2A5D4 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+2A5D8 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+2A5DC */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+2A5E0 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+2A5E4 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+2A5E8 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+2A5EC */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+2A5F0 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+2A5F4 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+2A5F8 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+2A5FC */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+2A600 */	"\x00\x00", "\x8A\xC9", "\x00\x00", "\x00\x00",
  /* U+2A604 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+2A608 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+2A60C */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+2A610 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+2A614 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+2A618 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+2A61C */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+2A620 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+2A624 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+2A628 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+2A62C */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+2A630 */	"\x00\x00", "\x00\x00", "\x8B\x4B", "\x00\x00",
  /* U+2A634 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+2A638 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+2A63C */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+2A640 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+2A644 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+2A648 */	"\x00\x00", "\x00\x00", "\x9F\xEC", "\x00\x00",
  /* U+2A64C */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+2A650 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+2A654 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+2A658 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x8A\xE2",
  /* U+2A65C */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+2A660 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+2A664 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+2A668 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+2A66C */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+2A670 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+2A674 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+2A678 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+2A67C */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+2A680 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+2A684 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+2A688 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+2A68C */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+2A690 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+2A694 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+2A698 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+2A69C */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+2A6A0 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+2A6A4 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+2A6A8 */	"\x00\x00", "\x9E\x75",

  /* U+2F825 */		    "\x98\x74", "\x00\x00", "\x00\x00",
  /* U+2F828 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+2F82C */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+2F830 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+2F834 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+2F838 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x9A\xC8",
  /* U+2F83C */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+2F840 */	"\xA0\x47", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+2F844 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+2F848 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+2F84C */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+2F850 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+2F854 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+2F858 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+2F85C */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+2F860 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+2F864 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+2F868 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+2F86C */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+2F870 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+2F874 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+2F878 */	"\x8B\xC3", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+2F87C */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+2F880 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+2F884 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+2F888 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+2F88C */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+2F890 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+2F894 */	"\xFC\x48", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+2F898 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+2F89C */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+2F8A0 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+2F8A4 */	"\x00\x00", "\x00\x00", "\xFC\x77", "\x00\x00",
  /* U+2F8A8 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+2F8AC */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+2F8B0 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+2F8B4 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+2F8B8 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+2F8BC */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+2F8C0 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+2F8C4 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+2F8C8 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+2F8CC */	"\x00\x00", "\x9C\x52",

  /* U+2F994 */	"\x8E\xFD", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+2F998 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+2F99C */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+2F9A0 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+2F9A4 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+2F9A8 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+2F9AC */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+2F9B0 */	"\x00\x00", "\x00\x00", "\x8F\xA8", "\x00\x00",
  /* U+2F9B4 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+2F9B8 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+2F9BC */	"\x95\x7A", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+2F9C0 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+2F9C4 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+2F9C8 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+2F9CC */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+2F9D0 */	"\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* U+2F9D4 */	"\x8F\xF0",
};

static struct
{
    /* Note: We are going to split this table so that we can use
       uint16_t for "from" and "to" again.  Anthony Fok, 2002-03-21 */
    uint32_t from;
    uint32_t to;
    uint32_t offset;
} from_ucs4_idx[] =
{
    {  0x00A1,  0x0305,     0 },
    {  0x0391,  0x0451,   613 },
    {  0x1EBE,  0x1EC1,   806 },
    {  0x2013,  0x203E,   810 },
    {  0x2103,  0x2307,   854 },
    {  0x2460,  0x247D,  1371 },
    {  0x2500,  0x2642,  1401 },
    {  0x273D,  0x273D,  1724 },
    {  0x2E80,  0x2F33,  1725 },
    {  0x3000,  0x319F,  1905 },
    {  0x3220,  0x32B0,  2321 },
    {  0x338E,  0x9FA5,  2466 },
    {  0xE000,  0xF848, 30138 },
    {  0xF8F8,  0xFA2D, 36355 },
    {  0xFE30,  0xFE6B, 36665 },
    {  0xFF01,  0xFFED, 36725 },
    { 0x20021, 0x20118, 36962 },
    { 0x201A9, 0x20676, 37210 },
    { 0x2070E, 0x20779, 38440 },
    { 0x2082C, 0x20B0D, 38548 },
    { 0x20B8F, 0x21A4B, 39286 },
    { 0x21B44, 0x21CAC, 43059 },
    { 0x21D46, 0x2227C, 43420 },
    { 0x22321, 0x22325, 44755 },
    { 0x223BD, 0x22993, 44760 },
    { 0x22A66, 0x22EEF, 46255 },
    { 0x22F74, 0x2331F, 47417 },
    { 0x233B4, 0x2383A, 48357 },
    { 0x239C2, 0x239C2, 49516 },
    { 0x23AA7, 0x23B5A, 49517 },
    { 0x23C63, 0x24AE9, 49697 },
    { 0x24B6E, 0x24B6E, 53416 },
    { 0x24BF5, 0x24C09, 53417 },
    { 0x24C9E, 0x24D13, 53438 },
    { 0x24DB8, 0x2509D, 53556 },
    { 0x2512B, 0x25313, 54298 },
    { 0x25419, 0x2546E, 54787 },
    { 0x25531, 0x259CC, 54873 },
    { 0x25A95, 0x25AE9, 56053 },
    { 0x25B74, 0x25D43, 56138 },
    { 0x25E0E, 0x25F4B, 56602 },
    { 0x25FE1, 0x26626, 56920 },
    { 0x266AF, 0x26A52, 58526 },
    { 0x26B05, 0x272E6, 59458 },
    { 0x2739A, 0x2739A, 61476 },
    { 0x27422, 0x27486, 61477 },
    { 0x27574, 0x277CC, 61578 },
    { 0x27858, 0x27B48, 62179 },
    { 0x27BF4, 0x27C12, 62932 },
    { 0x27D2F, 0x27E4F, 62963 },
    { 0x27FF9, 0x2838A, 63252 },
    { 0x28412, 0x28713, 64166 },
    { 0x28804, 0x2882B, 64936 },
    { 0x28933, 0x28C3B, 64976 },
    { 0x28CCA, 0x28EE7, 65753 },
    { 0x28FC5, 0x28FC5, 66295 },
    { 0x29088, 0x291EB, 66296 },
    { 0x2941D, 0x29448, 66652 },
    { 0x294D0, 0x294E7, 66696 },
    { 0x2959E, 0x295F4, 66720 },
    { 0x29720, 0x29732, 66807 },
    { 0x297D4, 0x29A4D, 66826 },
    { 0x29B05, 0x29B0E, 67460 },
    { 0x29BD5, 0x29BD5, 67470 },
    { 0x29C73, 0x29CAD, 67471 },
    { 0x29D3E, 0x29F30, 67530 },
    { 0x29FB7, 0x2A2FF, 68029 },
    { 0x2A3A9, 0x2A3A9, 68870 },
    { 0x2A434, 0x2A45B, 68871 },
    { 0x2A5C6, 0x2A6A9, 68911 },
    { 0x2F825, 0x2F8CD, 69139 },
    { 0x2F994, 0x2F9D4, 69308 },
};

/* Definitions used in the body of the `gconv' function.  */
#define CHARSET_NAME		"BIG5HKSCS//"
#define FROM_LOOP		from_big5
#define TO_LOOP			to_big5
#define DEFINE_INIT		1
#define DEFINE_FINI		1
#define MIN_NEEDED_FROM		1
#define MAX_NEEDED_FROM		2
#define MIN_NEEDED_TO		4


/* First define the conversion function from Big5-HKSCS to UCS4.  */
#define MIN_NEEDED_INPUT	MIN_NEEDED_FROM
#define MAX_NEEDED_INPUT	MAX_NEEDED_FROM
#define MIN_NEEDED_OUTPUT	MIN_NEEDED_TO
#define LOOPFCT			FROM_LOOP
#define BODY \
  {									      \
    uint32_t ch = *inptr;						      \
									      \
    if (ch >= 0x81 && ch <= 0xfe)					      \
      {									      \
	/* Two-byte character.  First test whether the next character	      \
	   is also available.  */					      \
	uint32_t ch2;							      \
	int idx;							      \
									      \
	if (__builtin_expect (inptr + 1 >= inend, 0))			      \
	  {								      \
	    /* The second character is not available.  */		      \
	    result = __GCONV_INCOMPLETE_INPUT;				      \
	    break;							      \
	  }								      \
									      \
	ch2 = inptr[1];							      \
	/* See whether the second byte is in the correct range.  */	      \
	if ((ch2 >= 0x40 && ch2 <= 0x7e) || (ch2 >= 0xa1 && ch2 <= 0xfe))     \
	  {								      \
	    if (ch >= 0x88)						      \
	      {								      \
		/* Look up the table */					      \
		idx = (ch - 0x88) * 157 + ch2 - (ch2 <= 0x7e ? 0x40 : 0x62);  \
		if ((ch = big5_hkscs_to_ucs[idx]) == 0)			      \
		  {							      \
		    /* This is illegal.  */				      \
		    STANDARD_FROM_LOOP_ERR_HANDLER (1);			      \
		  }							      \
	      }								      \
	    else							      \
	      {								      \
		/* 0x81..0x87 in UDA3, currently maps linearly to PUA */      \
		ch = (ch - 0x81) * 157 + ch2 - (ch2 <= 0x7e ? 0x40 : 0x62)    \
		      + 0xeeb8;						      \
	      }								      \
	  }								      \
	else								      \
	  {								      \
	    /* This is illegal.  */					      \
	    STANDARD_FROM_LOOP_ERR_HANDLER (1);				      \
	  }								      \
									      \
	inptr += 2;							      \
      }									      \
    else if (__builtin_expect (ch == 0xff, 0))				      \
      {									      \
	STANDARD_FROM_LOOP_ERR_HANDLER (1);				      \
      }									      \
    else  /* 0x00 to 0x80 */						      \
      ++inptr;								      \
									      \
    put32 (outptr, ch);							      \
    outptr += 4;							      \
  }
#define LOOP_NEED_FLAGS
#define ONEBYTE_BODY \
  {									      \
    if (c <= 0x80)							      \
      return c;								      \
    else								      \
      return WEOF;							      \
  }
#include <iconv/loop.c>


/* Next, define the other direction.  */
#define MIN_NEEDED_INPUT	MIN_NEEDED_TO
#define MIN_NEEDED_OUTPUT	MIN_NEEDED_FROM
#define MAX_NEEDED_OUTPUT	MAX_NEEDED_FROM
#define LOOPFCT			TO_LOOP
#define BODY \
  {									      \
    uint32_t ch = get32 (inptr);					      \
    const unsigned char *cp = "";						      \
    unsigned char b5ch[2] = "\0\0";					      \
    int i;								      \
    									      \
    for (i = 0;								      \
	 i < (int) (sizeof (from_ucs4_idx) / sizeof (from_ucs4_idx[0]));      \
	 ++i)								      \
      {									      \
	if (ch < from_ucs4_idx[i].from)					      \
	  break;							      \
	if (from_ucs4_idx[i].to >= ch)					      \
	  {								      \
	    cp = ucs4_to_big5_hkscs[from_ucs4_idx[i].offset		      \
			  + ch - from_ucs4_idx[i].from];		      \
	    break;							      \
	  }								      \
      }									      \
									      \
    if (ch <= 0x80)							      \
      {									      \
	b5ch[0] = ch;							      \
	cp = b5ch;							      \
      }									      \
									      \
    if (cp[0] == '\0' && ch != 0)					      \
      {									      \
	UNICODE_TAG_HANDLER (ch, 4);					      \
									      \
	/* Illegal character.  */					      \
	STANDARD_TO_LOOP_ERR_HANDLER (4);				      \
      }									      \
    else								      \
      {									      \
	/* See whether there is enough room for the second byte we write.  */ \
	if (__builtin_expect (cp[1] != '\0', 1)				      \
	    && __builtin_expect (outptr + 1 >= outend, 0))		      \
	  {								      \
	    /* We have not enough room.  */				      \
	    result = __GCONV_FULL_OUTPUT;				      \
	    break;							      \
	  }								      \
									      \
	*outptr++ = cp[0];						      \
	if (__builtin_expect (cp[1] != '\0', 1))			      \
	  *outptr++ = cp[1];						      \
      }									      \
									      \
    inptr += 4;								      \
  }
#define LOOP_NEED_FLAGS
#include <iconv/loop.c>


/* Now define the toplevel functions.  */
#include <iconv/skeleton.c>
